
with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Text_IO;                       use Ada.Text_IO;
with Ada.Unchecked_Deallocation;
with Allegro;                           use Allegro;
with Allegro.Audio;                     use Allegro.Audio;
with Allegro.Audio.Codecs;              use Allegro.Audio.Codecs;
with Allegro.System;                    use Allegro.System;
with Allegro.Time;                      use Allegro.Time;
with GNAT.OS_Lib;
with Interfaces.C;                      use Interfaces.C;

--
-- Milan Mimica
-- Audio example that plays multiple files at the same time
-- Originlly derived from the ex_acodec example.
--
package body Ex_Acodec_Multi is

    procedure Ada_Main is
        type Allegro_Sample_Array is array (Integer range <>) of A_Allegro_Sample;
        type A_Allegro_Sample_Array is access all Allegro_Sample_Array;

        procedure Free is new Ada.Unchecked_Deallocation( Allegro_Sample_Array,
                                                          A_Allegro_Sample_Array );

        type Allegro_Sample_Instance_Array is array (Integer range <>) of A_Allegro_Sample_Instance;
        type A_Allegro_Sample_Instance_Array is access all Allegro_Sample_Instance_Array;

        procedure Free is new Ada.Unchecked_Deallocation( Allegro_Sample_Instance_Array,
                                                          A_Allegro_Sample_Instance_Array );

        sample_data    : A_Allegro_Sample_Array;
        sample         : A_Allegro_Sample_Instance_Array;
        mixer          : A_Allegro_Mixer;
        voice          : A_Allegro_Voice;
        longest_sample : Float;
        sample_time    : Float;
    begin
        if Argument_Count < 1 then
            Put_Line( "Usage: ex_acodec_multi {audio_files}" );
            GNAT.OS_Lib.OS_Exit( 1 );
        end if;

        if not Al_Initialize then
            Put_Line( "Could not init Allegro." );
            GNAT.OS_Lib.OS_Exit( 1 );
        end if;

        if not Al_Init_Acodec_Addon then
            Put_Line( "Could not init audio codecs." );
            GNAT.OS_Lib.OS_Exit( 1 );
        end if;

        if not Al_Install_Audio then
            Put_Line( "Could not init sound!" );
            GNAT.OS_Lib.OS_Exit( 1 );
        end if;

        sample := new Allegro_Sample_Instance_Array(1..Argument_Count);
        sample_data := new Allegro_Sample_Array(1..Argument_Count);

        -- a voice is used for playback
        voice := Al_Create_Voice( 44100, ALLEGRO_AUDIO_DEPTH_INT16, ALLEGRO_CHANNEL_CONF_2 );
        if voice = null then
            Put_Line( "Could not create ALLEGRO_VOICE from sample" );
            GNAT.OS_Lib.OS_Exit( 1 );
        end if;

        mixer := Al_Create_Mixer( 44100, ALLEGRO_AUDIO_DEPTH_FLOAT32, ALLEGRO_CHANNEL_CONF_2 );
        if mixer = null then
            Put_Line( "al_create_mixer failed." );
            GNAT.OS_Lib.OS_Exit( 1 );
        end if;

        if not Al_Attach_Mixer_To_Voice( mixer, voice ) then
            Put_Line( "al_attach_mixer_to_voice failed." );
            GNAT.OS_Lib.OS_Exit( 1 );
        end if;

        for i in 1..Argument_Count loop
            sample(i) := null;

            -- loads the entire sound file from disk into sample data
            sample_data(i) := al_load_sample( Argument( i ) );
            if sample_data(i) = null then
                Put_Line( "Could not load sample from '" & Argument( i ) & "'!" );
            end if;

            if sample_data(i) /= null then
                sample(i) := Al_Create_Sample_Instance( sample_data(i) );
                if sample(i) = null then
                    Put_Line( "Could not create sample!" );
                    Al_Destroy_Sample( sample_data(i) );
                end if;

                if sample_data(i) /= null and then
                   not Al_Attach_Sample_Instance_To_Mixer( sample(i), mixer )
                then
                    Put_Line( "al_attach_sample_instance_to_mixer failed." );
                end if;
            end if;
        end loop;

        longest_sample := 0.0;

        for i in 1..Argument_Count loop
            if sample(i) /= null then
                -- play each sample once
                if not Al_Play_Sample_Instance( sample(i) ) then
                    Put_Line( "Al_Play_Sample_Instance failed" );
                end if;

                sample_time := Al_Get_Sample_Instance_Time( sample(i) );
                Put_Line( "Playing '" & Argument( i ) & "' (" & sample_time'Img & " seconds)" );

                if sample_time > longest_sample then
                    longest_sample := sample_time;
                end if;
            end if;
        end loop;

        Al_Rest( double(longest_sample*1.1) );

        for i in 1..Argument_Count loop
            -- free the memory allocated when creating the sample + voice
            if sample(i) /= null then
                if not Al_Stop_Sample_Instance( sample(i) ) then
                    Put_Line( "Failed to stop sample" & i'Img );
                end if;
                Al_Destroy_Sample_Instance( sample(i) );
                Al_Destroy_Sample( sample_data(i) );
            end if;
        end loop;
        Al_Destroy_Mixer( mixer );
        Al_Destroy_Voice( voice );

        Free( sample );
       Free( sample_data );

        Al_Uninstall_Audio;
    end Ada_Main;

begin
    Set_Ada_Main( Ada_Main'Access );
end Ex_Acodec_Multi;