
with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Ada.Text_IO;                       use Ada.Text_IO;
with Allegro;                           use Allegro;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Allegro.Bitmaps.Drawing;           use Allegro.Bitmaps.Drawing;
with Allegro.Bitmaps.Image_IO;          use Allegro.Bitmaps.Image_IO;
with Allegro.Color;                     use Allegro.Color;
with Allegro.Displays;                  use Allegro.Displays;
with Allegro.Drawing;                   use Allegro.Drawing;
with Allegro.Events;                    use Allegro.Events;
with Allegro.Image_Addon;               use Allegro.Image_Addon;
with Allegro.Keyboard;                  use Allegro.Keyboard;
with Allegro.Mouse;                     use Allegro.Mouse;
with Allegro.System;                    use Allegro.System;
with Allegro.Time;                      use Allegro.Time;
with Allegro.Timers;                    use Allegro.Timers;
with Common;                            use Common;
with Interfaces;                        use Interfaces;
with Interfaces.C;                      use Interfaces.C;

package body Ex_Bitmap is

    procedure Ada_Main is
        filename  : Unbounded_String;
        display   : A_Allegro_Display;
        membitmap : A_Allegro_Bitmap;
        bitmap    : A_Allegro_Bitmap;
        timer     : A_Allegro_Timer;
        queue     : A_Allegro_Event_Queue;
        redraw    : Boolean := True;
        zoom      : Float := 1.0;
        t0, t1    : double;
        event     : aliased Allegro_Event;
    begin
        if Argument_Count > 0 then
            filename := To_Unbounded_String( Argument(1) );
        else
            filename := To_Unbounded_String( "data/mysha.pcx" );
        end if;

        if not Al_Initialize then
            Abort_Example( "Could not init Allegro." );
        end if;

        if Argument_Count > 1 then
            Al_Set_New_Display_Adapter( Integer'Value( Argument(2) ) );
        end if;

        if not Al_Install_Mouse then
            Abort_Example( "Could not install mouse." );
        end if;
        if not Al_Install_Keyboard then
            Abort_Example( "Could not install keyboard." );
        end if;

        if not Al_Init_Image_Addon then
            Abort_Example( "Could not init image addon." );
        end if;

        display := Al_Create_Display( 640, 480 );
        if display = null then
            Abort_Example( "Error creating display" );
        end if;

        Al_Set_Window_Title( display, To_String( filename ) );

        -- We load the bitmap into a memory bitmap, because creating a
        -- display bitmap could fail if the bitmap is too big to fit into a
        -- single texture.
        -- FIXME: Or should A5 automatically created multiple display bitmaps?
        Al_Set_New_Bitmap_Flags( ALLEGRO_MEMORY_BITMAP );
        t0 := Al_Get_Time;
        membitmap := Al_Load_Bitmap( To_String( filename ) );
        t1 := Al_Get_Time;
        if membitmap = null then
            Abort_Example( To_String( filename ) & " not found or failed to load" );
        end if;
        Al_Set_New_Bitmap_Flags( ALLEGRO_VIDEO_BITMAP );

        Put_Line( "Loading took" & Integer'Image( Integer((t1 - t0) * 1000.0) ) & " milliseconds" );

        bitmap := Al_Clone_Bitmap( membitmap );
        if bitmap = null then
            bitmap := membitmap;
        end if;

        timer := Al_Create_Timer( 1.0 / 30.0 );
        queue := Al_Create_Event_Queue;
        Al_Register_Event_Source( queue, Al_Get_Keyboard_Event_Source );
        Al_Register_Event_Source( queue, Al_Get_Display_Event_Source( display ) );
        Al_Register_Event_Source( queue, Al_Get_Timer_Event_Source( timer ) );
        Al_Start_Timer( timer );

        loop
            Al_Wait_For_Event( queue, event );

            if event.any.typ = ALLEGRO_EVENT_DISPLAY_ORIENTATION then
                if event.display.orientation = ALLEGRO_DISPLAY_ORIENTATION_0_DEGREES then
                    Put_Line( "0 degrees" );
                elsif event.display.orientation = ALLEGRO_DISPLAY_ORIENTATION_90_DEGREES then
                    Put_Line( "90 degrees" );
                elsif event.display.orientation = ALLEGRO_DISPLAY_ORIENTATION_180_DEGREES then
                    Put_Line( "180 degrees" );
                elsif event.display.orientation = ALLEGRO_DISPLAY_ORIENTATION_270_DEGREES then
                    Put_Line( "270 degrees" );
                elsif event.display.orientation = ALLEGRO_DISPLAY_ORIENTATION_FACE_UP then
                    Put_Line( "Face up" );
                elsif event.display.orientation = ALLEGRO_DISPLAY_ORIENTATION_FACE_DOWN then
                    Put_Line( "Face down" );
                end if;

            elsif event.any.typ = ALLEGRO_EVENT_DISPLAY_CLOSE then
                exit;

            elsif event.any.typ = ALLEGRO_EVENT_KEY_CHAR then
                if event.keyboard.keycode = ALLEGRO_KEY_ESCAPE then
                    exit;
                elsif event.keyboard.unichar = Character'Pos( '1' ) then
                    zoom := 1.0;
                elsif event.keyboard.unichar = Character'Pos( '=' ) then
                    zoom := zoom * 1.1;
                elsif event.keyboard.unichar = Character'Pos( '-' ) then
                    zoom := zoom / 1.1;
                elsif event.keyboard.unichar = Character'Pos( 'f' ) then
                    zoom := Float(Al_Get_Display_Width( display )) /
                            Float(Al_Get_Bitmap_Width( bitmap ));
                end if;

            elsif event.any.typ = ALLEGRO_EVENT_TIMER then
                redraw := True;
            end if;

            if redraw and then Al_Is_Event_Queue_Empty( queue ) then
                redraw := False;
                Al_Clear_To_Color( Al_Map_RGB_f( 0.0, 0.0, 0.0 ) );
                if zoom = 1.0 then
                    Al_Draw_Bitmap( bitmap, 0.0, 0.0, 0 );
                else
                    Al_Draw_Scaled_Rotated_Bitmap( bitmap, 0.0, 0.0, 0.0, 0.0, zoom, zoom, 0.0, 0 );
                end if;
                Al_Flip_Display;
            end if;
        end loop;

        Al_Stop_Timer( timer );

        Al_Unregister_Event_Source( queue, Al_Get_Timer_Event_Source( timer ) );
        Al_Unregister_Event_Source( queue, Al_Get_Display_Event_Source( display ) );
        Al_Unregister_Event_Source( queue, Al_Get_Keyboard_Event_Source );
        Al_Destroy_Event_Queue( queue );

        Al_Destroy_Timer( timer );

        Al_Destroy_Bitmap( bitmap );

        Al_Destroy_Display( display );
    end Ada_Main;

begin
    Set_Ada_Main( Ada_Main'Access );
end Ex_Bitmap;
