
with Ada.Numerics.Elementary_Functions; use Ada.Numerics.Elementary_Functions;
with Allegro;                           use Allegro;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Allegro.Bitmaps.Drawing;           use Allegro.Bitmaps.Drawing;
with Allegro.Blending;                  use Allegro.Blending;
with Allegro.Color;                     use Allegro.Color;
with Allegro.Displays;                  use Allegro.Displays;
with Allegro.Drawing;                   use Allegro.Drawing;
with Allegro.Drawing.Primitives;        use Allegro.Drawing.Primitives;
with Allegro.Events;                    use Allegro.Events;
with Allegro.Fonts;                     use Allegro.Fonts;
with Allegro.Image_Addon;               use Allegro.Image_Addon;
with Allegro.Keyboard;                  use Allegro.Keyboard;
with Allegro.System;                    use Allegro.System;
with Allegro.Time;                      use Allegro.Time;
with Common;                            use Common;
with Interfaces;                        use Interfaces;
with Interfaces.C;                      use Interfaces.C;

-- An example comparing FPS when drawing to a bitmap with the
-- ALLEGRO_FORCE_LOCKING flag and without. Mainly meant as a test how much
-- speedup direct drawing can give over the slow locking.
package body Ex_Bitmap_Target is

    procedure Ada_Main is
        W  : constant Float := 300.0;  -- size of target bitmap.
        H  : constant Float := 300.0;
        RW : constant Float := 50.0;
        RH : constant Float := 50.0;   -- Size of rectangle we draw to it.
        display   : A_Allegro_Display;
        target    : A_Allegro_Bitmap;  -- The target bitmap.
        x, y,
        dx, dy    : Float;             -- Position and velocity of moving rectangle.
        last_time : double;            -- For controling speed.
        quit      : Boolean := False;  -- Flag to record Esc key or X button.
        myfont    : A_Allegro_Font;        -- Our font.
        queue     : A_Allegro_Event_Queue; -- Our events queue.

        ------------------------------------------------------------------------

        -- Print some text with a shadow.
        procedure Print( x, y : Float; message : String ) is
        begin
            Al_Set_Blender( ALLEGRO_ADD, ALLEGRO_ONE, ALLEGRO_INVERSE_ALPHA );
            Al_Draw_Text( myfont, Al_Map_RGB( 0, 0, 0 ), x + 2.0, y + 2.0, 0, message );
            Al_Draw_Text( myfont, Al_Map_RGB( 255, 255, 255 ), x, y, 0, message );
        end Print;

        ------------------------------------------------------------------------

        -- Draw our example scene.
        procedure Draw is
            t      : constant double := Al_Get_Time;
            dt     : double := 0.0;
            xs, ys : Float;
            angle  : Float;
        begin
            if last_time > 0.0 then
                dt := t - last_time;
            end if;
            last_time := t;

            Al_Set_Target_Bitmap( target );
            Al_Set_Blender( ALLEGRO_ADD, ALLEGRO_ALPHA, ALLEGRO_INVERSE_ALPHA );

            Al_Draw_Filled_Rectangle( x, y, x + RW, y + RH, Al_Map_RGBA_f(1.0, 0.0, 0.0, 1.0 ) );
            Al_Draw_Filled_Rectangle( 0.0, 0.0, W, H, Al_Map_RGBA_f( 1.0, 1.0, 0.0, 0.1 ) );

            x := x + dx * Float(dt);
            if x < 0.0 then
                x := 0.0;
                dx := -dx;
            elsif x + RW > W then
                x := W - RW;
                dx := -dx;
            end if;

            y := y + dy * Float(dt);
            if y < 0.0 then
                y := 0.0;
                dy := -dy;
            elsif y + RH > H then
                y := H - RH;
                dy := -dy;
            end if;

            Al_Set_Target_Backbuffer( display );
            Al_Set_Blender( ALLEGRO_ADD, ALLEGRO_ONE, ALLEGRO_ZERO );
            Al_Clear_To_Color( Al_Map_RGBA_f( 0.0, 0.0, 1.0, 1.0 ) );
            xs := 1.0 + 0.2 * Sin( Float(t * ALLEGRO_PI * 2.0) );
            ys := 1.0 + 0.2 * Sin( Float(t * ALLEGRO_PI * 2.0) );
            angle := Float(t) * ALLEGRO_PI * 2.0 / 3.0;
            Al_Draw_Scaled_Rotated_Bitmap( target, W / 2.0, H / 2.0, 320.0, 240.0, xs, ys, angle, 0 );
        end Draw;

        ------------------------------------------------------------------------

        -- Run the FPS test.
        procedure Run is
            event   : aliased Allegro_Event;
            frames  : Integer := 0;
            start   : double;
            restart : Boolean := False;
        begin
            target := Al_Create_Bitmap( Integer(W), Integer(H) );
            Al_Set_Target_Bitmap( target );
            Al_Clear_To_Color( Al_Map_RGBA_f( 1.0, 1.0, 0.0, 1.0 ) );

            Al_Set_Target_Backbuffer( display );

            dx := 81.0;
            dy := 63.0;

            start := Al_Get_Time;
            loop
                -- Check for ESC key or close button event and quit in either case.
                if not Al_Is_Event_Queue_Empty( queue ) then
                    while Al_Get_Next_Event( queue, event'Access ) loop
                        case event.any.typ is
                            when ALLEGRO_EVENT_DISPLAY_CLOSE =>
                                quit := True;
                                exit;
                            when ALLEGRO_EVENT_KEY_DOWN =>
                                if event.keyboard.keycode = ALLEGRO_KEY_ESCAPE then
                                    quit := True;
                                    exit;
                                elsif event.keyboard.keycode = ALLEGRO_KEY_SPACE then
                                    restart := True;
                                    exit;
                                end if;
                            when others =>
                                null;
                        end case;
                    end loop;
                end if;

                if restart or quit then
                    exit;
                end if;

                Draw;
                Print( 0.0, 0.0, "FPS: " & Integer'Image( Integer(double(frames) / (Al_Get_Time - start)) ) );
                if (Al_Get_New_Bitmap_Flags and ALLEGRO_FORCE_LOCKING) /= 0 then
                    Print( 0.0, Float(Al_Get_Font_Line_Height( myfont )), "using forced bitmap locking" );
                else
                    Print( 0.0, Float(Al_Get_Font_Line_Height( myfont )), "drawing directly to bitmap" );
                end if;
                Print( 0.0, Float(Al_Get_Font_Line_Height( myfont ) * 2), "Press SPACE to toggle drawing method." );
                Al_Flip_Display;
                frames := frames + 1;
            end loop;

           Al_Destroy_Bitmap( target );
        end Run;

    begin
        if not Al_Initialize then
            Abort_Example( "Could not init Allegro." );
        end if;

        if not Al_Init_Primitives_Addon then
            Abort_Example( "Could not init Allegro primitives addon." );
        end if;
        if not Al_Install_Keyboard then
            Abort_Example( "Could not install keyboard." );
        end if;
        if not Al_Init_Image_Addon then
            Abort_Example( "Could not init Allegro IIO addon." );
        end if;
        Al_Init_Font_Addon;

        display := Al_Create_Display( 640, 480 );
        if display = null then
            Abort_Example( "Error creating display" );
        end if;

        queue := Al_Create_Event_Queue;
        Al_Register_Event_Source( queue, Al_Get_Keyboard_Event_Source);
        Al_Register_Event_Source( queue, Al_Get_Display_Event_Source( display ) );

        myfont := Al_Load_Font( "data/font.tga", 0, 0 );
        if myfont = null then
            Abort_Example( "font.tga not found" );
        end if;

        while not quit loop
            if (Al_Get_New_Bitmap_Flags and ALLEGRO_FORCE_LOCKING) /= 0 then
                Al_Set_New_Bitmap_Flags( ALLEGRO_VIDEO_BITMAP );
            else
                Al_Set_New_Bitmap_Flags( ALLEGRO_FORCE_LOCKING );
            end if;
            Run;
        end loop;

        Al_Destroy_Font( myfont );
        Al_Destroy_Event_Queue( queue );
        Al_Destroy_Display( display );
    end Ada_Main;

begin
    Set_Ada_Main( Ada_Main'Access );
end Ex_Bitmap_Target;
