--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Directories;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Streams;                       use Ada.Streams;
with Ada.Streams.Stream_IO;             use Ada.Streams.Stream_IO;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Text_IO;                       use Ada.Text_IO;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;
with Zip_Streams;                       use Zip_Streams;
with Zip.Compress;                      use Zip.Compress;
with Zip.Create;

procedure Zipfile is

    ----------------------------------------------------------------------------

    function Add_Extension( s : String; ext : String ) return String is
    begin
        if s'Length < ext'Length + 1 or else
           To_Upper( s(s'Last-ext'Length..s'Last) ) = '.' & To_Upper( ext )
        then
            return s;
        else
            return s & '.' & ext;
        end if;
    end Add_Extension;

    ----------------------------------------------------------------------------

    function Get_Archive_Path( path : String ) return String is
        colon : Integer := Index( path, ":" );
    begin
        if colon >= path'First then
            return path(colon+1..path'Last);
        end if;
        return path;
    end Get_Archive_Path;

    ----------------------------------------------------------------------------

    function Get_Disk_Path( path : String ) return String is
        colon : Integer := Index( path, ":" );
    begin
        if colon >= path'First then
            return path(path'First..colon-1);
        end if;
        return path;
    end Get_Disk_Path;

    ----------------------------------------------------------------------------

    method : Compression_Method := Shrink;
    curArg : Integer := 1;
begin
    if Argument_Count < 2 then
        Put_Line( "Usage: zipfile [new_zipfile.zip] [file1] [file2] [...]" );
        OS_Exit( 1 );
    end if;

    declare
        myStream      : aliased ZipFile_Stream;
        zipFileStream : constant Zipstream_Class := myStream'Unchecked_Access;
        info          : Zip.Create.Zip_Create_Info;
        zipName       : STRING := Add_Extension( Argument( curArg ), "zip" );
        success       : BOOLEAN;
    begin
        -- delete the existing zip file
        if Is_Regular_File( zipName ) then
            Delete_File( zipName, success );
            if not success then
                Put_Line( "zipfile: Error deleting file: " & zipName );
                OS_Exit( 3 );
            end if;
        end if;

        Zip.Create.Create( info, zipFileStream, zipName, method );

        for i in curArg+1..Argument_Count loop
            declare
                srcPath     : constant String := Get_Disk_Path( Argument( i ) );
                destPath    : constant String := Get_Archive_Path( Argument( i ) );
                inStream    : aliased ZipFile_Stream;
                streamFile  : constant Zipstream_Class := inStream'Unchecked_Access;
            begin
                -- verify the file exists
                if not Is_Regular_File( srcPath ) then
                    Put_Line( "zipfile: File not found: " & srcPath );
                    OS_Exit( 4 );
                end if;

                -- add the file to the zip
                SetName( streamFile, srcPath );
                SetTime( streamFile, Ada.Directories.Modification_Time( srcPath ) );
                Zip_Streams.Open( inStream, In_File );
                SetName( streamFile, destPath );
                Zip.Create.Add_Stream( info, streamFile );
                Close( inStream );
            end;
        end loop;

        Zip.Create.Finish( info );
    end;

exception
    when e : others =>
        Put_Line( "zipfile: Exception: " & Exception_Name( e ) & " -- " & Exception_Message( e ) );
        OS_Exit( 2 );
end Zipfile;
