--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Text_IO;
with Allegro;                           use Allegro;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Allegro.Color;                     use Allegro.Color;
with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Engine_Debugging;                  use Engine_Debugging;
with Entities;
with Events.Manager;
with GNAT.OS_Lib;
with Themes.Palette;
with Tiles.Libraries.Manager;

package body Applications.Gui.Games is

    overriding
    procedure Finalize( this : access Game_Application ) is
    begin
        Delete( this.game );
        Delete( this.view );
        Delete( this.pman );

        Entities.Finalize;
        Tiles.Libraries.Manager.Finalize;

        Gui_Application(this.all).Finalize;
    end Finalize;

    ----------------------------------------------------------------------------

    overriding
    function Initialize( this : access Game_Application ) return Boolean is
    begin
        if not Gui_Application(this.all).Initialize then
            return False;
        end if;

        Tiles.Libraries.Manager.Initialize;

        -- Graphics have been initialized; display a loading image
        declare
            buf : A_Allegro_Bitmap;
        begin
            Al_Set_Target_Backbuffer( this.display );    -- set the display active
            buf := Al_Create_Bitmap( Al_Get_Display_Width( this.display ),
                                     Al_Get_Display_Height( this.display ) );
            Set_Target_Bitmap( buf );
            Clear_To_Color( Al_Map_RGB( 0, 0, 0 ) );
            Push_State;
            A_Game_Application(this).Draw_Loading_Screen;
            Pop_State;
            Set_Target_Bitmap( Al_Get_Backbuffer( this.display ) );
            Al_Set_Target_Backbuffer( this.display );
            Draw_Bitmap_Stretched( buf,
                                   0, 0,
                                   Get_Target_Width, Get_Target_Height,
                                   proportional => True );
            Al_Destroy_Bitmap( buf );
            Al_Flip_Display;
            Al_Set_Target_Backbuffer( null );            -- release the display
        end;

        -- Initialize systems after Allegro
        Themes.Palette.Initialize;
        begin
            -- may raise an exception (resources missing)
            Entities.Initialize;
        exception
            when e : others =>
                Dbg( "Error: Entities.Initialize: " &
                     Exception_Name( e ) & " - " & Exception_Message( e ),
                     D_INIT, ERROR );
                this.Show_Error( "Application Error: " & Exception_Message( e ) );
                raise INIT_EXCEPTION with Exception_Message( e );
        end;

        A_Game_Application(this).Load_Resources;   -- preload game resources

        return True;
    exception
        when e : others =>
            Dbg( "Game_Application.Initialize Exception: ...", D_INIT, ERROR );
            Dbg( e, D_INIT, ERROR );

            this.Finalize;
            return False;
    end Initialize;

    ----------------------------------------------------------------------------

    -- Runs the Application and returns a numeric code that will be returned to
    -- the operating system (0 on success). If an exception is raised, the
    -- caller should assume a return value of non-zero.
    overriding
    function Run( this : access Game_Application ) return Integer is
    begin
        -- may raise exception
        this.view := Create_Game_View( this.display );
        this.game := Create_Game;

        this.game.Initialize;
        this.view.Initialize;

        Events.Manager.Run;

        this.view.Finalize;
        this.game.Finalize;

        return this.errorCode;
    exception
        when e : others =>
            this.Show_Error( "Application Error: " & Exception_Message( e ) );
            this.errorCode := 1;
            return this.errorCode;
    end Run;

    ----------------------------------------------------------------------------

    procedure Stop( this      : not null access Game_Application'Class;
                    errorCode : Integer := NO_ERROR ) is
    begin
        this.errorCode := errorCode;
        Events.Manager.Stop;
    end Stop;

    --==========================================================================

    procedure Terminate_Application( error : String ) is
        app : A_Application;
    begin
        app := Get_Application;
        if app /= null then
            app.Show_Error( "Application Error: " & error );
            A_Game_Application(app).Stop;
        else
            -- the application object hasn't been created yet or isn't being
            -- used.
            Ada.Text_IO.Put_Line( "Application Error: " & error );
            GNAT.OS_Lib.OS_Exit( 1 );
        end if;
    end Terminate_Application;

end Applications.Gui.Games;
