--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Text_IO;
pragma Warnings( Off, Ada.Text_IO );
with Allegro;                           use Allegro;
with Allegro.Color;                     use Allegro.Color;
with Allegro.Displays.Monitors;         use Allegro.Displays.Monitors;
with Allegro.Drawing.Primitives;        use Allegro.Drawing.Primitives;
with Allegro.Keyboard;                  use Allegro.Keyboard;
with Allegro.Mouse;                     use Allegro.Mouse;
with Allegro.Mouse.Cursors;             use Allegro.Mouse.Cursors;
with Allegro.Native_Dialogs;            use Allegro.Native_Dialogs;
with Allegro.State;                     use Allegro.State;
with Allegro.System;                    use Allegro.System;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Fonts.Manager;
with Preferences;                       use Preferences;
with Support;                           use Support;
with xBR;

pragma Elaborate_All( Preferences );

package body Applications.Gui is

    MIN_WIN_WIDTH  : constant := 64;
    MIN_WIN_HEIGHT : constant := 64;
    MIN_FS_WIDTH   : constant := 640;
    MIN_FS_HEIGHT  : constant := 400;

    COLOR_DEPTH    : constant := 32;   -- bits per pixel (BPP)

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    logWindow : A_Allegro_Textlog := null;

    procedure Debug_Output( str : String );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    not overriding
    procedure Construct( this            : access Gui_Application;
                         company         : String;
                         name            : String;
                         userDir         : String;
                         defaultWidth,
                         defaultHeight   : Natural;
                         defaultScale    : Natural;
                         defaultWindowed : Boolean ) is
    begin
        Application(this.all).Construct( company, name, userDir );
        this.defaultWidth := defaultWidth;
        this.defaultHeight := defaultHeight;
        this.defaultScale := defaultScale;
        this.defaultWindowed := defaultWindowed;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Finalize( this : access Gui_Application ) is
    begin
        Fonts.Manager.Finalize;
        Al_Uninstall_Mouse;
        if this.interactive then
            pragma Debug( Dbg( "Shutting down keyboard module", D_INIT, INFO ) );
            Al_Uninstall_Keyboard;
        end if;
        Al_Shutdown_Primitives_Addon;

        pragma Debug( Dbg( "Shutting down Allegro", D_INIT, INFO ) );
        Al_Uninstall_System;
        pragma Debug( Dbg( "Application finalized", D_INIT, INFO ) );

        Application(this.all).Finalize;
    end Finalize;

    ----------------------------------------------------------------------------

    function Get_Display( this : not null access Gui_Application'Class ) return A_Allegro_Display is
    begin
        return this.display;
    end Get_Display;

    ----------------------------------------------------------------------------

    overriding
    function Initialize( this : access Gui_Application ) return Boolean is
        useLogWindow : Boolean := False;
    begin
        if not Application(this.all).Initialize then
            return False;
        end if;

        -- Initialize Allegro
        if not Al_Initialize then
            this.Show_Error( "Error initializing Allegro" );
            Dbg( "Error: Al_Initialize failed", D_INIT, ERROR );
            raise INIT_EXCEPTION;
        end if;

        -- Initialize native dialogs and log window
        if Al_Init_Native_Dialog_Addon then
            for i in 1..Argument_Count loop
                if Argument( i ) = "-log" then
                    useLogWindow := True;
                    exit;
                end if;
            end loop;
            if useLogWindow then
                logWindow := Al_Open_Native_Text_Log( To_String( this.name ) & " log",
                                                      ALLEGRO_TEXTLOG_NO_CLOSE );
                if logWindow = null then
                    this.Show_Error( "Log window disabled - Failed to open log window" );
                end if;
            end if;
        else
            this.Show_Error( "Log window disabled - Failed to init native dialogs" );
        end if;
        Debugging.Set_Output( Debug_Output'Access );

        -- Suggest the display color depth
        declare
            bpp : Integer := Get_Pref( "application", "bpp", COLOR_DEPTH );
        begin
            if bpp /= 15 and then
               bpp /= 16 and then
               bpp /= 24 and then
               bpp /= 32
            then
                bpp := COLOR_DEPTH;
            end if;
            pragma Debug( Dbg( "Application.Initialize: Color depth: " &
                               Image( bpp ) & " bpp", D_INIT, INFO ) );
            Al_Set_New_Display_Option( ALLEGRO_COLOR_SIZE, bpp, ALLEGRO_SUGGEST );
        end;

        -- Set the display size
        --
        -- Use the width/height preferences constrained to the nearest allowed
        -- resolution: For fullscreen, it must be one of the supported
        -- resolutions; for windowed, it must be smaller than the desktop and
        -- larger than the minimum size.
        --
        -- If no width/height preferences are given, use the default size and
        -- the scale preference, constrained to the nearest allowed resolution.
        -- If no scale preference is given, use the default scale.
        declare
            monitor       : aliased Allegro_Monitor_Info;
            bpp           : Integer;
            importance    : Option_Importance;
            desktopWidth,
            desktopHeight : Integer := 0;
            displayWidth,
            displayHeight : Integer := 0;
            scale         : Integer := 0;
        begin
            -- Detect the monitor to use (use the widest)
            -- Determine the desktop resolution
            for i in 0..Al_Get_Num_Video_Adapters loop
                if Al_Get_Monitor_Info( i, monitor'Access ) then
                    if monitor.x2 - monitor.x1 > desktopWidth then
                        Al_Set_New_Display_Adapter( i );
                        desktopWidth := monitor.x2 - monitor.x1;
                        desktopHeight := monitor.y2 - monitor.y1;
                    end if;
                end if;
            end loop;
            if desktopWidth = 0 then
                -- unable to get the desktop resolution!
                desktopWidth := this.defaultWidth;
                desktopHeight := this.defaultHeight;
                pragma Debug( Dbg( "Application.Initialize Warning: " &
                                   "Unable to get desktop resolution",
                                   D_INIT, WARNING ) );
            end if;

            if Get_Pref( "application", "windowed", this.defaultWindowed ) then
                -- use the width/height preferences and scale preferences to
                -- determine the display size, and constrain it to the size of
                -- the desktop.
                Al_Set_New_Display_Flags( ALLEGRO_WINDOWED );
                displayWidth := Get_Pref( "application", "xres", 0 );
                displayHeight := Get_Pref( "application", "yres", 0 );
                if displayWidth /= 0 and then displayHeight /= 0 then
                    -- found width/height preferences
                    --
                    -- use them as the display size. constrain the dimensions
                    -- independently to the desktop size. the proportions may
                    -- not be preserved.
                    displayWidth := Constrain( displayWidth, MIN_WIN_WIDTH, desktopWidth );
                    displayHeight := Constrain( displayHeight, MIN_WIN_HEIGHT, desktopHeight );
                else
                    -- no width/height preferences
                    --
                    -- calculate the display dimensions from the default display
                    -- size and the scale preference. if the default scale is
                    -- non-zero, default proportions will be preserved when
                    -- constrained to the desktop.
                    scale := Constrain( Get_Pref( "application", "scale", this.defaultScale ), 1, 100 );
                    if this.defaultScale = 0 then
                        -- constrain the dimensions independently to the desktop
                        -- size. default proportions may not be preserved.
                        displayWidth := Constrain( this.defaultWidth * scale, MIN_WIN_WIDTH, desktopWidth );
                        displayHeight := Constrain( this.defaultHeight * scale, MIN_WIN_HEIGHT, desktopHeight );
                    else
                        -- preserve the default proportions if the scaling
                        -- factor is too large for the desktop.
                        displayWidth := this.defaultWidth * scale;
                        displayHeight := this.defaultHeight * scale;
                        while displayWidth > desktopWidth or else displayHeight > desktopHeight loop
                            displayWidth := displayWidth - this.defaultWidth;
                            displayHeight := displayHeight - this.defaultHeight;
                        end loop;
                        if displayWidth <= 0 then
                            -- the default size is too large for the desktop in
                            -- at least one dimension. use it anyway to preserve
                            -- default proportions.
                            displayWidth := this.defaultWidth;
                            displayHeight := this.defaultHeight;
                        end if;
                    end if;
                end if;

                if displayWidth = desktopWidth and then displayHeight = desktopHeight then
                    -- if the window size matches the desktop resolution, make
                    -- it as a fullscreen window instead.
                    Al_Set_New_Display_Flags( ALLEGRO_FULLSCREEN_WINDOW );
                end if;
                -- 1. attempt to create the display at the requested resolution
                this.display := Al_Create_Display( displayWidth, displayHeight );
                if this.display = null then
                   -- 2. attempt to create the display at default resolution
                   this.display := Al_Create_Display( this.defaultWidth, this.defaultHeight );
                end if;
            else
                -- use the width/height preferences and scale preferences to
                -- determine the display size, and constrain it to a valid
                -- fullscreen resolution.
                Al_Set_New_Display_Flags( ALLEGRO_FULLSCREEN );
                Al_Get_New_Display_Option( ALLEGRO_COLOR_SIZE, bpp, importance );
                displayWidth := Get_Pref( "application", "xres", 0 );
                displayHeight := Get_Pref( "application", "yres", 0 );
                if displayWidth /= 0 and then displayHeight /= 0 then
                    -- found width/height preferences
                    --
                    -- use them as the display size.
                    displayWidth := Constrain( displayWidth, MIN_FS_WIDTH, 4096 );
                    displayHeight := Constrain( displayHeight, MIN_FS_HEIGHT, 4096 );
                else
                    -- no width/height preferences
                    --
                    -- use the application's default display size and scale.
                    scale := Constrain( Get_Pref( "application", "scale", this.defaultScale ), 1, 100 );
                    displayWidth := this.defaultWidth * scale;
                    displayHeight := this.defaultHeight * scale;
                end if;

                -- 1. attempt to create the display at the requested resolution
                this.display := Al_Create_Display( displayWidth, displayHeight );
                if this.display = null then
                    -- 2. attempt to create the display at default resolution
                    this.display := Al_Create_Display( this.defaultWidth, this.defaultHeight );
                end if;
                if this.display = null then
                    -- 3. attempt to create the display at desktop resolution
                    this.display := Al_Create_Display( desktopWidth, desktopHeight );
                end if;
                if this.display = null then
                    Al_Set_New_Display_Flags( ALLEGRO_FULLSCREEN_WINDOW );
                    -- 4. attempt to create the display at desktop resolution
                    --    as a fullscreen window
                    this.display := Al_Create_Display( desktopWidth, desktopHeight );
                end if;
            end if;

            if this.display = null then
                -- failed to create a display
                Dbg( "Error: Failed to create display [" &
                     Image( displayWidth ) & "x" & Image( displayHeight ) &
                     "]: Error " & Image( Al_Get_Errno ),
                     D_INIT, ERROR );
                this.Show_Error( "Unable to create graphics display window." );
                raise INIT_EXCEPTION with "Failed to create display [" &
                                          Image( displayWidth ) & "x" &
                                          Image( displayHeight ) & "]";
            end if;

            -- release the display so it's not active for the startup thread
            Al_Set_Target_Bitmap( null );

            pragma Debug( Dbg( "Info: Created display: " &
                               Image( Al_Get_Display_Width( this.display ) ) & "x" &
                               Image(  Al_Get_Display_Height( this.display ) ) & ", " &
                               Image( Al_Get_Pixel_Format_Bits( Al_Get_Display_Format( this.display ) ) ) & "bpp, " &
                               Image( Al_Get_Display_Refresh_Rate( this.display ) ) & "Hz",
                               D_INIT, INFO ) );
        end;

        -- Set the initial window title
        Al_Set_Window_Title( this.display, To_String( this.name ) );

        -- Inhibit the screensaver
        if not Al_Inhibit_Screensaver( True ) then
            Dbg( "Warning: Failed to inhibit screensaver", D_INIT, WARNING );
        end if;

        -- Install keyboard handler
        if this.interactive and then not Al_Install_Keyboard then
            Dbg( "Error: Install_Keyboard: " &
                 "Error " & Image( Al_Get_Errno ), D_INIT, ERROR );
            this.Show_Error( "Error installing keyboard support." );
            raise INIT_EXCEPTION with "Error installing keyboard support";
        end if;

        -- Setup mouse driver and cursor
        if not Al_Install_Mouse then
            Dbg( "Error: Install_Mouse: " &
                 "Error " & Image( Al_Get_Errno ), D_INIT, ERROR );
            this.Show_Error( "Error installing mouse support." );
            raise INIT_EXCEPTION with "Error installing mouse support";
        end if;
        if this.interactive and then this.useMouse then
            Al_Set_System_Mouse_Cursor( this.display, ALLEGRO_SYSTEM_MOUSE_CURSOR_ARROW );
        end if;

        -- Initialize the primitives drawing addon
        if not Al_Init_Primitives_Addon then
            Dbg( "Error: Al_Init_Primitives_Addon: " &
                 "Error " & Image( Al_Get_Errno ), D_INIT, ERROR );
            this.Show_Error( "Error initializing primitives drawing." );
            raise INIT_EXCEPTION with "Error initializing primitives drawing";
        end if;

        -- Initialize systems after Allegro
        begin
            xBR.Initialize;
            Fonts.Manager.Initialize;
        exception
            when e : others =>
                Dbg( "Error: Fonts.Manager.Initialize: " &
                     Exception_Name( e ) & " - " & Exception_Message( e ),
                     D_INIT, ERROR );
                this.Show_Error( "Error initializing font support." );
                raise INIT_EXCEPTION with "Error initializing font support";
        end;

        return True;
    exception
        when e : others =>
            Dbg( "Gui_Application.Initialize Exception: ...", D_INIT, ERROR );
            Dbg( e, D_INIT, ERROR );

            this.Finalize;
            return False;
    end Initialize;

    ----------------------------------------------------------------------------

    function Is_Mouse_Enabled( this : not null access Gui_Application'Class ) return Boolean is
    begin
        return this.useMouse;
    end Is_Mouse_Enabled;

    ----------------------------------------------------------------------------

    overriding
    procedure Show_Error( this : access Gui_Application; text : String ) is
    begin
        if 0 = Al_Show_Native_Message_Box( this.display,
                                           "Error!", "An error occurred", text,
                                           "", ALLEGRO_MESSAGEBOX_ERROR )
        then
            -- failed to show the message box
            Application(this.all).Show_Error( text );
        end if;
    end Show_Error;

    --==========================================================================

    procedure Debug_Output( str : String ) is
    begin
        if logWindow /= null then
            Al_Append_Native_Text_Log( logWindow, str & ASCII.LF );
        else
            null;
#if DEBUG then
            Ada.Text_IO.Put_Line( str );
#end if;
        end if;
    end Debug_Output;

end Applications.Gui;
