--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Allegro.Image_Addon;               use Allegro.Image_Addon;
with Ada.Text_IO;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with GNAT.OS_Lib;
with Preferences;                       use Preferences;

package body Applications is

    protected Singleton is

        -- Creates the global application if it doesn't exist. An exception is
        -- raised on error.
        procedure Create( app : out A_Application );
        pragma Postcondition( app /= null );

        -- Deletes the global application.
        procedure Delete;

        -- Returns the global application; may return null.
        function Get return A_Application;

        -- Sets the function used to allocate the global application.
        procedure Set_Allocator( alloc : not null Allocator );

    private
        globalApp : A_Application := null;
        allocate  : Allocator := null;
    end Singleton;

    ----------------------------------------------------------------------------

    protected body Singleton is

        procedure Create( app : out A_Application ) is
        begin
            if globalApp = null then
                -- Initialize systems here
                if allocate /= null then
                    globalApp := allocate.all;
                    pragma Debug( Dbg( "Created application <" &
                                       globalApp.Get_Name & '>',
                                       D_INIT, INFO ) );
                else
                    raise USE_ERROR with "Application allocator not registered";
                end if;
            end if;
            app := globalApp;
        end Create;

        ------------------------------------------------------------------------

        procedure Delete is
        begin
            if globalApp /= null then
                pragma Debug( Dbg( "Deleting application", D_INIT, INFO ) );
                Delete( globalApp );
            end if;
        end Delete;

        ------------------------------------------------------------------------

        function Get return A_Application is
        begin
            return globalApp;
        end Get;

        ------------------------------------------------------------------------

        procedure Set_Allocator( alloc : not null Allocator ) is
        begin
            allocate := alloc;
        end Set_Allocator;

    end Singleton;

    --==========================================================================

    procedure Create_Application is
        app : A_Application;
    begin
        Singleton.Create( app );
    end Create_Application;

    ----------------------------------------------------------------------------

    procedure Delete_Application is
    begin
        Singleton.Delete;
    end Delete_Application;

    ----------------------------------------------------------------------------

    function Get_Application return A_Application is
    begin
        return Singleton.Get;
    end Get_Application;

    ----------------------------------------------------------------------------

    procedure Run_Application( returnCode : out Integer ) is
        app : constant A_Application := Singleton.Get;
    begin
        if app.Initialize then
            begin
                pragma Debug( Dbg( "** Running application **", D_INIT, INFO ) );
                returnCode := app.Run;
                pragma Debug( Dbg( "** Application run complete **", D_INIT, INFO ) );
            exception
                when e : others =>
                    returnCode := 1;
                    pragma Debug( Dbg( "** Application run failed **", D_INIT, INFO ) );
                    pragma Debug( Dbg( "Application.Run Exception: ..." ) );
                    pragma Debug( Dbg( e ) );
            end;
            app.Finalize;
        else
            returnCode := -1;
        end if;
    end Run_Application;

    ----------------------------------------------------------------------------

    procedure Register_Allocator( allocate : not null Allocator ) is
    begin
        Singleton.Set_Allocator( allocate );
    end Register_Allocator;

    --==========================================================================

    not overriding
    procedure Finalize( this : access Application ) is
        pragma Unreferenced( this );
    begin
        Preferences.Finalize;
    end Finalize;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this    : access Application;
                         company : String;
                         name    : String;
                         userDir : String ) is
    begin
        Limited_Object(this.all).Construct;
        this.company := To_Unbounded_String( company );
        this.name    := To_Unbounded_String( name );
        this.userDir := To_Unbounded_String( userDir );
        this.lock    := new Locking_Object;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Application ) is
    begin
        Delete( this.lock );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Get_Company( this : not null access Application'Class ) return String is
    begin
        return To_String( this.company );
    end Get_Company;

    ----------------------------------------------------------------------------

    function Get_Name( this : not null access Application'Class ) return String is
    begin
        return To_String( this.name );
    end Get_Name;

    ----------------------------------------------------------------------------

    function Get_User_Directory( this : not null access Application'Class ) return String is
    begin
        return To_String( this.userDir );
    end Get_User_Directory;

    ----------------------------------------------------------------------------

    not overriding
    function Initialize( this : access Application ) return Boolean is
    begin
        -- Initialize Preferences
        if GNAT.OS_Lib.Is_Regular_File( A_Application(this).Get_Name & ".cfg" ) then
            -- a config file exists in the working directory; use that one
            pragma Debug( Dbg( "Application.Initialize Info: Using local config <" &
                               A_Application(this).Get_Name & ".cfg>", D_INIT, INFO ) );
            Preferences.Initialize( A_Application(this).Get_Name & ".cfg" );
        else
            -- use the config file in the user's directory, or write one there
            -- later if preferences change.
            pragma Debug( Dbg( "Application.Initialize Info: Using config <" &
                               To_String( this.userDir ) &
                               A_Application(this).Get_Name & ".cfg>",
                               D_INIT, INFO ) );
            Preferences.Initialize( To_String( this.userDir ) &
                                    A_Application(this).Get_Name & ".cfg" );
        end if;

        if not Al_Init_Image_Addon then
            -- failed to initialize Allegro image loaders
            Dbg( "Error: Al_Init_Image_Addon failed", D_INIT, ERROR );
            this.Show_Error( "Error initializing Allegro image addon." );
            raise INIT_EXCEPTION with "Error initializing Allegro image addon";
        end if;

        return True;
    exception
        when e : others =>
            Dbg( "Application.Initialize Exception: ...", D_INIT, ERROR );
            Dbg( e, D_INIT, ERROR );

            this.Finalize;
            return False;
    end Initialize;

    ----------------------------------------------------------------------------

    function Run( this : access Application ) return Integer is
        pragma Unreferenced( this );
    begin
        raise USE_ERROR with "Function Application.Run must be implemented";
        return 1;
    end Run;

    ----------------------------------------------------------------------------

    procedure Show_Error( this : access Application; text : String ) is
        pragma Unreferenced( this );
    begin
        Ada.Text_IO.Put_Line( Ada.Text_IO.Standard_Error, "Error: " & text );
    end Show_Error;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Application ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Applications;
