--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Containers.Doubly_Linked_Lists;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Unchecked_Deallocation;
with Allegro;                           use Allegro;
with Allegro.Audio.Codecs;              use Allegro.Audio.Codecs;
with Allegro.File_IO;                   use Allegro.File_IO;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Audio;                      use Events.Audio;
with Preferences;                       use Preferences;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;

pragma Elaborate_All( Preferences );

package body Audio_Players is

    STREAM_BUFFER_COUNT   : constant := 4;
    STREAM_BUFFER_SAMPLES : constant := 2048;

    initialized : Boolean := False;

    -- Initializes the audio driver.
    procedure Initialize;

    ----------------------------------------------------------------------------

    package Sample_Instance_Lists is new
        Ada.Containers.Doubly_Linked_Lists( A_Allegro_Sample_Instance, "=" );

    type Sound is
        record
            resource  : A_Resource_File := null;
            sample    : A_Allegro_Sample := null;
            instances : Sample_Instance_Lists.List;
        end record;

    ----------------------------------------------------------------------------

    -- Called when a Play_Music event is received to begin playing a music track.
    -- Only one music track can play at a time so this will stop any previously
    -- playing music. Music tracks will loop.
    procedure Play_Music( this : not null access Audio_Player'Class; name : String );
    pragma Precondition( name'Length > 0 );

    -- Called when a Play_Sound event is received to begin playing a sound once
    -- through. Multiple sounds can play simultaneously.
    procedure Play_Sound( this : not null access Audio_Player'Class; name : String );
    pragma Precondition( name'Length > 0 );

    -- Called when a Stop_Music event is received. Any currently looping music
    -- will be stopped.
    procedure Stop_Music( this : not null access Audio_Player'Class );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    -- Initializes the audio driver.
    procedure Initialize is
    begin
        if not initialized then
            pragma Debug( Dbg( "Initializing audio system", D_INIT, INFO ) );
            if Al_Install_Audio then
                if not Al_Init_ACodec_Addon then
                    Dbg( "Error: Unable to init audio codecs", D_INIT, ERROR );
                end if;
            else
                Dbg( "Error: Unable to install audio", D_INIT, ERROR );
            end if;

            initialized := True;
        end if;
    end Initialize;

    --==========================================================================

    function Create_Audio_Player( corral : not null A_Corral ) return A_Audio_Player is
        this : constant A_Audio_Player := new Audio_Player;
    begin
        this.Construct( corral );
        return this;
    end Create_Audio_Player;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Audio_Player; corral : not null A_Corral ) is
    begin
        Limited_Object(this.all).Construct;
        Audio_Players.Initialize;
        this.voice := Al_Create_Voice( 44100,
                                       ALLEGRO_AUDIO_DEPTH_INT16,
                                       ALLEGRO_CHANNEL_CONF_2 );
        if this.voice /= null then
            this.mixer := Al_Create_Mixer( 44100,
                                           ALLEGRO_AUDIO_DEPTH_FLOAT32,
                                           ALLEGRO_CHANNEL_CONF_2 );
            if this.mixer /= null then
                if not Al_Attach_Mixer_To_Voice( this.mixer, this.voice ) then
                    Dbg( "Error: Audio_Player: Unable to attach mixer to voice",
                         D_INIT, ERROR );
                end if;
            else
                Dbg( "Error: Audio_Player: Unable to allocate mixer", D_INIT, ERROR );
            end if;
        else
            Dbg( "Error: Audio_Player: Unable to allocate voice", D_INIT, ERROR );
        end if;
        this.corral := corral;
        this.corral.Add_Listener( A_Event_Listener(this), "Play_Music" );
        this.corral.Add_Listener( A_Event_Listener(this), "Play_Sound" );
        this.corral.Add_Listener( A_Event_Listener(this), "Stop_Music" );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Audio_Player ) is

        ------------------------------------------------------------------------

        procedure Delete_Sounds is

            procedure Delete is new Ada.Unchecked_Deallocation( Sound, A_Sound );

            pos      : Sound_Maps.Cursor := this.sounds.First;
            sound    : A_Sound;
            instance : A_Allegro_Sample_Instance;
        begin
            -- loop over all sounds
            while Sound_Maps.Has_Element( pos ) loop
                sound := Sound_Maps.Element( pos );

                -- loop over all instances
                while not sound.instances.Is_Empty loop
                    -- delete each instance
                    instance := sound.instances.First_Element;
                    Al_Detach_Sample_Instance( instance );
                    Al_Destroy_Sample_Instance( instance );
                    sound.instances.Delete_First;
                end loop;

                -- delete the sound
                Al_Destroy_Sample( sound.sample );
                Release( sound.resource );
                Delete( sound );

                Sound_Maps.Next( pos );
            end loop;
            this.sounds.Clear;
        end Delete_Sounds;

        ------------------------------------------------------------------------

    begin
        this.corral.Remove_Listener( A_Event_Listener(this), "Stop_Music" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Play_Sound" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Play_Music" );

        this.Stop_Music;
        Delete_Sounds;

        Al_Destroy_Mixer( this.mixer );
        Al_Destroy_Voice( this.voice );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access Audio_Player ) return String is
        pragma Unreferenced( this );
    begin
        return "Audio_Player";
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Audio_Player;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        if evt.Get_Id = PLAY_SOUND_ID then
            this.Play_Sound( A_Play_Sound_Event(evt).Get_Sound_Name );
        elsif evt.Get_Id = PLAY_MUSIC_ID then
            this.Play_Music( A_Play_Music_Event(evt).Get_Music_Name );
        elsif evt.Get_Id = STOP_MUSIC_ID then
            this.Stop_Music;
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    procedure Play_Music( this : not null access Audio_Player'Class; name : String ) is
        filename : constant String := Add_Extension( name, "ogg" );
    begin
        this.Stop_Music;
        pragma Debug( Dbg( "Audio_Player.Play_Music <" & name & '>', D_AUDIO, INFO ) );
        if Get_Pref( "application", "music" ) then

            -- load the file resource
            this.musicRes := Load_Resource( filename, "audio" );
            if this.musicRes = null then
                raise RESOURCE_ERROR with "File not found <" & filename & '>';
            end if;

            -- load the audio and create the stream
            this.musicStream := Al_Load_Audio_Stream_f( this.musicRes.Get_File,
                                                        "." & Get_Extension( filename ),
                                                        STREAM_BUFFER_COUNT,
                                                        STREAM_BUFFER_SAMPLES );
            if this.musicStream = null then
                raise RESOURCE_ERROR with "File format error in <" & filename & '>';
            end if;

            -- begin playing the music stream
            if not Al_Attach_Audio_Stream_To_Mixer( this.musicStream, this.mixer ) or else
               not Al_Set_Audio_Stream_Playmode( this.musicStream, ALLEGRO_PLAYMODE_LOOP ) or else
               not Al_Set_Audio_Stream_Playing( this.musicStream, True )
            then
                raise RESOURCE_ERROR with "Unable to play music <" & filename & '>';
            end if;

        end if;
    exception
        when e : others =>
            Dbg( "Error playing music <" & filename & ">: " &
                  Exception_Name( e ) & " -- " & Exception_Message( e ),
                  D_AUDIO, ERROR );
            this.Stop_Music;
    end Play_Music;

    ----------------------------------------------------------------------------

    procedure Play_Sound( this : not null access Audio_Player'Class; name : String ) is

        ------------------------------------------------------------------------

        function Load_Sound( filename : String ) return A_Sound is
            res    : A_Resource_File;
            file   : A_Allegro_File;
            sample : A_Allegro_Sample;
            result : A_Sound;
        begin
            -- load the file resources
            res := Load_Resource( filename, "audio", keep => True );
            if res = null then
                raise RESOURCE_ERROR with "File not found <" & filename & '>';
            end if;

            -- load the audio from the file
            file := res.Get_File;
            sample := Al_Load_Sample_f( file, "." & Get_Extension( filename ) );
            Al_Fclose( file );
            if sample = null then
                raise RESOURCE_ERROR with "File format error in <" & filename & '>';
            end if;

            result := new Sound;
            result.resource := res;
            result.sample := sample;
            return result;
        exception
            when others =>
                Release( res );
                raise;
        end Load_Sound;

        ------------------------------------------------------------------------

        filename : constant String := Add_Extension( name, "wav" );
        pos      : Sound_Maps.Cursor;
        snd      : A_Sound;
        instance : A_Allegro_Sample_Instance;
    begin
        pragma Debug( Dbg( "Audio_Player.Play_Sound <" & name & '>', D_AUDIO, INFO ) );
        if Get_Pref( "application", "soundfx" ) then

            pos := this.sounds.Find( filename );
            if Sound_Maps.Has_Element( pos ) then
                -- retrieve the Sound from the loaded sounds map
                snd := Sound_Maps.Element( pos );
            else
                -- create a Sound by load a sample from resources
                -- throws exception on error
                snd := Load_Sound( filename );
                this.sounds.Insert( filename, snd );
            end if;

            -- create and play a new instance of the sample
            instance := Al_Create_Sample_Instance( snd.sample );
            if Al_Attach_Sample_Instance_To_Mixer( instance, this.mixer ) and then
               Al_Set_Sample_Instance_Playmode( instance, ALLEGRO_PLAYMODE_ONCE ) and then
               Al_Play_Sample_Instance( instance )
            then
                -- success! add it to the sound's instance list
                snd.instances.Append( instance );
            else
                Dbg( "Unable to play <" & filename & '>', D_AUDIO, ERROR );
                Al_Detach_Sample_Instance( instance );
                Al_Destroy_Sample_Instance( instance );
            end if;

        end if;
    exception
        when e : others =>
            -- failed to load the sound
            Dbg( "Error playing sound <" & filename & ">: " &
                 Exception_Name( e ) & " -- " & Exception_Message( e ),
                 D_AUDIO, ERROR );
    end Play_Sound;

    ----------------------------------------------------------------------------

    procedure Stop_Music( this : not null access Audio_Player'Class ) is
    begin
        pragma Debug( Dbg( "Audio_Player.Stop_Music", D_AUDIO, INFO ) );
        if this.musicStream /= null then
            Al_Detach_Audio_Stream( this.musicStream );
            Al_Destroy_Audio_Stream( this.musicStream );
        end if;
        Release( this.musicRes );
    end Stop_Music;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Audio_Player; time : Tick_Time ) is
        pragma Unreferenced( time );
        sPos     : Sound_Maps.Cursor := this.sounds.First;
        iPos,
        iNext    : Sample_Instance_Lists.Cursor;
        sound    : A_Sound;
        instance : A_Allegro_Sample_Instance;
    begin
        -- loop over all sounds
        while Sound_Maps.Has_Element( sPos ) loop
            sound := Sound_Maps.Element( sPos );

            -- check the playing state of each sample instance
            iPos := sound.instances.First;
            while Sample_Instance_Lists.Has_Element( iPos ) loop
                instance := Sample_Instance_Lists.Element( iPos );
                iNext := Sample_Instance_Lists.Next( iPos );

                -- remove the instance if it finished playing
                if not Al_Get_Sample_Instance_Playing( instance ) then
                    Al_Detach_Sample_Instance( instance );
                    Al_Destroy_Sample_Instance( instance );
                    sound.instances.Delete( iPos );
                end if;

                iPos := iNext;
            end loop;

            Sound_Maps.Next( sPos );
        end loop;
    end Tick;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Audio_Player ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

begin

    Set_Default( "application", "music", True );     -- music enabled?
    Set_Default( "application", "soundfx", True );   -- sound effects enabled?

end Audio_Players;
