--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Calendar;                      use Ada.Calendar;
with Ada.Calendar.Formatting;           use Ada.Calendar.Formatting;
with Ada.Containers.Indefinite_Ordered_Maps;
with Ada.Streams.Stream_IO;             use Ada.Streams.Stream_IO;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Ada.Text_IO;                       use Ada.Text_IO;
with Ada.Text_IO.Unbounded_IO;          use Ada.Text_IO.Unbounded_IO;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with GNAT.OS_Lib;
with GNAT.SHA1;                         use GNAT.SHA1;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;

-- these are referenced in RELEASE builds
pragma Warnings( Off, Debugging );
pragma Warnings( Off, Engine_Debugging );

package body Content_Signing is

#if DEBUG then
    -- If the executable is built with debugging, allow unsigned content to be
    -- loaded by default. This is helpful for development work.
    ALLOW_UNSIGNED_DEFAULT : constant Boolean := True;
#else
    -- If the executable is built for release, don't allow unsigned content to
    -- be loaded by default. The application must specifically decide it's ok to
    -- load unsigned content.
    ALLOW_UNSIGNED_DEFAULT : constant Boolean := False;
#end if;

    SIGNATURE           : constant String := "-- Signature: ";
    SECRET_TEXT         : constant String := "Billy Blaze, eight year-old genius,";
    DICTIONARY_FILENAME : constant String := "files.digest";

    CHUNK_SIZE : constant Stream_Element_Offset := 4096;

    package Hash_Maps is new Ada.Containers.Indefinite_Ordered_Maps( String, String, "<", "=" );
    use Hash_Maps;

    ----------------------------------------------------------------------------

    function Valid_Signature( filename : String ) return Boolean;

    ----------------------------------------------------------------------------

    protected Dictionary is

        procedure Allow_Unsigned_Content( enabled : Boolean );

        procedure Is_Signed( filename : String; signed : out Boolean );

        procedure Match( filename : String;
                         hash     : GNAT.SHA1.Message_Digest;
                         matches  : out Boolean );

    private
        hashes        : Hash_Maps.Map;
        loaded        : Boolean := False;
        allowUnsigned : Boolean := ALLOW_UNSIGNED_DEFAULT;
    end Dictionary;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Valid_Signature( filename : String ) return Boolean is
        valid : Boolean := False;
        fl    : Ada.Text_IO.File_Type;
        line  : Unbounded_String;
        state : GNAT.SHA1.Context;
    begin
        Open( fl, In_File, filename );
        while not End_Of_File( fl ) loop
            Get_Line( fl, line );
            if Head( line, SIGNATURE'Length ) = SIGNATURE then
                -- found the signature line...
                -- replace the hash on the signature line with the secret text
                -- and include it in the hash.
                Update( state, SIGNATURE & SECRET_TEXT &
                               Tail( line, Length( line ) - (SIGNATURE'Length + Message_Digest'Length) ) );
                -- check that there isn't any more to the file and that the hash
                -- of the file matches the hash recorded on the signature line
                valid := End_Of_File( fl ) and then
                         Digest( state ) = Slice( line,
                                                  SIGNATURE'Length + 1,
                                                  SIGNATURE'Length + Message_Digest'Length );
                exit;
            else
                Update( state, To_String( line ) );
            end if;
        end loop;
        Close( fl );
        return valid;
    exception
        when others =>
            return False;
    end Valid_Signature;

    --==========================================================================

    protected body Dictionary is

        procedure Load_Dictionary is
            fl   : Ada.Text_IO.File_Type;
            line : Unbounded_String;
            pos  : Natural;
        begin
            loaded := True;
            if not Valid_Signature( DICTIONARY_FILENAME ) then
                return;
            end if;

            Open( fl, In_File, DICTIONARY_FILENAME );
            while not End_Of_File( fl ) loop
                Get_Line( fl, line );
                if Head( line, SIGNATURE'Length ) = SIGNATURE then
                    -- found the signature line, stop reading
                    exit;
                else
                    -- add the recorded hash
                    pos := Index( line, " " );
                    if pos = GNAT.SHA1.Message_Digest'Length + 1 then
                        hashes.Insert( Slice( line, pos + 2, Length( line ) ),
                                       Slice( line, 1, pos - 1 ) );
                    end if;
                end if;
            end loop;
            Close( fl );
        exception
            when others =>
                null;
        end Load_Dictionary;

        ------------------------------------------------------------------------

        procedure Allow_Unsigned_Content( enabled : Boolean ) is
        begin
            allowUnsigned := enabled;
        end Allow_Unsigned_Content;

        ------------------------------------------------------------------------

        procedure Is_Signed( filename : String; signed : out Boolean ) is
        begin
            if not loaded then
                Load_Dictionary;
            end if;
            signed := hashes.Contains( Get_Filename( filename ) );
        end Is_Signed;

        ------------------------------------------------------------------------

        procedure Match( filename : String;
                         hash     : GNAT.SHA1.Message_Digest;
                         matches  : out Boolean ) is
        begin
            if not loaded then
                Load_Dictionary;
            end if;

            -- If a hash is found, it must match the given hash
            declare
                pos : constant Hash_Maps.Cursor := hashes.Find( Get_Filename( filename ) );
            begin
                matches := Has_Element( pos ) and then
                           (Message_Digest(Element( pos )) = hash);
            end;
            if not matches then
                -- Hash doesn't match or no hash could be found; allow it anyway
                -- if  unsigned content is allowed
                if allowUnsigned then
#if not DEBUG then
                    -- Don't display this warning if the executable is built in
                    -- debug mode. It would only be displayed to a developer and
                    -- he doesn't care.
                    Dbg( "Warning: File " & filename &  " is unsigned", D_RES, WARNING );
#end if;
                    matches := True;
                end if;
            end if;
        end Match;

    end Dictionary;

    --==========================================================================

    procedure Allow_Unsigned_Content( enabled : Boolean ) is
    begin
        Dictionary.Allow_Unsigned_Content( enabled );
    end Allow_Unsigned_Content;

    ----------------------------------------------------------------------------

    function Is_Signed( filename : String ) return Boolean is
        result : Boolean;
    begin
        Dictionary.Is_Signed( filename, result );
        return result;
    end Is_Signed;

    ----------------------------------------------------------------------------

    function Is_Valid( filename : String;
                       data     : Stream_Element_Array ) return Boolean is
        result : Boolean;
        cntxt  : GNAT.SHA1.Context;
        pos    : Stream_Element_Offset := data'First;
    begin
        while pos <= data'Last loop
            Update( cntxt, data(pos..Stream_Element_Offset'Min( pos + CHUNK_SIZE - 1, data'Last )) );
            pos := pos + CHUNK_SIZE;
        end loop;

        Dictionary.Match( filename, Digest( cntxt ), result );
        return result;
    end Is_Valid;

    ----------------------------------------------------------------------------

    procedure Make_Dictionary( filename  : String;
                               inFiles   : String_List;
                               authority : String;
                               progress  : access procedure( filename   : String;
                                                             completion : Natural ) := null ) is
        outFile  : Ada.Text_IO.File_Type;
        outState : GNAT.SHA1.Context;
        data     : aliased Stream_Element_Array(1..CHUNK_SIZE);
        ok       : Boolean;
    begin
        -- create the dictionary and the digest of its content that we'll write
        begin
            Create( outFile, Out_File, filename );
        exception
            when others =>
                raise File_Error with "Error creating " &  filename;
        end;

        -- loop across all input files, calculating hash values for each. as we
        -- write to the dictionary file we also update its digest value. Line
        -- terminators are not included, only the text. This is ok because the
        -- signature is not strictly a hash of the dictionary contents; it's
        -- only based on them. We include secret known text in the hash to
        -- prevent someone from modifying the dictionary and recomputing the
        -- signature.
        for i in inFiles'Range loop
            declare
                procedure Write_To_Dictionary( line : String ) is
                begin
                    if progress /= null then
                        progress.all( inFiles(i).all,
                                      Natural(Float(i+1-inFiles'First) /
                                      Float(inFiles'Length) * 100) );
                    end if;
                    Put_Line( outFile, line );
                    Update( outState, line );
                end Write_To_Dictionary;

                inFile  : Ada.Streams.Stream_IO.File_Type;
                inState : GNAT.SHA1.Context;
                last    : Stream_Element_Offset;
            begin
                Open( inFile, In_File, inFiles(i).all );
                while not End_Of_File( inFile ) loop
                    Read( inFile, data, last );
                    Update( inState, data(data'First..last) );
                end loop;
                Close( inFile );
                Write_To_Dictionary( Digest( inState ) & " *" & Get_Filename( inFiles(i).all ) );
            exception
                when others =>
                    raise File_Error with "File not found: " & inFiles(i).all;
            end;
        end loop;

        -- write the signature to the dictionary
        declare
            sigTail : constant String := " by " & authority &
                                         " on " & Image( Clock ) & " UTC --";
        begin
            Update( outState, SIGNATURE & SECRET_TEXT & sigTail );
            Put_Line( outFile, SIGNATURE & Digest( outState ) & sigTail );
        end;

        Close( outFile );
    exception
        when others =>
            if Is_Open( outFile ) then
                Close( outFile );
                GNAT.OS_Lib.Delete_File( filename, ok );
            end if;
            raise;
    end Make_Dictionary;

end Content_Signing;
