--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Containers.Doubly_Linked_Lists;
with Ada.Unchecked_Deallocation;
with Allegro.Bitmaps.Drawing;           use Allegro.Bitmaps.Drawing;
with Allegro.Blending;                  use Allegro.Blending;
with Allegro.Displays;                  use Allegro.Displays;
with Allegro.Drawing;                   use Allegro.Drawing;
with Allegro.Drawing.Primitives;        use Allegro.Drawing.Primitives;
with Allegro.State;                     use Allegro.State;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;

package body Drawing is

    -- setting an unaccelerated target
    unaccleratedTargetWarning : Boolean := False;
    pragma Warnings( Off, unaccleratedTargetWarning );

    -- drawing an unaccelerated bitmap
    unaccleratedDrawWarning : Boolean := False;
    pragma Warnings( Off, unaccleratedDrawWarning );

    package State_Stacks is new Ada.Containers.Doubly_Linked_Lists( A_Allegro_State, "=" );
    stateStack : State_Stacks.List;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    procedure Push_State is
        use State_Stacks;
        state : constant A_Allegro_State := new Allegro_State;
    begin
        Al_Store_State( state.all, ALLEGRO_STATE_TARGET_BITMAP or
                                   ALLEGRO_STATE_BLENDER or
                                   ALLEGRO_STATE_TRANSFORM );
        stateStack.Append( state );
    end Push_State;

    ----------------------------------------------------------------------------

    procedure Pop_State is
        use State_Stacks;
        procedure Delete is new Ada.Unchecked_Deallocation( Allegro_State, A_Allegro_State );
        state : A_Allegro_State;
    begin
        if not stateStack.Is_Empty then
            state := stateStack.Last_Element;
            stateStack.Delete_Last;
            Al_Restore_State( state.all );
            Delete( state );
        else
            Dbg( "Error: Can't pop empty graphics state stack", D_GUI, ERROR );
        end if;
    end Pop_State;

    --==========================================================================

    -- Sets the drawing target bitmap.
    procedure Set_Target_Bitmap( bmp : not null A_Allegro_Bitmap ) is
    begin
#if DEBUG then
            -- Notify the developer that a bitmap being drawn is not
            -- accelerated. It's likely that the bitmap was not created within
            -- the context of the currently active display. This is a
            -- development error and should be fixed immediately because
            -- unaccelerated drawing is extremely slow.
            if not Al_Is_Compatible_Bitmap( bmp ) and then not unaccleratedTargetWarning then
                Dbg( "Warning: Setting unaccelerated target bitmap!", D_GUI, WARNING );
                unaccleratedTargetWarning := True;
            end if;
#end if;
        Al_Set_Target_Bitmap( bmp );
    end Set_Target_Bitmap;

    ----------------------------------------------------------------------------

    -- Returns the drawing target bitmap.
    function Get_Target_Bitmap return A_Allegro_Bitmap is
    begin
        return Al_Get_Target_Bitmap;
    end Get_Target_Bitmap;

    ----------------------------------------------------------------------------

    function Get_Target_Height return Natural is
    begin
        return Al_Get_Bitmap_Height( Al_Get_Target_Bitmap );
    end Get_Target_Height;

    ----------------------------------------------------------------------------

    function Get_Target_Width return Natural is
    begin
        return Al_Get_Bitmap_Width( Al_Get_Target_Bitmap );
    end Get_Target_Width;

    --==========================================================================

    procedure Clear_To_Color( color : Allegro_Color ) is
    begin
        Al_Clear_To_Color( color );
    end Clear_To_Color;

    ----------------------------------------------------------------------------

    procedure Draw_Bitmap( bmp : A_Allegro_Bitmap; x, y : Integer ) is
    begin
        if bmp /= null then
#if DEBUG then
            -- Notify the developer that a bitmap being drawn is not
            -- accelerated. It's likely that the bitmap was not created within
            -- the context of the currently active display. This is a
            -- development error and should be fixed immediately because
            -- unaccelerated drawing is extremely slow.
            if not Al_Is_Compatible_Bitmap( bmp ) and then not unaccleratedDrawWarning then
                Dbg( "Warning: Drawing unaccelerated bitmaps!", D_GUI, WARNING );
                unaccleratedDrawWarning := True;
            end if;
#end if;
            Al_Draw_Bitmap( bmp, Float(x), Float(y), 0 );
        end if;
    end Draw_Bitmap;

    ----------------------------------------------------------------------------

    procedure Draw_Bitmap_Region_Stretched( bmp        : A_Allegro_Bitmap;
                                            srcX,
                                            srcY       : Integer;
                                            srcWidth,
                                            srcHeight  : Positive;
                                            destX,
                                            destY      : Integer;
                                            destWidth,
                                            destHeight : Positive ) is
    begin
        if bmp /= null then
            Al_Draw_Scaled_Bitmap( bmp, Float(srcX), Float(srcY),
                                   Float(srcWidth), Float(srcHeight),
                                   Float(destX), Float(destY),
                                   Float(destWidth), Float(destHeight), 0 );
        end if;
    end Draw_Bitmap_Region_Stretched;

    ----------------------------------------------------------------------------

    procedure Draw_Bitmap_Stretched( bmp           : A_Allegro_Bitmap;
                                     x, y          : Integer;
                                     width, height : Positive;
                                     proportional  : Boolean := False ) is
            srcWidth, srcHeight : Positive;
            newWidth, newHeight : Natural;
            srcRatio, destRatio : Float;
    begin
        if bmp = null then
            return;
        end if;

        srcWidth := Al_Get_Bitmap_Width( bmp );
        srcHeight := Al_Get_Bitmap_Height( bmp );

        if proportional then
            destRatio := Float(width) / Float(height);
            srcRatio  := Float(srcWidth) / Float(srcHeight);
            if srcRatio = destRatio then
                -- proportions match
                Al_Draw_Scaled_Bitmap( bmp,
                                       0.0, 0.0,
                                       Float(srcWidth), Float(srcHeight),
                                       Float(x), Float(y),
                                       Float(width), Float(height),
                                       0 );
            elsif srcRatio < destRatio then
                -- bmp is taller; adjust width and shift right
                newWidth := width * srcWidth / srcHeight;
                Al_Draw_Scaled_Bitmap( bmp,
                                       0.0, 0.0,
                                       Float(srcWidth), Float(srcHeight),
                                       Float(x + ((width - newWidth) / 2)), Float(y),
                                       Float(newWidth), Float(height),
                                       0 );
            else
                -- bmp is wider; adjust height and shift down
                newHeight := height * srcHeight / srcWidth;
                Al_Draw_Scaled_Bitmap( bmp,
                                       0.0, 0.0,
                                       Float(srcWidth), Float(srcHeight),
                                       Float(x), Float(y + ((height - newHeight) / 2)),
                                       Float(width), Float(newHeight),
                                       0 );
            end if;
        else
            Al_Draw_Scaled_Bitmap( bmp,
                                   0.0, 0.0, Float(srcWidth), Float(srcHeight),
                                   Float(x), Float(y), Float(width), Float(height),
                                   0 );
        end if;
    end Draw_Bitmap_Stretched;

    ----------------------------------------------------------------------------

    procedure Draw_String( str   : String;
                           x, y  : Integer;
                           font  : A_Font;
                           color : Allegro_Color ) is
    begin
        if font /= null then
            font.Draw_String( str, Float(x), Float(y), color );
        end if;
    end Draw_String;

    ----------------------------------------------------------------------------

    procedure Line_H( x1, x2 : Integer; y : Integer; color : Allegro_Color ) is
        l, r : Integer;
    begin
        if x1 < x2 then
            l := x1;
            r := x2;
        else
            l := x2;
            r := x1;
        end if;

        Al_Draw_Line( Float(l),     Float(y) + 0.5,
                      Float(r + 1), Float(y) + 0.5,
                      color,
                      1.0 );
    end Line_H;

    ----------------------------------------------------------------------------

    procedure Line_V( x : Integer; y1, y2 : Integer; color : Allegro_Color ) is
        t, b : Integer;
    begin
        if y1 < y2 then
            t := y1;
            b := y2;
        else
            t := y2;
            b := y1;
        end if;

        Al_Draw_Line( Float(x) + 0.5, Float(t),
                      Float(x) + 0.5, Float(b + 1),
                      color,
                      1.0 );
    end Line_V;

    ----------------------------------------------------------------------------

    procedure Rect( x1, y1, x2, y2 : Integer; color : Allegro_Color ) is
        l, r, t, b : Integer;
    begin
        if x1 < x2 then
            l := x1;
            r := x2;
        else
            l := x2;
            r := x1;
        end if;

        if y1 < y2 then
            t := y1;
            b := y2;
        else
            t := y2;
            b := y1;
        end if;

        -- draw a pixel-perfect rectangle that is inclusive of the coordinates
        Al_Draw_Rectangle( Float(l) + 0.5, Float(t) + 0.5,
                           Float(r) + 0.5, Float(b) + 0.5,
                           color,
                           1.0 );
    end Rect;

    ----------------------------------------------------------------------------

    procedure Rectfill( x1, y1, x2, y2 : Integer; color : Allegro_Color ) is
        l, r, t, b : Integer;
    begin
        if x1 < x2 then
            l := x1;
            r := x2;
        else
            l := x2;
            r := x1;
        end if;

        if y1 < y2 then
            t := y1;
            b := y2;
        else
            t := y2;
            b := y1;
        end if;

        -- draw a pixel-perfect rectangle that is inclusive of the coordinates
        Al_Draw_Filled_Rectangle( Float(l), Float(t),
                                  Float(r + 1), Float(b + 1),
                                  color );
    end Rectfill;

    ----------------------------------------------------------------------------

    procedure Rectfill_Additive( x1, y1, x2, y2 : Integer; color : Allegro_Color ) is
        l, r, t, b : Integer;
        bo         : Allegro_Blend_Operations;
        bs, bd     : Allegro_Blend_Mode;
    begin
        if x1 < x2 then
            l := x1;
            r := x2;
        else
            l := x2;
            r := x1;
        end if;

        if y1 < y2 then
            t := y1;
            b := y2;
        else
            t := y2;
            b := y1;
        end if;

        -- draw a pixel-perfect rectangle that is inclusive of the coordinates
        Al_Get_Blender( bo, bs, bd );
        Al_Set_Blender( ALLEGRO_ADD, ALLEGRO_ONE, ALLEGRO_ONE );
        Al_Draw_Filled_Rectangle( Float(l), Float(t),
                                  Float(r + 1), Float(b + 1),
                                  color );
        Al_Set_Blender( bo, bs, bd );
    end Rectfill_Additive;

    ----------------------------------------------------------------------------

    procedure Triangle( x1, y1 : Integer;
                        x2, y2 : Integer;
                        x3, y3 : Integer;
                        color  : Allegro_Color ) is
        a1 : Integer := x1;
        a2 : Integer := x2;
        a3 : Integer := x3;
        b1 : Integer := y1;
        b2 : Integer := y2;
        b3 : Integer := y3;
    begin
        if x1 >= x2 and then x1 >= x3 then
            -- x1 is on the right-most edge
            a1 := x1 + 1;
        end if;
        if x2 >= x1 and then x2 >= x3 then
            -- x2 is on the right-most edge
            a2 := x2 + 1;
        end if;
        if x3 >= x1 and then x3 >= x2 then
            -- x3 is on the right-most edge
            a3 := x3 + 1;
        end if;
        if y1 >= y2 and then y1 >= y3 then
            -- y1 is on the bottom-most edge
            b1 := y1 + 1;
        end if;
        if y2 >= y1 and then y2 >= y3 then
            -- y2 is on the bottom-most edge
            b2 := y2 + 1;
        end if;
        if y3 >= y1 and then y3 >= y2 then
            -- y3 is on the bottom-most edge
            b3 := y3 + 1;
        end if;
        Al_Draw_Triangle( Float(a1), Float(b1),
                          Float(a2), Float(b2),
                          Float(a3), Float(b3),
                          color,
                          1.0 );
    end Triangle;

    ----------------------------------------------------------------------------

    procedure Triangle_Filled( x1, y1 : Integer;
                               x2, y2 : Integer;
                               x3, y3 : Integer;
                               color  : Allegro_Color ) is
        a1 : Integer := x1;
        a2 : Integer := x2;
        a3 : Integer := x3;
        b1 : Integer := y1;
        b2 : Integer := y2;
        b3 : Integer := y3;
    begin
        if x1 >= x2 and then x1 >= x3 then
            -- x1 is on the right-most edge
            a1 := x1 + 1;
        end if;
        if x2 >= x1 and then x2 >= x3 then
            -- x2 is on the right-most edge
            a2 := x2 + 1;
        end if;
        if x3 >= x1 and then x3 >= x2 then
            -- x3 is on the right-most edge
            a3 := x3 + 1;
        end if;
        if y1 >= y2 and then y1 >= y3 then
            -- y1 is on the bottom-most edge
            b1 := y1 + 1;
        end if;
        if y2 >= y1 and then y2 >= y3 then
            -- y2 is on the bottom-most edge
            b2 := y2 + 1;
        end if;
        if y3 >= y1 and then y3 >= y2 then
            -- y3 is on the bottom-most edge
            b3 := y3 + 1;
        end if;
        Al_Draw_Filled_Triangle( Float(a1), Float(b1),
                                 Float(a2), Float(b2),
                                 Float(a3), Float(b3),
                                 color );
    end Triangle_Filled;

    --==========================================================================

    function To_Filter( name    : String;
                        default : Filter_Type := Filter_Type'First
                      ) return Filter_Type is
        filter : Filter_Type;
    begin
        filter := Filter_Type'Value( "Filter_" & name );
        if not filter'Valid then
            return default;
        end if;
        return filter;
    exception
        when others =>
            return default;
    end To_Filter;

    ----------------------------------------------------------------------------

    function To_Name( filter : Filter_Type ) return String is
        result : constant String := Filter_Type'Image( filter );
    begin
        -- "Filter_NAME"
        return result(result'First+7..result'Last);
    end To_Name;

end Drawing;
