--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Exceptions;                    use Ada.Exceptions;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Entities.Players;                  use Entities.Players;
with Resources;                         use Resources;
with Tiles.Libraries;                   use Tiles.Libraries;
with Values.Booleans;                   use Values.Booleans;
with Values.Strings;                    use Values.Strings;
with Worlds;                            use Worlds;

package body Entities.Triggers is

    DEFAULT_WIDTH  : constant := 32;
    DEFAULT_HEIGHT : constant := 32;

    ----------------------------------------------------------------------------

    -- Evaluates the trigger's condition expression and then executes the
    -- trigger's action script if True.
    procedure Conditionally_Execute_Actions( this : not null access Trigger'Class );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    not overriding
    procedure Construct( this    : access Trigger;
                         libName : String;
                         icon    : String ) is
    begin
        Entity(this.all).Construct( DEFAULT_WIDTH, DEFAULT_HEIGHT, libName );
        this.frame := this.lib.Get_Id( icon );
        if this.frame = 0 then
            raise RESOURCE_ERROR with "Image <" & icon & "> not found in " &
                                      "tile library <" & libName & ">";
        end if;
        this.metaphysical := True;
        this.physical := False;
        this.clipped := False;

        -- set the default attributes
        this.attrs.Get.Set( ATTR_ACTION,    Create( "" ) );
        this.attrs.Get.Set( ATTR_CONDITION, Create( "" ) );
        this.attrs.Get.Set( ATTR_WHEN,      Create( "OnActivate" ) );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Trigger ) is
    begin
        Delete( this.evaluator );
        Entity(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Conditionally_Execute_Actions( this : not null access Trigger'Class ) is

        -- Evaluates the trigger's condition expression and returns the result
        -- as a boolean. If it cannot be evaluated, False will be returned.
        function Evaluate_Condition return Boolean is
            condition : constant String_Ptr := As_String( this.attrs.Get.Get( ATTR_CONDITION ) );
        begin
            if condition = Strings.Nul or else condition.Get.Length = 0 then
                return True;
            end if;

            return As_Boolean( this.evaluator.Evaluate_Expression( condition.Get.To_String,
                                                                   A_Evaluation_Node(this) ) ) = True;
        exception
            when e : others =>
                Dbg( To_String( this ) & ": Error in expression: " &
                     Exception_Name( e ) & " - " & Exception_Message( e ),
                     D_EVAL, ERROR );
                return False;
        end Evaluate_Condition;

    begin
        -- verify the trigger's action condition
        if Evaluate_Condition then
            -- execute the action script
            declare
                action : constant String_Ptr := As_String( this.attrs.Get.Get( ATTR_ACTION ) );
            begin
                if action /= Strings.Nul then
                    this.evaluator.Evaluate_Script( action.Get.To_String,
                                                    A_Evaluation_Node(this) );
                end if;
            exception
                when e : others =>
                    -- An error occurred evaluating the action script
                    Dbg( To_String( this ) & ": Error in action: " &
                         Exception_Name( e ) & " - " & Exception_Message( e ),
                         D_EVAL, ERROR );
            end;
        end if;
    end Conditionally_Execute_Actions;

    ----------------------------------------------------------------------------

    overriding
    function Evaluate_Function( this      : access Trigger;
                                name      : String;
                                arguments : Value_Array ) return Value_Ptr is
    begin
        -- the base Trigger doesn't define any script functions
        return this.world.Evaluate_Function( name, arguments );
    end Evaluate_Function;

    ----------------------------------------------------------------------------

    overriding
    function Evaluate_Symbol( this : access Trigger; symbol : String ) return Value_Ptr is
    begin
        -- No symbols exist at the Trigger scope at this time
        return this.world.Evaluate_Symbol( symbol );
    end Evaluate_Symbol;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Trigger ) is
    begin
        Entity'Read( stream, Entity(obj) );
        obj.activatable := Boolean'Input( stream );
        if obj.attrs.Get.Get( ATTR_ACTION ).Get.Get_Type /= V_STRING then
            obj.attrs.Get.Set( ATTR_ACTION, Create( "" ) );
        end if;
        if obj.attrs.Get.Get( ATTR_CONDITION ).Get.Get_Type /= V_STRING then
            obj.attrs.Get.Set( ATTR_CONDITION, Create( "" ) );
        end if;
        if obj.attrs.Get.Get( ATTR_WHEN ).Get.Get_Type /= V_STRING then
            obj.attrs.Get.Set( ATTR_WHEN, Create( "" ) );
        end if;
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Trigger ) is
    begin
        Entity'Write( stream, Entity(obj) );
        Boolean'Output( stream, obj.activatable );
    end Object_Write;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Activate( this      : access Trigger;
                           activator : not null A_Entity ) is
    begin
        if this.activatable and then
           not this.editingMode and then
           activator.all in Player'Class and then
           As_String( this.attrs.Get.Get( ATTR_WHEN ) ) = "OnActivate"
        then
            this.activator := activator;
            this.Conditionally_Execute_Actions;
            this.activator := null;
        end if;
    end On_Activate;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Collide( this : access Trigger; e : not null A_Entity ) is
    begin
        if this.activatable and then
           not this.editingMode and then
           e.all in Player'Class and then
           As_String( this.attrs.Get.Get( ATTR_WHEN ) ) = "OnTouch"
        then
            this.activator := e;
            this.Conditionally_Execute_Actions;
            this.activator := null;
        end if;
    end On_Collide;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Load( this : access Trigger ) is
    begin
        Entity(this.all).On_Load;

        -- reset the activatable flag so that a trigger can be activated again
        -- if it was activated once and then the game session was saved or the
        -- world was stored to memory and later resumed.
        this.activatable := True;

        if this.activatable and then
           not this.editingMode and then
           As_String( this.attrs.Get.Get( ATTR_WHEN ) ) = "OnLoad"
        then
            this.Conditionally_Execute_Actions;
        end if;
    end On_Load;

end Entities.Triggers;
