--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Tags;                          use Ada.Tags;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Entities;                   use Events.Entities;
with GNAT.Regexp;                       use GNAT.Regexp;
with Preferences;                       use Preferences;
with Resources;                         use Resources;
with Support;                           use Support;
with Tiles.Libraries.Manager;           use Tiles.Libraries.Manager;
with Values.Booleans;                   use Values.Booleans;
with Worlds;                            use Worlds;

package body Entities is

    function Unique_Entity_Id return Entity_Id;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    overriding
    function "<"( l, r : Entity_Id ) return Boolean is
    begin
        return Unsigned_64(l) < Unsigned_64(r);
    end "<";

    ----------------------------------------------------------------------------

    function Image( id : Entity_Id ) return String is
        result : constant String := Unsigned_64'Image( Unsigned_64(id) );
    begin
        return result(result'First + 1..result'Last);
    end Image;

    ----------------------------------------------------------------------------

    function To_Entity_Id( val : Id_Ptr ) return Entity_Id is
    begin
        if val /= Ids.Nul then
            return Entity_Id(val.Get.To_Unsigned_64);
        end if;
        return INVALID_ID;
    end To_Entity_Id;

    ----------------------------------------------------------------------------

    function To_Id_Value( id : Entity_Id ) return Id_Ptr is
    begin
        return Create( Unsigned_64(id) );
    end To_Id_Value;

    ----------------------------------------------------------------------------

    function Unique_Entity_Id return Entity_Id is
        pragma Warnings( Off );
        result : Entity_Id := INVALID_ID;
        res64  : Unsigned_64;
                 for res64'Address use result'Address;
        sc     : Seconds_Count;
        ts     : Time_Span;
    begin
        Split( Clock, sc, ts );
        -- to avoid overflowing an integer (if uptime > 49.7 days) when we get
        -- the number of millis in 'ts' by division later, we must make sure
        -- that 'ts' is never more than 2,147,483 (or 2^31) seconds.
        ts := ts + Seconds( Integer(sc mod Seconds_Count(2_147_482)) );

        -- An entity ID is a 64-bit unique ID based on machine up-time and a
        -- random number. The machine uptime is the lowest 32 bits of the number
        -- of milliseconds since boot.
        --
        -- D = | [32 bits] uptime ms | [32 bits] random number |
        res64 := Shift_Left( Unsigned_64(ts / Milliseconds( 1 )), 32 ) + Unsigned_64(Random_32);

        return result;
    end Unique_Entity_Id;

    --==========================================================================

    function Eq( l : Directive_Type; r : Directive_Type ) return Boolean is
    begin
        return l.name = r.name;
    end Eq;

    ----------------------------------------------------------------------------

    function Hash( d : Directive_Type ) return Hash_Type is
    begin
        return Hash( d.name );
    end Hash;

    --==========================================================================

    overriding
    procedure Adjust( this : access Entity ) is
    begin
        raise COPY_NOT_ALLOWED;
    end Adjust;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this    : access Entity;
                         width,
                         height  : Natural;
                         libName : String ) is
    begin
        Object(this.all).Construct;
        this.id := Unique_Entity_Id;
        this.width := width;
        this.height := height;
        this.attrs := Create_Assoc;
        this.editingMode := Get_Pref( "application", "editing_mode" );
        if libName'Length > 0 then
            this.lib := Load_Library( libName, bitmaps => False );
            if this.lib = null then
                raise RESOURCE_ERROR with "Error loading tile library <" & libName & ">";
            end if;
        end if;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Entity ) is
        use Entity_Lists;
        crsr : Entity_Lists.Cursor := this.touching.First;
    begin
        while Has_Element( crsr ) loop
            Element( crsr ).Separated( this'Unrestricted_Access );
            Next( crsr );
        end loop;
        this.touching.Clear;
        this.directives.Clear;
        Unload_Library( this.lib );
        this.attrs := Associations.Nul;
        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Activate( this : not null access Entity'Class; activator : not null A_Entity ) is
    begin
        this.On_Activate( activator );
    end Activate;

    ----------------------------------------------------------------------------

    procedure Cancel_Directives( this : not null access Entity'Class ) is
    begin
        this.directives.Clear;
    end Cancel_Directives;

    ----------------------------------------------------------------------------

    procedure Collided( this : not null access Entity'Class; e : not null A_Entity ) is
    begin
        pragma Debug( this.touching.Contains( e ),
                      Dbg( "Entity " & To_String( this ) & " collided with " &
                           To_String( e ) & " without separation" ) );
        this.touching.Append( e );
        pragma Debug( Dbg( "Entity " & To_String( this ) &
                           " collided with " & To_String( e ), D_PHYSICS, INFO ) );
        this.On_Collide( e );
    end Collided;

    ----------------------------------------------------------------------------

    function Directive( this : not null access Entity'Class;
                        name : Hashed_String ) return Boolean is
    begin
        return this.directives.Contains( Directive_Type'(name, Inactive)  );
    end Directive;

    ----------------------------------------------------------------------------

    function Get_Attributes( this : not null access Entity'Class ) return Assoc_Ptr is
    begin
        return this.attrs;
    end Get_Attributes;

    ----------------------------------------------------------------------------

    function Get_Direction( this : not null access Entity'Class ) return Direction_Type is
    begin
        return this.dir;
    end Get_Direction;

    ----------------------------------------------------------------------------

    function Get_Frame( this : not null access Entity'Class ) return Integer is
    begin
        return this.frame;
    end Get_Frame;

    ----------------------------------------------------------------------------

    function Get_Height( this : not null access Entity'Class ) return Integer is
    begin
        return this.height;
    end Get_Height;

    ----------------------------------------------------------------------------

    function Get_Id( this : not null access Entity'Class ) return Entity_Id is
    begin
        return this.id;
    end Get_Id;

    ----------------------------------------------------------------------------

    function Get_Lib_Name( this : not null access Entity'Class ) return String is
    begin
        if this.lib /= null then
            return this.lib.Get_Name;
        else
            return "";
        end if;
    end Get_Lib_Name;

    ----------------------------------------------------------------------------

    function Get_Width( this : not null access Entity'Class ) return Integer is
    begin
        return this.width;
    end Get_Width;

    ----------------------------------------------------------------------------

    function Get_X( this : not null access Entity'Class ) return Float is
    begin
        return this.x;
    end Get_X;

    ----------------------------------------------------------------------------

    function Get_XV( this : not null access Entity'Class ) return Float is
    begin
        return this.xv;
    end Get_XV;

    ----------------------------------------------------------------------------

    function Get_Y( this : not null access Entity'Class ) return Float is
    begin
        return this.y;
    end Get_Y;

    ----------------------------------------------------------------------------

    function Get_YV( this : not null access Entity'Class ) return Float is
    begin
        return this.yv;
    end Get_YV;

    ----------------------------------------------------------------------------

    procedure Hit_Wall( this         : not null access Entity'Class;
                        dir          : Cardinal_Direction;
                        firstContact : Boolean ) is
    begin
        this.On_Hit_Wall( dir, firstContact );
    end Hit_Wall;

    ----------------------------------------------------------------------------

    function Is_Clipped( this : not null access Entity'Class ) return Boolean is
    begin
        return this.clipped;
    end Is_Clipped;

    ----------------------------------------------------------------------------

    function Is_Metaphysical( this : not null access Entity'Class ) return Boolean is
    begin
        return this.metaphysical;
    end Is_Metaphysical;

    ----------------------------------------------------------------------------

    function Is_Physical( this : not null access Entity'Class ) return Boolean is
    begin
        return this.physical;
    end Is_Physical;

    ----------------------------------------------------------------------------

    procedure Iterate_Touching( this    : not null access Entity'Class;
                                examine : not null access procedure( e : A_Entity ) ) is
        use Entity_Lists;

        procedure Local_Examine( pos : Entity_Lists.Cursor ) is
        begin
            examine( Element( pos ) );
        end Local_Examine;

    begin
        this.touching.Iterate( Local_Examine'Access );
    end Iterate_Touching;

    ----------------------------------------------------------------------------

    not overriding
    procedure On_Load( this : access Entity ) is
    begin
        this.touching.Clear;
    end On_Load;

    ----------------------------------------------------------------------------

    procedure Separated( this : not null access Entity'Class; e : not null A_Entity ) is
        crsr : Entity_Lists.Cursor := this.touching.Find( e );
    begin
        if Entity_Lists.Has_Element( crsr ) then
            pragma Debug( Dbg( "Entity " & To_String( this ) &
                               " separated from " & To_String( e ),
                               D_PHYSICS, INFO ) );
            this.touching.Delete( crsr );
            this.On_Separate( e );
        else
            null;
            pragma Debug( Dbg( "Entity " & To_String( this ) &
                               " separated from " & To_String( e ) &
                               " without collision",
                               D_PHYSICS, INFO ) );
        end if;
    end Separated;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Attribute( this : access Entity;
                             name : String;
                             val  : Value_Ptr'Class ) is
        lname : constant String := To_Lower( name );
    begin
        if lname = "physical" then
            if val.Get.Get_Type = V_BOOLEAN then
                this.physical := As_Boolean( val ).Get.To_Boolean;
            end if;
        elsif lname = "clipped" then
            if val.Get.Get_Type = V_BOOLEAN then
                this.clipped := As_Boolean( val ).Get.To_Boolean;
            end if;
        else
            this.attrs.Get.Set( name, val );
        end if;
        Queue_Entity_Attribute_Changed( this.id, name, val );
    end Set_Attribute;

    ----------------------------------------------------------------------------

    procedure Set_Directive( this      : not null access Entity;
                             directive : Hashed_String;
                             mode      : Directive_Mode ) is
    begin
        pragma Debug( Dbg( "Entity " & To_String( this ) &
                           " directive '" & To_String( directive ) & "' : " &
                           Directive_Mode'Image( mode ),
                           D_ENTITIES, INFO ) );
        case mode is
            when Once | Ongoing =>
                this.directives.Include( Directive_Type'(directive, mode) );
            when Inactive =>
                this.directives.Exclude( Directive_Type'(directive, mode) );
        end case;
    end Set_Directive;

    ----------------------------------------------------------------------------

    procedure Set_Frame( this   : not null access Entity'Class;
                         frame  : Natural;
                         notify : Boolean := True ) is
    begin
        if this.frame /= frame then
            this.frame := frame;
            if notify and then frame > 0 then
                Queue_Frame_Changed( this.id, frame );
            end if;
        end if;
    end Set_Frame;

    ----------------------------------------------------------------------------

    procedure Set_Grounded( this : not null access Entity'Class; grounded : Boolean ) is
    begin
        this.grounded := grounded;
        this.Update_Frame;
    end Set_Grounded;

    ----------------------------------------------------------------------------

    procedure Set_Location( this : not null access Entity'Class; x, y : Float ) is
    begin
        this.x := x;
        this.y := y;
    end Set_Location;

    ----------------------------------------------------------------------------

    procedure Set_Size( this   : not null access Entity'Class;
                        width,
                        height : Natural ) is
    begin
        this.width := width;
        this.height := height;
    end Set_Size;

    ----------------------------------------------------------------------------

    procedure Set_Velocity_X( this : not null access Entity'Class; xv : Float ) is
    begin
        this.xv := xv;
    end Set_Velocity_X;

    ----------------------------------------------------------------------------

    procedure Set_Velocity_Y( this : not null access Entity'Class; yv : Float ) is
    begin
        this.yv := yv;
    end Set_Velocity_Y;

    ----------------------------------------------------------------------------

    procedure Set_World( this  : not null access Entity'Class;
                         world : not null access Worlds.World_Object'Class ) is
    begin
        if this.world /= null then
            raise Constraint_Error with "Entity " & To_String( this ) &
                                        " already has a world reference!";
        end if;
        this.world := world;
    end Set_World;

    ----------------------------------------------------------------------------

    not overriding
    procedure Tick( this : access Entity; time : Tick_Time ) is
    begin
        this.age := this.age + time.elapsed;

        -- let the entity update itself based on its state or directives
        A_Entity(this).Update( time );

        -- clear all non-ongoing directives (Once mode)
        declare
            use Directive_Sets;
            pos : Cursor := this.directives.First;
        begin
            while Has_Element( pos ) loop
                if Element( pos ).mode = Once then
                    this.directives.Delete( pos );
                    pos := this.directives.First;    -- restart the checking
                else
                    pos := Next( pos );
                end if;
            end loop;
        end;

        A_Entity(this).Update_Frame;
    end Tick;

    ----------------------------------------------------------------------------

    overriding
    function To_String( this : access Entity ) return String is
    begin
        return '<' & this.Get_Class_Name & " id=" & Image( this.id ) & '>';
    end To_String;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Entity ) is
    begin
        Object'Read( stream, Object(obj) );
        obj.id := Entity_Id'Input( stream );
        obj.age := Time_Span'Input( stream );
        obj.width := Integer'Input( stream );
        obj.height := Integer'Input( stream );
        obj.x := Float'Input( stream );
        obj.y := Float'Input( stream );
        obj.xv := Float'Input( stream );
        obj.yv := Float'Input( stream );
        obj.dir := Direction_Type'Input( stream );
        obj.frame := Natural'Input( stream );
        obj.attrs := As_Assoc( Value_Ptr'Input( stream ) );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Entity ) is
    begin
        Object'Write( stream, Object(obj) );
        Entity_Id'Output( stream, obj.id );
        Time_Span'Output( stream, obj.age );
        Integer'Output( stream, obj.width );
        Integer'Output( stream, obj.height );
        Float'Output( stream, obj.x );
        Float'Output( stream, obj.y );
        Float'Output( stream, obj.xv );
        Float'Output( stream, obj.yv );
        Direction_Type'Output( stream, obj.dir );
        Natural'Output( stream, obj.frame );
        Value_Ptr'Output( stream, obj.attrs.As_Value );
    end Object_Write;

    --==========================================================================

    function A_Entity_Input( stream : access Root_Stream_Type'Class ) return A_Entity is
        obj : A_Entity := null;
    begin
        if Boolean'Input( stream ) then
            A_Entity'Read( stream, obj );
        end if;
        return obj;
    end A_Entity_Input;

    ----------------------------------------------------------------------------

    procedure A_Entity_Output( stream : access Root_Stream_Type'Class; obj : A_Entity ) is
    begin
        Boolean'Output( stream, obj /= null );
        if obj /= null then
            A_Entity'Write( stream, obj );
        end if;
    end A_Entity_Output;

    ----------------------------------------------------------------------------

    procedure A_Entity_Read( stream : access Root_Stream_Type'Class; obj : out A_Entity ) is
    begin
        obj := new Entity'Class'(Entity'Class'Input( stream ));
    exception
        when e : Tag_Error =>
            declare
                msg : constant String := Exception_Message( e );
            begin
                raise Constraint_Error with "Found unknown entity class " &
                      "<" & To_Lower( msg((Index( msg, ".", Backward ) + 1)..msg'Last) ) & ">";
            end;
    end A_Entity_Read;

    ----------------------------------------------------------------------------

    procedure A_Entity_Write( stream : access Root_Stream_Type'Class; obj : A_Entity ) is
    begin
        Entity'Class'Output( stream, obj.all );
    end A_Entity_Write;

    ----------------------------------------------------------------------------

    function Allocate( id : String ) return A_Entity is
        e : A_Entity;
    begin
        e := Factory.Allocate( id );
        if e = null then
            Dbg( "Error: Attempted to create unrecognized entity " &
                 "class <" & id & ">", D_ENTITIES, ERROR );
        end if;
        return e;
    exception
        when e : RESOURCE_ERROR =>
            Dbg( "Error: Failed to create entity of class <" & id & ">: " &
                 Exception_Message( e ), D_ENTITIES, ERROR );
            return null;
    end Allocate;

    ----------------------------------------------------------------------------

    function Copy( src : A_Entity ) return A_Entity is
    begin
        return A_Entity(Copy( A_Object(src) ));
    end Copy;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Entity ) is
    begin
        Delete( A_Object(this) );
    end Delete;

    ----------------------------------------------------------------------------

    function In_Class( id : String; pattern : String ) return Boolean is
        exp : Regexp;
    begin
        exp := Compile( pattern, Glob => True, Case_Sensitive => False );
        return Match( id, exp );
    end In_Class;

    ----------------------------------------------------------------------------

    procedure Iterate_Classes( pattern : String;
                               examine : access procedure( id : String ) ) is
    begin
        Factory.Iterate_Classes( pattern, examine );
    end Iterate_Classes;

    ----------------------------------------------------------------------------

    function Template( id : String ) return A_Entity renames Factory.Template;

    ----------------------------------------------------------------------------

    procedure Initialize is
    begin
        Factory.Initialize( "entity" );    -- may raise an exception
    end Initialize;

    ----------------------------------------------------------------------------

    procedure Finalize renames Factory.Finalize;

end Entities;
