--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Directions;                        use Directions;
with Events;
with Entities;                          use Entities;
with Hashed_Strings;                    use Hashed_Strings;
with Values;                            use Values;
with Values.Associations;               use Values.Associations;

pragma Elaborate_All( Events );

package Events.Entities is

    -- A command to delete an entity by Entity_Id.
    DELETE_ENTITY_ID : constant Event_Id := To_Event_Id( "Delete_Entity" );

    -- A notification that an entity was deleted.
    ENTITY_DELETED_ID : constant Event_Id := To_Event_Id( "Entity_Deleted" );

    -- A command to the game view follow an entity with the view's viewport.
    FOLLOW_ENTITY_ID : constant Event_Id := To_Event_Id( "Follow_Entity" );

    -- An event involving an entity.
    type Entity_Event is new Event with private;
    type A_Entity_Event is access all Entity_Event'Class;

    -- Returns the Entity_Id of the entity involved in the event.
    function Get_Id( this : not null access Entity_Event'Class ) return Entity_Id;

    ----------------------------------------------------------------------------

    -- A notification that two entities collided (began touching).
    ENTITIES_COLLIDED_ID: constant Event_Id := To_Event_Id( "Entities_Collided" );

    -- A notification that two entities separated (are no longer touching).
    ENTITIES_SEPARATED_ID: constant Event_Id := To_Event_Id( "Entities_Separated" );

    -- An event involving the interaction of two entities.
    type Entities_Event is new Event with private;
    type A_Entities_Event is access all Entities_Event'Class;

    -- Returns the Entity_Id of the first entity involved in the interaction,
    -- deemed entity A.
    function Get_A( this : not null access Entities_Event'Class ) return Entity_Id;

    -- Returns the Entity_Id of the second entity involved in the interaction,
    -- deemed entity B.
    function Get_B( this : not null access Entities_Event'Class ) return Entity_Id;

    ----------------------------------------------------------------------------

    ACCELERATE_ID : constant Event_Id := To_Event_Id( "Accelerate" );

    -- A command to change the acceleration of an entity by Entity_Id.
    type Accelerate_Event is new Entity_Event with private;
    type A_Accelerate_Event is access all Accelerate_Event'Class;

    function Get_Acceleration( this : not null access Accelerate_Event'Class ) return Float;
    pragma Postcondition( Get_Acceleration'Result >= 0.0 );

    function Get_Direction( this : not null access Accelerate_Event'Class ) return Cardinal_Direction;

    function Get_Velocity( this : not null access Accelerate_Event'Class ) return Float;
    pragma Postcondition( Get_Velocity'Result >= 0.0 );

    ----------------------------------------------------------------------------

    -- A notification that an entity attribute has changed.
    ENTITY_ATTRIBUTE_CHANGED_ID : constant Event_Id := To_Event_Id( "Entity_Attribute_Changed" );

    -- A command to set an attribute of an entity.
    SET_ENTITY_ATTRIBUTE_ID : constant Event_Id := To_Event_Id( "Set_Entity_Attribute" );

    -- An event involving an entity attribute.
    type Entity_Attribute_Event is new Entity_Event with private;
    type A_Entity_Attribute_Event is access all Entity_Attribute_Event'Class;

    -- Returns the attribute name.
    function Get_Attribute( this : access Entity_Attribute_Event ) return String;
    pragma Postcondition( Get_Attribute'Result'Length > 0 );

    -- Returns a copy of the attribute value.
    function Get_Value( this : access Entity_Attribute_Event ) return Value_Ptr;
    pragma Postcondition( Get_Value'Result /= Values.Nul );

    ----------------------------------------------------------------------------

    ENTITY_CREATED_ID : constant Event_Id := To_Event_Id( "Entity_Created" );

    -- A notification that a new entity has been created.
    type Entity_Created_Event is new Entity_Event with private;
    type A_Entity_Created_Event is access all Entity_Created_Event'Class;

    -- Returns a copy of the entity's attributes as an association.
    function Get_Attributes( this : not null access Entity_Created_Event'Class ) return Assoc_Ptr;
    pragma Postcondition( Get_Attributes'Result /= Associations.Nul );

    -- Returns the entity's class name.
    function Get_Class( this : not null access Entity_Created_Event'Class ) return String;
    pragma Postcondition( Get_Class'Result'Length > 0 );

    -- Returns the entity's current display frame or 0 for none.
    function Get_Frame( this : not null access Entity_Created_Event'Class ) return Natural;

    -- Returns the physical height of the entity in world units.
    function Get_Height( this : not null access Entity_Created_Event'Class ) return Natural;

    -- Returns the name of the library that contains the entity's frames.
    function Get_Lib_Name( this : not null access Entity_Created_Event'Class ) return String;

    -- Returns the physical width of the entity in world units.
    function Get_Width( this : not null access Entity_Created_Event'Class ) return Natural;

    -- Returns the X coordinate of the entity in world coordinates.
    function Get_X( this : not null access Entity_Created_Event'Class ) return Float;

    -- Returns the X velocity of the entity in world units per second.
    function Get_XV( this : not null access Entity_Created_Event'Class ) return Float;

    -- Returns the Y coordinate of the entity in world coordinates.
    function Get_Y( this : not null access Entity_Created_Event'Class ) return Float;

    -- Returns the Y velocity of the entity in world units per second.
    function Get_YV( this : not null access Entity_Created_Event'Class ) return Float;

    -- Returns True if the entity is clipped to the solid walls in the world.
    function Is_Clipped( this : not null access Entity_Created_Event'Class ) return Boolean;

    -- Returns True if the entity is invisible, existing metaphysically like a
    -- Trigger.
    function Is_Metaphysical( this : not null access Entity_Created_Event'Class ) return Boolean;

    -- Returns True if physics apply to the entity, ie: gravity.
    function Is_Physical( this : not null access Entity_Created_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    -- A notification that part of an entity's bounding rectangle entered a tile.
    ENTITY_ENTERED_TILE_ID : constant Event_Id := To_Event_Id( "Entity_Entered_Tile" );

    -- A notification that no part of an entity's bounding rectangle is now
    -- touching a tile.
    ENTITY_EXITED_TILE_ID : constant Event_Id := To_Event_Id( "Entity_Exited_Tile" );

    -- An event involving the interaction of an entity with a tile.
    type Entity_Tile_Event is new Entity_Event with private;
    type A_Entity_Tile_Event is access all Entity_Tile_Event;

    -- Returns the X location in tile coordinates.
    function Get_X( this : not null access Entity_Tile_Event'Class ) return Natural;

    -- Returns the Y location in tile coordinates.
    function Get_Y( this : not null access Entity_Tile_Event'Class ) return Natural;

    ----------------------------------------------------------------------------

    ENTITY_GROUNDED_ID : constant Event_Id := To_Event_Id( "Entity_Grounded" );

    -- A notification that an entity has landed on/left the ground.
    type Entity_Grounded_Event is new Entity_Event with private;
    type A_Entity_Grounded_Event is access all Entity_Grounded_Event'Class;

    -- Returns True if the entity landed on the ground or False if the entity
    -- left the ground.
    function Is_Grounded( this : not null access Entity_Grounded_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    ENTITY_HIT_WALL_ID : constant Event_Id := To_Event_Id( "Entity_Hit_Wall" );

    -- A notification that a clipped entity collided with a wall and was stopped.
    type Entity_Hit_Wall_Event is new Entity_Event with private;
    type A_Entity_Hit_Wall_Event is access all Entity_Hit_Wall_Event'Class;

    -- Returns the direction that the entity was travelling when it hit the wall.
    function Get_Direction( this : not null access Entity_Hit_Wall_Event'Class ) return Cardinal_Direction;

    -- Returns True if this is the entity's first contact with the wall in that
    -- direction. If the entity hits the wall and continues accelerating into
    -- it, pressing against it, then this will return False until the entity
    -- moves away from the wall and hits it again.
    function Is_First_Contact( this : not null access Entity_Hit_Wall_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    ENTITY_MOVED_ID : constant Event_Id := To_Event_Id( "Entity_Moved" );

    -- A notification that an entity moved.
    type Entity_Moved_Event is new Entity_Event with private;
    type A_Entity_Moved_Event is access all Entity_Moved_Event'Class;

    -- Returns the entity's new X coordinate in world units.
    function Get_X( this : not null access Entity_Moved_Event'Class ) return Float;

    -- Returns the entity's X velocity in world units per second.
    function Get_XV( this : not null access Entity_Moved_Event'Class ) return Float;

    -- Returns the entity's new Y coordinate in world units.
    function Get_Y( this : not null access Entity_Moved_Event'Class ) return Float;

    -- Returns the entity's Y velocity in world units per second.
    function Get_YV( this : not null access Entity_Moved_Event'Class ) return Float;

    ----------------------------------------------------------------------------

    ENTITY_RESIZED_ID : constant Event_Id := To_Event_Id( "Entity_Resized" );

    -- A notification that an entity's physical size changed.
    type Entity_Resized_Event is new Entity_Event with private;
    type A_Entity_Resized_Event is access all Entity_Resized_Event'Class;

    -- Returns the entity's new height in world units.
    function Get_Height( this : not null access Entity_Resized_Event'Class ) return Natural;

    -- Returns the entity's new width in world units.
    function Get_Width( this : not null access Entity_Resized_Event'Class ) return Natural;

    ----------------------------------------------------------------------------

    FRAME_CHANGED_ID : constant Event_Id := To_Event_Id( "Frame_Changed" );

    -- A notification that an entity's display frame has changed.
    type Frame_Changed_Event is new Entity_Event with private;
    type A_Frame_Changed_Event is access all Frame_Changed_Event'Class;

    -- Returns the id of the new display frame or 0 for none.
    function Get_Frame( this : not null access Frame_Changed_Event'Class ) return Natural;

    ----------------------------------------------------------------------------

    ENTITY_DIRECTIVE_ID : constant Event_Id := To_Event_Id( "Entity_Directive" );

    -- A command sent to an entity, either as a one-time action or ongoing until
    -- deactivated. The entity may choose to act on the directive depending on
    -- its current state.
    type Entity_Directive_Event is new Entity_Event with private;
    type A_Entity_Directive_Event is access all Entity_Directive_Event'Class;

    -- Returns the name of the directive being sent to the entity.
    function Get_Directive( this : not null access Entity_Directive_Event'Class ) return Hashed_String;

    -- Returns the name of the directive being sent to the entity.
    function Get_Mode( this : not null access Entity_Directive_Event'Class ) return Directive_Mode;

    ----------------------------------------------------------------------------

    MOVE_ENTITY_ID : constant Event_Id := To_Event_Id( "Move_Entity" );

    -- A command to move an entity.
    type Move_Entity_Event is new Entity_Event with private;
    type A_Move_Entity_Event is access all Move_Entity_Event'Class;

    -- Returns the new X coordinate for the entity in world units.
    function Get_X( this : not null access Move_Entity_Event'Class ) return Float;

    -- Returns the new Y coordinate for the entity in world units.
    function Get_Y( this : not null access Move_Entity_Event'Class ) return Float;

    ----------------------------------------------------------------------------

    RESIZE_ENTITY_ID : constant Event_Id := To_Event_Id( "Resize_Entity" );

    -- A command to change the physical size of an entity.
    type Resize_Entity_Event is new Entity_Event with private;
    type A_Resize_Entity_Event is access all Resize_Entity_Event'Class;

    -- Returns the new height for the entity in world units.
    function Get_Height( this : not null access Resize_Entity_Event'Class ) return Natural;

    -- Returns the new width for the entity in world units.
    function Get_Width( this : not null access Resize_Entity_Event'Class ) return Natural;

    ----------------------------------------------------------------------------

    SPAWN_ENTITY_ID : constant Event_Id := To_Event_Id( "Spawn_Entity" );

    -- A command to create a new entity.
    type Spawn_Entity_Event is new Event with private;
    type A_Spawn_Entity_Event is access all Spawn_Entity_Event;

    -- Returns the physical height for the entity in world units.
    function Get_Height( this : access Spawn_Entity_Event ) return Integer;

    -- Returns class id of the entity to create.
    function Get_Id( this : access Spawn_Entity_Event ) return String;

    -- Returns the physical width for the entity in world units.
    function Get_Width( this : access Spawn_Entity_Event ) return Integer;

    -- Returns the X coordinate for the entity in world units.
    function Get_X( this : access Spawn_Entity_Event ) return Float;

    -- Returns the X velocity for the entity in world units per second.
    function Get_XV( this : access Spawn_Entity_Event ) return Float;

    -- Returns the Y coordinate for the entity in world units.
    function Get_Y( this : access Spawn_Entity_Event ) return Float;

    -- Returns the Y velocity for the entity in world units per second.
    function Get_YV( this : access Spawn_Entity_Event ) return Float;

    ----------------------------------------------------------------------------

    -- Queues an Accelerate event.
    procedure Queue_Accelerate( id  : Entity_Id;
                                dir : Cardinal_Direction;
                                vel : Float;
                                acc : Float );
    pragma Precondition( vel >= 0.0 );
    pragma Precondition( acc >= 0.0 );

    -- Queues a Delete_Entity event.
    procedure Queue_Delete_Entity( id : Entity_Id );

    -- Queues an Entities_Collided event.
    procedure Queue_Entities_Collided( a, b : Entity_Id );

    -- Queues an Entities_Separated event.
    procedure Queue_Entities_Separated( a, b : Entity_Id );

    -- Queues an Entity_Attribute_Changed event.
    procedure Queue_Entity_Attribute_Changed( id        : Entity_Id;
                                              attribute : String;
                                              val       : Value_Ptr'Class );
    pragma Precondition( attribute'Length > 0 );
    pragma Precondition( Value_Ptr(val) /= Values.Nul );

    -- Queues an Entity_Created event. Argument 'attributes' is copied and not
    -- modified.
    procedure Queue_Entity_Created( id           : Entity_Id;
                                    class        : String;
                                    physical,
                                    metaphysical,
                                    clipped      : Boolean;
                                    width,
                                    height       : Natural;
                                    x, y         : Float;
                                    xv, yv       : Float;
                                    libName      : String;
                                    frame        : Natural;
                                    attributes   : Assoc_Ptr );
    pragma Precondition( class'Length > 0 );

    -- Queues an Entity_Deleted event.
    procedure Queue_Entity_Deleted( id : Entity_Id );

    -- Queues an Entity_Entered_Tile event.
    procedure Queue_Entity_Entered_Tile( id : Entity_Id; x, y : Natural );

    -- Queues an Entity_Exited_Tile event.
    procedure Queue_Entity_Exited_Tile( id : Entity_Id; x, y : Natural );

    -- Queues an Entity_Grounded event.
    procedure Queue_Entity_Grounded( id : Entity_Id; grounded : Boolean );

    -- Queues an Entity_Hit_Wall event. Set 'firstContact' True if the entity
    -- bumped into the wall, or False if the entity already made contact with
    -- the wall and is pressing itself against it.
    procedure Queue_Entity_Hit_Wall( id           : Entity_Id;
                                     dir          : Cardinal_Direction;
                                     firstContact : Boolean );

    -- Queues an Entity_Moved event.
    procedure Queue_Entity_Moved( id : Entity_Id; x, y, xv, yv : Float );

    -- Queues an Entity_Resized event.
    procedure Queue_Entity_Resized( id : Entity_Id; width, height : Natural );

    -- Queues an Entity_Directive event to an entity.
    procedure Queue_Entity_Directive( id        : Entity_Id;
                                      directive : Hashed_String;
                                      mode      : Directive_Mode );

    -- Queues a Follow_Entity event.
    procedure Queue_Follow_Entity( id : Entity_Id );

    -- Queues a Frame_Changed event.
    procedure Queue_Frame_Changed( id : Entity_Id; frame : Natural );

    -- Queues a Move_Entity event.
    procedure Queue_Move_Entity( id : Entity_Id; x, y : Float );

    -- Queues a Resize_Entity event.
    procedure Queue_Resize_Entity( id : Entity_Id; width, height : Natural );

    -- Queues a Set_Entity_Attribute event with a value of any type.
    procedure Queue_Set_Entity_Attribute( id        : Entity_Id;
                                          attribute : String;
                                          val       : Value_Ptr'Class );
    pragma Precondition( attribute'Length > 0 );
    pragma Precondition( Value_Ptr(val) /= Values.Nul );

    -- Queues a Spawn_Entity event.
    procedure Queue_Spawn_Entity( id     : String;
                                  x, y   : Float;
                                  width,
                                  height : Natural := 0;
                                  xv, yv : Float := 0.0 );
    pragma Precondition( id'Length > 0 );

private

    type Entity_Event is new Event with
        record
            id : Entity_Id := INVALID_ID;
        end record;

    procedure Construct( this : access Entity_Event; name : String; id : Entity_Id );
    pragma Precondition( name'Length > 0 );

    function To_String( this : access Entity_Event ) return String;

    ----------------------------------------------------------------------------

    type Entities_Event is new Event with
        record
            a, b : Entity_Id := INVALID_ID;
        end record;

    procedure Construct( this : access Entities_Event; name : String; a, b : Entity_Id );
    pragma Precondition( name'Length > 0 );

    function To_String( this : access Entities_Event ) return String;

    ----------------------------------------------------------------------------

    type Accelerate_Event is new Entity_Event with
        record
            dir : Cardinal_Direction := Left;
            vel : Float := 0.0;
            acc : Float := 0.0;
        end record;

    procedure Construct( this : access Accelerate_Event;
                         id   : Entity_Id;
                         dir  : Cardinal_Direction;
                         vel  : Float;
                         acc  : Float );

    ----------------------------------------------------------------------------

    type Entity_Attribute_Event is new Entity_Event with
        record
            attribute : Unbounded_String;
            val       : Value_Ptr;
        end record;

    procedure Adjust( this : access Entity_Attribute_Event );

    -- 'name' is event name
    -- 'attribute' is the name of the attribute
    -- 'val' is the value of the attribute
    procedure Construct( this      : access Entity_Attribute_Event;
                         name      : String;
                         id        : Entity_Id;
                         attribute : String;
                         val       : Value_Ptr'Class );
    pragma Precondition( name'Length > 0 );
    pragma Precondition( attribute'Length > 0 );
    pragma Precondition( Value_Ptr(val) /= Values.Nul );

    procedure Delete( this : in out Entity_Attribute_Event );

    ----------------------------------------------------------------------------

    type Entity_Created_Event is new Entity_Event with
        record
            class        : Unbounded_String;
            physical     : Boolean := True;
            metaphysical : Boolean := False;
            clipped      : Boolean := True;
            width,
            height       : Natural := 0;
            x, y         : Float := 0.0;
            xv, yv       : Float := 0.0;
            libName      : Unbounded_String;
            frame        : Natural := 0;
            attributes   : Assoc_Ptr;
        end record;

    procedure Adjust( this : access Entity_Created_Event );

    procedure Construct( this         : access Entity_Created_Event;
                         id           : Entity_Id;
                         class        : String;
                         physical,
                         metaphysical,
                         clipped      : Boolean;
                         width,
                         height       : Natural;
                         x, y         : Float;
                         xv, yv       : Float;
                         libName      : String;
                         frame        : Natural;
                         attributes   : Assoc_Ptr );

    procedure Delete( this : in out Entity_Created_Event );

    function To_String( this : access Entity_Created_Event ) return String;

    ----------------------------------------------------------------------------

    type Entity_Tile_Event is new Entity_Event with
        record
            x, y : Natural;
        end record;

    procedure Construct( this : access Entity_Tile_Event;
                         name : String;
                         id   : Entity_Id;
                         x, y : Natural );

    ----------------------------------------------------------------------------

    type Entity_Grounded_Event is new Entity_Event with
        record
            grounded : Boolean := False;
        end record;

    procedure Construct( this     : access Entity_Grounded_Event;
                         id       : Entity_Id;
                         grounded : Boolean );

    ----------------------------------------------------------------------------

    type Entity_Hit_Wall_Event is new Entity_Event with
        record
            dir          : Cardinal_Direction;
            firstContact : Boolean;
        end record;

    procedure Construct( this         : access Entity_Hit_Wall_Event;
                         id           : Entity_Id;
                         dir          : Cardinal_Direction;
                         firstContact : Boolean );

    ----------------------------------------------------------------------------

    type Entity_Moved_Event is new Entity_Event with
        record
            x, y   : Float := 0.0;
            xv, yv : Float := 0.0;
        end record;

    procedure Construct( this   : access Entity_Moved_Event;
                         id     : Entity_Id;
                         x, y   : Float;
                         xv, yv : Float );

    ----------------------------------------------------------------------------

    type Entity_Resized_Event is new Entity_Event with
        record
            width, height : Natural := 0;
        end record;

    procedure Construct( this   : access Entity_Resized_Event;
                         id     : Entity_Id;
                         width,
                         height : Natural );

    ----------------------------------------------------------------------------

    type Frame_Changed_Event is new Entity_Event with
        record
            frame : Natural;
        end record;

    procedure Construct( this  : access Frame_Changed_Event;
                         id    : Entity_Id;
                         frame : Natural );

    ----------------------------------------------------------------------------

    type Entity_Directive_Event is new Entity_Event with
        record
            directive : Hashed_String;
            mode      : Directive_Mode := Inactive;
        end record;

    procedure Construct( this      : access Entity_Directive_Event;
                         id        : Entity_Id;
                         directive : Hashed_String;
                         mode      : Directive_Mode );

    function To_String( this : access Entity_Directive_Event ) return String;

    ----------------------------------------------------------------------------

    type Move_Entity_Event is new Entity_Event with
        record
            x, y : Float := 0.0;
        end record;

    procedure Construct( this : access Move_Entity_Event;
                         id   : Entity_Id;
                         x, y : Float );

    ----------------------------------------------------------------------------

    type Resize_Entity_Event is new Entity_Event with
        record
            width, height : Natural := 0;
        end record;

    procedure Construct( this   : access Resize_Entity_Event;
                         id     : Entity_Id;
                         width,
                         height : Natural );

    ----------------------------------------------------------------------------

    type Spawn_Entity_Event is new Event with
        record
            id     : Unbounded_String;
            x, y   : Float;
            width,
            height : Natural;
            xv, yv : Float;
        end record;

    procedure Construct( this   : access Spawn_Entity_Event;
                         id     : String;
                         x, y   : Float;
                         width,
                         height : Natural;
                         xv, yv : Float );
    pragma Precondition( id'Length > 0 );

    function To_String( this : access Spawn_Entity_Event ) return String;

end Events.Entities;
