--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Events;
with Values;                            use Values;

pragma Elaborate_All( Events );

package Events.Game is

    -- A command to end the game session.
    END_GAME_ID : constant Event_Id := To_Event_Id( "End_Game" );

    type End_Game_Event is new Event with private;
    type A_End_Game_Event is access all End_Game_Event'Class;

    function Is_Completed( this : not null access End_Game_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    GAME_PAUSED_ID : constant Event_Id := To_Event_Id( "Game_Paused" );

    -- A notification that the game paused state has changed.
    type Game_Paused_Event is new Event with private;
    type A_Game_Paused_Event is access all Game_Paused_Event'Class;

    -- Returns True if the gameplay has been paused or False if it has been resumed.
    function Is_Paused( this : not null access Game_Paused_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    GAME_STATE_ID : constant Event_Id := To_Event_Id( "Game_State" );

    -- A notification that the gameplay state has changed. This happens when a
    -- game session begins and ends.
    type Game_State_Event is new Event with private;
    type A_Game_State_Event is access all Game_State_Event'Class;

    -- Returns True if gameplay is not in progress now because the game session
    -- was interrupted by the player, instead of ending naturally (winning or
    -- losing.) The game session can be interrupted to start a new game, load a
    -- saved game, or just end the current game session by choice.
    function Is_Interrupted( this : not null access Game_State_Event'Class ) return Boolean;

    -- Returns True if gameplay is in progress.
    function Is_Playing( this : not null access Game_State_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    GAME_VAR_CHANGED_ID : constant Event_Id := To_Event_Id( "Game_Var_Changed" );

    -- A notification that a game session variable has changed.
    type Game_Var_Changed_Event is new Event with private;
    type A_Game_Var_Changed_Event is access all Game_Var_Changed_Event'Class;

    -- Returns a copy of the variable's new value.
    function Get_Value( this : access Game_Var_Changed_Event ) return Value_Ptr;
    pragma Postcondition( Get_Value'Result /= Values.Nul );

    -- Returns the name of the variable that changed.
    function Get_Var( this : not null access Game_Var_Changed_Event'Class ) return String;

    ----------------------------------------------------------------------------

    -- A notification that the loading of a world by the game logic has begun or
    -- ended. If loading has ended, an error message is included on failure.
    LOADING_WORLD_ID : constant Event_Id := To_Event_Id( "Loading_World" );

    -- A notification regarding the loading state of a resource.
    type Loading_Event is new Event with private;
    type A_Loading_Event is access all Loading_Event'Class;

    -- Returns an error message if the loading has ended due to a failure. An
    -- empty string will be returned otherwise.
    function Get_Error( this : not null access Loading_Event'Class ) return String;

    -- Returns True if loading is beginning or False if it has ended. If loading
    -- has ended, check Get_Error to determine if the load was successful.
    function Is_Loading( this : not null access Loading_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    -- A command to begin a new game session.
    NEW_GAME_ID : constant Event_Id := To_Event_Id( "New_Game" );

    ----------------------------------------------------------------------------

    PAUSE_GAME_ID : constant Event_Id := To_Event_Id( "Pause_Game" );

    -- A command to change the game paused state.
    type Pause_Game_Event is new Event with private;
    type A_Pause_Game_Event is access all Pause_Game_Event'Class;

    -- Returns True if the command is to pause the gameplay, or False if the
    -- command is to resume gameplay.
    function Is_Paused( this : not null access Pause_Game_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    -- A notification that the player died.
    PLAYER_DIED_ID : constant Event_Id := To_Event_Id( "Player_Died" );

    ----------------------------------------------------------------------------

    SCROLL_VIEW_ID : constant Event_Id := To_Event_Id( "Scroll_View" );

    -- A command to scroll the gaame view's viewport of the world.
    type Scroll_View_Event is new Event with private;
    type A_Scroll_View_Event is access all Scroll_View_Event'Class;

    -- Amount to scroll view in the X axis in world coordinates.
    function Get_X( this : not null access Scroll_View_Event'Class ) return Float;

    -- Amount to scroll view in the Y axis in world coordinates.
    function Get_Y( this : not null access Scroll_View_Event'Class ) return Float;

    ----------------------------------------------------------------------------

    -- A notification that the view is ready for game play to begin after a new
    -- world has been loaded. After loading a new world, the game waits in the
    -- Loaded state until the view notifies it with this event.
    VIEW_READY_ID : constant Event_Id := To_Event_Id( "View_Ready" );

    ----------------------------------------------------------------------------

    -- Queues an End_Game event. 'completed' indicates the game session ended naturally,
    -- one way or another. If the game was aborted by the player via the menu, completed
    -- must be False.
    procedure Queue_End_Game( completed : Boolean );

    -- Queues a Game_Paused event.
    procedure Queue_Game_Paused( paused : Boolean );

    -- Queues a Game_State event.
    procedure Queue_Game_State( playing : Boolean; interrupted : Boolean := False );

    -- Queues a Game_Var_Changed event.
    procedure Queue_Game_Var_Changed( var : String; val : Value_Ptr'Class );
    pragma Precondition( var'Length > 0 );
    pragma Precondition( Value_Ptr(val) /= Values.Nul );

    -- Queues a Loading_World event.
    procedure Queue_Loading_World( loading : Boolean; error : String := "" );
    pragma Precondition( not loading or else error'Length = 0 );

    -- Queues a New_Game event.
    procedure Queue_New_Game;

    -- Queues a Pause_Game event.
    procedure Queue_Pause_Game( paused : Boolean );

    -- Queues a Player_Died event.
    procedure Queue_Player_Died;

    -- Queues a Scroll_View event.
    procedure Queue_Scroll_View( x, y : Float );

    -- Queues a View_Ready event.
    procedure Queue_View_Ready;

private

    type End_Game_Event is new Event with
        record
            completed : Boolean := True;
        end record;

    procedure Construct( this : access End_Game_Event; completed : Boolean );

    function To_String( this : access End_Game_Event ) return String;

    ----------------------------------------------------------------------------

    type Game_Paused_Event is new Event with
        record
            paused : Boolean := True;
        end record;

    procedure Construct( this : access Game_Paused_Event; paused : Boolean );

    function To_String( this : access Game_Paused_Event ) return String;

    ----------------------------------------------------------------------------

    type Game_State_Event is new Event with
        record
            playing     : Boolean := True;
            interrupted : Boolean := False;
        end record;

    procedure Construct( this        : access Game_State_Event;
                         playing     : Boolean;
                         interrupted : Boolean );

    function To_String( this : access Game_State_Event ) return String;

    ----------------------------------------------------------------------------

    type Game_Var_Changed_Event is new Event with
        record
            var : Unbounded_String;
            val : Value_Ptr;
        end record;

    procedure Adjust( this : access Game_Var_Changed_Event );

    procedure Construct( this : access Game_Var_Changed_Event;
                         var  : String;
                         val  : Value_Ptr'Class );
    pragma Precondition( var'Length > 0 );
    pragma Precondition( Value_Ptr(val) /= Values.Nul );

    function To_String( this : access Game_Var_Changed_Event ) return String;

    ----------------------------------------------------------------------------

    type Loading_Event is new Event with
        record
            loading : Boolean;
            error   : Unbounded_String;
        end record;

    procedure Construct( this    : access Loading_Event;
                         name    : String;
                         loading : Boolean;
                         error   : String );

    function To_String( this : access Loading_Event ) return String;

    ----------------------------------------------------------------------------

    type Pause_Game_Event is new Event with
        record
            paused : Boolean := True;
        end record;

    procedure Construct( this : access Pause_Game_Event; paused : Boolean );

    ----------------------------------------------------------------------------

    type Scroll_View_Event is new Event with
        record
            x, y : Float := 0.0;
        end record;

    procedure Construct( this : access Scroll_View_Event; x, y : Float );

end Events.Game;
