--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Events;
with Keyboard;                          use Keyboard;
with Mouse;                             use Mouse;

pragma Elaborate_All( Events );

package Events.Input is

    -- A notification that a key was pressed.
    KEY_PRESS_ID : constant Event_Id := To_Event_Id( "Key_Press" );

    -- A notification that a key was released.
    KEY_RELEASE_ID : constant Event_Id := To_Event_Id( "Key_Release" );

    -- A notification of a keyboard-based event.
    type Key_Event is new Event with private;
    type A_Key_Event is access all Key_Event'Class;

    -- Returns the Allegro key code of the key involved.
    function Get_Key( this : not null access Key_Event'Class ) return Positive;

    -- Returns a boolean array of modifier keys that were pressed/not pressed at
    -- the time of the event.
    procedure Get_Modifiers( this      : not null access Key_Event'Class;
                             modifiers : out Modifiers_Array );

    -- Returns True if no modifier keys were pressed at the time of the event.
    function No_Modifiers( this : not null access Key_Event'Class ) return Boolean;

    -- Returns True if the Alt key was the only pressed modifier at the time of
    -- the event.
    function Only_Alt( this : not null access Key_Event'Class ) return Boolean;

    -- Returns True if the Ctrl key was the only pressed modifier at the time of
    -- the event.
    function Only_Ctrl( this : not null access Key_Event'Class ) return Boolean;

    -- Returns True if the Shift key was the only pressed modifier at the time
    -- of the event.
    function Only_Shift( this : not null access Key_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    -- A notification that a key was typed or held down for the repeat delay.
    KEY_TYPED_ID : constant Event_Id := To_Event_Id( "Key_Typed" );

    -- A character was typed, or repeated due to holding down the key.
    type Key_Typed_Event is new Key_Event with private;
    type A_Key_Typed_Event is access all Key_Typed_Event'Class;

    -- Returns the character that was typed, or nul if not a visible character.
    function Get_Char( this : not null access Key_Typed_Event'Class ) return Character;

    ----------------------------------------------------------------------------

    -- A notification that the mouse moved.
    MOUSE_MOVE_ID : constant Event_Id := To_Event_Id( "Mouse_Move" );

    -- A notification of a mouse-based event.
    type Mouse_Event is new Event with private;
    type A_Mouse_Event is access all Mouse_Event'Class;

    -- Returns the X location of the mouse at the time of the event.
    function Get_X( this : not null access Mouse_Event'Class ) return Integer;

    -- Returns the Y location of the mouse within the window at the time of the
    -- event.
    function Get_Y( this : not null access Mouse_Event'Class ) return Integer;

    -- Sets the location of the mouse at the time of the event. This can be
    -- called by an event handler to filter the mouse even location for later
    -- mouse event handlers, to make the event location relative to some widget.
    procedure Set_XY( this : not null access Mouse_Event'Class; x, y : Integer );

    ----------------------------------------------------------------------------

    MOUSE_SCROLL_ID : constant Event_Id := To_Event_Id( "Mouse_Scroll" );

    -- A notification that the mouse wheel was scrolled.
    type Mouse_Scroll_Event is new Mouse_Event with private;
    type A_Mouse_Scroll_Event is access all Mouse_Scroll_Event'Class;

    -- Returns the amount that the mouse wheel was scrolled.
    function Get_Amount( this : not null access Mouse_Scroll_Event'Class ) return Integer;
    pragma Postcondition( Get_Amount'Result /= 0 );

    ----------------------------------------------------------------------------

    -- A notification that a mouse button was clicked (pressed and released rapidly).
    MOUSE_CLICK_ID : constant Event_Id := To_Event_Id( "Mouse_Click" );

    -- A notification that a mouse button was double-clicked (clicked twice, rapidly).
    MOUSE_DOUBLECLICK_ID : constant Event_Id := To_Event_Id( "Mouse_Doubleclick" );

    -- A notification that a mouse button was held down for the mouse button
    -- repeat rate delay.
    MOUSE_HELD_ID : constant Event_Id := To_Event_Id( "Mouse_Held" );

    -- A notification that a mouse button was pressed.
    MOUSE_PRESS_ID : constant Event_Id := To_Event_Id( "Mouse_Press" );

    -- A notification that a mouse button was released.
    MOUSE_RELEASE_ID : constant Event_Id := To_Event_Id( "Mouse_Release" );

    -- A notification of a mouse button-based event.
    type Mouse_Button_Event is new Mouse_Event with private;
    type A_Mouse_Button_Event is access all Mouse_Button_Event'Class;

    -- Returns the mouse button involved in the event.
    function Get_Button( this : not null access Mouse_Button_Event'Class ) return Mouse_Button;

    -- Returns a boolean array of keyboard modifiers that were pressed at the
    -- time of the event.
    function Get_Modifiers( this : not null access Mouse_Button_Event'Class ) return Modifiers_Array;

    ----------------------------------------------------------------------------

    -- Queues a Key_Typed event.
    procedure Queue_Key_Typed( key       : Positive;
                               char      : Character;
                               modifiers : Modifiers_Array );

    -- Queues a Key_Press event.
    procedure Queue_Key_Press( key : Positive; modifiers : Modifiers_Array );

    -- Queues a Key_Release event.
    procedure Queue_Key_Release( key : Positive; modifiers : Modifiers_Array );

    -- Creates and returns a Key_Release event without queueing it.
    function Generate_Key_Release( key : Positive; modifiers : Modifiers_Array ) return A_Key_Event;

    -- Queues a Mouse_Click event.
    procedure Queue_Mouse_Click( x, y : Integer; btn : Mouse_Button );

    -- Queues a Mouse_Doubleclick event.
    procedure Queue_Mouse_Doubleclick( x, y : Integer; btn : Mouse_Button );

    -- Queues a Mouse_Held event.
    procedure Queue_Mouse_Held( x, y : Integer; btn : Mouse_Button );

    -- Queues a Mouse_Move event.
    procedure Queue_Mouse_Move( x, y : Integer );

    -- Queues a Mouse_Press event.
    procedure Queue_Mouse_Press( x, y      : Integer;
                                 btn       : Mouse_Button;
                                 modifiers : Modifiers_Array );

    -- Queues a Mouse_Release event.
    procedure Queue_Mouse_Release( x, y : Integer; btn : Mouse_Button );

    -- Creates a Mouse_Release_Event instance without queueing it. This is used
    -- only for a special case in the Window widget. The event is never queued;
    -- it's passed directly to a handler procedure to simulate an event.
    function Create_Mouse_Release( x, y : Integer;
                                   btn  : Mouse_Button ) return A_Mouse_Button_Event;

    -- Queues a Mouse_Scroll event.
    procedure Queue_Mouse_Scroll( x, y, amount : Integer );
    pragma Precondition( amount /= 0 );

private

    type Key_Event is new Event with
        record
            key       : Positive := 1;
            modifiers : Modifiers_Array := (others => False);
        end record;

    procedure Construct( this      : access Key_Event;
                         evtName   : String;
                         key       : Positive;
                         modifiers : Modifiers_Array );

    function To_String( this : access Key_Event ) return String;

    ----------------------------------------------------------------------------

    type Key_Typed_Event is new Key_Event with
        record
            char : Character := ASCII.NUL;
        end record;

    procedure Construct( this      : access Key_Typed_Event;
                         evtName   : String;
                         key       : Positive;
                         char      : Character;
                         modifiers : Modifiers_Array );

    ----------------------------------------------------------------------------

    type Mouse_Event is new Event with
        record
            x, y : Integer := 0;
        end record;

    procedure Construct( this    : access Mouse_Event;
                         evtName : String;
                         x, y    : Integer );

    function To_String( this : access Mouse_Event ) return String;

    ----------------------------------------------------------------------------

    type Mouse_Scroll_Event is new Mouse_Event with
        record
            amount : Integer := 0;
        end record;

    procedure Construct( this    : access Mouse_Scroll_Event;
                         evtName : String;
                         x, y    : Integer;
                         amount  : Integer );
    pragma Precondition( amount /= 0 );

    ----------------------------------------------------------------------------

    type Mouse_Button_Event is new Mouse_Event with
        record
            btn       : Mouse_Button := Mouse_Left;
            modifiers : Modifiers_Array := Modifiers_Array'(others=>False);
        end record;

    procedure Construct( this      : access Mouse_Button_Event;
                         evtName   : String;
                         x, y      : Integer;
                         btn       : Mouse_Button;
                         modifiers : Modifiers_Array );

    function To_String( this : access Mouse_Button_Event ) return String;

end Events.Input;
