--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Containers.Doubly_Linked_Lists;
with Ada.Containers.Indefinite_Hashed_Maps;
with Ada.Strings.Hash_Case_Insensitive;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events;
with Mutable_Lists;

package body Events.Manager is

    package Event_Lists is new Ada.Containers.Doubly_Linked_Lists( A_Event, "=" );
    use Event_Lists;

    protected Event_Queue is

        procedure Add( evt : in out A_Event );

        entry Remove( evt : out A_Event );

        procedure Shutdown;

    private
        empty   : Boolean := True;
        queue   : Event_Lists.List;
        stopped : Boolean := False;
    end Event_Queue;

    ----------------------------------------------------------------------------

    package Corral_Lists is new Mutable_Lists( A_Corral, "=" );

    type Corral_List is
        record
            evtName : Unbounded_String;
            corrals : access Corral_Lists.List := new Corral_Lists.List;
        end record;
    type A_Corral_List is access all Corral_List;

    procedure Construct( this : access Corral_List; evt : String );

    procedure Queue_Event( this : access Corral_List; evt : in out A_Event );

    procedure Register( this : access Corral_List; corral : not null A_Corral );

    procedure Trigger_Event( this     : access Corral_List;
                             evt      : in out A_Event;
                             response : out Response_Type );


    procedure Unregister( this : access Corral_List; corral : not null A_Corral );

    ----------------------------------------------------------------------------

    package Type_Map is new
        Ada.Containers.Indefinite_Hashed_Maps( String,
                                               A_Corral_List,
                                               Ada.Strings.Hash_Case_Insensitive,
                                               "=", "=" );
    use Type_Map;

    protected Corral_Registry is

        -- Returns a reference to the list of corrals for the given event name.
        -- If no corrals have been registered for the name yet, an empty list
        -- is returned. The reference to the corral list that is returned is
        -- owned by the corral registry. Do not delete it.
        procedure Get_List( evtName : String; list : out A_Corral_List );

    private
        registry : Type_Map.Map;
    end Corral_Registry;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    protected body Event_Queue is

        procedure Add( evt : in out A_Event ) is
        begin
            if not stopped then
                pragma Debug( Dbg( "Queuing " & To_String( evt ), D_EVENTS, INFO ) );
                queue.Append( evt );
                evt := null;
                empty := False;
            else
                --pragma Debug( Dbg( "Ignoring queue request for " & To_String( evt ) &
                --                   ", event manager is stopped", D_EVENTS, WARNING ) );
                Delete( evt );
            end if;
        end Add;

        ------------------------------------------------------------------------

        entry Remove( evt : out A_Event ) when not empty or else stopped is
        begin
            if not empty then
                evt := queue.First_Element;
                queue.Delete_First;
                empty := Integer(Length( queue )) = 0;
            else
                evt := null;
            end if;
        end Remove;

        ------------------------------------------------------------------------

        procedure Shutdown is
            evt : A_Event;
        begin
            while not queue.Is_Empty loop
                evt := queue.First_Element;
                queue.Delete_First;
                Delete( evt );
            end loop;
            empty := True;
            stopped := True;
        end Shutdown;

    end Event_Queue;

    --==========================================================================

    procedure Construct( this : access Corral_List; evt : String ) is
    begin
        this.evtName := To_Unbounded_String( evt );
    end Construct;

    ----------------------------------------------------------------------------

    procedure Queue_Event( this : access Corral_List; evt : in out A_Event ) is
        remaining : Natural := 0;

        procedure Send_To_Corral( corral : A_Corral ) is
            dup : A_Event;
        begin
            if remaining > 1 then
                dup := Copy( evt );
                corral.Queue_Event( dup );
            else
                corral.Queue_Event( evt );
            end if;
            remaining := remaining - 1;
        end Send_To_Corral;

    begin
        remaining := this.corrals.Length;
        if remaining > 0 then
            this.corrals.Iterate( Send_To_Corral'Access );
        end if;
        Delete( evt );
    end Queue_Event;

    ----------------------------------------------------------------------------

    procedure Register( this : access Corral_List; corral : not null A_Corral ) is
        inserted : Boolean;
    begin
        this.corrals.Append_No_Duplicate( corral, inserted );
        if not inserted then
            null;
            pragma Debug( Dbg( "Corral <" & corral.Get_Name &
                               "> already registered for event <" &
                               To_String( this.evtName ) & ">", D_EVENTS, WARNING ) );
        end if;
    end Register;

    ----------------------------------------------------------------------------

    procedure Trigger_Event( this     : access Corral_List;
                             evt      : in out A_Event;
                             response : out Response_Type ) is

        procedure Trigger_Corral( corral : A_Corral; quit : in out Boolean ) is
        begin
            --pragma Debug( Dbg( "Triggering " & To_String( evt ) &
            --                   " on corral " & Element( pos ).Get_Name, D_EVENTS, INFO );
            corral.Trigger_Event( evt, response );
            quit := evt = null;
        end Trigger_Corral;

    begin
        this.corrals.Iterate_With_Quit( Trigger_Corral'Access );
    end Trigger_Event;

    ----------------------------------------------------------------------------

    procedure Unregister( this : access Corral_List; corral : not null A_Corral ) is
        crsr : Corral_Lists.Cursor;
    begin
        this.corrals.Find( corral, crsr );
        if Corral_Lists.Has_Element( crsr ) then
            this.corrals.Remove( crsr );
        else
            null;
            pragma Debug( Dbg( "Unable to unregister corral <" &
                               corral.Get_Name & "> for event <" &
                               To_String( this.evtName ) &
                               ">: Corral not registered", D_EVENTS, WARNING ) );
        end if;
    end Unregister;

    --==========================================================================

    protected body Corral_Registry is

        procedure Get_List( evtName : String; list : out A_Corral_List ) is
            crsr : constant Type_Map.Cursor := registry.Find( evtName );
        begin
             if Type_Map.Has_Element( crsr ) then
                 list := Type_Map.Element( crsr );
             else
                 list := new Corral_List;
                 Construct( list, evtName );
                 registry.Insert( evtName, list );
             end if;
        end Get_List;

    end Corral_Registry;

    --==========================================================================

    procedure Queue_Event( evt : in out A_Event ) is
    begin
        Event_Queue.Add( evt );
    end Queue_Event;

    ----------------------------------------------------------------------------

    procedure Register_Listener( corral : not null A_Corral; evtName : String ) is
        list : A_Corral_List;
    begin
        Corral_Registry.Get_List( evtName, list );
        Register( list, corral );
    end Register_Listener;

    ----------------------------------------------------------------------------

    procedure Run is
        evt  : A_Event;
        list : A_Corral_List;
    begin
        loop
            Event_Queue.Remove( evt );
            exit when evt = null;
            Corral_Registry.Get_List( evt.Get_Name, list );
            Queue_Event( list, evt );
        end loop;
    end Run;

    ----------------------------------------------------------------------------

    procedure Stop is
    begin
        Event_Queue.Shutdown;
    end Stop;

    ----------------------------------------------------------------------------

    procedure Trigger_Event( evt : in out A_Event ) is
        response : Response_Type;
        list     : A_Corral_List;
    begin
        Corral_Registry.Get_List( evt.Get_Name, list );
        pragma Debug( Dbg( "Triggering " & To_String( evt ), D_EVENTS, INFO ) );
        Trigger_Event( list, evt, response );
    end Trigger_Event;

    ----------------------------------------------------------------------------

    procedure Trigger_Event( evt : in out A_Event; response : out Response_Type ) is
        list : A_Corral_List;
    begin
        Corral_Registry.Get_List( evt.Get_Name, list );
        pragma Debug( Dbg( "Triggering " & To_String( evt ), D_EVENTS, INFO ) );
        Trigger_Event( list, evt, response );
    end Trigger_Event;

    ----------------------------------------------------------------------------

    procedure Unregister_Listener( corral : not null A_Corral; evtName : String ) is
        list : A_Corral_List;
    begin
        Corral_Registry.Get_List( evtName, list );
        Unregister( list, corral );
    end Unregister_Listener;

end Events.Manager;
