--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Events;
with Maps;                              use Maps;
with Values;                            use Values;

pragma Elaborate_All( Events );

package Events.World is

    CREATE_WORLD_ID : constant Event_Id := To_Event_Id( "Create_World" );

    -- A command to create a new, empty world.
    type Create_World_Event is new Event with private;
    type A_Create_World_Event is access all Create_World_Event'Class;

    -- Returns the domain of the world to create (map, platform, etc.)
    function Get_Domain( this : not null access Create_World_Event'Class ) return String;
    pragma Postcondition( Get_Domain'Result'Length > 0 );

    -- Returns the height of the new world in tiles.
    function Get_Height( this : not null access Create_World_Event'Class ) return Positive;

    -- Returns the name of the library to be used for tiles.
    function Get_Library_Name( this : not null access Create_World_Event'Class ) return String;
    pragma Postcondition( Get_Library_Name'Result'Length > 0 );

    -- Returns the width of the new world in tiles.
    function Get_Width( this : not null access Create_World_Event'Class ) return Positive;

    ----------------------------------------------------------------------------

    LOAD_WORLD_ID : constant Event_Id := To_Event_Id( "Load_World" );

    -- A command to load a world from disk.
    type Load_World_Event is new Event with private;
    type A_Load_World_Event is access all Load_World_Event'Class;

    -- Returns the filename of the world to load.
    function Get_Filename( this : not null access Load_World_Event'Class ) return String;

    ----------------------------------------------------------------------------

    WORLD_LOADED_ID : constant Event_Id := To_Event_Id( "World_Loaded" );

    -- A notification that a world has been loaded/created/resized.
    type World_Loaded_Event is new Event with private;
    type A_World_Loaded_Event is access all World_Loaded_Event'Class;

    -- Returns the height of the world in tiles.
    function Get_Height( this : not null access World_Loaded_Event'Class ) return Positive;

    -- Returns the world map as an array of tile layers.
    function Get_Layers( this : not null access World_Loaded_Event'Class ) return A_Layer_Array;
    pragma Postcondition( Get_Layers'Result /= null );

    -- Returns the name of the library used for tiles.
    function Get_Library_Name( this : not null access World_Loaded_Event'Class ) return String;
    pragma Postcondition( Get_Library_Name'Result'Length > 0 );

    -- Returns the width of a tile in the world in pixels.
    function Get_Tile_Width( this : not null access World_Loaded_Event'Class ) return Positive;

    -- Returns the width of the world in tiles.
    function Get_Width( this : not null access World_Loaded_Event'Class ) return Positive;

    ----------------------------------------------------------------------------

    -- A command to set the value of a world property.
    SET_WORLD_PROPERTY_ID : constant Event_Id := To_Event_Id( "Set_World_Property" );

    -- A notification that a world property changed.
    WORLD_PROPERTY_CHANGED_ID : constant Event_Id := To_Event_Id( "World_Property_Changed" );

    -- An event involving a world property value.
    type World_Property_Event is new Event with private;
    type A_World_Property_Event is access all World_Property_Event'Class;

    -- Returns the name of the property.
    function Get_Property_Name( this : not null access World_Property_Event'Class ) return String;
    pragma Postcondition( Get_Property_Name'Result'Length > 0 );

    -- Returns the property's value.
    function Get_Value( this : not null access World_Property_Event'Class ) return Value_Ptr;
    pragma Postcondition( Get_Value'Result /= Values.Nul );

    ----------------------------------------------------------------------------

    RESIZE_WORLD_ID : constant Event_Id := To_Event_Id( "Resize_World" );

    -- A command to resize the world map.
    type Resize_World_Event is new Event with private;
    type A_Resize_World_Event is access all Resize_World_Event'Class;

    -- Returns the new height of the map in tiles.
    function Get_Height( this : not null access Resize_World_Event'Class ) return Positive;

    -- Returns the new width of the map in tiles.
    function Get_Width( this : not null access Resize_World_Event'Class ) return Positive;

    ----------------------------------------------------------------------------

    SET_TILE_ID : constant Event_Id := To_Event_Id( "Set_Tile" );

    -- A command to set the tile id at a specific location.
    type Set_Tile_Event is new Event with private;
    type A_Set_Tile_Event is access all Set_Tile_Event'Class;

    -- Returns the layer of the tile to set.
    function Get_Layer( this : not null access Set_Tile_Event'Class ) return Integer;

    -- Returns new tile id to set.
    function Get_Tile_ID( this : not null access Set_Tile_Event'Class ) return Natural;

    -- Returns the X coordinate of the tile to set.
    function Get_X( this : not null access Set_Tile_Event'Class ) return Float;

    -- Returns the Y coordinate of the tile to set.
    function Get_Y( this : not null access Set_Tile_Event'Class ) return Float;

    ----------------------------------------------------------------------------

    TILE_CHANGED_ID : constant Event_Id := To_Event_Id( "Tile_Changed" );

    -- A notification that a tile in the map changed.
    type Tile_Changed_Event is new Event with private;
    type A_Tile_Changed_Event is access all Tile_Changed_Event'Class;

    -- Returns the map layer of the tile that changed.
    function Get_Layer( this : not null access Tile_Changed_Event'Class ) return Integer;

    -- Returns the new tile id.
    function Get_Tile_ID( this : not null access Tile_Changed_Event'Class ) return Natural;

    -- Returns the X coordinate of the tile that changed.
    function Get_X( this : not null access Tile_Changed_Event'Class ) return Natural;

    -- Returns the Y coordinate of the tile that changed.
    function Get_Y( this : not null access Tile_Changed_Event'Class ) return Natural;

    ----------------------------------------------------------------------------

    WORLD_MODIFIED_ID : constant Event_Id := To_Event_Id( "World_Modified" );

    -- A notification that the world has been modified somehow.
    type World_Modified_Event is new Event with private;
    type A_World_Modified_Event is access all World_Modified_Event'Class;

    ----------------------------------------------------------------------------

    -- Creates a new, empty world synchronously. An exception is raised on error.
    procedure Trigger_Create_World( width,
                                    height : Positive;
                                    libName,
                                    domain : String );
    pragma Precondition( libName'Length > 0 );
    pragma Precondition( domain'Length > 0 );

    -- Load a world from disk asynchronously.
    procedure Queue_Load_World( filename : String );

    -- Loads a world from disk synchronously. An exception is raised on error.
    procedure Trigger_Load_World( filename : String );
    pragma Precondition( filename'Length > 0 );

    -- A world has been loaded into the Game, by loading from disk, by creating
    -- a new world, etc.
    procedure Queue_World_Loaded( width,
                                  height,
                                  tileWidth : Positive;
                                  layers    : not null A_Layer_Array;
                                  libName   : String );
    pragma Precondition( libName'Length > 0 );

    -- Resizes the world sychronously. An exception is raised on failure.
    -- (ie: parameters rejected)
    procedure Trigger_Resize_World( width, height : Positive );

    -- Set a tile in the world by location (layer, x, y).
    procedure Queue_Set_Tile( layer   : Integer;
                              x, y    : Float;      -- world pixel coordinates
                              tile_id : Natural );

    -- Asynchronously sets a world property 'name' to 'value'. The value will be
    -- copied.
    procedure Queue_Set_World_Property( name : String; value : Value_Ptr'Class );
    pragma Precondition( name'Length > 0 );

    -- A tile in the world has changed.
    procedure Queue_Tile_Changed( layer   : Integer;
                                  x, y    : Natural;
                                  tile_id : Natural );

    -- Something about the world has changed. This is an indication that the
    -- world has changed since it was loaded from disk.
    procedure Queue_World_Modified;

    -- Notifies listeners that the property 'name' of the world has changed. The
    -- property's new value is 'value'. The value will be copied.
    procedure Queue_World_Property_Changed( name : String; value : Value_Ptr'Class );
    pragma Precondition( name'Length > 0 );

private

    type Create_World_Event is new Event with
        record
            width,
            height  : Positive := 1;
            libName : Unbounded_String;
            domain  : Unbounded_String;
        end record;

    procedure Construct( this    : access Create_World_Event;
                         width,
                         height  : Positive;
                         libName,
                         domain  : String );
    pragma Precondition( libName'Length > 0 );
    pragma Precondition( domain'Length > 0 );

    ----------------------------------------------------------------------------

    type Load_World_Event is new Event with
        record
            filename : Unbounded_String;
        end record;

    procedure Construct( this : access Load_World_Event; filename : String );

    function To_String( this : access Load_World_Event ) return String;

    ----------------------------------------------------------------------------

    type World_Loaded_Event is new Event with
        record
            width,
            height,
            tileWidth : Positive := 1;
            layers    : A_Layer_Array := null;
            libName   : Unbounded_String;
        end record;

    procedure Adjust( this : access World_Loaded_Event );

    procedure Construct( this      : access World_Loaded_Event;
                         width,
                         height,
                         tileWidth : Positive;
                         layers    : not null A_Layer_Array;
                         libName   : String );
    pragma Precondition( libName'Length > 0 );

    procedure Delete( this : in out World_Loaded_Event );

    ----------------------------------------------------------------------------

    type World_Property_Event is new Event with
        record
            name  : Unbounded_String;
            value : Value_Ptr;
        end record;

    procedure Construct( this    : access World_Property_Event;
                         evtName : String;
                         name    : String;
                         value   : Value_Ptr'Class );
    pragma Precondition( name'Length > 0 );

    function To_String( this : access World_Property_Event ) return String;

    ----------------------------------------------------------------------------

    type Resize_World_Event is new Event with
        record
            width,
            height : Positive := 1;
        end record;

    procedure Construct( this   : access Resize_World_Event;
                         width,
                         height : Positive );

    ----------------------------------------------------------------------------

    type Set_Tile_Event is new Event with
        record
            layer   : Integer := 0;
            x, y    : Float := 0.0;
            tile_id : Natural := 0;
        end record;

    procedure Construct( this    : access Set_Tile_Event;
                         layer   : Integer;
                         x, y    : Float;
                         tile_id : Natural );

    ----------------------------------------------------------------------------

    type Tile_Changed_Event is new Event with
        record
            layer   : Integer := 0;
            x, y    : Natural := 0;
            tile_id : Natural := 0;
        end record;

    procedure Construct( this    : access Tile_Changed_Event;
                         layer   : Integer;
                         x, y    : Natural;
                         tile_id : Natural );

    ----------------------------------------------------------------------------

    type World_Modified_Event is new Event with null record;

end Events.World;
