--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Objects;                           use Objects;
with Statuses;                          use Statuses;

private with Ada.Containers;
private with Ada.Strings.Unbounded;
private with Ada.Real_Time;

package Events is

    type Event_Id is private;

    function "="( l, r : Event_Id ) return Boolean;

    function To_Event_Id( evtName : String ) return Event_Id;
    pragma Precondition( evtName'Length > 0 );

    type Event_Type is private;

    function "="( l, r : Event_Type ) return Boolean;

    ----------------------------------------------------------------------------

    -- A Response_Type respresents a response to an event by an event listener.
    type Response_Type is private;

    -- Returns the message string returned by the event listener in it's response.
    function Get_Message( response : Response_Type ) return String;

    -- Returns the status returned by the event listener in it's response.
    function Get_Status( response : Response_Type ) return Status;

    -- Sets the Response's status and message. This is to be called by an event
    -- listener returning a response.
    procedure Set_Response( response : in out Response_Type;
                            stat     : Status;
                            msg      : String := "" );

    -- The default no-response with a status of NONE and no message.
    No_Response : constant Response_Type;

    ----------------------------------------------------------------------------

    -- An Event object indicates an event/change has taken place and it contains
    -- all of the information required to describe that change.
    type Event is new Object with private;
    type A_Event is access all Event'Class;

    -- Returns the name of the event which ocurred. All event instances of the
    -- same class share the same name.
    function Get_Name( this : not null access Event'Class ) return String;
    pragma Postcondition( Get_Name'Result'Length > 0 );

    -- Returns the event identifier. This is a hash of the event's name for
    -- quick event type comparison. All event instances of the same class share
    -- the same event identifier. However, in the case of a hash collision, it
    -- is not correct to assume that instances of one event class will not have
    -- the same id as instances of another event class with a different name.
    function Get_Id( this : not null access Event'Class ) return Event_Id;

    -- Returns a deep copy of Event 'src'. All events are required to be
    -- copyable.
    function Copy( src : A_Event ) return A_Event;
    pragma Postcondition( Copy'Result /= src or else src = null );

    -- Deletes the Event.
    procedure Delete( this : in out A_Event );
    pragma Postcondition( this = null );

private

    use Ada.Real_Time;
    use Ada.Strings.Unbounded;

    type Event_Id is new Ada.Containers.Hash_Type;

    type Event_Type is
        record
            id   : Event_Id := Event_Id'First;
            name : Unbounded_String;
        end record;

    type Response_Type is
        record
            stat : Status := ST_NONE;
            msg  : Unbounded_String;
        end record;

    No_Response : constant Response_Type := Response_Type'(others => <>);

    ----------------------------------------------------------------------------

    type Event is new Object with
        record
            eType : Event_Type;
            eTime : Time := Time_First;
        end record;

    procedure Construct( this : access Event; name : String );
    pragma Precondition( name'Length > 0 );

    function To_String( this : access Event ) return String;

end Events;
