--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

--with Debugging;                         use Debugging;
with Allegro.Fonts.TTF;                 use Allegro.Fonts.TTF;

package body Fonts is

    -- Loads the Allegro_Font object from the object's resource file if it has
    -- not already been loaded. This is a part of a workaround for Allegro 5.0's
    -- limitation of loading fonts only on the drawing thread. It should be
    -- fixed by Allegro 5.1 in the future. A RESOURCE_ERROR will be raised on
    -- error (ex: font file format error.)
    procedure Load_From_Resource( this : not null access Font'Class );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Load_Font( filename : String; size : Positive ) return A_Font is
        this : A_Font := new Font;
    begin
        Font(this.all).Construct( filename, size );
        return this;
    exception
        when others =>
            Delete( this );
            raise RESOURCE_ERROR with "Error loading font <" & filename & ">";
    end Load_Font;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this     : access Font;
                         filename : String;
                         size     : Positive ) is
    begin
        Limited_Object(this.all).Construct;
        this.size := size;
        this.res := Load_Resource( filename, "resources", keep => True );
        if this.res = null then
            raise RESOURCE_ERROR with "Error loading font <" & filename & '>';
        end if;

        -- this is probably not the drawing thread, but load it anyway so the
        -- font size functions will work. the font will be loaded again before
        -- actual drawing occurs.
        this.Load_From_Resource;     -- may raise RESOURCE_ERROR
    exception
        when others =>
            Release( this.res );
            raise;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Font ) is
    begin
        Al_Destroy_Font( this.ptr );
        Release( this.res );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Load_From_Resource( this : not null access Font'Class ) is
    begin
        if this.ptr = null then
            this.ptr := Al_Load_TTF_Font_f( this.res.Get_File,
                                            this.res.Get_Filename,
                                            this.size, 0 );
            if this.ptr = null then
                raise RESOURCE_ERROR with "Bad font file <" & this.res.Get_Filename & '>';
            end if;
        end if;
    end Load_From_Resource;

    ----------------------------------------------------------------------------

    procedure Draw_String( this  : not null access Font'Class;
                           str   : String;
                           x, y  : Float;
                           color : Allegro_Color ) is
    begin
        if not this.drawn then
            -- if this font has not been drawn yet, load it again on this thread
            -- that is doing the drawing. this works around a limitation in
            -- Allegro 5.0 that only supports image loading on the rendering
            -- thread.
            Al_Destroy_Font( this.ptr );
            this.Load_From_Resource;    -- may raise RESOURCE_ERROR
            this.drawn := True;
        end if;
        Al_Draw_Text( this.ptr, color, x, y, ALLEGRO_ALIGN_LEFT, str );
    end Draw_String;

    ----------------------------------------------------------------------------

    function Text_Height( this : not null access Font'Class ) return Positive is
    begin
        return Al_Get_Font_Line_Height( this.ptr );
    end Text_Height;

    ----------------------------------------------------------------------------

    function Text_Length( this : not null access Font'Class;
                          str  : String ) return Natural is
    begin
        if str'Length = 0 then
            return 0;
        end if;
        return Al_Get_Text_Width( this.ptr, str );
    end Text_Length;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Font ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Fonts;
