--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Containers;                    use Ada.Containers;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Applications;                      use Applications;
with Applications.Gui.Games;            use Applications.Gui.Games;
with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Application;                use Events.Application;
with Events.Game;                       use Events.Game;
with Events.World;                      use Events.World;
with Preferences;                       use Preferences;
with Support;                           use Support;
with Themes;

package body Game_Views is

    allocate : Allocator := null;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Game_View( display : not null A_Allegro_Display ) return A_Game_View is
    begin
        if allocate /= null then
            -- set the application's display active on this thread temporarily
            -- so any bitmaps loaded by the view will be associated with it. we
            -- will set it back again after creating the Game_View. the view's
            -- thread will take the display later when it is initialized.
            declare
                oldTarget : constant A_Allegro_Bitmap := Al_Get_Target_Bitmap;
                view      : A_Game_View;
            begin
                Al_Set_Target_Backbuffer( display );
                view := allocate.all( display );
                Al_Set_Target_Bitmap( oldTarget );
                return view;
            exception
                when others =>
                    Al_Set_Target_Bitmap( oldTarget );
                    raise;
            end;
        else
            raise Constraint_Error with "Game_View allocator not registered";
        end if;
    end Create_Game_View;

    ----------------------------------------------------------------------------

    procedure Attach( this    : not null access Game_View'Class;
                      process : not null A_Process ) is
    begin
        this.pman.Attach_Async( process );
    end Attach;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this          : access Game_View;
                         display       : not null A_Allegro_Display;
                         width, height : Natural ) is
    begin
        Limited_Object(this.all).Construct;
        this.display := display;
        this.width := width;
        this.height := height;

        this.queue := Al_Create_Event_Queue;
        Al_Register_Event_Source( this.queue, Al_Get_Display_Event_Source( this.display ) );

        this.corral := Create_Corral( "game view" );
        this.pman := Create_Process_Manager( "view", Constrain( Get_Pref( "application", "max_render_fps" ), 0, 120 ) );

        this.audioPlayer := Create_Audio_Player( this.corral );
        this.inhandler := Create_Input_Handler( A_Game_Application(Get_Application).Is_Mouse_Enabled );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Game_View ) is
    begin
        Delete( this.renderer );
        Delete( this.win );

        Delete( this.inhandler );
        Delete( this.audioPlayer );

        Delete( this.pman );
        Delete( this.corral );

        Al_Destroy_Event_Queue( this.queue );

        declare
            pos : Widget_Registry.Cursor := this.widgets.First;
        begin
            pragma Debug( this.widgets.Length > 0,
                          Dbg( "Game_Views.Delete Warning: " &
                               Image( Integer(this.widgets.Length) ) &
                               " widgets were abandoned...",
                               D_VIEW, WARNING ) );
            while Has_Element( pos ) loop
                pragma Debug( Dbg( "  Widget ID = '" & Element( pos ).Get_Id & "'",
                                   D_VIEW, INFO ) );
                pos := Next( pos );
            end loop;
        end;

        Themes.Delete_Theme;

        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Detach( this    : not null access Game_View'Class;
                      process : not null A_Process ) is
    begin
        this.pman.Detach_Async( process );
    end Detach;

    ----------------------------------------------------------------------------

    procedure End_Game( this : not null access Game_View'Class ) is
    begin
        if this.gameInSession then
            Queue_End_Game( completed => False );
        end if;
    end End_Game;

    ----------------------------------------------------------------------------

    procedure Finalize( this : not null access Game_View'Class ) is
    begin
        if this.initialized and then not this.finalized then
            pragma Debug( Dbg( "Finalizing game view", D_VIEW, INFO ) );

            this.On_Finalize;

            this.Detach( A_Process(this.audioPlayer) );
            this.Detach( A_Process(this.renderer) );
            this.Detach( A_Process(this.corral) );
            this.Detach( A_Process(this) );

            this.pman.Stop;
            this.inhandler.Stop;

            this.corral.Remove_Listener( A_Event_Listener(this), "World_Property_Changed" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Loading_World" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Game_State" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Game_Paused" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Exit_Application" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Close_Window" );

            this.finalized := True;
        end if;
    end Finalize;

    ----------------------------------------------------------------------------

    function Get_Corral( this : not null access Game_View'Class ) return A_Corral is
    begin
        return this.corral;
    end Get_Corral;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access Game_View ) return String is
        pragma Unreferenced( this );
    begin
        return "Game View";
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    function Get_Display( this : not null access Game_View'Class ) return A_Allegro_Display is
    begin
        return this.display;
    end Get_Display;

    ----------------------------------------------------------------------------

    function Get_Widget( this : not null access Game_View'Class;
                         id   : String ) return A_Widget is
        pos : Widget_Registry.Cursor;
    begin
        pos := this.widgets.Find( id );
        if Has_Element( pos ) then
            return Element( pos );
        else
            if id'Length > 0 then
                pragma Debug( Dbg( "Warning: Widget id '" & id & "' not found", D_GUI, WARNING ) );
                raise ID_NOT_FOUND with "Widget not found, id: " & id;
            else
                raise ID_NOT_FOUND with "Widget not found, no id given";
            end if;
        end if;
    end Get_Widget;

    ----------------------------------------------------------------------------

    function Get_Window( this : not null access Game_View'Class ) return A_Window is
    begin
        return this.win;
    end Get_Window;

    ----------------------------------------------------------------------------

    procedure Handle_Close_Window( this : not null access Game_View'Class ) is
        close : Boolean := True;
    begin
        A_Game_View(this).On_Close_Window( close );
        if close then
            Queue_Exit_Application;
        end if;
    end Handle_Close_Window;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Game_View;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
        id : constant Event_Id := evt.Get_Id;
    begin
        if evt.Get_Id = WORLD_PROPERTY_CHANGED_ID then
            A_Game_View(this).On_World_Property_Changed( A_World_Property_Event(evt).Get_Property_Name,
                                                         A_World_Property_Event(evt).Get_Value );

        elsif id = GAME_PAUSED_ID then
            if this.paused /= A_Game_Paused_Event(evt).Is_Paused then
                this.paused := A_Game_Paused_Event(evt).Is_Paused;
                A_Game_View(this).On_Game_Paused( this.paused );
            end if;

        elsif evt.Get_Id = GAME_STATE_ID then
            this.gameInSession := A_Game_State_Event(evt).Is_Playing;
            if this.gameInSession then
                -- a new game is beginning
                this.paused := False;
            end if;
            A_Game_View(this).On_Game_State_Changed( A_Game_State_Event(evt).Is_Playing,
                                                     A_Game_State_Event(evt).Is_Interrupted );

        elsif id = LOADING_WORLD_ID then
            if A_Loading_Event(evt).Is_Loading then
                A_Game_View(this).On_Loading_Begin;
            else
                A_Game_View(this).On_Loading_End( success => A_Loading_Event(evt).Get_Error'Length = 0,
                                                  message => A_Loading_Event(evt).Get_Error );
            end if;

        elsif id = CLOSE_WINDOW_ID then
            A_Game_View(this).Handle_Close_Window;

        elsif id = EXIT_APPLICATION_ID then
            A_Game_Application(Get_Application).Stop;

        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    procedure Initialize( this : not null access Game_View'Class ) is
        winWidth, winHeight : Natural;
    begin
        if not this.initialized then
            pragma Debug( Dbg( "Initializing game view", D_VIEW, INFO ) );
            Al_Set_Target_Backbuffer( this.display );

            Themes.Create_Theme;

            if this.width > 0 and then this.height > 0 then
                winWidth := this.width;
                winHeight := this.height;
            else
                winWidth := Al_Get_Display_Width( this.display );
                winHeight := Al_Get_Display_Height( this.display );
            end if;

            this.win := Create_Window( this, "win", winWidth, winHeight,
                                       To_Filter( Get_Pref( "scaling_filter" ), Filter_xBR ) );
            this.renderer := Create_Renderer( this.display,
                                              this.win,
                                              Constrain( Get_Pref( "application",
                                                                   "max_render_fps" ),
                                                         1, 60 ) );

            this.corral.Add_Listener( A_Event_Listener(this), "Close_Window" );
            this.corral.Add_Listener( A_Event_Listener(this), "Exit_Application" );
            this.corral.Add_Listener( A_Event_Listener(this), "Game_Paused" );
            this.corral.Add_Listener( A_Event_Listener(this), "Game_State" );
            this.corral.Add_Listener( A_Event_Listener(this), "Loading_World" );
            this.corral.Add_Listener( A_Event_Listener(this), "World_Property_Changed" );

            this.Attach( A_Process(this) );
            this.Attach( A_Process(this.corral) );
            this.Attach( A_Process(this.renderer) );
            this.Attach( A_Process(this.audioPlayer) );

            A_Game_View(this).Initialize_Widgets( this.win );

            -- release the display; it will be taken by the view's process manager
            Al_Set_Target_Backbuffer( null );

            this.inhandler.Start;
            this.pman.Start;

            this.On_Initialize;

            this.initialized := True;
        end if;
    end Initialize;

    ----------------------------------------------------------------------------

    function Is_Game_In_Session( this : not null access Game_View'Class ) return Boolean is
    begin
        return this.gameInSession;
    end Is_Game_In_Session;

    ----------------------------------------------------------------------------

    function Is_Paused( this : not null access Game_View'Class ) return Boolean is
    begin
        return this.gameInSession and then this.paused;
    end Is_Paused;

    ----------------------------------------------------------------------------

    procedure New_Game( this : not null access Game_View'Class ) is
    begin
        if this.gameInSession then
            this.End_Game;
        end if;
        Queue_New_Game;
    end New_Game;

    ----------------------------------------------------------------------------

    procedure Pause_Game( this  : not null access Game_View'Class;
                          pause : Boolean ) is
    begin
        if this.gameInSession then
            Queue_Pause_Game( pause );

            -- don't change .paused here, this is only where the command is
            -- issued. the state is changed only by the Game_Paused event.
        end if;
    end Pause_Game;

    ----------------------------------------------------------------------------

    procedure Quit( this : not null access Game_View'Class ) is
        pragma Unreferenced( this );
    begin
        Queue_Exit_Application;
    end Quit;

    ----------------------------------------------------------------------------

    procedure Ready_For_Play( this : not null access Game_View'Class ) is
    begin
        if this.gameInSession then
            Queue_View_Ready;
        end if;
    end Ready_For_Play;

    ----------------------------------------------------------------------------

    procedure Register( this   : not null access Game_View'Class;
                        widget : not null A_Widget ) is
        id       : constant String := widget.Get_Id;
        pos      : Widget_Registry.Cursor;
        inserted : Boolean;
    begin
        this.widgets.Insert( widget.Get_Id, widget, pos, inserted );
        if not inserted then
            raise DUPLICATE_ID with "Duplicate widget id: " & id ;
        end if;
    end Register;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Game_View; time : Tick_Time ) is
        pragma Unreferenced( time );
        event : aliased Allegro_Event;
    begin
        if not this.acquiredDisplay then
            -- Immediately attempt to set the display active on this thread
            -- (the game view process manager's thread)
            Al_Set_Target_Backbuffer( this.display );
            this.acquiredDisplay := True;
        end if;

        while Al_Get_Next_Event( this.queue, event'Access ) loop
            case event.any.typ is
                when ALLEGRO_EVENT_DISPLAY_CLOSE      => Queue_Close_Window;
                when ALLEGRO_EVENT_DISPLAY_SWITCH_IN  => Queue_App_Focus;
                when ALLEGRO_EVENT_DISPLAY_SWITCH_OUT => Queue_App_Blur;
                when others =>
                    null;  -- display events that aren't supported
            end case;
        end loop;
    end Tick;

    ----------------------------------------------------------------------------

    procedure Unregister( this : not null access Game_View'Class; id : String ) is
        pos : Widget_Registry.Cursor := this.widgets.Find( id );
    begin
        if Has_Element( pos ) then
            this.widgets.Delete( pos );
        end if;
    end Unregister;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Game_View ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    procedure Register_Allocator( allocate : not null Allocator ) is
    begin
        Game_Views.allocate := allocate;
    end Register_Allocator;

end Game_Views;
