--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Exceptions;                    use Ada.Exceptions;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Game;                       use Events.Game;
with Events.World;                      use Events.World;
with Games.Sessions;                    use Games.Sessions;
with Statuses;                          use Statuses;

package body Games.Sessions is

    allocate : Allocator := null;

    ----------------------------------------------------------------------------

    procedure Start_Gameplay( this : not null access Game_Session'Class );

    procedure Stop_Gameplay( this : not null access Game_Session'Class );

    -- Notifies the game session that the view has finished loading its
    -- resources after a world change and it is now ready for game play.
    procedure View_Ready( this : not null access Game_Session'Class );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    -- package private
    function Create_Game_Session( game : not null A_Game ) return A_Game_Session is
    begin
        if allocate /= null then
            return allocate.all( game );
        else
            raise Constraint_Error with "Game_Session allocator not registered";
        end if;
    end Create_Game_Session;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Game_Session;
                         game : not null A_Game ) is
    begin
        Object(this.all).Construct;
        this.game := game;
        this.vars := Create_Assoc;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Game_Session ) is
        use Worlds_Cache;
        crsr  : Worlds_Cache.Cursor;
        world : A_World;
    begin
        this.End_Game( completed => False );    -- may have already ended

        -- delete all the cached worlds
        crsr := this.worlds.First;
        while Has_Element( crsr ) loop
            world := Element( crsr );
            Delete( world );
            Next( crsr );
        end loop;
        this.worlds.Clear;

        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure End_Game( this      : not null access Game_Session'Class;
                        completed : Boolean ) is
    begin
        if this.state = Ended then
            return;
        end if;

        pragma Debug( Dbg( "Game session: End Game", D_GAME, INFO ) );

        -- notify the View that the game session ended
        Queue_Game_State( playing => False, interrupted => not completed );

        -- allow the subclass to handle end-game actions
        this.On_End_Game( completed );

        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "View_Ready" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Pause_Game" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Load_World" );

        -- stop gameplay and deactivate the active world
        if this.state = Playing then
            this.Stop_Gameplay;
        end if;
        if this.state /= Initial then
            this.activeWorld.Detach_From_Framework;
        end if;
        this.activeWorld := null;

        this.state := Ended;
    end End_Game;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access Game_Session ) return String is
        pragma Unreferenced( this );
    begin
        return "Game Session";
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    overriding
    function Get_Var( this : access Game_Session;
                      name : String ) return Value_Ptr is
        value : constant Value_Ptr := this.vars.Get.Get( name );
    begin
        if value.Get.Get_Type = V_NULL then
            Dbg( "Error: Game session var """ & name & """ not found", D_EVAL, ERROR );
        end if;
        return value;
    end Get_Var;

    ----------------------------------------------------------------------------

    function Get_World( this : not null access Game_Session'Class ) return A_World is
    begin
        return this.activeWorld;
    end Get_World;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Game_Session;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
    begin
        if this.state = Ended then
            -- this shouldn't happen because the object should by unregistered
            -- as an event listener before the state changes to Ended.
            return;
        end if;

        if evt.Get_Id = LOAD_WORLD_ID then
            begin
                if this.state = Loading then
                    raise Constraint_Error with "Loading already in progress";
                else
                    -- may raise exception
                    A_Game_Session(this).On_Load_World( A_Load_World_Event(evt).Get_Filename );
                    Set_Response( resp, ST_SUCCESS );
                end if;
            exception
                when e : others =>
                    -- Loading the world from disk failed. The current state of
                    -- the game session remains unchanged.
                    Set_Response( resp, ST_FAILED, Exception_Message( e ) );
            end;

        elsif evt.Get_Id = PAUSE_GAME_ID then
            if (this.state = Playing and then A_Pause_Game_Event(evt).Is_Paused) or else
               (this.state = Paused and then not A_Pause_Game_Event(evt).Is_Paused)
            then
                A_Game_Session(this).On_Pause( A_Pause_Game_Event(evt).Is_Paused );
            end if;

        elsif evt.Get_Id = VIEW_READY_ID then
            this.View_Ready;

        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    procedure Load_World( this     : not null access Game_Session'Class;
                          filename : String;
                          fresh    : Boolean ) is
        use Worlds_Cache;
        crsr  : Worlds_Cache.Cursor;
        world : A_World;
    begin
        case this.state is
            when Loading =>
                raise Constraint_Error with "Load already in progress";

            when Playing =>
                -- stop gameplay, preventing the passage of game time. gameplay
                -- will resume if loading fails, or when the view notifies us
                -- that it is ready.
                this.Stop_Gameplay;

            when Initial | Paused | Ended =>
                null;
        end case;

        -- loading of the world has begun
        Queue_Loading_World( True );

        -- load the world from disk or retrieve its current state
        crsr := this.worlds.Find( filename );
        if Has_Element( crsr ) and then not fresh then
            -- use the existing world
            world := Element( crsr );
            pragma Debug( Dbg( "Game session: Activating cached world <""" &
                               filename & """>", D_GAME, INFO ) );
        else
            -- need to load the world from disk; this may raise an exception
            pragma Debug( Dbg( "Game session: Loading world <""" & filename &
                               """> from disk", D_GAME, INFO ) );
            world := Worlds.Load_World( filename );
        end if;

        -- NO MORE EXCEPTIONS BEYOND THIS POINT --
        -- any failure that may occur MUST occur before the current world is
        -- deactivated, so that gameplay may resume if loading fails.

        -- deactivate the old world
        if this.activeWorld /= null then
            this.activeWorld.Detach_From_Framework;
            pragma Debug( Dbg( "Game session: Deactivated world <" &
                               Image( this.activeWorld.Get_Property( "filename" ) ) & ">",
                               D_GAME, INFO ) );
        end if;

        -- activate the new world (view begins loading too)
        this.activeWorld := world;
        this.activeWorld.Attach_To_Framework( A_Game_State(this), this.game.corral );
        pragma Debug( Dbg( "Game session: Activated world <" &
                           Image( this.activeWorld.Get_Property( "filename" ) ) & ">",
                           D_GAME, INFO ) );

        -- update the cache
        if fresh and then Has_Element( crsr ) then
            -- replace the old world
            world := Element( crsr );
            Delete( world );
            this.worlds.Replace_Element( crsr, this.activeWorld );
            pragma Debug( Dbg( "Game session: Replaced world <""" & filename &
                               """> in cache", D_GAME, INFO ) );
        elsif not Has_Element( crsr ) then
            -- insert the new world
            this.worlds.Insert( filename, this.activeWorld );
            pragma Debug( Dbg( "Game session: Stored world <""" & filename &
                               """> in cache", D_GAME, INFO ) );
        end if;

        -- loading the world has successfully completed
        Queue_Loading_World( False );

        -- enter the Loading state to wait for the View_Ready event
        this.state := Loading;
    exception
        when e : others =>
            -- loading the world has ended on failure
            Queue_Loading_World( False, error => Exception_Message( e ) );

            -- resume gameplay if the session was in play before the load
            if this.state = Playing then
                this.Start_Gameplay;
            end if;
            raise;
    end Load_World;

    ----------------------------------------------------------------------------

    procedure Load_World( this     : not null access Game_Session'Class;
                          filename : String;
                          world    : in out A_World ) is
        use Worlds_Cache;
        crsr  : Worlds_Cache.Cursor;
    begin
        case this.state is
            when Loading =>
                raise Constraint_Error with "Load already in progress";

            when Playing =>
                -- stop gameplay, preventing the passage of game time. gameplay
                -- will resume if loading fails, or when the view notifies us
                -- that it is ready.
                this.Stop_Gameplay;

            when Initial | Paused | Ended =>
                null;
        end case;

        -- deactivate the old world
        if this.activeWorld /= null then
            this.activeWorld.Detach_From_Framework;
            pragma Debug( Dbg( "Game session: Deactivated world <" &
                               Image( this.activeWorld.Get_Property( "filename" ) ) & ">",
                               D_GAME, INFO ) );
        end if;

        -- activate the new world (view begins loading too)
        this.activeWorld := world;
        world := null;
        this.activeWorld.Attach_To_Framework( A_Game_State(this), this.game.corral );
        pragma Debug( Dbg( "Game session: Activated world <" &
                           Image( this.activeWorld.Get_Property( "filename" ) ) & ">",
                           D_GAME, INFO ) );

        -- update the cache
        crsr := this.worlds.Find( filename );
        if Has_Element( crsr ) then
            -- replace the old world
            world := Element( crsr );
            Delete( world );
            this.worlds.Replace_Element( crsr, this.activeWorld );
            pragma Debug( Dbg( "Game session: Replaced world <""" & filename &
                               """> in cache", D_GAME, INFO ) );
        elsif not Has_Element( crsr ) then
            -- insert the new world
            this.worlds.Insert( filename, this.activeWorld );
            pragma Debug( Dbg( "Game session: Stored world <""" & filename &
                               """> in cache", D_GAME, INFO ) );
        end if;

        this.state := Loading;
    end Load_World;

    ----------------------------------------------------------------------------

    procedure New_Game( this : not null access Game_Session'Class ) is
    begin
        -- notify the View that the game session began
        Queue_Game_State( playing => True );

        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Load_World" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Pause_Game" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "View_Ready" );

        pragma Debug( Dbg( "Game session: New game", D_GAME, INFO ) );

        -- allow the subclass to setup the initial game state
        A_Game_Session(this).On_New_Game;
    end New_Game;

    ----------------------------------------------------------------------------

    not overriding
    procedure On_Load_World( this : access Game_Session; filename : String ) is
    begin
        this.Load_World( filename, fresh => True );
    end On_Load_World;

    ----------------------------------------------------------------------------

    not overriding
    procedure On_Pause( this : access Game_Session; enabled : Boolean ) is
    begin
        this.Pause( enabled );
    end On_Pause;

    ----------------------------------------------------------------------------

    procedure Pause( this    : not null access Game_Session'Class;
                     enabled : Boolean ) is
    begin
        case this.state is
            when Playing =>
                -- pause the game play
                if enabled then
                    this.Stop_Gameplay;
                    Queue_Game_Paused( enabled );
                    pragma Debug( Dbg( "Game session: Gameplay paused", D_GAME, INFO ) );

                    this.state := Paused;
                end if;

            when Paused =>
                -- resume the game play
                if not enabled then
                    this.Start_Gameplay;
                    Queue_Game_Paused( enabled );
                    pragma Debug( Dbg( "Game session: Gameplay resumed", D_GAME, INFO ) );

                    this.state := Playing;
                end if;

            when others => null;
        end case;
    end Pause;

    ----------------------------------------------------------------------------

    overriding
    procedure Set_Var( this  : access Game_Session;
                       name  : String;
                       value : Value_Ptr'Class ) is
    begin
        pragma Debug( Dbg( "Game session: Set " & name & " := " & Image( value ),
                           D_GAME, INFO ) );
        this.vars.Get.Set( name, value );
        Queue_Game_Var_Changed( name, value );
    end Set_Var;

    ----------------------------------------------------------------------------

    -- package private
    procedure Start_Gameplay( this : not null access Game_Session'Class ) is
    begin
        this.game.pman.Attach_Async( A_Process(this) );
        this.game.pman.Attach_Async( A_Process(this.game.physics) );
        this.game.pman.Attach_Async( A_Process(this.activeWorld) );
    end Start_Gameplay;

    ----------------------------------------------------------------------------

    -- package private
    procedure Stop_Gameplay( this : not null access Game_Session'Class ) is
    begin
        this.game.pman.Detach_Async( A_Process(this.activeWorld) );
        this.game.pman.Detach_Async( A_Process(this.game.physics) );
        this.game.pman.Detach_Async( A_Process(this) );
    end Stop_Gameplay;

    ----------------------------------------------------------------------------

    procedure View_Ready( this : not null access Game_Session'Class ) is
    begin
        case this.state is
            when Loading =>
                this.Start_Gameplay;
                this.state := Playing;

            when others =>
                Dbg( "Warning: Unexpected View_Ready in game session", D_GAME, WARNING );
        end case;
    end View_Ready;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Game_Session ) is
    begin
        Delete( A_Object(this) );
    end Delete;

    --==========================================================================

    procedure Register_Allocator( allocate : not null Allocator ) is
    begin
        Games.Sessions.allocate := allocate;
    end Register_Allocator;

end Games.Sessions;
