--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Game;                       use Events.Game;
with Games.Sessions;                    use Games.Sessions;
with Preferences;                       use Preferences;
with Support;                           use Support;

package body Games is

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Game return A_Game is
        this : constant A_Game := new Game;
    begin
        this.Construct;
        return this;
    end Create_Game;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Game ) is
    begin
        Limited_Object(this.all).Construct;
        this.corral := Create_Corral( "game logic" );
        this.physics := Create_Physics( this.corral );
        this.pman := Create_Process_Manager( "game",
                                             Constrain( Get_Pref( "application",
                                                                  "max_logic_fps" ),
                                                        0, 120 ), 15 );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Game ) is
    begin
        if this.initialized and then not this.finalized then
            Dbg( "Error: Game object not finalized", D_GAME, ERROR );
        end if;

        -- Delete other owned objects
        Delete( this.pman );
        Delete( this.physics );
        Delete( this.corral );

        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Finalize( this : not null access Game'Class ) is
    begin
        if this.initialized and then not this.finalized then
            this.finalized := True;
            Delete( A_Game_Session(this.session) );

            this.pman.Detach_Async( A_Process(this.corral) );

            this.corral.Remove_Listener( A_Event_Listener(this), "New_Game" );
            this.corral.Remove_Listener( A_Event_Listener(this), "End_Game" );

            this.pman.Stop;
        end if;
    end Finalize;

    ----------------------------------------------------------------------------

    function Get_Corral( this : not null access Game'Class ) return A_Corral is
    begin
        return this.corral;
    end Get_Corral;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Game;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        if evt.Get_Id = END_GAME_ID then
            if this.session /= null then
                this.session.End_Game( A_End_Game_Event(evt).Is_Completed );
                Delete( A_Game_Session(this.session) );
            end if;

        elsif evt.Get_Id = NEW_GAME_ID then
            if this.session = null then
                this.session := Create_Game_Session( A_Game(this) );
                this.session.New_Game;
            else
                Dbg( "Error: Game session in progress", D_GAME, ERROR );
            end if;

        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    procedure Initialize( this : not null access Game'Class ) is
    begin
        if not this.initialized then
            this.initialized := True;
            this.corral.Add_Listener( A_Event_Listener(this), "End_Game" );
            this.corral.Add_Listener( A_Event_Listener(this), "New_Game" );

            this.pman.Attach_Async( A_Process(this.corral) );
            this.pman.Start;
        end if;
    end Initialize;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Game ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Games;
