--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Task_Identification;           use Ada.Task_Identification;
with Ada.Unchecked_Deallocation;
with Allegro;                           use Allegro;
with Allegro.Events;                    use Allegro.Events;
with Allegro.Keyboard;                  use Allegro.Keyboard;
with Allegro.Mouse;                     use Allegro.Mouse;
with Allegro.Time;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Input;                      use Events.Input;
with Interfaces;                        use Interfaces;
with Keyboard;                          use Keyboard;
with Mouse;                             use Mouse;

package body Input_Handlers is

    UNKNOWN : constant := 16#7FFFFFFF#;

    ----------------------------------------------------------------------------

    function Is_Visible_Keycode( keycode : Integer ) return Boolean is
    begin
        case keycode is
            when ALLEGRO_KEY_A |
                 ALLEGRO_KEY_B |
                 ALLEGRO_KEY_C |
                 ALLEGRO_KEY_D |
                 ALLEGRO_KEY_E |
                 ALLEGRO_KEY_F |
                 ALLEGRO_KEY_G |
                 ALLEGRO_KEY_H |
                 ALLEGRO_KEY_I |
                 ALLEGRO_KEY_J |
                 ALLEGRO_KEY_K |
                 ALLEGRO_KEY_L |
                 ALLEGRO_KEY_M |
                 ALLEGRO_KEY_N |
                 ALLEGRO_KEY_O |
                 ALLEGRO_KEY_P |
                 ALLEGRO_KEY_Q |
                 ALLEGRO_KEY_R |
                 ALLEGRO_KEY_S |
                 ALLEGRO_KEY_T |
                 ALLEGRO_KEY_U |
                 ALLEGRO_KEY_V |
                 ALLEGRO_KEY_W |
                 ALLEGRO_KEY_X |
                 ALLEGRO_KEY_Y |
                 ALLEGRO_KEY_Z |
                 ALLEGRO_KEY_0 |
                 ALLEGRO_KEY_1 |
                 ALLEGRO_KEY_2 |
                 ALLEGRO_KEY_3 |
                 ALLEGRO_KEY_4 |
                 ALLEGRO_KEY_5 |
                 ALLEGRO_KEY_6 |
                 ALLEGRO_KEY_7 |
                 ALLEGRO_KEY_8 |
                 ALLEGRO_KEY_9 |
                 ALLEGRO_KEY_PAD_0 |
                 ALLEGRO_KEY_PAD_1 |
                 ALLEGRO_KEY_PAD_2 |
                 ALLEGRO_KEY_PAD_3 |
                 ALLEGRO_KEY_PAD_4 |
                 ALLEGRO_KEY_PAD_5 |
                 ALLEGRO_KEY_PAD_6 |
                 ALLEGRO_KEY_PAD_7 |
                 ALLEGRO_KEY_PAD_8 |
                 ALLEGRO_KEY_PAD_9 |
                 ALLEGRO_KEY_ESCAPE |
                 ALLEGRO_KEY_TILDE |
                 ALLEGRO_KEY_MINUS |
                 ALLEGRO_KEY_EQUALS |
                 ALLEGRO_KEY_OPENBRACE |
                 ALLEGRO_KEY_CLOSEBRACE |
                 ALLEGRO_KEY_SEMICOLON |
                 ALLEGRO_KEY_QUOTE |
                 ALLEGRO_KEY_BACKSLASH |
                 ALLEGRO_KEY_BACKSLASH2 |
                 ALLEGRO_KEY_COMMA |
                 ALLEGRO_KEY_FULLSTOP |
                 ALLEGRO_KEY_SLASH |
                 ALLEGRO_KEY_SPACE |
                 ALLEGRO_KEY_PAD_SLASH |
                 ALLEGRO_KEY_PAD_ASTERISK |
                 ALLEGRO_KEY_PAD_MINUS |
                 ALLEGRO_KEY_PAD_PLUS |
                 ALLEGRO_KEY_AT |
                 ALLEGRO_KEY_CIRCUMFLEX |
                 ALLEGRO_KEY_COLON2 |
                 ALLEGRO_KEY_PAD_EQUALS |
                 ALLEGRO_KEY_BACKQUOTE |
                 ALLEGRO_KEY_SEMICOLON2 =>
                     return True;
             when others =>
                 return False;
         end case;
    end Is_Visible_Keycode;

    ----------------------------------------------------------------------------

    type Click_Count is mod 2;

    type MB_State is
        record
            pressed   : Boolean := False;
            pressTime : Ada.Real_Time.Time := Clock;
            pressX,
            pressY    : Integer := UNKNOWN;
            lastHeld  : Ada.Real_Time.Time := Clock;
            clickTime : Ada.Real_Time.Time := Clock;
            clickX,
            clickY    : Integer := UNKNOWN;
            clicks    : Click_Count := 0;
        end record;

    type MB_States is array (Mouse_Button) of MB_State;

    ----------------------------------------------------------------------------

    -- allowed mouse drift during click
    CLICK_DRIFT : constant := 10;

    -- delay between press and release
    CLICK_DELAY : constant Time_Span := Milliseconds( 400 );

    -- delay between 1st click and 2nd click
    DOUBLECLICK_DELAY : constant Time_Span := Milliseconds( 600 );

    -- delay between mouse held events
    HELD_REPEAT_DELAY : constant Time_Span := Milliseconds( 1000 / 25 );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Input_Handler( mouseEnabled : Boolean ) return A_Input_Handler is
        this : constant A_Input_Handler := new Input_Handler;
    begin
        this.Construct( mouseEnabled );
        return this;
    end Create_Input_Handler;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this         : access Input_Handler;
                         mouseEnabled : Boolean ) is
    begin
        Limited_Object(this.all).Construct;
        this.mouseEnabled := mouseEnabled;
        this.process := new Input_Task;
        this.process.Init( A_Input_Handler(this) );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Input_Handler ) is
    begin
        this.Stop;
        Delete( this.process );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Start( this : not null access Input_Handler'Class ) is
    begin
        if not this.started and then not this.stopped then
            this.started := True;
            this.process.Start;
        end if;
    end Start;

    ----------------------------------------------------------------------------

    procedure Stop( this : not null access Input_Handler'Class ) is
    begin
        if not this.stopped then
            this.stopped := True;
            this.process.Stop;

            -- wait one second for the process to shut down
            for count in 1..10 loop
                if Is_Terminated( this.process'Identity ) then
                    exit;
                end if;
                delay 0.1;
            end loop;
            if not Is_Terminated( this.process'Identity ) then
                Dbg( "Input_Handler.Stop: Input_Task failed to stop", D_INIT, ERROR );
            end if;
        end if;
    exception
        when e : others =>
            Dbg( "Input_Handler.Stop Exception: ...", D_INIT, ERROR );
            Dbg( e, D_INIT, ERROR );
    end Stop;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Input_Handler ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    task body Input_Task is
        mbtn : MB_States;

        ------------------------------------------------------------------------

        -- Processes all waiting input events.
        procedure Process_Events( queue : A_Allegro_Event_Queue ) is
            start     : constant Ada.Real_Time.Time := Clock;
            now       : Ada.Real_Time.Time;
            kbstate   : Allegro_Keyboard_State;
            modifiers : Modifiers_Array;
            mstate    : Allegro_Mouse_State;
            event     : aliased Allegro_Event;
            timeout   : Allegro.Time.Allegro_Timeout;
        begin
            -- process all events waiting since the last input frame
            Allegro.Time.Al_Init_Timeout( timeout, 0.1 );
            while Al_Wait_For_Event_Until( queue, event'Access, timeout ) loop
                now := Clock;

                -- set the modifiers for any input events, using current kb state
                Al_Get_Keyboard_State( kbstate );
                modifiers(ALT) := Al_Key_Down( kbstate, ALLEGRO_KEY_ALT ) or else
                                  Al_Key_Down( kbstate, ALLEGRO_KEY_ALTGR );
                modifiers(CTRL) := Al_Key_Down( kbstate, ALLEGRO_KEY_LCTRL ) or else
                                   Al_Key_Down( kbstate, ALLEGRO_KEY_RCTRL );
                modifiers(SHIFT) := Al_Key_Down( kbstate, ALLEGRO_KEY_LSHIFT ) or else
                                    Al_Key_Down( kbstate, ALLEGRO_KEY_RSHIFT );
                modifiers(CMD) := Al_Key_Down( kbstate, ALLEGRO_KEY_LWIN ) or else
                                  Al_Key_Down( kbstate, ALLEGRO_KEY_RWIN ) or else
                                  Al_Key_Down( kbstate, ALLEGRO_KEY_COMMAND );

                case event.any.typ is
                    when ALLEGRO_EVENT_KEY_DOWN =>
                        Queue_Key_Press( event.keyboard.keycode, modifiers );

                    when ALLEGRO_EVENT_KEY_UP =>
                        Queue_Key_Release( event.keyboard.keycode, modifiers );

                    when ALLEGRO_EVENT_KEY_CHAR =>
                        -- a character was typed on the keyboard or auto-repeated
                        -- key char events come with modifiers included
                        modifiers(ALT) := (event.keyboard.modifiers and (KEYMOD_ALT or KEYMOD_ALTGR)) /= 0;
                        modifiers(CTRL) := (event.keyboard.modifiers and KEYMOD_CTRL) /= 0;
                        modifiers(SHIFT) := (event.keyboard.modifiers and KEYMOD_SHIFT) /= 0;
                        modifiers(CMD) := (event.keyboard.modifiers and (KEYMOD_LWIN or KEYMOD_RWIN or KEYMOD_COMMAND)) /= 0;
                        if Is_Visible_Keycode( event.keyboard.keycode ) then
                            Queue_Key_Typed( event.keyboard.keycode,
                                             Character'Val(event.keyboard.unichar),
                                             modifiers );
                        else
                            Queue_Key_Typed( event.keyboard.keycode, ASCII.NUL, modifiers );
                        end if;

                    when ALLEGRO_EVENT_MOUSE_AXES =>
                        if event.mouse.dz /= 0 then
                            Queue_Mouse_Scroll( event.mouse.x, event.mouse.y, event.mouse.dz );
                        end if;
                        if event.mouse.dx /= 0 or else event.mouse.dy /= 0 then
                            Queue_Mouse_Move( event.mouse.x, event.mouse.y );
                        end if;

                    when ALLEGRO_EVENT_MOUSE_BUTTON_DOWN =>
                        if event.mouse.button = 1 then
                            -- event: left mouse press
                            Queue_Mouse_Press( event.mouse.x, event.mouse.y, Mouse_Left, modifiers );
                            mbtn(Mouse_Left).pressed := True;
                            mbtn(Mouse_Left).pressTime := now;
                            mbtn(Mouse_Left).pressX := event.mouse.x;
                            mbtn(Mouse_Left).pressY := event.mouse.y;
                            mbtn(Mouse_Left).lastHeld := now;
                        elsif event.mouse.button = 2 then
                            -- event: right mouse press
                            Queue_Mouse_Press( event.mouse.x, event.mouse.y, Mouse_Right, modifiers );
                            mbtn(Mouse_Right).pressed := True;
                            mbtn(Mouse_Right).pressTime := now;
                            mbtn(Mouse_Right).pressX := event.mouse.x;
                            mbtn(Mouse_Right).pressY := event.mouse.y;
                            mbtn(Mouse_Right).lastHeld := now;
                        elsif event.mouse.button = 3 then
                            -- event: middle mouse press
                            Queue_Mouse_Press( event.mouse.x, event.mouse.y, Mouse_Middle, modifiers );
                            mbtn(Mouse_Middle).pressed := True;
                            mbtn(Mouse_Middle).pressTime := now;
                            mbtn(Mouse_Middle).pressX := event.mouse.x;
                            mbtn(Mouse_Middle).pressY := event.mouse.y;
                            mbtn(Mouse_Middle).lastHeld := now;
                        end if;

                    when ALLEGRO_EVENT_MOUSE_BUTTON_UP =>
                        if event.mouse.button = 1 then
                            -- event: left mouse release
                            Queue_Mouse_Release( event.mouse.x, event.mouse.y, Mouse_Left );
                            mbtn(Mouse_Left).pressed := False;
                            if now - mbtn(Mouse_Left).pressTime < CLICK_DELAY and then
                               abs( mbtn(Mouse_Left).pressX - event.mouse.x ) < CLICK_DRIFT and then
                               abs( mbtn(Mouse_Left).pressY - event.mouse.y ) < CLICK_DRIFT
                            then
                                -- event: left mouse click
                                Queue_Mouse_Click( event.mouse.x, event.mouse.y, Mouse_Left );
                                if mbtn(Mouse_Left).clicks = 1 and then
                                   now - mbtn(Mouse_Left).clickTime < DOUBLECLICK_DELAY and then
                                   abs( mbtn(Mouse_Left).clickX - event.mouse.x ) < CLICK_DRIFT and then
                                   abs( mbtn(Mouse_Left).clickY - event.mouse.y ) < CLICK_DRIFT
                                then
                                    -- event: left mouse double click
                                    Queue_Mouse_Doubleclick( event.mouse.x, event.mouse.y, Mouse_Left );
                                end if;
                                mbtn(Mouse_Left).clicks := mbtn(Mouse_Left).clicks + 1;
                                mbtn(Mouse_Left).clickTime := now;
                                mbtn(Mouse_Left).clickX := event.mouse.x;
                                mbtn(Mouse_Left).clickY := event.mouse.y;
                            end if;

                        elsif event.mouse.button = 2 then
                            -- event: right mouse release
                            Queue_Mouse_Release( event.mouse.x, event.mouse.y, Mouse_Right );
                            mbtn(Mouse_Right).pressed := False;
                            if now - mbtn(Mouse_Right).pressTime < CLICK_DELAY and then
                               abs( mbtn(Mouse_Right).pressX - event.mouse.x ) < CLICK_DRIFT and then
                               abs( mbtn(Mouse_Right).pressY - event.mouse.y ) < CLICK_DRIFT
                            then
                                -- event: right mouse click
                                Queue_Mouse_Click( event.mouse.x, event.mouse.y, Mouse_Right );
                                if mbtn(Mouse_Right).clicks = 1 and then
                                   now - mbtn(Mouse_Right).clickTime < DOUBLECLICK_DELAY and then
                                   abs( mbtn(Mouse_Right).clickX - event.mouse.x ) < CLICK_DRIFT and then
                                   abs( mbtn(Mouse_Right).clickY - event.mouse.y ) < CLICK_DRIFT
                                then
                                    -- event: right mouse double click
                                    Queue_Mouse_Doubleclick( event.mouse.x, event.mouse.y, Mouse_Right );
                                end if;
                                mbtn(Mouse_Right).clicks := mbtn(Mouse_Right).clicks + 1;
                                mbtn(Mouse_Right).clickTime := now;
                                mbtn(Mouse_Right).clickX := event.mouse.x;
                                mbtn(Mouse_Right).clickY := event.mouse.y;
                            end if;

                        elsif event.mouse.button = 3 then
                            -- event: middle mouse release
                            Queue_Mouse_Release( event.mouse.x, event.mouse.y, Mouse_Middle );
                            mbtn(Mouse_Middle).pressed := False;
                            if now - mbtn(Mouse_Middle).pressTime < CLICK_DELAY and then
                               abs( mbtn(Mouse_Middle).pressX - event.mouse.x ) < CLICK_DRIFT and then
                               abs( mbtn(Mouse_Middle).pressY - event.mouse.y ) < CLICK_DRIFT
                            then
                                -- event: middle mouse click
                                Queue_Mouse_Click( event.mouse.x, event.mouse.y, Mouse_Middle );
                                if mbtn(Mouse_Middle).clicks = 1 and then
                                   now - mbtn(Mouse_Middle).clickTime < DOUBLECLICK_DELAY and then
                                   abs( mbtn(Mouse_Middle).clickX - event.mouse.x ) < CLICK_DRIFT and then
                                   abs( mbtn(Mouse_Middle).clickY - event.mouse.y ) < CLICK_DRIFT
                                then
                                    -- event: middle mouse double click
                                    Queue_Mouse_Doubleclick( event.mouse.x, event.mouse.y, Mouse_Middle );
                                end if;
                                mbtn(Mouse_Middle).clicks := mbtn(Mouse_Middle).clicks + 1;
                                mbtn(Mouse_Middle).clickTime := now;
                                mbtn(Mouse_Middle).clickX := event.mouse.x;
                                mbtn(Mouse_Middle).clickY := event.mouse.y;
                            end if;

                        end if;

                    when others =>
                        null;
                end case;

                -- send mouse button held events for all buttons held down
                Al_Get_Mouse_State( mstate );
                for mb in mbtn'Range loop
                    if mbtn(mb).pressed and then
                       now - mbtn(mb).pressTime >= CLICK_DELAY and then
                       now - mbtn(mb).lastHeld >= HELD_REPEAT_DELAY
                    then
                        -- event: mouse button still held
                        Queue_Mouse_Held( mstate.x, mstate.y, mb );
                        mbtn(mb).lastHeld := now;
                    end if;
                end loop;

                -- break out of the Allegro event processing loop to check for
                -- the Stop rendezvous.
                if now - start > Milliseconds( 100 ) then
                    exit;
                end if;
            end loop;
        end Process_Events;

        ------------------------------------------------------------------------

        handler  : A_Input_Handler;
        done     : Boolean := False;
        queue    : A_Allegro_Event_Queue;
    begin
        pragma Debug( Dbg( "Input_Handlers.Input_Task created", D_INIT, INFO ) );

        accept Init( this : A_Input_Handler ) do
            handler := this;
        end Init;

        select
            accept Start do
                null;
                pragma Debug( Dbg( "Input_Handlers.Input_Task starting", D_INIT, INFO ) );
            end Start;
        or
            accept Stop do
                pragma Debug( Dbg( "Input_Handlers.Input_Task stopping", D_INIT, INFO ) );
                done := True;
            end Stop;
        end select;

        -- create an event queue for listening to input events
        queue := Al_Create_Event_Queue;
        if queue = null then
            raise Constraint_Error with "Event queue creation failed";
        end if;

        -- register for keyboard events
        Al_Register_Event_Source( queue, Al_Get_Keyboard_Event_Source );

        -- register for mouse events
        if handler.mouseEnabled then
            declare
                source : constant A_Allegro_Event_Source := Al_Get_Mouse_Event_Source;
            begin
                if source /= null then
                    Al_Register_Event_Source( queue, source );
                end if;
            end;
        end if;

        -- Process_Events runs repeatedly to process incoming Allegro events,
        -- breaking to check for the Stop rendezvous at least every 100 ms.
        while not done loop
            select
                accept Stop do
                    pragma Debug( Dbg( "Input_Handlers.Input_Task stopping", D_INIT, INFO ) );
                    done := True;
                end Stop;
            else
                -- Processes all pending Allegro events, waiting up to 100ms for
                -- the first one, if none are queued.
                Process_Events( queue );
            end select;
        end loop;

        Al_Destroy_Event_Queue( queue );

        pragma Debug( Dbg( "Input_Handlers.Input_Task complete", D_INIT, INFO ) );
    exception
        when e : others =>
            Dbg( "Input_Handlers.Input_Task Exception: ...", D_INPUT, ERROR );
            Dbg( e, D_INPUT, ERROR );
            Al_Destroy_Event_Queue( queue );
    end Input_Task;

    ----------------------------------------------------------------------------

    procedure Delete( intask : in out A_Input_Task ) is
        procedure Free is new Ada.Unchecked_Deallocation( Input_Task, A_Input_Task );
    begin
        Free( intask );
    end Delete;

end Input_Handlers;
