--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Containers.Indefinite_Hashed_Maps;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Strings.Hash_Case_Insensitive;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with GNAT.Regexp;                       use GNAT.Regexp;

package body Object_Factory is

    type Object_Class_Type is
        record
            allocate : Allocator := null;
            instance : Object_Access := null;
        end record;

    package Class_Registry is new
        Ada.Containers.Indefinite_Hashed_Maps( String, Object_Class_Type,
                                               Ada.Strings.Hash_Case_Insensitive,
                                               "=", "=" );
    use Class_Registry;

    registry    : Class_Registry.Map;
    name        : Unbounded_String;
    initialized : Boolean := False;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    procedure Initialize( factoryName : String ) is

        procedure Init_Class( pos : Class_Registry.Cursor ) is

            procedure Examine( id : String; elem : in out Object_Class_Type ) is
            begin
                elem.instance := elem.allocate.all;
            exception
                when e : others =>
                    Dbg( "Error initializing " & To_String( name ) &
                         " class <" & id & ">: " & Exception_Name( e ) & " - " &
                         Exception_Message( e ),
                         D_INIT, ERROR );
                    raise;
            end Examine;

        begin
            registry.Update_Element( pos, Examine'Access );
        end Init_Class;

    begin
        if not initialized then
            initialized := True;
            name := To_Unbounded_String( factoryName );
            pragma Debug( Dbg( "Initializing " & To_String( name ) &
                               " object factory", D_INIT, INFO ) );
            registry.Iterate( Init_Class'Access );
        end if;
    end Initialize;

    ----------------------------------------------------------------------------

    procedure Finalize is

        procedure Delete_Class( pos : Class_Registry.Cursor ) is

            procedure Examine( id : String; elem : in out Object_Class_Type ) is
                pragma Unreferenced( id );
            begin
                Delete( elem.instance );
            end Examine;

        begin
            registry.Update_Element( pos, Examine'Access );
        end Delete_Class;

    begin
        if initialized then
            pragma Debug( Dbg( "Finalizing " & To_String( name ) &
                               " object factory", D_INIT, INFO ) );
            registry.Iterate( Delete_Class'Access );
            initialized := False;
        end if;
    end Finalize;

    ----------------------------------------------------------------------------

    function Allocate( id : String ) return Object_Access is
        crsr : constant Class_Registry.Cursor := registry.Find( To_Upper( id ) );
    begin
        if Has_Element( crsr ) then
            return Element( crsr ).allocate.all;
        end if;
        return null;
    end Allocate;

    ----------------------------------------------------------------------------

    procedure Iterate_Classes( pattern : String;
                               examine : access procedure( id : String ) ) is
        exp : constant Regexp := Compile( pattern, Glob => True, Case_Sensitive => False );

        ------------------------------------------------------------------------

        procedure Examine_Class( pos : Class_Registry.Cursor ) is
            id : constant String := Key( pos );
        begin
            if pattern'Length = 0 or else Match( id, exp ) then
                examine( id );
            end if;
        end Examine_Class;

        ------------------------------------------------------------------------

    begin
        registry.Iterate( Examine_Class'Access );
    end Iterate_Classes;

    ----------------------------------------------------------------------------

    procedure Register_Class( id : String; allocate : not null Allocator ) is
        upper : constant String := To_Upper( id );
    begin
        if not registry.Contains( upper ) then
            declare
                elem : Object_Class_Type;
            begin
                elem.allocate := allocate;
                elem.instance := null;
                registry.Insert( upper, elem );
                return;
            end;
        end if;
        pragma Debug( Dbg( "Attempted to register duplicate " &
                           To_String( name ) & " allocator for class " &
                           "<" & upper & ">",
                           D_INIT, WARNING ) );
    end Register_Class;

    ----------------------------------------------------------------------------

    function Template( id : String ) return Object_Access is
        crsr : constant Class_Registry.Cursor := registry.Find( To_Upper( id ) );
    begin
        if Has_Element( crsr ) then
            return Element( crsr ).instance;
        end if;
        return null;
    end Template;

end Object_Factory;
