--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Events.Corrals;                    use Events.Corrals;
with Events.Listeners;                  use Events.Listeners;
with Objects;                           use Objects;
with Processes;                         use Processes;

private with Ada.Containers.Doubly_Linked_Lists;
private with Ada.Containers.Ordered_Maps;
private with Entities;
private with Events;
private with Physics.Bodies;
private with Physics.Clip_Maps;

package Physics.Managers is

    -- The Physics_Manager class is responsible for simulating physical
    -- interaction between the entities and between entities and the world. A
    -- physics manager listens for world and entity events. It represents the
    -- world's tile map with a Clip_Map object contains physical information
    -- about the world based on tile attributes: solid walls, slopes, etc.
    -- Entities in the world are represented as Corpus objects, modelling only
    -- the location, size, and physical attributes of their corresponding
    -- entities. The physics manager runs as a Process that updates position,
    -- velocity, and acceleration of entities, simulates the effects of gravity
    -- and friction, and generates events when entities collide and separate.
    type Physics_Manager is new Object and Process and Event_Listener with private;
    type A_Physics is access all Physics_Manager'Class;

    -- Creates a new physics manager attached to the given Corral.
    function Create_Physics( corral : not null A_Corral ) return A_Physics;
    pragma Postcondition( Create_Physics'Result /= null );

    -- Sets the friction in the X axis for entities handled by this physics
    -- manager.
    function Friction_X( this : not null access Physics_Manager'Class ) return Float;

    -- Sets the friction in the Y axis for entities handled by this physics
    -- manager.
    function Friction_Y( this : not null access Physics_Manager'Class ) return Float;

    -- Sets the gravitional acceleration for entities handled by this physics
    -- manager.
    function Gravity( this : not null access Physics_Manager'Class ) return Float;

    -- Sets the maximum velocity in the X axis for entities handled by this
    -- physics manager.
    function Max_VX( this : not null access Physics_Manager'Class ) return Float;

    -- Sets the maximum velocity in the Y axis for entities handled by this
    -- physics manager.
    function Max_VY( this : not null access Physics_Manager'Class ) return Float;

    -- Deletes the physics manager and all its data (clip map, corps, etc.)
    procedure Delete( this : in out A_Physics );
    pragma Postcondition( this = null );

private

    use Entities;
    use Events;
    use Physics.Bodies;
    use Physics.Clip_Maps;

    package Corpus_Map is new
        Ada.Containers.Ordered_Maps( Entity_Id, A_Corpus, "<", "=" );

    package Corpus_Lists is new
        Ada.Containers.Doubly_Linked_Lists( A_Corpus, "=" );

    ----------------------------------------------------------------------------

    type Physics_Manager is new Object and Process and Event_Listener with
        record
            corral : A_Corral := null;
            map    : A_Clip_Map := null;
            corps  : Corpus_Map.Map;
            grav   : Float := 0.0;
            fricX,
            fricY  : Float := 0.0;
            maxVX,
            maxVY  : Float := 0.0;
        end record;

    procedure Construct( this   : access Physics_Manager;
                         corral : not null A_Corral );

    procedure Delete( this : in out Physics_Manager );

    -- Returns a name identifying the physics manager as a process.
    function Get_Process_Name( this : access Physics_Manager ) return String;
    pragma Postcondition( Get_Process_Name'Result'Length > 0 );

    -- Handles an event received by the physics manager, dispatching it to the
    -- appropriate Handle procedure.
    procedure Handle_Event( this : access Physics_Manager;
                            evt  : in out A_Event;
                            resp : out Response_Type );
    pragma Precondition( evt /= null );

    -- Ticks every Corpus to update its state and then detects collisions.
    procedure Tick( this : access Physics_Manager; time : Tick_Time );

end Physics.Managers;
