--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

--with Debugging;                         use Debugging;

package body Physics is

    type Solidity_Array is array (Clip_Type) of Integer;

    -- mapping of clip_type to solidity values for the "or" operator.
    solidity : constant Solidity_Array :=
        Solidity_Array'(
            Passive => 0,
            OneWay  => 50,

            Slope_22_Up_Floor_Thin     => 70,
            Slope_22_Up_Ceiling_Thin   => 70,
            Slope_22_Down_Floor_Thin   => 70,
            Slope_22_Down_Ceiling_Thin => 70,

            Slope_45_Up_Floor     => 75,
            Slope_45_Up_Ceiling   => 75,
            Slope_45_Down_Floor   => 75,
            Slope_45_Down_Ceiling => 75,

            Slope_22_Up_Floor_Wide    => 80,
            Slope_22_Up_Ceiling_Wide  => 80,
            Slope_22_Down_Floor_Wide  => 80,
            Slope_22_Down_Ceiling_Wide => 80,

            Wall => 100
        );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function "or"( l, r : Clip_Type ) return Clip_Type is
    begin
        -- OR-ing two clip values gives you the greater value, like binary
        if solidity(r) > solidity(l) then
            return r;
        else
            return l;
        end if;
    end "or";

    ----------------------------------------------------------------------------

    procedure Clip_Downward( clip    : Clip_Type;
                             size    : Positive;
                             x, y    : Natural;
                             adjustY : out Integer ) is
        slope : Integer;
    begin
        case clip is
            when Passive =>
                slope := size;
            when Slope_45_Up_Floor =>
                slope := (size - 1) - x;
            when Slope_45_Down_Floor =>
                slope := x;
            when Slope_22_Up_Floor_Thin =>
                slope := (size - 1) - x / 2;
            when Slope_22_Up_Floor_Wide =>
                slope := (size / 2 - 1) - x / 2;
            when Slope_22_Down_Floor_Wide =>
                slope := x / 2;
            when Slope_22_Down_Floor_Thin =>
                slope := (size / 2 - 1) + x / 2;
            when Wall                       |
                 OneWay                     |
                 Slope_45_Up_Ceiling        |
                 Slope_45_Down_Ceiling      |
                 Slope_22_Up_Ceiling_Thin   |
                 Slope_22_Up_Ceiling_Wide   |
                 Slope_22_Down_Ceiling_Thin |
                 Slope_22_Down_Ceiling_Wide =>
                slope := 0;
        end case;

        if y < slope then
            adjustY := 0;
        else
            adjustY := slope - y - 1;
        end if;
    end Clip_Downward;

    ----------------------------------------------------------------------------

    procedure Clip_Upward( clip    : Clip_Type;
                           size    : Positive;
                           x, y    : Natural;
                           adjustY : out Integer ) is
        slope : Integer;
    begin
        case clip is
            when Passive |
                 OneWay  =>
                slope := -1;
            when Slope_45_Up_Ceiling =>
                slope := (size - 1) - x;
            when Slope_45_Down_Ceiling =>
                slope := x;
            when Slope_22_Up_Ceiling_Thin =>
                slope := (size / 2 - 1) - x / 2;
            when Slope_22_Up_Ceiling_Wide =>
                slope := (size - 1) - x / 2;
            when Slope_22_Down_Ceiling_Wide =>
                slope := (size / 2 - 1) + x / 2;
            when Slope_22_Down_Ceiling_Thin =>
                slope := x / 2;
            when Wall                     |
                 Slope_45_Up_Floor        |
                 Slope_45_Down_Floor      |
                 Slope_22_Up_Floor_Thin   |
                 Slope_22_Up_Floor_Wide   |
                 Slope_22_Down_Floor_Thin |
                 Slope_22_Down_Floor_Wide =>
                slope := size - 1;
        end case;

        if y > slope then
            adjustY := 0;
        else
            adjustY := slope - y + 1;
        end if;
    end Clip_Upward;

    ----------------------------------------------------------------------------

    procedure To_Clip_Type( str   : String;
                            clip  : out Clip_Type;
                            valid : out Boolean ) is
    begin
        clip := Clip_Type'Value( str );
        if clip'Valid then
            valid := True;
        else
            clip := Clip_Type'First;
            valid := False;
        end if;
    exception
        when others =>
            clip := Clip_Type'First;
            valid := False;
    end To_Clip_Type;

end Physics;
