
--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Task_Identification;           use Ada.Task_Identification;
with Ada.Unchecked_Deallocation;
with Applications;                      use Applications;
with Applications.Gui.Games;            use Applications.Gui.Games;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Support;                           use Support;
with Support.Real_Time;                 use Support.Real_Time;

package body Processes.Managers is

    -- if the process manager is consistently running slowly and playing catchup
    -- for this amount of time, a warning will be periodically issued.
    SLOW_WARNING_TIMEOUT : constant Time_Span := Seconds( 10 );

    ----------------------------------------------------------------------------

    -- Deletes the Execution.
    procedure Delete( exec : in out A_Execution );

    ----------------------------------------------------------------------------

    -- Deletes the Operation.
    procedure Delete( this : in out A_Operation );

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    -- Appends a process to the execution list.
    type Attach_Operation is new Operation with null record;

    procedure Execute( this     : access Attach_Operation;
                       now      : Time;
                       execList : in out Execution_Lists.List );

    function Create_Attach_Operation( process : not null A_Process ) return A_Operation;

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    -- Removes a process from the execution list.
    type Detach_Operation is new Operation with
        record
            destroy : A_Destructor := null;
        end record;

    -- Detaches the operation's process from 'execList'. If a destructor was
    -- provided, it will be called with the detached process.
    procedure Execute( this     : access Detach_Operation;
                       now      : Time;
                       execList : in out Execution_Lists.List );

    -- If 'destructor' is not null, the detached process will be passed to it.
    function Create_Detach_Operation( process    : not null A_Process;
                                      destructor : A_Destructor ) return A_Operation;

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    -- Pauses the execution of a process before its next Tick.
    type Pause_Operation is new Operation with null record;

    procedure Execute( this     : access Pause_Operation;
                       now      : Time;
                       execList : in out Execution_Lists.List );

    function Create_Pause_Operation( process : not null A_Process ) return A_Operation;

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    -- Resumes the execution of a process on its next Tick.
    type Resume_Operation is new Operation with null record;

    procedure Execute( this     : access Resume_Operation;
                       now      : Time;
                       execList : in out Execution_Lists.List );

    function Create_Resume_Operation( process : not null A_Process ) return A_Operation;

    ----------------------------------------------------------------------------

    -- Deletes the Ticker_Task.
    procedure Delete( ticker : in out A_Ticker_Task );

    --==========================================================================

    function Create_Process_Manager( name     : String;
                                     hertz    : Natural := 0;
                                     minHertz : Natural := 0
                                   ) return A_Process_Manager is
        this : constant A_Process_Manager := new Process_Manager;
    begin
        this.Construct( name, hertz, minHertz );
        return this;
    end Create_Process_Manager;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this     : access Process_Manager;
                         name     : String;
                         hertz,
                         minHertz : Natural ) is
    begin
        Limited_Object(this.all).Construct;
        this.lock := new Locking_Object;
        this.name := To_Unbounded_String( name );
        this.ticker := new Ticker_Task;
        this.ticker.Init( A_Process_Manager(this), hertz, minHertz );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Process_Manager ) is

        -- Delete the operation without executing it
        procedure Delete_Op( pos : Operation_Lists.Cursor ) is
            op : A_Operation;
        begin
            op := Operation_Lists.Element( pos );
            Delete( op );
        end Delete_Op;

    begin
        this.Stop;
        this.lock.Lock;
        this.operations.Iterate( Delete_Op'Access );
        this.operations.Clear;
        Delete( this.lock );
        Delete( this.ticker );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Attach_Async( this : not null access Process_Manager'Class;
                            proc : not null A_Process ) is
        op : A_Operation;
    begin
        op := Create_Attach_Operation( proc );
        this.lock.Lock;
        if not this.stopped then
            pragma Debug( Dbg( "Attaching process <" &
                                op.process.Get_Process_Name & "> to manager " &
                                "<" & this.Get_Name & ">",
                                D_PROCS, INFO ) );
            this.operations.Append( op );
            this.lock.Unlock;
        else
            this.lock.Unlock;
            Delete( op );
            raise Constraint_Error with
                "Error: Attempted to attach process <" &
                proc.Get_Process_Name & "> to manager <" & this.Get_Name &
                "> after Stop";
        end if;
    end Attach_Async;

    ----------------------------------------------------------------------------

    procedure Detach_Async( this : not null access Process_Manager'Class;
                            proc : not null A_Process ) is
        op : A_Operation;
    begin
        op := Create_Detach_Operation( proc, destructor => null );
        this.lock.Lock;
        pragma Debug( Dbg( "Detaching process <" &
                            op.process.Get_Process_Name & "> from manager " &
                            "<" & this.Get_Name & ">",
                            D_PROCS, INFO ) );
        if not this.stopped then
            this.operations.Append( op );
            this.lock.Unlock;
        else
            this.lock.Unlock;
            Delete( op );
        end if;
    end Detach_Async;

    ----------------------------------------------------------------------------

    procedure Detach_Async( this       : not null access Process_Manager'Class;
                            proc       : in out A_Process;
                            destructor : not null A_Destructor ) is
        op : A_Operation;
    begin
        op := Create_Detach_Operation( proc, destructor );
        this.lock.Lock;
        pragma Debug( Dbg( "Detaching process <" &
                            op.process.Get_Process_Name & "> from manager " &
                            "<" & this.Get_Name & ">",
                            D_PROCS, INFO ) );
        if not this.stopped then
            this.operations.Append( op );
            this.lock.Unlock;
            proc := null;
        else
            this.lock.Unlock;
            Delete( op );

            -- no need to detach processes after Stop.
            begin
                destructor.all( proc );
            exception
                when e : others =>
                    Dbg( "Error: Exception destroying process on detach: " &
                         Exception_Name( e ) & ": " & Exception_Message( e ),
                         D_PROCS, ERROR );
            end;
            proc := null;
        end if;
    end Detach_Async;

    ----------------------------------------------------------------------------

    function Get_Name( this : not null access Process_Manager'Class ) return String is
    begin
        return To_String( this.name );
    end Get_Name;

    ----------------------------------------------------------------------------

    function Get_Rate( this : not null access Process_Manager'Class ) return Natural is
        result : Natural := 0;
    begin
        this.lock.Lock;
        if this.trackRate then
            result := this.rate;
        end if;
        this.lock.Unlock;
        return result;
    end Get_Rate;

    ----------------------------------------------------------------------------

    procedure Pause_Async( this   : not null access Process_Manager'Class;
                           proc   : not null A_Process;
                           paused : Boolean )  is
        op : A_Operation;
    begin
        if paused then
            op := Create_Pause_Operation( proc );
        else
            op := Create_Resume_Operation( proc );
        end if;
        this.lock.Lock;
        if not this.stopped then
            this.operations.Append( op );
            this.lock.Unlock;
        else
            this.lock.Unlock;
            Delete( op );
            raise Constraint_Error with
                "Error: Attempted to pause/resume process " &
                proc.Get_Process_Name & " in manager " & this.Get_Name &
                " after Stop";
        end if;
    end Pause_Async;

    ----------------------------------------------------------------------------

    procedure Start( this : not null access Process_Manager'Class ) is
    begin
        this.lock.Lock;
        if not this.started and then not this.stopped then
            this.started := True;
            this.ticker.Start;
        end if;
        this.lock.Unlock;
    end Start;

    ----------------------------------------------------------------------------

    procedure Stop( this : not null access Process_Manager'Class ) is
    begin
        this.lock.Lock;
        if not this.stopped then
            this.stopped := True;

            -- the ticker task needs the lock, so we must release it before
            -- attempting to rendezvous with ticker to prevent a deadlock.
            this.lock.Unlock;

            this.ticker.Stop;

            for count in 1..10 loop
                if Is_Terminated( this.ticker'Identity ) then
                    exit;
                end if;
                delay 0.1;
            end loop;
            if not Is_Terminated( this.ticker'Identity ) then
                Dbg( "Error: Process manager <" & this.Get_Name &
                     "> Ticker_Task failed to stop", D_PROCS, ERROR );
            end if;
        else
            this.lock.Unlock;
        end if;
    end Stop;

    ----------------------------------------------------------------------------

    procedure Track_Rate( this    : not null access Process_Manager'Class;
                          enabled : Boolean ) is
    begin
        this.lock.Lock;
        this.trackRate := enabled;
        this.lock.Unlock;
    end Track_Rate;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Process_Manager ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    procedure Delete( exec : in out A_Execution ) is
        procedure Free is new Ada.Unchecked_Deallocation( Execution, A_Execution );
    begin
        Free( exec );
    end Delete;

    ----------------------------------------------------------------------------

    function Eq( l, r : A_Execution ) return Boolean is
    begin
        if l = null then
            return r = null;
        elsif r = null then
            return False;
        else
            return l.process = r.process;
        end if;
    end Eq;

    --==========================================================================

    procedure Delete( this : in out A_Operation ) is
        procedure Free is new Ada.Unchecked_Deallocation( Operation'Class, A_Operation );
    begin
        Free( this );
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Execute( this     : access Attach_Operation;
                       now      : Time;
                       execList : in out Execution_Lists.List ) is
        crsr : Execution_Lists.Cursor;
        exec : A_Execution := new Execution'(process   => this.process,
                                             firstTick => now,
                                             lastTick  => now,
                                             others    => <>);
    begin
        crsr := execList.Find( exec );
        if Has_Element( crsr ) then
            -- already attached; if the process is flagged for detach, clear
            -- the detach flag. otherwise, ignore the duplicate attach
            -- operation. this covers a special case where a detach and an
            -- attach could be queued back to back.
            Delete( exec );
            exec := Element( crsr );
            if exec.detachMe then
                exec.detachMe := False;
                exec.destructor := null;
            end if;
        else
            pragma Debug( Dbg( "Attached process <" &
                               exec.process.Get_Process_Name & ">",
                               D_PROCS, INFO ) );
            execList.Append( exec );
        end if;
    end Execute;

    ----------------------------------------------------------------------------

    overriding
    procedure Execute( this     : access Detach_Operation;
                       now      : Time;
                       execList : in out Execution_Lists.List ) is
        pragma Unreferenced( now );
        exec : A_Execution := new Execution'(process => this.process, others => <>);
        crsr : Execution_Lists.Cursor;
    begin
        crsr := execList.Find( exec );
        if Execution_Lists.Has_Element( crsr ) then
            Delete( exec );
            exec := Execution_Lists.Element( crsr );
            exec.detachMe := True;
            exec.destructor := this.destroy;
        else
            Delete( exec );
        end if;
    end Execute;

    ----------------------------------------------------------------------------

    overriding
    procedure Execute( this     : access Pause_Operation;
                       now      : Time;
                       execList : in out Execution_Lists.List ) is
        exec : A_Execution := new Execution'(process => this.process, others => <>);
        crsr : Execution_Lists.Cursor;
    begin
        crsr := execList.Find( exec );
        if Execution_Lists.Has_Element( crsr ) then
            Delete( exec );
            exec := Execution_Lists.Element( crsr );
            exec.paused := True;
            exec.pauseTime := now;
            pragma Debug( Dbg( "Paused process <" &
                               exec.process.Get_Process_Name & ">",
                               D_PROCS, INFO ) );
        else
            Delete( exec );
        end if;
    end Execute;

    ----------------------------------------------------------------------------

    overriding
    procedure Execute( this     : access Resume_Operation;
                       now      : Time;
                       execList : in out Execution_Lists.List ) is
        exec : A_Execution := new Execution'(process => this.process, others => <>);
        crsr : Execution_Lists.Cursor;
    begin
        crsr := execList.Find( exec );
        if Execution_Lists.Has_Element( crsr ) then
            Delete( exec );
            exec := Execution_Lists.Element( crsr );
            pragma Debug( Dbg( "Resuming process <" &
                               exec.process.Get_Process_Name & ">",
                               D_PROCS, INFO ) );
            if exec.paused then
                exec.paused := False;
                exec.lastTick := now - (exec.pauseTime - exec.lastTick);
            end if;
        else
            Delete( exec );
        end if;
    end Execute;

    ----------------------------------------------------------------------------

    function Create_Attach_Operation( process : not null A_Process ) return A_Operation is
        op : constant A_Operation := new Attach_Operation;
    begin
        op.process := process;
        return op;
    end Create_Attach_Operation;

    ----------------------------------------------------------------------------

    function Create_Detach_Operation( process    : not null A_Process;
                                      destructor : A_Destructor ) return A_Operation is
        op : constant A_Operation := new Detach_Operation;
    begin
        op.process := process;
        Detach_Operation(op.all).destroy := destructor;
        return op;
    end Create_Detach_Operation;

    ----------------------------------------------------------------------------

    function Create_Pause_Operation( process : not null A_Process ) return A_Operation is
        op : constant A_Operation := new Pause_Operation;
    begin
        op.process := process;
        return op;
    end Create_Pause_Operation;

    ----------------------------------------------------------------------------

    function Create_Resume_Operation( process : not null A_Process ) return A_Operation is
        op : constant A_Operation := new Resume_Operation;
    begin
        op.process := process;
        return op;
    end Create_Resume_Operation;

    --==========================================================================

    task body Ticker_Task is
        this : A_Process_Manager;
        fps  : aliased Frame_Timer;   -- for frame rate counter

        ------------------------------------------------------------------------

#if DEBUG then
        nextUsageUpdate : Time := Clock;
        usageTotal      : Natural := 0;
        usageCount      : Natural := 0;

        -- Track the CPU time used by the process manager. The percentage used
        -- can be calculated by checking the difference between the requested
        -- time between frames and the amount of time remaining for sleep after
        -- ticking all processes in the frame. This code is debugging only so it
        -- doesn't slow down the release version. The average CPU use each
        -- second is sent to debugging output and can be seen with INFO level
        -- output for the PROCS system.
        procedure Track_CPU_Usage( tickDelta : Time_Span;
                                   nextTick  : Time ) is
            now       : constant Time := Clock;
            remaining : Time_Span;
            usageNow  : Natural := 0;
        begin
            if now > nextTick then
                usageNow := 100;
            else
                remaining := nextTick - now;
                usageNow := 100 - Integer(Float(100.0) * Float(To_Duration( remaining ) / To_Duration( tickDelta )));
            end if;
            usageTotal := usageTotal + usageNow;
            usageCount := usageCount + 1;

            if now > nextUsageUpdate then
                Dbg( "Process manager <" & this.Get_Name & "> CPU: " &
                     Image( usageTotal / usageCount ) & "%",
                     D_PROCS, INFO );
                usageTotal := 0;
                usageCount := 0;
                nextUsageUpdate := now + Seconds( 1 );
            end if;
        end Track_CPU_Usage;
#end if;

        ------------------------------------------------------------------------

        procedure Track_Rate is
        begin
            this.lock.Lock;
            if this.trackRate then
                Tick( fps );
                if Is_Updated( fps ) then
                    this.rate := Rate( fps'Access );
                end if;
            end if;
            this.lock.Unlock;
        end Track_Rate;

        ------------------------------------------------------------------------

        procedure Execute_Pending_Ops( execList : in out Execution_Lists.List;
                                       now      : Time ) is

            procedure Execute_Op( pos : Operation_Lists.Cursor ) is
                op : A_Operation := Operation_Lists.Element( pos );
            begin
                op.Execute( now, execList );
                Delete( op );
            end Execute_Op;

        begin
            this.lock.Lock;
            this.operations.Iterate( Execute_Op'Access );
            this.operations.Clear;
            this.lock.Unlock;
        end Execute_Pending_Ops;

        ------------------------------------------------------------------------

        lastTick     : Time := Time_First;
        catchUpStart : Time := Time_First;

        procedure Tick_All( execList : in out Execution_Lists.List;
                            maxDelta : Time_Span ) is
            current  : constant Time := Clock;
            tickTime : Time;

            --------------------------------------------------------------------

            procedure Tick_Process( exec : not null A_Execution ) is
                dt : Time_Span;
            begin
                dt := tickTime - exec.lastTick;
                if maxDelta > Time_Span_Zero and then dt > maxDelta then
                    dt := maxDelta;
                end if;
                exec.process.Tick( Tick_Time'(total => tickTime - exec.firstTick, elapsed => dt) );
                exec.lastTick := tickTime;
            exception
                when e : others =>
                    -- the process threw an unhandled exception. terminate the
                    -- application.
                    Terminate_Application( "Exception in process <" &
                                           exec.process.Get_Process_Name & "> of " &
                                           "process manager <" & this.Get_Name & ">: " &
                                           Exception_Message( e ) );
            end Tick_Process;

            --------------------------------------------------------------------

            crsr  : Execution_Lists.Cursor;
            crsr2 : Execution_Lists.Cursor;
            exec  : A_Execution;
        begin
            if lastTick > Time_First and then
               maxDelta > Time_Span_Zero and then
               current - lastTick > maxDelta
            then
                -- The last time the processes were ticked was longer ago than
                -- we allow. Another OS process may have been running and slowed
                -- us down, or we performed some logic last time that took longer
                -- that the allowed time between ticks.
                --
                -- To account for this large delay, we'll run Tick as though the
                -- time is 'maxDelta' beyond the last Tick, to prevent the
                -- processes from seeing a time delta too large. We'll then
                -- increment the time ('tickTime') by 'maxDelta' again and
                -- repeat until tickTime catches up with the actually 'current'
                -- time.
                --
                -- However, if the processes are just running too slowly, we
                -- might never catch up. We need to detect this case and warn
                -- the user.
                tickTime := lastTick + maxDelta;
            else
                tickTime := current;
            end if;

            loop
                -- execute any pending process management operations queued
                -- since the previous frame
                Execute_Pending_Ops( execList, tickTime );

                crsr := execList.First;

                -- execute each of the running processes
                while Has_Element( crsr ) loop
                    exec := Execution_Lists.Element( crsr );

                    -- check if the process was previously flagged for detach
                    if exec.detachMe then
                        -- detach the process
                        crsr2 := Next( crsr );
                        execList.Delete( crsr );
                        crsr := crsr2;

                        -- delete it if necessary
                        if exec.destructor /= null then
                            exec.destructor( exec.process );
                        end if;
                        Delete( exec );
                    else
                        -- tick the process
                        if not exec.paused then
                            Tick_Process( exec );
                        end if;

                        -- execute any pending process management operations
                        -- this may cause any processes to be flagged for
                        -- detachment on their next tick.
                        Execute_Pending_Ops( execList, tickTime );

                        -- move to the next process
                        Next( crsr );
                    end if;
                end loop;

                -- we haven't caught up yet so we need to do another iteration
                if tickTime < current then
                    if tickTime + maxDelta <= current then
                        -- we need to do another full iteration
                        tickTime := tickTime + maxDelta;
                    else
                        -- this is the last iteration and dt < maxDelta
                        tickTime := current;
                    end if;
                else
                    -- we're caught up
                    exit;
                end if;
            end loop;

            if lastTick > Time_First and then      -- not the first Tick
               maxDelta > Time_Span_Zero and then  -- have a maxDelta constraint
               tickTime - lastTick > maxDelta
            then
                -- the last time Tick was called, it took too long so we had to
                -- catch up this time.
                if catchUpStart = Time_First then
                    -- first time we had to catch up
                    catchUpStart := lastTick;
                elsif tickTime - catchUpStart > SLOW_WARNING_TIMEOUT then
                    -- SLOW_WARNING_TIMEOUT time has elapsed since the processes
                    -- started running too slowly. warn that the application is
                    -- running too slowly. the processes need to execute for
                    -- smaller periods of time, fewer processes should be
                    -- attached to this process manager, or the rate of
                    -- execution should be reduced.
                    Dbg( "Process Manager <" & this.Get_Name &
                         "> is running too slowly", D_PROCS, ERROR );
                    catchUpStart := lastTick;
                end if;
            else
                catchUpStart := Time_First;
            end if;

            -- record the time of the last iteration
            lastTick := tickTime;
        end Tick_All;

        ------------------------------------------------------------------------

        stopped   : Boolean := False;
        nextTick  : Time;
        tickDelta,
        maxDelta  : Time_Span := Time_Span_Zero;
        execList  : aliased Execution_Lists.List;
    begin
        accept Init( pman : A_Process_Manager; hertz, minHertz : Natural ) do
            this := pman;
            if hertz > 0 then
                tickDelta := Seconds( 1 ) / hertz;
            end if;
            if minHertz > 0 then
                maxDelta := Seconds( 1 ) / minHertz;
            end if;
            pragma Debug( Dbg( "Process manager <" & this.Get_Name & "> initialized", D_PROCS, INFO ) );
        end Init;

        select
            accept Start do
                null;
                pragma Debug( Dbg( "Process manager <" & this.Get_Name & "> starting", D_PROCS, INFO ) );
            end Start;
        or
            accept Stop do
                pragma Debug( Dbg( "Process manager <" & this.Get_Name & "> stopping", D_PROCS, INFO ) );
                stopped := True;
            end Stop;
        end select;

        nextTick := Clock + tickDelta;
        while not stopped loop

            -- track the execution rate (in hertz)
            Track_Rate;

            -- execute all our processes (one frame)
            Tick_All( execList, maxDelta );

            -- track CPU usage (determined by now much of tickDelta we use)
#if DEBUG then
            Track_CPU_Usage( tickDelta, nextTick );
#end if;

            -- delay until the next frame time
            select
                accept Stop do
                    pragma Debug( Dbg( "Process manager <" & this.Get_Name & "> stopping", D_PROCS, INFO ) );
                    stopped := True;
                end Stop;
            or
                delay until nextTick;
            end select;

            nextTick := nextTick + tickDelta;
        end loop;

        -- Execute any remaining queued operations. This is necessary because
        -- they may flag certain Processes for detachment with a destructor,
        -- leaving us reponsible for deleting those Processes.
        Execute_Pending_Ops( execList, nextTick );

        declare
            crsr : Execution_Lists.Cursor := execList.First;
            exec : A_Execution;
        begin
            -- delete any processed flagged for detachment, having a destructor
            while Has_Element( crsr ) loop
                exec := Element( crsr );
                if exec.detachMe and then exec.destructor /= null then
                    exec.destructor( exec.process );
                end if;
                Next( crsr );
            end loop;
            execList.Clear;
        end;

        pragma Debug( Dbg( "Process manager <" & this.Get_Name & "> complete", D_PROCS, INFO ) );
    exception
        when e : others =>
            Dbg( "Process manager <" & this.Get_Name & "> Ticker_Task Exception: ...", D_PROCS, ERROR );
            Dbg( e, D_PROCS, ERROR );
            A_Game_Application(Get_Application).Stop( ERROR_UNEXPECTED_EXCEPTION );
    end Ticker_Task;

    ----------------------------------------------------------------------------

    procedure Delete( ticker : in out A_Ticker_Task ) is
        procedure Free is new Ada.Unchecked_Deallocation( Ticker_Task, A_Ticker_Task );
    begin
        Free( ticker );
    end Delete;

end Processes.Managers;
