--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Allegro.Displays;                  use Allegro.Displays;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Objects;                           use Objects;
with Processes;                         use Processes;
with Widgets.Containers.Windows;        use Widgets.Containers.Windows;

private with Ada.Real_Time;

package Renderers is

    -- A Renderer is a member of the Game View system and is responsible for
    -- drawing a Window widget to the screen. Only one renderer exists per
    -- application. It implements the Process interface to draw on a tick when
    -- the specified framerate allows. The renderer uses double-buffering for
    -- updating frames.
    type Renderer is new Limited_Object and Process with private;
    type A_Renderer is access all Renderer'Class;

    -- Create a Renderer object that draws 'win' to 'display' at a maximum
    -- frequency of 'fps'. For best performance, do not attach the Renderer to
    -- a Process_Manager running at a frequency less than 'fps'; it's best if
    -- the Process is ticked at the same frequency, or a multiple of it.
    function Create_Renderer( display : A_Allegro_Display;
                              win     : not null A_Window;
                              fps     : Positive ) return A_Renderer;
    pragma Postcondition( Create_Renderer'Result /= null );

    -- Deletes the Renderer.
    procedure Delete( this : in out A_Renderer );
    pragma Postcondition( this = null );

private

    use Ada.Real_Time;

    -- an array of video bitmaps for double/triple-buffering
    type Page_Array is array(Natural range <>) of A_Allegro_Bitmap;
    type A_Page_Array is access all Page_Array;

    ----------------------------------------------------------------------------

    type Renderer is new Limited_Object and Process with
        record
            display    : A_Allegro_Display := null;
            buffer     : A_Allegro_Bitmap := null;
            win        : A_Window := null;
            frameDelta : Time_Span := Time_Span_Zero;
            lastFrame  : Time := Time_First;
        end record;

    procedure Construct( this    : access Renderer;
                         display : A_Allegro_Display;
                         win     : not null A_Window;
                         fps     : Positive );

    procedure Delete( this : in out Renderer );

    function Get_Process_Name( this : access Renderer ) return String;

    -- Updates the screen.
    procedure Tick( this : access Renderer; time : Tick_Time );

end Renderers;
