--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Streams.Stream_IO;             use Ada.Streams.Stream_IO;
--with Debugging;                         use Debugging;
with Resources.Archives;                use Resources.Archives;
with Support.Paths;                     use Support.Paths;
with UnZip.Streams;                     use UnZip.Streams;
with Zip_Streams;

pragma Elaborate_All( Resources.Archives );

package body Resources.Archives.Zip_Archives is

    -- Loads a .zip from a Resource_File. Raises an exception on error.
    function Load_Zip( resource : not null A_Resource_File ) return A_Archive;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Load_Zip( resource : not null A_Resource_File ) return A_Archive is
        this : A_Archive := new Zip_Archive;
    begin
        Zip_Archive(this.all).Construct( resource );
        return this;
    exception
        when others =>
            Delete( this );
            raise;
    end Load_Zip;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this     : access Zip_Archive;
                         resource : not null A_Resource_File ) is
        buffer : Zip_Streams.Array_Streams.A_SEA := null;
        stream : A_Buffer_Stream;
    begin
        Archive(this.all).Construct( resource.Get_Id );

        -- Copy the file contents in memory
        buffer := new Stream_Element_Array(1..Stream_Element_Offset(resource.Size));
        resource.Get_Stream( stream );
        Stream_Element_Array'Read( stream, buffer.all );
        Close( stream );

        -- Cache the zip info catalog
        begin
            this.zipArray := new Array_Stream;
            this.zipArray.Set( buffer );
            Load( this.zipInfo, Zip_Streams.Zipstream_Class(this.zipArray) );
        exception
            when e : others =>
                raise READ_ERROR with "Error reading archive: " &
                                      Exception_Name( e ) & " - " &
                                      Exception_Message( e );
        end;
    exception
        when others =>
            Close( stream );
            Delete( buffer );
            raise;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Zip_Archive ) is
    begin
        Delete( this.zipInfo );
        Delete( this.zipArray );
        Archive(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    function Exists( this     : access Zip_Archive;
                     filename : String ) return Boolean is
        index        : Ada.Streams.Stream_IO.Positive_Count;
        csize, usize : File_size_type;
    begin
        Find_offset( this.zipInfo, filename, False, index, csize, usize );
        return True;
    exception
        when File_name_not_found =>
            return False;
    end Exists;

    ----------------------------------------------------------------------------

    overriding
    function Load_Raw( this     : access Zip_Archive;
                       filename : String ) return access Stream_Element_Array is
        zippedFile : Zipped_File_Type;
        compSize,
        uncompSize : File_size_type;
        fileData   : Zip_Streams.Array_Streams.A_SEA := null;
    begin
        Open( zippedFile, this.zipInfo, filename );
        Get_sizes( this.zipInfo, filename, False, compSize, uncompSize );
        fileData := new Stream_Element_Array(0..Stream_Element_Offset(uncompSize - 1));
        Stream_Element_Array'Read( Stream( zippedFile ), fileData.all );
        Close( zippedFile );

        return fileData;
    exception
        when others =>
            Delete( fileData );
            if Is_Open( zippedFile ) then
                Close( zippedFile );
            end if;
            raise;
    end Load_Raw;

    ----------------------------------------------------------------------------

    function Load_Resource( this     : access Zip_Archive;
                            filename : String ) return A_Resource_File is
        data : Streams.A_SEA;
    begin
        -- may raise exception
        data := Streams.A_SEA(this.Load_Raw( filename ));

        return resource : A_Resource_File do
            Create_Resource( resource, To_String( this.path ) & Slash & filename, data );
        end return;
    end Load_Resource;

begin

    Register_Format( "gfx", Load_Zip'Access );
    Register_Format( "zip", Load_Zip'Access );

end Resources.Archives.Zip_Archives;
