--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Containers.Indefinite_Hashed_Maps;
with Ada.Strings.Equal_Case_Insensitive;
with Ada.Strings.Hash_Case_Insensitive;
with Ada.Streams;                       use Ada.Streams;
with Ada.Streams.Stream_IO;             use Ada.Streams.Stream_IO;
with Applications;                      use Applications;
with Content_Signing;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;
with Locking_Objects;                   use Locking_Objects;
with Preferences;                       use Preferences;
with Resources.Archives;                use Resources.Archives;
with Resources.Cache;                   use Resources.Cache;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;

package body Resources.Cache is

    package Maps is new
        Ada.Containers.Indefinite_Hashed_Maps( String,
                                               A_Resource_File,
                                               Ada.Strings.Hash_Case_Insensitive,
                                               Ada.Strings.Equal_Case_Insensitive,
                                               "=" );

    lock : Locking_Object;
    map  : Maps.Map;
    size : Unsigned_32 := 0;

    ----------------------------------------------------------------------------

    function Load_From_Disk( filepath : String; group : String ) return A_SEA;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Load_From_Disk( filepath : String; group : String ) return A_SEA is

        ------------------------------------------------------------------------

        function Read_File( path : String ) return A_SEA is
            len    : constant Long_Integer := File_Length( path );
            ft     : File_Type;
            buffer : A_SEA := null;
        begin
            if len >= 0 then
                Open( ft, In_File, path );
                if Is_Open( ft ) then
                    pragma Debug( Dbg( group & ':' & filepath & " -> " & path,
                                       D_RES, INFO ) );
                    buffer := new Stream_Element_Array(0..Stream_Element_Offset(len-1));
                    Stream_Element_Array'Read( Stream( ft ), buffer.all );
                    Close( ft );
                    if not Content_Signing.Is_Valid( path, buffer.all ) then
                        Delete( buffer );
                        Dbg( "Error: " & path & " signature does not match",
                             D_RES, ERROR );
                    end if;
                end if;
            end if;
            return buffer;
        exception
            when others =>
                Delete( buffer );
                if Is_Open( ft ) then
                    Close( ft );
                end if;
                return null;
        end Read_File;

        ------------------------------------------------------------------------

        -- 'path' is the path of the archive and 'filename' is the name of the
        -- file within the archive to read.
        function Read_From_Archive( path     : String;
                                    filename : String ) return A_SEA is
            resource : A_Resource_File;
            archive  : A_Archive;
            buffer   : A_SEA := null;
        begin
            resource := Load( path, "", keep => True, needLock => False );
            if resource /= null then
                archive := Load_Archive( resource );
                buffer := A_SEA(archive.Load_Raw( filename ));   -- may raise exception
                pragma Debug( Dbg( group & ':' & filepath & " -> " &
                                   path & "/" & filename, D_RES, INFO ) );
                Delete( archive );
                Unload( resource, needLock => False );
            end if;
            return buffer;
        exception
            when others =>
                Delete( buffer );
                Delete( archive );
                Unload( resource, needLock => False );
                return null;
        end Read_From_Archive;

        ------------------------------------------------------------------------

        mediaDir : constant String := Get_Pref( "application", "media" );
        userDir  : constant String := Get_Application.Get_User_Directory;
        data  : A_SEA;
    begin
        -- 1. Look at absolute filepath
        if Is_Absolute_Path( filepath ) and then Is_Regular_File( filepath ) then
            return Read_File( filepath );
        end if;

        -- 2. Look at working_dir/filename
        if data = null and then Is_Regular_File( Get_Filename( filepath ) ) then
             data := Read_File( Get_Filename( filepath ) );
        end if;

        -- 3. Look at working_dir/group/filename
        if data = null and then
           group'Length > 0 and then
           Is_Regular_File( group & Slash & Get_Filename( filepath ) )
        then
            data := Read_File( group & Slash & Get_Filename( filepath ) );
        end if;

        -- 4. Look at working_dir/group.archive/filename
        if data = null and then
           group'Length > 0 and then
           Is_Regular_File( group & '.' & ARCHIVE_EXTENSION )
        then
            data := Read_From_Archive( group & '.' & ARCHIVE_EXTENSION,
                                       Get_Filename( filepath ) );
        end if;

        -- 5. Look at user_dir/filename
        if data = null and then Is_Regular_File( userDir & Get_Filename( filepath ) ) then
            data := Read_File( userDir & Get_Filename( filepath ) );
        end if;

        -- 6. Look at media_dir/filename
        if data = null and then Is_Regular_File( mediaDir & Get_Filename( filepath ) ) then
            data := Read_File( mediaDir & Get_Filename( filepath ) );
        end if;

        -- 7. Look at media_dir/group/filename
        if data = null and then
           group'Length > 0 and then
           Is_Regular_File( mediaDir & group & Slash & Get_Filename( filepath ) )
        then
            data := Read_File( mediaDir & group & Slash & Get_Filename( filepath ) );
        end if;

        -- 8. Look at media_dir/group.archive/filename
        if data = null and then
           group'Length > 0 and then
           Is_Regular_File( mediaDir & group & '.' & ARCHIVE_EXTENSION )
        then
            data := Read_From_Archive( mediaDir & group & '.' & ARCHIVE_EXTENSION,
                                       Get_Filename( filepath ) );
        end if;

        return data;
    end Load_From_Disk;

    ----------------------------------------------------------------------------

    function Load( filepath : String;
                   group    : String;
                   keep     : Boolean;
                   needLock : Boolean ) return A_Resource_File is
        resource : A_Resource_File;
        pos      : Maps.Cursor;
        tmp      : Boolean;
        data     : A_SEA;
    begin
        if needLock then
            lock.Lock;
        end if;
        pos := map.Find( group & ':' & filepath );
        if Maps.Has_Element( pos ) then
            resource := Maps.Element( pos );
            resource.Reference;
        else
            data := Load_From_Disk( filepath, group );
            if data /= null then
                Create_Resource( resource, filepath, group, keep, data );
                map.Insert( group & ':' & filepath, resource, pos, tmp );
                size := size + resource.Size;
                pragma Debug( Dbg( "Resources: Cache size:" &
                                   Unsigned_32'Image( size / 1024 ) & "KB",
                                   D_RES, INFO ) );
            else
                resource := null;
            end if;
        end if;
        if needLock then
            lock.Unlock;
        end if;
        return resource;
    end Load;

    ----------------------------------------------------------------------------

    procedure Preload( filepath : String; group : String; needLock : Boolean ) is
        pos      : Maps.Cursor;
        resource : A_Resource_File;
        tmp      : Boolean;
        data     : A_SEA;
    begin
        if needLock then
            lock.Lock;
        end if;
        pos := map.Find( group & ':' & filepath );
        if not Maps.Has_Element( pos ) then
            data := Load_From_Disk( filepath, group );
            if data /= null then
                Create_Resource( resource, filepath, group, True, data );
                resource.Unreference;   -- don't count this reference
                map.Insert( group & ':' & filepath, resource, pos, tmp );
                size := size + resource.Size;
                pragma Debug( Dbg( "Resources: Cache size:" &
                                   Unsigned_32'Image( size / 1024 ) & "KB",
                                   D_RES, INFO ) );
            end if;
        end if;
        if needLock then
            lock.Unlock;
        end if;
    end Preload;

    ----------------------------------------------------------------------------

    procedure Unload( resource : in out A_Resource_File; needLock : Boolean ) is
        pos : Maps.Cursor;
    begin
        if needLock then
            lock.Lock;
        end if;

        -- remove the resource from the cache, if it's cached
        -- if 'keep' was set true, the resource will not be unloaded
        if not resource.keep then
            -- search for the resource in the cache by reference
            pos := map.First;
            while Maps.Has_Element( pos ) loop
                if resource = Maps.Element( pos ) then
                    resource.Unreference;                -- remove cache's reference
                    map.Delete( pos );                   -- remove from cache
                    size := size - resource.Size;
                    exit;
                end if;
                Maps.Next( pos );
            end loop;
        end if;

        -- delete the resource if no references remain
        if resource.refs = 0 then
            Delete( resource );
        else
            resource := null;
        end if;

        if needLock then
            lock.Unlock;
        end if;
    end Unload;

end Resources.Cache;
