--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Unchecked_Deallocation;
with Applications;                      use Applications;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;
with Preferences;                       use Preferences;
with Resources.Archives;                use Resources.Archives;
with Resources.Allegro_Files;           use Resources.Allegro_Files;
with Resources.Cache;                   use Resources.Cache;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;

package body Resources is

    function Find_On_Disk( filepath : String; group : String ) return String;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    procedure Create_Resource( this     : out A_Resource_File;
                               filepath : String;
                               group    : String;
                               keep     : Boolean;
                               data     : in out A_SEA ) is
    begin
        this := new Resource_File;
        this.filepath := To_Unbounded_String( filepath );
        this.group := To_Unbounded_String( group );
        this.keep := keep;
        this.data := data;
        data := null;
    end Create_Resource;

    ----------------------------------------------------------------------------

    procedure Create_Resource( this     : out A_Resource_File;
                               filepath : String;
                               data     : in out A_SEA ) is
    begin
        this := new Resource_File;
        this.filepath := To_Unbounded_String( filepath );
        this.keep := False;
        this.data := data;
        data := null;
    end Create_Resource;

    ----------------------------------------------------------------------------

    function Get_Address( this : not null access Resource_File'Class ) return Address is
    begin
        if this.data'Length > 0 then
            return this.data(this.data'First)'Address;
        end if;
        return Null_Address;
    end Get_Address;

    ----------------------------------------------------------------------------

    function Get_Filename( this : not null access Resource_File'Class ) return String is
    begin
        return Get_Filename( To_String( this.filepath ) );
    end Get_Filename;

    ----------------------------------------------------------------------------

    function Get_Id( this : not null access Resource_File'Class ) return String is
    begin
        return To_String( this.group ) & ':' & To_String( this.filepath );
    end Get_Id;

    ----------------------------------------------------------------------------

    function Get_File( this : not null access Resource_File'Class ) return A_Allegro_File is
    begin
        return Allegro_Files.Open_Memfile( this.data, consume => False );
    end Get_File;

    ----------------------------------------------------------------------------

    procedure Get_Stream( this : not null access Resource_File'Class;
                          strm : out A_Buffer_Stream ) is
    begin
        strm := Stream( this.data );
    end Get_Stream;

    ----------------------------------------------------------------------------

    procedure Reference( this : not null access Resource_File'Class ) is
    begin
        this.refs := this.refs + 1;
        this.lastUse := Clock;
    end Reference;

    ----------------------------------------------------------------------------

    function Size( this : not null access Resource_File'Class ) return Unsigned_32 is
    begin
        return Unsigned_32(this.data'Length);
    end Size;

    ----------------------------------------------------------------------------

    procedure Unreference( this : not null access Resource_File'Class ) is
    begin
        if this.refs > 0 then
            this.refs := this.refs - 1;
        else
            Dbg( "Error: Reference count error for <" & this.Get_Id & '>',
                 D_RES, ERROR );
        end if;
    end Unreference;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Resource_File ) is
        procedure Free is new Ada.Unchecked_Deallocation( Resource_File'Class,
                                                          A_Resource_File );
    begin
        if this /= null then
            if this.refs = 0 then
                Delete( this.data );
                Free( this );
            else
                Dbg( "Error: References remain for <" & this.Get_Id & '>',
                     D_RES, ERROR );
            end if;
        end if;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Release( resource : in out A_Resource_File ) is
    begin
        if resource /= null then
            Cache.Unload( resource, needLock => True );
        end if;
    end Release;

    --==========================================================================

    function Find_On_Disk( filepath : String; group : String ) return String is

        ------------------------------------------------------------------------

        -- 'path' is the path of the archive to test and 'filename' is the name
        -- of the file within the archive to test.
        function Exists_In_Archive( path     : String;
                                    filename : String ) return Boolean is
            resource : A_Resource_File;
            archive  : A_Archive;
            exists   : Boolean := False;
        begin
           resource := Load_Resource( path, "", keep => False );
            if resource /= null then
                archive := Load_Archive( resource );
                exists := archive.Exists( filename );
                Delete( archive );
                Release( resource );
            end if;
            return exists;
        exception
            when others =>
                Delete( archive );
                Release( resource );
                return False;
        end Exists_In_Archive;

        ------------------------------------------------------------------------

        mediaDir : constant String := Get_Pref( "application", "media" );
        userDir  : constant String := Get_Application.Get_User_Directory;
    begin
        -- 1. Look at filepath (if it is absolute)
        if Is_Absolute_Path( filepath ) and then Is_Regular_File( filepath ) then
            return filepath;
        end if;

        -- 2. Look at working_dir/filename
        if Is_Regular_File( Get_Filename( filepath ) ) then
            return Get_Filename( filepath );
        end if;

        -- 3. Look at working_dir/group/filename
        if group'Length > 0 and then
           Is_Regular_File( group & Slash & Get_Filename( filepath ) )
       then
           return group & Slash & Get_Filename( filepath );
        end if;

        -- 4. Look at working_dir/group.archive/filename
        if group'Length > 0 and then
           Is_Regular_File( group & '.' & ARCHIVE_EXTENSION ) and then
           Exists_In_Archive( group & '.' & ARCHIVE_EXTENSION, Get_Filename( filepath ) )
       then
           return group & '.' & ARCHIVE_EXTENSION;
        end if;

        -- 5. Look at user_dir/filename
        if Is_Regular_File( userDir & Get_Filename( filepath ) ) then
            return userDir & Get_Filename( filepath );
        end if;

        -- 6. Look at media_dir/filename
        if Is_Regular_File( mediaDir & Get_Filename( filepath ) ) then
            return mediaDir & Get_Filename( filepath );
        end if;

        -- 7. Look at media_dir/group/filename
        if Is_Regular_File( mediaDir & group & Slash & Get_Filename( filepath ) ) then
            return mediaDir & group & Slash & Get_Filename( filepath );
        end if;

        -- 8. Look at media_dir/group.archive/filename
        if group'Length > 0 and then
           Is_Regular_File( mediaDir & group & '.' & ARCHIVE_EXTENSION ) and then
           Exists_In_Archive( mediaDir & group & '.' & ARCHIVE_EXTENSION, Get_Filename( filepath ) )
        then
            return mediaDir & group & '.' & ARCHIVE_EXTENSION;
        end if;

        -- Resource not found
        return "";
    end Find_On_Disk;

    ----------------------------------------------------------------------------

    function Load_Resource( filepath : String;
                            group    : String;
                            keep     : Boolean := False ) return A_Resource_File is
    begin
        return Cache.Load( filepath, group, keep, needLock => True );
    end Load_Resource;

    ----------------------------------------------------------------------------

    procedure Preload_Resource( filepath : String; group : String ) is
    begin
        Cache.Preload( filepath, group, needLock => True );
    end Preload_Resource;

    ----------------------------------------------------------------------------

    function Write_Path( filepath : String; group : String ) return String is
        found : constant String := Find_On_Disk( filepath, group );
    begin
        if found'Length > 0 then
            if Get_Filename( found ) = Get_Filename( filepath ) then
                -- found the naked resource file
                return Normalize_Pathname( found );
            else
                -- found it in an archive; return the file as though it were
                -- in the same directory as the archive. if it's not writable,
                -- the caller can always fallback to the user directory or
                -- current working directory.
                return Normalize_Pathname( Get_Directory( found ) & Get_Filename( filepath ) );
            end if;
        else
            -- didn't find the file; return a path to the working directory
            return Normalize_Pathname( Get_Filename( filepath ) );
        end if;
    end Write_Path;

end Resources;
