--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Streams.Stream_IO;             use Ada.Streams.Stream_IO;
with Scripting.Expressions;             use Scripting.Expressions;
with Scripting.Scripts;                 use Scripting.Scripts;
with Streams.Buffers;                   use Streams.Buffers;
with Tokens;                            use Tokens;

package body Scripting.Evaluators is

    function Create_Evaluator return A_Evaluator is
        this : constant A_Evaluator := new Evaluator;
    begin
        this.Construct;
        return this;
    end Create_Evaluator;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Evaluator ) is
    begin
        Delete( this.parser );
        Delete( this.scanner );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Evaluate_Expression( this       : not null access Evaluator'Class;
                                  expression : String;
                                  evalNode   : A_Evaluation_Node ) return Value_Ptr is
        input   : A_Buffer_Stream := Stream( expression );
        exp     : A_Expression := null;
        context : A_Eval_Context := null;
        result  : Value_Ptr;
    begin
        this.scanner.Set_Input( Stream_Access(input) );
        exp := this.parser.Scan_Expression( this.scanner );  -- may raise exception
        this.scanner.Expect( TK_EOF );                       -- may raise exception
        if exp /= null then
            context := Create_Eval_Context( evalNode );
            result := exp.Evaluate( context );               -- may raise exception
            Delete( context );
            Delete( exp );
        end if;
        this.scanner.Set_Input( null );
        Close( input );
        return result;
    exception
        when others =>
            Delete( context );
            Delete( exp );
            this.scanner.Set_Input( null );
            Close( input );
            raise;
    end Evaluate_Expression;

    ----------------------------------------------------------------------------

    procedure Evaluate_Script( this     : not null access Evaluator'Class;
                               text     : String;
                               evalNode : A_Evaluation_Node ) is
        input   : A_Buffer_Stream := Stream( text );
        script  : A_Script;
        context : A_Eval_Context;
    begin
        this.scanner.Set_Input( Stream_Access(input) );
        script := this.parser.Scan_Script( this.scanner );   -- may raise exception
        this.scanner.Expect( TK_EOF );                       -- may raise exception
        if script /= null then
            context := Create_Eval_Context( evalNode );
            script.Evaluate( context );                      -- may raise exception
            Delete( context );
            Delete( script );
        end if;
        this.scanner.Set_Input( null );
        Close( input );
    exception
        when others =>
            Delete( context );
            Delete( script );
            this.scanner.Set_Input( null );
            Close( input );
            raise;
    end Evaluate_Script;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Evaluator ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Scripting.Evaluators;
