--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

package Scripting.Expressions.Operators is

    -- A Binary_Operator has two operands upon which it operates. Evaluating it
    -- produces a single resultant value.
    --
    -- Relative order of operations (highest priority first)
    -- * / %           | Multiplication, division, modulo
    -- + -             | Addition, subtraction
    -- = != < <= > >=  | Comparators
    -- && ||           | Logical operators
    type Binary_Operator is abstract new Expression with private;
    type A_Binary_Operator is access all Binary_Operator'Class;

    -- Creates a new Binary_Operator of a concrete type depending on 'token'.
    function Create_Binary_Operator( token : not null A_Token ) return A_Binary_Operator;

    function Get_Precedence( this : not null access Binary_Operator'Class ) return Positive;

    -- Sets the 'left' and 'right' operands of the operator.
    procedure Set_Operands( this        : not null access Binary_Operator'Class;
                            left, right : in out A_Expression );
    pragma Precondition( left /= null );
    pragma Precondition( right /= null );
    pragma Postcondition( left = null );
    pragma Postcondition( right = null );

    -- Returns True if the token type represents a binary operator.
    function Is_Binary( tokenType : Token_Type ) return Boolean;

    ----------------------------------------------------------------------------

    -- A Unary_Operator has only one operand upon which it operates. Evaluating
    -- it produces a single resultant value.
    type Unary_Operator is abstract new Expression with private;
    type A_Unary_Operator is access all Unary_Operator'Class;

    -- Creates a new Unary_Operator of a concrete type depending on 'token'.
    function Create_Unary_Operator( token : not null A_Token ) return A_Unary_Operator;

    -- Sets the 'right' (and only) operand of the operator.
    procedure Set_Operand( this  : not null access Unary_Operator'Class;
                           right : in out A_Expression );
    pragma Precondition( right /= null );
    pragma Postcondition( right = null );

    -- Returns True if the token type represents a unary operator.
    function Is_Unary( tokenType : Token_Type ) return Boolean;

private

    type Binary_Operator is abstract new Expression with
        record
            precedence : Positive;
            left       : A_Expression;
            right      : A_Expression;
        end record;

    procedure Construct( this       : access Binary_Operator;
                         loc        : Token_Location;
                         precedence : Positive );

    procedure Delete( this : in out Binary_Operator );

    ----------------------------------------------------------------------------

    type Unary_Operator is abstract new Expression with
        record
            right : A_Expression;
        end record;

    procedure Delete( this : in out Unary_Operator );

end Scripting.Expressions.Operators;
