--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Containers;                    use Ada.Containers;
with Scripting.Expressions.Functions;   use Scripting.Expressions.Functions;
with Scripting.Expressions.Literals;    use Scripting.Expressions.Literals;
with Scripting.Expressions.Operators;   use Scripting.Expressions.Operators;
with Scripting.Expressions.Variables;   use Scripting.Expressions.Variables;
with Scripting.Statements;              use Scripting.Statements;
with Scripting.Statements.Methods;      use Scripting.Statements.Methods;
with Tokens;                            use Tokens;

package body Scripting.Parsers is

    -- Scans an expression or raises Parse_Exception if one is not found. If a
    -- boolean operator is encountered, it must have an operator precedence
    -- greater than or equal to 'precedence'. Parse_Exception will be raised if
    -- a parsing error occurs. Token_Exception will be raised if an unrecognized
    -- token is encountered by the scanner.
    --
    -- The grammar implemented by this function is:
    -- expression := <term>
    -- expression := <term> <binary_op> <term>
    function Expect_Expression( this       : not null access Parser'Class;
                                scanner    : not null A_Token_Scanner;
                                precedence : Natural ) return A_Expression;
    pragma Postcondition( Expect_Expression'Result /= null );

    -- Scans an expression or returns null if one is not found. If a boolean
    -- operator is encountered, it must have an operator precedence greater than
    -- or equal to 'precedence'. Parse_Exception will be raised if a parsing
    -- error occurs. Token_Exception will be raised if an unrecognized token is
    -- encountered by the scanner.
    --
    -- The grammar implemented by this function is:
    -- expression := <term>
    -- expression := <term> <binary_op> <term>
    function Scan_Expression( this       : not null access Parser'Class;
                              scanner    : not null A_Token_Scanner;
                              precedence : Natural ) return A_Expression;

    -- Scans a comma-delimited list of expressions or returns null if no
    -- expression is found. Parse_Exception will be raised if a parsing error
    -- occurs. Token_Exception will be raised if an unrecognized token is
    -- encounter by the scanner.
    --
    -- The grammar implemented by this function is:
    -- expression_list := <expression>
    -- expression_list := <expression> ',' <expression_list>
    function Scan_Expression_List( this    : not null access Parser'Class;
                                   scanner : not null A_Token_Scanner ) return Expression_Lists.List;

    -- Scans a function or returns null if one is not found. Parse_Exception
    -- will be raised if a parsing error occurs. Token_Exception will be raised
    -- if an unrecognized token is encounter by the scanner.
    --
    -- The grammar implemented by this function is:
    -- function := _identifier_ '(' [expression_list] ')'
    function Scan_Function( this    : not null access Parser'Class;
                            scanner : not null A_Token_Scanner ) return A_Expression;

    -- Scans an operand or returns null if one is not found. Parse_Exception
    -- will be raised if a parsing error occurs. Token_Exception will be raised
    -- if an unrecognized token is encountered by the scanner.
    --
    -- The grammar implemented by this function is:
    -- operand := _number_
    -- operand := _string_
    -- operand := _identifier_
    -- operand := '(' <expression> ')'
    function Scan_Operand( this    : not null access Parser'Class;
                           scanner : not null A_Token_Scanner ) return A_Expression;

    -- Scans any kind of statement or returns null if one is not found.
    -- Parse_Exception will be raised if a parsing error occurs. Token_Exception
    -- will be raised if an unrecognized token is encountered by the scanner.
    --
    -- The grammar implemented by this function is:
    -- statement := <method> ';'
    function Scan_Statement( this    : not null access Parser'Class;
                             scanner : not null A_Token_Scanner ) return A_Statement;

    -- Scans a list of statements terminated by semicolons, or returns null if
    -- none are found. Parse_Exception will be raised if a parsing error occurs.
    -- Token_Exception will be raised if an unrecognized token is encountered by
    -- the scanner.
    --
    -- The grammar implemented by this function is:
    -- statement_list := <statement> ';'
    -- statement_list := <statement> ';' <statement_list>
    function Scan_Statement_List( this    : not null access Parser'Class;
                                  scanner : not null A_Token_Scanner
                                ) return Statement_Lists.List;

    -- Scans a term or returns null if one is not found. Parse_Exception will
    -- be raised if a parsing error occurs. Token_Exception will be raised if an
    -- unrecognized token is encountered by the scanner.
    --
    -- The grammar implemented by this function is:
    -- term := <operand>
    -- term := <unary_op> <term>
    function Scan_Term( this    : not null access Parser'Class;
                        scanner : not null A_Token_Scanner ) return A_Expression;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Parser return A_Parser is
        this : constant A_Parser := new Parser;
    begin
        this.Construct;
        return this;
    end Create_Parser;

    ----------------------------------------------------------------------------

    function Expect_Expression( this    : not null access Parser'Class;
                                scanner : not null A_Token_Scanner ) return A_Expression is
    begin
        return this.Expect_Expression( scanner, 0 );
    end Expect_Expression;

    ----------------------------------------------------------------------------

    function Expect_Expression( this       : not null access Parser'Class;
                                scanner    : not null A_Token_Scanner;
                                precedence : Natural ) return A_Expression is
        e : constant A_Expression := this.Scan_Expression( scanner, precedence );
    begin
        if e = null then
            raise Parse_Exception with "Expected expression at " &
                To_String( scanner.Get_Location );
        end if;
        return e;
    end Expect_Expression;

    ----------------------------------------------------------------------------

    function Scan_Expression( this    : not null access Parser'Class;
                              scanner : not null A_Token_Scanner ) return A_Expression is
    begin
        return this.Scan_Expression( scanner, 0 );
    end Scan_Expression;

    ----------------------------------------------------------------------------

    -- expression := <term>
    -- expression := <term> <binary_op> <term>
    function Scan_Expression( this       : not null access Parser'Class;
                              scanner    : not null A_Token_Scanner;
                              precedence : Natural ) return A_Expression is
        token : A_Token := null;
        left  : A_Expression := null;
        right : A_Expression := null;
        bop   : A_Binary_Operator := null;
    begin
        left := this.Scan_Term( scanner );
        if left /= null then

            -- we have a left term; check for a binary operator following it
            token := scanner.Accept_Binary_Op;
            while token /= null loop

                -- found a binary operator; check the precedence of it
                bop := Create_Binary_Operator( token );
                if bop.Get_Precedence <= precedence then
                    -- the operator's precedence is lower than, or the same as,
                    -- the current scan level. put it back in the queue so we
                    -- see it later. It will operate on an ancestor of this
                    -- expression (after this expression is evaluated).
                    scanner.Push_Back( token );
                    Delete( A_Expression(bop) );
                    return left;
                else
                    -- found an operator for term 'left'. Require a right-hand
                    -- expression for it.
                    Delete( token );
                    right := this.Expect_Expression( scanner, bop.Get_Precedence );
                    bop.Set_Operands( left, right );   -- consumes both
                    left := A_Expression(bop);
                    bop := null;
                end if;

                token := scanner.Accept_Binary_Op;
            end loop;

        end if;
        return left;
    exception
        when others =>
            Delete( A_Expression(bop) );
            Delete( token );
            Delete( left );
            raise;
    end Scan_Expression;

    ----------------------------------------------------------------------------

    -- expression_list := <expression>
    -- expression_list := <expression> ',' <expression_list>
    function Scan_Expression_List( this    : not null access Parser'Class;
                                   scanner : not null A_Token_Scanner ) return Expression_Lists.List is
        result : Expression_Lists.List;
        e      : A_Expression;
    begin
        -- scan the first expression
        e := this.Scan_Expression( scanner );

        if e /= null then
            loop
                result.Append( e );
                e := null;

                -- if a comma follows, an expression is expected
                if not scanner.Accept_Token( TK_COMMA ) then
                    exit;
                end if;

                e := this.Expect_Expression( scanner );
            end loop;
        end if;

        return result;
    exception
        when others =>
            Delete_Contents( result );
            raise;
    end Scan_Expression_List;

    ----------------------------------------------------------------------------

    -- function := _identifier_ '(' [expression_list] ')'
    function Scan_Function( this    : not null access Parser'Class;
                            scanner : not null A_Token_Scanner ) return A_Expression is
        func  : A_Function;
        args  : Expression_Lists.List;
        name  : A_Identifier_Token;
    begin
        -- look for the function name identifier
        name := scanner.Accept_Identifier;
        if name /= null then
            -- look for the opening parenthesis
            if scanner.Accept_Token( TK_LEFT_PARENTHESIS ) then
                -- we know we found a function now
                -- scan an expression list of arguments
                args := this.Scan_Expression_List( scanner );

                -- expect the closing parenthesis
                scanner.Expect( TK_RIGHT_PARENTHESIS );

                func := Create_Function( name );   -- create the function
                func.Set_Arguments( args );        -- 'args' is consumed

                Delete( A_Token(name) );
            else
                -- an opening parenthesis didn't follow the name so this isn't
                -- a function. put the name back.
                scanner.Push_Back( A_Token(name) );
            end if;
        end if;

        return A_Expression(func);
    exception
        when others =>
            Delete( A_Expression(func) );
            Delete_Contents( args );
            Delete( A_Token(name) );
            raise;
    end Scan_Function;

    ----------------------------------------------------------------------------

    -- method := <function>
    function Scan_Method( this    : not null access Parser'Class;
                          scanner : not null A_Token_Scanner ) return A_Statement is
        method : A_Method;
        func   : A_Function;
    begin
        -- check for a function
        func := A_Function(this.Scan_Function( scanner ));
        if func /= null then
            method := Create_Method;
            method.Set_Function( func );     -- 'func' is consumed
        end if;
        return A_Statement(method);
    end Scan_Method;

    ----------------------------------------------------------------------------

    -- operand := <number>
    -- operand := <string>
    -- operand := <function>
    -- operand := <identifier>
    -- operand := '(' <expression> ')'
    function Scan_Operand( this    : not null access Parser'Class;
                           scanner : not null A_Token_Scanner ) return A_Expression is
        token : A_Token := null;
        left  : A_Expression := null;
    begin
        -- accept a number literal
        token := scanner.Accept_Token( TK_NUMBER );
        if token /= null then
            left := A_Expression(Create_Literal( token ));
        else
            -- accept a string literal
            token := scanner.Accept_Token( TK_STRING );
            if token /= null then
                left := A_Expression(Create_Literal( token ));
            else
                -- accept a function call
                left := this.Scan_Function( scanner );
                if left = null then
                    -- accept a variable name
                    token := scanner.Accept_Token( TK_IDENTIFIER );
                    if token /= null then
                        left := A_Expression(Create_Variable( A_Identifier_Token(token) ));
                    else
                        -- accept a parenthesized expression
                        if scanner.Accept_Token( TK_LEFT_PARENTHESIS ) then
                            left := this.Expect_Expression( scanner );
                            scanner.Expect( TK_RIGHT_PARENTHESIS );
                        else
                            -- no operand was found
                            null;
                        end if;
                    end if;
                end if;
            end if;
        end if;

        Delete( token );
        return left;
    exception
        when others =>
            Delete( token );
            Delete( left );
            raise;
    end Scan_Operand;

    ----------------------------------------------------------------------------

    -- script := <statement_list>
    function Scan_Script( this    : not null access Parser'Class;
                          scanner : not null A_Token_Scanner ) return A_Script is
        use Statement_Lists;

        statements : Statement_Lists.List;
        script     : A_Script;
    begin
        statements := this.Scan_Statement_List( scanner );
        if Length( statements ) > 0 then
            script := Create_Script;                 -- create the script
            script.Set_Statements( statements );     -- 'statements' is consumed
        end if;
        return script;
    end Scan_Script;

    ----------------------------------------------------------------------------

    -- statement := <method>
    function Scan_Statement( this    : not null access Parser'Class;
                             scanner : not null A_Token_Scanner ) return A_Statement is
        statement : A_Statement;
    begin
        -- check for a method statement (function with no return value)
        statement := this.Scan_Method( scanner );
        -- As more statements are added, scan for them here if statement is null
        return statement;
    end Scan_Statement;

    ----------------------------------------------------------------------------

    -- statement_list := <statement> ';'
    -- statement_list := <statement> ';' <statement_list>
    function Scan_Statement_List( this    : not null access Parser'Class;
                                  scanner : not null A_Token_Scanner
                                ) return Statement_Lists.List is
        result    : Statement_Lists.List;
        statement : A_Statement;
    begin
        loop
            -- optionally accept a statement
            statement := this.Scan_Statement( scanner );
            if statement = null then
                exit;
            end if;

            -- if we got a statement, expect a semicolon
            scanner.Expect( TK_SEMICOLON );

            result.Append( statement );
            statement := null;
        end loop;

        return result;
    exception
        when others =>
            Delete( statement );
            Delete_Contents( result );
            raise;
    end Scan_Statement_List;

    ----------------------------------------------------------------------------

    -- term := <operand>
    -- term := <unary_op> <term>
    function Scan_Term( this    : not null access Parser'Class;
                        scanner : not null A_Token_Scanner ) return A_Expression is
        token : A_Token := null;
        uop   : A_Unary_Operator := null;
        right : A_Expression := null;
    begin
        token := scanner.Accept_Unary_Op;
        if token /= null then
            uop := Create_Unary_Operator( token );
            Delete( token );
            right := this.Scan_Term( scanner );
            if right = null then
                raise Parse_Exception with "Expected expression at " &
                    To_String( scanner.Get_Location );
            end if;
            uop.Set_Operand( right );
            right := A_Expression(uop);
            uop := null;
            return right;
        else
            return this.Scan_Operand( scanner );
        end if;
    exception
        when others =>
            Delete( A_Expression(uop) );
            Delete( token );
            raise;
    end Scan_Term;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Parser ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Scripting.Parsers;
