--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
--with Debugging;                         use Debugging;
with GNAT.Directory_Operations;         use GNAT.Directory_Operations;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;

package body Support.Paths is

    -- Returns the location of the '.' character which begins the path's file
    -- extension. If no extension is found, path'First - 1 is returned.
    function Find_Extension( path : String ) return Integer is

        ------------------------------------------------------------------------

        function Is_Dir_Sep( c : Character ) return Boolean is
        begin
            return c = '/' or else c = '\' or else c = '#';
        end Is_Dir_Sep;

        ------------------------------------------------------------------------

        pos : Integer := path'Last;
    begin
        while pos > path'First and then
              not Is_Dir_Sep( path(pos) ) and then
              path(pos) /= '.'
        loop
            pos := pos - 1;
        end loop;
        -- to find the extension, we must be somewhere after the first character,
        -- at or before the last character, at a dot character and the character
        -- before the dot can't be a directory separator.
        if pos + 1 <= path'Last and then pos > path'First and then
           path(pos) = '.' and then not Is_Dir_Sep( path(pos-1) )
        then
           return pos;
        end if;
        return path'First - 1;
    end Find_Extension;

    ----------------------------------------------------------------------------

    function Add_Extension( path,
                            ext   : String;
                            force : Boolean := False ) return String is
    begin
        if ext'Length > 0 then
            if Get_Extension( path )'Length = 0 or else
               (force and then not Ends_With( To_Lower( path ), "." & To_Lower( ext ) ))
            then
                -- add the extension if we're forcing it, or if there isn't an
                -- extension
                return path & "." & ext;
            end if;
        end if;
        return path;
    end Add_Extension;

    ----------------------------------------------------------------------------

    function Ancestor_Dir( fullpath, filename : String ) return String is
    begin
        if Ends_With( fullpath, filename ) then
            return fullpath(fullpath'First..fullpath'Last-filename'Length);
        end if;
        return fullpath;
    end Ancestor_Dir;

    ----------------------------------------------------------------------------

    function App_Data_Directory return String renames OS.App_Data_Directory;

    ----------------------------------------------------------------------------

    function Executable_Extension return String renames OS.Executable_Path;

    ----------------------------------------------------------------------------

    function Execution_Directory return String renames OS.Execution_Directory;

    ----------------------------------------------------------------------------

    function Get_Directory( path : String ) return String is
        pos : Integer := Index( path, Slash, Backward );
    begin
        if pos < path'First then
            pos := Index( path, Other_Slash, Backward );
        end if;
        if pos >= path'First then
            return path(path'First..pos);
        end if;
        return "";
    end Get_Directory;

    ----------------------------------------------------------------------------

    function Get_Extension( path : String ) return String is
        pos : constant Integer := Find_Extension( path );
    begin
        if pos >= path'First then
            return path(pos+1..path'Last);
        end if;
        return "";
    end Get_Extension;

    ----------------------------------------------------------------------------

    function Get_Filename( path : String ) return String is
        pos : Integer := Index( path, Slash, Backward );
    begin
        if pos < path'First then
            pos := Index( path, Other_Slash, Backward );
        end if;
        if pos >= path'First and then pos < path'Last then
            return path(pos+1..path'Last);
        end if;
        return path;
    end Get_Filename;

    ----------------------------------------------------------------------------

    function Home_Directory return String renames OS.Home_Directory;

    ----------------------------------------------------------------------------

    function Make_Dir( path : String ) return Boolean is
        pos : Integer;
    begin
        if path'Length > 0 then
            if not GNAT.OS_Lib.Is_Directory( path ) then
                -- attempt to create 'path' as given
                GNAT.Directory_Operations.Make_Dir( path );
            end if;
            return True;
        else
            return False;
        end if;
    exception
        when Directory_Error =>
            -- failed, try to create its parent directory first
            pos := path'Last;
            -- skip any trailing directory separators
            while pos > path'First and then (path(pos) = '/' or else path(pos) = '\') loop
                pos := pos - 1;
            end loop;
            -- skip to the last directory separator on the path and discard
            -- everything after it
            while pos > path'First and then
                  path(pos) /= '/' and then
                  path(pos) /= '\'
            loop
                if path(pos) = ':' then
                    -- found the drive letter
                    return False;
                end if;
                pos := pos - 1;
            end loop;
            if Make_Dir( path(path'First..pos) ) then
                -- try to make the child directory again
                begin
                    GNAT.Directory_Operations.Make_Dir( path );
                    return True;
                exception
                    when Directory_Error =>
                        return False;
                end;
            else
                return False;
            end if;
    end Make_Dir;

    ----------------------------------------------------------------------------

    function Other_Slash return String is
    begin
        if GNAT.OS_Lib.Directory_Separator = '/' then
            return "\";
        else
            return "/";
        end if;
    end Other_Slash;

    ----------------------------------------------------------------------------

    function Path_Append( basePath, extra : String ) return String is

        ------------------------------------------------------------------------

        function No_Leading( str : String ) return String is
        begin
            if str'Length > 0 and then
               (str(str'First) = '/' or else str(str'First) = '\')
            then
                return No_Leading( str(str'First+1..str'Last) );
            end if;
            return str;
        end No_Leading;

        ------------------------------------------------------------------------

        function No_Trailing( str : String ) return String is
        begin
            if str'Length > 0 and then
               (str(str'Last) = '/' or else str(str'Last) = '\')
            then
                return No_Trailing( str(str'First..str'Last-1) );
            end if;
            return str;
        end No_Trailing;

        ------------------------------------------------------------------------

    begin
        if basePath'Length = 0 and then extra'Length = 0 then
            return "";
        elsif extra'Length = 0 then
            return basePath;
        elsif basePath'Length = 0 then
            return extra;
        end if;

        return No_Trailing( basePath ) & Slash & No_Leading( extra );
    end Path_Append;

    ----------------------------------------------------------------------------

    function Remove_Extension( path : String ) return String is
        pos : constant Integer := Find_Extension( path );
    begin
        if pos >= path'First then
            return path(path'First..pos-1);
        end if;
        return path;
    end Remove_Extension;

    ----------------------------------------------------------------------------

    function Remove_Trailing_Slash( path : String ) return String is
        pos : Integer := path'Last;
    begin
        while pos >= path'First loop
            exit when path(pos) /= '/' and then path(pos) /= '\';
            pos := pos - 1;
        end loop;
        return path(path'First..pos);
    end Remove_Trailing_Slash;

    ----------------------------------------------------------------------------

    procedure Reveal_Path( path : String ) renames OS.Reveal_Path;

    ----------------------------------------------------------------------------

    function Slash return String is
        result : constant String (1..1) := (1 => GNAT.OS_Lib.Directory_Separator);
    begin
        return result;
    end Slash;

    ----------------------------------------------------------------------------

    function System_Font_Directory return String renames OS.System_Font_Directory;

    ----------------------------------------------------------------------------

    function Temp_Directory return String renames OS.Temp_Directory;

end Support.Paths;
