--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

package Support.Paths is

    -- If the path has no extension, or if 'force' is True then the path with
    -- the appended extension is returned. Otherwise, the path is returned
    -- as-is. The value for 'ext' should not include a leading dot character.
    function Add_Extension( path,
                            ext   : String;
                            force : Boolean := False ) return String;
    pragma Precondition( path'Length > 0 );
    pragma Precondition( ext'Length > 0 );

    -- If fullpath ends with filename, the ancestor part of fullpath that
    -- contains filename will be returned. Otherwise, fullpath will be returned.
    --
    -- Example:
    --   fullpath => "C:\my_dir\something\file.dat"
    --   filename => "something\file.dat"
    --   returns "C:\my_dir\"
    function Ancestor_Dir( fullpath, filename : String ) return String;
    pragma Precondition( fullpath'Length > 0 );

    -- Returns the current platform's common application data directory. This is
    -- where the standard application resources may have been installed.
    function App_Data_Directory return String;

    -- Returns the file extension for executable files on the current OS without
    -- a leading dot character.
    function Executable_Extension return String;

    -- Returns the directory containing the executable file, or, if the executable is
    -- running on OS X inside an .app bundle, the bundle's Resources directory.
    function Execution_Directory return String;

    -- Returns the directory portion of a full path, including a directory
    -- separator at the end.
    function Get_Directory( path : String ) return String;

    -- Returns the path's file extension without a leading dot character, or an
    -- empty string if the path has no file extension.
    function Get_Extension( path : String ) return String;

    -- Returns the path's filename if it has one, otherwise an empty string is
    -- returned.
    function Get_Filename( path : String ) return String;

    -- Returns the platform's home directory. The string will have a trailing
    -- directory separator.
    function Home_Directory return String;

    -- Recursively creates directory 'path', returning True on success or False
    -- on error. If the directory already exists, True will be returned.
    function Make_Dir( path : String ) return Boolean;

    -- Returns the opposite slash of the current platform's directory separator
    -- slash character.
    function Other_Slash return String;

    -- Appends 'extra' onto 'basePath', following path naming conventions by
    -- ensuring that extra one directory separator character is inserted between
    -- the two path parts. If both 'basePath' and 'extra' are empty then an
    -- empty string is returned, otherwise if 'basePath' is empty then 'extra'
    -- is returned, or if 'extra' is empty, then 'basePath' is returned.
    function Path_Append( basePath, extra : String ) return String;

    -- Returns a path without its file extension. If the path is a directory or
    -- the file has no extension, it will be returned as-is.
    function Remove_Extension( path : String ) return String;

    -- Removes all trailing slashes (forward and backward) from tail of 'path'.
    function Remove_Trailing_Slash( path : String ) return String;

    -- Opens the directory of 'path' in the platform's GUI and selects the
    -- specific file in 'path', if it points to an existing file. If the
    -- directory of 'path' does not exist then nothing happens.
    procedure Reveal_Path( path : String );

    -- Returns the current platform's directory separator character.
    function Slash return String;
    pragma Postcondition( Slash'Result'Length = 1 );

    -- Returns the current platform's directory for system fonts.
    function System_Font_Directory return String;

    -- Returns the current platform's temporary directory for temporary files.
    function Temp_Directory return String;

end Support.Paths;
