--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Numerics.Elementary_Functions; use Ada.Numerics.Elementary_Functions;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Interfaces;                        use Interfaces;
with Support;                           use Support;

package body Themes.Palette is

    initialized : Boolean := False;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    procedure Initialize is
    begin
        if not initialized then
            pragma Debug( Dbg( "Initializing palette", D_INIT, INFO ) );

            base_colors(Black)       := Al_Map_RGB( 16#00#, 16#00#, 16#00# );
            base_colors(White)       := Al_Map_RGB( 16#FF#, 16#FF#, 16#FF# );
            base_colors(Transparent) := Al_Map_RGBA( 16#00#, 16#00#, 16#00#, 16#00# );
            base_colors(Magenta)     := Al_Map_RGB( 16#FE#, 16#00#, 16#FE# );

            base_colors(Butter1) := Al_Map_RGB( 16#FC#, 16#E9#, 16#4F# );
            base_colors(Butter2) := Al_Map_RGB( 16#ED#, 16#D4#, 16#00# );
            base_colors(Butter3) := Al_Map_RGB( 16#C4#, 16#A0#, 16#00# );

            base_colors(Orange1) := Al_Map_RGB( 16#FC#, 16#AF#, 16#3E# );
            base_colors(Orange2) := Al_Map_RGB( 16#F5#, 16#79#, 16#00# );
            base_colors(Orange3) := Al_Map_RGB( 16#CE#, 16#5C#, 16#00# );

            base_colors(Chocolate1) := Al_Map_RGB( 16#E9#, 16#B9#, 16#6E# );
            base_colors(Chocolate2) := Al_Map_RGB( 16#C1#, 16#7D#, 16#11# );
            base_colors(Chocolate3) := Al_Map_RGB( 16#8F#, 16#59#, 16#02# );

            base_colors(Chameleon1) := Al_Map_RGB( 16#8A#, 16#E2#, 16#34# );
            base_colors(Chameleon2) := Al_Map_RGB( 16#73#, 16#D2#, 16#16# );
            base_colors(Chameleon3) := Al_Map_RGB( 16#4E#, 16#9A#, 16#06# );

            base_colors(Skyblue1) := Al_Map_RGB( 16#72#, 16#9F#, 16#CF# );
            base_colors(Skyblue2) := Al_Map_RGB( 16#34#, 16#65#, 16#A4# );
            base_colors(Skyblue3) := Al_Map_RGB( 16#20#, 16#4A#, 16#87# );

            base_colors(Plum1) := Al_Map_RGB( 16#AD#, 16#7F#, 16#A8# );
            base_colors(Plum2) := Al_Map_RGB( 16#75#, 16#50#, 16#7B# );
            base_colors(Plum3) := Al_Map_RGB( 16#5C#, 16#35#, 16#66# );

            base_colors(Scarlet1) := Al_Map_RGB( 16#EF#, 16#29#, 16#29# );
            base_colors(Scarlet2) := Al_Map_RGB( 16#CC#, 16#00#, 16#00# );
            base_colors(Scarlet3) := Al_Map_RGB( 16#A4#, 16#00#, 16#00# );

            base_colors(Chrome1) := Al_Map_RGB( 16#EE#, 16#EE#, 16#EC# );
            base_colors(Chrome2) := Al_Map_RGB( 16#D3#, 16#D7#, 16#CF# );
            base_colors(Chrome3) := Al_Map_RGB( 16#BA#, 16#BD#, 16#B6# );
            base_colors(Chrome4) := Al_Map_RGB( 16#88#, 16#8A#, 16#85# );
            base_colors(Chrome5) := Al_Map_RGB( 16#55#, 16#57#, 16#53# );
            base_colors(Chrome6) := Al_Map_RGB( 16#2E#, 16#34#, 16#36# );

            initialized := True;
        end if;
    end Initialize;

    ----------------------------------------------------------------------------

    function Compare( a, b : Allegro_Color; tolerance : Natural := 0 ) return Boolean is
        a_r, a_g, a_b, a_a : Unsigned_8;
        b_r, b_g, b_b, b_a : Unsigned_8;
    begin
        if a = b then
            return True;
        end if;

        Al_Unmap_RGBA( a, a_r, a_g, a_b, a_a );
        Al_Unmap_RGBA( b, b_r, b_g, b_b, b_a );

        return abs (Integer(a_r) - Integer(b_r)) <= Integer(tolerance) and then
               abs (Integer(a_g) - Integer(b_g)) <= Integer(tolerance) and then
               abs (Integer(a_b) - Integer(b_b)) <= Integer(tolerance) and then
               abs (Integer(a_a) - Integer(b_a)) <= Integer(tolerance);
    end Compare;

    ----------------------------------------------------------------------------

    function Contrast( fg, bg : Allegro_Color; ratio : Float := 1.75 ) return Allegro_Color is

        ------------------------------------------------------------------------

        function Luminance( c : Allegro_Color ) return Float is

            function R( v : Unsigned_8 ) return Float is
            begin
                return (Float(v) / 255.0) ** 2.2;
            end R;

            c_r, c_g, c_b, c_a : Unsigned_8;
        begin
            Al_Unmap_RGBA( c, c_r, c_g, c_b, c_a );
            return (0.2126 * R( c_r ) +
                    0.7152 * R( c_g ) +
                    0.0722 * R( c_b ) + 0.05) / 1.05;
        end Luminance;

        ------------------------------------------------------------------------

        -- todo: Re-examine this later and use Al_Unmap_RGBA_f

        function Luminance_Adjust( c : Allegro_Color; val : Float ) return Allegro_Color is

            function Adjust( v : Unsigned_8; c : Float ) return Unsigned_8 is
            begin
                return Unsigned_8(Float'Max(0.0, Float'Min(255.0, Float(v) - 255.0 * val * c)));
            end Adjust;

            r, g, b, a : Unsigned_8;
        begin
            Al_Unmap_RGBA( c, r, g, b, a );
            return Al_Map_RGBA( Adjust( r, 0.2126 ),
                                Adjust( g, 0.7152 ),
                                Adjust( b, 0.0722 ),
                                a );
        end Luminance_Adjust;

        ------------------------------------------------------------------------

        bg_lum : constant Float := Luminance( bg );
        fg_lum : constant Float := Luminance( fg );
    begin
        if Float'Max( bg_lum, fg_lum ) / Float'Min( bg_lum, fg_lum ) > ratio then
            return fg;
        end if;
        return Luminance_Adjust( fg, Float'Min( Float'Max( bg_lum, fg_lum ), 1.0 ) );
    end Contrast;

    ----------------------------------------------------------------------------

    function Hard_Contrast( bg : Allegro_Color ) return Allegro_Color is

        function Is_Dark( color : Allegro_Color ) return Boolean is

            function Luminance( r, g, b : Float ) return Float is
                weight_red   : constant Float := 0.3;
                weight_green : constant Float := 0.59;
                weight_blue  : constant Float := 0.11;
            begin
                return weight_red * r +
                       weight_green * g +
                       weight_blue * b;
            end Luminance;

            r, g, b : Unsigned_8;
        begin
            Al_Unmap_RGB( color, r, g, b );
            return Luminance( Float(r), Float(g), Float(b) ) <= 128.0;
        end Is_Dark;

        r, g, b, a : Unsigned_8;
    begin
        Al_Unmap_RGBA( bg, r, g, b, a );
        if Is_Dark( bg ) then
            return Al_Map_RGBA( 255, 255, 255, a );
        else
            return Al_Map_RGBA( 0, 0, 0, a );
        end if;
    end Hard_Contrast;

    ----------------------------------------------------------------------------

    function Lighten( color : Allegro_Color; factor : Float ) return Allegro_Color is

        function "*"( left : Unsigned_8; right : Float ) return Unsigned_8 is
        begin
            return Unsigned_8(Integer'Min( Integer(Float(left) * right), 255 ));
        end "*";
        pragma Inline_Always( "*" );

        r, g, b, a : Unsigned_8;
    begin
        Al_Unmap_RGBA( color, r, g, b, a );
        return Al_Map_RGBA( r * factor, g * factor, b * factor, a );
    end Lighten;

    ----------------------------------------------------------------------------

    function Make_Grey( brightness : Natural ) return Allegro_Color is
        b : constant Unsigned_8 := Unsigned_8(Constrain( brightness, 0, 255 ));
    begin
        return Al_Map_RGBA( b, b, b, 255 );
    end Make_Grey;

end Themes.Palette;
