--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Unchecked_Deallocation;
with Allegro.Bitmaps.Drawing;           use Allegro.Bitmaps.Drawing;
with Allegro.Bitmaps.Image_IO;          use Allegro.Bitmaps.Image_IO;
with Allegro.Blending;                  use Allegro.Blending;
with Allegro.Color;                     use Allegro.Color;
with Allegro.Drawing;                   use Allegro.Drawing;
with Allegro.State;                     use Allegro.State;
--with Debugging;                         use Debugging;
with Support;                           use Support;
with Tiles.Atlases.Bins;                use Tiles.Atlases.Bins;

package body Tiles.Atlases is

    function Create_Atlas( display    : not null A_Allegro_Display;
                           pageWidth,
                           pageHeight : Positive ) return A_Atlas is
        this : constant A_Atlas := new Atlas;
    begin
        this.Construct( display, pageWidth, pageHeight );
        return this;
    end Create_Atlas;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this       : in out Atlas;
                         display    : not null A_Allegro_Display;
                         pageWidth,
                         pageHeight : Positive ) is
    begin
        Limited_Object(this).Construct;
        this.display := display;
        this.pageWidth := Integer'Min( pageWidth, Al_Get_Display_Option( display, ALLEGRO_MAX_BITMAP_SIZE ) );
        this.pageHeight := Integer'Min( pageHeight, Al_Get_Display_Option( display, ALLEGRO_MAX_BITMAP_SIZE ) );
        this.Add_Page;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Atlas ) is
        procedure Free is new Ada.Unchecked_Deallocation( Page, A_Page );
        p : A_Page;
    begin
        while not this.pages.Is_Empty loop
            p := this.pages.First_Element;
            Al_Destroy_Bitmap( p.bmp );
            Delete( A_Bin(p.bin) );
            Free( p );

            this.pages.Delete_First;
        end loop;

        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Add_Bitmap( this : not null access Atlas'Class;
                          bmp  : in out A_Allegro_Bitmap ) is

        ------------------------------------------------------------------------

        procedure Draw_And_Fetch( page     : A_Page;
                                  bmp      : in out A_Allegro_Bitmap;
                                  location : Rect ) is
            bo     : Allegro_Blend_Operations;
            bs, bd : Allegro_Blend_Mode;
            state  : aliased Allegro_State;
        begin
            Al_Store_State( state, ALLEGRO_STATE_ALL );

            Al_Set_Target_Bitmap( page.bmp );
            Al_Get_Blender( bo, bs, bd );
            Al_Set_Blender( ALLEGRO_ADD, ALLEGRO_ONE, ALLEGRO_ZERO );
            Al_Draw_Bitmap( bmp, Float(location.x), Float(location.y), 0 );
            Al_Set_Blender( bo, bs, bd );

            Al_Destroy_Bitmap( bmp );
            bmp := Al_Create_Sub_Bitmap( page.bmp,
                                         location.x, location.y,
                                         location.width, location.height );

            Al_Restore_State( state );
        end Draw_And_Fetch;

        ------------------------------------------------------------------------

        pos : Page_Lists.Cursor;
        p   : A_Page;
        r   : Rect;
    begin
        if bmp = null then
            return;
        end if;

        pos := this.pages.First;
        while Page_Lists.Has_Element( pos ) loop
            p := Page_Lists.Element( pos );
            if not p.full then
                r := p.bin.Insert( Al_Get_Bitmap_Width( bmp ),
                                   Al_Get_Bitmap_Height( bmp ) );
                if r.height /= 0 then
                    Draw_And_Fetch( p, bmp, r );
                    return;
                elsif Integer'Max( Al_Get_Bitmap_Width( bmp ),
                                   Al_Get_Bitmap_Height( bmp ) ) <= 16 then
                    -- If a bitmap this small didn't fit, then consider it full
                    p.full := True;
                end if;
            end if;
            Page_Lists.Next( pos );
        end loop;

        this.Add_Page;
        p := this.pages.Last_Element;
        r := p.bin.Insert( Al_Get_Bitmap_Width( bmp ), Al_Get_Bitmap_Height( bmp ) );
        if r.height /= 0 then
            Draw_And_Fetch( p, bmp, r );
        else
            -- the bitmap still didn't fit on a page; this shouldn't happen but
            -- if it does, just don't put it in the atlas.
            null;
        end if;
    end Add_Bitmap;

    ----------------------------------------------------------------------------

    procedure Add_Page( this : not null access Atlas'Class ) is
        p     : constant A_Page := new Page;
        state : aliased Allegro_State;
    begin
        Al_Store_State( state, ALLEGRO_STATE_ALL );
        p.bin := Create_Bin( this.pageWidth, this.pageHeight );
        p.bmp := Al_Create_Bitmap( this.pageWidth, this.pageHeight );
        Al_Set_Target_Bitmap( p.bmp );
        Al_Clear_To_Color( Al_Map_RGB( 255, 0, 255 ) );
        this.pages.Append( p );
        Al_Restore_State( state );
    end Add_Page;

    ----------------------------------------------------------------------------

    procedure Save_Bitmaps( this : not null access Atlas'Class;
                            name : String ) is
        pos : Page_Lists.Cursor;
        n : Integer := 1;
    begin
        pos := this.pages.First;
        while Page_Lists.Has_Element( pos ) loop
            if Al_Save_Bitmap( name & Image( n ) & ".png",
                               Page_Lists.Element( pos ).bmp ) then
                null;
            end if;
            n := n + 1;
            Page_Lists.Next( pos );
        end loop;
    end Save_Bitmaps;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Atlas ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Tiles.Atlases;
