--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Tags;                          use Ada.Tags;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;

package body Tiles.Catalogs is

    -- A string that identifies the tile catalog file format. It is the first
    -- data found in the file.
    CATALOG_IDENTIFIER : constant String := "cat";

    -- The file format version number for tile catalogs.
    FORMAT_VERSION : constant := 6;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Catalog return A_Catalog is
        this : constant A_Catalog := new Catalog;
    begin
        this.Construct;
        return this;
    end Create_Catalog;

    ----------------------------------------------------------------------------

    procedure Adjust( this : access Catalog ) is
        pragma Unreferenced( this );
    begin
        raise COPY_NOT_ALLOWED;
    end Adjust;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out Catalog ) is

        ------------------------------------------------------------------------

        procedure Delete_Tile( pos : Tile_Maps.Cursor ) is
            tile : A_Tile := Tile_Maps.Element( pos );
        begin
            Delete( tile );
            this.idMap.Replace_Element( pos, null );
        end Delete_Tile;

        ------------------------------------------------------------------------

        procedure Delete_Matrix( pos : Matrix_Vectors.Cursor ) is
            matrix : A_Tile_Matrix := Matrix_Vectors.Element( pos );
        begin
            Delete( matrix );
        end Delete_Matrix;

        ------------------------------------------------------------------------

    begin
        this.idMap.Iterate( Delete_Tile'Access );
        this.slotArray.Clear;
        this.matArray.Iterate( Delete_Matrix'Access );
        this.matArray.Clear;
        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Add_Matrix( this   : not null access Catalog'Class;
                          matrix : in out A_Tile_Matrix ) is
    begin
        this.matArray.Append( matrix );
        matrix := null;
    end Add_Matrix;

    ----------------------------------------------------------------------------

    procedure Add_Tile( this : not null access Catalog'Class;
                        tile : in out A_Tile ) is
    begin
        if tile /= null then
            if tile.Get_Id /= 0 then
                if not this.idMap.Contains( tile.Get_Id ) then
                    this.idMap.Insert( tile.Get_Id, tile );
                else
                    raise DUPLICATE_TILE with
                          "Duplicate tile ID:" & Natural'Image( tile.Get_Id );
                end if;
            end if;
            this.slotArray.Append( tile );
            tile := null;
        else
            this.slotArray.Append( null );
        end if;
    end Add_Tile;

    ----------------------------------------------------------------------------

    function Find_Tile( this : not null access Catalog'Class;
                        name : String ) return A_Tile is

        function Search( name : String ) return A_Tile is
            lname : constant String := To_Lower( name );
            crsr  : Tile_Maps.Cursor := this.idMap.First;
            tile  : A_Tile;
        begin
            while Tile_Maps.Has_Element( crsr ) loop
                tile := Tile_Maps.Element( crsr );
                if lname = To_Lower( tile.Get_Name ) then
                    return tile;
                end if;
                crsr := Tile_Maps.Next( crsr );
            end loop;
            return null;
        end Search;

        tile : A_Tile;
    begin
        if Get_Extension( name )'Length = 0 then
            tile := Search( Add_Extension( name, "png" ) );
            if tile = null then
                tile := Search( Add_Extension( name, "bmp" ) );
            end if;
        else
            tile := Search( name );
        end if;
        return tile;
    end Find_Tile;

    ----------------------------------------------------------------------------

    function Get_Matrix( this  : not null access Catalog'Class;
                         index : Natural ) return A_Tile_Matrix is
    begin
        if index > 0 and then index <= Integer(this.matArray.Length) then
            return this.matArray.Element( index );
        end if;
        return null;
    end Get_Matrix;

    ----------------------------------------------------------------------------

    function Get_Matrix_Count( this : not null access Catalog'Class ) return Natural is
    begin
        return Natural(this.matArray.Length);
    end Get_Matrix_Count;

    ----------------------------------------------------------------------------

    function Get_Tile( this : not null access Catalog'Class;
                       id   : Natural ) return A_Tile is
    begin
        return this.idMap.Element( id );
    exception
        when Constraint_Error =>
            return null;
    end Get_Tile;

    ----------------------------------------------------------------------------

    procedure Iterate_By_Id( this    : not null access Catalog'Class;
                             examine : not null access procedure( tile : not null A_Tile ) ) is

        procedure Local_Examine( pos : Tile_Maps.Cursor ) is
        begin
            examine.all( Tile_Maps.Element( pos ) );
        end Local_Examine;

    begin
        this.idMap.Iterate( Local_Examine'Access );
    end Iterate_By_Id;

    ----------------------------------------------------------------------------

    procedure Iterate_By_Slot( this    : not null access Catalog'Class;
                               examine : not null access procedure( slot : Positive;
                                                                    tile : A_Tile ) ) is
    begin
        for i in 1..this.slotArray.Length loop
            examine.all( Positive(i), this.slotArray.Element( Positive(i) ) );
        end loop;
    end Iterate_By_Slot;

    ----------------------------------------------------------------------------

    function Object_Input( stream : access Root_Stream_Type'Class ) return Catalog is
        this : aliased Catalog;
    begin
        this.Construct;
        Catalog'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    procedure Object_Output( stream : access Root_Stream_Type'Class; obj : Catalog ) is
    begin
        Catalog'Write( stream, obj );
    end Object_Output;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class;
                           obj    : out Catalog ) is
    begin
        Object'Read( stream, Object(obj) );

        -- read the tile list
        declare
            count : constant Natural := Natural'Input( stream );
            tile  : A_Tile;
        begin
            for i in 1..count loop
                tile := A_Tile'Input( stream );
                obj.Add_Tile( tile );
            end loop;
        end;

        -- read the matrix list
        declare
            count  : constant Natural := Natural'Input( stream );
            matrix : A_Tile_Matrix;
        begin
            for i in 1..count loop
                matrix := A_Tile_Matrix'Input( stream );
                obj.Add_Matrix( matrix );
            end loop;
        end;
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class;
                            obj    : Catalog ) is
    begin
        Object'Write( stream, Object(obj) );

        -- write the tile list
        Natural'Output( stream, Natural(obj.slotArray.Length) );
        for i in 1..obj.slotArray.Length loop
            A_Tile'Output( stream, obj.slotArray.Element( Positive(i) ) );
        end loop;

        -- write the matrix list
        Natural'Output( stream, Natural(obj.matArray.Length) );
        for i in 1..obj.matArray.Length loop
            A_Tile_Matrix'Output( stream, obj.matArray.Element( Positive(i) ) );
        end loop;
    end Object_Write;

    --==========================================================================

    function A_Catalog_Input( stream : access Root_Stream_Type'Class ) return A_Catalog is
        catalog : A_Catalog := null;
        ident   : String(1..CATALOG_IDENTIFIER'Length);
        version : Integer := 0;
    begin
        -- verify the header info
        String'Read( stream, ident );
        if ident = CATALOG_IDENTIFIER then
            version := Integer'Input( stream );
            if version = Catalog_Format_Version then
                if Boolean'Input( stream ) then
                    A_Catalog'Read( stream, catalog );
                end if;
            else
                -- unrecognized version
                raise Constraint_Error with "Unrecognized catalog file version: " &
                                            Image( version );
            end if;
        else
            -- invalid identifier
            raise Constraint_Error with "Unrecognized file format";
        end if;

        return catalog;
    exception
        when Constraint_Error =>
            raise;
        when e : others =>
            -- file too short
            Dbg( "Error: Tiles.Catalogs Exception: " & Exception_Name( e ) &
                 " - " & Exception_Message( e ), D_TILES, ERROR );
            raise Constraint_Error with "Unrecognized file format";
    end A_Catalog_Input;

    ----------------------------------------------------------------------------

    procedure A_Catalog_Output( stream : access Root_Stream_Type'Class;
                                obj    : A_Catalog ) is
    begin
        -- write the header info
        String'Write( stream, CATALOG_IDENTIFIER );
        Integer'Output( stream, Catalog_Format_Version );
        Boolean'Output( stream, obj /= null );
        if stream /= null then
            A_Catalog'Write( stream, obj );
        end if;
    end A_Catalog_Output;

    ----------------------------------------------------------------------------

    procedure A_Catalog_Read( stream : access Root_Stream_Type'Class;
                              obj    : out A_Catalog ) is
    begin
        obj := new Catalog'Class'(Catalog'Class'Input( stream ));
    exception
        when e : Tag_Error =>
            declare
                msg : constant String := Exception_Message( e );
            begin
                raise Constraint_Error with "Found unknown class " &
                      "<" & msg((Index( msg, ".", Backward ) + 1)..msg'Last) & ">";
            end;
    end A_Catalog_Read;

    ----------------------------------------------------------------------------

    procedure A_Catalog_Write( stream : access Root_Stream_Type'Class;
                               obj    : A_Catalog ) is
    begin
        Catalog'Class'Output( stream, obj.all );
    end A_Catalog_Write;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Catalog ) is
    begin
        Delete( A_Object(this) );
    end Delete;

    --==========================================================================

    function Catalog_Format_Version return Natural is
    begin
        return FORMAT_VERSION;
    end Catalog_Format_Version;

end Tiles.Catalogs;
