--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Calendar;                      use Ada.Calendar;
with Ada.Calendar.Formatting;
with Ada.Directories;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Real_Time;                     use Ada.Real_Time;
with Ada.Streams.Stream_IO;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Text_IO;
with Applications;                      use Applications;
with GNAT.OS_Lib;
with Resources.Images;                  use Resources.Images;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;
with Values.Booleans;                   use Values.Booleans;
with Values.Lists;                      use Values.Lists;
with Values.Numbers;                    use Values.Numbers;
with Values.Strings;                    use Values.Strings;
with Version;
with Zip;                               use Zip;
with Zip.Compress;
with Zip.Create;                        use Zip.Create;
with Zip_Streams;                       use Zip_Streams;

package body Tiles.Libraries.Compilers is

    MAX_COLUMNS_PER_LINE : constant := 10;

    type String_Array is array (Integer range <>) of Unbounded_String;

    ----------------------------------------------------------------------------

    -- Parses 'str' as a value. If it begins with a digit, it is assumed
    -- to be a number. If it begins with '[', it is assumed to be a
    -- list, otherwise it is a string (no quotes necessary). The parsed Value
    -- will be returned, or an ATTRIBUTE_ERROR will be raised on error.
    function Parse_Value( str : String ) return Value_Ptr is

        ------------------------------------------------------------------------

        function Parse_Id_List( str : String ) return Value_Ptr is
            idx : Integer := str'First;

            --------------------------------------------------------------------

            procedure Expect( c : Character ) is
            begin
                if idx > str'Last or else str(idx) /= c then
                    raise ATTRIBUTE_ERROR with "Expected '" & c & "'";
                end if;
                idx := idx + 1;
            end Expect;

            --------------------------------------------------------------------

            result     : constant List_Ptr := Create_List;
            start      : Integer := str'First + 1;
            digitCount : Natural := 0;   -- number of digits seen in an id
        begin
            Expect( '[' );
            while idx <= str'Last and then str(idx) /= ']' loop
                if str(idx) = ',' then
                    if digitCount = 0 then
                        -- found a comma following the opening parenthesis
                        -- or another comma. we expected to find an id.
                        raise ATTRIBUTE_ERROR with "Expected tile id";
                    end if;
                    -- record the tile id prior to the comma
                    result.Get.Append( Numbers.Create( Integer'Max( 0, Integer'Value( str(start..idx-1) ) ) ) );
                    digitCount := 0;
                elsif '0' <= str(idx) and then str(idx) <= '9' then
                    digitCount := digitCount + 1;
                    if digitCount = 1 then
                        start := idx;
                    elsif digitCount > 9 then
                        -- way too many digits in this tile id number
                        raise ATTRIBUTE_ERROR with "Tile id too large";
                    end if;
                else
                    -- found an invalid character in the list
                    -- (only digits and commas are allowed)
                    raise ATTRIBUTE_ERROR with "Invalid character '" & str(idx) & "'";
                end if;
                idx := idx + 1;
            end loop;
            if digitCount = 0 then
                raise ATTRIBUTE_ERROR with "Expected tile id";
            else
                -- record the tile id prior to the closing bracket
                result.Get.Append( Numbers.Create( Integer'Max( 0, Integer'Value( str(start..idx-1) ) ) ) );
            end if;
            Expect( ']' );
            if idx <= str'Last then
                raise ATTRIBUTE_ERROR with "Unexpected characters after list";
            end if;

            return result.As_Value;
        end Parse_Id_List;

        ------------------------------------------------------------------------

    begin
        if str(str'First) = '[' then
            return Parse_Id_List( str );
        elsif str(str'First) = '-' or else ('0' <= str(str'First) and then str(str'First) <= '9') then
            return Value_Ptr(Numbers.Create( Integer'Value( str ) ));
        else
            return Value_Ptr(Strings.Create( str ));
        end if;
    end Parse_Value;

    ----------------------------------------------------------------------------

    procedure Split_On_Spaces( str   : Unbounded_String;
                               items : out String_Array;
                               last  : out Integer ) is
        i    : Integer := items'First;
        low  : Integer := 1;
        high : Integer := Index( str, " ", low ) - 1;
    begin
        if high < low then
            high := Length( str );
        end if;

        while low <= high and then i <= items'Last loop
            -- grab the current item
            items(i) := Unbounded_Slice( str, low, high );

            -- skip whitespace
            low := high + 1;
            while low <= Length( str ) and then Element( str, low ) = ' ' loop
                low := low + 1;
            end loop;

            -- find the end of the next item
            high := Index( str, " ", low ) - 1;
            if high < low then
                high := Length( str );
            end if;

            i := i + 1;
        end loop;

        -- check if there were more tokens on the line that we had to skip
        if low <= high then
           last := i;
        else
           last := i - 1;
        end if;
    end Split_On_Spaces;

    ----------------------------------------------------------------------------

    function Trim_Whitespace_And_Comments( str : String ) return Unbounded_String is
        last : Integer := Index( str, "--" );
    begin
        if last < str'First then
           last := str'Last;
        else
            last := last - 1;
        end if;
        if last >= str'First and then str(last) = ASCII.CR then
            -- remove Windows CR characters on *nix systems
            last := last - 1;
        end if;
        return To_Unbounded_String( Trim( str(str'First..last), Both ) );
    end Trim_Whitespace_And_Comments;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Compiler( output : A_Output_Proc ) return A_Compiler is
        this : constant A_Compiler := new Compiler;
    begin
        this.Construct( output );
        return this;
    end Create_Compiler;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Compiler; output : A_Output_Proc ) is
    begin
        Limited_Object(this.all).Construct;
        this.output := output;
    end Construct;

    ----------------------------------------------------------------------------

    function Add_File( this      : not null access Compiler'Class;
                       pathInZip : String;
                       filePath  : String ) return Boolean is
        use String_Maps;
        pos      : String_Maps.Cursor;
        inserted : Boolean;
    begin
        this.files.Insert( pathInZip, filePath, pos, inserted );
        if not inserted then
            -- does the existing record match the source file? (not an error)
            return Ada.Strings.Equal_Case_Insensitive( Element( pos ), filePath );
        else
            -- inserted ok
            if Is_Format_Supported( Get_Extension( filePath ) ) then
                this.imageCount := this.imageCount + 1;
            end if;
            return True;
        end if;
    end Add_File;

    ----------------------------------------------------------------------------

    function Compile( this           : not null access Compiler'Class;
                      libPath        : String;
                      tileListPath   : String;
                      matrixListPath : String;
                      compress       : Boolean ) return Boolean is
    begin
        this.error := False;
        this.imageCount := 0;
        this.tileCount := 0;
        this.matrixCount := 0;
        this.files.Clear;

        if not Equal_Case_Insensitive( Get_Extension( libPath ), Library_Extension ) then
            -- invalid library extension
            this.Put_Error( "Invalid library file extension: " &
                            Get_Extension( libPath ) );
            return False;
        end if;

        -- - - - - - - - - - - - - - - -
        -- Setup directories

        -- ensure the output directory for the archive exists
        this.outputDir := To_Unbounded_String( Get_Directory( libPath ) );
        if not GNAT.OS_Lib.Is_Directory( To_String( this.outputDir ) ) then
            if not Make_Dir( To_String( this.outputDir ) ) then
                this.Put_Error( "Failed to create directory: " &
                                To_String( this.outputDir ) );
                return False;
            end if;
        end if;

        -- create a temporary directory
        this.tempDir := To_Unbounded_String( Temp_Directory & "tmp" & Image( Random_32 ) & Slash );
        while GNAT.OS_Lib.Is_Directory( To_String( this.tempDir ) ) loop
            this.tempDir := To_Unbounded_String( Temp_Directory & "tmp" & Image( Random_32 ) & Slash );
        end loop;
        if not Make_Dir( To_String( this.tempDir ) ) then
            this.Put_Error( "Failed to create temporary directory" );
            return False;
        end if;

        -- - - - - - - - - - - - - - - -
        -- Create library from sources

        this.lib := Create_Library( Remove_Extension( Get_Filename( libPath ) ) );

        if not this.error then
            this.Read_Tile_List( tileListPath );
        end if;

        if not this.error then
             this.Read_Matrix_List( matrixListPath );
        end if;

        -- - - - - - - - - - - - - - - -
        -- Write library files

        if not this.error then
            this.Write_Catalog_File( To_String( this.tempDir ) & CATALOG_FILENAME );
        end if;
        if not this.error then
            this.Write_Info_File( To_String( this.tempDir ) & INFO_FILENAME );
        end if;

        if not this.error then
            this.Write_Library_File( libPath, compress );
        end if;

        -- - - - - - - - - - - - - - - -
        -- Cleanup

        Delete( this.lib );

        begin
            Ada.Directories.Delete_Tree( To_String( this.tempDir ) );
        exception
            when others =>
                -- failed to clean up temporary directory
                this.Put_Line( "Failed to delete temporary directory", error => False );
        end;

        return not this.error;
    exception
        when e : others =>
            this.Put_Error( "Exception: " & Exception_Name( e ) & " - " &
                            Exception_Message( e ) );
            Delete( this.lib );
            return False;
    end Compile;

    ----------------------------------------------------------------------------

    procedure Put_Line( this  : not null access Compiler'Class;
                        msg   : String;
                        error : Boolean := False ) is
    begin
        this.error := this.error or error;
        if this.output /= null then
            this.output.all( msg, error );
        end if;
    end Put_Line;

    ----------------------------------------------------------------------------

    procedure Put_Error( this : not null access Compiler'Class;
                         msg  : String ) is
    begin
        this.Put_Line( msg, error => True );
    end Put_Error;

    ----------------------------------------------------------------------------

    procedure Read_Matrix_List( this : not null access Compiler'Class;
                                path : String ) is
        line : Positive := 1;

        ------------------------------------------------------------------------

        procedure Parse_Dimensions( text   : Unbounded_String;
                                    width,
                                    height : out Natural ) is
            tokens : String_Array(1..3);
            last   : Integer;
        begin
            Split_On_Spaces( text, tokens, last );
            width := 0;
            height := 0;

            if last = 3 then
                -- parse the width
                width := Natural'Value( To_String( tokens(1) ) );
                if width > 0 then
                    -- parse the 'x'
                    if To_String( tokens(2) ) = "x" or else To_String( tokens(2) ) = "X" then
                        height := Natural'Value( To_String( tokens(3) ) );
                        if height = 0 then
                            this.Put_Error( path & ":" & Image( line ) &
                                            ": Expected positive height" );
                        elsif width > MAX_MATRIX_SIZE or else height > MAX_MATRIX_SIZE then
                            this.Put_Error( path & ":" & Image( line ) &
                                            ": Matrix dimensions too large" );
                        end if;
                    else
                        this.Put_Error( path & ":" & Image( line ) &
                                        ": Expected token ""x""" );
                    end if;
                else
                    this.Put_Error( path & ":" & Image( line ) &
                                    ": Expected positive width" );
                end if;
            else
                this.Put_Error( path & ":" & Image( line ) &
                                ": Expected matrix dimensions" );
            end if;
        exception
            when Constraint_Error =>
                this.Put_Error( path & ":" & Image( line ) & ": Syntax error" );
        end Parse_Dimensions;

        ------------------------------------------------------------------------

        procedure Parse_Row( text   : Unbounded_String;
                             matrix : in out Tile_Matrix;
                             row    : Positive ) is
            tokens : String_Array(matrix'Range(2));
            last   : Integer;
            id     : Natural;
        begin
            Split_On_Spaces( text, tokens, last );
            if last = tokens'Last then
                for c in tokens'Range loop
                    begin
                        id := Natural'Value( To_String( tokens(c) ) );
                        if id = 0 or else this.lib.Get_Tile( id ) /= null then
                            matrix(row, c) := id;
                        else
                            this.Put_Error( path & ":" & Image( line ) &
                                            ": Undefined id in column " & Image( c ) );
                        end if;
                    exception
                        when Constraint_Error =>
                            this.Put_Error( path & ":" & Image( line ) &
                                            ": Expected id in column " & Image( c ) );
                    end;
                    exit when this.error;
                end loop;
            else
                this.Put_Error( path & ":" & Image( line ) & ": Expected " &
                                Image( Integer(matrix'Length(2)) ) &
                                " matrix columns" );
            end if;
        end Parse_Row;

        ------------------------------------------------------------------------

        use Ada.Text_IO;

        file   : Ada.Text_IO.File_Type;
        text   : Unbounded_String;
        width,
        height : Natural;
        matrix : A_Tile_Matrix;
        row    : Positive;
    begin
        -- open the matrix list file
        Open( file, In_File, path );    -- throws Name_Error on failure

        -- loop over matrices in the file (outer loop)
        while not End_Of_File( file ) and not this.error loop
            text := Trim_Whitespace_And_Comments( Get_Line( file ) );
            if Length( text ) > 0 then

                -- accept matrix dimensions
                Parse_Dimensions( text, width, height );
                if this.error then
                    exit;
                end if;

                matrix := new Tile_Matrix(1..height, 1..width);
                row := 1;

                -- loop over the rows in the matrix (inner loop)
                while not End_Of_File( file ) and not this.error loop
                    text := Trim_Whitespace_And_Comments( Get_Line( file ) );
                    if Length( text ) > 0 then

                        Parse_Row( text, matrix.all, row );
                        if not this.error then
                            row := row + 1;
                            if row > matrix'Length(1) then
                                -- inner row loop completed; break out
                                this.lib.Add_Matrix( matrix );
                                this.matrixCount := this.matrixCount + 1;
                                exit;
                            end if;
                        end if;

                    end if;
                    line := line + 1;
                end loop;

                -- if the matrix exists after the loop, it wasn't completed. if
                -- it's because the file ended, that's an error.
                if matrix /= null then
                    if not this.error then
                        line := line + 1;
                        this.Put_Error( path & ":" & Image( line ) &
                                        ": Unexpected end of file" );
                    end if;
                    Delete( matrix );
                end if;

            end if;
            line := line + 1;
        end loop;

        -- display the progress
        if this.matrixCount > 0 and then not this.error then
            this.Put_Line( "Parsed " & Image( this.matrixCount ) & " matrices" );
        end if;

        Close( file );
    exception
        when Name_Error =>
            -- File not found; not an error
            null;
    end Read_Matrix_List;

    ----------------------------------------------------------------------------

    procedure Read_Tile_List( this : not null access Compiler'Class;
                              path : String ) is
        line : Positive := 1;

        ------------------------------------------------------------------------

        function Scan_Tile( text : Unbounded_String ) return A_Tile is
            tokens : String_Array(1..MAX_COLUMNS_PER_LINE);
            last   : Integer;
            tile   : A_Tile := null;
        begin
            Split_On_Spaces( text, tokens, last );

            if last > tokens'Last then
                this.Put_Error( path & ":" & Image( line ) & ": Too many attributes" );
                return null;
            end if;

            tile := Create_Tile;

            -- expect id number
            begin
                tile.Set_Id( Natural'Value( To_String( tokens(tokens'First) ) ) );
            exception
                when Constraint_Error =>
                    this.Put_Error( path & ":" & Image( line ) & ": Expected tile id" );
                    Delete( tile );
                    return null;
            end;

            if tile.Get_Id > 0 then
                -- expect image filename
                if last < tokens'First + 1 then
                    this.Put_Error( path & ":" & Image( line ) & ": Expected filename" );
                    Delete( tile );
                    return null;
                end if;

                -- ensure the filename is real
                if not GNAT.OS_Lib.Is_Regular_File( Get_Directory( path ) &
                                                    To_String( tokens(tokens'First + 1) ) )
                then
                    this.Put_Error( path & ":" & Image( line ) & ": File not found: " &
                                    Get_Directory( path ) &
                                    To_String( tokens(tokens'First + 1) ) );
                    Delete( tile );
                    return null;
                end if;

                -- strip path information from the tile's name
                tile.Set_Name( Get_Filename( To_String( tokens(tokens'First + 1) ) ) );

                -- mark the tile's image to be added to the library
                if not this.Add_File( "images/" & Get_Filename( To_String( tokens(tokens'First + 1) ) ),
                                      Get_Directory( path ) & To_String( tokens(tokens'First + 1) ) )
                then
                    this.Put_Error( path & ":" & Image( line ) & ": Tile filename collision" );
                    Delete( tile );
                    return null;
                end if;

                 -- accept attributes
                for i in tokens'First+2..last loop
                    declare
                        attr  : constant String  := To_String( tokens(i) );
                        eqPos : constant Integer := Index( attr, "=" );
                        name  : constant String  := attr(attr'First..eqPos-1);
                        value : constant String  := attr(eqPos+1..attr'Last);
                    begin
                        -- expect attribute name
                        if eqPos < attr'First then
                            -- found an attribute name (implicit value is True)
                            tile.Set_Attribute( attr, Booleans.Create( True ) );
                        elsif eqPos > attr'First then
                            -- found a valued attribute
                            tile.Set_Attribute( name, Parse_Value( value ) );
                        else
                            -- found an equals sign instead of an attribute name
                            raise ATTRIBUTE_ERROR with "Expected attribute name";
                        end if;
                    exception
                        when e : ATTRIBUTE_ERROR =>
                            this.Put_Error( path & ":" & Image( line ) & ": " &
                                            Exception_Message( e ) );
                            Delete( tile );
                            return null;
                    end;
                end loop;
            else
                -- nothing is expected past the id 0
                if last > tokens'First then
                    -- found unexpected text beyond the id 0
                    this.Put_Error( path & ":" & Image( line ) & ": Unexpected token" );
                    Delete( tile );
                    return null;
                end if;
            end if;

            return tile;
        exception
            when e : others =>
                this.Put_Error( path & ":" & Image( line ) & ": Syntax error: " &
                                Exception_Name( e ) & " - " & Exception_Message( e ) );
                Delete( tile );
                return null;
        end Scan_Tile;

        ------------------------------------------------------------------------

        use Ada.Text_IO;

        file : Ada.Text_IO.File_Type;
        text : Unbounded_String;
        tile : A_Tile;
    begin
        -- open the tile list file
        Open( file, In_File, path );
        if not Is_Open( file ) then
            this.Put_Error( path & ": File not found" );
            return;
        end if;

        -- read line by line
        while not End_Of_File( file ) loop
            text := Trim_Whitespace_And_Comments( Get_Line( file ) );
            if Length( text ) > 0 then
                -- parse the line into a tile with attributes
                tile := Scan_Tile( text );
                begin
                    if tile /= null then
                        this.lib.Add_Tile( tile );
                        this.tileCount := this.tileCount + 1;
                    end if;
                exception
                    when e : DUPLICATE_TILE =>
                        Delete( tile );
                        this.Put_Error( path & ":" & Image( line ) & ": " &
                                        Exception_Message( e ) );
                    when e : others =>
                        Delete( tile );
                        this.Put_Error( path & ":" & Image( line ) & ": " &
                                        "Unknown error: " & Exception_Name( e ) &
                                        " - " & Exception_Message( e ) );
                end;
            end if;
            line := line + 1;
        end loop;

        -- display the progress
        if not this.error then
            this.Put_Line( "Parsed " & Image( this.tileCount ) & " tiles" );
        end if;

        Close( file );
    end Read_Tile_List;

    ----------------------------------------------------------------------------

    procedure Write_Catalog_File( this : not null access Compiler'Class;
                                  path : String ) is
        use Ada.Streams.Stream_IO;
        file : File_Type;
    begin
        Ada.Streams.Stream_IO.Create( file, Out_File, path );
        if not Is_Open( file ) then
            this.Put_Error( "Unable to create file " & path );
            return;
        end if;

        A_Catalog'Output( Stream( file ), this.lib.catalog );
        Close( file );

        if this.Add_File( Get_Filename( path ), path ) then
            -- should always succeed
             null;
        end if;
    exception
        when e : others =>
            if Is_Open( file ) then
                Close( file );
            end if;
            this.Put_Error( "Error writing " & path & ": " & Exception_Name( e ) &
                            " - " & Exception_Message( e ) );
    end Write_Catalog_File;

    ----------------------------------------------------------------------------

    procedure Write_Info_File( this : not null access Compiler'Class;
                               path : String ) is
        use Ada.Text_IO;
        file : File_Type;
    begin
        Ada.Text_IO.Create( file, Out_File, path );
        if not Is_Open( file ) then
            this.Put_Error( "Unable to create file: " & path );
            return;
        end if;

        Put_Line( file, "Version    : " & Image( Catalog_Format_Version ) );
        Put_Line( file, "Created By : " & Get_Application.Get_Name & " " & Version.Release );
        Put_Line( file, "Date       : " & Ada.Calendar.Formatting.Image( Clock ) & " UTC" );
        Put_Line( file, "Images     : " & Image( this.imageCount ) );
        Put_Line( file, "Tiles      : " & Image( this.tileCount ) );
        Put_Line( file, "Matrices   : " & Image( this.matrixCount ) );
        Close( file );

        if this.Add_File( Get_Filename( path ), path ) then
            -- should always succeed
            null;
        end if;
    end Write_Info_File;

    ----------------------------------------------------------------------------

    procedure Write_Library_File( this     : not null access Compiler'Class;
                                  path     : String;
                                  compress : Boolean ) is
        use String_Maps;

        archive            : Zip_Create_info;
        fileStream         : aliased File_Zipstream;
        filesRemaining     : Natural;
        pos                : String_Maps.Cursor;
        lastProgressReport : Ada.Real_Time.Time;
        success            : Boolean;
    begin
        -- Delete the old file first, if it exists
        GNAT.OS_Lib.Delete_File( path, success );

        -- Create a new zip file
        if compress then
            Zip.Create.Create( archive, fileStream'Unrestricted_Access, path, Zip.Compress.Shrink );
        else
            Zip.Create.Create( archive, fileStream'Unrestricted_Access, path, Zip.Compress.Store );
        end if;

        -- Add all of the files to the archive
        filesRemaining := Natural(this.files.Length);
        this.Put_Line( "Adding " & Image( filesRemaining ) & " files..." );
        lastProgressReport := Clock;
        pos := this.files.First;
        while Has_Element( pos ) loop
            begin
                Add_File( archive, Element( pos ), Key( pos ) );
            exception
                when e : others =>
                    this.Put_Error( "Error adding " & Element( pos ) & ": " &
                                    Exception_Name( e ) );
            end;

            Next( pos );
            filesRemaining := filesRemaining - 1;

            -- report the progress every 5 seconds and at the end
            if Clock - Seconds( 5 ) > lastProgressReport then
                if filesRemaining > 0 then
                    this.Put_Line( Image( 100 - (filesRemaining * 100 / Integer(this.files.Length)) ) & "% complete" );
                end if;
                lastProgressReport := Clock;
            end if;
        end loop;
        this.Put_Line( "100% complete" );

        -- Finalize and close the zip file
        Finish( archive );

        -- Clean up on error
        if this.error then
            GNAT.OS_Lib.Delete_File( path, success );
        end if;
    end Write_Library_File;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Compiler ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Tiles.Libraries.Compilers;
