--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Containers.Indefinite_Hashed_Maps;
with Ada.Strings.Hash_Case_Insensitive;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;

package body Tiles.Libraries.Manager is

    initialized : Boolean := False;
    finalized   : Boolean := False;

    ----------------------------------------------------------------------------

    -- Set this to True to keep tile libraries in the cache after all references
    -- have been removed. This will speed up loading for libraries that are
    -- frequently loaded and unloaded, like the tile library for the world map.
    -- Libraries will still be unloaded at finalization and when unreferenced
    -- after finalization.
    --
    -- Set this to False to immediately unload libraries when they are not being
    -- used. This will reduce memory usage at the cost of additional load time.
    CACHE_UNLOAD_DELAYED : constant Boolean := True;

    type Cache_Element is
        record
            lib  : A_Tile_Library := null;
            refs : Integer := 1;
        end record;

    package Cache_Elements is new
        Ada.Containers.Indefinite_Hashed_Maps( String,
                                               Cache_Element,
                                               Ada.Strings.Hash_Case_Insensitive,
                                               "=" );

    -- The tile library cache maintains reference counts of all the currently
    -- loaded tile libraries and unloads them when they become unreferenced.
    protected Library_Cache is

        -- Finalizes the tile library cache on application shutdown. All
        -- unreferenced libraries in the cache will be unloaded at this point
        -- and all libraries that become unreferenced after this will be
        -- unloaded. No libraries should remain in the cache on application
        -- finalization if it shut down cleanly.
        procedure Finalize;

        -- Loads a library into memory or just returns a reference if it's
        -- already in the cache.
        procedure Load_Library( name : String; lib : out A_Tile_Library );

        -- Unreferences a library in the cache. If this is the last reference
        -- remaining to the library, it will be unloaded (unless the
        -- CACHE_UNLOAD_DELAYED is set.)
        procedure Unload_Library( lib : in out A_Tile_Library );

    private

        map       : Cache_Elements.Map;
        finalized : Boolean := False;
    end Library_Cache;

    -- END PRIVATE DECLARATAIONS --
    --==========================================================================

    protected body Library_Cache is

        procedure Finalize is
            crsr : Cache_Elements.Cursor := map.First;
            next : Cache_Elements.Cursor;
            elem : Cache_Element;
        begin
            while Cache_Elements.Has_Element( crsr ) loop
                elem := Cache_Elements.Element( crsr );
                next := Cache_Elements.Next( crsr );
                if elem.refs = 0 then
                    pragma Debug( Dbg( "Library <" & elem.lib.Get_Name & "> unloaded",
                                       D_TILES, INFO ) );
                    Delete( elem.lib );
                    map.Delete( crsr );
                else
                    -- one of the tile libraries still has references remaining.
                    -- it will be unloaded on the last unreference, which should
                    -- happen before application shutdown. Otherwise, the memory
                    -- will appear to be leaked.
                    null;
                end if;
                crsr := next;
            end loop;
        end Finalize;

        ------------------------------------------------------------------------

        procedure Load_Library( name : String; lib : out A_Tile_Library ) is
            crsr : Cache_Elements.Cursor;
            elem : Cache_Element;
        begin
            if not finalized then
                if map.Contains( name ) then
                    crsr := map.Find( name );
                    elem := Cache_Elements.Element( crsr );
                    elem.refs := elem.refs + 1;
                    lib := elem.lib;
                    map.Replace_Element( crsr, elem );
                else
                    if name'Length > 0 then
                        elem.lib := Tiles.Libraries.Load_Library( name, bitmaps => False );
                        elem.refs := 1;
                        lib := elem.lib;
                        if elem.lib /= null then
                            map.Insert( name, elem );
                        end if;
                    else
                        lib := null;
                    end if;
                end if;
            else
                lib := null;
                pragma Debug( Dbg( "Tiles.Libraries.Cache: Won't load <" &
                                   elem.lib.Get_Name & "> - cache finalized",
                                   D_TILES, WARNING ) );
            end if;
        end Load_Library;

        ------------------------------------------------------------------------

        procedure Unload_Library( lib : in out A_Tile_Library ) is
            crsr : Cache_Elements.Cursor;
            elem : Cache_Element;
        begin
            if lib /= null then
                crsr := map.Find( To_String( lib.name ) );
                if Cache_Elements.Has_Element( crsr ) then
                    elem := Cache_Elements.Element( crsr );
                    if elem.refs > 0 then
                        elem.refs := elem.refs - 1;
                    end if;
                    if elem.refs = 0 and then (not CACHE_UNLOAD_DELAYED or else finalized) then
                        -- unload the tile library if the cache was finalized or
                        -- if we're not configured to delay cache unloads
                    pragma Debug( Dbg( "Library <" & elem.lib.Get_Name & "> unloaded",
                                       D_TILES, INFO ) );
                        Delete( elem.lib );
                        map.Delete( crsr );
                    else
                        map.Replace_Element( crsr, elem );
                    end if;
                end if;
                lib := null;
            end if;
        end Unload_Library;

    end Library_Cache;

    --==========================================================================

    procedure Initialize is
    begin
        if not initialized then
            pragma Debug( Dbg( "Initializing tile manager", D_INIT, INFO ) );
            initialized := True;
        end if;
    end Initialize;

    ----------------------------------------------------------------------------

    procedure Finalize is
    begin
        if initialized and then not finalized then
            pragma Debug( Dbg( "Finalizing tile manager", D_INIT, INFO ) );
            Library_Cache.Finalize;
            finalized := True;
        end if;
    end Finalize;

    ----------------------------------------------------------------------------

    function Load_Library( name : String; bitmaps : Boolean ) return A_Tile_Library is
        lib : A_Tile_Library;
    begin
        Library_Cache.Load_Library( name, lib );
        if lib /= null then
            if bitmaps then
                lib.Load_Bitmaps;
            end if;
        end if;
        return lib;
    end Load_Library;

    ----------------------------------------------------------------------------

    procedure Unload_Library( lib : in out A_Tile_Library ) is
    begin
        Library_Cache.Unload_Library( lib );
    end Unload_Library;

end Tiles.Libraries.Manager;
