--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Applications;                      use Applications;
with Applications.Gui;                  use Applications.Gui;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Resources;                         use Resources;
with Streams.Buffers;                   use Streams.Buffers;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;
with Support.Real_Time;                 use Support.Real_Time;

package body Tiles.Libraries is

    ATLAS_MIN_SIZE  : constant := 256;
    ATLAS_INCREMENT : constant := 128;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Library( name : String ) return A_Tile_Library is
        this : constant A_Tile_Library := new Tile_Library;
    begin
        this.Construct;
        this.name := To_Unbounded_String( name );
        this.catalog := Create_Catalog;
        return this;
    end Create_Library;

    ----------------------------------------------------------------------------

    function Load_Library( name : String; bitmaps : Boolean ) return A_Tile_Library is
        libFilename : constant String := Add_Extension( name, Library_Extension, force => True );
        this        : A_Tile_Library := null;
    begin
        if name'Length = 0 then
            return null;
        end if;

        -- create the tile library instance
        this := new Tile_Library;
        this.Construct;
        this.name := To_Unbounded_String( name );

        -- load the tile library's .gfx file
        declare
            resource : A_Resource_File := null;
        begin
            pragma Debug( Dbg( "Loading library <" & name & ">", D_TILES, INFO ) );
            resource := Load_Resource( libFilename, "graphics" );
            if resource /= null then
                this.file := Load_Archive( resource );     -- may raise exception
                Release( resource );
                -- successfully loaded the library archive file
            else
                -- unable to find archive
                Dbg( "Error: Library <" & name & ">: File not found: " &
                     libFilename, D_TILES, ERROR );
                Release( resource );
                Delete( A_Limited_Object(this) );
            end if;
        exception
            when e : others =>
                -- Load_Archive failed: unable to open archive
                Dbg( "Error: Unable to load library <" & name & ">: Exception...",
                     D_TILES, ERROR );
                Dbg( e, D_TILES, ERROR );
                Release( resource );
                Delete( A_Limited_Object(this) );
        end;

        -- load the catalog file
        if this /= null then
            declare
                resource : A_Resource_File := null;
                stream   : A_Buffer_Stream := null;
            begin
                resource := this.file.Load_Resource( CATALOG_FILENAME );
                resource.Get_Stream( stream );
                this.catalog := A_Catalog'Input( stream );
                if this.catalog = null then
                    raise Constraint_Error with "Bad file format";
                end if;
                Close( stream );
                Release( resource );
            exception
                when e : others =>
                    -- unable to load catalog file (format error or not found)
                    Dbg( "Error: Unable to load catalog file <" &
                         CATALOG_FILENAME & "> in library <" & name & ">",
                         D_TILES, ERROR );
                    Dbg( e, D_TILES, ERROR );
                    Close( stream );
                    Release( resource );
                    Delete( A_Limited_Object(this) );
            end;
        end if;

        if bitmaps and then this /= null then
            this.Load_Bitmaps;
        end if;

        -- returns null if load failed
        return this;
    end Load_Library;

    ----------------------------------------------------------------------------

    procedure Add_Matrix( this   : not null access Tile_Library'Class;
                          matrix : in out A_Tile_Matrix ) is
    begin
        if matrix /= null then
            this.catalog.Add_Matrix( matrix );
        end if;
    end Add_Matrix;

    ----------------------------------------------------------------------------

    procedure Add_Tile( this : not null access Tile_Library'Class;
                        tile : in out A_Tile ) is
    begin
        if tile /= null then
            this.catalog.Add_Tile( tile );
        end if;
    end Add_Tile;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Tile_Library ) is
    begin
        Delete( this.catalog );
        Delete( this.atlas );
        Delete( this.file );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Get_Id( this : not null access Tile_Library'Class;
                     name : String ) return Natural is
        tile : constant A_Tile := this.catalog.Find_Tile( name );
    begin
        if tile /= null then
            return tile.Get_Id;
        end if;
        return 0;
    end Get_Id;

    ----------------------------------------------------------------------------

    function Get_Matrix( this  : not null access Tile_Library'Class;
                         index : Natural ) return A_Tile_Matrix is
    begin
        return this.catalog.Get_Matrix( index );
    end Get_Matrix;

    ----------------------------------------------------------------------------

    function Get_Matrix_Count( this : not null access Tile_Library'Class ) return Natural is
    begin
        return this.catalog.Get_Matrix_Count;
    end Get_Matrix_Count;

    ----------------------------------------------------------------------------

    function Get_Name( this : not null access Tile_Library'Class ) return String is
    begin
        return To_String( this.name );
    end Get_Name;

    ----------------------------------------------------------------------------

    function Get_Tile( this : not null access Tile_Library'Class;
                       id   : Natural ) return A_Tile is
    begin
        if id > 0 then
            return this.catalog.Get_Tile( id );
        end if;
        return null;
    end Get_Tile;

    ----------------------------------------------------------------------------

    function Is_Loaded( this : not null access Tile_Library'Class ) return Boolean is
    begin
        this.loadLock.Lock;
        return result : constant Boolean := this.loaded do
            this.loadLock.Unlock;
        end return;
    end Is_Loaded;

    ----------------------------------------------------------------------------

    procedure Iterate_By_Id( this    : not null access Tile_Library'Class;
                             examine : not null access procedure( tile : not null A_Tile ) ) is
    begin
        this.catalog.Iterate_By_Id( examine );
    end Iterate_By_Id;

    ----------------------------------------------------------------------------

    procedure Iterate_By_Slot( this    : not null access Tile_Library'Class;
                               examine : not null access procedure( slot : Positive;
                                                                    tile : A_Tile ) ) is
    begin
        this.catalog.Iterate_By_Slot( examine );
    end Iterate_By_Slot;

    ----------------------------------------------------------------------------

    procedure Load_Bitmaps( this : not null access Tile_Library'Class ) is
        timer : Timer_Type;
    begin
        this.loadLock.Lock;
        if this.loaded then
            this.loadLock.Unlock;
            return;
        end if;

        declare
            maxTileWidth  : Natural := 0;
            maxTileHeight : Natural := 0;
            pixels        : Natural := 0;

            --------------------------------------------------------------------

            procedure Load( tile : not null A_Tile ) is
            begin
                tile.Load_Bitmap( this.file );
                if tile.bmp /= null then
                    if Al_Get_Bitmap_Width( tile.bmp ) > maxTileWidth then
                        maxTileWidth := Al_Get_Bitmap_Width( tile.bmp );
                    end if;
                    if Al_Get_Bitmap_Height( tile.bmp ) > maxTileHeight then
                        maxTileHeight := Al_Get_Bitmap_Height( tile.bmp );
                    end if;
                    pixels := pixels + Al_Get_Bitmap_Width( tile.bmp ) *
                                       Al_Get_Bitmap_Height( tile.bmp );
                end if;
            end Load;

            --------------------------------------------------------------------

            procedure Add_To_Atlas( tile : not null A_Tile ) is
            begin
                this.atlas.Add_Bitmap( tile.bmp );
            end Add_To_Atlas;

            --------------------------------------------------------------------

            atlasPageWidth  : Positive := 1;
            atlasPageHeight : Positive := 1;
        begin
            Start( timer );
            this.catalog.Iterate_By_Id( Load'Unrestricted_Access );

            if Get_Application.all in Gui_Application'Class then
                -- smallest atlas page size
                atlasPageWidth := ATLAS_MIN_SIZE;
                atlasPageHeight := ATLAS_MIN_SIZE;

                -- enlarge it to fit the largest tile
                while atlasPageWidth < maxTileWidth loop
                    atlasPageWidth := atlasPageWidth + ATLAS_INCREMENT;
                end loop;
                while atlasPageHeight < maxTileHeight loop
                    atlasPageHeight := atlasPageHeight + ATLAS_INCREMENT;
                end loop;

                -- enlarge the page to fit all the pixels
                while (pixels * 6) / 5 > atlasPageWidth * atlasPageHeight loop
                    -- enlarge the shorter dimension until it's square, then
                    -- continue to enlarge it as a square.
                    if atlasPageHeight >= atlasPageWidth then
                        atlasPageWidth := atlasPageWidth + ATLAS_INCREMENT;
                    end if;
                    if atlasPageWidth + ATLAS_INCREMENT >= atlasPageHeight then
                        atlasPageHeight := atlasPageHeight + ATLAS_INCREMENT;
                    end if;
                end loop;

                this.atlas := Create_Atlas( A_Gui_Application(Get_Application).Get_Display,
                                            atlasPageWidth, atlasPageHeight );
                this.catalog.Iterate_By_Id( Add_To_Atlas'Unrestricted_Access );

                -- uncomment the following line to write the atlas's pages to
                -- disk for debugging purposes.
                --this.atlas.Save_Bitmaps( To_String( this.name ) );
            end if;

            pragma Debug( Dbg( "Loaded library <" & this.Get_Name & "> bitmaps - " &
                               Image( To_Milliseconds( Elapsed( timer ) ) ) & "[ms]",
                               D_TILES, INFO ) );
        exception
            when e : others =>
                Dbg( "Exception: Failed to load bitmaps in library " &
                     "<" & this.Get_Name & ">", D_TILES, ERROR );
                Dbg( e, D_TILES, ERROR );
        end;

        this.loaded := True;
        this.loadLock.Unlock;
    end Load_Bitmaps;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Tile_Library ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    function Library_Extension return String is
    begin
        return "gfx";
    end Library_Extension;

end Tiles.Libraries;
