--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Values.Nulls;                      use Values.Nulls;
with Resources;                         use Resources;
with Resources.Images;                  use Resources.Images;

package body Tiles is

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Tile return A_Tile is
        this : constant A_Tile := new Tile_Object;
    begin
        this.Construct;
        return this;
    end Create_Tile;

    ----------------------------------------------------------------------------

    overriding
    procedure Adjust( this : access Tile_Object ) is
        pragma Unreferenced( this );
    begin
        raise COPY_NOT_ALLOWED;
    end Adjust;
    pragma No_Return( Adjust );

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Tile_Object ) is
    begin
        Object(this.all).Construct;
        this.lock := new Locking_Object;
        this.loadState := new Completion;
        this.attributes := Create_Assoc;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Tile_Object ) is
    begin
        this.lock.Lock;
        Al_Destroy_Bitmap( this.bmp );
        Delete( this.loadState );
        Delete( this.lock );
        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Get_Attribute( this : access Tile_Object'Class;
                            name : String ) return Value_Ptr is
    begin
        if this /= null then
            return this.attributes.Get.Get( name );
        end if;
        return Create_Null;
    end Get_Attribute;

    ----------------------------------------------------------------------------

    function Get_Bitmap( this : access Tile_Object'Class ) return A_Allegro_Bitmap is
        bmp : A_Allegro_Bitmap := null;
    begin
        if this /= null then
            this.lock.Lock;
            if this.loadState.Is_Complete then
                bmp := this.bmp;
            end if;
            this.lock.Unlock;
        end if;
        return bmp;
    end Get_Bitmap;

    ----------------------------------------------------------------------------

    function Get_Id( this : not null access Tile_Object'Class ) return Natural is
    begin
        return this.id;
    end Get_Id;

    ----------------------------------------------------------------------------

    function Get_Name( this : not null access Tile_Object'Class ) return String is
    begin
        return To_String( this.name );
    end Get_Name;

    ----------------------------------------------------------------------------

    function Is_Loaded( this : not null access Tile_Object'Class ) return Boolean is
    begin
        return this.loadState.Is_Complete;
    end Is_Loaded;

    ----------------------------------------------------------------------------

    procedure Load_Bitmap( this    : not null access Tile_Object'Class;
                           archive : not null A_Archive ) is
    begin
        this.lock.Lock;
        if not this.loadState.Is_Complete then
            declare
                resource : A_Resource_File;
            begin
                resource := archive.Load_Resource( "images/" & To_String( this.name ) );
                this.bmp := Load_Image( resource );
                Release( resource );
            exception
                when e : others =>
                    Release( resource );
                    Dbg( "Tiles.Load_Bitmap Exception: ...", D_TILES, ERROR );
                    Dbg( e, D_TILES, ERROR );
                    Dbg( "Unable to find <" & To_String( this.name ) &
                         "> in tile library", D_TILES, ERROR );
            end;
            this.loadState.Notify_Complete;
        end if;
        this.lock.Unlock;
    end Load_Bitmap;

    ----------------------------------------------------------------------------

    not overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Tile_Object is
        this : aliased Tile_Object;
    begin
        this.Construct;
        Tile_Object'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    not overriding
    procedure Object_Output( stream : access Root_Stream_Type'Class; obj : Tile_Object ) is
    begin
        Tile_Object'Write( stream, obj );
    end Object_Output;

    ----------------------------------------------------------------------------

    procedure Set_Attribute( this : not null access Tile_Object'Class;
                             name : String;
                             val  : Value_Ptr'Class ) is
    begin
        -- offsetX=<pixels>
        -- offsetY=<pixels>
        -- clip=<Physics.Clip_Type enum values>
        -- delay=<milliseconds>
        -- frames=[<tile_id>[,<tile_id>]*]
        -- next=<tile_id>
        this.attributes.Get.Set( name, val );
    end Set_Attribute;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Id( this : access Tile_Object; id : Natural ) is
    begin
        this.id := id;
    end Set_Id;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Name( this : access Tile_Object; name : String ) is
    begin
        this.name := To_Unbounded_String( name );
    end Set_Name;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Tile ) is
    begin
        Delete( A_Object(this) );
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Tile_Object ) is
    begin
        Object'Read( stream, Object(obj) );
        obj.id := Integer'Input( stream );
        obj.name := To_Unbounded_String( String'Input( stream ) );
        obj.attributes := As_Assoc( Value_Ptr'Input( stream ) );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Tile_Object ) is
    begin
        Object'Write( stream, Object(obj) );
        Integer'Output( stream, obj.id );
        String'Output( stream, To_String( obj.name ) );
        Value_Ptr'Output( stream, obj.attributes.As_Value );
    end Object_Write;

    ----------------------------------------------------------------------------

    function A_Tile_Input( stream : access Root_Stream_Type'Class ) return A_Tile is
        tile : A_Tile := null;
    begin
        if Boolean'Input( stream ) then
            A_Tile'Read( stream, tile );
        end if;
        return tile;
    end A_Tile_Input;

    ----------------------------------------------------------------------------

    procedure A_Tile_Output( stream : access Root_Stream_Type'Class; tile : A_Tile ) is
    begin
        Boolean'Output( stream, tile /= null );
        if tile /= null then
            A_Tile'Write( stream, tile );
        end if;
    end A_Tile_Output;

    ----------------------------------------------------------------------------

    procedure A_Tile_Read( stream : access Root_Stream_Type'Class; tile : out A_Tile ) is
    begin
        tile := new Tile_Object'Class'(Tile_Object'Class'Input( stream ));
    end A_Tile_Read;

    ----------------------------------------------------------------------------

    procedure A_Tile_Write( stream : access Root_Stream_Type'Class; tile : A_Tile ) is
    begin
        Tile_Object'Class'Output( stream, tile.all );
    end A_Tile_Write;

    --==========================================================================

    function A_Tile_Id_Array_Input( stream : access Root_Stream_Type'Class ) return A_Tile_Id_Array is
        tia : A_Tile_Id_Array := null;
        len : Natural;
    begin
        if Boolean'Input( stream ) then
            len := Natural'Input( stream );
            tia := new Tile_Id_Array(1..len);
            for i in tia'Range loop
                tia(i) := Natural'Input( stream );
            end loop;
        end if;
        return tia;
    exception
        when others =>
            Delete( tia );
            raise;
    end A_Tile_Id_Array_Input;

    ----------------------------------------------------------------------------

    procedure A_Tile_Id_Array_Output( stream : access Root_Stream_Type'Class; tia : A_Tile_Id_Array ) is
    begin
        Boolean'Output( stream, tia /= null );
        if tia /= null then
            Natural'Output( stream, tia'Length );
            for i in tia'Range loop
                Natural'Output( stream, tia(i) );
            end loop;
        end if;
    end A_Tile_Id_Array_Output;

    ----------------------------------------------------------------------------

    function Copy( src : A_Tile_Id_Array ) return A_Tile_Id_Array is
    begin
        if src /= null then
            return new Tile_Id_Array'(src.all);
        end if;
        return null;
    end Copy;

end Tiles;
