--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Unchecked_Deallocation;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Ada.Streams;                       use Ada.Streams;
with Objects;                           use Objects;
with Values;                            use Values;
with Values.Associations;               use Values.Associations;

private with Ada.Containers.Doubly_Linked_Lists;
private with Ada.Strings.Unbounded;
private with Completions;
private with Locking_Objects;
private with Resources.Archives;

package Tiles is

    -- An array of tile ids
    type Tile_Id_Array is array (Integer range <>) of Natural;
    type A_Tile_Id_Array is access all Tile_Id_Array;

    function A_Tile_Id_Array_Input( stream : access Root_Stream_Type'Class ) return A_Tile_Id_Array;
    for A_Tile_Id_Array'Input use A_Tile_Id_Array_Input;

    procedure A_Tile_Id_Array_Output( stream : access Root_Stream_Type'Class; tia : A_Tile_Id_Array );
    for A_Tile_Id_Array'Output use A_Tile_Id_Array_Output;

    -- Returns a copy of 'src'.
    function Copy( src : A_Tile_Id_Array ) return A_Tile_Id_Array;
    pragma Postcondition( Copy'Result /= src or else src = null );

    -- Deletes the Tile_Id_Array.
    procedure Delete is new Ada.Unchecked_Deallocation( Tile_Id_Array, A_Tile_Id_Array );

    ----------------------------------------------------------------------------

    -- A Tile_Object represents a tile, the most basic element of a world map.
    -- It is uniquely identified within its library by an integer Id. Each tile
    -- is represented by a bitmap, can be animated using bitmaps from other
    -- tiles, and contains a set of attributes the define how entities interact
    -- with it in the world.
    type Tile_Object is new Object with private;
    type A_Tile is access all Tile_Object'Class;

    -- Creates a new tile using register allocator. Returns null if no allocator
    -- has been registered.
    function Create_Tile return A_Tile;

    -- Returns the value of the 'name' attribute of the tile, or Null if the
    -- attribute has not been defined.
    function Get_Attribute( this : access Tile_Object'Class;
                            name : String ) return Value_Ptr;

    -- Returns null if the tile's bitmap file isn't found or hasn't loaded yet.
    -- If 'this' is null, null will be returned.
    function Get_Bitmap( this : access Tile_Object'Class ) return A_Allegro_Bitmap;

    -- Returns the id of the tile within its library.
    function Get_Id( this : not null access Tile_Object'Class ) return Natural;

    -- Returns the name (bitmap filename) of the tile. Tile names are not unique
    -- within a library.
    function Get_Name( this : not null access Tile_Object'Class ) return String;

    -- Returns True if the tile's bitmap has been loaded, successfully or otherwise.
    function Is_Loaded( this : not null access Tile_Object'Class ) return Boolean;

    -- Reads a Tile_Object from a stream.
    function Object_Input( stream : access Root_Stream_Type'Class ) return Tile_Object;
    for Tile_Object'Input use Object_Input;

    -- Writes a Tile_Object to a stream.
    procedure Object_Output( stream : access Root_Stream_Type'Class; obj : Tile_Object );
    for Tile_Object'Output use Object_Output;

    -- Sets an attribute on the tile by name. The value may be of any type.
    procedure Set_Attribute( this : not null access Tile_Object'Class;
                             name : String;
                             val  : Value_Ptr'Class );
    pragma Precondition( name'Length > 0 );
    pragma Precondition( Value_Ptr(val) /= Values.Nul );

    -- Sets the id of the tile. Tile ids are unique within a library. Adding
    -- this tile to a library that already contains a tile with the same id will
    -- result in an error.
    procedure Set_Id( this : access Tile_Object; id : Natural );

    -- Sets the name (bitmap filename) of the tile. Tile names are not unique
    -- within a library.
    procedure Set_Name( this : access Tile_Object; name : String );
    pragma Precondition( name'Length > 0 );

    -- Deletes the Tile.
    procedure Delete( this : in out A_Tile );
    pragma Postcondition( this = null );

    ATTRIBUTE_ERROR : exception;

private

    use Ada.Strings.Unbounded;
    use Completions;
    use Locking_Objects;
    use Resources.Archives;

    type Tile_Object is new Object with
        record
            -- ** these fields are streamed **
            id         : Natural := 0;
            name       : Unbounded_String;
            attributes : Assoc_Ptr;
            -- ** end streamed fields **

            lock       : A_Locking_Object := null;    -- protects bmp, priority
            bmp        : A_Allegro_Bitmap := null;
            priority   : Integer := 0;
            loadState  : A_Completion := null;
        end record;

    -- Raises COPY_NOT_ALLOWED.
    procedure Adjust( this : access Tile_Object );

    procedure Construct( this : access Tile_Object );

    procedure Delete( this : in out Tile_Object );

    -- Loads the tile's bitmap from an archive.
    procedure Load_Bitmap( this    : not null access Tile_Object'Class;
                           archive : not null A_Archive );

    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Tile_Object );
    for Tile_Object'Read use Object_Read;

    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Tile_Object );
    for Tile_Object'Write use Object_Write;

    ----------------------------------------------------------------------------

    function A_Tile_Input( stream : access Root_Stream_Type'Class ) return A_Tile;
    for A_Tile'Input use A_Tile_Input;

    procedure A_Tile_Output( stream : access Root_Stream_Type'Class; tile : A_Tile );
    for A_Tile'Output use A_Tile_Output;

    procedure A_Tile_Read( stream : access Root_Stream_Type'Class; tile : out A_Tile );
    for A_Tile'Read use A_Tile_Read;

    procedure A_Tile_Write( stream : access Root_Stream_Type'Class; tile : A_Tile );
    for A_Tile'Write use A_Tile_Write;

    -- Defines a list of Tile objects.
    package Tile_Lists is new Ada.Containers.Doubly_Linked_Lists( A_Tile, "=" );
    use Tile_Lists;

end Tiles;
