--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Support;                           use Support;

package body Tokens is

    function To_String( loc : Token_Location ) return String is
    begin
        return Image( loc.line ) & ':' & Image( loc.col );
    end To_String;

    ----------------------------------------------------------------------------

    function Create_Token( text : String; loc : Token_Location ) return A_Token is
        token : A_Token := null;
    begin
        for i in Token_Type'Range loop
            case i is
                when TK_IDENTIFIER =>
                    token := Create_Identifier_Token( text, loc );
                when TK_NUMBER =>
                    token := Create_Number_Token( text, loc );
                when TK_STRING =>
                    token := Create_String_Token( text, loc );
                when others =>
                    if text = tokenImage(i).all then
                       token := Create_Token( i, loc );
                   end if;
            end case;

            exit when token /= null;
        end loop;
        return token;
    end Create_Token;

    --==========================================================================

    function Create_Token( tokenType : Token_Type;
                           loc       : Token_Location ) return A_Token is
        this : constant A_Token := new Token;
    begin
        this.Construct( tokenType, loc );
        return this;
    end Create_Token;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this      : access Token;
                         tokenType : Token_Type;
                         loc       : Token_Location ) is
    begin
        Limited_Object(this.all).Construct;
        this.tokenType := tokenType;
        this.loc := loc;
    end Construct;

    ----------------------------------------------------------------------------

    function Get_Location( this : not null access Token'Class ) return Token_Location is
    begin
        return this.loc;
    end Get_Location;

    ----------------------------------------------------------------------------

    function Get_Type( this : not null access Token'Class ) return Token_Type is
    begin
        return this.tokenType;
    end Get_Type;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Token ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    function Create_Identifier_Token( text : String;
                                      loc  : Token_Location ) return A_Token is
        this : A_Identifier_Token := null;
    begin
        if Match( identPattern, text ) then
            this := new Identifier_Token;
            this.Construct( text, loc );
        end if;
        return A_Token(this);
    end Create_Identifier_Token;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Identifier_Token;
                         name : String;
                         loc  : Token_Location ) is
    begin
        Token(this.all).Construct( TK_IDENTIFIER, loc );
        this.name := To_Unbounded_String( name );
    end Construct;

    ----------------------------------------------------------------------------

    function Get_Name( this : not null access Identifier_Token'Class ) return String is
    begin
        return To_String( this.name );
    end Get_Name;

    --==========================================================================

    function Create_Number_Token( text : String;
                                  loc  : Token_Location ) return A_Token is
        this : A_Number_Token := null;
        val  : Long_Float;
    begin
        if Match( numPattern, text ) then
            val := Long_Float'Value( text );
            if val'Valid then
                this := new Number_Token;
                this.Construct( val, loc );
            end if;
        end if;
        return A_Token(this);
    exception
        when others =>
            -- raised by Long_Float'Value
            return null;
    end Create_Number_Token;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Number_Token;
                         val  : Long_Float;
                         loc  : Token_Location ) is
    begin
        Token(this.all).Construct( TK_NUMBER, loc );
        this.val := val;
    end Construct;

    ----------------------------------------------------------------------------

    function Get_Value( this : not null access Number_Token'Class ) return Long_Float is
    begin
        return this.val;
    end Get_Value;

    --==========================================================================

    function Create_String_Token( text : String;
                                  loc  : Token_Location ) return A_Token is
        this : A_String_Token := null;
    begin
        if text'Length > 1 and then
           text(text'First) = '"' and then
           text(text'Last) = text(text'First)
        then
            this := new String_Token;
            this.Construct( text(text'First+1..text'Last-1), loc );
        end if;
        return A_Token(this);
    end Create_String_Token;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access String_Token;
                         val  : String;
                         loc  : Token_Location ) is
    begin
        Token(this.all).Construct( TK_STRING, loc );
        this.val := To_Unbounded_String( val );
    end Construct;

    ----------------------------------------------------------------------------

    function Get_Value( this : not null access String_Token'Class ) return String is
    begin
        return To_String( this.val );
    end Get_Value;

end Tokens;
