--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with GNAT.Regpat;                       use GNAT.Regpat;
with Objects;                           use Objects;

package Tokens is

    -- Each Token_Type value represents a token in a grammar.
    type Token_Type is (
        TK_AND,
        TK_BANG,
        TK_COMMA,
        TK_EQUALS,
        TK_GREATER,
        TK_GREATER_EQUALS,
        TK_LEFT_PARENTHESIS,
        TK_LESS,
        TK_LESS_EQUALS,
        TK_MINUS,
        TK_NOT_EQUALS,
        TK_OR,
        TK_PERCENT,
        TK_PLUS,
        TK_RIGHT_PARENTHESIS,
        TK_SEMICOLON,
        TK_SLASH,
        TK_STAR,

        TK_IDENTIFIER,
        TK_NUMBER,
        TK_STRING,
        TK_EOF
    );

    subtype Static_Token_Type is Token_Type range TK_AND..TK_STAR;

    -- Encapsulates a token's position in an input stream. A token's location is
    -- determined by the location of it's first character.
    type Token_Location is
        record
            line : Natural := 0;
            col  : Natural := 0;
        end record;

    -- Returns a string representation in the format 'line:column'.
    function To_String( loc : Token_Location ) return String;

    ----------------------------------------------------------------------------

    -- A Token represents a recognized element of syntax. It is a terminal
    -- element in a script's context free grammar.
    type Token is new Limited_Object with private;
    type A_Token is access all Token'Class;

    -- Creates a new Token of the type matching 'text'. If 'text' is not
    -- recognized as a valid token, null will be returned.
    function Create_Token( text : String; loc : Token_Location ) return A_Token;

    -- Returns the location of the token in the input stream.
    function Get_Location( this : not null access Token'Class ) return Token_Location;

    -- Returns the token's type.
    function Get_Type( this : not null access Token'Class ) return Token_Type;

    -- Deletes the Token.
    procedure Delete( this : in out A_Token );
    pragma Postcondition( this = null );

    ----------------------------------------------------------------------------

    -- An Identifier_Token represents an identifier or variable name.
    type Identifier_Token is new Token with private;
    type A_Identifier_Token is access all Identifier_Token'Class;

    -- Returns the name of the identifier specified by the token.
    function Get_Name( this : not null access Identifier_Token'Class ) return String;

    ----------------------------------------------------------------------------

    -- A Number_Token represents a literal number, either decimal or integer.
    type Number_Token is new Token with private;
    type A_Number_Token is access all Number_Token'Class;

    -- Returns the numeric value of the token.
    function Get_Value( this : not null access Number_Token'Class ) return Long_Float;

    ----------------------------------------------------------------------------

    -- A String_Token represents a literal string.
    type String_Token is new Token with private;
    type A_String_Token is access all String_Token'Class;

    -- Returns the string value of the token.
    function Get_Value( this : not null access String_Token'Class ) return String;

private

    type TokenStringArray is array (Token_Type) of access String;

    -- An array of text images of all token types. Entries with a null value are
    -- pattern tokens that don't have a single static image.
    tokenImage : constant TokenStringArray := (
        TK_AND               => new String'("&&"),
        TK_BANG              => new String'("!"),
        TK_COMMA             => new String'(","),
        TK_EQUALS            => new String'("="),
        TK_GREATER           => new String'(">"),
        TK_GREATER_EQUALS    => new String'(">="),
        TK_LEFT_PARENTHESIS  => new String'("("),
        TK_LESS              => new String'("<"),
        TK_LESS_EQUALS       => new String'("<="),
        TK_MINUS             => new String'("-"),
        TK_NOT_EQUALS        => new String'("!="),
        TK_OR                => new String'("||"),
        TK_PERCENT           => new String'("%"),
        TK_PLUS              => new String'("+"),
        TK_RIGHT_PARENTHESIS => new String'(")"),
        TK_SEMICOLON         => new String'(";"),
        TK_SLASH             => new String'("/"),
        TK_STAR              => new String'("*"),

        TK_IDENTIFIER        => null,
        TK_NUMBER            => null,
        TK_STRING            => null,
        TK_EOF               => new String'("")
    );

    ----------------------------------------------------------------------------

    type Token is new Limited_Object with
        record
            tokenType : Token_Type;
            loc       : Token_Location;
        end record;

    procedure Construct( this      : access Token;
                         tokenType : Token_Type;
                         loc       : Token_Location );

    -- Creates a new Token of type 'tokenType'. The type must be one of the
    -- token types that has a static representation.
    function Create_Token( tokenType : Token_Type;
                           loc       : Token_Location ) return A_Token;

    ----------------------------------------------------------------------------

    -- Regular expression for recognizing an identifier token. An identifier
    -- must start with a letter and optionally be followed by any number of
    -- letters, numbers, underscores and dots.
    identPattern : constant Pattern_Matcher := Compile("^[a-zA-Z]+[a-zA-Z0-9_.]*$");

    type Identifier_Token is new Token with
        record
            name : Unbounded_String;
        end record;

    procedure Construct( this : access Identifier_Token;
                         name : String;
                         loc  : Token_Location );

    -- Creates a new Identifier_Token from 'text'. Returns null if 'text' is
    -- not an identifier string. An identifier is a letter followed by any
    -- number of letters, digits or underscores.
    function Create_Identifier_Token( text : String;
                                      loc  : Token_Location ) return A_Token;

    ----------------------------------------------------------------------------

    -- Regular expression for recognizing a number token. A number must start
    -- with any number of digits and may optionally by followed by a dot and
    -- another series of digits.
    numPattern : constant Pattern_Matcher := Compile("^([0-9]+)|([0-9]*\.[0-9]+)$");

    type Number_Token is new Token with
        record
            val : Long_Float;
        end record;

    procedure Construct( this : access Number_Token;
                         val  : Long_Float;
                         loc  : Token_Location );

    -- Creates a new Number_Token from 'text'. Returns null if 'text' is not a
    -- valid number. A valid number is a digit sequence, optionally followed
    -- by a dot and another digit sequence.
    function Create_Number_Token( text : String;
                                  loc  : Token_Location ) return A_Token;

    ----------------------------------------------------------------------------

    type String_Token is new Token with
        record
            val : Unbounded_String;
        end record;

    procedure Construct( this : access String_Token;
                         val  : String;
                         loc  : Token_Location );

    -- Creates a new String_Token from 'text'. Returns null if 'text' is not a
    -- valid string literal. A string is any sequence of characters within
    -- matching double-quote or single-quote characters.
    function Create_String_Token( text : String;
                                  loc  : Token_Location ) return A_Token;

end Tokens;
