--
-- Copyright (c) 2012-2013 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Values.Nulls;                      use Values.Nulls;
with Values.Strings;                    use Values.Strings;

package body Values.Associations is

    function Create_Assoc return Assoc_Ptr is
        this : constant A_Naked_Assoc := new Assoc_Value;
        ref  : Assoc_Ptr;
    begin
        ref.Set( this );
        return ref;
    end Create_Assoc;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out Assoc_Value ) is
    begin
        this.val.Clear;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    function Clone( this : access Assoc_Value ) return Value_Ptr'Class is
        use Value_Maps;
        copy : constant Assoc_Ptr := Create_Assoc;
        pos  : Cursor;
    begin
        pos := this.val.First;
        while Has_Element( pos ) loop
            copy.Get.Set( Key( pos ), Element( pos ).Clone );
            Next( pos );
        end loop;
        return copy;
    end Clone;

    ----------------------------------------------------------------------------

    function Compare( this : Assoc_Value; other : Value'Class ) return Integer is
        use Value_Maps;
        thisIter,
        otherIter : Cursor;
        tmp       : Integer;
    begin
        if this.Get_Type = other.Get_Type then
            thisIter := this.val.First;
            otherIter := Assoc_Value(other).val.First;

            while Has_Element( thisIter ) and then
                  Has_Element( otherIter )
            loop
                -- compare the keys
                if Key( thisIter ) < Key( otherIter ) then
                    return -1;
                elsif Key( thisIter ) > Key( otherIter ) then
                    return 1;
                end if;

                -- compare the values
                -- one or both of the values might be a null pointer
                if Element( thisIter ) /= Values.Nul then
                    -- compare the non-null values
                    tmp := Element( thisIter ).Get.Compare( Element( otherIter ).Get );
                    if tmp /= 0 then
                        return tmp;
                    end if;
                elsif Element( otherIter ) /= Values.Nul then
                    -- this is null and the other is not
                    return -1;
                end if;

                Next( thisIter );
                Next( otherIter );
            end loop;

            if Has_Element( thisIter ) then
                -- this association has more keys; this is greater
                return 1;
            elsif Has_Element( otherIter ) then
                -- other association has more keys; this is less
                return -1;
            else
                -- both associations have the same keys and values; equal
                return 0;
            end if;
        elsif this.Get_Type < other.Get_Type then
            return -1;
        else
            return 1;
        end if;
    end Compare;

    ----------------------------------------------------------------------------

    function Get( this : Assoc_Value; field : String ) return Value_Ptr is
        pos : Value_Maps.Cursor;
    begin
        pos := this.val.Find( field );
        if Value_Maps.Has_Element( pos ) then
            return Value_Ptr(Value_Maps.Element( pos ).Get.Clone);
        end if;
        return Create_Null;
    end Get;

    ----------------------------------------------------------------------------

    function Get_Type( this : Assoc_Value ) return Value_Type is
        pragma Unreferenced( this );
    begin
        return V_ASSOCIATION;
    end Get_Type;

    ----------------------------------------------------------------------------

    overriding
    function Image( this : Assoc_Value ) return String is
        use Value_Maps;
        result : Unbounded_String;
        pos    : Cursor := this.val.First;
        nxt    : Cursor;
    begin
        Append( result, "{" );
        while Has_Element( pos ) loop
            Append( result, """" );
            Append( result, Values.Strings.Escape( Key( pos ) ) );
            Append( result, """:" );
            Append( result, Image( Element( pos ) ) );

            nxt := Next( pos );
            if Has_Element( nxt ) then
                Append( result, "," );
            end if;

            pos := nxt;
        end loop;
        Append( result, "}" );
        return To_String( result );
    end Image;

    ----------------------------------------------------------------------------

    procedure Iterate( this    : Assoc_Value;
                       examine : not null access procedure( key   : String;
                                                            value : Value_Ptr ) ) is
        use Value_Maps;
        pos  : Value_Maps.Cursor := this.val.First;
    begin
        while Has_Element( pos ) loop
            examine.all( Value_Maps.Key( pos ), Element( pos ).Clone );
            Next( pos );
        end loop;
    end Iterate;

    ----------------------------------------------------------------------------

    function Is_Empty( this : Assoc_Value ) return Boolean is
    begin
        return this.val.Is_Empty;
    end Is_Empty;

    ----------------------------------------------------------------------------

    function Get_Keys( this : Assoc_Value ) return List_Ptr is
        keys : constant List_Ptr := Create_List;
        pos  : Value_Maps.Cursor := this.val.First;
    begin
        while Value_Maps.Has_Element( pos ) loop
            keys.Get.Append( Strings.Create( Value_Maps.Key( pos ) ).As_Value );
            Value_Maps.Next( pos );
        end loop;
        return keys;
    end Get_Keys;

    ----------------------------------------------------------------------------

    function Size( this : Assoc_Value ) return Natural is
    begin
        return Natural(this.val.Length);
    end Size;

    ----------------------------------------------------------------------------

    procedure Set( this  : in out Assoc_Value;
                   field : String;
                   val   : Value_Ptr'Class ) is
        pos : Value_Maps.Cursor;
    begin
        pos := this.val.Find( field );
        if Value_Maps.Has_Element( pos ) then
            -- replace it if the value is not null, or remove the key if it is
            if Value_Ptr(val) /= Values.Nul and then val.Get.Get_Type /= V_NULL then
                this.val.Replace_Element( pos, Value_Ptr(val.Get.Clone) );
            else
                this.val.Delete( pos );
            end if;
        else
            -- insert if the value is not null
            if Value_Ptr(val) /= Values.Nul and then val.Get.Get_Type /= V_NULL then
                this.val.Insert( field, Value_Ptr(val.Get.Clone) );
            end if;
        end if;
    end Set;

    ----------------------------------------------------------------------------

    overriding
    function Value_Input( stream : access Root_Stream_Type'Class ) return Assoc_Value is
        this : aliased Assoc_Value;
    begin
        Assoc_Value'Read( stream, this );
        return this;
    end Value_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Value_Read( stream : access Root_Stream_Type'Class; this : out Assoc_Value ) is
        len   : Integer;
        key   : Unbounded_String;
        value : Value_Ptr;
    begin
        len := Integer'Input( stream );
        for i in 1..len loop
            key := To_Unbounded_String( String'Input( stream ) );
            value := Value_Ptr'Input( stream );
            begin
                -- may raise an exception on duplicate key
                this.val.Insert( To_String( key ), value );
            exception
                when others =>
                    null;
            end;
        end loop;
    end Value_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Value_Write( stream : access Root_Stream_Type'Class; this : Assoc_Value ) is
        use Value_Maps;
        pos : Value_Maps.Cursor;
    begin
        Natural'Output( stream, Natural(this.val.Length) );
        pos := this.val.First;
        while Has_Element( pos ) loop
            String'Output( stream, Key( pos ) );
            Value_Ptr'Output( stream, Element( pos ) );
            Next( pos );
        end loop;
    end Value_Write;

    --==========================================================================

    function As_Assoc( ptr : Value_Ptr'Class ) return Assoc_Ptr is
        result : Assoc_Ptr;
    begin
        if ptr.target /= null then
            if ptr.target.Get_Type = V_ASSOCIATION then
                result.Set( ptr.target );
            end if;
        end if;
        return result;
    end As_Assoc;

    ----------------------------------------------------------------------------

    function As_Value( this : Assoc_Ptr ) return Value_Ptr is
        result : Value_Ptr;
    begin
        result.Set( this.target );
        return result;
    end As_Value;

    ----------------------------------------------------------------------------

    function Get( this : Assoc_Ptr ) return access Assoc_Value'Class is
    begin
        return A_Naked_Assoc(this.target);
    end Get;

end Values.Associations;
