--
-- Copyright (c) 2012-2013 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Values.Lists;                      use Values.Lists;

private with Ada.Containers.Indefinite_Ordered_Maps;

package Values.Associations is

    type Assoc_Value is new Value with private;
    type Assoc_Ptr is new Value_Ptr with private;

    -- Constructor
    function Create_Assoc return Assoc_Ptr;

    overriding
    function Clone( this : access Assoc_Value ) return Value_Ptr'Class;

    overriding
    function Compare( this : Assoc_Value; other : Value'Class ) return Integer;

    overriding
    function Get_Type( this : Assoc_Value ) return Value_Type;

    overriding
    function Image( this : Assoc_Value ) return String;

    -- Access and modify the internal value

    -- Returns the corresponding value for the key named 'field', or a Null
    -- value if the key is not defined.
    function Get( this : Assoc_Value; field : String ) return Value_Ptr;

    -- Returns True if there are no keys defined in the association.
    function Is_Empty( this : Assoc_Value ) return Boolean;

    -- Iterates across each key/value pair in the association, calling 'examine'
    -- once for each pair. The order of the iteration is not defined.
    procedure Iterate( this    : Assoc_Value;
                       examine : not null access procedure( key   : String;
                                                            value : Value_Ptr ) );

    -- Returns a list of the keys in the association as a list of string values.
    -- The order of the keys is not defined.
    function Get_Keys( this : Assoc_Value ) return List_Ptr;

    -- Returns the number of keys in the association.
    function Size( this : Assoc_Value ) return Natural;

    -- Sets or updates the key named 'field' in the association. If 'val' is a
    -- null pointer or a Null value then the key will be removed from the
    -- association.
    procedure Set( this  : in out Assoc_Value;
                   field : String;
                   val   : Value_Ptr'Class );

    ----------------------------------------------------------------------------

    -- Casts a Value_Ptr down to an Assoc_Value. Returns Nul on failure.
    function As_Assoc( ptr : Value_Ptr'Class ) return Assoc_Ptr;

    -- Casts an Assoc_Value up to a Value_Ptr.
    function As_Value( this : Assoc_Ptr ) return Value_Ptr;

    -- Returns an access to the Assoc_Value, or null if no target.
    function Get( this : Assoc_Ptr ) return access Assoc_Value'Class;

    Nul : constant Assoc_Ptr;

private

    package Value_Maps is new Ada.Containers.Indefinite_Ordered_Maps( String, Value_Ptr, "<", "=" );

    type Assoc_Value is new Value with
        record
            val : Value_Maps.Map;
        end record;
    type A_Naked_Assoc is access all Assoc_Value'Class;

    overriding
    procedure Delete( this : in out Assoc_Value );

    overriding
    function Value_Input( stream : access Root_Stream_Type'Class ) return Assoc_Value;
    for Assoc_Value'Input use Value_Input;

    overriding
    procedure Value_Read( stream : access Root_Stream_Type'Class; this : out Assoc_Value );
    for Assoc_Value'Read use Value_Read;

    procedure Value_Write( stream : access Root_Stream_Type'Class; this : Assoc_Value );
    for Assoc_Value'Write use Value_Write;


    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    type Assoc_Ptr is new Value_Ptr with null record;

    Nul : constant Assoc_Ptr := (Value_Ptr with others => <>);

end Values.Associations;
