--
-- Copyright (c) 2012-2013 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Containers;                    use Ada.Containers;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Values.Nulls;                      use Values.Nulls;

package body Values.Lists is

    function Create_List return List_Ptr is
        this : constant A_Naked_List := new List_Value;
        ref  : List_Ptr;
    begin
        ref.Set( this );
        return ref;
    end Create_List;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out List_Value ) is
    begin
        this.val.Clear;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Append( this : in out List_Value; val : Value_Ptr'Class ) is
    begin
        if Value_Ptr(val) /= Values.Nul then
            this.val.Append( Value_Ptr(val.Get.Clone) );
        else
            this.val.Append( Create_Null );
        end if;
    end Append;

    ----------------------------------------------------------------------------

    overriding
    function Clone( this : access List_Value ) return Value_Ptr'Class is
        copy : constant List_Ptr := Create_List;
    begin
        for i in 1..this.val.Length loop
            copy.Get.Append( this.val.Element( Positive(i) ).Clone );
        end loop;
        return copy;
    end Clone;

    ----------------------------------------------------------------------------

    overriding
    function Compare( this : List_Value; other : Value'Class ) return Integer is
        tmp : Integer;
    begin
        if this.Get_Type = other.Get_Type then
            for i in 1..this.val.Length loop
                if i > List_Value(other).val.Length then
                    -- longer than 'other'
                    return 1;
                else
                    -- compare each element
                    tmp := this.val.Element( Positive(i) ).Get.Compare( List_Value(other).val.Element( Positive(i) ).Get.all );
                    if tmp /= 0 then
                        -- the elements aren't equal
                        return tmp;
                    end if;
                end if;
            end loop;
            if this.val.Length = List_Value(other).val.Length then
                -- same length, same contents
                return 0;
            else
                -- 'other' is longer
                return -1;
            end if;
        elsif this.Get_Type < other.Get_Type then
            return -1;
        else
            return 1;
        end if;
    end Compare;

    ----------------------------------------------------------------------------

    function Get( this : List_Value; index : Positive ) return Value_Ptr is
    begin
        if Integer(index) > Integer(this.val.Length) then
            return Create_Null;
        end if;
        return Value_Ptr(this.val.Element( index ).Get.Clone);
    end Get;

    ----------------------------------------------------------------------------

    overriding
    function Get_Type( this : List_Value ) return Value_Type is
        pragma Unreferenced( this );
    begin
        return V_LIST;
    end Get_Type;

    ----------------------------------------------------------------------------

    overriding
    function Image( this : List_Value ) return String is
        result : Unbounded_String;
    begin
        Append( result, "[" );
        if this.val.Length > 0 then
            Append( result, Image( this.val.Element( 1 ) ) );
            for i in 2..this.val.Length loop
                Append( result, "," );
                Append( result, Image( this.val.Element( Integer(i) ) ) );
            end loop;
        end if;
        Append( result, "]" );
        return To_String( result );
    end Image;

    ----------------------------------------------------------------------------

    function Length( this : List_Value ) return Natural is
    begin
        return Natural(this.val.Length);
    end Length;

    ----------------------------------------------------------------------------

    procedure Set( this : in out List_Value; index : Positive; val : Value_Ptr'Class ) is
    begin
        if Integer(index) > Integer(this.val.Length) then
            this.val.Append( Create_Null, Count_Type(index) - this.val.Length );
        end if;
        if Value_Ptr(val) /= Values.Nul then
            this.val.Replace_Element( index, Value_Ptr(val.Get.Clone) );
        else
            this.val.Replace_Element( index, Create_Null );
        end if;
    end Set;

    ----------------------------------------------------------------------------

    overriding
    function Value_Input( stream : access Root_Stream_Type'Class ) return List_Value is
        this : aliased List_Value;
    begin
        List_Value'Read( stream, this );
        return this;
    end Value_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Value_Read( stream : access Root_Stream_Type'Class; this : out List_Value ) is
        len : Integer;
    begin
        len := Integer'Input( stream );
        for i in 1..len loop
            this.val.Append( Value_Ptr'Input( stream ) );
        end loop;
    end Value_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Value_Write( stream : access Root_Stream_Type'Class; this : List_Value ) is
    begin
        Natural'Output( stream, Natural(this.val.Length) );
        for i in 1..this.val.Length loop
            Value_Ptr'Output( stream, this.val.Element( Natural(i) ) );
        end loop;
    end Value_Write;

    --==========================================================================

    function As_List( ptr : Value_Ptr'Class ) return List_Ptr is
        result : List_Ptr;
    begin
        if ptr.target /= null then
            if ptr.target.Get_Type = V_LIST then
                result.Set( ptr.target );
            end if;
        end if;
        return result;
    end As_List;

    ----------------------------------------------------------------------------

    function As_Value( this : List_Ptr ) return Value_Ptr is
        result : Value_Ptr;
    begin
        result.Set( this.target );
        return result;
    end As_Value;

    ----------------------------------------------------------------------------

    function Get( this : List_Ptr ) return access List_Value'Class is
    begin
        return A_Naked_List(this.target);
    end Get;

end Values.Lists;
