--
-- Copyright (c) 2012-2013 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Support;                           use Support;

package body Values.Numbers is

    function Create( val : Integer ) return Number_Ptr is
        this : constant A_Naked_Value := new Number_Value'(Value with val => Long_Float(val) );
        ref  : Number_Ptr;
    begin
        ref.Set( this );
        return ref;
    end Create;

    ----------------------------------------------------------------------------

    function Create( val : Long_Float ) return Number_Ptr is
        this : constant A_Naked_Value := new Number_Value'(Value with val => val );
        ref  : Number_Ptr;
    begin
        ref.Set( this );
        return ref;
    end Create;

    ----------------------------------------------------------------------------

    overriding
    function Clone( this : access Number_Value ) return Value_Ptr'Class is
    begin
        return Numbers.To_Ptr(this);
    end Clone;

    ----------------------------------------------------------------------------

    overriding
    function Compare( this : Number_Value; other : Value'Class ) return Integer is
    begin
        if this.Get_Type = other.Get_Type then
            if this.val = Number_Value(other).val then
                return 0;
            elsif this.val < Number_Value(other).val then
                return -1;
            else
                return 1;
            end if;
        elsif this.Get_Type < other.Get_Type then
            return -1;
        else
            return 1;
        end if;
    end Compare;

    ----------------------------------------------------------------------------

    overriding
    function Get_Type( this : Number_Value ) return Value_Type is
        pragma Unreferenced( this );
    begin
        return V_NUMBER;
    end Get_Type;

    ----------------------------------------------------------------------------

    overriding
    function Image( this : Number_Value ) return String is
    begin
        if this.val = Long_Float'Rounding( this.val ) then
            return Image( Integer(this.val) );
        else
            return Image( Float(this.val), precision => 3 );
        end if;
    end Image;

    ----------------------------------------------------------------------------

    function To_Float( this : Number_Value ) return Float is
    begin
        return Float(this.val);
    end To_Float;

    ----------------------------------------------------------------------------

    function To_Int( this : Number_Value ) return Integer is
    begin
        return Integer(this.val);
    end To_Int;

    ----------------------------------------------------------------------------

    overriding
    function Value_Input( stream : access Root_Stream_Type'Class ) return Number_Value is
        this : aliased Number_Value;
    begin
        Number_Value'Read( stream, this );
        return this;
    end Value_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Value_Read( stream : access Root_Stream_Type'Class; this : out Number_Value ) is
    begin
        this.val := Long_Float'Input( stream );
    end Value_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Value_Write( stream : access Root_Stream_Type'Class; this : Number_Value ) is
    begin
        Long_Float'Output( stream, this.val );
    end Value_Write;

    ----------------------------------------------------------------------------

    function   "-"( l : Number_Value ) return Number_Ptr'Class is begin return Create(   - l.val ); end "-";
    function "abs"( l : Number_Value ) return Number_Ptr'Class is begin return Create( abs l.val ); end "abs";

    function  "+"( l, r : Number_Value ) return Number_Ptr'Class is begin return Create( l.val + r.val ); end "+";
    function  "-"( l, r : Number_Value ) return Number_Ptr'Class is begin return Create( l.val - r.val ); end "-";
    function  "*"( l, r : Number_Value ) return Number_Ptr'Class is begin return Create( l.val * r.val ); end "*";
    function  "/"( l, r : Number_Value ) return Number_Ptr'Class is begin return Create( l.val / r.val ); end "/";

    function "**"( l, r : Number_Value ) return Number_Ptr'Class is
    begin
        if r.val < 0.0 then
            raise Constraint_Error with "Negative exponent";
        end if;
        return Create( l.val ** Natural(r.val) );
    end "**";

    function "mod"( l, r : Number_Value ) return Number_Ptr'Class is
    begin
        return Create( Integer(l.val) mod Integer(r.val) );
    end "mod";

    function "rem"( l, r : Number_Value ) return Number_Ptr'Class is
    begin
        return Create( Integer(l.val) rem Integer(r.val) );
    end "rem";

    --==========================================================================

    function As_Number( ptr : Value_Ptr'Class ) return Number_Ptr is
        result : Number_Ptr;
    begin
        if ptr.target /= null then
            if ptr.target.Get_Type = V_NUMBER then
                result.Set( ptr.target );
            end if;
        end if;
        return result;
    end As_Number;

    ----------------------------------------------------------------------------

    function As_Value( this : Number_Ptr ) return Value_Ptr is
        result : Value_Ptr;
    begin
        result.Set( this.target );
        return result;
    end As_Value;

    ----------------------------------------------------------------------------

    function Get( this : Number_Ptr ) return access Number_Value'Class is
    begin
        return A_Naked_Number(this.target);
    end Get;

    ----------------------------------------------------------------------------

    function   "-"( l : Number_Ptr ) return Number_Ptr is begin return Number_Ptr(  - l.Get.all); end "-";
    function "abs"( l : Number_Ptr ) return Number_Ptr is begin return Number_Ptr(abs l.Get.all); end "abs";

    function   "+"( l, r : Number_Ptr ) return Number_Ptr is begin return Number_Ptr(l.Get.all +   r.Get.all); end "+";
    function   "-"( l, r : Number_Ptr ) return Number_Ptr is begin return Number_Ptr(l.Get.all -   r.Get.all); end "-";
    function   "*"( l, r : Number_Ptr ) return Number_Ptr is begin return Number_Ptr(l.Get.all *   r.Get.all); end "*";
    function   "/"( l, r : Number_Ptr ) return Number_Ptr is begin return Number_Ptr(l.Get.all /   r.Get.all); end "/";
    function  "**"( l, r : Number_Ptr ) return Number_Ptr is begin return Number_Ptr(l.Get.all **  r.Get.all); end "**";
    function "mod"( l, r : Number_Ptr ) return Number_Ptr is begin return Number_Ptr(l.Get.all mod r.Get.all); end "mod";
    function "rem"( l, r : Number_Ptr ) return Number_Ptr is begin return Number_Ptr(l.Get.all mod r.Get.all); end "rem";

    ----------------------------------------------------------------------------

    function  "="( l : Number_Ptr; r : Integer    ) return Boolean is begin return l =  Create( r ); end "=";
    function  "="( l : Number_Ptr; r : Long_Float ) return Boolean is begin return l =  Create( r ); end "=";
    function  "<"( l : Number_Ptr; r : Integer    ) return Boolean is begin return l <  Create( r ); end "<";
    function  "<"( l : Number_Ptr; r : Long_Float ) return Boolean is begin return l <  Create( r ); end "<";
    function  ">"( l : Number_Ptr; r : Integer    ) return Boolean is begin return l >  Create( r ); end ">";
    function  ">"( l : Number_Ptr; r : Long_Float ) return Boolean is begin return l >  Create( r ); end ">";
    function "<="( l : Number_Ptr; r : Integer    ) return Boolean is begin return l <= Create( r ); end "<=";
    function "<="( l : Number_Ptr; r : Long_Float ) return Boolean is begin return l <= Create( r ); end "<=";
    function ">="( l : Number_Ptr; r : Integer    ) return Boolean is begin return l >= Create( r ); end ">=";
    function ">="( l : Number_Ptr; r : Long_Float ) return Boolean is begin return l >= Create( r ); end ">=";

    function   "+"( l : Number_Ptr; r : Integer    ) return Number_Ptr is begin return l +   Create( r ); end "+";
    function   "+"( l : Number_Ptr; r : Long_Float ) return Number_Ptr is begin return l +   Create( r ); end "+";
    function   "-"( l : Number_Ptr; r : Integer    ) return Number_Ptr is begin return l -   Create( r ); end "-";
    function   "-"( l : Number_Ptr; r : Long_Float ) return Number_Ptr is begin return l -   Create( r ); end "-";
    function   "*"( l : Number_Ptr; r : Integer    ) return Number_Ptr is begin return l *   Create( r ); end "*";
    function   "*"( l : Number_Ptr; r : Long_Float ) return Number_Ptr is begin return l *   Create( r ); end "*";
    function   "/"( l : Number_Ptr; r : Integer    ) return Number_Ptr is begin return l /   Create( r ); end "/";
    function   "/"( l : Number_Ptr; r : Long_Float ) return Number_Ptr is begin return l /   Create( r ); end "/";
    function  "**"( l : Number_Ptr; r : Integer    ) return Number_Ptr is begin return l **  Create( r ); end "**";
    function  "**"( l : Number_Ptr; r : Long_Float ) return Number_Ptr is begin return l **  Create( r ); end "**";
    function "mod"( l : Number_Ptr; r : Integer    ) return Number_Ptr is begin return l mod Create( r ); end "mod";
    function "rem"( l : Number_Ptr; r : Integer    ) return Number_Ptr is begin return l rem Create( r ); end "rem";

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    function  "="( l : Integer;    r : Number_Ptr ) return Boolean is begin return Create( l ) =  r; end "=";
    function  "="( l : Long_Float; r : Number_Ptr ) return Boolean is begin return Create( l ) =  r; end "=";
    function  "<"( l : Integer;    r : Number_Ptr ) return Boolean is begin return Create( l ) <  r; end "<";
    function  "<"( l : Long_Float; r : Number_Ptr ) return Boolean is begin return Create( l ) <  r; end "<";
    function  ">"( l : Integer;    r : Number_Ptr ) return Boolean is begin return Create( l ) >  r; end ">";
    function  ">"( l : Long_Float; r : Number_Ptr ) return Boolean is begin return Create( l ) >  r; end ">";
    function "<="( l : Integer;    r : Number_Ptr ) return Boolean is begin return Create( l ) <= r; end "<=";
    function "<="( l : Long_Float; r : Number_Ptr ) return Boolean is begin return Create( l ) <= r; end "<=";
    function ">="( l : Integer;    r : Number_Ptr ) return Boolean is begin return Create( l ) >= r; end ">=";
    function ">="( l : Long_Float; r : Number_Ptr ) return Boolean is begin return Create( l ) >= r; end ">=";

    function   "+"( l : Integer;    r : Number_Ptr ) return Number_Ptr is begin return Create( l ) +   r; end "+";
    function   "+"( l : Long_Float; r : Number_Ptr ) return Number_Ptr is begin return Create( l ) +   r; end "+";
    function   "-"( l : Integer;    r : Number_Ptr ) return Number_Ptr is begin return Create( l ) -   r; end "-";
    function   "-"( l : Long_Float; r : Number_Ptr ) return Number_Ptr is begin return Create( l ) -   r; end "-";
    function   "*"( l : Integer;    r : Number_Ptr ) return Number_Ptr is begin return Create( l ) *   r; end "*";
    function   "*"( l : Long_Float; r : Number_Ptr ) return Number_Ptr is begin return Create( l ) *   r; end "*";
    function   "/"( l : Integer;    r : Number_Ptr ) return Number_Ptr is begin return Create( l ) /   r; end "/";
    function   "/"( l : Long_Float; r : Number_Ptr ) return Number_Ptr is begin return Create( l ) /   r; end "/";
    function  "**"( l : Integer;    r : Number_Ptr ) return Number_Ptr is begin return Create( l ) **  r; end "**";
    function  "**"( l : Long_Float; r : Number_Ptr ) return Number_Ptr is begin return Create( l ) **  r; end "**";
    function "mod"( l : Integer;    r : Number_Ptr ) return Number_Ptr is begin return Create( l ) mod r; end "mod";
    function "rem"( l : Integer;    r : Number_Ptr ) return Number_Ptr is begin return Create( l ) rem r; end "rem";

end Values.Numbers;
