--
-- Copyright (c) 2012-2013 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;

package Values.Strings is

    type String_Value is new Value with private;
    type String_Ptr is new Value_Ptr with private;

    -- Constructors
    function Create( val : String ) return String_Ptr;
    function Create( val : Unbounded_String ) return String_Ptr;

    overriding
    function Clone( this : access String_Value ) return Value_Ptr'Class;

    overriding
    function Compare( this : String_Value; other : Value'Class ) return Integer;

    overriding
    function Get_Type( this : String_Value ) return Value_Type;

    overriding
    function Image( this : String_Value ) return String;

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    -- Access the internal value

    function Length( this : String_Value ) return Natural;

    function To_String( this : String_Value ) return String;

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    -- String operators
    function "&"( l, r : String_Value ) return String_Ptr'Class;

    ----------------------------------------------------------------------------

    -- Casts a Value_Ptr down to a String_Value. Returns Nul on failure.
    function As_String( ptr : Value_Ptr'Class ) return String_Ptr;

    -- Casts a String_Value up to a Value_Ptr.
    function As_Value( this : String_Ptr ) return Value_Ptr;

    -- Returns an access to the String_Value, or null if no target.
    function Get( this : String_Ptr ) return access String_Value'Class;

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    -- The following string operators are equivalent to "l.Get [op] r.Get"
    -- Both 'l' and 'r' must have targets or the behavior is undefined.
    function "&"( l, r : String_Ptr ) return String_Ptr; pragma Inline( "&" );

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    -- The following operators are useful for mixing String_Ptrs with primitive
    -- string types. Equivalent to: "l [op] Create( r )" or "Create( r ) [op] l"
    function  "="( l : String_Ptr; r : String           ) return Boolean; pragma Inline( "=" );
    function  "="( l : String_Ptr; r : Unbounded_String ) return Boolean; pragma Inline( "=" );
    function  "<"( l : String_Ptr; r : String           ) return Boolean; pragma Inline( "<" );
    function  "<"( l : String_Ptr; r : Unbounded_String ) return Boolean; pragma Inline( "<" );
    function  ">"( l : String_Ptr; r : String           ) return Boolean; pragma Inline( ">" );
    function  ">"( l : String_Ptr; r : Unbounded_String ) return Boolean; pragma Inline( ">" );
    function "<="( l : String_Ptr; r : String           ) return Boolean; pragma Inline( "<=" );
    function "<="( l : String_Ptr; r : Unbounded_String ) return Boolean; pragma Inline( "<=" );
    function ">="( l : String_Ptr; r : String           ) return Boolean; pragma Inline( ">=" );
    function ">="( l : String_Ptr; r : Unbounded_String ) return Boolean; pragma Inline( ">=" );

    function "&"( l : String_Ptr; r : String           ) return String_Ptr; pragma Inline( "&" );
    function "&"( l : String_Ptr; r : Unbounded_String ) return String_Ptr; pragma Inline( "&" );

    function  "="( l : String;           r : String_Ptr ) return Boolean; pragma Inline( "=" );
    function  "="( l : Unbounded_String; r : String_Ptr ) return Boolean; pragma Inline( "=" );
    function  "<"( l : String;           r : String_Ptr ) return Boolean; pragma Inline( "<" );
    function  "<"( l : Unbounded_String; r : String_Ptr ) return Boolean; pragma Inline( "<" );
    function  ">"( l : String;           r : String_Ptr ) return Boolean; pragma Inline( ">" );
    function  ">"( l : Unbounded_String; r : String_Ptr ) return Boolean; pragma Inline( ">" );
    function "<="( l : String;           r : String_Ptr ) return Boolean; pragma Inline( "<=" );
    function "<="( l : Unbounded_String; r : String_Ptr ) return Boolean; pragma Inline( "<=" );
    function ">="( l : String;           r : String_Ptr ) return Boolean; pragma Inline( ">=" );
    function ">="( l : Unbounded_String; r : String_Ptr ) return Boolean; pragma Inline( ">=" );

    function "&"( l : String;           r : String_Ptr ) return String_Ptr; pragma Inline( "&" );
    function "&"( l : Unbounded_String; r : String_Ptr ) return String_Ptr; pragma Inline( "&" );

    Nul : constant String_Ptr;

    -- Escapes special characters in a string with backslash notation.
    -- Tab => \t
    -- LF  => \n
    -- CR  => \r
    -- \   => \\
    -- "   => \"
    function Escape( str : String ) return String;

private

    type String_Value is new Value with
        record
            val : Unbounded_String;
        end record;
    type A_Naked_String is access all String_Value'Class;

    overriding
    function Value_Input( stream : access Root_Stream_Type'Class ) return String_Value;
    for String_Value'Input use Value_Input;

    overriding
    procedure Value_Read( stream : access Root_Stream_Type'Class; this : out String_Value );
    for String_Value'Read use Value_Read;

    procedure Value_Write( stream : access Root_Stream_Type'Class; this : String_Value );
    for String_Value'Write use Value_Write;

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    type String_Ptr is new Value_Ptr with null record;

    Nul : constant String_Ptr := (Value_Ptr with others => <>);

end Values.Strings;
