--
-- Copyright (c) 2012-2013 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Tags;                          use Ada.Tags;
with Ada.Unchecked_Deallocation;
with Reference_Counting;                use Reference_Counting;

package body Values is

    function "<"( l, r : Value_Type ) return Boolean is
    begin
        return Value_Type'Pos( l ) < Value_Type'Pos( r );
    end "<";

    --==========================================================================

    function Compare( this : Value'Class; other : access Value'Class ) return Integer is
    begin
        if other /= null then
            return this.Compare( other.all );
        else
            -- other is null and this isn't
            return 1;
        end if;
    end Compare;

    ----------------------------------------------------------------------------

    function "="( this : Value; other : Value ) return Boolean is
    begin
        return Value'Class(this).Compare( Value'Class(other) ) = 0;
    end "=";

    function "<"( this : Value'Class; other : Value'Class ) return Boolean is
    begin
        return this.Compare( other ) < 0;
    end "<";

    function ">"( this : Value'Class; other : Value'Class ) return Boolean is
    begin
        return this.Compare( other ) > 0;
    end ">";

    function "<="( this : Value'Class; other : Value'Class ) return Boolean is
    begin
        return this.Compare( other ) <= 0;
    end "<=";

    function ">="( this : Value'Class; other : Value'Class ) return Boolean is
    begin
        return this.Compare( other ) >= 0;
    end ">=";

    --==========================================================================

    overriding
    procedure Adjust( this : in out Value_Ptr ) is
        tmp : Integer_32;
        pragma Unreferenced( tmp );
    begin
        if this.target /= null then
            tmp := Atomic.Add( this.target.refs'Access, 1 );
        end if;
    end Adjust;

    ----------------------------------------------------------------------------

    overriding
    procedure Finalize( this : in out Value_Ptr ) is
        procedure Free is new Ada.Unchecked_Deallocation( Value'Class,
                                                          A_Naked_Value );

        target : A_Naked_Value := this.target;
    begin
        this.target := null;

        if target /= null then
            if Atomic.Add( target.refs'Access, -1 ) = 0 then
                Delete( target.all );
                Free( target );
            end if;
        end if;
    end Finalize;

    ----------------------------------------------------------------------------

    function Clone( this : Value_Ptr'Class ) return Value_Ptr is
    begin
        if Value_Ptr(this) /= Nul then
            return Value_Ptr(this.Get.Clone);
        end if;
        return Nul;
    end Clone;

    ----------------------------------------------------------------------------

    function Get( this : Value_Ptr ) return access Value'Class is
    begin
        return this.target;
    end Get;

    ----------------------------------------------------------------------------

    function Refcount( this : Value_Ptr'Class ) return Natural is
    begin
        if this.target /= null then
            return Natural(this.target.refs);
        end if;
        return 0;
    end Refcount;

    ----------------------------------------------------------------------------

    function Image( this : Value_Ptr'Class ) return String is
    begin
        if this.target /= null then
            return this.target.Image;
        else
            return "";
        end if;
    end Image;

    ----------------------------------------------------------------------------

    function Is_Null( this : Value_Ptr'Class ) return Boolean is
    begin
        return this.target = null;
    end Is_Null;

    ----------------------------------------------------------------------------

    function Not_Null( this : Value_Ptr'Class ) return Boolean is
    begin
        return this.target /= null;
    end Not_Null;

    ----------------------------------------------------------------------------

    procedure Set( this : in out Value_Ptr; target : access Value'Class ) is
    begin
        if this.target = A_Naked_Value(target) then
            -- the target isn't changing
            return;
        end if;

        if this.target /= null then
            this.Finalize;     -- decrement reference count
        end if;

        if target /= null then
            this.target := A_Naked_Value(target);
            this.Adjust;       -- increment reference count
        end if;
    end Set;

    ----------------------------------------------------------------------------

    function To_Ptr( val : access Value'Class ) return Value_Ptr is
        this : Value_Ptr;
    begin
        this.Set( val );
        return this;
    end To_Ptr;

    ----------------------------------------------------------------------------

    function Value_Ptr_Input( stream : access Root_Stream_Type'Class ) return Value_Ptr is
        result : Value_Ptr;
    begin
        if Boolean'Input( stream ) then
            Value_Ptr'Read( stream, result );
        end if;
        return result;
    end Value_Ptr_Input;

    ----------------------------------------------------------------------------

    procedure Value_Ptr_Read( stream : access Root_Stream_Type'Class; this : out Value_Ptr ) is
        target : A_Naked_Value := null;
    begin
        target := new Value'Class'(Value'Class'Input( stream ));
        this.Set( target );
    exception
        when e : Tag_Error =>
            declare
                msg : constant String := Exception_Message( e );
            begin
                raise Constraint_Error with "Found unknown value class " &
                      "<" & msg((Index( msg, ".", Backward ) + 1)..msg'Last) & ">";
            end;
    end Value_Ptr_Read;

    ----------------------------------------------------------------------------

    procedure Value_Ptr_Output( stream : access Root_Stream_Type'Class; this : Value_Ptr ) is
    begin
        Boolean'Output( stream, this /= Nul );
        if this /= Nul then
            Value_Ptr'Write( stream, this );
        end if;
    end Value_Ptr_Output;

    ----------------------------------------------------------------------------

    procedure Value_Ptr_Write( stream : access Root_Stream_Type'Class; this : Value_Ptr ) is
    begin
        Value'Class'Output( stream, this.target.all );
    end Value_Ptr_Write;

    ----------------------------------------------------------------------------

    overriding
    function "="( l, r : Value_Ptr ) return Boolean is
    begin
        if l.target = null then
            return r.target = null;
        end if;

        -- compare pointers; if different, compare value contents
        return r.target /= null and then
              (l.target = r.target or else l.target.all = r.target.all);
    end "=";

    ----------------------------------------------------------------------------

    function "<"( l, r : Value_Ptr ) return Boolean is
    begin
        -- compare the targets to each other
        if l /= Nul and then r /= Nul then
            return l.Get /= r.Get and then l.Get.all < r.Get.all;
        end if;

        -- if this target is null then the other is not
        return l = Nul;
    end "<";

    ----------------------------------------------------------------------------

    function ">"( l, r : Value_Ptr ) return Boolean is
    begin
        -- compare the targets to each other
        if l /= Nul and then r /= Nul then
            return l.Get /= r.Get and then l.Get.all > r.Get.all;
        end if;

        -- if this target is null then the other is not
        return l /= Nul;
    end ">";

    ----------------------------------------------------------------------------

    function "<="( l, r : Value_Ptr ) return Boolean is
    begin
        return not (l > r);
    end "<=";

    ----------------------------------------------------------------------------

    function ">="( l,r : Value_Ptr ) return Boolean is
    begin
        return not (l < r);
    end ">=";

end Values;
