--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Allegro.Keyboard;                  use Allegro.Keyboard;
with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Engine_Debugging;                  use Engine_Debugging;
with Mouse;                             use Mouse;
with Themes.Palette;                    use Themes.Palette;

package body Widgets.Buttons.Checkboxes is

    function Create_Checkbox( view : not null access Game_Views.Game_View'Class;
                              id   : String;
                              text : String := "";
                              icon : String := "" ) return A_Checkbox is
        this : A_Checkbox := new Checkbox;
    begin
        this.Construct( view, id, text, icon );
        this.align := Align_Left;
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Checkbox;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Checkbox ) is
        theme     : constant A_Theme := Themes.Get;
        icon      : constant A_Allegro_Bitmap := theme.Get_Bitmap( this.icon );
        text      : constant String := To_String( this.text );
        lblwidth  : constant Integer := theme.Label_Width( icon, text,
                                                           To_String( this.fontName ),
                                                           this.fontSize );
        textColor : Allegro_Color := this.colors(Themes.Text);
        chkicon   : A_Allegro_Bitmap := null;
        border    : Border_Type;
        drawX     : Integer;
    begin
        case this.align is
            when Align_Center =>
                -- left to right, centered: ... BOX spacing LABEL ...
                drawX := (this.cwidth / 2);
                if lblwidth > 0 then
                    drawX := drawX - (lblwidth + SPACING) / 2;
                end if;
                drawX := drawX - BOX_SIZE / 2;
            when Align_Left =>
                -- left to right: | BORDER spacing BOX spacing LABEL ...
                drawX := theme.Border_Width + SPACING;
            when Align_Right =>
                -- left to right: ... BOX spacing LABEL spacing BORDER |
                drawX := (this.cwidth - 1) - (theme.Border_Width + SPACING + BOX_SIZE);
                if lblwidth > 0 then
                    drawX := drawX - (lblwidth + SPACING);
                end if;
        end case;

        if this.hover then
            border := Border_Line;
        else
            border := this.border;
            if this.on then
                if this.border = Border_Raised then
                    border := Border_Sunk;
                end if;
            end if;
        end if;

        if this.contrastText then
            textColor := Hard_Contrast( this.colors(Foreground) );
        end if;

        if not this.transparent then
            Rectfill( 0, 0, this.cwidth - 1, this.cheight - 1, this.colors(Foreground) );
        end if;

        -- draw border for the check box
        theme.Draw_Box( drawX, this.cheight / 2 - BOX_SIZE / 2,
                        drawX + BOX_SIZE - 1, this.cheight / 2 + BOX_SIZE / 2 - 1,
                        this.colors(Foreground), border );
        if this.focused then
            Rect( drawX - 1, this.cheight / 2 - BOX_SIZE / 2 - 1,
                  drawX + BOX_SIZE, this.cheight / 2 + BOX_SIZE / 2,
                  this.colors(Selected) );
        end if;

        -- draw the check mark (icon or square)
        if this.on then
            if this.checkIcon > 0 then
                chkicon := theme.Get_Bitmap( this.checkIcon );
            end if;
            if chkicon /= null then
                Draw_Bitmap( chkicon, drawX, this.cheight / 2 - BOX_SIZE / 2 );
            else
                Rectfill( drawX + 2, this.cheight / 2 - BOX_SIZE / 2 + 2,
                          drawX + BOX_SIZE - 3, this.cheight / 2 + BOX_SIZE / 2 - 3,
                          this.colors(Selected) );
            end if;
        end if;
        drawX := drawX + BOX_SIZE + SPACING;

        -- draw the checkbox's label
        if lblwidth > 0 then
            theme.Draw_Label( drawX, theme.Border_Width, drawX + lblwidth - 1,
                              this.cheight - 1 - theme.Border_Width,
                              icon, text,
                              To_String( this.fontName ), this.fontSize,
                              textColor, Align_Left );
        end if;
    end Draw_Content;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Height( this : access Checkbox ) return Natural is
        theme : A_Theme;
    begin
        if this.minHeight > 0 then
            return this.minHeight;
        else
            theme := Themes.Get;
            return theme.Border_Width * 2 +
                   Integer'Max( BOX_SIZE,
                                theme.Label_Height( theme.Get_Bitmap( this.icon ),
                                                    To_String( this.text ),
                                                    To_String( this.fontName ),
                                                    this.fontSize ) );
        end if;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Width( this : access Checkbox ) return Natural is
        theme : A_Theme;
    begin
        if this.minWidth > 0 then
            return this.minWidth;
        else
            theme := Themes.Get;
            return theme.Border_Width * 2 + BOX_SIZE + SPACING +
                   theme.Label_Width( theme.Get_Bitmap( this.icon ),
                                      To_String( this.text ),
                                      To_String( this.fontName ), this.fontSize );
        end if;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Key_Press( this    : access Checkbox;
                            evt     : not null A_Key_Event;
                            handled : in out Boolean ) is
    begin
        if evt.Get_Key = ALLEGRO_KEY_SPACE or else
           evt.Get_Key = ALLEGRO_KEY_ENTER or else
           evt.Get_Key = ALLEGRO_KEY_PAD_ENTER
        then
            A_Button(this).Set_State( not this.on );
            handled := True;
        end if;
    end On_Key_Press;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Press( this : access Checkbox;
                              evt  : not null A_Mouse_Button_Event ) is
    begin
        if evt.Get_Button = Mouse_Left then
            A_Button(this).Set_State( not this.on );
        end if;
    end On_Mouse_Press;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Check_Icon( this : access Checkbox; icon : String ) is
    begin
        if icon'Length > 0 then
            this.checkIcon := Themes.Get.Get_ID( icon );
            pragma Debug( this.checkIcon = 0,
                          Dbg( "Warning: Icon <" & icon & "> not found for widget " &
                               To_String( this ),
                               D_GUI, WARNING ) );
        else
            this.checkIcon := 0;
        end if;
    end Set_Check_Icon;

end Widgets.Buttons.Checkboxes;
