--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Engine_Debugging;                  use Engine_Debugging;
with Themes.Palette;                    use Themes.Palette;

package body Widgets.Buttons is

    not overriding
    procedure Construct( this : access Button;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String;
                         text : String;
                         icon : String ) is
    begin
        Widget(this.all).Construct( view, id );
        this.border := Border_Raised;
        this.Set_Text( text );
        this.Set_Icon( icon );
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add_Listener( this     : access Button;
                            listener : not null A_Button_Listener ) is
    begin
        this.Add_Listener( "Button_Listener", A_Action_Listener(listener) );
    end Add_Listener;

    ----------------------------------------------------------------------------

    procedure Contrast_Text( this : not null access Button'Class; enabled : Boolean ) is
    begin
        if this.contrastText /= enabled then
            this.contrastText := enabled;
        end if;
    end Contrast_Text;

    ----------------------------------------------------------------------------

    not overriding
    procedure Dispatch_Action( this : access Button; id : Action_Id ) is
        action : A_Button_Action := new Button_Action;

        procedure Dispatch( listener : A_Action_Listener ) is
        begin
            A_Button_Listener(listener).Handle_Action( action );
        end Dispatch;

    begin
        action.Construct( id, A_Widget(this) );
        this.Iterate_Listeners( "Button_Listener", Dispatch'Access );
        Delete( A_Object(action) );
    end Dispatch_Action;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Button ) is
        theme     : constant A_Theme := Themes.Get;
        icon      : constant A_Allegro_Bitmap := theme.Get_Bitmap( this.icon );
        border    : Border_Type := this.border;
        bgColor   : Allegro_Color;
        textColor : Allegro_Color := this.colors(Text);
        bw        : constant Natural := theme.Border_Width;
        pad       : Natural := 0;
    begin
        if not this.enabled then
            bgColor := Lighten( this.colors(Foreground), this.shadeDisabled );
        elsif this.on then
            bgColor := Lighten( this.colors(Foreground), this.shadeOn );
            if border = Border_Raised then
                border := Border_Sunk;
            end if;
        else
            bgColor := this.colors(Foreground);
            if this.hover then
                bgColor := Lighten( bgColor, this.shadeHover );
            end if;
        end if;

        if this.contrastText then
            textColor := Hard_Contrast( bgColor );
        end if;

        if not this.transparent then
            theme.Draw_Box( 0, 0, this.cwidth - 1, this.cheight - 1, bgColor, border );
        end if;

        if this.focused and then this.border /= Border_None then
            Rect( bw, bw, this.cwidth - 1 - bw, this.cheight - 1 - bw,
                  this.colors(Selected) );
        end if;
        if this.border /= Border_None then
            pad := bw + SPACING;
        end if;
        theme.Draw_Label( pad, pad,
                          this.cwidth - 1 - pad,
                          this.cheight - 1 - pad,
                          icon, To_String( this.text ),
                          To_String( this.fontName ), this.fontSize,
                          textColor, this.align );
    end Draw_Content;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Height( this : access Button ) return Natural is
        theme : A_Theme;
        icon  : A_Allegro_Bitmap;
        pad   : Natural := 0;
    begin
        if this.minHeight > 0 then
            return this.minHeight;
        else
            theme := Themes.Get;
            icon := theme.Get_Bitmap( this.icon );
            if this.border /= Border_None then
                pad := theme.Border_Width * 2;
            end if;
            return pad + theme.Label_Height( icon, To_String( this.text ),
                                             To_String( this.fontName ), this.fontSize );
        end if;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Width( this : access Button ) return Natural is
        theme : A_Theme;
        icon  : A_Allegro_Bitmap;
        pad   : Natural := 0;
    begin
        if this.minWidth > 0 then
            return this.minWidth;
        else
            theme := Themes.Get;
            icon := theme.Get_Bitmap( this.icon );
            if this.border /= Border_None then
                pad := theme.Border_Width * 2;
            end if;
            return pad + theme.Label_Width( icon, To_String( this.text ),
                                            To_String( this.fontName ), this.fontSize );
        end if;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    function Get_State( this : not null access Button'Class ) return Boolean is
    begin
        return this.on;
    end Get_State;

    ----------------------------------------------------------------------------

    function Get_Text( this : not null access Button'Class ) return String is
    begin
        return To_String( this.text );
    end Get_Text;

    ----------------------------------------------------------------------------

    not overriding
    procedure Remove_Listener( this     : access Button;
                               listener : not null A_Button_Listener ) is
    begin
        this.Remove_Listener( "Button_Listener", A_Action_Listener(listener) );
    end Remove_Listener;

    ----------------------------------------------------------------------------

    procedure Set_Align( this : not null access Button'Class; align : Align_Type ) is
    begin
        if this.align /= align then
            this.align := align;
        end if;
    end Set_Align;

    ----------------------------------------------------------------------------

    overriding
    procedure Set_Color( this    : access Button;
                         purpose : Color_Purpose;
                         color   : Allegro_Color ) is
    begin
        Widget(this.all).Set_Color( purpose, color );
        this.transparent := purpose = Foreground and then
                            color = base_colors(Transparent);
    end Set_Color;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Icon( this : access Button; icon : String ) is
    begin
        if icon'Length > 0 then
            this.icon := Themes.Get.Get_ID( icon );
            pragma Debug( this.icon = 0,
                          Dbg( "Warning: Icon <" & icon & "> not found for widget " &
                               To_String( this ),
                               D_GUI, WARNING ) );
        else
            this.icon := 0;
        end if;
    end Set_Icon;

    ----------------------------------------------------------------------------

    procedure Set_Shade_Disabled( this : not null access Button'Class; shade : Float ) is
    begin
        if shade /= this.shadeDisabled then
            this.shadeDisabled := shade;
        end if;
    end Set_Shade_Disabled;

    ----------------------------------------------------------------------------

    procedure Set_Shade_On(this : not null access Button'Class; shade : Float ) is
    begin
        if shade /= this.shadeOn then
            this.shadeOn := shade;
        end if;
    end Set_Shade_On;

    ----------------------------------------------------------------------------

    procedure Set_Shade_Hover( this : not null access Button'Class; shade : Float ) is
    begin
        if shade /= this.shadeHover then
            this.shadeHover := shade;
        end if;
    end Set_Shade_Hover;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_State( this : access Button; on : Boolean ) is
    begin
        if this.enabled and then this.on /= on then
            this.on := on;
            if this.on then
                this.Dispatch_Action( Press );
            else
                this.Dispatch_Action( Release );
            end if;
        end if;
    end Set_State;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Text( this : access Button; text : String ) is
    begin
        if text /= To_String( this.text ) then
            this.text := To_Unbounded_String( text );
        end if;
    end Set_Text;

    ----------------------------------------------------------------------------

    not overriding
    procedure Toggle_State( this : access Button ) is
    begin
        A_Button(this).Set_State( not this.on );
    end Toggle_State;

    ----------------------------------------------------------------------------

    overriding
    function To_String( this : access Button ) return String is
    begin
        return "<" & this.Get_Class_Name &
                     " : id=" & To_String( this.id ) &
                     ", text='" & To_String( this.text ) & "'" &
               ">";
    end To_String;

end Widgets.Buttons;
