--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Allegro;
with Allegro.Displays;                  use Allegro.Displays;
with Allegro.State;                     use Allegro.State;
with Allegro.Transformations;           use Allegro.Transformations;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Game_Views;                        use Game_Views;
with Interfaces;                        use Interfaces;
with Themes.Palette;                    use Themes.Palette;
with Widgets.Containers;                use Widgets.Containers;
with xBR;

package body Widgets.Containers.Windows is

    use Widget_Lists;

    -- Dispatches a mouse click event to the widget that the mouse clicked on.
    -- If the button release occurred over a different widget than the press,
    -- then the click will be ignored.
    procedure Dispatch_Click( this : not null access Window'Class;
                              evt  : not null A_Mouse_Button_Event );

    -- Dispatches a key press event to the focused widget.
    procedure Dispatch_Key_Press( this    : not null access Window'Class;
                                  evt     : not null A_Key_Event;
                                  handled : in out Boolean );

    -- Dispatches a key release event to the widget that received the key press
    -- event, regardless of input focus.
    procedure Dispatch_Key_Release( this    : not null access Window'Class;
                                    evt     : not null A_Key_Event;
                                    handled : in out Boolean );

    -- Dispatches a key typed event to the focused widget, if the key was
    -- pressed after the widget received focus.
    procedure Dispatch_Key_Typed( this    : not null access Window'Class;
                                  evt     : not null A_Key_Typed_Event;
                                  handled : in out Boolean );

    -- Dispatches a mouse held event to the widget that the mouse was over at
    -- the time the mouse button was pressed.
    procedure Dispatch_Mouse_Held( this : not null access Window'Class;
                                   evt  : not null A_Mouse_Button_Event );

    -- Dispatches a mouse move event. If the mouse is over a different widget
    -- than last move then widget exit and widget enter events are dispatched
    -- appropriately first.
    --
    -- If the left mouse button is being held down, a mouse move event is sent
    -- to the widget that the mouse was over when the left button was pressed.
    -- If the mouse is currently over a widget that isn't the one that the left
    -- mouse button was pressed over, then that widget will also receive a mouse
    -- move event.
    procedure Dispatch_Mouse_Move( this : not null access Window'Class;
                                   evt  : not null A_Mouse_Event );

    -- Dispatches a mouse press event to the widget the mouse is currently over.
    -- If a popup widget (or stack of popup widgets) is being displayed, then
    -- the popup widgets will be popped off the stack and hidden unless they are
    -- under the mouse. This allows context menus, etc, to be hidden when the
    -- user clicks somewhere else in the Window. If a modal dialog widget is
    -- being displayed, then the press will only be dispatched to descendents of
    -- the modal widget.
    procedure Dispatch_Mouse_Press( this : not null access Window'Class;
                                    evt  : not null A_Mouse_Button_Event );

    -- Dispatches a mouse release event to the widget the mouse is over.
    procedure Dispatch_Mouse_Release( this : not null access Window'Class;
                                      evt  : not null A_Mouse_Button_Event );

    -- Dispatches a mouse release event of button 'btn' to the widget on the
    -- screen at 'x', 'y'.
    procedure Dispatch_Mouse_Release( this : not null access Window'Class;
                                      x, y : Integer;
                                      btn  : Mouse_Button );

    -- Dispatches a scroll wheel event to the widget the mouse is over.
    procedure Dispatch_Mouse_Scroll( this    : not null access Window'Class;
                                     evt     : not null A_Mouse_Scroll_Event;
                                     handled : in out Boolean );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Window( view   : access Game_Views.Game_View'Class;
                            id     : String;
                            width,
                            height : Natural;
                            filter : Filter_Type := Filter_Nearest
                          ) return A_Window is
        this : A_Window := new Window;
    begin
        this.Construct( view, id, width, height, filter );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Window;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this   : access Window;
                         view   : access Game_Views.Game_View'Class;
                         id     : String;
                         width,
                         height : Natural;
                         filter : Filter_Type ) is
        state : Allegro_State;
    begin
        Container(this.all).Construct( view, id );
        this.cwidth := width;
        this.cheight := height;
        this.scale := Integer'Max( 1, Integer'Min( Al_Get_Display_Width( view.Get_Display ) / width,
                                                   Al_Get_Display_Height( view.Get_Display ) / height ) );
        this.filter := filter;
        this.x1 := 0;
        this.y1 := 0;
        this.x2 := this.cwidth - 1;
        this.y2 := this.cheight - 1;
        this.focused := True;
        this.focus := A_Widget(this);
        this.modal := A_Widget(this);
        this.outerBorder := base_colors(Black);

        if this.scale > 1 then
            Al_Store_State( state, ALLEGRO_STATE_ALL );
            if this.filter = Filter_Linear then
                Al_Set_New_Bitmap_Flags( ALLEGRO_MIN_LINEAR or
                                         ALLEGRO_MAG_LINEAR );
            end if;
            this.scaleBmp := Al_Create_Bitmap( width, height );
            Al_Restore_State( state );
        end if;

        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Key_Press" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Key_Release" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Key_Typed" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Mouse_Held" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Mouse_Move" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Mouse_Press" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Mouse_Release" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Mouse_Click" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Mouse_Doubleclick" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Mouse_Scroll" );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Window ) is
    begin
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Mouse_Scroll" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Mouse_Doubleclick" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Mouse_Click" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Mouse_Release" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Mouse_Press" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Mouse_Move" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Mouse_Held" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Key_Typed" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Key_Release" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Key_Press" );

        while not this.popups.Is_Empty loop
            this.Pop_Popup;
        end loop;
        Al_Destroy_Bitmap( this.scaleBmp );
        Delete( A_Limited_Object(this.menu) );

        -- clear the widget references so hiding the child widgets before
        -- deleting them won't cause us to reference widgets that were deleted
        -- earlier in the child traversal.
        this.mouseOver := null;
        for i in this.mousePressedOn'Range loop
            this.mousePressedOn(i) := null;
        end loop;
        for i in this.mouseClickedOn'Range loop
            this.mouseClickedOn(i) := null;
        end loop;
        this.focus := null;
        this.modal := null;

        Container(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add_Widget( this    : access Window;
                          child   : in out A_Widget;
                          consume : Boolean := True ) is
    begin
        this.Add_Child( child, consume );
    end Add_Widget;

    ----------------------------------------------------------------------------

    procedure Dispatch_Click( this : not null access Window'Class;
                              evt  : not null A_Mouse_Button_Event ) is
        btn  : constant Mouse_Button := evt.Get_Button;
        evtX : constant Integer := evt.Get_X;
        evtY : constant Integer := evt.Get_Y;
        cx,
        cy   : Integer;
    begin
        -- only send the click to the widget if a press and release occurred
        -- over the same widget
        if this.mouseClickedOn(btn) /= null then
            this.mouseClickedOn(btn).Translate_To_Content( evtX, evtY, cx, cy );
            evt.Set_XY( cx, cy );
            this.mouseClickedOn(btn).On_Click( evt );
            evt.Set_XY( evtX, evtY );
        end if;
    end Dispatch_Click;

    ----------------------------------------------------------------------------

    procedure Dispatch_Key_Press( this    : not null access Window'Class;
                                  evt     : not null A_Key_Event;
                                  handled : in out Boolean ) is
        key : constant Integer := evt.Get_Key;
    begin
        this.keyPressSentTo(key) := this.focus;
        this.keyPressSentTo(key).Handle_Key_Press( evt, handled );
    end Dispatch_Key_Press;

    ----------------------------------------------------------------------------

    procedure Dispatch_Key_Release( this    : not null access Window'Class;
                                    evt     : not null A_Key_Event;
                                    handled : in out Boolean ) is
        key : constant Integer := evt.Get_Key;
    begin
        if this.keyPressSentTo(key) /= null then
            this.keyPressSentTo(key).Handle_Key_Release( evt, handled );
            this.keyPressSentTo(key) := null;
        end if;
    end Dispatch_Key_Release;

    ----------------------------------------------------------------------------

    procedure Dispatch_Key_Typed( this    : not null access Window'Class;
                                  evt     : not null A_Key_Typed_Event;
                                  handled : in out Boolean ) is
        key : constant Integer := evt.Get_Key;
    begin
        if this.keyPressSentTo(key) /= null then
            this.keyPressSentTo(key).Handle_Key_Typed( evt, handled );
        end if;
    end Dispatch_Key_Typed;

    ----------------------------------------------------------------------------

    procedure Dispatch_Mouse_Held( this : not null access Window'Class;
                                   evt  : not null A_Mouse_Button_Event ) is
        btn  : constant Mouse_Button := evt.Get_Button;
        evtX : constant Integer := evt.Get_X;
        evtY : constant Integer := evt.Get_Y;
        cx,
        cy   : Integer;
    begin
        -- if the mouse button was pressed on a widget, then we send a held
        -- event to that widget, even if the mouse isn't currently over it.
        if this.mousePressedOn(btn) /= null then
            this.mousePressedOn(btn).Translate_To_Content( evtX, evtY, cx, cy );
            evt.Set_XY( cx, cy );
            this.mousePressedOn(btn).On_Mouse_Held( evt );
            evt.Set_XY( evtX, evtY );
        end if;
    end Dispatch_Mouse_Held;

    ----------------------------------------------------------------------------

    procedure Dispatch_Mouse_Move( this : not null access Window'Class;
                                   evt  : not null A_Mouse_Event ) is
        evtX    : constant Integer := evt.Get_X;
        evtY    : constant Integer := evt.Get_Y;
        current : A_Widget;
        widgetX,              -- translated location of mouse on the found
        widgetY : Integer;    --   widget, in widget content coordinates
    begin
        this.Find_Widget_At( evtX, evtY, widgetX, widgetY, current );
        if this.mouseOver /= current then
            -- handle the exit and enter events
            if this.mouseOver /= null then
                this.mouseOver.On_Exit;
            end if;
            if current /= null then
                current.On_Enter;
            end if;
            this.mouseOver := current;
        end if;

        declare
            target : A_Widget := this.mousePressedOn(Mouse_Left);
            cx, cy : Integer;
        begin
            -- if the left mouse button was pressed on a widget, then we
            -- send a move event to that widget, even if the mouse isn't
            -- currently over it.
            if target /= null and then target.Is_Visible then
                target.Translate_To_Content( evtX, evtY, cx, cy );
                evt.Set_XY( cx, cy );
                target.On_Mouse_Move( evt );
                evt.Set_XY( evtX, evtY );
            end if;

            -- if the mouse is over a different wiget that the one we just
            -- sent the mouse move to, then also send a mouse move to the widget
            -- the mouse is currently over.
            target := this.mouseOver;
            if target /= null and then
               target /= this.mousePressedOn(Mouse_Left)
            then
                target.Translate_To_Content( evtX, evtY, cx, cy );
                evt.Set_XY( widgetX, widgetY );
                target.On_Mouse_Move( evt );
                evt.Set_XY( evtX, evtY );
            end if;
        end;
    end Dispatch_Mouse_Move;

    ----------------------------------------------------------------------------

    procedure Dispatch_Mouse_Press( this : not null access Window'Class;
                                    evt  : not null A_Mouse_Button_Event ) is
        btn  : constant Mouse_Button := evt.Get_Button;
        evtX : constant Integer := evt.Get_X;
        evtY : constant Integer := evt.Get_Y;
        cx,
        cy   : Integer;
    begin
        if this.mouseOver = null then
            this.Find_Widget_At( evtX, evtY, cx, cy, this.mouseOver );
        end if;
        this.mousePressedOn(btn) := this.mouseOver;

        if this.mousePressedOn(btn) /= null then
            -- iterate through the popup stack, popping each off until we find
            -- a widget that is an ancestor of mousePressedOn(btn).
            while not this.popups.Is_Empty loop
                exit when this.mousePressedOn(btn).Is_Descendant_Of( this.popups.Last_Element );
                this.Pop_Popup;
            end loop;

            -- if the widget the mouse was pressed on is not within the current
            -- modal widget then don't dispatch the press event
            if not this.mousePressedOn(btn).Is_Descendant_Of( this.modal ) then
                this.mousePressedOn(btn) := null;
                this.mouseOver := null;
                return;
            end if;
        end if;

        -- send the press to the widget the mouse is currently over
        if this.mousePressedOn(btn) /= null then
            -- change the focused widget
            this.Give_Focus( this.mousePressedOn(btn) );
            this.mousePressedOn(btn).Translate_To_Content( evtX, evtY, cx, cy );
            evt.Set_XY( cx, cy );
            this.mousePressedOn(btn).On_Mouse_Press( evt );
            evt.Set_XY( evtX, evtY );
        end if;
    end Dispatch_Mouse_Press;

    ----------------------------------------------------------------------------

    procedure Dispatch_Mouse_Release( this : not null access Window'Class;
                                      evt  : not null A_Mouse_Button_Event ) is
    begin
        A_Window(this).Dispatch_Mouse_Release( evt.Get_X, evt.Get_Y, evt.Get_Button );
    end Dispatch_Mouse_Release;

    ----------------------------------------------------------------------------

    procedure Dispatch_Mouse_Release( this : not null access Window'Class;
                                      x, y : Integer;
                                      btn  : Mouse_Button ) is
        evt    : A_Mouse_Button_Event;
        cx, cy : Integer;   -- widget content coordinates
    begin
        -- send the release event to the same widget that received the press
        if this.mousePressedOn(btn) /= null then
            this.mousePressedOn(btn).Translate_To_Content( x, y, cx, cy );
            evt := Create_Mouse_Release( cx, cy, btn );
            this.mousePressedOn(btn).On_Mouse_Release( evt );
            Delete( A_Event(evt) );
        end if;

        -- indicate that the last press and release were over the same widget.
        -- this does not mean an actual click occurred because the release
        -- could have delayed too long.
        if this.mousePressedOn(btn) = this.mouseOver then
            this.mouseClickedOn(btn) := this.mouseOver;
        else
            this.mouseClickedOn(btn) := null;
        end if;

        this.mousePressedOn(btn) := null;
    end Dispatch_Mouse_Release;

    ----------------------------------------------------------------------------

    procedure Dispatch_Mouse_Scroll( this    : not null access Window'Class;
                                     evt     : not null A_Mouse_Scroll_Event;
                                     handled : in out Boolean ) is
        evtX    : constant Integer := evt.Get_X;
        evtY    : constant Integer := evt.Get_Y;
        cx, cy  : Integer;
    begin
        if this.mouseOver /= null then
            this.mouseOver.Translate_To_Content( evtX, evtY, cx, cy );
            evt.Set_XY( cx, cy );
            this.mouseOver.Handle_Mouse_Scroll( evt, handled );
            evt.Set_XY( evtX, evtY );
        end if;
    end Dispatch_Mouse_Scroll;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw( this : access Window ) is
        bmp : constant A_Allegro_Bitmap := Get_Target_Bitmap;

        ------------------------------------------------------------------------

        procedure Draw_Child( pos : Widget_Lists.Cursor ) is
            child        : constant A_Widget := Element( pos );
            clipX, clipY,
            clipW, clipH : Integer;
            trans        : Allegro_Transform;
        begin
            if child.Is_Visible and then not this.popups.Contains( child ) then
                Al_Get_Clipping_Rectangle( clipX, clipY, clipW, clipH );
                Al_Set_Clipping_Rectangle( clipX + Integer'Max( 0, child.x1 - this.vx1 ),
                                           clipY + Integer'Max( 0, child.y1 - this.vy1 ),
                                           child.Get_Width, child.Get_Height );
                Push_State;
                Al_Copy_Transform( trans, Al_Get_Current_Transform.all );
                Al_Translate_Transform( trans,
                                        Float(child.x1) - Float(child.vx1) * child.zoom,
                                        Float(child.y1) - Float(child.vy1) * child.zoom );
                Al_Use_Transform( trans );
                child.Draw;
                Pop_State;
                Al_Set_Clipping_Rectangle( clipX, clipY, clipW, clipH );
            end if;
        end Draw_Child;

        ------------------------------------------------------------------------

        -- TO DO: replace this with Draw_Child instead
        -- the only reason this is done is so that popup widgets are drawn after
        -- the menu bar. they are not drawn in the first set of Draw_Child calls
        -- but in the second this.popups.Iterate, after the menu bar is drawn.
        procedure Draw_Popup( pos : Widget_Lists.Cursor ) is
            popup        : constant A_Widget := Element( pos );
            clipX, clipY,
            clipW, clipH : Integer;
            trans        : Allegro_Transform;
        begin
            Al_Get_Clipping_Rectangle( clipX, clipY, clipW, clipH );
            Al_Set_Clipping_Rectangle( clipX + Integer'Max( 0, popup.x1 - this.vx1 ),
                                       clipY + Integer'Max( 0, popup.y1 - this.vy1 ),
                                       popup.Get_Width, popup.Get_Height );

            Push_State;
            Al_Copy_Transform( trans, Al_Get_Current_Transform.all );
            Al_Translate_Transform( trans,
                                    Float(popup.x1) - Float(popup.vx1) * popup.zoom,
                                    Float(popup.y1) - Float(popup.vy1) * popup.zoom );
            Al_Use_Transform( trans );
            popup.Draw;
            Pop_State;
            Al_Set_Clipping_Rectangle( clipX, clipY, clipW, clipH );
        end Draw_Popup;

        ------------------------------------------------------------------------

        trans        : Allegro_Transform;
        scaled       : Boolean := False;
        clipX, clipY,
        clipW, clipH : Integer;
    begin
        if this.visible then
            if this.scale > 1 then
                -- no transformation needed; scaleBmp is the same size at the
                -- window's content.
                Al_Set_Target_Bitmap( this.scaleBmp );
            else
                -- drawing to bmp, the drawing must be translated by
                -- (win.x1, win.y1) in case the display is larger than the
                -- scaled window content.
                Al_Identity_Transform( trans );
                Al_Translate_Transform( trans,
                                        Float(this.x1 - this.vx1),
                                        Float(this.y1 - this.vy1) );
                Al_Use_Transform( trans );
                Al_Get_Clipping_Rectangle( clipX, clipY, clipW, clipH );
                Al_Set_Clipping_Rectangle( clipX + Integer'Max( 0, this.x1 - this.vx1 ),
                                           clipY + Integer'Max( 0, this.y1 - this.vy1 ),
                                           this.Get_Width, this.Get_Height );
            end if;

            -- draw all the window's children
            this.children.Iterate( Draw_Child'Access );

            -- draw the menubar widget at the top
            if this.menu /= null then
                Push_State;
                A_Widget(this.menu).Draw;
                Pop_State;
            end if;

            -- draw the popup widgets
            this.popups.Iterate( Draw_Popup'Access );

            -- stretch the window widget's content (xRes * yRes) the actual OS
            -- window's resolution.
            if this.scale > 1 then
                -- xBR scaling 2x, 3x, 4x
                if this.filter = Filter_xBR then
                    if this.scale = 2 then
                        xBR.xBR2( this.scaleBmp, bmp, this.x1, this.y1 );
                        scaled := True;
                    elsif this.scale = 3 then
                        xBR.xBR3( this.scaleBmp, bmp, this.x1, this.y1 );
                        scaled := True;
                    elsif this.scale = 4 then
                        xBR.xBR4( this.scaleBmp, bmp, this.x1, this.y1 );
                        scaled := True;
                    end if;
                end if;

                -- Allegro linear scaling
                if not scaled then
                    Set_Target_Bitmap( bmp );
                    Draw_Bitmap_Region_Stretched( this.scaleBmp,
                                                  0, 0,
                                                  this.Get_Width, this.Get_Height,
                                                  this.x1, this.y1,
                                                  this.Get_Width * this.scale,
                                                  this.Get_Height * this.scale );
                end if;
            else
                -- reset the clipping rectangle and transformation
                Al_Set_Clipping_Rectangle( clipX, clipY, clipW, clipH );
                Al_Identity_Transform( trans );
                Al_Use_Transform( trans );
            end if;
        end if;
    end Draw;

    ----------------------------------------------------------------------------

    overriding
    procedure Find_Widget_At( this   : access Window;
                              sx, sy : Integer;
                              wx, wy : out Integer;
                              found  : out A_Widget ) is
        winX : constant Integer := (sx - this.x1) / this.scale;
        winY : constant Integer := (sy - this.y1) / this.scale;
    begin
        found := null;
        if this.menu /= null then
            A_Widget(this.menu).Find_Widget_At( winX + A_Widget(this.menu).vx1,
                                                winY + A_Widget(this.menu).vy1,
                                                wx, wy, found );
        end if;
        if found = null then
            Container(this.all).Find_Widget_At( winX, winY, wx, wy, found );
        end if;
        if found /= null and then not found.Is_Descendant_Of( this.modal ) then
            -- if we have a modal widget and the widget we found is not a
            -- descendant of it, then forget that we found anything.
            found := null;
        end if;
    end Find_Widget_At;

    ----------------------------------------------------------------------------

    function Get_Outer_Border( this : not null access Window'Class ) return Allegro_Color is
    begin
        return this.outerBorder;
    end Get_Outer_Border;

    ----------------------------------------------------------------------------

    function Get_Scale( this : not null access Window'Class ) return Positive is
    begin
        return this.scale;
    end Get_Scale;

    ----------------------------------------------------------------------------

    overriding
    function Get_Window( this : access Window ) return access Window'Class is
    begin
        return A_Window(this);
    end Get_Window;

    ----------------------------------------------------------------------------

    overriding
    procedure Give_Focus( this : access Window; target : not null A_Widget ) is
        tryFocus : A_Widget := null;
        release  : A_Key_Event;
        tmp      : Boolean := False;
    begin
        if target /= this.focus then
            tryFocus := target;
            loop
                -- check if tryFocus accepts focus. if it can't accept and it
                -- doesn't have a parent (because it's the window) then it will
                -- get focus anyway.
                if tryFocus.Is_Focusable or else tryFocus.parent = null then
                    -- check if the new focus widget is actually different
                    if this.focus /= tryFocus then
                        -- find any keys pressed on the old focus widget and
                        -- clear their hooks so that future held and release
                        -- events of those keys will not be sent to any widget
                        -- without another key press event.
                        for i in this.keyPressSentTo'Range loop
                            if this.keyPressSentTo(i) = this.focus then
                                -- generate a Key_Release event for the widget
                                -- that is losing focus.
                                tmp := False;
                                release := Generate_Key_Release( i, (others=>False) );
                                 pragma Debug( Dbg( "Dispatching generated key release " &
                                                    A_Object(release).To_String & " to " &
                                                    this.focus.To_String,
                                                    D_GUI, INFO ) );
                                this.Dispatch_Key_Release( release, tmp );
                                Delete( A_Event(release) );
                            end if;
                        end loop;

                        -- blur the old widget and focus the new widget
                        this.focus.Set_Focused( False );
                        this.focus.On_Blur;
                        this.focus := tryFocus;
                        pragma Debug( Dbg( "Focused " & this.focus.To_String, D_GUI, INFO ) );
                        this.focus.Set_Focused( True );
                        this.focus.On_Focus;
                    end if;
                    exit;
                end if;

                -- This widget doesn't accept focus, try to give it to its "next"
                -- (if it has one that will accept focus), otherwise try giving
                -- focus to its parent. if we run out of parents, the window
                -- will have to take focus.
                if Length( tryFocus.nextId ) > 0 then
                    declare
                        next : A_Widget;
                    begin
                        next := this.Get_View.Get_Widget( To_String( tryFocus.nextId ) );
                        if next.Is_Focusable then
                            tryFocus := next;
                        else
                            -- 'next' widget doesn't accept focus
                            tryFocus := tryFocus.parent;
                        end if;
                    exception
                        when others =>
                            -- 'next' widget doesn't exist
                            tryFocus := tryFocus.parent;
                    end;
                else
                    -- no 'next' widget specified
                    tryFocus := tryFocus.parent;
                end if;
            end loop;
        end if;
    end Give_Focus;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Window;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
        id      : constant Event_Id := evt.Get_Id;
        handled : Boolean := False;
    begin
        if id = KEY_PRESS_ID then
            this.Dispatch_Key_Press( A_Key_Event(evt), handled );

        elsif id = KEY_RELEASE_ID then
            this.Dispatch_Key_Release( A_Key_Event(evt), handled );

        elsif id = KEY_TYPED_ID then
            this.Dispatch_Key_Typed( A_Key_Typed_Event(evt), handled );

        elsif id = MOUSE_HELD_ID then
            this.Dispatch_Mouse_Held( A_Mouse_Button_Event(evt) );
            handled := True;

        elsif id = MOUSE_MOVE_ID then
            this.Dispatch_Mouse_Move( A_Mouse_Event(evt) );
            handled := True;

        elsif id = MOUSE_PRESS_ID then
            this.Dispatch_Mouse_Press( A_Mouse_Button_Event(evt) );
            handled := True;

        elsif id = MOUSE_RELEASE_ID then
            this.Dispatch_Mouse_Release( A_Mouse_Button_Event(evt) );
            handled := True;

        elsif id = MOUSE_CLICK_ID then
            this.Dispatch_Click( A_Mouse_Button_Event(evt) );
            handled := True;

        elsif id = MOUSE_SCROLL_ID then
            this.Dispatch_Mouse_Scroll( A_Mouse_Scroll_Event(evt), handled );
        end if;

        if handled then
            -- consume the input event. there should be any other listeners.
            Delete( evt );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    overriding
    function Is_Rooted( this : access Window ) return Boolean is
        pragma Unreferenced( this );
    begin
        return True;
    end Is_Rooted;

    ----------------------------------------------------------------------------

    overriding
    function Is_Showing( this : access Window ) return Boolean is
    begin
        return this.visible and then this.Get_View.Get_Window = A_Window(this);
    end Is_Showing;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Descendant_Shown( this       : access Window;
                                   descendant : not null A_Widget;
                                   shown      : Boolean ) is
    begin
        if not shown then
            if this.mouseOver /= null and then
               this.mouseOver.Is_Descendant_Of( descendant )
            then
                -- mouse was over the widget that was hidden, or a descendant
                -- of the widget that was hidden
                this.mouseOver.On_Exit;
                this.mouseOver := null;
            end if;
            for i in this.mousePressedOn'Range loop
                if this.mousePressedOn(i) /= null and then
                   this.mousePressedOn(i).Is_Descendant_Of( descendant )
                then
                    -- a mouse button was pressed on the widget that was hidden, or
                    -- on a descendant of the widget that was hidden. fake a mouse
                    -- button release event. handling the event will cause
                    -- this.mousePressedOn(i) to be set to null.
                    this.Dispatch_Mouse_Release( -1, -1, i );
                end if;
            end loop;
            for i in this.mouseClickedOn'Range loop
                if this.mouseClickedOn(i) /= null and then
                   this.mouseClickedOn(i).Is_Descendant_Of( descendant )
                then
                    -- a mouse button was clicked on the widget that was hidden, or
                    -- on a descendant of the widget that was hidden. we don't do
                    -- anything special here because widgets don't have any special
                    -- state for this case.
                    this.mouseClickedOn(i) := null;
                end if;
            end loop;

            if this.modal = descendant then
                this.Set_Modal( null );
            elsif this.focus /= null and then
                  this.focus.Is_Descendant_Of( descendant )
            then
                -- the widget that was hidden had focus, or a descendant of it had
                -- focus, so we give focus to the parent of the widget that was
                -- hidden because all widgets between the focused and the descendant
                -- (if any) are being hidden as well.
                this.Give_Focus( descendant.parent );
            end if;
        end if;
    end On_Descendant_Shown;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Key_Press( this    : access Window;
                            evt     : not null A_Key_Event;
                            handled : in out Boolean ) is
    begin
        if evt.Get_Key = ALLEGRO_KEY_ESCAPE then
            -- if nothing else has handled the ESC key, default behavior of the
            -- window is to pop the top popup widget.
            if not this.popups.Is_Empty then
                this.Pop_Popup;
                handled := True;
            end if;
        end if;
    end On_Key_Press;

    ----------------------------------------------------------------------------

    overriding
    procedure Pack( this : access Window ) is
    begin
        Container(this.all).Pack;
        if this.menu /= null then
            A_Widget(this.menu).Pack;
        end if;
    end Pack;

    ----------------------------------------------------------------------------

    not overriding
    procedure Pop_Popup( this : access Window ) is
        popup : A_Widget;
    begin
        if not this.popups.Is_Empty then
            -- hide then delete the top popup
            popup := this.popups.Last_Element;
            this.popups.Delete_Last;
            popup.Set_Visible( False );
            this.Remove_Child( popup );
            -- todo: decide what to do with the popup
            --Delete( A_Object(popup) );

            -- give focus to the next popup, the modal widget, or the window
            if not this.popups.Is_Empty then
                this.Give_Focus( this.popups.Last_Element );
            elsif this.modal /= null then
                this.Give_Focus( this.modal );
            else
                this.Give_Focus( A_Widget(this) );
            end if;
        end if;
    end Pop_Popup;

    ----------------------------------------------------------------------------

    not overriding
    procedure Pop_Popup( this : access Window; popup : not null A_Widget ) is
    begin
        if this.popups.Contains( popup ) then
            loop
                if this.popups.Last_Element = popup then
                    this.Pop_Popup;
                    exit;
                end if;
                this.Pop_Popup;
            end loop;
        end if;
    end Pop_Popup;

    ----------------------------------------------------------------------------

    not overriding
    procedure Push_Popup( this : access Window; popup : not null A_Widget ) is
        popup2 : A_Widget := popup;
    begin
        popup.Set_Visible( True );
        this.popups.Append( popup );
        this.Add_Widget( popup2 );

        this.Give_Focus( popup );
    end Push_Popup;

    ----------------------------------------------------------------------------

    not overriding
    procedure Remove_Widget( this : access Window; child : not null A_Widget ) is
    begin
        this.Remove_Child( child );
    end Remove_Widget;

    ----------------------------------------------------------------------------

    procedure Set_Filter( this : not null access Window'Class; filter : Filter_Type ) is
        bmp   : A_Allegro_Bitmap;
        state : Allegro_State;
    begin
        if this.filter /= filter then
            this.filter := filter;

            -- recreate the scaling bitmap to set or clear the linear filtering
            -- flags on it.
            if this.scale >  1 then
                Al_Store_State( state, ALLEGRO_STATE_ALL );
                if this.filter = Filter_Linear then
                    Al_Set_New_Bitmap_Flags( ALLEGRO_MIN_LINEAR or
                                             ALLEGRO_MAG_LINEAR );
                end if;
                bmp := Al_Create_Bitmap( Al_Get_Bitmap_Width( this.scaleBmp ),
                                         Al_Get_Bitmap_Height( this.scaleBmp ) );
                Al_Restore_State( state );
                Al_Destroy_Bitmap( this.scaleBmp );
                this.scaleBmp := bmp;
            end if;
        end if;
    end Set_Filter;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Menubar( this : access Window; menu : in out A_Menubar ) is
    begin
        Delete( A_Limited_Object(this.menu) );
        if menu.all in Menubar'Class then
            this.menu := menu;
            A_Widget(this.menu).Set_Parent( A_Widget(this) );
        end if;
        menu := null;
    end Set_Menubar;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Modal( this : access Window; modal : A_Widget ) is
    begin
        if this.modal /= A_Widget(this) and then this.modal.Is_Visible then
            -- if we already have a modal widget and it's visible then hide it
            -- before setting the new modal.
            this.modal.Set_Visible( False );
        end if;

        -- set the new widget modal, make it visible and give it focus
        if modal /= null then
            this.modal := modal;
            this.modal.Bring_To_Front;
            this.modal.Set_Visible( True );
        else
            this.modal := A_Widget(this);
        end if;
        this.Give_Focus( this.modal );

        -- clear the whole mouse state
        if this.mouseOver /= null then
            this.mouseOver.On_Exit;
            this.mouseOver := null;
        end if;
        for i in Mouse_Widget_Array'Range loop
            this.mousePressedOn(i) := null;
            this.mouseClickedOn(i) := null;
        end loop;
    end Set_Modal;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Offset( this : access Window; x, y : Integer ) is
    begin
        this.x2 := x + (this.x2 - this.x1);
        this.y2 := y + (this.y2 - this.y1);
        this.x1 := x;
        this.y1 := y;
    end Set_Offset;

    ----------------------------------------------------------------------------

    procedure Set_Outer_Border( this  : not null access Window'Class;
                                color : Allegro_Color ) is
    begin
        this.outerBorder := color;
    end Set_Outer_Border;

    ----------------------------------------------------------------------------

    procedure Set_Title( this : not null access Window'Class; title : String ) is
    begin
        Al_Set_Window_Title( this.Get_View.Get_Display, title );
    end Set_Title;

    ----------------------------------------------------------------------------

    overriding
    procedure Translate_To_Content( this   : access Window;
                                    sx, sy : Integer;
                                    cx, cy : out Integer ) is
    begin
        cx := Integer((sx - this.x1) / this.scale);
        cy := Integer((sy - this.y1) / this.scale);
    end Translate_To_Content;

    ----------------------------------------------------------------------------

    overriding
    procedure Translate_To_Window( this   : access Window;
                                   cx, cy : Integer;
                                   wx, wy : out Integer ) is
        pragma Unreferenced( this );
    begin
        -- no further translation necessary
        wx := cx;
        wy := cy;
    end Translate_To_Window;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Window ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Widgets.Containers.Windows;
