--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Allegro.Drawing;                   use Allegro.Drawing;
with Allegro.Transformations;           use Allegro.Transformations;
with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Engine_Debugging;                  use Engine_Debugging;
--with Scaling;                           use Scaling;
with Support;                           use Support;
with Themes.Palette;                    use Themes.Palette;
--with Tanskanen;
with Widgets.Layouts;                   use Widgets.Layouts;

package body Widgets.Containers is

    use Widget_Lists;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Container ) is
    begin
        this.Delete_Children;
        Delete( this.childLayout );
        Widget(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add_Child( this    : access Container;
                         child   : in out A_Widget;
                         consume : Boolean := True ) is
    begin
        child.Set_Parent( A_Widget(this) );
        this.children.Append( child );
        if consume then
            child := null;
        end if;
    end Add_Child;

    ----------------------------------------------------------------------------

    not overriding
    procedure Apply_Container_Layout( this  : access Container;
                                      child : not null A_Widget ) is
    begin
        if this.childLayout /= null then
            this.childLayout.Apply( child );
        end if;
    end Apply_Container_Layout;

    ----------------------------------------------------------------------------

    not overriding
    procedure Bring_To_Front( this  : access Container;
                              child : not null A_Widget ) is
        pos : Widget_Lists.Cursor := this.children.Find( child );
    begin
        if Has_Element( pos ) then
            this.children.Delete( pos );
            this.children.Append( child );
        else
            raise WIDGET_NOT_FOUND with "Widget is not a child";
        end if;
    end Bring_To_Front;

    ----------------------------------------------------------------------------

    not overriding
    procedure Delete_Child( this  : access Container;
                            child : in out A_Widget ) is
        pos : constant Widget_Lists.Cursor := this.children.Find( child );
    begin
        if Has_Element( pos ) then
            child.Set_Visible( False );
            A_Container(this).Remove_Child( child );
            Delete( child );
        else
            raise WIDGET_NOT_FOUND with "Widget is not a child";
        end if;
    end Delete_Child;

    ----------------------------------------------------------------------------

    not overriding
    procedure Delete_Children( this : access Container ) is

        procedure Examine( pos : Widget_Lists.Cursor ) is
            child : A_Widget := Widget_Lists.Element( pos );
        begin
            child.Set_Visible( False );
            Delete( A_Limited_Object(child) );
        end Examine;

    begin
        this.children.Iterate( Examine'Access );
        this.children.Clear;
    end Delete_Children;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw( this : access Container ) is
        doScaling : constant Boolean := this.zoom /= 1.0 and this.buffer /= null;
        bmp       : constant A_Allegro_Bitmap := Get_Target_Bitmap;
        trans     : Allegro_Transform;
    begin
        if this.visible then
            if doScaling then
                -- setup 'this.buffer' as a temporary drawing buffer
                Set_Target_Bitmap( this.buffer );
                Al_Identity_Transform( trans );

                -- reapply the transformation relative to the widget's viewport,
                -- so that drawing at 0,0 maps to -this.vx1, -this.vy1 in
                -- 'this.buffer', which maps to 0,0 in 'bmp'.
                Al_Translate_Transform( trans, Float(-this.vx1), Float(-this.vy1) );

                Al_Use_Transform( trans );
                Al_Clear_To_Color( base_colors(Transparent) );
            end if;

            Push_State;
            A_Container(this).Draw_Content;
            Pop_State;
            Push_State;
            A_Container(this).Draw_Children;
            Pop_State;
            Push_State;
            A_Container(this).Draw_Content_Foreground;
            Pop_State;

            if doScaling then
                -- scale 'this.buffer' to 'bmp'
                --
                -- 'this.buffer' is only the size of the viewport, whereas 'bmp'
                -- represents the widget's entire content area. the bitmap of
                -- the viewport we drew must be pasted into the widget's content
                -- at the location of the viewport (vx1*zoom, vy1*zoom).
                Set_Target_Bitmap( bmp );
                Draw_Bitmap_Region_Stretched( this.buffer,
                                              0, 0,
                                              this.vx2 - this.vx1 + 1,
                                              this.vy2 - this.vy1 + 1,
                                              Integer(Float(this.vx1) * this.zoom),
                                              Integer(Float(this.vy1) * this.zoom),
                                              Integer(Float(this.vx2 - this.vx1 + 1) * this.zoom),
                                              Integer(Float(this.vy2 - this.vy1 + 1) * this.zoom) );
            end if;
        end if;
    end Draw;

    ----------------------------------------------------------------------------

    not overriding
    procedure Draw_Children( this : access Container ) is

        procedure Draw_Child( pos : Widget_Lists.Cursor ) is
            child        : constant A_Widget := Element( pos );
            clipX, clipY,
            clipW, clipH : Integer;
            trans        : Allegro_Transform;
        begin
            if Intersect( this.vx1, this.vy1, this.vx2, this.vy2,
                          child.x1, child.y1, child.x2, child.y2 )
            then
                Al_Get_Clipping_Rectangle( clipX, clipY, clipW, clipH );
                Al_Set_Clipping_Rectangle( clipX + Integer'Max( 0, child.x1 - this.vx1 ),
                                           clipY + Integer'Max( 0, child.y1 - this.vy1 ),
                                           child.Get_Width, child.Get_Height );
                Push_State;
                Al_Copy_Transform( trans, Al_Get_Current_Transform.all );
                Al_Translate_Transform( trans,
                                        Float(child.x1) - Float(child.vx1) * child.zoom,
                                        Float(child.y1) - Float(child.vy1) * child.zoom );
                Al_Use_Transform( trans );
                child.Draw;
                Pop_State;
                Al_Set_Clipping_Rectangle( clipX, clipY, clipW, clipH );
            end if;
        end Draw_Child;

    begin
        this.children.Iterate( Draw_Child'Access );
    end Draw_Children;

    ----------------------------------------------------------------------------

    overriding
    procedure Find_Widget_At( this   : access Container;
                              x, y   : Integer;
                              wx, wy : out Integer;
                              found  : out A_Widget ) is
        -- translate viewport coordinates into content coordinates
        cx : constant Integer := this.vx1 + Integer(x / this.zoom) ;
        cy : constant Integer := this.vy1 + Integer(y / this.zoom);

        ------------------------------------------------------------------------

        procedure Examine( pos : Widget_Lists.Cursor ) is
            child : constant A_Widget := Element( pos );
        begin
            if found = null then
                if cx >= child.x1 and then cx <= child.x2 and then
                   cy >= child.y1 and then cy <= child.y2
                then
                    child.Find_Widget_At( cx - child.x1, cy - child.y1, wx, wy, found );
                end if;
            end if;
        end Examine;

        ------------------------------------------------------------------------

    begin
        found := null;

        if this.visible and then this.enabled then
            -- iterate in reverse to check in Z-order front to back
            this.children.Reverse_Iterate( Examine'Access );
        end if;

        if found = null then
            Widget(this.all).Find_Widget_At( x, y, wx, wy, found );
        end if;
    end Find_Widget_At;

    ----------------------------------------------------------------------------

    not overriding
    procedure Give_Focus( this : access Container; target : not null A_Widget ) is
    begin
        if this.parent /= null then
            A_Container(this.parent).Give_Focus( target );
        else
            Dbg( "Error: Unable to focus " & target.To_String & " - not rooted",
                 D_GUI, ERROR );
        end if;
    end Give_Focus;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Rooted( this : access Container; rooted : Boolean ) is

        procedure Examine( pos : Widget_Lists.Cursor ) is
            child : constant A_Widget := Element( pos );
        begin
            child.Handle_Rooted( rooted );
        end Examine;

    begin
        this.children.Iterate( Examine'Access );
        Widget(this.all).Handle_Rooted( rooted );
    end Handle_Rooted;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Shown( this : access Container; shown : Boolean ) is

        procedure Examine( pos : Widget_Lists.Cursor ) is
            child : constant A_Widget := Element( pos );
        begin
            if child.Is_Visible then
                child.Handle_Shown( shown );
            else
                -- if the child isn't visible then it's not affected by showing
                -- notifications; skip it.
                null;
            end if;
        end Examine;

    begin
        this.children.Iterate( Examine'Access );
        Widget(this.all).Handle_Shown( shown );
    end Handle_Shown;

    ----------------------------------------------------------------------------

    overriding
    procedure Pack( this : access Container ) is

        procedure Examine( pos : Widget_Lists.Cursor ) is
        begin
            Element( pos ).Pack;
        end Examine;

    begin
        if this.Is_Rooted then
            Widget(this.all).Pack;
            this.children.Iterate( Examine'Access );
        end if;
    end Pack;

    ----------------------------------------------------------------------------

    not overriding
    procedure Remove_Child( this  : access Container;
                            child : not null A_Widget ) is
        pos : Widget_Lists.Cursor := this.children.Find( child );
    begin
        if Has_Element( pos ) then
            this.children.Delete( pos );
            child.Set_Parent( null );
        else
            raise WIDGET_NOT_FOUND with "Widget is not a child";
        end if;
    end Remove_Child;

end Widgets.Containers;
