--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Allegro.Keyboard;                  use Allegro.Keyboard;
--with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Fonts;                             use Fonts;
with Themes.Palette;                    use Themes.Palette;

package body Widgets.Input_Boxes is

    PADDING : constant := 2;

    function Unconstrained( curstr, newstr : String ) return String;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Input_Box( view : not null access Game_Views.Game_View'Class;
                               id   : String ) return A_Input_Box is
        this : A_Input_Box := new Input_Box;
    begin
        this.Construct( view, id );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Input_Box;

    ----------------------------------------------------------------------------

    procedure Add_Character( this : not null access Input_Box'Class;
                             char : Character ) is
        len   : constant Natural := Length( this.text );
        left  : constant String := Slice( this.text, 1, this.cursor );
        right : constant String := Slice( this.text, this.cursor + 1, len );
    begin
        if len < this.maxlen then
            this.text := To_Unbounded_String( this.constrain( To_String( this.text ),
                                                              left & char & right ) );
        end if;
        if len < Length( this.text ) then
            -- a character was indeed added
            this.cursor := this.cursor + 1;
            this.Adjust_Visible_Text;
        end if;
    end Add_Character;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add_Listener( this     : access Input_Box;
                            listener : not null A_Input_Listener ) is
    begin
        this.Add_Listener( "Input_Listener", A_Action_Listener(listener) );
    end Add_Listener;

    ----------------------------------------------------------------------------

    procedure Adjust_Visible_Text( this : not null access Input_Box'Class ) is
        theme : constant A_Theme := Themes.Get;
        bw    : constant Natural := theme.Border_Width;
        len   : Natural := 0;
    begin
        -- if we don't have a widget size yet (not yet attached to the window)
        -- then don't compute visible because: 1. It's not necessary, and
        -- 2. the visible character range will be scrunched to nothing.
        if this.cwidth = 0 then
            return;
        end if;

        if this.cursor < this.firstchar then
            this.firstchar := Integer'Max( 0, this.cursor - 5 );
        end if;

        this.lastchar := this.firstchar;
        for i in this.firstchar+1..Length( this.text ) loop
            len := theme.Get_Text_Width( Slice( this.text, this.firstchar+1, i ), To_String( this.fontName ), this.fontSize );
            if len > this.cwidth - (bw + PADDING) * 2 then
                exit;
            end if;
            this.lastchar := i;
        end loop;

        if this.cursor > this.lastchar then
            this.lastchar := Integer'Min( Length( this.text ), this.cursor + 5 );
            this.firstchar := this.lastchar;
            for i in reverse 1..this.lastchar loop
                len := theme.Get_Text_Width( Slice( this.text, i, this.lastchar ), To_String( this.fontName ), this.fontSize );
                if len > this.cwidth - (bw + PADDING) * 2 then
                    exit;
                end if;
                this.firstchar := i - 1;
            end loop;
        end if;
    end Adjust_Visible_Text;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Input_Box;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
    begin
        Widget(this.all).Construct( view, id );
        this.border := Border_Sunk;
        this.colors(Background) := base_colors(White);
        this.constrain := Unconstrained'Access;
        this.cursor := Length( this.text );
        this.Adjust_Visible_Text;
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    procedure Dispatch_Action( this : access Input_Box; id : Action_Id ) is
        action : A_Input_Action := new Input_Action;

        procedure Dispatch( listener : A_Action_Listener ) is
        begin
            A_Input_Listener(listener).Handle_Action( action );
        end Dispatch;

    begin
        action.Construct( id, A_Widget(this) );
        this.Iterate_Listeners( "Input_Listener", Dispatch'Access );
        Delete( A_Object(action) );
    end Dispatch_Action;

    ----------------------------------------------------------------------------

    procedure Do_Backspace( this : not null access Input_Box'Class ) is
        len : constant Natural := Length( this.text );
    begin
        if this.cursor > 0 then
            this.text := To_Unbounded_String( this.constrain( To_String( this.text ),
                                                              Slice( this.text, 1, this.cursor - 1 ) &
                                                              Slice( this.text,
                                                                     this.cursor + 1,
                                                                     len ) ) );
            if len > Length( this.text ) then
                -- a character was indeed removed
                this.cursor := this.cursor - 1;
                this.Adjust_Visible_Text;
            end if;
        end if;
    end Do_Backspace;

    ----------------------------------------------------------------------------

    procedure Do_Delete( this : not null access Input_Box'Class ) is
        len : constant Natural := Length( this.text );
    begin
        if this.cursor < len then
            this.text := To_Unbounded_String( this.constrain( To_String( this.text ),
                                                              Slice( this.text, 1, this.cursor ) &
                                                              Slice( this.text, this.cursor + 2, len ) ) );
            if len > Length( this.text ) then
                -- a character was indeed removed
                this.Adjust_Visible_Text;
            end if;
        end if;
    end Do_Delete;

    ----------------------------------------------------------------------------

    procedure Do_Enter( this : not null access Input_Box'Class ) is
    begin
        this.Dispatch_Action( Entered );
    end Do_Enter;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Input_Box ) is
        theme : constant A_Theme := Themes.Get;
        bw    : constant Natural := theme.Border_Width;
        font  : constant A_Font := theme.Get_Font( To_String( this.fontName ), this.fontSize );
    begin
        theme.Draw_Box( 0, 0, this.cwidth - 1, this.cheight - 1,
                        this.colors(Background), this.border );
        if this.focused then
            Rect( 0, 0, this.cwidth - 1, this.cheight - 1, this.colors(Selected) );
            Rect( 1, 1, this.cwidth - 2, this.cheight - 2, this.colors(Selected) );
        elsif this.hover then
            theme.Draw_Border( Border_Line,
                               0, 0,
                               this.cwidth - 1, this.cheight - 1,
                               Al_Map_RGB( 0, 0, 0 ) );
        end if;
        if this.lastchar - this.firstchar > 0 then
            Draw_String( Slice( this.text, 1 + this.firstchar, this.lastchar ),
                     bw + PADDING, this.cheight / 2 - font.Text_Height / 2,
                     font, Hard_Contrast( this.colors(Background) ) );
        end if;
        if this.focused then
            declare
                textH  : constant Natural := font.Text_Height;
                height : constant Natural := Natural(Float(textH) * 0.8);
                cx     : constant Natural := bw + PADDING + font.Text_Length( Slice( this.text, 1 + this.firstchar, this.cursor ) );
                cy     : constant Natural := this.cheight / 2 - textH / 2 + 1;
            begin
                Line_V( cx, cy, cy + height,
                        Hard_Contrast( this.colors(Background) ) );
            end;
        end if;
    end Draw_Content;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Text( this : access Input_Box ) return String is
    begin
        return To_String( this.text );
    end Get_Text;

    ----------------------------------------------------------------------------

    procedure Move_Cursor( this : not null access Input_Box'Class; dir : Move_Dir ) is
    begin
        case dir is
            when Go_First =>
                if this.cursor > 0 then
                    this.cursor := 0;
                    if this.cursor < this.firstchar then
                        this.Adjust_Visible_Text;
                    end if;
                end if;
            when Go_Left =>
                if this.cursor > 0 then
                    this.cursor := this.cursor - 1;
                    if this.cursor < this.firstchar then
                        this.Adjust_Visible_Text;
                    end if;
                end if;
            when Go_Right =>
                if this.cursor < Length( this.text ) then
                    this.cursor := this.cursor + 1;
                    if this.cursor > this.lastchar then
                        this.Adjust_Visible_Text;
                    end if;
                end if;
            when Go_Last =>
                if this.cursor < Length( this.text ) then
                    this.cursor := Length( this.text );
                    if this.cursor > this.lastchar then
                        this.Adjust_Visible_Text;
                    end if;
                end if;
        end case;
    end Move_Cursor;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Key_Typed( this    : access Input_Box;
                            evt     : not null A_Key_Typed_Event;
                            handled : in out Boolean ) is
        mods : Modifiers_Array;
    begin
        evt.Get_Modifiers( mods );
        if mods(ALT) or else mods(CTRL) then
            -- don't handle control keys / shortcuts
            return;
        end if;

        handled := True;
        case evt.Get_Key is
            when ALLEGRO_KEY_RIGHT =>
                this.Move_Cursor( Go_Right );

            when ALLEGRO_KEY_LEFT =>
                this.Move_Cursor( Go_Left );

            when ALLEGRO_KEY_HOME =>
                this.Move_Cursor( Go_First );

            when ALLEGRO_KEY_END =>
                this.Move_Cursor( Go_Last );

            when ALLEGRO_KEY_BACKSPACE =>
                this.Do_Backspace;

            when ALLEGRO_KEY_DELETE =>
                this.Do_Delete;

            when ALLEGRO_KEY_ENTER | ALLEGRO_KEY_PAD_ENTER =>
                this.Do_Enter;

            when others =>
                if evt.Get_Char /= ASCII.NUL then
                    this.Add_Character(evt.Get_Char);
                else
                    handled := False;
                end if;

        end case;
    end On_Key_Typed;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Press( this : access Input_Box;
                              evt  : not null A_Mouse_Button_Event ) is
        theme : constant A_Theme := Themes.Get;
        bw    : constant Natural := theme.Border_Width;
        x     : constant Integer := evt.Get_X;
    begin
       if x >= (bw + PADDING) then
            this.cursor := this.lastchar;
            for i in this.firstchar+1..this.lastchar loop
                if (bw + PADDING) + theme.Get_Text_Width( Slice( this.text, 1 + this.firstchar, i ), To_String( this.fontName ), this.fontSize ) > x then
                    this.cursor := i - 1;
                    exit;
                end if;
            end loop;
        end if;
    end On_Mouse_Press;

    ----------------------------------------------------------------------------

    overriding
    procedure Pack( this : access Input_Box ) is
    begin
        if this.Is_Rooted then
            Widget(this.all).Pack;
            this.Adjust_Visible_Text;
        end if;
    end Pack;

    ----------------------------------------------------------------------------

    not overriding
    procedure Remove_Listener( this     : access Input_Box;
                               listener : not null A_Input_Listener ) is
    begin
        A_Widget(this).Remove_Listener( "Input_Listener", A_Action_Listener(listener) );
    end Remove_Listener;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Constraint( this       : access Input_Box;
                              constraint : A_Constrain_Func ) is
        len : constant Natural := Length( this.text );
    begin
        this.constrain := constraint;
        if this.constrain = null then
            this.constrain := Unconstrained'Access;
        end if;
        this.text := To_Unbounded_String( this.constrain( "", To_String( this.text ) ) );
        if this.cursor > len then
            this.cursor := len;
        end if;
        this.Adjust_Visible_Text;
    end Set_Constraint;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Max_Length( this : access Input_Box; maxlen : Positive ) is
    begin
        this.maxlen := maxlen;
        if Length( this.text ) > this.maxlen then
            this.text := Unbounded_Slice( this.text, 1, this.maxlen );
            if this.cursor > Length( this.text ) then
                this.cursor := Length( this.text );
            end if;
            this.Adjust_Visible_Text;
        end if;
    end Set_Max_Length;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Text( this : access Input_Box; text : String ) is
    begin
        this.text := To_Unbounded_String(
                this.constrain( To_String( this.text ),
                                text(text'First..Integer'Min( text'Last, text'First+this.maxlen-1 )) ) );
        this.cursor := Length( this.text );
        this.Adjust_Visible_Text;
    end Set_Text;

    --==========================================================================

    function Unconstrained( curstr, newstr : String ) return String is
        pragma Unreferenced( curstr );
    begin
        return newstr;
    end Unconstrained;

end Widgets.Input_Boxes;
