--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Real_Time;                     use Ada.Real_Time;

private with Values.Lists;

package Widgets.Labels is

    -- A Label widget is a simple text string with an optional icon to the left
    -- of the text. A label is intended for displaying static strings and isn't
    -- editable by the interface user.
    type Label is new Widget and Animated with private;
    type A_Label is access all Label'Class;

    -- Creates a new Label within 'view' with id 'id'. Both 'text' and 'icon'
    -- are optional. 'icon' is the name of a frame in the Theme's tile library.
    -- The icon image can have an alpha channel for transparency.
    function Create_Label( view : not null access Game_Views.Game_View'Class;
                           id   : String;
                           text : String := "";
                           icon : String := "" ) return A_Label;
    pragma Precondition( id'Length > 0 );
    pragma Postcondition( Create_Label'Result /= null );

    -- Returns the label's text.
    function Get_Text( this : not null access Label'Class ) return String;

    -- Sets the label's text alignment. Both the text and icon will be aligned
    -- according to 'align'.
    procedure Set_Align( this : not null access Label'Class; align : Align_Type );

    -- Sets the color for a particular purpose when drawing the label. To remove
    -- the color behind the icon text, set the Foreground color to transparent.
    procedure Set_Color( this    : access Label;
                         purpose : Color_Purpose;
                         color   : Allegro_Color );

    -- Sets the label's icon. Pass an empty string to remove it.
    procedure Set_Icon( this : not null access Label'Class; icon : String );

    -- Sets the label's text.
    procedure Set_Text( this : not null access Label'Class; text : String );

private

    use Values.Lists;

    type Label is new Widget and Animated with
        record
            align          : Align_Type := Align_Left;
            text           : Unbounded_String;
            icon           : Natural := 0;     -- the icon that was set

            -- for icon animation --
            iconStart      : Time_Span;        -- start time of the animation
            iconFrame      : Natural := 0;     -- current image being displayed
            iconFrameDelay : Time_Span;        -- delay between frames
            iconFrameLoop  : List_Ptr;         -- list of frames in loop
        end record;

    procedure Construct( this : access Label;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String;
                         text : String;
                         icon : String );

    procedure Draw_Content( this : access Label );

    -- Returns the label's minimum height.
    function Get_Min_Height( this : access Label ) return Natural;

    -- Returns the label's minimum width.
    function Get_Min_Width( this : access Label ) return Natural;

    -- Updates the label's icon, if it's animated.
    procedure Tick( this : access Label; time : Tick_Time );

end Widgets.Labels;
