--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

--with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Mouse;                             use Mouse;
with Support;                           use Support;
with Themes.Palette;                    use Themes.Palette;

package body Widgets.Scrollbars is

    overriding
    procedure Construct( this : access Scrollbar;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
    begin
        Widget(this.all).Construct( view, id );
        this.acceptFocus := False;
        this.border := Border_Raised;
        this.colors(Background) := Lighten( this.colors(Foreground), 0.5 );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Scrollbar ) is
    begin
        if this.client /= null then
            this.client.Remove_Listener( A_Resize_Listener(this) );
            this.client := null;
        end if;
        Widget(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Action( this : access Scrollbar; action : A_Resize_Action ) is
        pragma Unreferenced( action );
    begin
        null;
    end Handle_Action;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Release( this : access Scrollbar;
                                evt  : not null A_Mouse_Button_Event ) is
    begin
        if evt.Get_Button = Mouse_Left then
            this.pageUp := False;
            this.pageDown := False;
            if this.dragging then
                this.dragging := False;
            end if;
        end if;
    end On_Mouse_Release;

    ----------------------------------------------------------------------------

    procedure Set_Client( this : not null access Scrollbar'Class; client : A_Widget ) is
    begin
        if this.client /= null then
            this.client.Remove_Listener( A_Resize_Listener(this) );
        end if;
        this.client := client;
        if this.client /= null then
            this.client.Add_Listener( A_Resize_Listener(this) );
        end if;
    end Set_Client;

    --==========================================================================

    function Create_H_Scrollbar( view : not null access Game_Views.Game_View'Class;
                                 id   : String ) return A_H_Scrollbar is
        this : A_H_Scrollbar := new H_Scrollbar;
    begin
        this.Construct( view, id );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_H_Scrollbar;

    ----------------------------------------------------------------------------

    overriding
    procedure Calculate_Button( this : access H_Scrollbar; min, max : out Integer ) is
        scale : Float;
    begin
        if this.client /= null then
            scale := Float(this.Get_Width) / Float(this.client.Get_Content_Width);
            min := Integer(this.client.vx1 * scale);
            max := Integer(this.client.vx2 * scale);
        else
            min := 0;
            max := this.cwidth - 1;
        end if;
    end Calculate_Button;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access H_Scrollbar ) is
        theme    : constant A_Theme := Themes.Get;
        border   : Border_Type := this.border;
        btncolor : Allegro_Color := this.colors(Foreground);
        left,
        right    : Integer;
    begin
        Clear_To_Color( this.colors(Background) );
        if this.client /= null then
            A_H_Scrollbar(this).Calculate_Button( left, right );
            if this.dragging then
                btncolor := this.colors(Background);
                if this.border = Border_Raised then
                    border := Border_Sunk;
                end if;
            end if;
            theme.Draw_Box( left, 0, right, this.cheight - 1, btncolor, border );
        end if;
    end Draw_Content;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Held( this : access H_Scrollbar;
                             evt  : not null A_Mouse_Button_Event ) is
        pragma Unreferenced( evt );
    begin
        if this.pageUp then
            this.client.Scroll( -this.client.Get_Width, 0 );
        elsif this.pageDown then
            this.client.Scroll( this.client.Get_Width, 0 );
        end if;
    end On_Mouse_Held;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Move( this : access H_Scrollbar;
                             evt  : not null A_Mouse_Event ) is
        scale : Float;
        left,
        right : Integer;
    begin
        if this.dragging and then this.client /= null then
            scale := Float(this.client.Get_Content_Width) / Float(this.Get_Width);
            A_H_Scrollbar(this).Calculate_Button( left, right );
            this.client.Scroll( Integer((evt.Get_X - (left + this.dragPos)) * scale), 0 );
        end if;
    end On_Mouse_Move;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Press( this : access H_Scrollbar;
                              evt  : not null A_Mouse_Button_Event ) is
        left,
        right : Integer;
    begin
        if evt.Get_Button = Mouse_Left and then this.client /= null then
            A_H_Scrollbar(this).Calculate_Button( left, right );
            if evt.Get_X < left then
                this.pageUp := True;
                this.client.Scroll( -this.client.Get_Width, 0 );
            elsif evt.Get_X > right then
                this.pageDown := True;
                this.client.Scroll( this.client.Get_Width, 0 );
            else
                this.dragging := True;
                this.dragPos := evt.Get_X - left;
            end if;
        end if;
    end On_Mouse_Press;

    --==========================================================================

    function Create_V_Scrollbar( view : not null access Game_Views.Game_View'Class;
                                 id   : String ) return A_V_Scrollbar is
        this : A_V_Scrollbar := new V_Scrollbar;
    begin
        this.Construct( view, id );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_V_Scrollbar;

    ----------------------------------------------------------------------------

    overriding
    procedure Calculate_Button( this : access V_Scrollbar; min, max : out Integer ) is
    begin
        if this.client /= null then
            min := Integer(Float(this.cheight - 1) * Float(this.client.vy1) / Float(this.client.cheight - 1));
            max := Integer(Float(this.cheight - 1) * Float(this.client.vy2) / Float(this.client.cheight - 1));
        else
            min := 0;
            max := this.cheight - 1;
        end if;
    end Calculate_Button;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access V_Scrollbar ) is
        theme    : constant A_Theme := Themes.Get;
        border   : Border_Type := this.border;
        btncolor : Allegro_Color := this.colors(Foreground);
        top,
        bottom   : Integer;
    begin
        Clear_To_Color( this.colors(Background) );
        if this.client /= null then
            A_V_Scrollbar(this).Calculate_Button( top, bottom );
            if this.dragging then
                btncolor := Lighten( this.colors(Foreground), 0.85 );
                if this.border = Border_Raised then
                    border := Border_Sunk;
                end if;
            end if;
            theme.Draw_Box( 0, top, this.cwidth - 1, bottom, btncolor, border );
        end if;
    end Draw_Content;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Held( this : access V_Scrollbar;
                             evt  : not null A_Mouse_Button_Event ) is
        pragma Unreferenced( evt );
    begin
        if this.pageUp then
            this.client.Scroll( 0, -this.client.Get_Height );
        elsif this.pageDown then
            this.client.Scroll( 0, this.client.Get_Height );
        end if;
    end On_Mouse_Held;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Move( this : access V_Scrollbar;
                             evt  : not null A_Mouse_Event ) is
        scale : Float;
        top,
        bottom : Integer;
    begin
        if this.dragging and then this.client /= null then
            scale := Float(this.client.Get_Content_Height) / Float(this.Get_Height);
            A_V_Scrollbar(this).Calculate_Button( top, bottom );
            this.client.Scroll( 0, Integer((evt.Get_Y - (top + this.dragPos)) * scale) );
        end if;
    end On_Mouse_Move;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Press( this : access V_Scrollbar;
                              evt  : not null A_Mouse_Button_Event ) is
        top,
        bottom : Integer;
    begin
        if evt.Get_Button = Mouse_Left and then this.client /= null then
            A_V_Scrollbar(this).Calculate_Button( top, bottom );
            if evt.Get_Y < top then
                this.pageUp := True;
                this.client.Scroll( 0, -this.client.Get_Height );
            elsif evt.Get_Y > bottom then
                this.pageDown := True;
                this.client.Scroll( 0, this.client.Get_Height );
            else
                this.dragging := True;
                this.dragPos := evt.Get_Y - top;
            end if;
        end if;
    end On_Mouse_Press;

end Widgets.Scrollbars;
