--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

package Widgets.Scrollbars is

    -- An abstract scrollbar widget. A scrollbar has a client widget that it is
    -- responsible for scrolling. The amount that can be scrolled is the
    -- difference between the client's content region and its viewport.
    type Scrollbar is abstract new Widget and Resize_Listener with private;

    -- Calculates where the scroll button will be drawn on the scrollbar, based
    -- on the size of the client widget and the location and size of its
    -- viewport. 'min' and 'max' are in the range of 0..[scroll_dimension]-1, in
    -- pixels. For example, if this is a horizontal scrollbar for a client that
    -- is completely visible in the horizontal dimension, then 'min' will return
    -- 0 and 'max' will return Scrollbar.Get_Width - 1. This is because the
    -- button should entirely fill the area scrollbar.
    procedure Calculate_Button( this : access Scrollbar; min, max : out Integer ) is abstract;

    -- Sets the client widget of the scrollbar; ie: the widget that will be
    -- controlled. 'client' will not belong to the scrollbar. The client of a
    -- scrollbar should be removed before its deleted. If 'client' is null, the
    -- scrollbar will have no client.
    procedure Set_Client( this : not null access Scrollbar'Class; client : A_Widget );

    ----------------------------------------------------------------------------

    -- A horizontal scrollbar widget.
    type H_Scrollbar is new Scrollbar with private;
    type A_H_Scrollbar is access all H_Scrollbar'Class;

    -- Creates a horizontal scrollbar within 'view' with widget id 'id'.
    function Create_H_Scrollbar( view : not null access Game_Views.Game_View'Class;
                                 id   : String ) return A_H_Scrollbar;
    pragma Precondition( id'Length > 0 );
    pragma Postcondition( Create_H_Scrollbar'Result /= null );

    ----------------------------------------------------------------------------

    -- A vertical scrollbar widget.
    type V_Scrollbar is new Scrollbar with private;
    type A_V_Scrollbar is access all V_Scrollbar'Class;

    -- Creates a vertical scrollbar within 'view' with widget id 'id'.
    function Create_V_Scrollbar( view : not null access Game_Views.Game_View'Class;
                                 id   : String ) return A_V_Scrollbar;
    pragma Precondition( id'Length > 0 );
    pragma Postcondition( Create_V_Scrollbar'Result /= null );

private

    type Scrollbar is abstract new Widget and Resize_Listener with
        record
            client : A_Widget := null;      -- widget to scroll
            pageUp,                         -- true when paging up
            pageDown,                       -- true when paging down
            dragging : Boolean := False;    -- true when button is dragged
            dragPos  : Integer := 0;
        end record;

    procedure Construct( this : access Scrollbar;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String );
    pragma Precondition( id'Length > 0 );

    procedure Delete( this : in out Scrollbar );

    procedure Handle_Action( this   : access Scrollbar;
                             action : A_Resize_Action );

    procedure On_Mouse_Release( this : access Scrollbar;
                                evt  : not null A_Mouse_Button_Event );

    ----------------------------------------------------------------------------

    type H_Scrollbar is new Scrollbar with null record;

    procedure Calculate_Button( this : access H_Scrollbar; min, max : out Integer );

    procedure Draw_Content( this : access H_Scrollbar );

    procedure On_Mouse_Held( this : access H_Scrollbar;
                             evt  : not null A_Mouse_Button_Event );

    procedure On_Mouse_Move( this : access H_Scrollbar;
                             evt  : not null A_Mouse_Event );

    procedure On_Mouse_Press( this : access H_Scrollbar;
                              evt  : not null A_Mouse_Button_Event );

    ----------------------------------------------------------------------------

    type V_Scrollbar is new Scrollbar with null record;

    procedure Calculate_Button( this : access V_Scrollbar; min, max : out Integer );

    procedure Draw_Content( this : access V_Scrollbar );

    procedure On_Mouse_Held( this : access V_Scrollbar;
                             evt  : not null A_Mouse_Button_Event );

    procedure On_Mouse_Move( this : access V_Scrollbar;
                             evt  : not null A_Mouse_Event );

    procedure On_Mouse_Press( this : access V_Scrollbar;
                              evt  : not null A_Mouse_Button_Event );

end Widgets.Scrollbars;
