--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

--with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Themes.Palette;                    use Themes.Palette;
with Tiles;                             use Tiles;
with Tiles.Libraries.Manager;           use Tiles.Libraries.Manager;
with Values.Casting;                    use Values.Casting;

package body Widgets.Sprites is

    function Create_Sprite( view     : not null access Game_Views.Game_View'Class;
                            eid      : Entity_Id;
                            class    : String;
                            x, y     : Float;
                            physical : Boolean;
                            libName  : String;
                            frame    : Natural ) return A_Sprite is
        this : A_Sprite := new Sprite;
    begin
        this.Construct( view, eid, class, x, y, physical, libName, frame );
        this.visible := False;
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Sprite;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this     : access Sprite;
                         view     : not null access Game_Views.Game_View'Class;
                         eid      : Entity_Id;
                         class    : String;
                         x, y     : Float;
                         physical : Boolean;
                         libName  : String;
                         frame    : Natural ) is
    begin
        Widget(this.all).Construct( view, "sprite " & Image( eid ) );
        this.colors(Selected) := base_colors(White);
        this.colors(Foreground) := base_colors(Chrome4);
        this.colors(Background) := base_colors(Transparent);
        this.transparent := True;
        this.eid := eid;
        this.class := To_Unbounded_String( class );
        this.x := x;
        this.y := y;
        this.physical := physical;
        this.lib := Load_Library( libName, bitmaps => True );
        this.frame := frame;
        if this.lib /= null then
            declare
                tile : constant A_Tile := this.lib.Get_Tile( this.frame );
            begin
                this.offsetX := Cast_Int( tile.Get_Attribute( "offsetX" ), 0 );
                this.offsetY := Cast_Int( tile.Get_Attribute( "offsetY" ), 0 );
            end;
        end if;

        -- position the sprite within its parent
        declare
            width  : constant Integer := A_Widget(this).Get_Min_Width;
            height : constant Integer := A_Widget(this).Get_Min_Height;
        begin
            this.Set_Layout_XYWH( Integer(this.x) - width / 2 + this.offsetX,
                                  Integer(this.y) - height / 2 + this.offsetY,
                                  width, height );
        end;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Sprite ) is
    begin
        Unload_Library( this.lib );
        Widget(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Center_At( this : not null access Sprite'Class; x, y : Float ) is
    begin
        this.x := x;
        this.y := y;
        this.Move_To( Integer(this.x) - this.Get_Width / 2 + this.offsetX,
                      Integer(this.y) - this.Get_Height / 2 + this.offsetY );
    end Center_At;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Sprite ) is
        bmp : A_Allegro_Bitmap := null;
    begin
        if this.lib /= null then
            bmp := this.lib.Get_Tile( this.frame ).Get_Bitmap;
        end if;
        if bmp /= null then
            -- center the bitmap in the sprite widget's content
            Draw_Bitmap( bmp,
                         (this.cwidth - Al_Get_Bitmap_Width( bmp )) / 2,
                         (this.cheight - Al_Get_Bitmap_Height( bmp )) / 2 );
            if this.border /= Border_None then
                Rect( 0, 0, this.cwidth - 1, this.cheight - 1, this.colors(Foreground) );
            end if;
        else
            -- we don't have a bitmap; draw the background color
            Rectfill( 0, 0, this.cwidth - 1, this.cheight - 1, this.colors(Background) );
            if this.lib /= null and then this.frame /= 0 then
                -- there is a frame for the sprite but it hasn't loaded yet
                -- hopefully it will be loaded on the next redraw.
                null;
            end if;
        end if;
        if this.interactive and then this.selected then
            Rect( 0, 0, this.cwidth - 1, this.cheight - 1, this.colors(Selected) );
        end if;

        -- Display a box showing the physical edges of the sprite's entity
        --declare
        --    physicalWidth  : constant Integer := this.Get_Min_Width;
        --    physicalHeight : constant Integer := this.Get_Min_Height;
        --begin
        --    Rect( (this.cwidth - physicalWidth) / 2,
        --          (this.cheight - physicalHeight) / 2,
        --          this.cwidth - (this.cwidth - physicalWidth) / 2 - 1,
        --          this.cheight - (this.cheight - physicalHeight) / 2 - 1,
        --          base_colors(Magenta) );
        --end;
    end Draw_Content;

    ----------------------------------------------------------------------------

    function Get_Class( this : not null access Sprite'Class ) return String is
    begin
        return To_String( this.class );
    end Get_Class;

    ----------------------------------------------------------------------------

    function Get_Entity_Id( this : not null access Sprite'Class ) return Entity_Id is
    begin
        return this.eid;
    end Get_Entity_Id;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Height( this : access Sprite ) return Natural is
        bmp : A_Allegro_Bitmap;
    begin
        if this.minHeight > 0 then
            return this.minHeight;
        elsif this.lib /= null then
            bmp := this.lib.Get_Tile( this.frame ).Get_Bitmap;
            if bmp /= null then
                return Al_Get_Bitmap_Height( bmp );
            end if;
        end if;
        return 0;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Width( this : access Sprite ) return Natural is
        bmp : A_Allegro_Bitmap;
    begin
        if this.minWidth > 0 then
            return this.minWidth;
        elsif this.lib /= null then
            bmp := this.lib.Get_Tile( this.frame ).Get_Bitmap;
            if bmp /= null then
                return Al_Get_Bitmap_Width( bmp );
            end if;
        end if;
        return 0;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    function Get_X( this : not null access Sprite'Class ) return Float is
    begin
        return this.x;
    end Get_X;

    ----------------------------------------------------------------------------

    function Get_Y( this : not null access Sprite'Class ) return Float is
    begin
        return this.y;
    end Get_Y;

    ----------------------------------------------------------------------------

    function Is_Updatable( this : not null access Sprite'Class ) return Boolean is
    begin
        return this.updatable = 0;
    end Is_Updatable;

    ----------------------------------------------------------------------------

    procedure Resize( this   : not null access Sprite'Class;
                      width,
                      height : Natural ) is
    begin
        if width /= this.Get_Width or else height /= this.Get_Height then
            this.Set_Layout_LTWH( Integer(this.x) - width / 2 + this.offsetX,
                                  Integer(this.y) - height / 2 + this.offsetY,
                                  width, height );
        end if;
    end Resize;

    ----------------------------------------------------------------------------

    procedure Set_Frame( this : not null access Sprite'Class; frame : Natural ) is
        new_width,
        new_height,
        old_width,
        old_height  : Natural;
        old_offsetX,
        old_offsetY : Integer;
    begin
        if frame /= this.frame then
            old_width := A_Widget(this).Get_Min_Width;
            old_height := A_Widget(this).Get_Min_Height;
            old_offsetX := this.offsetX;
            old_offsetY := this.offsetY;

            this.frame := frame;
            this.offsetX := 0;
            this.offsetY := 0;
            if this.lib /= null then
                declare
                    tile : constant A_Tile := this.lib.Get_Tile( this.frame );
                begin
                    this.offsetX := Cast_Int( tile.Get_Attribute( "offsetX" ), 0 );
                    this.offsetY := Cast_Int( tile.Get_Attribute( "offsetY" ), 0 );
                end;
            end if;

            -- update the sprite's layout if its frame size or frame offset has
            -- changed
            new_width := A_Widget(this).Get_Min_Width;
            new_height := A_Widget(this).Get_Min_Height;
            if new_width /= old_width or else
               new_height /= old_height or else
               this.offsetX /= old_offsetX or else
               this.offsetY /= old_offsetY
            then
                this.Set_Layout_LTWH( this.x1 - (new_width  - old_width)  / 2 + (this.offsetX - old_offsetX),
                                      this.y1 - (new_height - old_height) / 2 + (this.offsetY - old_offsetY),
                                      new_width, new_height );
            end if;
        end if;
    end Set_Frame;

    ----------------------------------------------------------------------------

    procedure Set_Selected( this : access Sprite; selected : Boolean ) is
    begin
        if selected /= this.selected then
            this.selected := selected;
        end if;
    end Set_Selected;

end Widgets.Sprites;
