--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Entities;                          use Entities;

limited with Game_Views;

private with Tiles.Libraries;

package Widgets.Sprites is

    -- A Sprite is the visible representation of an Entity. It is a partially
    -- transparent widget, drawn as a tile from a tile library, that is added to
    -- a Scene container widget.
    type Sprite is new Widget with private;
    type A_Sprite is access all Sprite'Class;

    -- Creates a new Sprite. 'eid' is the id of the entity it represents. Class
    -- is the class of entity it represents. 'x', 'y' are the location of the
    -- center of the widget in pixels. 'physical' indicates if the entity obeys
    -- the laws of physics. 'libName' is the name of the tile library to be used
    -- for the sprite's images. 'frame' is the id of the current frame in the
    -- tile library to display.
    function Create_Sprite( view     : not null access Game_Views.Game_View'Class;
                            eid      : Entity_Id;
                            class    : String;
                            x, y     : Float;
                            physical : Boolean;
                            libName  : String;
                            frame    : Natural ) return A_Sprite;
    pragma Precondition( class'Length > 0 );
    pragma Precondition( libName'Length > 0 );
    pragma Postcondition( Create_Sprite'Result /= null );

    -- Centers the sprite at 'x', 'y' in pixels. The sprite will update its
    -- known location of its entity.
    procedure Center_At( this : not null access Sprite'Class; x, y : Float );

    -- Returns the sprite's entity's class.
    function Get_Class( this : not null access Sprite'Class ) return String;
    pragma Postcondition( Get_Class'Result'Length > 0 );

    -- Returns the Entity_Id of the Sprite's entity.
    function Get_Entity_Id( this : not null access Sprite'Class ) return Entity_Id;

    -- Returns the x location of the Sprite's entity; located at the center of
    -- the widget.
    function Get_X( this : not null access Sprite'Class ) return Float;

    -- Returns the y location of the Sprite's entity; located at the center of
    -- the widget.
    function Get_Y( this : not null access Sprite'Class ) return Float;

    -- Sets the tile id of the Sprite's current frame.
    procedure Set_Frame( this : not null access Sprite'Class; frame : Natural );

    function Is_Updatable( this : not null access Sprite'Class ) return Boolean;

    -- Resizes the Sprite to 'width', 'height' while keeping the widget centered
    -- at its entity's location. If 'width' or 'height' is 0, the value will not
    -- be changed. (ex: width => 10, height => 0 will change only the width.)
    procedure Resize( this   : not null access Sprite'Class;
                      width,
                      height : Natural );

    -- Notifies the sprite of its selected state. This doesn't actually
    -- select/unselect the widget; that can only be done by its Window.
    procedure Set_Selected( this : access Sprite; selected : Boolean );

private

    use Tiles.Libraries;

    type Sprite is new Widget with
        record
            eid         : Entity_Id := INVALID_ID;
            class       : Unbounded_String;
            x, y        : Float := 0.0;
            physical    : Boolean := False;
            lib         : A_Tile_Library := null;
            frame       : Natural := 0;
            offsetX,
            offsetY     : Integer := 0;       -- the center's offset from x, y
            updatable   : Integer := 0;       -- accepts updates if = 0
            interactive : Boolean := False;
            selected    : Boolean := False;
        end record;

    procedure Construct( this     : access Sprite;
                         view     : not null access Game_Views.Game_View'Class;
                         eid      : Entity_Id;
                         class    : String;
                         x, y     : Float;
                         physical : Boolean;
                         libName  : String;
                         frame    : Natural );
    pragma Precondition( class'Length > 0 );
    pragma Precondition( libName'Length > 0 );

    procedure Delete( this : in out Sprite );

    procedure Draw_Content( this : access Sprite );

    function Get_Min_Height( this : access Sprite ) return Natural;

    function Get_Min_Width( this : access Sprite ) return Natural;

end Widgets.Sprites;
