--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Unchecked_Conversion;
with Allegro.Bitmaps.Drawing;           use Allegro.Bitmaps.Drawing;
with Allegro.Keyboard;                  use Allegro.Keyboard;
with Allegro.State;                     use Allegro.State;
with Allegro.Transformations;           use Allegro.Transformations;
with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Engine_Debugging;                  use Engine_Debugging;
with Game_Views;                        use Game_Views;
with Interfaces;                        use Interfaces;
with Mouse;                             use Mouse;
--with Scaling;                           use Scaling;
with Support;                           use Support;
with Themes.Palette;                    use Themes.Palette;
with Widgets.Containers;                use Widgets.Containers;
with Widgets.Containers.Windows;
with Widgets.Layouts;                   use Widgets.Layouts;

package body Widgets is

    -- This is called when a descendent of the widget is shown or hidden on the
    -- screen. This notification traverses up the widget tree from the widget
    -- 'descendent' that was shown or hidden.
    procedure Handle_Descendant_Shown( this       : not null access Widget'Class;
                                       descendant : not null A_Widget;
                                       shown      : Boolean );

    -- Updates the .buffer field that contains the widget's zoom bitmap buffer
    -- when the size of the viewport changes. If the widget's .zoom field equals
    -- 1.0, the .buffer will be null. Otherwise, the dimensions of the bitmap
    -- will be: viewport_width x viewport_height
    procedure Update_Zoom_Buffer( this : not null access Widget'Class );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    not overriding
    procedure Activate_Popup( this   : access Widget;
                              popup  : not null A_Widget;
                              corner : Widget_Corner_Type;
                              x, y   : Integer;
                              width  : Natural := 0;
                              height : Natural := 0;
                              scale  : Boolean := False ) is
        pragma Unreferenced( scale );
        winX, winY : Integer;           -- window content coordinates for the corner
        winVX1,
        winVX2,
        winVY1,
        winVY2     : Integer;
        realCorner : Widget_Corner_Type := corner;
        realWidth  : Integer := width;  --
        realHeight : Integer := height;
    begin
        -- determine the location of the specified corner of the popup
        A_Widget(this).Translate_To_Window( x, y, winX, winY );

        -- determines the visible area of the Window the popup will go in
        this.Get_Window.Get_Viewport( winVX1, winVY1, winVX2, winVY2 );

        -- determine the size of the popup
        if width = 0 then
            realWidth := popup.Get_Min_Width;
        end if;
        if height = 0 then
            realHeight := popup.Get_Min_Height;
        end if;
        -- todo: apply size scaling if scale is true

        -- determine if the specified popup corner can be placed at x, y without
        -- the popup going off the edge of its parent.
        if realCorner = Widget_Top_Left or else realCorner = Widget_Bottom_Left then
            -- check for widget going off the right side
            if winX + realWidth - 1 >= winVX2 then
                if realCorner = Widget_Top_Left then
                    realCorner := Widget_Top_Right;
                else
                    realCorner := Widget_Bottom_Right;
                end if;
            end if;
        else
            -- check for widget going off the left side
            if winX - realWidth < winVX1 then
                if realCorner = Widget_Top_Right then
                    realCorner := Widget_Top_Left;
                else
                    realCorner := Widget_Bottom_Left;
                end if;
            end if;
        end if;
        if realCorner = Widget_Top_Left or else realCorner = Widget_Top_Right then
            -- check for widget going off the bottom side
            if winY + realHeight - 1 >= winVY2 then
                if realCorner = Widget_Top_Left then
                    realCorner := Widget_Bottom_Left;
                else
                    realCorner := Widget_Bottom_Right;
                end if;
            end if;
        else
            -- check for widget going off the top side
            if winY - realHeight < winVY1 then
                if realCorner = Widget_Bottom_Left then
                    realCorner := Widget_Top_Left;
                else
                    realCorner := Widget_Top_Right;
                end if;
            end if;
        end if;

        -- apply the calculated layout to the popup
        case realCorner is
            when Widget_Top_Left =>
                popup.Set_Layout_LTWH( winX, winY, realWidth, realHeight );
            when Widget_Top_Right =>
                popup.Set_Layout_LTWH( winX - realWidth, winY, realWidth, realHeight );
            when Widget_Bottom_Left =>
                popup.Set_Layout_LTWH( winX, winY - realHeight, realWidth, realHeight );
            when Widget_Bottom_Right =>
                popup.Set_Layout_LTWH( winX - realWidth, winY - realHeight, realWidth, realHeight );
        end case;

        -- push the popup widget to the window
        this.Get_Window.Push_Popup( popup );
    end Activate_Popup;

    ----------------------------------------------------------------------------

    procedure Add_Listener( this     : not null access Widget'Class;
                            listener : not null A_Key_Listener ) is
    begin
        this.Add_Listener( "Key_Listener", A_Action_Listener(listener) );
    end Add_Listener;

    ----------------------------------------------------------------------------

    procedure Add_Listener( this     : not null access Widget'Class;
                            listener : not null A_Resize_Listener ) is
    begin
        this.Add_Listener( "Resize_Listener", A_Action_Listener(listener) );
    end Add_Listener;

    ----------------------------------------------------------------------------

    procedure Add_Listener( this     : not null access Widget'Class;
                            listener : not null A_Visibility_Listener ) is
    begin
        this.Add_Listener( "Visibility_Listener", A_Action_Listener(listener) );
    end Add_Listener;

    ----------------------------------------------------------------------------

    procedure Add_Listener( this         : not null access Widget'Class;
                            listenerType : String;
                            listener     : not null A_Action_Listener ) is

        --------------------------------------------------------------------

        procedure Update( key     : String;
                          element : in out Action_Listeners.List ) is
            pragma Unreferenced( key );
        begin
            -- Verify the listener hasn't already been added for this type
            if not element.Contains( listener ) then
                element.Append( listener );
            end if;
        end Update;

        --------------------------------------------------------------------

        crsr : constant Listener_Maps.Cursor := this.listeners.Find( listenerType );
    begin
        if Listener_Maps.Has_Element( crsr ) then
            this.listeners.Update_Element( crsr, Update'Access );
        else
            -- Insert a new listener list of the given type with the listener
            -- as the first element.
            declare
                list : Action_Listeners.List;
            begin
                list.Append( listener );
                this.listeners.Insert( listenerType, list );
            end;
        end if;
    end Add_Listener;

    ----------------------------------------------------------------------------

    not overriding
    procedure Bring_To_Front( this : access Widget ) is
    begin
        if this.parent /= null then
            A_Container(this.parent).Bring_To_Front( A_Widget(this) );
        end if;
    end Bring_To_Front;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Widget;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is

        function To_GV2 is new Ada.Unchecked_Conversion( A_Game_View, A_Game_View2 );

        theme : constant A_Theme := Themes.Get;
    begin
        Limited_Object(this.all).Construct;
        this.view := To_GV2(A_Game_View(view));
        this.id := To_Unbounded_String( id );
        this.attrs := Create_Assoc;
        this.layout := Create_Layout_LTRB( 0, 0, 0, 0 );
        for c in this.colors'Range loop
            this.colors(c) := theme.Color( c );
        end loop;
        this.visible := True;
        view.Register( A_Widget(this) );   -- raises exception on duplicate id
        this.registered := True;
    end Construct;

    ----------------------------------------------------------------------------

    procedure Deactivate_Popup( this : not null access Widget'Class; popup : not null A_Widget ) is
    begin
        this.Get_Window.Pop_Popup( popup );
    end Deactivate_Popup;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Widget ) is
    begin
        Al_Destroy_Bitmap( this.buffer );

        -- delete the simple key listeners because this widget owns them. other
        -- listener classes are not owned by this widget and should not be
        -- deleted.
        declare
            use Listener_Maps;

            --------------------------------------------------------------------

            procedure Delete_Listeners( key  : String;
                                        elem : in out Action_Listeners.List ) is
                use Action_Listeners;
                pragma Unreferenced( key );

                ----------------------------------------------------------------

                procedure Delete_Listener( elem : in out A_Action_Listener ) is
                begin
                    -- if the listener extends Simple_Action_Listener then
                    -- delete it. this will leave a null in the list but that's
                    -- ok because we will clear the list after iterating.
                    if elem.all in Simple_Action_Listener'Class then
                        Delete( A_Limited_Object(elem) );
                    end if;
                end Delete_Listener;

                ----------------------------------------------------------------

                crsr : Action_Listeners.Cursor := elem.First;
            begin
                -- loop through each listener in the list
                while Has_Element( crsr ) loop
                    if Element( crsr ) /= null then
                        elem.Update_Element( crsr, Delete_Listener'Access );
                    end if;
                    crsr := Next( crsr );
                end loop;
                elem.Clear;
            end Delete_Listeners;

            --------------------------------------------------------------------

            pos : Listener_Maps.Cursor := this.listeners.First;
        begin
            -- loop through all listener types
            while Has_Element( pos ) loop
                this.listeners.Update_Element( pos, Delete_Listeners'Access );
                pos := Next( pos );
            end loop;
        end;

        if this.registered then
            this.Get_View.Unregister( this.Get_Id );
        end if;
        Delete( this.layout );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Dispatch_Key_Action( this      : not null access Widget'Class;
                                  id        : Action_Id;
                                  key       : Integer;
                                  modifiers : Modifiers_Array ) return Boolean is
        action  : A_Key_Action;
        handled : Boolean := False;

        ------------------------------------------------------------------------

        procedure Examine( listener : A_Action_Listener ) is
            hndld : Boolean;
        begin
            if listener.all in Key_Listener'Class then
                A_Key_Listener(listener).Handle_Action( action, hndld );
                if hndld then
                    handled := True;
                end if;
            end if;
        end Examine;

        ------------------------------------------------------------------------

    begin
        action := new Key_Action;
        action.Construct( id, A_Widget(this), key, modifiers );
        this.Iterate_Listeners( "Key_Listener", Examine'Access );
        Delete( A_Object(action) );
        return handled;
    end Dispatch_Key_Action;

    ----------------------------------------------------------------------------

    procedure Dispatch_Resize_Action( this : not null access Widget'Class ) is
        action  : A_Resize_Action;

        ------------------------------------------------------------------------

        procedure Examine( listener : A_Action_Listener ) is
        begin
            if listener.all in Resize_Listener'Class then
                A_Resize_Listener(listener).Handle_Action( action );
            end if;
        end Examine;

        ------------------------------------------------------------------------

    begin
        action := new Resize_Action;
        action.Construct( Widget_Resized, A_Widget(this) );
        this.Iterate_Listeners( "Resize_Listener", Examine'Access );
        Delete( A_Object( action ) );
    end Dispatch_Resize_Action;

    ----------------------------------------------------------------------------

    procedure Dispatch_Visibility_Action( this    : not null access Widget'Class;
                                          visible : Boolean ) is
        action  : A_Visibility_Action;

        procedure Examine( listener : A_Action_Listener ) is
        begin
            if listener.all in Visibility_Listener'Class then
                A_Visibility_Listener(listener).Handle_Action( action );
            end if;
        end Examine;

    begin
        action := new Visibility_Action;
        action.Construct( Change_Visibility, A_Widget(this), visible );
        this.Iterate_Listeners( "Visibility_Listener", Examine'Access );
        Delete( A_Object(action) );
    end Dispatch_Visibility_Action;

    ----------------------------------------------------------------------------

    not overriding
    procedure Draw( this : access Widget ) is
        doScaling : constant Boolean := this.zoom /= 1.0 and this.buffer /= null;
        bmp       : constant A_Allegro_Bitmap := Get_Target_Bitmap;
        trans     : Allegro_Transform;
    begin
        if this.visible then
            if doScaling then
                -- setup 'this.buffer' as a temporary drawing buffer
                Set_Target_Bitmap( this.buffer );
                Al_Identity_Transform( trans );

                -- reapply the transformation relative to the widget's viewport,
                -- so that drawing at 0,0 maps to -this.vx1, -this.vy1 in
                -- 'this.buffer', which maps to 0,0 in the target bitmap.
                Al_Translate_Transform( trans, Float(-this.vx1), Float(-this.vy1) );

                Al_Use_Transform( trans );
                Clear_To_Color( base_colors(Transparent) );
            end if;

            Push_State;
            A_Widget(this).Draw_Content;
            Pop_State;

            if doScaling then
                -- scale 'this.buffer' to 'bmp'
                --
                -- 'this.buffer' is only the size of the viewport, whereas 'bmp'
                -- represents the widget's entire content area. the bitmap of
                -- the viewport we drew must be pasted into the widget's content
                -- at the location of the viewport (vx1*zoom, vy1*zoom).
                Set_Target_Bitmap( bmp );
                Al_Draw_Scaled_Bitmap( this.buffer,
                                       0.0, 0.0,
                                       Float(this.vx2 - this.vx1 + 1),
                                       Float(this.vy2 - this.vy1 + 1),
                                       Float(this.vx1) * this.zoom,
                                       Float(this.vy1) * this.zoom,
                                       Float(this.vx2 - this.vx1 + 1) * this.zoom,
                                       Float(this.vy2 - this.vy1 + 1) * this.zoom,
                                       0 );
            end if;
        end if;
    end Draw;

    ----------------------------------------------------------------------------

    not overriding
    procedure Find_Widget_At( this   : access Widget;
                              x, y   : Integer;
                              wx, wy : out Integer;
                              found  : out A_Widget ) is
    begin
        -- verify x, y is within the viewport
        if this.visible and then this.enabled and then
           x >= 0 and then x <= this.x2 - this.x1 and then
           y >= 0 and then y <= this.y2 - this.y1
        then
            -- translate the viewport coordinates into content coordinates
            wx := this.vx1 + Integer(x / this.zoom);
            wy := this.vy1 + Integer(y / this.zoom);
            found := A_Widget(this);
        else
           wx := 0;
           wy := 0;
           found := null;
        end if;
    end Find_Widget_At;

    ----------------------------------------------------------------------------

    function Focus_Prev( this : not null access Widget'Class ) return Boolean is
        wdgt : A_Widget := null;
    begin
        wdgt := this.Get_View.Get_Widget( To_String( this.prevId ) );
        while wdgt /= A_Widget(this) loop
            -- loop through the list of prev widgets until one of
            -- them accepts focus or we get back to this widget.
            if wdgt.Is_Focusable then
                wdgt.Request_Focus;
                return True;
            end if;
            -- raises ID_NOT_FOUND if no prev widget
            wdgt := this.Get_View.Get_Widget( To_String( wdgt.prevId ) );
        end loop;
        -- the prev widget is this widget. return that it has been focused.
        return True;
    exception
        when ID_NOT_FOUND =>
            return False;
    end Focus_Prev;

    ----------------------------------------------------------------------------

    function Focus_Next( this : not null access Widget'Class ) return Boolean is
        wdgt : A_Widget := null;
    begin
        wdgt := this.Get_View.Get_Widget( To_String( this.nextId ) );
        while wdgt /= A_Widget(this) loop
            -- loop through the list of next widgets until one of
            -- them accepts focus or we get back to this widget.
            if wdgt.Is_Focusable then
                wdgt.Request_Focus;
                return True;
            end if;
            -- raises ID_NOT_FOUND if no next widget
            wdgt := this.Get_View.Get_Widget( To_String( wdgt.nextId ) );
        end loop;
        -- the next widget is this widget. return that it has been focused.
        return True;
    exception
        when ID_NOT_FOUND =>
            return False;
    end Focus_Next;

    ----------------------------------------------------------------------------

    function Get_Attribute( this : not null access Widget'Class;
                            name : String ) return Value_Ptr is
    begin
        return this.attrs.Get.Get( name );
    end Get_Attribute;

    ----------------------------------------------------------------------------

    function Get_Border( this : not null access Widget'Class ) return Border_Type is
    begin
        return this.border;
    end Get_Border;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Color( this    : access Widget;
                        purpose : Color_Purpose ) return Allegro_Color is
    begin
        return this.colors(purpose);
    end Get_Color;

    ----------------------------------------------------------------------------

    function Get_Content_Height( this : not null access Widget'Class ) return Natural is
    begin
        return this.cheight;
    end Get_Content_Height;

    ----------------------------------------------------------------------------

    function Get_Content_Width( this : not null access Widget'Class ) return Natural is
    begin
        return this.cwidth;
    end Get_Content_Width;

    ----------------------------------------------------------------------------

    function Get_Font_Name( this : not null access Widget'Class ) return String is
    begin
        return To_String( this.fontName );
    end Get_Font_Name;

    ----------------------------------------------------------------------------

    function Get_Font_Size( this : not null access Widget'Class ) return Positive is
    begin
        return this.fontSize;
    end Get_Font_Size;

    ----------------------------------------------------------------------------

    function Get_Height( this : not null access Widget'Class ) return Integer is
    begin
        return this.y2 - this.y1 + 1;
    end Get_Height;

    ----------------------------------------------------------------------------

    function Get_Id( this : not null access Widget'Class ) return String is
    begin
        return To_String( this.id );
    end Get_Id;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Min_Height( this : access Widget ) return Natural is
    begin
        return this.minHeight;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Min_Width( this : access Widget ) return Natural is
    begin
        return this.minWidth;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Process_Name( this : access Widget ) return String is
    begin
        return A_Widget(this).To_String;
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    not overriding
    function Get_View( this : access Widget ) return access Game_Views.Game_View'Class is
        function To_GV is new Ada.Unchecked_Conversion( A_Game_View2, A_Game_View );
    begin
        return To_GV(this.view);
    end Get_View;

    ----------------------------------------------------------------------------

    procedure Get_Viewport( this   : not null access Widget'Class;
                            x1, y1,
                            x2, y2 : out Integer ) is
    begin
        x1 := this.x1;
        y1 := this.y1;
        x2 := this.x2;
        y2 := this.y2;
    end Get_Viewport;

    ----------------------------------------------------------------------------

    function Get_Width( this : not null access Widget'Class ) return Natural is
    begin
        return this.x2 - this.x1 + 1;
    end Get_Width;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Window( this : access Widget
                       ) return access Widgets.Containers.Windows.Window'Class is
    begin
        if this.parent /= null then
            return this.parent.Get_Window;
        end if;
        return null;
    end Get_Window;

    ----------------------------------------------------------------------------

    function Get_Zoom( this : not null access Widget'Class ) return Float is
    begin
        return this.zoom;
    end Get_Zoom;

    ----------------------------------------------------------------------------

    procedure Handle_Descendant_Shown( this       : not null access Widget'Class;
                                       descendant : not null A_Widget;
                                       shown      : Boolean ) is
    begin
        A_Widget(this).On_Descendant_Shown( descendant, shown );
        if this.parent /= null then
            this.parent.Handle_Descendant_Shown( descendant, shown );
        end if;
    end Handle_Descendant_Shown;

    ----------------------------------------------------------------------------

    procedure Handle_Key_Press( this    : not null access Widget'Class;
                                evt     : not null A_Key_Event;
                                handled : in out Boolean ) is
        modifiers : Modifiers_Array;
    begin
        evt.Get_Modifiers( modifiers );

        -- Try dispatching the key to an action handler
        handled := this.Dispatch_Key_Action( Key_Press, evt.Get_Key, modifiers );
        if handled then
            return;
        end if;

        -- Try dispatching the event to the current widget
        A_Widget(this).On_Key_Press( evt, handled );
        if handled then
            return;
        end if;

        -- Check if the widget uses the key internally for focus changes
        if evt.Get_Key = ALLEGRO_KEY_TAB and then
           not modifiers(ALT) and then
           not modifiers(CTRL)
        then
            if modifiers(SHIFT) then
                -- shift-tab to previous widget
                if this.Focus_Prev then
                    handled := True;
                end if;
            else
                -- tab to next widget
                if this.Focus_Next then
                    handled := True;
                end if;
            end if;
        end if;

        -- Check if the parent can handle the event
        if this.parent /= null then
            this.parent.Handle_Key_Press( evt, handled );
        end if;
    end Handle_Key_Press;

    ----------------------------------------------------------------------------

    procedure Handle_Key_Release( this    : not null access Widget'Class;
                                  evt     : not null A_Key_Event;
                                  handled : in out Boolean ) is
        modifiers : Modifiers_Array;
    begin
        evt.Get_Modifiers( modifiers );

        handled := this.Dispatch_Key_Action( Key_Release, evt.Get_Key, modifiers );
        if handled then
            return;
        end if;

        -- Try dispatching the event to the current widget
        A_Widget(this).On_Key_Release( evt, handled );
        if handled then
            return;
        end if;

        if this.parent /= null then
            this.parent.Handle_Key_Release( evt, handled );
        end if;
    end Handle_Key_Release;

    ----------------------------------------------------------------------------

    procedure Handle_Key_Typed( this    : not null access Widget'Class;
                                evt     : not null A_Key_Typed_Event;
                                handled : in out Boolean ) is
        modifiers : Modifiers_Array;
    begin
        evt.Get_Modifiers( modifiers );

        -- Try dispatching the key to an action handler
        handled := this.Dispatch_Key_Action( Key_Typed, evt.Get_Key, modifiers );
        if handled then
            return;
        end if;

        -- Try dispatching the event to the current widget
        A_Widget(this).On_Key_Typed( evt, handled );
        if handled then
            return;
        end if;

        -- Check if the parent can handle the event
        if this.parent /= null then
            this.parent.Handle_Key_Typed( evt, handled );
        end if;
    end Handle_Key_Typed;

    ----------------------------------------------------------------------------

    procedure Handle_Mouse_Scroll( this    : not null access Widget'Class;
                                   evt     : not null A_Mouse_Scroll_Event;
                                   handled : in out Boolean ) is
        evtX : constant Integer := evt.Get_X;
        evtY : constant Integer := evt.Get_Y;
    begin
        A_Widget(this).On_Mouse_Scroll( evtX, evtY, evt.Get_Amount, handled );
        if not handled then
            if this.parent /= null then
                -- translate the x, y of the event into the parent's content coordinates
                evt.Set_XY( Integer(Float(evtX) * this.zoom) + this.x1,
                            Integer(Float(evtY) * this.zoom) + this.y1 );
                -- give the widget an opportunity to handle the event
                this.parent.Handle_Mouse_Scroll( evt, handled );
                -- restore the given x, y
                evt.Set_XY( evtX, evtY );
            end if;
        end if;
    end Handle_Mouse_Scroll;

    ----------------------------------------------------------------------------

    procedure Handle_Resize( this : not null access Widget'Class ) is
    begin
        -- update the .buffer field to match the new viewport dimensions
        this.Update_Zoom_Buffer;

        -- allow the widget subclass to handle the event
        A_Widget(this).On_Resize;
    end Handle_Resize;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Rooted( this : access Widget; rooted : Boolean ) is
    begin
        A_Widget(this).On_Rooted( rooted );
    end Handle_Rooted;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Shown( this : access Widget; shown : Boolean ) is
    begin
        A_Widget(this).Handle_Descendant_Shown( A_Widget(this), shown );
        if A_Widget(this).all in Animated'Class then
            if shown then
                this.Get_View.Attach( A_Process(A_Widget(this)) );
            else
                this.Get_View.Detach( A_Process(A_Widget(this)) );
            end if;
        end if;

        A_Widget(this).On_Shown( shown );

        -- (before the vis action was dispatched) using Handle_[Un]Hidden
        this.Dispatch_Visibility_Action( visible => shown );
    end Handle_Shown;

    ----------------------------------------------------------------------------

    function Is_Descendant_Of( this     : not null access Widget'Class;
                               ancestor : A_Widget ) return Boolean is
    begin
        if ancestor = null then
            return False;
        elsif ancestor = A_Widget(this) or else ancestor = this.parent then
            return True;
        elsif this.parent = null then
            return False;
        else
            return this.parent.Is_Descendant_Of( ancestor );
        end if;
    end Is_Descendant_Of;

    ----------------------------------------------------------------------------

    function Is_Enabled( this : not null access Widget'Class ) return Boolean is
    begin
        return this.enabled and then (this.parent = null or else this.parent.Is_Enabled);
    end Is_Enabled;

    ----------------------------------------------------------------------------

    function Is_Focusable( this : not null access Widget'Class ) return Boolean is
    begin
        return this.acceptFocus and then this.Is_Enabled and then this.Is_Showing;
    end Is_Focusable;

    ----------------------------------------------------------------------------

    not overriding
    function Is_Rooted( this : access Widget ) return Boolean is
    begin
        if this.parent /= null then
            return this.parent.Is_Rooted;
        end if;
        return False;
    end Is_Rooted;

    ----------------------------------------------------------------------------

    not overriding
    function Is_Showing( this : access Widget ) return Boolean is
    begin
        if this.parent /= null then
            return this.Is_Visible and then this.parent.Is_Showing;
        end if;
        return False;
    end Is_Showing;

    ----------------------------------------------------------------------------

    function Is_Visible( this : not null access Widget'Class ) return Boolean is
    begin
        return this.visible;
    end Is_Visible;

    ----------------------------------------------------------------------------

    procedure Iterate_Listeners( this         : not null access Widget'Class;
                                 listenerType : String;
                                 examine      : not null access procedure( listener : A_Action_Listener ) ) is
        use Action_Listeners;
        use Listener_Maps;

        procedure Local_Examine( pos : Action_Listeners.Cursor ) is
            elem : constant A_Action_Listener := Element( pos );
        begin
            if elem /= null then
                examine.all( elem );
            end if;
        end Local_Examine;

        crsr : constant Listener_Maps.Cursor := this.listeners.Find( listenerType );
    begin
        if Has_Element( crsr ) then
            Element( crsr ).Iterate( Local_Examine'Access );
        end if;
    end Iterate_Listeners;

    ----------------------------------------------------------------------------

    procedure Move( this  : not null access Widget'Class;
                    xdist,
                    ydist : Integer ) is
    begin
        if this.parent /= null and then (xdist /= 0 or else ydist /= 0) then
            this.Set_Layout_XYWH( this.x1 + xdist, this.y1 + ydist, this.Get_Width, this.Get_Height );
        end if;
    end Move;

    ----------------------------------------------------------------------------

    procedure Move_To( this : not null access Widget'Class; x, y : Integer ) is
    begin
        if this.parent /= null then
            this.Set_Layout_XYWH( x, y, this.Get_Width, this.Get_Height );
        end if;
    end Move_To;

    ----------------------------------------------------------------------------

    not overriding
    procedure On_Enter( this : access Widget ) is
    begin
        this.hover := True;
    end On_Enter;

    ----------------------------------------------------------------------------

    not overriding
    procedure On_Exit( this : access Widget ) is
    begin
        this.hover := False;
    end On_Exit;

    ----------------------------------------------------------------------------

    not overriding
    procedure On_Mouse_Press( this : access Widget;
                              evt  : not null A_Mouse_Button_Event ) is
    begin
        if evt.Get_Button = Mouse_Left then
            this.pressed := True;
        end if;
    end On_Mouse_Press;

    ----------------------------------------------------------------------------

    not overriding
    procedure On_Mouse_Release( this : access Widget;
                                evt  : not null A_Mouse_Button_Event ) is
    begin
        if evt.Get_Button = Mouse_Left then
            this.pressed := False;
        end if;
    end On_Mouse_Release;

    ----------------------------------------------------------------------------

    not overriding
    procedure Pack( this : access Widget ) is
        oldCW : constant Integer := this.cwidth;
        oldCH : constant Integer := this.cheight;
        oldVW : constant Integer := this.vx2 - this.vx1 + 1;
        oldVH : constant Integer := this.vy2 - this.vy1 + 1;
        newW,
        newH  : Integer;
    begin
        if this.Is_Rooted then
            this.x1 := 0;
            this.y1 := 0;
            this.x2 := this.parent.cwidth - 1;
            this.y2 := this.parent.cheight - 1;

            A_Container(this.parent).Apply_Container_Layout( A_Widget(this) );
            this.layout.Apply( A_Widget(this) );

            newW := this.x2 - this.x1 + 1;
            newH := this.y2 - this.y1 + 1;
            this.cwidth := Integer'Max( A_Widget(this).Get_Min_Width, newW );
            this.cheight := Integer'Max( A_Widget(this).Get_Min_Height, newH );

            -- the following lines are commented out because the vx2, vy2
            -- corner of the widget view was being clamped to less than the
            -- vx1, vy1 corner if vx1, vy1 was off the bottom/right edge of the
            -- widget content at pack time. this happens when we change the
            -- contents of a scene widget and the viewport was in the lower right
            -- corner of a much bigger content area than the new map.
--            this.vx2 := Integer'Min( this.vx1 + Floor(newW / this.zoom) - 1, this.cwidth - 1 );
--            this.vy2 := Integer'Min( this.vy1 + Floor(newH / this.zoom) - 1, this.cheight - 1 );
            this.vx2 := this.vx1 + Floor( newW / this.zoom ) - 1;
            this.vy2 := this.vy1 + Floor( newH / this.zoom ) - 1;

            if oldVW /= this.vx2 - this.vx1 + 1 or else
               oldVH /= this.vy2 - this.vy1 + 1
            then
                A_Widget(this).Handle_Resize;
            end if;

            if oldCW /= this.cwidth or else oldCH /= this.cheight then
                this.Dispatch_Resize_Action;
            end if;
        end if;
    end Pack;

    ----------------------------------------------------------------------------

    procedure Prepend_Listener( this         : not null access Widget'Class;
                                listenerType : String;
                                listener     : not null A_Action_Listener ) is

        --------------------------------------------------------------------

        procedure Update( key     : String;
                          element : in out Action_Listeners.List ) is
            use Action_Listeners;
            pragma Unreferenced( key );
            crsr : Action_Listeners.Cursor := element.Find( listener );
        begin
            -- Prepend the listener to the list. If it's already in the list,
            -- move it to the front.
            if Has_Element( crsr ) then
                element.Delete( crsr );
            end if;
            element.Prepend( listener );
        end Update;

        --------------------------------------------------------------------

        crsr : constant Listener_Maps.Cursor := this.listeners.Find( listenerType );
    begin
        if Listener_Maps.Has_Element( crsr ) then
            this.listeners.Update_Element( crsr, Update'Access );
        else
            -- Insert a new listener list of the given type with the listener
            -- as the first element.
            declare
                list : Action_Listeners.List;
            begin
                list.Prepend( listener );
                this.listeners.Insert( listenerType, list );
            end;
        end if;
    end Prepend_Listener;

    ----------------------------------------------------------------------------

    procedure Remove_Listener( this     : not null access Widget'Class;
                               listener : not null A_Key_Listener ) is
    begin
        this.Remove_Listener( "Key_Listener", A_Action_Listener(listener) );
    end Remove_Listener;

    ----------------------------------------------------------------------------

    procedure Remove_Listener( this     : not null access Widget'Class;
                               listener : not null A_Resize_Listener ) is
    begin
        this.Remove_Listener( "Resize_Listener", A_Action_Listener(listener) );
    end Remove_Listener;

    ----------------------------------------------------------------------------

    procedure Remove_Listener( this     : not null access Widget'Class;
                               listener : not null A_Visibility_Listener ) is
    begin
        this.Remove_Listener( "Visibility_Listener", A_Action_Listener(listener) );
    end Remove_Listener;

    ----------------------------------------------------------------------------

    procedure Remove_Listener( this         : not null access Widget'Class;
                               listenerType : String;
                               listener     : not null A_Action_Listener ) is

        procedure Remove_Listener( key  : String;
                                   list : in out Action_Listeners.List ) is
            pragma Unreferenced( key );
            crsr : Action_Listeners.Cursor := list.Find( listener );
        begin
            -- remove the specified listener
            if Action_Listeners.Has_Element( crsr ) then
                list.Delete( crsr );
            end if;
        end Remove_Listener;

        crsr : constant Listener_Maps.Cursor := this.listeners.Find( listenerType );
    begin
        -- check if we have any listeners registered as the given type
        if Listener_Maps.Has_Element( crsr ) then
            this.listeners.Update_Element( crsr, Remove_Listener'Access );
        end if;
    end Remove_Listener;

    ----------------------------------------------------------------------------

    procedure Request_Focus( this : not null access Widget'Class ) is
    begin
        if this.parent /= null then
            A_Container(this.parent).Give_Focus( A_Widget(this) );
        else
            Dbg( "Error: Unable to focus " & this.To_String & " - not rooted",
                 D_GUI, ERROR );
        end if;
    end Request_Focus;

    ----------------------------------------------------------------------------

    procedure Scroll( this : not null access Widget'Class; x, y : Integer ) is
    begin
        if y /= 0 and then this.cheight > this.y2 - this.y1 + 1 then
            this.vy1 := this.vy1 + y;
            this.vy2 := this.vy2 + y;
            if this.vy1 < 0 then
                -- bound the viewport at the top of the content
                this.vy2 := this.vy2 + (0 - this.vy1);
                this.vy1 := 0;
            elsif this.vy2 >= this.cheight then
                -- bound the viewport at the bottom of the content
                this.vy1 := this.vy1 - (this.vy2 - (this.cheight - 1));
                this.vy2 := this.cheight - 1;
            end if;
        end if;
        if x /= 0 and then this.cwidth > this.x2 - this.x1 + 1 then
            this.vx1 := this.vx1 + x;
            this.vx2 := this.vx2 + x;
            if this.vx1 < 0 then
                -- bound the viewport at the left of the content
                this.vx2 := this.vx2 + (0 - this.vx1);
                this.vx1 := 0;
            elsif this.vx2 >= this.cwidth then
                -- bound the viewport at the right of the content
                this.vx1 := this.vx1 - (this.vx2 - (this.cwidth - 1));
                this.vx2 := this.cwidth - 1;
            end if;
        end if;
    end Scroll;

    ----------------------------------------------------------------------------

    procedure Scroll_To( this : not null access Widget'Class; x, y : Integer ) is
        vwidth  : Integer;
        vheight : Integer;
    begin
        if x /= this.vx1 then
            vwidth := this.vx2 - this.vx1 + 1;
            this.vx1 := x;
            if this.clampView then
                -- prevent scrolling past the right edge of the content if the
                -- content is wider than the viewport
                if this.vx1 + vwidth >= this.cwidth then
                    this.vx1 := this.cwidth - vwidth;
                end if;
            end if;
            if this.vx1 < 0 then
                this.vx1 := 0;
            end if;
            this.vx2 := this.vx1 + vwidth - 1;
        end if;
        if y /= this.vy1 then
            vheight := this.vy2 - this.vy1 + 1;
            this.vy1 := y;
            if this.clampView then
                -- prevent scrolling past the bottom edge of the content if the
                -- content is taller than the viewport
                if this.vy1 + vheight >= this.cheight then
                    this.vy1 := this.cheight - vheight;
                end if;
            end if;
            if this.vy1 < 0 then
                this.vy1 := 0;
            end if;
            this.vy2 := this.vy1 + vheight - 1;
        end if;
    end Scroll_To;

    ----------------------------------------------------------------------------

    procedure Set_Attribute( this : not null access Widget'Class;
                             name : String;
                             val  : Value_Ptr'Class ) is
    begin
        this.attrs.Get.Set( name, val );
    end Set_Attribute;

    ----------------------------------------------------------------------------

    procedure Set_Border( this : access Widget; border : Border_Type ) is
    begin
        if this.border /= border then
            this.border := border;
        end if;
    end Set_Border;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Color( this    : access Widget;
                         purpose : Color_Purpose;
                         color   : Allegro_Color ) is
    begin
        if this.colors(purpose) /= color then
            this.colors(purpose) := color;
        end if;
    end Set_Color;

    ----------------------------------------------------------------------------

    procedure Set_Enabled( this : not null access Widget'Class; enabled : Boolean ) is
    begin
        if enabled = not this.enabled then
            this.enabled := enabled;
            if enabled then
                A_Widget(this).On_Enabled;
            else
                A_Widget(this).On_Disabled;
            end if;
        end if;
    end Set_Enabled;

    ----------------------------------------------------------------------------

    procedure Set_Focusable( this      : not null access Widget'Class;
                             focusable : Boolean ) is
    begin
        this.acceptFocus := focusable;
    end Set_Focusable;

    ----------------------------------------------------------------------------

    procedure Set_Focused( this : not null access Widget'Class; focused : Boolean ) is
    begin
        if this.focused /= focused then
            this.focused := focused;
        end if;
    end Set_Focused;

    ----------------------------------------------------------------------------

    procedure Set_Font( this : not null access Widget'Class;
                        name : String := "";
                        size : Natural := 0 ) is
    begin
        if name'Length > 0 then
            this.fontName := To_Unbounded_String( name );
        end if;
        if size > 0 then
            this.fontSize := size;
        end if;
    end Set_Font;

    ----------------------------------------------------------------------------

    procedure Set_Layout( this   : not null access Widget'Class;
                          layout : in out A_Layout ) is
    begin
        Delete( this.layout );
        this.layout := layout;
        if this.parent /= null then
            A_Widget(this).Pack;
        end if;
        layout := null;
    end Set_Layout;

    ----------------------------------------------------------------------------

    procedure Set_Layout_Center( this   : not null access Widget'Class;
                                 width,
                                 height : Natural := 0 ) is
        layout : A_Layout := Create_Layout_Center( width, height );
    begin
        this.Set_Layout( layout );
    end Set_Layout_Center;

    ----------------------------------------------------------------------------

    procedure Set_Layout_CenterH( this   : not null access Widget'Class;
                                  width  : Natural;
                                  top,
                                  bottom : Integer ) is
        layout : A_Layout := Create_Layout_CenterH( width, top, bottom );
    begin
        this.Set_Layout( layout );
    end Set_Layout_CenterH;

    ----------------------------------------------------------------------------

    procedure Set_Layout_CenterHY( this   : not null access Widget'Class;
                                   width  : Natural;
                                   y1, y2 : Integer ) is
        layout : A_Layout := Create_Layout_CenterHY( width, y1, y2 );
    begin
        this.Set_Layout( layout );
    end Set_Layout_CenterHY;

    ----------------------------------------------------------------------------

    procedure Set_Layout_LTRB( this   : not null access Widget'Class;
                               left,
                               top,
                               right,
                               bottom : Integer ) is
        layout : A_Layout := Create_Layout_LTRB( left, top, right, bottom );
    begin
        this.Set_Layout( layout );
    end Set_Layout_LTRB;

    ----------------------------------------------------------------------------

    procedure Set_Layout_LTWH( this   : not null access Widget'Class;
                               left,
                               top    : Integer;
                               width,
                               height : Natural ) is
        layout : A_Layout := Create_Layout_LTWH( left, top, width, height );
    begin
        this.Set_Layout( layout );
    end Set_Layout_LTWH;

    ----------------------------------------------------------------------------

    procedure Set_Layout_XYWH( this   : not null access Widget'Class;
                               x, y   : Integer;
                               width,
                               height : Natural ) is
        layout : A_Layout := Create_Layout_XYWH( x, y, width, height );
    begin
        this.Set_Layout( layout );
    end Set_Layout_XYWH;

    ----------------------------------------------------------------------------

    procedure Set_Min_Height( this   : not null access Widget'Class;
                              height : Natural ) is
    begin
        this.minHeight := height;
        A_Widget(this).Pack;
    end Set_Min_Height;

    ----------------------------------------------------------------------------

    procedure Set_Min_Width( this  : not null access Widget'Class;
                             width : Natural ) is
    begin
        this.minWidth := width;
        A_Widget(this).Pack;
    end Set_Min_Width;

    ----------------------------------------------------------------------------

    procedure Set_Next( this : not null access Widget'Class; id : String ) is
    begin
        this.nextId := To_Unbounded_String( id );
    end Set_Next;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Parent( this : access Widget; parent : A_Widget ) is
        wasRooted : Boolean;
        wasShown  : Boolean;
    begin
        -- NOTE: This procedure only sets the new parent of the widget, it does
        --       not break the current parent's association with this widget as
        --       a child.
        if parent = this.parent then
            return;
        end if;

        wasRooted := A_Widget(this).Is_Rooted;
        wasShown := wasRooted and then A_Widget(this).Is_Showing;

        -- if the parent has already been set, remove it from its widget tree
        -- before setting its new parent.
        if this.parent /= null then
            this.parent := null;
            if wasRooted then
                A_Widget(this).Handle_Rooted( False );
                if wasShown then
                    A_Widget(this).Handle_Shown( False );
                end if;
            end if;
        end if;

        -- set the widget's new parent
        if parent /= null then
            this.parent := parent;
            if A_Widget(this).Is_Rooted then
                A_Widget(this).Pack;
                A_Widget(this).Handle_Rooted( True );
                if A_Widget(this).Is_Showing then
                    A_Widget(this).Handle_Shown( True );
                end if;
            end if;
        end if;
    end Set_Parent;

    ----------------------------------------------------------------------------

    procedure Set_Prev( this : not null access Widget'Class; id : String ) is
    begin
        this.prevId := To_Unbounded_String( id );
    end Set_Prev;

    ----------------------------------------------------------------------------

    procedure Set_Transparent( this        : not null access Widget'Class;
                               transparent : Boolean ) is
    begin
        if this.transparent /= transparent then
            this.transparent := transparent;
        end if;
    end Set_Transparent;

    ----------------------------------------------------------------------------

    procedure Set_Visible( this : not null access Widget'Class; visible : Boolean ) is
        wasShown : Boolean;
    begin
        if visible = this.visible then
            return;
        end if;

        wasShown := A_Widget(this).Is_Showing;

        this.visible := visible;
        if A_Widget(this).Is_Showing then
            -- the widget became shown
            A_Widget(this).Handle_Shown( True );
        elsif wasShown then
            -- the widget became unshown
            A_Widget(this).Handle_Shown( False );
        end if;
    end Set_Visible;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Zoom( this : access Widget; zoom : Float ) is
        vw, vh : Integer;    -- view width, height
    begin
        if this.zoomable and then this.zoom /= zoom then
            this.zoom := zoom;

            vw := Integer(Float'Ceiling( Float(this.x2 - this.x1 + 1) / zoom ));
            vh := Integer(Float'Ceiling( Float(this.y2 - this.y1 + 1) / zoom ));

            this.vx2 := Integer'Min( this.vx1 + vw, this.cwidth ) - 1;
            this.vy2 := Integer'Min( this.vy1 + vh, this.cheight ) - 1;
            this.vx1 := this.vx2 - vw + 1;
            this.vy1 := this.vy2 - vh + 1;
            if this.vx1 < 0 then
                this.vx1 := 0;
                this.vx2 := vw - 1;
            end if;
            if this.vy1 < 0 then
                this.vy1 := 0;
                this.vy2 := vh - 1;
            end if;

            -- update the .buffer field to match the new viewport dimensions
            this.Update_Zoom_Buffer;
        end if;
    end Set_Zoom;

    ----------------------------------------------------------------------------

    overriding
    function To_String( this : access Widget ) return String is
    begin
        return '<' & this.Get_Class_Name & " : id=" & To_String( this.id ) & '>';
    end To_String;

    ----------------------------------------------------------------------------

    not overriding
    procedure Translate_To_Content( this   : access Widget;
                                    sx, sy : Integer;
                                    cx, cy : out Integer ) is
    begin
        if this.parent /= null then
            this.parent.Translate_To_Content( sx, sy, cx, cy );
        else
            cx := sx;
            cy := sy;
        end if;
        cx := Integer(Float'Ceiling((cx - this.x1) / this.zoom)) + this.vx1;
        cy := Integer(Float'Ceiling((cy - this.y1) / this.zoom)) + this.vy1;
    end Translate_To_Content;

    ----------------------------------------------------------------------------

    not overriding
    procedure Translate_To_Window( this   : access Widget;
                                   cx, cy : Integer;
                                   wx, wy : out Integer ) is
        parentX, parentY : Integer;
    begin
        -- translate the content coordinates of this widget to the content
        -- coordinates of the parent
        wx := Integer(this.zoom * (cx - this.vx1)) + this.x1;
        wy := Integer(this.zoom * (cy - this.vy1)) + this.y1;

        -- translate the content coordinates of the parent to screen coordinates
        if this.parent /= null then
            parentX := wx;
            parentY := wy;
            this.parent.Translate_To_Window( parentX, parentY, wx, wy );
        end if;
    end Translate_To_Window;

    ----------------------------------------------------------------------------

    procedure Update_Zoom_Buffer( this : not null access Widget'Class ) is
        vw    : constant Integer := this.vx2 - this.vx1 + 1;
        vh    : constant Integer := this.vy2 - this.vy1 + 1;
        state : Allegro_State;
    begin
        Al_Destroy_Bitmap( this.buffer );
        if vw > 0 and then vh > 0 then
            Al_Store_State( state, ALLEGRO_STATE_ALL );
            Al_Set_New_Bitmap_Flags( ALLEGRO_MIN_LINEAR or
                                     ALLEGRO_MAG_LINEAR );
            this.buffer := Al_Create_Bitmap( vw, vh );
            Al_Restore_State( state );
        end if;
    end Update_Zoom_Buffer;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Widget ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    ----------------------------------------------------------------------------

    function Lt( l, r : A_Widget ) return Boolean is
    begin
        if l = null then
            return r /= null;
        elsif r /= null then
            return l.id < r.id;
        end if;

        -- l /= null and r = null
        return False;
    end Lt;

    --==========================================================================

    not overriding
    procedure Apply( this : access Layout; widget : not null A_Widget ) is
        pragma Unreferenced( this, widget );
    begin
        raise Constraint_Error with "Invoked abstract method Layout.Apply";
    end Apply;
    pragma No_Return( Apply );

    --==========================================================================

    not overriding
    procedure Construct( this      : access Key_Action;
                         id        : Action_Id;
                         source    : not null A_Widget;
                         key       : Integer;
                         modifiers : Modifiers_Array ) is
    begin
        Action(this.all).Construct( id, source );
        this.key := key;
        this.modifiers := modifiers;
    end Construct;

    ----------------------------------------------------------------------------

    function Get_Key( this : not null access Key_Action'Class ) return Integer is
    begin
        return this.key;
    end Get_Key;

    ----------------------------------------------------------------------------

    function Get_Modifiers( this : not null access Key_Action'Class ) return Modifiers_Array is
    begin
        return this.modifiers;
    end Get_Modifiers;

    --==========================================================================

    not overriding
    procedure Construct( this    : access Visibility_Action;
                         id      : Action_Id;
                         source  : not null A_Widget;
                         visible : Boolean ) is
    begin
        Action(this.all).Construct( id, source );
        this.visible := visible;
    end Construct;

    ----------------------------------------------------------------------------

    function Is_Visible( this : not null access Visibility_Action'Class ) return Boolean is
    begin
        return this.visible;
    end Is_Visible;

end Widgets;
