--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Assertions;                    use Ada.Assertions;
with Ada.Containers.Indefinite_Hashed_Sets;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Streams.Stream_IO;             use Ada.Streams.Stream_IO;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Strings.Hash;
with Ada.Tags;                          use Ada.Tags;
with Applications;                      use Applications;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.World;                      use Events.World;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;
with Hashed_Strings;                    use Hashed_Strings;
with Resources;                         use Resources;
with Streams.Buffers;                   use Streams.Buffers;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;
with Tiles.Libraries.Manager;           use Tiles.Libraries.Manager;
with Values.Casting;                    use Values.Casting;
with Values.Ids;                        use Values.Ids;
with Values.Nulls;                      use Values.Nulls;
with Values.Numbers;                    use Values.Numbers;
with Values.Strings;                    use Values.Strings;

package body Worlds is

    package String_Sets is new
        Ada.Containers.Indefinite_Hashed_Sets( String, Ada.Strings.Hash, "=", "=" );

    IDENTIFIER    : constant String := "galaxy";
    READ_VERSION  : constant Natural := 22;
    WRITE_VERSION : constant Natural := 22;

    domains : String_Sets.Set;

    ----------------------------------------------------------------------------

    -- Handles an Entities_Collided event.
    procedure Handle_Entities_Collided( this : not null access World_Object'Class;
                                        evt  : not null A_Entities_Event );

    -- Handles an Entities_Separated event.
    procedure Handle_Entities_Separated( this : not null access World_Object'Class;
                                         evt  : not null A_Entities_Event );

    -- Handles an Entity_Directive event.
    procedure Handle_Entity_Directive( this : not null access World_Object'Class;
                                       evt  : not null A_Entity_Directive_Event );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Equivalent( l, r : Animated_Info ) return Boolean is
    begin
        return l.layer = r.layer and then l.x = r.x and then l.y = r.y;
    end Equivalent;

    ----------------------------------------------------------------------------

    function Get_Frame( ai : Animated_Info; index : Integer ) return Natural is
        id : Number_Ptr;
    begin
        -- .frames should be a list of natural (non-negative) numbers
        if ai.frames.Not_Null and then index >= 1 then
            id := As_Number( ai.frames.Get.Get( index ) );
            if id.Not_Null then
                Assert( id.Get.To_Int >= 0 );
                return Natural(id.Get.To_Int);
            end if;
        end if;
        return 0;
    end Get_Frame;

    ----------------------------------------------------------------------------

    function Hash( ai : Animated_Info ) return Hash_Type is
    begin
        return Hash_Type((ai.layer * MAX_WIDTH * MAX_HEIGHT) + (ai.y * MAX_WIDTH) + ai.x);
    end Hash;

    ----------------------------------------------------------------------------

    function Is_Looping( ai : Animated_Info ) return Boolean is
    begin
        return ai.frames.Not_Null;
    end Is_Looping;

    ----------------------------------------------------------------------------

    function Animated_Info_Input( stream : access Root_Stream_Type'Class ) return Animated_Info is
        info : Animated_Info;
    begin
        info.layer := Integer'Input( stream );
        info.x := Integer'Input( stream );
        info.y := Integer'Input( stream );
        info.nextUpdate := Clock + Time_Span'Input( stream );
        info.frameDelay := Time_Span'Input( stream );
        info.frames := As_List( Value_Ptr'Input( stream ) );
        return info;
    end Animated_Info_Input;

    ----------------------------------------------------------------------------

    procedure Animated_Info_Output( stream : access Root_Stream_Type'Class;
                                    info   : Animated_Info ) is
        clk : constant Time := Clock;
    begin
        Integer'Output( stream, info.layer );
        Integer'Output( stream, info.x );
        Integer'Output( stream, info.y );
        if info.nextUpdate <= clk then
            Time_Span'Output( stream, Time_Span_Zero );
        else
            Time_Span'Output( stream, info.nextUpdate - clk );
        end if;
        Time_Span'Output( stream, info.frameDelay );
        Value_Ptr'Output( stream, info.frames.As_Value );
    end Animated_Info_Output;

    --==========================================================================

    function Load_World( stream : Stream_Access ) return A_World is
        world : A_World;
    begin
        world := A_World'Input( stream );
        if world /= null then
            world.lib := Load_Library( Cast_String( world.Get_Property( "library" ) ),
                                       bitmaps => False );
            if world.lib = null then
                raise RESOURCE_ERROR with "Unable to open library <" &
                                          Cast_String( world.Get_Property( "library" ) ) & '>';
            end if;
        end if;
        return world;
    exception
        when others =>
            Delete( world );
            raise;
    end Load_World;

    ----------------------------------------------------------------------------

    function Load_World( name : String ) return A_World is
        filename : constant String := Add_Extension( name, World_Extension );
        resource : A_Resource_File;
        stream   : A_Buffer_Stream;
        this     : A_World;
    begin
        pragma Debug( Dbg( "Loading world: " & filename, D_GAME, INFO ) );
        resource := Load_Resource( filename, "worlds" );
        if resource /= null then
            resource.Get_Stream( stream );
            this := Load_World( Stream_Access(stream) );
            Close( stream );
            Release( resource );
            if this /= null then
                this.Initialize;
                this.properties.Get.Set( "filename", Create( filename ) );
            else
                raise READ_EXCEPTION with "No data in file: " & filename;
            end if;
        else
            raise FILE_NOT_FOUND with "File not found: " & filename;
        end if;
        return this;
    exception
        when others =>
            Delete( this );
            Close( stream );
            Release( resource );
            raise;
    end Load_World;

    ----------------------------------------------------------------------------

    function Create_World( width,
                           height         : Positive;
                           physicalLayers : Maps.Boolean_Array;
                           library,
                           domain         : String ) return A_World is
        this : A_World := new World_Object;
    begin
        this.Construct( width, height, physicalLayers, library, domain );
        return this;
    exception
        when others =>
            Delete( this );
            raise;
    end Create_World;

    ----------------------------------------------------------------------------

    overriding
    procedure Adjust( this : access World_Object ) is
        pragma Unreferenced( this );
    begin
        raise COPY_NOT_ALLOWED;
    end Adjust;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access World_Object ) is
    begin
        Object(this.all).Construct;
        this.properties := Create_Assoc;
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this           : access World_Object;
                         width,
                         height         : Positive;
                         physicalLayers : Maps.Boolean_Array;
                         library,
                         domain         : String ) is
    begin
        this.Construct;

        this.lib := Load_Library( library, bitmaps => False );
        if this.lib = null then
            raise RESOURCE_ERROR with "Unable to open tile library " &
                                      '<' & library & '>';
        end if;
        this.Set_Property( "library", Create( library ) );

        -- domain length zero is a special case where we are reading the
        -- world from a stream and don't yet know the domain.
        if domain'Length > 0 and then not Valid_Domain( domain ) then
            raise Constraint_Error with "Invalid world domain <" & domain & '>';
        end if;
        this.properties.Get.Set( "domain", Create( domain ) );
        this.properties.Get.Set( "filename", Create( "" ) );

        this.map := Create_Map( width, height, physicalLayers );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out World_Object ) is

        procedure Delete_Entity( pos : Entity_Map.Cursor ) is
            e : A_Entity := Entity_Map.Element( pos );
        begin
            Delete( e );
        end Delete_Entity;

    begin
        pragma Debug( this.attached,
                      Dbg( "Warning: Deleting world attached to framework", D_GAME, WARNING ) );
        Delete( this.map );
        Unload_Library( this.lib );
        this.entities.Iterate( Delete_Entity'Access );
        this.entities.Clear;
        this.animated.Clear;
        this.properties := Associations.Nul;

        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Attach_To_Framework( this      : not null access World_Object'Class;
                                   gameState : not null A_Game_State;
                                   corral    : not null A_Corral ) is
    begin
        if not this.attached then
            this.gameState := gameState;
            this.corral := corral;
            this.attached := True;
            A_World(this).On_Attach;
        else
            null;
            pragma Debug( Dbg( "Warning: World already attached to framework",
                               D_GAME, WARNING ) );
        end if;
    end Attach_To_Framework;

    ----------------------------------------------------------------------------

    procedure Detach_From_Framework( this : not null access World_Object'Class ) is

        procedure Cancel( e : not null A_Entity ) is
        begin
            e.Cancel_Directives;
        end Cancel;

    begin
        if this.attached then
            -- cancel all ongoing directives to entities; the world may be
            -- reattached later and they will be given new directives at that
            -- time.
            this.Examine_Entities( Cancel'Unrestricted_Access );

            -- let the world implementation handle the event
            A_World(this).On_Detach;

            this.corral := null;
            this.attached := False;
        else
            null;
            pragma Debug( Dbg( "Warning: World not attached to framework", D_GAME, WARNING ) );
        end if;
    end Detach_From_Framework;

    ----------------------------------------------------------------------------

    overriding
    function Evaluate_Function( this      : access World_Object;
                                name      : String;
                                arguments : Value_Array ) return Value_Ptr is
        pragma Unreferenced( this, name, arguments );
    begin
        return Create_Null;
    end Evaluate_Function;

    ----------------------------------------------------------------------------

    overriding
    function Evaluate_Symbol( this : access World_Object; symbol : String ) return Value_Ptr is
    begin
        return this.gameState.Get_Var( symbol );
    end Evaluate_Symbol;

    ----------------------------------------------------------------------------

    procedure Examine_Entities( this    : not null access World_Object'Class;
                                examine : not null access procedure( e : not null A_Entity ) ) is
        crsr : Entity_Map.Cursor := this.entities.First;
    begin
        while Entity_Map.Has_Element( crsr ) loop
            examine.all( Entity_Map.Element( crsr ) );
            crsr := Entity_Map.Next( crsr );
        end loop;
    end Examine_Entities;

    ----------------------------------------------------------------------------

    function Get_Entity( this : not null access World_Object'Class; id : Entity_Id ) return A_Entity is
        pos : constant Entity_Map.Cursor := this.entities.Find( id );
    begin
        if Entity_Map.Has_Element( pos ) then
            return Entity_Map.Element( pos );
        end if;
        return null;
    end Get_Entity;

    ----------------------------------------------------------------------------

    function Get_Game_State( this : not null access World_Object'Class ) return A_Game_State is
    begin
        return this.gameState;
    end Get_Game_State;

    ----------------------------------------------------------------------------

    function Get_Height( this : not null access World_Object'Class ) return Positive is
        result : Integer := 0;
    begin
        if this.map /= null then
            result := this.map.Get_Height * this.map.Tile_Width;
        end if;
        return result;
    end Get_Height;

    ----------------------------------------------------------------------------

    function Get_Height_Tiles( this : not null access World_Object'Class ) return Positive is
        result : Integer := 0;
    begin
        if this.map /= null then
            result := this.map.Get_Height;
        end if;
        return result;
    end Get_Height_Tiles;

    ----------------------------------------------------------------------------

    function Get_Layers( this : not null access World_Object'Class ) return Positive is
    begin
        return this.map.Get_Layers;
    end Get_Layers;

    ----------------------------------------------------------------------------

    function Get_Library( this : not null access World_Object'Class ) return A_Tile_Library is
    begin
        return this.lib;
    end Get_Library;

    ----------------------------------------------------------------------------

    function Get_Player( this : not null access World_Object'Class ) return A_Player is
    begin
        return A_Player(this.Get_Entity( To_Entity_Id( As_Id( this.Get_Property( "playerId" ) ) ) ));
    end Get_Player;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access World_Object ) return String is
    begin
        if Cast_String( this.Get_Property( "filename" ) )'Length = 0 then
            return "World('Untitled World')";
        else
            return "World('" & Cast_String( this.Get_Property( "filename" ) ) & "')";
        end if;
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    function Get_Property( this : not null access World_Object'Class;
                           name : String ) return Value_Ptr is
    begin
        return this.properties.Get.Get( name );
    end Get_Property;

    ----------------------------------------------------------------------------

    function Get_Tile_Id( this  : not null access World_Object'Class;
                          layer : Positive;
                          x, y  : Integer ) return Natural is
        anmInfo : Animated_Info := Animated_Info'(layer, x, y,
                                                  Time_First,
                                                  Time_Span_Zero,
                                                  Lists.Nul);
        pos : constant Animated_Set.Cursor := this.animated.Find( anmInfo );
    begin
        if Animated_Set.Has_Element( pos ) then
            -- if there's a looping animation at this location, return the first
            -- frame instead of whatever happens to be showing. for one-shot
            -- animations, we'll return the actual current tile.
            anmInfo := Animated_Set.Element( pos );
            if Is_Looping( anmInfo ) then
                return Get_Frame( anmInfo, 1 );
            end if;
        end if;
        return this.map.Get_Tile_Id( layer, x, y );
    end Get_Tile_Id;

    ----------------------------------------------------------------------------

    function Get_Width( this : not null access World_Object'Class ) return Positive is
        result : Integer := 0;
    begin
        if this.map /= null then
            result := this.map.Get_Width * this.map.Tile_Width;
        end if;
        return result;
    end Get_Width;

    ----------------------------------------------------------------------------

    function Get_Width_Tiles( this : not null access World_Object'Class ) return Positive is
        result : Integer := 0;
    begin
        if this.map /= null then
            result := this.map.Get_Width;
        end if;
        return result;
    end Get_Width_Tiles;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Delete_Entity( this : access World_Object;
                                    evt  : not null A_Entity_Event ) is
        pos : Entity_Map.Cursor := this.entities.Find( evt.Get_Id );
        e   : A_Entity;
    begin
        if Entity_Map.Has_Element( pos ) then
            e := Entity_Map.Element( pos );
            this.entities.Delete( pos );
            Delete( e );
            Queue_Entity_Deleted( evt.Get_Id );
            Queue_World_Modified;
        end if;
    end Handle_Delete_Entity;

    ----------------------------------------------------------------------------

    procedure Handle_Entities_Collided( this : not null access World_Object'Class;
                                        evt  : not null A_Entities_Event ) is
        entityA : constant A_Entity := this.Get_Entity( evt.Get_A );
        entityB : constant A_Entity := this.Get_Entity( evt.Get_B );
    begin
        -- check if the entities exist because they may have been deleted by an
        -- event just prior to this one
        if entityA /= null and then entityB /= null then
            entityA.Collided( entityB );
            entityB.Collided( entityA );
        end if;
    end Handle_Entities_Collided;

    ----------------------------------------------------------------------------

    procedure Handle_Entities_Separated( this : not null access World_Object'Class;
                                         evt  : not null A_Entities_Event ) is
        entityA : constant A_Entity := this.Get_Entity( evt.Get_A );
        entityB : constant A_Entity := this.Get_Entity( evt.Get_B );
    begin
        -- check if the entities exist because they may have been deleted by an
        -- event just prior to this one
        if entityA /= null and then entityB /= null then
            entityA.Separated( entityB );
            entityB.Separated( entityA );
        end if;
    end Handle_Entities_Separated;

    ----------------------------------------------------------------------------

    procedure Handle_Entity_Directive( this : not null access World_Object'Class;
                                       evt  : not null A_Entity_Directive_Event ) is
        e : constant A_Entity := this.Get_Entity( evt.Get_Id );
    begin
        if e /= null then
            pragma Debug( Dbg( "World: Directive [" & Hashed_Strings.To_String( evt.Get_Directive ) &
                               ":" & evt.Get_Mode'Img & "] -> Entity " & Entities.Image( evt.Get_Id ),
                               D_GAME, INFO ) );
            e.Set_Directive( evt.Get_Directive, evt.Get_Mode );
            Queue_World_Modified;
        end if;
    end Handle_Entity_Directive;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Entity_Grounded( this : access World_Object;
                                      evt  : not null A_Entity_Grounded_Event ) is
        e : constant A_Entity := this.Get_Entity( evt.Get_Id );
    begin
        if e /= null then
            e.Set_Grounded( evt.Is_Grounded );
        end if;
    end Handle_Entity_Grounded;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Entity_Hit_Wall( this : access World_Object;
                                      evt  : not null A_Entity_Hit_Wall_Event ) is
        e : constant A_Entity := this.Get_Entity( evt.Get_Id );
    begin
        if e /= null then
            e.Hit_Wall( evt.Get_Direction, evt.Is_First_Contact );
        end if;
    end Handle_Entity_Hit_Wall;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Entity_Moved( this : access World_Object;
                                   evt  : not null A_Entity_Moved_Event ) is
        e : constant A_Entity := this.Get_Entity( evt.Get_Id );
    begin
        if e /= null then
            e.Set_Location( evt.Get_X, evt.Get_Y );
            e.Set_Velocity_X( evt.Get_XV );
            e.Set_Velocity_Y( evt.Get_YV );
            Queue_World_Modified;
        end if;
    end Handle_Entity_Moved;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Entity_Resized( this : access World_Object;
                                     evt  : not null A_Entity_Resized_Event ) is
        e : constant A_Entity := this.Get_Entity( evt.Get_Id );
    begin
        if e /= null then
            e.Set_Size( evt.Get_Width, evt.Get_Height );
            Queue_World_Modified;
        end if;
    end Handle_Entity_Resized;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Set_Entity_Attribute( this : access World_Object;
                                           evt  : not null A_Entity_Attribute_Event ) is
        e : constant A_Entity := this.Get_Entity( evt.Get_Id );
    begin
        if e /= null then
            e.Set_Attribute( evt.Get_Attribute, evt.Get_Value );
            Queue_World_Modified;
        end if;
    end Handle_Set_Entity_Attribute;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Spawn_Entity( this : access World_Object;
                                   evt  : not null A_Spawn_Entity_Event ) is
        eid : Entity_Id;
        pragma Warnings( Off, eid );
    begin
        eid := A_World(this).Spawn_Entity( evt.Get_Id,
                                           evt.Get_X, evt.Get_Y,
                                           evt.Get_Width, evt.Get_Height,
                                           evt.Get_XV, evt.Get_YV );
    end Handle_Spawn_Entity;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access World_Object;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        if evt.Get_Id = ENTITY_MOVED_ID then
            A_World(this).Handle_Entity_Moved( A_Entity_Moved_Event(evt) );
        elsif evt.Get_Id = ENTITY_DIRECTIVE_ID then
            A_World(this).Handle_Entity_Directive( A_Entity_Directive_Event(evt) );
        elsif evt.Get_Id = ENTITIES_COLLIDED_ID then
            A_World(this).Handle_Entities_Collided( A_Entities_Event(evt) );
        elsif evt.Get_Id = ENTITIES_SEPARATED_ID then
            A_World(this).Handle_Entities_Separated( A_Entities_Event(evt) );
        elsif evt.Get_Id = ENTITY_HIT_WALL_ID then
            A_World(this).Handle_Entity_Hit_Wall( A_Entity_Hit_Wall_Event(evt) );
        elsif evt.Get_Id = ENTITY_GROUNDED_ID then
            A_World(this).Handle_Entity_Grounded( A_Entity_Grounded_Event(evt) );
        elsif evt.Get_Id = DELETE_ENTITY_ID then
            A_World(this).Handle_Delete_Entity( A_Entity_Event(evt) );
        elsif evt.Get_Id = ENTITY_RESIZED_ID then
            A_World(this).Handle_Entity_Resized( A_Entity_Resized_Event(evt) );
        elsif evt.Get_Id = SET_ENTITY_ATTRIBUTE_ID then
            A_World(this).Handle_Set_Entity_Attribute( A_Entity_Attribute_Event(evt) );
        elsif evt.Get_Id = SPAWN_ENTITY_ID then
            A_World(this).Handle_Spawn_Entity( A_Spawn_Entity_Event(evt) );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    not overriding
    procedure Initialize( this : access World_Object ) is

        procedure Examine( pos : Entity_Map.Cursor ) is
        begin
            Entity_Map.Element( pos ).Set_World( this );
        end Examine;

    begin
        -- Pass a reference to this world to each entity in it
        this.entities.Iterate( Examine'Access );
    end Initialize;

    ----------------------------------------------------------------------------

    not overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return World_Object is
        this : aliased World_Object;
    begin
        this.Construct;
        World_Object'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out World_Object ) is

        ------------------------------------------------------------------------

        procedure Read_Entities is
            count  : constant Natural := Natural'Input( stream );
            entity : A_Entity;
        begin
            for i in 1..count loop
                entity := A_Entity'Input( stream );
                if entity /= null then
                    obj.entities.Insert( entity.Get_Id, entity );
                end if;
            end loop;
        end Read_Entities;

        ------------------------------------------------------------------------

        procedure Read_Animated_Info is
            count : constant Natural := Natural'Input( stream );
            info  : Animated_Info;
        begin
            for i in 1..count loop
                info := Animated_Info'Input( stream );
                obj.animated.Include( info );
            end loop;
        end Read_Animated_Info;

        ------------------------------------------------------------------------

    begin
        Object'Read( stream, Object(obj) );

        obj.properties := As_Assoc( Value_Ptr'Input( stream ) );
        obj.map := A_Map'Input( stream );
        Read_Animated_Info;
        Read_Entities;

        -- validate the world domain
        if not Valid_Domain( Cast_String( obj.properties.Get.Get( "domain" ) ) ) then
            raise Constraint_Error with "Invalid world domain <" &
                  Cast_String( obj.properties.Get.Get( "domain" ) ) & '>';
        end if;
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : World_Object ) is

        ------------------------------------------------------------------------

        procedure Write_Entity( pos : Entity_Map.Cursor ) is
            e : constant A_Entity := Entity_Map.Element( pos );
        begin
            A_Entity'Output( stream, e );
        end Write_Entity;

        ------------------------------------------------------------------------

        procedure Write_Animated_Info( pos : Animated_Set.Cursor ) is
            info : constant Animated_Info := Animated_Set.Element( pos );
        begin
            Animated_Info'Output( stream, info );
        end Write_Animated_Info;

        ------------------------------------------------------------------------

        properties : Assoc_Ptr;
    begin
        Object'Write( stream, Object(obj) );

        properties := As_Assoc( obj.properties.Clone );
        properties.Get.Set( "filename", Create_Null ); -- don't store "filename"
        Value_Ptr'Output( stream, properties.As_Value );

        A_Map'Output( stream, obj.map );

        Natural'Output( stream, Natural( Animated_Set.Length(obj.animated) ) );
        obj.animated.Iterate( Write_Animated_Info'Access );

        Natural'Output( stream, Natural( Entity_Map.Length(obj.entities) ) );
        obj.entities.Iterate( Write_Entity'Access );
    end Object_Write;

    ----------------------------------------------------------------------------

    not overriding
    procedure On_Attach( this : access World_Object ) is

        procedure On_Load( pos : Entity_Map.Cursor ) is
        begin
            Entity_Map.Element( pos ).On_Load;
        end On_Load;

        playerId : Id_Ptr;
    begin
        pragma Debug( Dbg( "Attaching world <" &
                           Cast_String( this.Get_Property( "filename" ) ) &
                           "> to framework", D_GAME, INFO ) );

        -- receive updates for things that have happened
        this.corral.Add_Listener( A_Event_Listener(this), "Entities_Collided" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entities_Separated" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entity_Grounded" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entity_Hit_Wall" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entity_Moved" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entity_Resized" );

        -- respond to commands to do something
        this.corral.Add_Listener( A_Event_Listener(this), "Delete_Entity" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entity_Directive" );
        this.corral.Add_Listener( A_Event_Listener(this), "Set_Entity_Attribute" );
        this.corral.Add_Listener( A_Event_Listener(this), "Spawn_Entity" );

        A_World(this).Queue_Load_Events;
        this.entities.Iterate( On_Load'Access );

        playerId := As_Id( this.properties.Get.Get( "playerId" ) );
        if playerId /= Ids.Nul then
            Queue_Follow_Entity( To_Entity_Id( playerId ) );
        end if;
    end On_Attach;

    ----------------------------------------------------------------------------

    not overriding
    procedure On_Detach( this : access World_Object ) is
    begin
        pragma Debug( Dbg( "Detaching world <" &
                           Cast_String( this.Get_Property( "filename" ) ) &
                           "> from framework", D_GAME, INFO ) );
        this.corral.Remove_Listener( A_Event_Listener(this), "Spawn_Entity" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Set_Entity_Attribute" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entity_Directive" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Delete_Entity" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entity_Resized" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entity_Moved" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entity_Hit_Wall" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entity_Grounded" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entities_Separated" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entities_Collided" );
    end On_Detach;

    ----------------------------------------------------------------------------

    procedure Queue_Load_Events( this : not null access World_Object'Class ) is

        ------------------------------------------------------------------------

        procedure Entity_Created( pos : Entity_Map.Cursor ) is
            e : constant A_Entity := Entity_Map.Element( pos );
        begin
            Queue_Entity_Created( e.Get_Id, e.Get_Class_Name( full=>True ),
                                  e.Is_Physical, e.Is_Metaphysical,
                                  e.Is_Clipped, e.Get_Width, e.Get_Height,
                                  e.Get_X, e.Get_Y, e.Get_XV, e.Get_YV,
                                  e.Get_Lib_Name, e.Get_Frame, e.Get_Attributes );
        end Entity_Created;

        ------------------------------------------------------------------------

        procedure Property_Changed( key   : String;
                                    value : Value_Ptr ) is
        begin
            Queue_World_Property_Changed( key, value );
        end Property_Changed;

        ------------------------------------------------------------------------

    begin
        Queue_World_Loaded( this.map.Get_Width,
                            this.map.Get_Height,
                            this.map.Tile_Width,
                            this.map.Get_Layers_Data,
                            Cast_String( this.Get_Property( "library" ) ) );
        this.entities.Iterate( Entity_Created'Access );
        this.properties.Get.Iterate( Property_Changed'Access );
    end Queue_Load_Events;

    ----------------------------------------------------------------------------

    procedure Resize( this   : not null access World_Object'Class;
                      width,
                      height : Positive ) is

        ------------------------------------------------------------------------

        -- cull animation info for tiles no longer on the map
        procedure Cull_Animations is
            pos,
            next    : Animated_Set.Cursor;
            anmInfo : Animated_Info;
        begin
            pos := this.animated.First;
            while Animated_Set.Has_Element( pos ) loop
                next := Animated_Set.Next( pos );
                anmInfo := Animated_Set.Element( pos );
                if anmInfo.x >= this.map.Get_Width or else
                   anmInfo.y >= this.map.Get_Height
                then
                    this.animated.Delete( pos );
                end if;
                pos := next;
            end loop;
        end Cull_Animations;

        ------------------------------------------------------------------------

        -- cull entities that are now off the map. keep the player entity but
        -- ensure it stays within the new map's bounds.
        procedure Cull_Entities is
            pos,
            next   : Entity_Map.Cursor;
            entity : A_Entity;
        begin
            pos := this.entities.First;
            while Entity_Map.Has_Element( pos ) loop
                next := Entity_Map.Next( pos );
                entity := Entity_Map.Element( pos );
                if entity.Get_X >= Float(this.Get_Width - 1) or else
                   entity.Get_Y >= Float(this.Get_Height - 1)
                then
                    if entity.all not in Player'Class then
                        -- the entity is not a player; delete it for being
                        -- outside of the new map boundaries.
                        this.entities.Delete( pos );
                        Delete( entity );
                    else
                        -- this is a player and can't be deleted. move it back
                        -- inside the map boundaries.
                        entity.Set_Location( Float'Min( entity.Get_X, Float(this.Get_Width - 1) ),
                                             Float'Min( entity.Get_Y, Float(this.Get_Height - 1) ) );
                    end if;
                end if;
                pos := next;
            end loop;
        end Cull_Entities;

        ------------------------------------------------------------------------

    begin
        if width /= this.map.Get_Width or else height /= this.map.Get_Height then
            this.map.Resize( width, height );
            Cull_Animations;
            Cull_Entities;
            if this.attached then
                A_World(this).Queue_Load_Events;
                Queue_World_Modified;
            end if;
        end if;
    end Resize;

    ----------------------------------------------------------------------------

    procedure Save( this      : not null access World_Object'Class;
                    name      : String;
                    overwrite : Boolean := True ) is
        filename : constant String := Add_Extension( name, World_Extension, force => True );
        file     : Ada.Streams.Stream_IO.File_Type;
        sa       : Stream_Access;
    begin
        -- attempt to create the file in a writable location
        begin
            -- attempt to find and overwrite or override the resource file if it
            -- exists, or write it to the current working directory if it does
            -- not exist yet.
            if Is_Regular_File( Write_Path( filename, "worlds" ) ) and then not overwrite then
                raise WRITE_EXCEPTION with "File already exists";
            end if;
            Create( file, Out_File, Write_Path( filename, "worlds" ) );
        exception
            when WRITE_EXCEPTION => raise;
            when others =>
                begin
                    -- attempt to write the world file to the current working
                    -- directory because the Write_Path target is not writable.
                    -- only happens if the application is installed.
                    if Is_Regular_File( filename ) and then not overwrite then
                        raise WRITE_EXCEPTION with "File already exists";
                    end if;
                    Create( file, Out_File, filename );
                exception
                    when WRITE_EXCEPTION => raise;
                    when others =>
                        -- attempt to write the world to the user directory
                        -- because the current working directory isn't
                        -- writable. only happens if the application is
                        -- installed.
                        if Is_Regular_File( Get_Application.Get_User_Directory & filename ) and then not overwrite then
                            raise WRITE_EXCEPTION with "File already exists";
                        end if;
                        if Make_Dir( Get_Application.Get_User_Directory ) then
                            Create( file, Out_File, Get_Application.Get_User_Directory & filename );
                        end if;
                end;
        end;

        sa := Stream( file );
        A_World'Output( sa, A_World(this) );
        this.Set_Property( "filename", Create( filename ) );
        pragma Debug( Dbg( "Wrote world <" & name & "> to disk at: " &
                           Ada.Streams.Stream_IO.Name( file ), D_GAME, INFO ) );
        Close( file );
    exception
        when WRITE_EXCEPTION =>
            -- file is not allowed to be overwritten
            raise;
        when Name_Error =>
            raise WRITE_EXCEPTION with "Bad filename: " & name;
        when e : others =>
            raise WRITE_EXCEPTION with "Error writing " & filename & ": " &
                                       Exception_Name( e ) & " - " &
                                       Exception_Message( e );
    end Save;

    ----------------------------------------------------------------------------

    procedure Set_Property( this  : not null access World_Object'Class;
                            name  : String;
                            value : Value_Ptr'Class ) is
    begin
        this.properties.Get.Set( name, value );
        if this.attached then
            Queue_World_Property_Changed( name, value );
            if name /= "filename" then
                -- "filename" isn't written to disk so it's not a modification
                Queue_World_Modified;
            end if;
        end if;
    end Set_Property;

    ----------------------------------------------------------------------------

    procedure Set_Tile( this   : not null access World_Object'Class;
                        layer  : Integer;
                        x, y   : Integer;
                        id     : Natural;
                        notify : Boolean := True ) is
        tile    : A_Tile;
        anmInfo : Animated_Info;
        pos     : Animated_Set.Cursor;
    begin
        if x >= 0 and then x < this.map.Get_Width and then
           y >= 0 and then y < this.map.Get_Height and then
           id /= this.map.Get_Tile_Id( layer, x, y )
        then
            tile := this.lib.Get_Tile( id );
            if tile /= null and then Cast_Int( tile.Get_Attribute( "delay" ), 0 ) > 0 then
                -- add/update animation info
                declare
                    anmDelay : constant Time_Span := Milliseconds( Cast_Int( tile.Get_Attribute( "delay" ), 0 ) );
                    frames   : List_Ptr := As_List( tile.Get_Attribute( "frames" ) );
                begin
                    if frames.Is_Null then
                        frames := Create_List;
                        frames.Get.Append( Create( id ) );
                    end if;
                    anmInfo := Animated_Info'(layer, x, y,
                                              Clock + anmDelay,
                                              anmDelay,
                                              frames);
                    this.animated.Include( anmInfo );
                end;
            else
                -- remove animation info, if any
                anmInfo := Animated_Info'(layer, x, y, Time_First, Time_Span_Zero, Lists.Nul);
                pos := this.animated.Find( anmInfo );
                if Animated_Set.Has_Element( pos ) then
                    this.animated.Delete( pos );
                end if;
            end if;
            this.map.Set_Tile( layer, x, y, id );
            if notify and then this.attached then
                Queue_Tile_Changed( layer, x, y, id );
                Queue_World_Modified;
            end if;
        end if;
    end Set_Tile;

    ----------------------------------------------------------------------------

    procedure Set_Tile( this   : not null access World_Object'Class;
                        layer  : Integer;
                        x, y   : Float;
                        id     : Natural;
                        notify : Boolean := True ) is
        tw : constant Positive := this.Tile_Width;
    begin
        this.Set_Tile( layer, Integer(x) / tw, Integer(y) / tw, id, notify );
    end Set_Tile;

    ----------------------------------------------------------------------------

    function Spawn_Entity( this   : not null access World_Object'Class;
                           id     : String;
                           x, y   : Float;
                           width,
                           height : Natural := 0;
                           xv, yv : Float := 0.0 ) return Entity_Id is
        entity : constant A_Entity := Entities.Allocate( id );
        w      : Natural := width;
        h      : Natural := height;
    begin
        if entity /= null then
            if w = 0 then
                w := entity.Get_Width;
            end if;
            if h = 0 then
                h := entity.Get_Height;
            end if;
            if width = 0 or else height = 0 then
                entity.Set_Size( w, h );
            end if;
            entity.Set_World( this );
            entity.Set_Location( x, y );
            entity.Set_Velocity_X( xv );
            entity.Set_Velocity_Y( yv );
            this.entities.Insert( entity.Get_Id, entity );
            if this.attached then
                Queue_Entity_Created( entity.Get_Id, entity.Get_Class_Name( full=>True ),
                                      entity.Is_Physical, entity.Is_Metaphysical,
                                      entity.Is_Clipped,  entity.Get_Width, entity.Get_Height,
                                      entity.Get_X, entity.Get_Y,
                                      entity.Get_XV, entity.Get_YV,
                                      entity.Get_Lib_Name, entity.Get_Frame,
                                      entity.Get_Attributes );
            end if;
            Queue_World_Modified;
            return entity.Get_Id;
        else
            Dbg( "Error spawning entity <" & id & "> at " &
                 Image( x ) & ", " & Image( y ),
                 D_GAME, ERROR );
            return INVALID_ID;
        end if;
    end Spawn_Entity;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access World_Object; time : Tick_Time ) is

        ------------------------------------------------------------------------

        procedure Tick_Animated is
            clk : constant Ada.Real_Time.Time := Clock;

            --------------------------------------------------------------------

            procedure Set_Tile( layer, x, y, id : Integer ) is
            begin
                this.map.Set_Tile( layer, x, y, id );
                Queue_Tile_Changed( layer, x, y, id );
            end Set_Tile;

            --------------------------------------------------------------------

            pos      : Animated_Set.Cursor := this.animated.First;
            nextpos  : Animated_Set.Cursor;
            anmInfo  : Animated_Info;
            anmDelay : Integer;
            id       : Integer;
            tile     : A_Tile;
        begin
            while Animated_Set.Has_Element( pos ) loop
                nextpos := Animated_Set.Next( pos );
                anmInfo := Animated_Set.Element( pos );
                if Is_Looping( anmInfo ) then
                    -- looped animation:
                    -- calculate frame using upTime modulus frame list size
                    id := Get_Frame( anmInfo, 1 + ((time.total / anmInfo.frameDelay) mod anmInfo.frames.Get.Length) );
                    if id /= this.map.Get_Tile_Id( anmInfo.layer, anmInfo.x, anmInfo.y ) then
                        Set_Tile( anmInfo.layer, anmInfo.x, anmInfo.y, id );
                    end if;
                else
                    -- one-shot animation:
                    -- move to next frame if the frame delay expired
                    if clk >= anmInfo.nextUpdate then
                        id := this.map.Get_Tile_Id( anmInfo.layer, anmInfo.x, anmInfo.y );
                        id := Cast_Int( this.lib.Get_Tile( id ).Get_Attribute( "next" ), 0 );
                        Set_Tile( anmInfo.layer, anmInfo.x, anmInfo.y, id );
                        tile := this.lib.Get_Tile( id );
                        anmDelay := Cast_Int( tile.Get_Attribute( "delay" ), 0 );
                        if anmDelay > 0 then
                            -- update the animated_info with the next update time
                            anmInfo.nextUpdate := clk + Milliseconds( anmDelay );
                            this.animated.Include( anmInfo );
                        else
                            -- the next frame isn't part of an animation, remove
                            -- the animated_info for this location
                            this.animated.Delete( pos );
                        end if;
                    end if;
                end if;
                pos := nextpos;
            end loop;
        end Tick_Animated;

        ------------------------------------------------------------------------

        procedure Tick_Entities is

            procedure Tick_Entity( pos : Entity_Map.Cursor ) is
            begin
                Entity_Map.Element( pos ).Tick( time );
            end Tick_Entity;

        begin
            this.entities.Iterate( Tick_Entity'Access );
        end Tick_Entities;

        ------------------------------------------------------------------------

    begin
        Tick_Animated;
        Tick_Entities;
    end Tick;

    ----------------------------------------------------------------------------

    function Tile_Width( this : not null access World_Object'Class ) return Positive is
    begin
        return this.map.Tile_Width;
    end Tile_Width;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_World ) is
    begin
        Delete( A_Object(this) );
    end Delete;

    ----------------------------------------------------------------------------

    function A_World_Input( stream : access Root_Stream_Type'Class ) return A_World is
        world : A_World := null;
        recg  : constant String := IDENTIFIER & ":" & World_Extension;
        recg2 : String(1..recg'Length);
        ver   : Integer;
    begin
        String'Read( stream, recg2 );
        if recg2 = recg then
            ver := Integer'Input( stream );
            if ver = READ_VERSION then
                if Boolean'Input( stream ) then
                    A_World'Read( stream, world );
                end if;
            else
                raise READ_EXCEPTION with "Unsupported file version " & Image( ver ) &
                                          "; current version is " & Image( READ_VERSION );
            end if;
        else
            raise READ_EXCEPTION with "Bad file format identifier";
        end if;
        return world;
    exception
        when others =>
            Delete( world );
            raise;
    end A_World_Input;

    ----------------------------------------------------------------------------

    procedure A_World_Output( stream : access Root_Stream_Type'Class; world : A_World ) is
    begin
        String'Write( stream, IDENTIFIER & ':' & World_Extension );
        Integer'Output( stream, WRITE_VERSION );
        Boolean'Output( stream, world /= null );
        if world /= null then
            A_World'Write( stream, world );
        end if;
    end A_World_Output;

    ----------------------------------------------------------------------------

    procedure A_World_Read( stream : access Root_Stream_Type'Class; world : out A_World ) is
    begin
        world := new World_Object'Class'(World_Object'Class'Input( stream ));
    exception
        when e : Tag_Error =>
            declare
                msg : constant String := Exception_Message( e );
            begin
                raise Constraint_Error with "Found unknown world class " &
                      "<" & msg((Index( msg, ".", Backward ) + 1)..msg'Last) & ">";
            end;
    end A_World_Read;

    ----------------------------------------------------------------------------

    procedure A_World_Write( stream : access Root_Stream_Type'Class; world : A_World ) is
    begin
        World_Object'Class'Output( stream, world.all );
    end A_World_Write;

    --==========================================================================

    procedure Register_Domain( domain : String ) is
    begin
        domains.Include( domain );
    end Register_Domain;

    ----------------------------------------------------------------------------

    function Valid_Domain( domain : String ) return Boolean is
    begin
        return domains.Contains( domain );
    end Valid_Domain;

    ----------------------------------------------------------------------------

    function World_Extension return String is
    begin
        return "world";
    end World_Extension;

end Worlds;
