--
-- Copyright (c) 2013 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--
--------------------------------------------------------------------------------

--
-- Hyllian's xBR v3.3
--
-- These scaling filters are based on the xBR scaling algorithm, as described by
-- Hyllian/Jararaca - sergiogdb@gmail.com.
--

with Allegro.Bitmaps.Locking;           use Allegro.Bitmaps.Locking;
with Allegro.Color;                     use Allegro.Color;
with Interfaces;                        use Interfaces;
with System;                            use System;
with System.Storage_Elements;           use System.Storage_Elements;

package body xBR is

    type YUV_Record is
        record
            weighted : Unsigned_32;
            eq       : Unsigned_32;
        end record;

    type YUV_Array is array (Integer range <>) of YUV_Record;

--    type Uint16_Array is array (Integer range<>) of Unsigned_16;

    type Uint8_Array is array (Integer range<>) of Unsigned_8;

    ----------------------------------------------------------------------------

    initialized : Boolean := False;

    -- for the slower, more accurate version
    RGBtoYUV : YUV_Array(0..65535);
    -- for the fast version
--    RGBtoYUV : Uint16_Array(0..65535);

    -- map a 5 bit space to an 8 bit space (output := input * 255 / 32)
    tbl_5_to_8 : constant Uint8_Array(0..31) :=
    (
          0,   8,  16,  25,  33,  41,  49,  58,
         66,  74,  82,  90,  99, 107, 115, 123,
        132, 140, 148, 156, 165, 173, 181, 189,
        197, 206, 214, 222, 230, 239, 247, 255
    );

    -- map a 6 bit space to an 8 bit space (output := input * 255 / 64)
    tbl_6_to_8 : constant Uint8_Array(0..63) :=
    (
         0,    4,   8,  12,  16,  20,  24,  28,
         32,  36,  40,  45,  49,  53,  57,  61,
         65,  69,  73,  77,  81,  85,  89,  93,
         97, 101, 105, 109, 113, 117, 121, 125,
        130, 134, 138, 142, 146, 150, 154, 158,
        162, 166, 170, 174, 178, 182, 186, 190,
        194, 198, 202, 206, 210, 215, 219, 223,
        227, 231, 235, 239, 243, 247, 251, 255
    );

    RED_MASK      : constant Unsigned_16 := 2#11111_000000_00000#;
    GREEN_MASK    : constant Unsigned_16 := 2#00000_111111_00000#;
    BLUE_MASK     : constant Unsigned_16 := 2#00000_000000_11111#;
    RED_BLUE_MASK : constant Unsigned_16 := RED_MASK or BLUE_MASK;

    RGB_MASK      : constant Unsigned_32 := 2#00000_111111_00000_11111_000000_11111#;
    LB_MASK       : constant Unsigned_16 := 2#11110_111110_11110#;

    ----------------------------------------------------------------------------

    procedure Blend_X_32( bg      : in out Unsigned_16;
                          fg      : Unsigned_16;
                          opacity : Unsigned_32 ) is
        fg2 : Unsigned_32 := Unsigned_32(fg);
        bg2 : Unsigned_32 := Unsigned_32(bg);
    begin
        bg2 := (bg2 or Shift_Left( bg2, 16 )) and RGB_MASK;
        fg2 := (fg2 or Shift_Left( fg2, 16 )) and RGB_MASK;
        bg2 := (Shift_Right( (fg2 - bg2) * opacity, 5 ) + bg2) and RGB_MASK;
        bg := Unsigned_16((bg2 or Shift_Right( bg2, 16 )) and 16#0000_FFFF#);
    end Blend_X_32;
    pragma Inline_Always( Blend_X_32 );

    ----------------------------------------------------------------------------

    -- Blend 'src' to 'dst' with 87.5% opacity (224 / 256 or 28 / 32)
    procedure Blend_224( bg : in out Unsigned_16; fg : Unsigned_16 ) is
    begin
        Blend_X_32( bg, fg, 28 );
    end Blend_224;
    pragma Inline_Always( Blend_224 );

    ----------------------------------------------------------------------------

    -- Blend 'src' to 'dst' with 75% opacity (192 / 256 or 24 / 32)
    procedure Blend_192( bg : in out Unsigned_16; fg : Unsigned_16 ) is
    begin
        Blend_X_32( bg, fg, 24 );
    end Blend_192;
    pragma Inline_Always( Blend_192 );

    ----------------------------------------------------------------------------

    -- Blend 'src' to 'dst' with 50% opacity (128 / 256)
    procedure Blend_128( bg : in out Unsigned_16; fg : Unsigned_16 ) is
    begin
        bg := Shift_Right( fg and LB_MASK, 1 ) + Shift_Right( bg and LB_MASK, 1 );
    end Blend_128;
    pragma Inline_Always( Blend_128 );

    ----------------------------------------------------------------------------

    -- Blend 'src' to 'dst' with 25% opacity (64 / 256)
    procedure Blend_64( bg : in out Unsigned_16; fg : Unsigned_16 ) is
    begin
        bg :=
            (RED_BLUE_MASK and Unsigned_16(16#0000_FFFF# and
                                           (Unsigned_32(bg and RED_BLUE_MASK) +
                                            Shift_Right( Unsigned_32(fg and RED_BLUE_MASK) -
                                                         Unsigned_32(bg and RED_BLUE_MASK), 2 )))) or
            (GREEN_MASK and ((bg and GREEN_MASK) +
                              Shift_Right( (fg and GREEN_MASK) -
                                           (bg and GREEN_MASK), 2 )));
    end Blend_64;
    pragma Inline_Always( Blend_64 );

    ----------------------------------------------------------------------------

    -- Blend 'src' to 'dst' with 12.5% opacity (32 / 256)
    procedure Blend_32( bg : in out Unsigned_16; fg : Unsigned_16 ) is
    begin
        bg :=
            (RED_BLUE_MASK and Unsigned_16(16#0000_FFFF# and
                                           (Unsigned_32(bg and RED_BLUE_MASK) +
                                            Shift_Right( Unsigned_32(fg and RED_BLUE_MASK) -
                                                         Unsigned_32(bg and RED_BLUE_MASK), 3 )))) or
            (GREEN_MASK and ((bg and GREEN_MASK) +
                              Shift_Right( (fg and GREEN_MASK) -
                                           (bg and GREEN_MASK), 3 )));
    end Blend_32;
    pragma Inline_Always( Blend_32 );

    ----------------------------------------------------------------------------

    -- This is the slower, more accurate version.
    function df( i, j : Unsigned_16 ) return Unsigned_32 is
        rgbDiff : Unsigned_16;
    begin
        rgbDiff :=
            (RED_BLUE_MASK and Unsigned_16(abs (Integer(i and RED_BLUE_MASK) -
                                                Integer(j and RED_BLUE_MASK)) )) or
            (GREEN_MASK and Unsigned_16(abs (Integer(i and GREEN_MASK) -
                                             Integer(j and GREEN_MASK)) ));

        return RGBtoYUV(Integer(rgbDiff)).weighted;
    end df;
    pragma Inline_Always( df );

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    -- This is the fast version.
--    function df( a, b : Unsigned_16 ) return Unsigned_32 is
--    begin
--        return Unsigned_32(abs (Integer(RGBtoYUV(Integer(a))) - Integer(RGBtoYUV(Integer(b)))));
--    end df;
--    pragma Inline_Always( df );

    ----------------------------------------------------------------------------

    -- This is the slower, more accurate version.
    function eq( i, j : Unsigned_16 ) return Boolean is
        rgbDiff : Unsigned_16;
    begin
        rgbDiff :=
            (RED_BLUE_MASK and Unsigned_16(abs (Integer(i and RED_BLUE_MASK) -
                                                Integer(j and RED_BLUE_MASK)) )) or
            (GREEN_MASK and Unsigned_16(abs (Integer(i and GREEN_MASK) -
                                             Integer(j and GREEN_MASK)) ));

        return RGBtoYUV(Integer(rgbDiff)).eq /= 0;
    end eq;
    pragma Inline_Always( eq );

    -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    -- This is the fast version.
--    function eq( a, b : Unsigned_16 ) return Boolean is
--    begin
--        return df( a, b ) < 155;
--    end eq;
--    pragma Inline_Always( eq );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    procedure Initialize is
        r, g, b : Unsigned_16;
        y, u, v : Float;
    begin
        if initialized then
            return;
        end if;
        initialized := True;

        -- 16 bit RGB 565 format
        for c in RGBtoYUV'Range loop
            r := Unsigned_16(tbl_5_to_8( Integer(Shift_Right( Unsigned_16(c) and   RED_MASK, 11 )) ));
            g := Unsigned_16(tbl_6_to_8( Integer(Shift_Right( Unsigned_16(c) and GREEN_MASK,  5 )) ));
            b := Unsigned_16(tbl_5_to_8( Integer(             Unsigned_16(c) and  BLUE_MASK      ) ));

            y := abs( 0.299 * Float(r) + 0.587 * Float(g) + 0.114 * Float(b) );
            u := abs(-0.169 * Float(r) - 0.331 * Float(g) + 0.500 * Float(b) );
            v := abs( 0.500 * Float(r) - 0.419 * Float(g) - 0.081 * Float(b) );
            RGBtoYUV(c).weighted := Unsigned_32(y * 48.0 + u * 7.0 + v * 6.0 );
            if y < 48.0 and then u < 7.0 and then v < 6.0 then
                RGBtoYUV(c).eq := 1;
            else
                RGBtoYUV(c).eq := 0;
            end if;

            -- for the fast version
--            y := Shift_Left( r, 4 ) + Shift_Left( g, 5 ) + Shift_Left(  b, 2 );
--            u :=            -r      - Shift_Left( g, 1 ) + Shift_Left(  b, 2 );
--            v := Shift_Left( r, 1 ) - Shift_Left( g, 1 ) - Shift_Right( b, 1 );
--            RGBtoYUV(c) := y + u + v;
        end loop;
    end Initialize;

    ----------------------------------------------------------------------------

    -- 2xBR v3.3 (a/b/c)
    procedure xBR2( input, output : not null A_Allegro_Bitmap;
                    destX, destY  : Integer ) is
        -- used by nested functions
        Eaddr : Address;

        ------------------------------------------------------------------------

        procedure Filter( PE, PI, PH, PF, PG, PC, PD, PB, F4, I4, H5, I5 : Unsigned_16;
                          N1, N2, N3 : Integer ) is

            E1     : Unsigned_16; for E1'Address use Eaddr + Storage_Offset(N1 * 2);
            E2     : Unsigned_16; for E2'Address use Eaddr + Storage_Offset(N2 * 2);
            E3     : Unsigned_16; for E3'Address use Eaddr + Storage_Offset(N3 * 2);

            ex2,
            ex3    : Boolean;
            pixel  : Unsigned_16;
            e, i   : Unsigned_32;
            ke, ki : Unsigned_32;
        begin
            if PE = PH or else PE = PF then
                return;
            end if;

            e := df( PE, PC ) + df( PE, PG ) + df( PI, H5 ) + df( PI, F4 ) + Shift_Left( df( PH, PF ), 2 );
            i := df( PH, PD ) + df( PH, I5 ) + df( PF, I4 ) + df( PF, PB ) + Shift_Left( df( PE, PI ), 2 );
            -- version 'a' (square)
            if e < i and then
                (
                    (not eq( PF, PB ) and then not eq( PF, PC )) or else
                    (not eq( PH, PD ) and then not eq( PH, PG )) or else
                    (
                        eq( PE, PI ) and then
                        (
                            (not eq( PF, F4 ) and then not eq( PF, I4 )) or else
                            (not eq( PH, H5 ) and then not eq( PH, I5 ))
                        )
                    ) or else
                    eq( PE, PG ) or else
                    eq( PE, PC )
                )

            -- version 'b' (semi-rounded)
--            if e < i and then
--                (
--                    (not eq( PF, PB ) and then not eq( PH, PD )) or else
--                    (
--                        eq( PE, PI ) and then
--                        (not eq( PF, I4 ) and then not eq( PH, I5 ))
--                    ) or else
--                    eq( PE, PG ) or else
--                    eq( PE, PC )
--                )

            -- version 'c' (rounded)
--            if e < i
            then
                ke := df( PF, PG );
                ki := df( PH, PC );
                ex2 := PE /= PC and then PB /= PC;
                ex3 := PE /= PG and then PD /= PG;
                if df( PE, PF ) <= df( PE, PH ) then
                    pixel := PF;
                else
                    pixel := PH;
                end if;

                if ex3 and then ex2 and then Shift_Left( ke, 1 ) <= ki and then ke >= Shift_Left( ki, 1 ) then
                    -- Left + Up 2
                    Blend_224( E3, pixel );
                    Blend_64( E2, pixel );
                    E1 := E2;
                elsif ex3 and then Shift_Left( ke, 1 ) <= ki then
                    -- Left 2
                    Blend_192( E3, pixel );
                    Blend_64( E2, pixel );
                elsif ex2 and then ke >= Shift_Left( ki, 1 ) then
                    -- Up 2
                    Blend_192( E3, pixel );
                    Blend_64( E1, pixel );
                else
                    -- Diagonal
                    Blend_128( E3, pixel );
                end if;
            elsif e <= i then
                if df( PE, PF ) <= df( PE, PH ) then
                    Blend_64( E3, PF );
                else
                    Blend_64( E3, PH );
                end if;
            end if;
        end Filter;
        pragma Inline_Always( Filter );

        ------------------------------------------------------------------------

        -- used outside the loops
        SCALE    : constant := 2;
        src, dst : A_Allegro_Locked_Region;
        nl1      : Integer;

        -- used inside the Y loop
        srcHeight : Integer;
        EaddrY    : Address;
        EaddrYinc : Storage_Offset;
        EaddrXinc : constant Storage_Offset := SCALE * 2;
        sa2Y      : Address;
        sa0, sa1,
        sa2, sa3,
        sa4       : Address;

        -- used inside the X loop
        srcWidth : Integer;
        pprev,
        pprev2   : Unsigned_8;
        C1addr,
        PCaddr,
        PFaddr,
        PIaddr,
        I5addr   : Address;
        C4addr,
        F4addr,
        I4addr   : Address;
    begin
        if not initialized then
            return;
        end if;

        srcWidth := Al_Get_Bitmap_Width( input );
        srcHeight := Al_Get_Bitmap_Height( input );
        if Al_Get_Bitmap_Width( output ) < destX + srcWidth * SCALE or else
           Al_Get_Bitmap_Height( output ) < destY + srcHeight * SCALE
        then
            -- Invalid destination bitmap size
            return;
        end if;

        src := Al_Lock_Bitmap( input, ALLEGRO_PIXEL_FORMAT_RGB_565, ALLEGRO_LOCK_READONLY );
        if src = null then
            -- Failed to lock input bitmap
            return;
        end if;

        if Al_Get_Bitmap_Width( output ) = srcWidth * SCALE and then
           Al_Get_Bitmap_Height( output) = srcHeight * SCALE
        then
            -- optimization when we're filling the entire destination bitmap.
            -- this won't copy the contents of 'output' down from the video
            -- card, saving some time. it starts as a completely black bitmap.
            dst := Al_Lock_Bitmap( output, ALLEGRO_PIXEL_FORMAT_RGB_565, ALLEGRO_LOCK_WRITEONLY );
        else
            -- there are pixels in 'output' that we're not overwriting, so we
            -- need to grab the image from the video card.
            dst := Al_Lock_Bitmap( output, ALLEGRO_PIXEL_FORMAT_RGB_565, ALLEGRO_LOCK_READWRITE );
        end if;
        if dst = null then
            -- Failed to lock destination bitmap
            Al_Unlock_Bitmap( input );
            return;
        end if;

        nl1 := dst.pitch / 2;

        EaddrYinc := Storage_Offset(dst.pitch * SCALE);
        EaddrY := dst.data +
                  Storage_Offset(dst.pitch * destY) +
                  Storage_Offset(destX * 2);                  -- output pos
        sa2Y := src.data - Storage_Offset(4);                 -- input pos
        for y in 0..srcHeight-1 loop
            -- output pixel (top left of 2x2)
            Eaddr := EaddrY;

            -- input pixel (middle)
            sa2 := sa2Y;
            -- up one
            sa1 := sa2 - Storage_Offset(src.pitch);
            -- up two
            sa0 := sa1 - Storage_Offset(src.pitch);
            -- down one
            sa3 := sa2 + Storage_Offset(src.pitch);
            -- down two
            sa4 := sa3 + Storage_Offset(src.pitch);

            if y <= 1 then
                sa0 := sa1;
                if y = 0 then
                    sa0 := sa2;
                    sa1 := sa2;
                end if;
            end if;

            if y >= srcHeight - 2 then
                sa4 := sa3;
                if y = srcHeight - 1 then
                    sa4 := sa2;
                    sa3 := sa2;
                end if;
            end if;

            pprev := 2;
            pprev2 := 2;

            for x in 0..srcWidth-1 loop
                if x >= srcWidth - 2 then
                    if x = srcWidth - 1 then
                        C1addr := sa0 + Storage_Offset(2 * 2);
                        PCaddr := sa1 + Storage_Offset(2 * 2);
                        PFaddr := sa2 + Storage_Offset(2 * 2);
                        PIaddr := sa3 + Storage_Offset(2 * 2);
                        I5addr := sa4 + Storage_Offset(2 * 2);
                    else
                        C1addr := sa0 + Storage_Offset(3 * 2);
                        PCaddr := sa1 + Storage_Offset(3 * 2);
                        PFaddr := sa2 + Storage_Offset(3 * 2);
                        PIaddr := sa3 + Storage_Offset(3 * 2);
                        I5addr := sa4 + Storage_Offset(3 * 2);
                    end if;
                    C4addr := PCaddr;
                    F4addr := PFaddr;
                    I4addr := PIaddr;
                else
                    C1addr := sa0 + Storage_Offset(3 * 2);
                    PCaddr := sa1 + Storage_Offset(3 * 2);
                    PFaddr := sa2 + Storage_Offset(3 * 2);
                    PIaddr := sa3 + Storage_Offset(3 * 2);
                    I5addr := sa4 + Storage_Offset(3 * 2);

                    C4addr := sa1 + Storage_Offset(4 * 2);
                    F4addr := sa2 + Storage_Offset(4 * 2);
                    I4addr := sa3 + Storage_Offset(4 * 2);
                end if;

                declare
                    B1 : Unsigned_16; for B1'Address use sa0 + Storage_Offset(2 * 2);
                    PB : Unsigned_16; for PB'Address use sa1 + Storage_Offset(2 * 2);
                    PE : Unsigned_16; for PE'Address use sa2 + Storage_Offset(2 * 2);
                    PH : Unsigned_16; for PH'Address use sa3 + Storage_Offset(2 * 2);
                    H5 : Unsigned_16; for H5'Address use sa4 + Storage_Offset(2 * 2);

                    A1 : Unsigned_16; for A1'Address use sa0 + Storage_Offset(pprev * 2);
                    PA : Unsigned_16; for PA'Address use sa1 + Storage_Offset(pprev * 2);
                    PD : Unsigned_16; for PD'Address use sa2 + Storage_Offset(pprev * 2);
                    PG : Unsigned_16; for PG'Address use sa3 + Storage_Offset(pprev * 2);
                    G5 : Unsigned_16; for G5'Address use sa4 + Storage_Offset(pprev * 2);

                    A0 : Unsigned_16; for A0'Address use sa1 + Storage_Offset(pprev2 * 2);
                    D0 : Unsigned_16; for D0'Address use sa2 + Storage_Offset(pprev2 * 2);
                    G0 : Unsigned_16; for G0'Address use sa3 + Storage_Offset(pprev2 * 2);

                    C1 : Unsigned_16; for C1'Address use C1addr;
                    PC : Unsigned_16; for PC'Address use PCaddr;
                    PF : Unsigned_16; for PF'Address use PFaddr;
                    PI : Unsigned_16; for PI'Address use PIaddr;
                    I5 : Unsigned_16; for I5'Address use I5addr;

                    C4 : Unsigned_16; for C4'Address use C4addr;
                    F4 : Unsigned_16; for F4'Address use F4addr;
                    I4 : Unsigned_16; for I4'Address use I4addr;

                    -- E0 E1
                    -- E2 E3
                    E0 : Unsigned_16; for E0'Address use Eaddr + Storage_Offset(0 * 2);
                    E1 : Unsigned_16; for E1'Address use Eaddr + Storage_Offset(1 * 2);
                    E2 : Unsigned_16; for E2'Address use Eaddr + Storage_Offset((nl1 + 0) * 2);
                    E3 : Unsigned_16; for E3'Address use Eaddr + Storage_Offset((nl1 + 1) * 2);
                begin
                    E0 := PE; E1 := PE;
                    E2 := PE; E3 := PE;

                    Filter( PE, PI, PH, PF, PG, PC, PD, PB, F4, I4, H5, I5,     1,   nl1, nl1+1 );
                    Filter( PE, PC, PF, PB, PI, PA, PH, PD, B1, C1, F4, C4,     0, nl1+1,     1 );
                    Filter( PE, PA, PB, PD, PC, PG, PF, PH, D0, A0, B1, A1,   nl1,     1,     0 );
                    Filter( PE, PG, PD, PH, PA, PI, PB, PF, H5, G5, D0, G0, nl1+1,     0,   nl1 );
                end;

                sa0 := sa0 + Storage_Offset(1 * 2);
                sa1 := sa1 + Storage_Offset(1 * 2);
                sa2 := sa2 + Storage_Offset(1 * 2);
                sa3 := sa3 + Storage_Offset(1 * 2);
                sa4 := sa4 + Storage_Offset(1 * 2);

                Eaddr := Eaddr + EaddrXinc;

                if pprev2 > 0 then
                    pprev2 := pprev2 - 1;
                    pprev := 1;
                end if;
            end loop;

            EaddrY := EaddrY + EaddrYinc;
            sa2Y := sa2Y + Storage_Offset(src.pitch);
        end loop;

        Al_Unlock_Bitmap( input );
        Al_Unlock_Bitmap( output );
    exception
        when others =>
            if Al_Is_Bitmap_Locked( input ) then
                Al_Unlock_Bitmap( input );
            end if;
            if Al_Is_Bitmap_Locked( output) then
                Al_Unlock_Bitmap( output );
            end if;
            raise;
    end xBR2;

    ----------------------------------------------------------------------------

    -- 3xBR v3.3 (a/b/c)
    procedure xBR3( input, output : not null A_Allegro_Bitmap;
                    destX, destY  : Integer ) is
        -- used by nested functions
        Eaddr : Address;

        ------------------------------------------------------------------------

        procedure Filter( PE, PI, PH, PF, PG, PC, PD, PB, F4, I4, H5, I5 : Unsigned_16;
                          N2, N5, N6, N7, N8 : Integer ) is

            E2     : Unsigned_16; for E2'Address use Eaddr + Storage_Offset(N2 * 2);
            E5     : Unsigned_16; for E5'Address use Eaddr + Storage_Offset(N5 * 2);
            E6     : Unsigned_16; for E6'Address use Eaddr + Storage_Offset(N6 * 2);
            E7     : Unsigned_16; for E7'Address use Eaddr + Storage_Offset(N7 * 2);
            E8     : Unsigned_16; for E8'Address use Eaddr + Storage_Offset(N8 * 2);

            ex2,
            ex3    : Boolean;
            pixel  : Unsigned_16;
            e, i   : Unsigned_32;
            ke, ki : Unsigned_32;
        begin
            if PE = PH or else PE = PF then
                return;
            end if;

            e := df( PE, PC ) + df( PE, PG ) + df( PI, H5 ) + df( PI, F4 ) + Shift_Left( df( PH, PF ), 2 );
            i := df( PH, PD ) + df( PH, I5 ) + df( PF, I4 ) + df( PF, PB ) + Shift_Left( df( PE, PI ), 2 );
            -- version 'a' (square)
            if e < i and then
                (
                    (not eq( PF, PB ) and then not eq( PF, PC )) or else
                    (not eq( PH, PD ) and then not eq( PH, PG )) or else
                    (
                        eq( PE, PI ) and then
                        (
                            (not eq( PF, F4 ) and then not eq( PF, I4 )) or else
                            (not eq( PH, H5 ) and then not eq( PH, I5 ))
                        )
                    ) or else
                    eq( PE, PG ) or else
                    eq( PE, PC )
                )

            -- version 'b' (semi-rounded)
--            if e < i and then
--                (
--                    (not eq( PF, PB ) and then not eq( PH, PD )) or else
--                    (
--                        eq( PE, PI ) and then
--                        (not eq( PF, I4 ) and then not eq( PH, I5 ))
--                    ) or else
--                    eq( PE, PG ) or else
--                    eq( PE, PC )
--                )

            -- version 'c' (rounded)
--            if e < i
            then
                ke := df( PF, PG );
                ki := df( PH, PC );
                ex2 := PE /= PC and then PB /= PC;
                ex3 := PE /= PG and then PD /= PG;
                if df( PE, PF ) <= df( PE, PH ) then
                    pixel := PF;
                else
                    pixel := PH;
                end if;

                if ex3 and then ex2 and then Shift_Left( ke, 1 ) <= ki and then ke >= Shift_Left( ki, 1 ) then
                    -- Left + Up 2
                    Blend_192( E7, pixel );
                    Blend_64( E6, pixel );
                    E5 := E7;
                    E2 := E6;
                    E8 := pixel;
                elsif ex3 and then Shift_Left( ke, 1 ) <= ki then
                    -- Left 2
                    Blend_192( E7, pixel );
                    Blend_64( E5, pixel );
                    Blend_64( E6, pixel );
                    E8 := pixel;
                elsif ex2 and then ke >= Shift_Left( ki, 1 ) then
                    -- Up 2
                    Blend_192( E5, pixel );
                    Blend_64( E7, pixel );
                    Blend_64( E2, pixel );
                    E8 := pixel;
                else
                    -- Diagonal
                    Blend_224( E8, pixel );
                    Blend_32( E5, pixel );
                    Blend_32( E7, pixel );
                end if;
            elsif e <= i then
                if df( PE, PF ) <= df( PE, PH ) then
                    Blend_64( E8, PF );
                else
                    Blend_64( E8, PH );
                end if;
            end if;
        end Filter;
        pragma Inline_Always( Filter );

        ------------------------------------------------------------------------

        -- used outside the loops
        SCALE     : constant := 3;
        src, dst  : A_Allegro_Locked_Region;
        nl1, nl2  : Integer;

        -- used inside the Y loop
        srcHeight : Integer;
        EaddrY    : Address;
        EaddrYinc : Storage_Offset;
        EaddrXinc : constant Storage_Offset := SCALE * 2;
        sa2Y      : Address;
        sa0, sa1,
        sa2, sa3,
        sa4       : Address;

        -- used inside the X loop
        srcWidth : Integer;
        pprev,
        pprev2   : Unsigned_8;
        C1addr,
        PCaddr,
        PFaddr,
        PIaddr,
        I5addr   : Address;
        C4addr,
        F4addr,
        I4addr   : Address;
    begin
        if not initialized then
            return;
        end if;

        srcWidth := Al_Get_Bitmap_Width( input );
        srcHeight := Al_Get_Bitmap_Height( input );
        if Al_Get_Bitmap_Width( output ) < destX + srcWidth * SCALE or else
           Al_Get_Bitmap_Height( output ) < destY + srcHeight * SCALE
        then
            -- Invalid destination bitmap size
            return;
        end if;

        src := Al_Lock_Bitmap( input, ALLEGRO_PIXEL_FORMAT_RGB_565, ALLEGRO_LOCK_READONLY );
        if src = null then
            -- Failed to lock input bitmap
            return;
        end if;

        if Al_Get_Bitmap_Width( output ) = srcWidth * SCALE and then
           Al_Get_Bitmap_Height( output) = srcHeight * SCALE
        then
            -- optimization when we're filling the entire destination bitmap.
            -- this won't copy the contents of 'output' down from the video
            -- card, saving some time. it starts as a completely black bitmap.
            dst := Al_Lock_Bitmap( output, ALLEGRO_PIXEL_FORMAT_RGB_565, ALLEGRO_LOCK_WRITEONLY );
        else
            -- there are pixels in 'output' that we're not overwriting, so we
            -- need to grab the image from the video card.
            dst := Al_Lock_Bitmap( output, ALLEGRO_PIXEL_FORMAT_RGB_565, ALLEGRO_LOCK_READWRITE );
        end if;
        if dst = null then
            -- Failed to lock destination bitmap
            Al_Unlock_Bitmap( input );
            return;
        end if;

        nl1 := dst.pitch / 2;    -- pixel offset of next line
        nl2 := nl1 + nl1;        -- pixel offset of line after nl1

        EaddrYinc := Storage_Offset(dst.pitch * SCALE);
        EaddrY := dst.data +
                  Storage_Offset(dst.pitch * destY) +
                  Storage_Offset(destX * 2);                  -- output pos
        sa2Y := src.data - Storage_Offset(4);                 -- input pos
        for y in 0..srcHeight-1 loop
            -- output pixel (top left of 3x3)
            Eaddr := EaddrY;

            -- input pixel (middle)
            sa2 := sa2Y;
            -- up one
            sa1 := sa2 - Storage_Offset(src.pitch);
            -- up two
            sa0 := sa1 - Storage_Offset(src.pitch);
            -- down one
            sa3 := sa2 + Storage_Offset(src.pitch);
            -- down two
            sa4 := sa3 + Storage_Offset(src.pitch);

            if y <= 1 then
                sa0 := sa1;
                if y = 0 then
                    sa0 := sa2;
                    sa1 := sa2;
                end if;
            end if;

            if y >= srcHeight - 2 then
                sa4 := sa3;
                if y = srcHeight - 1 then
                    sa4 := sa2;
                    sa3 := sa2;
                end if;
            end if;

            pprev := 2;
            pprev2 := 2;

            for x in 0..srcWidth-1 loop
                if x >= srcWidth - 2 then
                    if x = srcWidth - 1 then
                        C1addr := sa0 + Storage_Offset(2 * 2);
                        PCaddr := sa1 + Storage_Offset(2 * 2);
                        PFaddr := sa2 + Storage_Offset(2 * 2);
                        PIaddr := sa3 + Storage_Offset(2 * 2);
                        I5addr := sa4 + Storage_Offset(2 * 2);
                    else
                        C1addr := sa0 + Storage_Offset(3 * 2);
                        PCaddr := sa1 + Storage_Offset(3 * 2);
                        PFaddr := sa2 + Storage_Offset(3 * 2);
                        PIaddr := sa3 + Storage_Offset(3 * 2);
                        I5addr := sa4 + Storage_Offset(3 * 2);
                    end if;
                    C4addr := PCaddr;
                    F4addr := PFaddr;
                    I4addr := PIaddr;
                else
                    C1addr := sa0 + Storage_Offset(3 * 2);
                    PCaddr := sa1 + Storage_Offset(3 * 2);
                    PFaddr := sa2 + Storage_Offset(3 * 2);
                    PIaddr := sa3 + Storage_Offset(3 * 2);
                    I5addr := sa4 + Storage_Offset(3 * 2);

                    C4addr := sa1 + Storage_Offset(4 * 2);
                    F4addr := sa2 + Storage_Offset(4 * 2);
                    I4addr := sa3 + Storage_Offset(4 * 2);
                end if;

                declare
                    B1 : Unsigned_16; for B1'Address use sa0 + Storage_Offset(2 * 2);
                    PB : Unsigned_16; for PB'Address use sa1 + Storage_Offset(2 * 2);
                    PE : Unsigned_16; for PE'Address use sa2 + Storage_Offset(2 * 2);
                    PH : Unsigned_16; for PH'Address use sa3 + Storage_Offset(2 * 2);
                    H5 : Unsigned_16; for H5'Address use sa4 + Storage_Offset(2 * 2);

                    A1 : Unsigned_16; for A1'Address use sa0 + Storage_Offset(pprev * 2);
                    PA : Unsigned_16; for PA'Address use sa1 + Storage_Offset(pprev * 2);
                    PD : Unsigned_16; for PD'Address use sa2 + Storage_Offset(pprev * 2);
                    PG : Unsigned_16; for PG'Address use sa3 + Storage_Offset(pprev * 2);
                    G5 : Unsigned_16; for G5'Address use sa4 + Storage_Offset(pprev * 2);

                    A0 : Unsigned_16; for A0'Address use sa1 + Storage_Offset(pprev2 * 2);
                    D0 : Unsigned_16; for D0'Address use sa2 + Storage_Offset(pprev2 * 2);
                    G0 : Unsigned_16; for G0'Address use sa3 + Storage_Offset(pprev2 * 2);

                    C1 : Unsigned_16; for C1'Address use C1addr;
                    PC : Unsigned_16; for PC'Address use PCaddr;
                    PF : Unsigned_16; for PF'Address use PFaddr;
                    PI : Unsigned_16; for PI'Address use PIaddr;
                    I5 : Unsigned_16; for I5'Address use I5addr;

                    C4 : Unsigned_16; for C4'Address use C4addr;
                    F4 : Unsigned_16; for F4'Address use F4addr;
                    I4 : Unsigned_16; for I4'Address use I4addr;

                    -- E0 E1 E2
                    -- E3 E4 E5
                    -- E6 E7 E8
                    E0 : Unsigned_16; for E0'Address use Eaddr;
                    E1 : Unsigned_16; for E1'Address use Eaddr + Storage_Offset(1 * 2);
                    E2 : Unsigned_16; for E2'Address use Eaddr + Storage_Offset(2 * 2);
                    E3 : Unsigned_16; for E3'Address use Eaddr + Storage_Offset( nl1      * 2);
                    E4 : Unsigned_16; for E4'Address use Eaddr + Storage_Offset((nl1 + 1) * 2);
                    E5 : Unsigned_16; for E5'Address use Eaddr + Storage_Offset((nl1 + 2) * 2);
                    E6 : Unsigned_16; for E6'Address use Eaddr + Storage_Offset( nl2      * 2);
                    E7 : Unsigned_16; for E7'Address use Eaddr + Storage_Offset((nl2 + 1) * 2);
                    E8 : Unsigned_16; for E8'Address use Eaddr + Storage_Offset((nl2 + 2) * 2);
                begin
                    E0 := PE; E1 := PE; E2 := PE;
                    E3 := PE; E4 := PE; E5 := PE;
                    E6 := PE; E7 := PE; E8 := PE;

                    Filter( PE, PI, PH, PF, PG, PC, PD, PB, F4, I4, H5, I5,     2, nl1+2,   nl2, nl2+1, nl2+2 );
                    Filter( PE, PC, PF, PB, PI, PA, PH, PD, B1, C1, F4, C4,     0,     1, nl2+2, nl1+2,     2 );
                    Filter( PE, PA, PB, PD, PC, PG, PF, PH, D0, A0, B1, A1,   nl2,   nl1,     2,     1,     0 );
                    Filter( PE, PG, PD, PH, PA, PI, PB, PF, H5, G5, D0, G0, nl2+2, nl2+1,     0,   nl1,   nl2 );
                end;

                sa0 := sa0 + Storage_Offset(1 * 2);
                sa1 := sa1 + Storage_Offset(1 * 2);
                sa2 := sa2 + Storage_Offset(1 * 2);
                sa3 := sa3 + Storage_Offset(1 * 2);
                sa4 := sa4 + Storage_Offset(1 * 2);

                Eaddr := Eaddr + EaddrXinc;

                if pprev2 > 0 then
                    pprev2 := pprev2 - 1;
                    pprev := 1;
                end if;
            end loop;

            EaddrY := EaddrY + EaddrYinc;
            sa2Y := sa2Y + Storage_Offset(src.pitch);
        end loop;

        Al_Unlock_Bitmap( input );
        Al_Unlock_Bitmap( output );
    exception
        when others =>
            if Al_Is_Bitmap_Locked( input ) then
                Al_Unlock_Bitmap( input );
            end if;
            if Al_Is_Bitmap_Locked( output) then
                Al_Unlock_Bitmap( output );
            end if;
            raise;
    end xBR3;

    ----------------------------------------------------------------------------

    -- 4xBR v3.3 (a/b/c)
    procedure xBR4( input, output : not null A_Allegro_Bitmap;
                    destX, destY  : Integer ) is
        -- used by nested functions
        Eaddr : Address;

        ------------------------------------------------------------------------

        procedure Filter( PE, PI, PH, PF, PG, PC, PD, PB, F4, I4, H5, I5 : Unsigned_16;
                          N15, N14, N11, N3, N7, N10, N13, N12, N5, N8 : Integer ) is

            E3  : Unsigned_16; for  E3'Address use Eaddr + Storage_Offset(N3  * 2);
            E5  : Unsigned_16; for  E5'Address use Eaddr + Storage_Offset(N5  * 2);
            E7  : Unsigned_16; for  E7'Address use Eaddr + Storage_Offset(N7  * 2);
            E8  : Unsigned_16; for  E8'Address use Eaddr + Storage_Offset(N8  * 2);
            E10 : Unsigned_16; for E10'Address use Eaddr + Storage_Offset(N10 * 2);
            E11 : Unsigned_16; for E11'Address use Eaddr + Storage_Offset(N11 * 2);
            E12 : Unsigned_16; for E12'Address use Eaddr + Storage_Offset(N12 * 2);
            E13 : Unsigned_16; for E13'Address use Eaddr + Storage_Offset(N13 * 2);
            E14 : Unsigned_16; for E14'Address use Eaddr + Storage_Offset(N14 * 2);
            E15 : Unsigned_16; for E15'Address use Eaddr + Storage_Offset(N15 * 2);

            ex2,
            ex3    : Boolean;
            pixel  : Unsigned_16;
            e, i   : Unsigned_32;
            ke, ki : Unsigned_32;
        begin
            if PE = PH or else PE = PF then
                return;
            end if;

            e := df( PE, PC ) + df( PE, PG ) + df( PI, H5 ) + df( PI, F4 ) + Shift_Left( df( PH, PF ), 2 );
            i := df( PH, PD ) + df( PH, I5 ) + df( PF, I4 ) + df( PF, PB ) + Shift_Left( df( PE, PI ), 2 );
            -- version 'a' (square)
            if e < i and then
                (
                    (not eq( PF, PB ) and then not eq( PF, PC )) or else
                    (not eq( PH, PD ) and then not eq( PH, PG )) or else
                    (
                        eq( PE, PI ) and then
                        (
                            (not eq( PF, F4 ) and then not eq( PF, I4 )) or else
                            (not eq( PH, H5 ) and then not eq( PH, I5 ))
                        )
                    ) or else
                    eq( PE, PG ) or else
                    eq( PE, PC )
                )

            -- version 'b' (semi-rounded)
--            if e < i and then
--                (
--                    (not eq( PF, PB ) and then not eq( PH, PD )) or else
--                    (
--                        eq( PE, PI ) and then
--                        (not eq( PF, I4 ) and then not eq( PH, I5 ))
--                    ) or else
--                    eq( PE, PG ) or else
--                    eq( PE, PC )
--                )

            -- version 'c' (rounded)
--            if e < i
            then
                ke := df( PF, PG );
                ki := df( PH, PC );
                ex2 := PE /= PC and then PB /= PC;
                ex3 := PE /= PG and then PD /= PG;
                if df( PE, PF ) <= df( PE, PH ) then
                    pixel := PF;
                else
                    pixel := PH;
                end if;

                if ex3 and then ex2 and then Shift_Left( ke, 1 ) <= ki and then ke >= Shift_Left( ki, 1 ) then
                    -- Left + Up 2
                    Blend_192( E13, pixel );
                    Blend_64( E12, pixel );
                    E15 := pixel;
                    E14 := pixel;
                    E11 := pixel;
                    E10 := E12;
                    E3 := E12;
                    E7 := E13;
                elsif ex3 and then Shift_Left( ke, 1 ) <= ki then
                    -- Left 2
                    Blend_192( E11, pixel );
                    Blend_192( E13, pixel );
                    Blend_64( E10, pixel );
                    Blend_64( E12, pixel );
                    E14 := pixel;
                    E15 := pixel;
                elsif ex2 and then ke >= Shift_Left( ki, 1 ) then
                    -- Up 2
                    Blend_192( E14, pixel );
                    Blend_192( E7, pixel );
                    Blend_64( E10, pixel );
                    Blend_64( E3, pixel );
                    E11 := pixel;
                    E15 := pixel;
                else
                    -- Diagonal
                    Blend_128( E11, pixel );
                    Blend_128( E14, pixel );
                    E15 := pixel;
                end if;
            elsif e <= i then
                if df( PE, PF ) <= df( PE, PH ) then
                    Blend_64( E15, PF );
                else
                    Blend_64( E15, PH );
                end if;
            end if;
        end Filter;
        pragma Inline_Always( Filter );

        ------------------------------------------------------------------------

        -- used outside the loops
        SCALE     : constant := 4;
        src, dst  : A_Allegro_Locked_Region;
        nl1, nl2,
        nl3       : Integer;

        -- used inside the Y loop
        srcHeight : Integer;
        EaddrY    : Address;
        EaddrYinc : Storage_Offset;
        EaddrXinc : constant Storage_Offset := SCALE * 2;
        sa2Y      : Address;
        sa0, sa1,
        sa2, sa3,
        sa4       : Address;

        -- used inside the X loop
        srcWidth : Integer;
        pprev,
        pprev2   : Unsigned_8;
        C1addr,
        PCaddr,
        PFaddr,
        PIaddr,
        I5addr   : Address;
        C4addr,
        F4addr,
        I4addr   : Address;
    begin
        if not initialized then
            return;
        end if;

        srcWidth := Al_Get_Bitmap_Width( input );
        srcHeight := Al_Get_Bitmap_Height( input );
        if Al_Get_Bitmap_Width( output ) < destX + srcWidth * SCALE or else
           Al_Get_Bitmap_Height( output ) < destY + srcHeight * SCALE
        then
            -- Invalid destination bitmap size
            return;
        end if;

        src := Al_Lock_Bitmap( input, ALLEGRO_PIXEL_FORMAT_RGB_565, ALLEGRO_LOCK_READONLY );
        if src = null then
            -- Failed to lock input bitmap
            return;
        end if;

        if Al_Get_Bitmap_Width( output ) = srcWidth * SCALE and then
           Al_Get_Bitmap_Height( output) = srcHeight * SCALE
        then
            -- optimization when we're filling the entire destination bitmap.
            -- this won't copy the contents of 'output' down from the video
            -- card, saving some time. it starts as a completely black bitmap.
            dst := Al_Lock_Bitmap( output, ALLEGRO_PIXEL_FORMAT_RGB_565, ALLEGRO_LOCK_WRITEONLY );
        else
            -- there are pixels in 'output' that we're not overwriting, so we
            -- need to grab the image from the video card.
            dst := Al_Lock_Bitmap( output, ALLEGRO_PIXEL_FORMAT_RGB_565, ALLEGRO_LOCK_READWRITE );
        end if;
        if dst = null then
            -- Failed to lock destination bitmap
            Al_Unlock_Bitmap( input );
            return;
        end if;

        nl1 := dst.pitch / 2;    -- pixel offset of next line
        nl2 := nl1 + nl1;        -- pixel offset of line after nl1
        nl3 := nl2 + nl1;        -- pixel offset of line after nl2

        EaddrYinc := Storage_Offset(dst.pitch * SCALE);
        EaddrY := dst.data +
                  Storage_Offset(dst.pitch * destY) +
                  Storage_Offset(destX * 2);                  -- output pos
        sa2Y := src.data - Storage_Offset(4);                 -- input pos
        for y in 0..srcHeight-1 loop
            -- output pixel (top left of 4x4)
            Eaddr := EaddrY;

            -- input pixel (middle)
            sa2 := sa2Y;
            -- up one
            sa1 := sa2 - Storage_Offset(src.pitch);
            -- up two
            sa0 := sa1 - Storage_Offset(src.pitch);
            -- down one
            sa3 := sa2 + Storage_Offset(src.pitch);
            -- down two
            sa4 := sa3 + Storage_Offset(src.pitch);

            if y <= 1 then
                sa0 := sa1;
                if y = 0 then
                    sa0 := sa2;
                    sa1 := sa2;
                end if;
            end if;

            if y >= srcHeight - 2 then
                sa4 := sa3;
                if y = srcHeight - 1 then
                    sa4 := sa2;
                    sa3 := sa2;
                end if;
            end if;

            pprev := 2;
            pprev2 := 2;

            for x in 0..srcWidth-1 loop
                if x >= srcWidth - 2 then
                    if x = srcWidth - 1 then
                        C1addr := sa0 + Storage_Offset(2 * 2);
                        PCaddr := sa1 + Storage_Offset(2 * 2);
                        PFaddr := sa2 + Storage_Offset(2 * 2);
                        PIaddr := sa3 + Storage_Offset(2 * 2);
                        I5addr := sa4 + Storage_Offset(2 * 2);
                    else
                        C1addr := sa0 + Storage_Offset(3 * 2);
                        PCaddr := sa1 + Storage_Offset(3 * 2);
                        PFaddr := sa2 + Storage_Offset(3 * 2);
                        PIaddr := sa3 + Storage_Offset(3 * 2);
                        I5addr := sa4 + Storage_Offset(3 * 2);
                    end if;
                    C4addr := PCaddr;
                    F4addr := PFaddr;
                    I4addr := PIaddr;
                else
                    C1addr := sa0 + Storage_Offset(3 * 2);
                    PCaddr := sa1 + Storage_Offset(3 * 2);
                    PFaddr := sa2 + Storage_Offset(3 * 2);
                    PIaddr := sa3 + Storage_Offset(3 * 2);
                    I5addr := sa4 + Storage_Offset(3 * 2);

                    C4addr := sa1 + Storage_Offset(4 * 2);
                    F4addr := sa2 + Storage_Offset(4 * 2);
                    I4addr := sa3 + Storage_Offset(4 * 2);
                end if;

                declare
                    B1 : Unsigned_16; for B1'Address use sa0 + Storage_Offset(2 * 2);
                    PB : Unsigned_16; for PB'Address use sa1 + Storage_Offset(2 * 2);
                    PE : Unsigned_16; for PE'Address use sa2 + Storage_Offset(2 * 2);
                    PH : Unsigned_16; for PH'Address use sa3 + Storage_Offset(2 * 2);
                    H5 : Unsigned_16; for H5'Address use sa4 + Storage_Offset(2 * 2);

                    A1 : Unsigned_16; for A1'Address use sa0 + Storage_Offset(pprev * 2);
                    PA : Unsigned_16; for PA'Address use sa1 + Storage_Offset(pprev * 2);
                    PD : Unsigned_16; for PD'Address use sa2 + Storage_Offset(pprev * 2);
                    PG : Unsigned_16; for PG'Address use sa3 + Storage_Offset(pprev * 2);
                    G5 : Unsigned_16; for G5'Address use sa4 + Storage_Offset(pprev * 2);

                    A0 : Unsigned_16; for A0'Address use sa1 + Storage_Offset(pprev2 * 2);
                    D0 : Unsigned_16; for D0'Address use sa2 + Storage_Offset(pprev2 * 2);
                    G0 : Unsigned_16; for G0'Address use sa3 + Storage_Offset(pprev2 * 2);

                    C1 : Unsigned_16; for C1'Address use C1addr;
                    PC : Unsigned_16; for PC'Address use PCaddr;
                    PF : Unsigned_16; for PF'Address use PFaddr;
                    PI : Unsigned_16; for PI'Address use PIaddr;
                    I5 : Unsigned_16; for I5'Address use I5addr;

                    C4 : Unsigned_16; for C4'Address use C4addr;
                    F4 : Unsigned_16; for F4'Address use F4addr;
                    I4 : Unsigned_16; for I4'Address use I4addr;

                    -- E0  E1  E2  E3
                    -- E4  E5  E6  E7
                    -- E8  E9  E10 E11
                    -- E12 E13 E14 E15
                    E0  : Unsigned_16; for  E0'Address use Eaddr;
                    E1  : Unsigned_16; for  E1'Address use Eaddr + Storage_Offset(1 * 2);
                    E2  : Unsigned_16; for  E2'Address use Eaddr + Storage_Offset(2 * 2);
                    E3  : Unsigned_16; for  E3'Address use Eaddr + Storage_Offset(3 * 2);
                    E4  : Unsigned_16; for  E4'Address use Eaddr + Storage_Offset( nl1      * 2);
                    E5  : Unsigned_16; for  E5'Address use Eaddr + Storage_Offset((nl1 + 1) * 2);
                    E6  : Unsigned_16; for  E6'Address use Eaddr + Storage_Offset((nl1 + 2) * 2);
                    E7  : Unsigned_16; for  E7'Address use Eaddr + Storage_Offset((nl1 + 3) * 2);
                    E8  : Unsigned_16; for  E8'Address use Eaddr + Storage_Offset( nl2      * 2);
                    E9  : Unsigned_16; for  E9'Address use Eaddr + Storage_Offset((nl2 + 1) * 2);
                    E10 : Unsigned_16; for E10'Address use Eaddr + Storage_Offset((nl2 + 2) * 2);
                    E11 : Unsigned_16; for E11'Address use Eaddr + Storage_Offset((nl2 + 3) * 2);
                    E12 : Unsigned_16; for E12'Address use Eaddr + Storage_Offset( nl3      * 2);
                    E13 : Unsigned_16; for E13'Address use Eaddr + Storage_Offset((nl3 + 1) * 2);
                    E14 : Unsigned_16; for E14'Address use Eaddr + Storage_Offset((nl3 + 2) * 2);
                    E15 : Unsigned_16; for E15'Address use Eaddr + Storage_Offset((nl3 + 3) * 2);
                begin
                    E0  := PE; E1  := PE; E2  := PE; E3  := PE;
                    E4  := PE; E5  := PE; E6  := PE; E7  := PE;
                    E8  := PE; E9  := PE; E10 := PE; E11 := PE;
                    E12 := PE; E13 := PE; E14 := PE; E15 := PE;

                    Filter( PE, PI, PH, PF, PG, PC, PD, PB, F4, I4, H5, I5, nl3+3, nl3+2, nl2+3,     3, nl1+3, nl2+2, nl3+1,   nl3, nl1+1,   nl2 );
                    Filter( PE, PC, PF, PB, PI, PA, PH, PD, B1, C1, F4, C4,     3, nl1+3,     2,     0,     1, nl1+2, nl2+3, nl3+3, nl2+1, nl3+2 );
                    Filter( PE, PA, PB, PD, PC, PG, PF, PH, D0, A0, B1, A1,     0,     1,   nl1,   nl3,   nl2, nl1+1,     2,     3, nl2+2, nl1+3 );
                    Filter( PE, PG, PD, PH, PA, PI, PB, PF, H5, G5, D0, G0,   nl3,   nl2, nl3+1, nl3+3, nl3+2, nl2+1,   nl1,     0, nl1+2,     1 );
                end;

                sa0 := sa0 + Storage_Offset(1 * 2);
                sa1 := sa1 + Storage_Offset(1 * 2);
                sa2 := sa2 + Storage_Offset(1 * 2);
                sa3 := sa3 + Storage_Offset(1 * 2);
                sa4 := sa4 + Storage_Offset(1 * 2);

                Eaddr := Eaddr + EaddrXinc;

                if pprev2 > 0 then
                    pprev2 := pprev2 - 1;
                    pprev := 1;
                end if;
            end loop;

            EaddrY := EaddrY + EaddrYinc;
            sa2Y := sa2Y + Storage_Offset(src.pitch);
        end loop;

        Al_Unlock_Bitmap( input );
        Al_Unlock_Bitmap( output );
    exception
        when others =>
            if Al_Is_Bitmap_Locked( input ) then
                Al_Unlock_Bitmap( input );
            end if;
            if Al_Is_Bitmap_Locked( output) then
                Al_Unlock_Bitmap( output );
            end if;
            raise;
    end xBR4;

end xBR;
