--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Ada.Text_IO;
with Applications;
with Preferences;
with Support.Paths;                     use Support.Paths;
with Tiles.Libraries;                   use Tiles.Libraries;
with Tiles.Libraries.Compilers;         use Tiles.Libraries.Compilers;

pragma Elaborate_All( Applications );
pragma Elaborate_All( Preferences );
pragma Elaborate_All( Support.Paths );

package body Applications.Shell.Gfxpack is

    function Create_Application return A_Application is
        this : constant A_Shell_Application := new Gfxpack_Application;
    begin
        this.Construct( company       => "Kevin Wellwood",
                        name          => "gfxpack",
                        configDir     => "",
                        enableAllegro => False );
        return A_Application(this);
    end Create_Application;

    ----------------------------------------------------------------------------

    overriding
    function Run( this : access Gfxpack_Application ) return Integer is

        returnCode : Integer := 0;
        quiet      : Boolean := False;

        ------------------------------------------------------------------------

        procedure Output_Line( str : String; error : Boolean := False ) is
        begin
            if not quiet or error then
                Ada.Text_IO.Put_Line( str );
            end if;
        end Output_Line;

        ------------------------------------------------------------------------

        procedure Display_Title is
        begin
            Output_Line( "" );
            Output_Line( this.Get_Announce );
            Output_Line( "" );
        end Display_Title;

        ------------------------------------------------------------------------

        procedure Display_Usage is
        begin
            Display_Title;
            Output_Line( "Usage:" );
            Output_Line( "  " & this.Get_Name & " [switches] <index_path> <library_path>" );
            Output_Line( "" );
            Output_Line( "Where:" );
            Output_Line( "  <index_path>   : Path of the tile index file" );
            Output_Line( "  <library_path> : Path of the output library file" );
            Output_Line( "" );
            Output_Line( "Switches:" );
            Output_Line( "  -m=? : Path of the tile matrix file (optional)" );
            Output_Line( "  -c   : Compress archive" );
            Output_Line( "  -q   : Quiet mode" );
            Output_Line( "" );
            returnCode := 1;
        end Display_Usage;

        ------------------------------------------------------------------------

        function Arg_Value( arg : String ) return String is
            eq : constant Integer := Index( arg, "=" );
        begin
            if eq >= arg'First and then eq < arg'Last then
                return arg(eq+1..arg'Last);
            else
                return "";
            end if;
        end Arg_Value;

        ------------------------------------------------------------------------

        function Match_Arg( arg : String; head : String ) return Boolean is
            larg : constant String := To_Lower( arg );
        begin
            return larg'Length >= head'Length and then
                   larg(larg'First..larg'First+head'Length-1) = To_Lower( head );
        end Match_Arg;

        ------------------------------------------------------------------------

        indexPath  : Unbounded_String;
        matrixPath : Unbounded_String;
        libPath    : Unbounded_String;
        compress   : Boolean := False;
        compiler   : A_Compiler;
    begin
        if Argument_Count < 2 then
            Display_Usage;
            return returnCode;
        end if;

        for i in 1..Argument_Count - 2 loop
            if Argument( i ) = "-c" then
                compress := True;
            elsif Argument( i ) = "-q" then
                quiet := True;
            elsif Match_Arg( Argument( i ), "-m=" ) then
                matrixPath := To_Unbounded_String( Arg_Value( Argument( i ) ) );
            else
                Display_Usage;
                return returnCode;
            end if;
        end loop;

        indexPath := To_Unbounded_String( Argument( Argument_Count - 1 ) );
        libPath := To_Unbounded_String( Argument( Argument_Count ) );

        Display_Title;

        -- create the library from individual resources
        compiler := Create_Compiler( Output_Line'Unrestricted_Access );
        if not compiler.Compile( To_String( libPath ),
                                 To_String( indexPath ),
                                 To_String( matrixPath ),
                                 compress )
        then
            returnCode := 1;
        end if;
        Delete( compiler );

        return returnCode;
    end Run;

    ----------------------------------------------------------------------------

begin

    Register_Allocator( Create_Application'Access );

    -- Look for media (and configuration) in the current working directory
    Preferences.Set_Default( "application", "media", "media" & Slash );

end Applications.Shell.Gfxpack;
