--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Actions;                           use Actions;
with Ada.Exceptions;                    use Ada.Exceptions;
with Allegro.Keyboard;                  use Allegro.Keyboard;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Application;                use Events.Application;
with Events.Game;                       use Events.Game;
with Events.Ked;                        use Events.Ked;
with Events.World;                      use Events.World;
with Game_Views;                        use Game_Views;
with Game_Views.Ked.Actions;
with Game_Views.Ked.Toolbar;
with Game_Views.Ked.Toolbar.Layers_Menu;
with Game_Views.Ked.Menubar;
with GNAT.Directory_Operations;
with Keyboard;                          use Keyboard;
with Preferences;                       use Preferences;
with Tools.Pointers;                    use Tools.Pointers;
with Support.Paths;                     use Support.Paths;
with Values.Casting;                    use Values.Casting;
with Widgets.Containers;                use Widgets.Containers;
with Widgets.Containers.Panels.Dialogs; use Widgets.Containers.Panels.Dialogs;
with Widgets.Containers.Panels.Dialogs.About;
 use Widgets.Containers.Panels.Dialogs.About;
with Widgets.Containers.Panels.Dialogs.Import_World;
 use Widgets.Containers.Panels.Dialogs.Import_World;
with Widgets.Containers.Panels.Dialogs.New_World;
 use Widgets.Containers.Panels.Dialogs.New_World;
with Widgets.Containers.Panels.Dialogs.Open_World;
 use Widgets.Containers.Panels.Dialogs.Open_World;
with Widgets.Containers.Panels.Dialogs.Properties;
 use Widgets.Containers.Panels.Dialogs.Properties;
with Widgets.Containers.Panels.Dialogs.Resize_World;
 use Widgets.Containers.Panels.Dialogs.Resize_World;
with Widgets.Containers.Panels.Dialogs.Save_Changes;
 use Widgets.Containers.Panels.Dialogs.Save_Changes;
with Widgets.Containers.Panels.Dialogs.Save_World;
 use Widgets.Containers.Panels.Dialogs.Save_World;
with Widgets.Containers.Scenes.Ked;     use Widgets.Containers.Scenes.Ked;
with Worlds;

pragma Elaborate_All( Game_Views );
pragma Elaborate_All( Preferences );

package body Game_Views.Ked is

    WINDOW_TITLE : constant String := "Keen Editor";

    function Is_Modified( this : not null access Ked_View'Class ) return Boolean;

    procedure Set_Filename( this : not null access Ked_View'Class; filename : String );

    procedure Set_Modified( this : not null access Ked_View'Class; modified : Boolean );

    procedure Update_Win_Title( this : not null access Ked_View'Class );

    ----------------------------------------------------------------------------

    procedure Zoom_Scene( this   : not null access Ked_View'Class;
                          action : A_Key_Action );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    procedure Zoom_Scene( this   : not null access Ked_View'Class;
                          action : A_Key_Action ) is
    begin
        case action.Get_Key is
            when ALLEGRO_KEY_F1     => A_Ked_Scene(this.Get_Widget( "scene" )).Zoom_With_Focus( 1.0 );
            when ALLEGRO_KEY_F2     => A_Ked_Scene(this.Get_Widget( "scene" )).Zoom_With_Focus( 2.0 );
            when ALLEGRO_KEY_F3     => A_Ked_Scene(this.Get_Widget( "scene" )).Zoom_With_Focus( 3.0 );
            when ALLEGRO_KEY_F4     => A_Ked_Scene(this.Get_Widget( "scene" )).Zoom_With_Focus( 4.0 );
            when ALLEGRO_KEY_0      => A_Ked_Scene(this.Get_Widget( "scene" )).Zoom_With_Focus( 1.0 );
            when ALLEGRO_KEY_MINUS  => A_Ked_Scene(this.Get_Widget( "scene" )).Zoom_Out;
            when ALLEGRO_KEY_EQUALS => A_Ked_Scene(this.Get_Widget( "scene" )).Zoom_In;
            when others     => null;  -- should never happen
        end case;
    end Zoom_Scene;

    --==========================================================================

    function Create_Game_View( display : not null A_Allegro_Display ) return A_Game_View is
        this : A_Ked_View := new Ked_View;
    begin
        this.Construct( display );
        return A_Game_View(this);
    exception
        when others =>
            Delete( A_Game_View(this) );
            raise;
    end Create_Game_View;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this    : access Ked_View;
                         display : not null A_Allegro_Display ) is
    begin
        Game_View(this.all).Construct( display, width => 0, height => 0 );
        this.tool := Create_Pointer;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Ked_View ) is
    begin
        Delete( this.tool );
        Delete( this.paletteGrp );
        Delete( this.layerGrp );
        Game_View(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Do_Dialog_Action( this        : not null access Ked_View'Class;
                                action      : Dialog_Action;
                                saveChanges : Boolean ) is
    begin
        if action = DA_Save then
            this.nextAction := DA_None;
            A_Dialog(this.Get_Widget( "pnlSaveWorld" )).Show;
        elsif not this.Is_Modified or else not saveChanges then
            case action is
                when DA_Import => A_Dialog(this.Get_Widget( "pnlImport" )).Show;
                when DA_New    => A_Dialog(this.Get_Widget( "pnlNewWorld" )).Show;
                when DA_Open   => A_Dialog(this.Get_Widget( "pnlOpenWorld" )).Show;
                when DA_Quit   => Queue_Exit_Application;
                when DA_Save   |
                     DA_None   => null;
            end case;
        else
            this.nextAction := action;
            A_Dialog(this.Get_Widget( "pnlAskSave" )).Show;
        end if;
    end Do_Dialog_Action;

    ----------------------------------------------------------------------------

    procedure Do_Dialog_Action( this        : not null access Ked_View'Class;
                                saveChanges : Boolean ) is
    begin
        if not this.Is_Modified or else not saveChanges then
            this.Do_Dialog_Action( this.nextAction, saveChanges => False );
        else
            if this.Get_Filename'Length > 0 then
                begin
                    this.Save_World( this.Get_Filename );
                    this.Do_Dialog_Action( this.nextAction, saveChanges => False );
                exception
                    when e : others =>
                        -- shouldn't happen unless the drive disconnected or the
                        -- file is read-only. don't do whatever dialog action
                        -- we were going to do, to avoid losing changes.
                        Dbg( "Error saving world: " & Exception_Message( e ),
                             D_VIEW, ERROR );
                        this.nextAction := DA_None;
                end;
            else
                A_Dialog(this.Get_Widget( "pnlSaveWorld" )).Show;
            end if;
        end if;
    end Do_Dialog_Action;

    ----------------------------------------------------------------------------

    function Get_Filename( this : not null access Ked_View'Class ) return String is
    begin
        return To_String( this.filename );
    end Get_Filename;

    ----------------------------------------------------------------------------

    function Get_Tool( this : not null access Ked_View'Class ) return Tool_Type is
    begin
        return this.tool.Get_Type;
    end Get_Tool;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Ked_View;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        if evt.Get_Id = WORLD_MODIFIED_ID then
            this.Set_Modified( True );

        elsif evt.Get_Id = SET_TOOL_ID then
            Delete( this.tool );
            this.tool := Copy( A_Set_Tool_Event(evt).Get_Tool );
            Toolbar.Handle_Tool_Changed( A_Ked_View(this), A_Set_Tool_Event(evt).Get_Tool );

        elsif evt.Get_Id = WORLD_LOADED_ID then
            this.loaded := True;
            this.Set_Modified( False );
            Menubar.Handle_New_Map( A_Ked_View(this) );
            Toolbar.Handle_New_Map( A_Ked_View(this) );
            Toolbar.Layers_Menu.Update_Menu( A_Ked_View(this), A_World_Loaded_Event(evt).Get_Layers );
            this.Ready_For_Play;

        else
            Game_View(this.all).Handle_Event( evt, resp );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    overriding
    procedure Initialize_Widgets( this : access Ked_View;
                                  win  : not null A_Window ) is
        scene  : A_Ked_Scene;
        dialog : A_Dialog;
    begin
        win.Set_Title( WINDOW_TITLE );

        -- [ scene ] --
        scene := Create_Scene( this, "scene" );
        scene.Set_Layout_LTRB( 0, Menubar.Get_Menubar_Height, Toolbar.Get_Toolbar_Width, 0 );
        win.Add_Widget( A_Widget(scene) );

        -- add key handlers to the window to control the scene
        win.Add_Listener( Listener( ALLEGRO_KEY_F1,     (others=>No), Key_Typed, this, Zoom_Scene'Access ) );
        win.Add_Listener( Listener( ALLEGRO_KEY_F2,     (others=>No), Key_Typed, this, Zoom_Scene'Access ) );
        win.Add_Listener( Listener( ALLEGRO_KEY_F3,     (others=>No), Key_Typed, this, Zoom_Scene'Access ) );
        win.Add_Listener( Listener( ALLEGRO_KEY_F4,     (others=>No), Key_Typed, this, Zoom_Scene'Access ) );
        win.Add_Listener( Listener( ALLEGRO_KEY_0,      (others=>No), Key_Typed, this, Zoom_Scene'Access ) );
        win.Add_Listener( Listener( ALLEGRO_KEY_MINUS,  (others=>No), Key_Typed, this, Zoom_Scene'Access ) );
        win.Add_Listener( Listener( ALLEGRO_KEY_EQUALS, (others=>No), Key_Typed, this, Zoom_Scene'Access ) );

        -- [ tool bar ] --
        Toolbar.Make_Toolbar( A_Ked_View(this), win );

        -- [ dialogs ] --
        dialog := Create_New_Dialog( this, "pnlNewWorld" );
        win.Add_Widget( A_Widget(dialog) );
        dialog := Create_Open_Dialog( this, "pnlOpenWorld" );
        win.Add_Widget( A_Widget(dialog) );
        dialog := Create_Save_Dialog( this, "pnlSaveWorld" );
        win.Add_Widget( A_Widget(dialog) );
        dialog := Create_Save_Changes_Dialog( this, "pnlAskSave" );
        win.Add_Widget( A_Widget(dialog) );
        dialog := Create_Import_Dialog( this, "pnlImport" );
        win.Add_Widget( A_Widget(dialog) );
        dialog := Create_Properties_Dialog( this, "pnlProp" );
        win.Add_Widget( A_Widget(dialog) );
        dialog := Create_Resize_Dialog( this, "pnlResize" );
        win.Add_Widget( A_Widget(dialog) );
        dialog := Create_About_Dialog( this, "pnlAbout" );
        win.Add_Widget( A_Widget(dialog) );

        -- [ menu bar ] --
        Menubar.Make_Menubar( A_Ked_View(this), win );
    end Initialize_Widgets;

    ----------------------------------------------------------------------------

    function Is_Modified( this : not null access Ked_View'Class ) return Boolean is
    begin
        return this.modified;
    end Is_Modified;

    ----------------------------------------------------------------------------

    function Is_World_Loaded( this : not null access Ked_View'Class ) return Boolean is
    begin
        return this.loaded;
    end Is_World_Loaded;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Close_Window( this    : access Ked_View;
                               allowed : in out Boolean ) is
    begin
        Actions.Action_Quit( A_Ked_View(this) );
        allowed := False;
    end On_Close_Window;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Initialize( this : access Ked_View ) is
    begin
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Set_Tool" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "World_Loaded" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "World_Modified" );
        Queue_New_Game;
    end On_Initialize;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Finalize( this : access Ked_View ) is
    begin
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "World_Modified" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "World_Loaded" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Set_Tool" );
    end On_Finalize;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Game_Paused( this : access Ked_View; paused : Boolean ) is
        pragma Unreferenced( this );
    begin
        null;
        pragma Debug( paused, Dbg( "Game_View.Ked: Game paused", D_VIEW, INFO ) );
        pragma Debug( not paused, Dbg( "Game_View.Ked: Game resumed", D_VIEW, INFO ) );
    end On_Game_Paused;

    ----------------------------------------------------------------------------

    overriding
    procedure On_World_Property_Changed( this  : access Ked_View;
                                         name  : String;
                                         value : Value_Ptr'Class ) is
    begin
        if name = "filename" then
            this.Set_Filename( Cast_String( value ) );
        end if;
    end On_World_Property_Changed;

    ----------------------------------------------------------------------------

    procedure Save_World( this : not null access Ked_View'Class; filename : String ) is
        l_filename : constant String := Add_Extension( filename, Worlds.World_Extension, force => True );
    begin
        Trigger_Save_World( l_filename );   -- may raise exception
        this.Set_Filename( l_filename );
        this.Set_Modified( False );
    end Save_World;

    ----------------------------------------------------------------------------

    procedure Set_Filename( this : not null access Ked_View'Class; filename : String ) is
    begin
        if filename'Length /= Length( this.filename ) or else
           filename /= To_String( this.filename )
        then
            this.filename := To_Unbounded_String( filename );
            this.Update_Win_Title;
        end if;
    end Set_Filename;

    ----------------------------------------------------------------------------

    procedure Set_Modified( this : not null access Ked_View'Class; modified : Boolean ) is
    begin
        if this.modified /= modified then
            this.modified := modified;
            this.Update_Win_Title;
        end if;
    end Set_Modified;

    ----------------------------------------------------------------------------

    procedure Update_Win_Title( this : not null access Ked_View'Class ) is

        ------------------------------------------------------------------------

        function Make_Filename return String is
        begin
            if this.Get_Filename'Length > 0 then
                return GNAT.Directory_Operations.File_Name( this.Get_Filename );
            else
                return "Untitled";
            end if;
        end Make_Filename;

        ------------------------------------------------------------------------

        function Make_Modified return String is
        begin
            if this.modified then
                return " (modified)";
            else
                return "";
            end if;
        end Make_Modified;

        ------------------------------------------------------------------------

    begin
        this.win.Set_Title( WINDOW_TITLE & " - " & Make_Filename & Make_Modified );
    end Update_Win_Title;

begin

    Preferences.Set_Default( "gridsize", 16 );
    Preferences.Set_Default( "gridsnap", True );
    Preferences.Set_Default( "pinkscene", False );
    Preferences.Set_Default( "outlines", True );
    Preferences.Set_Default( "draw_clipping", False );
    Preferences.Set_Default( "scaling_filter", "Quality" );

    Register_Allocator( Create_Game_View'Access );

end Game_Views.Ked;
