--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Exceptions;                    use Ada.Exceptions;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Entities;                          use Entities;
with Events.Ked;                        use Events.Ked;
with Events.World;                      use Events.World;
with Games;                             use Games;
with Games.Sessions;                    use Games.Sessions;
with Statuses;                          use Statuses;
with Worlds.Importing;                  use Worlds.Importing;

pragma Elaborate_All( Games.Sessions );

package body Games.Sessions.Ked is

    procedure Handle_Apply_Tool( this : not null access Ked_Session'Class;
                                 evt  : not null A_Apply_Tool_Map_Event );

    procedure Handle_Create_World( this : not null access Ked_Session'Class;
                                   evt  : not null A_Create_World_Event;
                                   resp : out Response_Type );

    procedure Handle_Import_World( this : not null access Ked_Session'Class;
                                   evt  : not null A_Import_World_Event;
                                   resp : out Response_Type );

    procedure Handle_Resize_World( this : not null access Ked_Session'Class;
                                   evt  : not null A_Resize_World_Event;
                                   resp : out Response_Type );

    procedure Handle_Save_World( this : not null access Ked_Session'Class;
                                 evt  : not null A_Save_World_Event;
                                 resp : out Response_Type );

    procedure Handle_Set_Tool( this : not null access Ked_Session'Class;
                               evt  : not null A_Set_Tool_Event );

    procedure Handle_Set_World_Property( this : not null access Ked_Session'Class;
                                         evt  : not null A_World_Property_Event );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Ked_Session( game : not null A_Game ) return A_Game_Session is
        this : constant A_Game_Session := new Ked_Session;
    begin
        this.Construct( game );
        return this;
    end Create_Ked_Session;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Ked_Session ) is
    begin
        Delete( this.tool );
        Game_Session(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Handle_Apply_Tool( this : not null access Ked_Session'Class;
                                 evt  : not null A_Apply_Tool_Map_Event ) is
    begin
        if this.state in Playing..Paused then
            this.tool.Apply( Tool_Context'(func      => evt.Get_Function,
                                           modifiers => evt.Get_Modifiers,
                                           first     => evt.Is_First,
                                           world     => this.Get_World,
                                           worldX    => evt.Get_X,
                                           worldY    => evt.Get_Y,
                                           layer     => evt.Get_Layer) );
        end if;
    end Handle_Apply_Tool;

    ----------------------------------------------------------------------------

    procedure Handle_Create_World( this : not null access Ked_Session'Class;
                                   evt  : not null A_Create_World_Event;
                                   resp : out Response_Type ) is
        newWorld : A_World;
        playerId : Entity_Id;
    begin
        -- may raise an exception
        newWorld := Create_World( evt.Get_Width,
                                  evt.Get_Height,
                                  (True, True, False),
                                  evt.Get_Library_Name,
                                  evt.Get_Domain );
        playerId := newWorld.Spawn_Entity( "player." & evt.Get_Domain,
                                           Float(newWorld.Get_Width) / 2.0,
                                           Float(newWorld.Get_Height) / 2.0 );
        if playerId = INVALID_ID then
            raise Constraint_Error with "Player allocator not registered for domain <" &
                                        evt.Get_Domain & '>';
        end if;
        newWorld.Set_Property( "playerId", To_Id_Value( playerId ) );

        this.Load_World( "", newWorld );     -- consumes newWorld
        Queue_World_Modified;
        Set_Response( resp, ST_SUCCESS );
    exception
        when e : others =>
            Delete( newWorld );
            Set_Response( resp, ST_FAILED, Exception_Message( e ) );
    end Handle_Create_World;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Ked_Session;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
    begin
        if evt.Get_Id = APPLY_TOOL_MAP_ID then
            this.Handle_Apply_Tool( A_Apply_Tool_Map_Event(evt) );

        elsif evt.Get_Id = SET_TOOL_ID then
            this.Handle_Set_Tool( A_Set_Tool_Event(evt) );

        elsif evt.Get_Id = SAVE_WORLD_ID then
            this.Handle_Save_World( A_Save_World_Event(evt), resp );

        elsif evt.Get_Id = CREATE_WORLD_ID then
            this.Handle_Create_World( A_Create_World_Event(evt), resp );

        elsif evt.Get_Id = IMPORT_WORLD_ID then
            this.Handle_Import_World( A_Import_World_Event(evt), resp );

        elsif evt.Get_Id = RESIZE_WORLD_ID then
            this.Handle_Resize_World( A_Resize_World_Event(evt), resp );

        elsif evt.Get_Id = SET_WORLD_PROPERTY_ID then
            this.Handle_Set_World_Property( A_World_Property_Event(evt) );

        else
            Game_Session(this.all).Handle_Event( evt, resp );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    procedure Handle_Import_World( this : not null access Ked_Session'Class;
                                   evt  : not null A_Import_World_Event;
                                   resp : out Response_Type ) is
        imported : A_World;
    begin
        -- may raise an exception
        imported := Import_World( evt.Get_Filename,
                                  evt.Get_Library_Name,
                                  evt.Get_Domain,
                                  evt.Get_Tolerance );
        this.Load_World( "", imported );  -- consumes imported
        Queue_World_Modified;
        Set_Response( resp, ST_SUCCESS );
    exception
        when e : others =>
            Delete( imported );
            Set_Response( resp, ST_FAILED, Exception_Message( e ) );
    end Handle_Import_World;

    ----------------------------------------------------------------------------

    procedure Handle_Resize_World( this : not null access Ked_Session'Class;
                                   evt  : not null A_Resize_World_Event;
                                   resp : out Response_Type ) is
    begin
        case this.state is
            when Playing | Paused =>
                this.Get_World.Resize( evt.Get_Width, evt.Get_Height );
                Set_Response( resp, ST_SUCCESS );

            when Loading =>
                Set_Response( resp, ST_FAILED, "Loading in progress" );

            when Initial | Ended =>
                Set_Response( resp, ST_FAILED, "No world to resize" );
        end case;
    exception
        when e : others =>
            Set_Response( resp, ST_FAILED, Exception_Message( e ) );
    end Handle_Resize_World;

    ----------------------------------------------------------------------------

    procedure Handle_Save_World( this : not null access Ked_Session'Class;
                                 evt  : not null A_Save_World_Event;
                                 resp : out Response_Type ) is
    begin
        case this.state is
            when Playing | Paused =>
                this.Get_World.Save( evt.Get_Filename );
                Set_Response( resp, ST_SUCCESS );

            when Loading =>
                Set_Response( resp, ST_FAILED, "Loading in progress" );

            when Initial | Ended =>
                Set_Response( resp, ST_FAILED, "No world to save" );
        end case;
    exception
        when e : others =>
            Set_Response( resp, ST_FAILED, Exception_Message( e ) );
    end Handle_Save_World;

    ----------------------------------------------------------------------------

    procedure Handle_Set_Tool( this : not null access Ked_Session'Class;
                               evt  : not null A_Set_Tool_Event ) is
    begin
        if this.state in Playing..Paused then
            Delete( this.tool );
            this.tool := Copy( evt.Get_Tool );
        end if;
    end Handle_Set_Tool;

    ----------------------------------------------------------------------------

    procedure Handle_Set_World_Property( this : not null access Ked_Session'Class;
                                         evt  : not null A_World_Property_Event ) is
    begin
        if this.state in Playing..Paused then
            this.Get_World.Set_Property( evt.Get_Property_Name, evt.Get_Value );
        end if;
    end Handle_Set_World_Property;

    ----------------------------------------------------------------------------

    overriding
    procedure On_New_Game( this : access Ked_Session ) is
    begin
        -- register to listen for events
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Apply_Tool_Map" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Create_World" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Import_World" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Resize_World" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Save_World" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Set_Tool" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Set_World_Property" );

        -- load a filename passed on the command line or the default world
        for i in 1..Argument_Count loop
             declare
                 mapName : constant String := Argument( i );
             begin
                 if mapName(mapName'First) /= '-' then
                     this.Load_World( mapName, fresh => True );
                     return;
                 end if;
             exception
                 when e : others =>
                     Dbg( "Error: Failed to load world <" & mapName & ">: " &
                          Exception_Message( e ), D_RES, ERROR );
             end;
        end loop;

        this.Load_World( "default", fresh => True );
    exception
        when others =>
            -- failed to load the default world; carry on
            null;
    end On_New_Game;

    ----------------------------------------------------------------------------

    overriding
    procedure On_End_Game( this      : access Ked_Session;
                           completed : Boolean ) is
        pragma Unreferenced( completed );
    begin
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Set_World_Property" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Set_Tool" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Save_World" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Resize_World" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Import_World" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Create_World" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Apply_Tool_Map" );
    end On_End_Game;

begin

    Register_Allocator( Create_Ked_Session'Access );

end Games.Sessions.Ked;
