--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Exceptions;                    use Ada.Exceptions;
with Allegro.Keyboard;                  use Allegro.Keyboard;
with Debugging;                         use Debugging;
with Game_Views;                        use Game_Views;
with Engine_Debugging;                  use Engine_Debugging;
with Events.World;                      use Events.World;
with Support;                           use Support;
with Themes.Palette;                    use Themes.Palette;
with Widgets.Buttons.Pushes;            use Widgets.Buttons.Pushes;
with Widgets.Containers.Panels;         use Widgets.Containers.Panels;
with Widgets.Input_Boxes.Constraints;   use Widgets.Input_Boxes.Constraints;
with Widgets.Labels;                    use Widgets.Labels;

package body Widgets.Containers.Panels.Dialogs.Open_World is

    procedure Do_Ok( this : not null access Open_Dialog'Class );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Open_Dialog( view : not null access Game_Views.Game_View'Class;
                                 id   : String ) return A_Dialog is
        this : A_Dialog := new Open_Dialog;
    begin
        this.Construct( view, id );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Open_Dialog;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Open_Dialog;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
        button    : A_Button;
        input_box : A_Input_Box;
        label     : A_Label;
    begin
        Dialog(this.all).Construct( view, id, "Open World...", "" );
        this.Set_Layout_Center( 250, 116 );

        label := Create_Label( view, this.id & ":lblFile", "Filename:" );
        label.Set_Layout_LTRB( 5, 5, 5, -23 );
        this.Add_Widget( A_Widget(label) );

        input_box := Create_Input_Box( view, this.id & ":ibFile" );
        input_box.Set_Layout_LTRB( 5, 26, 5, -44 );
        input_box.Set_Constraint( Constraints.Filename_Only'Access );
        input_box.Set_Max_Length( 255 );
        input_box.Add_Listener( A_Input_Listener(this) );
        input_box.Set_Prev( this.id & ":btnCancel" );
        input_box.Set_Next( this.id & ":btnOk" );
        this.Add_Widget( A_Widget(input_box) );

        label := Create_Label( view, this.id & ":lblError" );
        label.Set_Layout_LTRB( 5, 47, 5, -65 );
        label.Set_Color( Text, base_colors(Scarlet1) );
        this.Add_Widget( A_Widget(label) );

        button := Create_Push_Button( view, this.id & ":btnOk", "Open" );
        button.Set_Layout_LTWH( -63, -5, 55, 20 );
        button.Add_Listener( A_Button_Listener(this) );
        button.Set_Prev( this.id & ":ibFile" );
        button.Set_Next( this.id & ":btnCancel" );
        this.Add_Widget( A_Widget(button) );

        button := Create_Push_Button( view, this.id & ":btnCancel", "Cancel" );
        button.Set_Layout_LTWH( -5, -5, 55, 20 );
        button.Add_Listener( A_Button_Listener(this) );
        button.Set_Prev( this.id & ":btnOk" );
        button.Set_Next( this.id & ":ibFile" );
        this.Add_Widget( A_Widget(button) );
    exception
        when others =>
            Delete( A_Widget(button) );
            Delete( A_Widget(input_box) );
            Delete( A_Widget(label) );
            raise;
    end Construct;

    ----------------------------------------------------------------------------

    procedure Do_Ok( this : not null access Open_Dialog'Class ) is
        filename : constant String := A_Input_Box(this.Get_View.Get_Widget( this.id & ":ibFile" )).Get_Text;
    begin
        if filename'Length > 0 then
            Trigger_Load_World( filename );
            this.Hide;
        end if;
    exception
        when e : others =>
            Dbg( "Error: Unable to open world <" & filename & ">: " &
                 Exception_Message( e ), D_VIEW, ERROR );
            this.Set_Error( Exception_Message( e ) );
    end Do_Ok;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Action( this   : access Open_Dialog;
                             action : A_Button_Action ) is
    begin
        if action.Get_Id = Click then
            if action.Get_Source.Get_Id = To_String( this.id & ":btnOk" ) then
                this.Do_Ok;
            elsif action.Get_Source.Get_Id = To_String( this.id & ":btnCancel" ) then
                this.Hide;
            end if;
        end if;
    end Handle_Action;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Action( this   : access Open_Dialog;
                             action : A_Input_Action ) is
    begin
        if action.Get_Id = Entered then
            this.Do_Ok;
        end if;
    end Handle_Action;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Action( this    : access Open_Dialog;
                             action  : A_Key_Action;
                             handled : out Boolean ) is
    begin
        if action.Get_Key = ALLEGRO_KEY_ESCAPE and then
           action.Get_Id = Key_Press
        then
            this.Hide;
            handled := True;
        else
            Dialog(this.all).Handle_Action( action, handled );
        end if;
    end Handle_Action;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Dialog_Show( this : access Open_Dialog ) is
    begin
        A_Input_Box(this.Get_View.Get_Widget( this.id & ":ibFile" )).Set_Text( "" );
        this.Set_Error( "" );
    end On_Dialog_Show;

    ----------------------------------------------------------------------------

    procedure Set_Error( this  : not null access Open_Dialog'Class;
                         error : String ) is
    begin
        A_Label(this.Get_View.Get_Widget( this.id & ":lblError" )).Set_Text( error );
    end Set_Error;

    ----------------------------------------------------------------------------

    overriding
    procedure Show( this : access Open_Dialog ) is
    begin
        Dialog(this.all).Show;
        this.Get_View.Get_Widget( this.id & ":ibFile" ).Request_Focus;
    end Show;

end Widgets.Containers.Panels.Dialogs.Open_World;
