--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
--with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Events.Ked;                        use Events.Ked;
with Game_Views;                        use Game_Views;
with Game_Views.Ked;                    use Game_Views.Ked;
with Mouse;                             use Mouse;
with Physics.Drawing;                   use Physics.Drawing;
with Preferences;                       use Preferences;
with Support;                           use Support;
with Themes.Palette;                    use Themes.Palette;
with Tools;                             use Tools;
with Widgets.Labels;                    use Widgets.Labels;
with Widgets.Sprites;                   use Widgets.Sprites;
with Widgets.Sprites.Ked;               use Widgets.Sprites.Ked;

pragma Elaborate_All( Preferences );

package body Widgets.Containers.Scenes.Ked is

    -- amount the zoom changes each frame
    ZOOM_FACTOR : constant := 0.03;
    MIN_ZOOM    : constant := 0.5;
    MAX_ZOOM    : constant := 4.0;

    -- Gets a bounding box for the scene's target such that the target will
    -- always be in the center of the scene when it is within these bounds.
    -- This procedure is only called by the target when it updates its location.
    procedure Get_Target_Bounds( this   : not null access Ked_Scene'Class;
                                 x1, y1,
                                 x2, y2 : out Float );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Scene( view : not null access Game_Views.Game_View'Class;
                           id   : String ) return A_Ked_Scene is
        this : A_Ked_Scene := new Ked_Scene;
    begin
        this.Construct( view, id );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Scene;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Ked_Scene;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
    begin
        Scene(this.all).Construct( view, id );

        -- allows the viewport to scroll partly off the bottom and right edges
        -- of the map
        this.clampView := False;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Ked_Scene ) is
    begin
        Delete( this.clipMap );
        Scene(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Layer( this   : access Ked_Scene;
                          layer  : Positive;
                          startX,
                          startY : Integer;
                          tileX1,
                          tileY1,
                          tileX2,
                          tileY2 : Integer ) is
        pinkScene  : constant Boolean := Get_Pref( "pinkscene" );
        outlines   : constant Boolean := Get_Pref( "outlines" );
        -- color for edging around current layer
        edgeColor  : constant Allegro_Color := Make_Grey( Integer(this.lightLevels(layer) ));
        -- amount of shading to apply to current layer
        shadeLevel : constant Integer := 255 - Integer(this.lightLevels(layer));

        drawY      : Integer := startY;
        drawX      : Integer := startX;
        mapOffset  : Integer := this.mapLayers(layer).data'First + tileY1 * this.mapWidth;
        mapPos     : Integer;
        tileId     : Integer;     -- id of the tile to draw
        bmp        : A_Allegro_Bitmap;
    begin
        if layer = this.mapLayers'First then
            if pinkScene then
                Rectfill( startX, startY,
                          (tileX2 + 1) * this.tileWidth,
                          (tileY2 + 1) * this.tileWidth,
                          base_colors(Magenta) );
            else
                Rectfill( this.vx1, this.vy1, this.vx2, this.vy2,
                          this.colors(Background) );
            end if;
        end if;

        for y in tileY1..tileY2 loop

            drawX := startX;
            for x in tileX1..tileX2 loop
                mapPos := mapOffset + x;

                tileId := this.mapLayers(layer).data(mapPos);
                if tileId > 0 then
                    -- look up the bitmap in the cache
                    if tileId = this.bitmapCache(layer)(mapPos).tileId then
                        bmp := this.bitmapCache(layer)(mapPos).bmp;
                    else
                        -- bitmap not cached, find it
                        bmp := this.tileLib.Get_Tile( tileId ).Get_Bitmap;
                        if bmp /= null then
                            -- found the bitmap, put it in the cache
                            this.bitmapCache(layer)(mapPos) := (tileId, bmp);
                        end if;
                    end if;

                    if bmp /= null then
                        -- draw the tile to the scene
                        Draw_Bitmap( bmp, drawX, drawY );
                    end if;

                    if shadeLevel < 255 then
                        Rectfill( drawX,
                                  drawY,
                                  drawX + this.tileWidth - 1,
                                  drawY + this.tileWidth - 1,
                                  Al_Map_RGBA( 0, 0, 0, Unsigned_8(shadeLevel) ) );
                    end if;
                elsif outlines then
                    if x < tileX2 then
                        -- check tile to the right
                        if this.mapLayers(layer).data(mapPos + 1) > 0 then
                            Line_V( drawX + this.tileWidth - 1,
                                    drawY,
                                    drawY + this.tileWidth - 1,
                                    edgeColor );
                        end if;
                    end if;
                    if x > tileX1 then
                        -- check tile to the left
                        if this.mapLayers(layer).data(mapPos - 1) > 0 then
                            Line_V( drawX,
                                    drawY,
                                    drawY + this.tileWidth - 1,
                                    edgeColor );
                        end if;
                    end if;
                    if y < tileY2 then
                        -- check tile to the bottom
                        if this.mapLayers(layer).data(mapPos + this.mapWidth) > 0 then
                            Line_H( drawX,
                                    drawX + this.tileWidth - 1,
                                    drawY + this.tileWidth - 1,
                                    edgeColor );
                        end if;
                    end if;
                    if y > tileY1 then
                        -- check tile to the top
                        if this.mapLayers(layer).data(mapPos - this.mapWidth) > 0 then
                            Line_H( drawX,
                                    drawX + this.tileWidth - 1,
                                    drawY,
                                    edgeColor );
                        end if;
                    end if;
                end if;

                drawX := drawX + this.tileWidth;
            end loop;

            drawY := drawY + this.tileWidth;
            mapOffset := mapOffset + this.mapWidth;
        end loop;
    end Draw_Layer;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Layer_Overlay( this   : access Ked_Scene;
                                  startX,
                                  startY : Integer;
                                  tileX1,
                                  tileY1,
                                  tileX2,
                                  tileY2 : Integer ) is
        drawClipping : constant Boolean := Get_Pref( "draw_clipping" );
        drawY        : Integer := startY;
        drawX        : Integer := startX;
    begin
        if not drawClipping then
            return;
        end if;

        for y in tileY1..tileY2 loop
            drawX := startX;
            for x in tileX1..tileX2 loop
                Draw( this.clipMap.Get( x, y ),
                      drawX, drawY, this.tileWidth, base_colors(White) );
                drawX := drawX + this.tileWidth;
            end loop;
            drawY := drawY + this.tileWidth;
        end loop;
    end Draw_Layer_Overlay;

    ----------------------------------------------------------------------------

    function Get_Active_Layer( this : not null access Ked_Scene'Class ) return Natural is
    begin
        return this.activeLayer;
    end Get_Active_Layer;

    ----------------------------------------------------------------------------

    function Get_Focus_X( this : not null access Ked_Scene'Class ) return Float is
    begin
        return this.wFocusX;
    end Get_Focus_X;

    ----------------------------------------------------------------------------

    function Get_Focus_Y( this : not null access Ked_Scene'Class ) return Float is
    begin
        return this.wFocusY;
    end Get_Focus_Y;

    ----------------------------------------------------------------------------

    procedure Get_Target_Bounds( this   : not null access Ked_Scene'Class;
                                 x1, y1,
                                 x2, y2 : out Float ) is
    begin
        -- the focus is always in the center of the viewport, so the lower bound
        -- is half of the widget's viewport width
        x1 := (Float(this.Get_Width) / 2.0) / this.zoom;
        y1 := (Float(this.Get_Height) / 2.0) / this.zoom;
        if this.mapLayers /= null then
            -- allow the right-most/bottom-most target bounds to be 75% of the
            -- width/height of the scene off the right/bottom edge.
            x2 := Float(this.mapWidth * this.tileWidth) + ((Float(this.Get_Width) * 0.25) / this.zoom);
            y2 := Float(this.mapHeight * this.tileWidth) + ((Float(this.Get_Height) * 0.25) / this.zoom);
            if x2 < x1 then
                x2 := x1;
            end if;
            if y2 < y1 then
                y2 := y1;
            end if;
        else
            x2 := x1;
            y2 := y1;
        end if;
    end Get_Target_Bounds;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Entity_Created( this : access Ked_Scene;
                                     evt  : not null A_Entity_Created_Event ) is
        sprite : A_Sprite;
    begin
        sprite := A_Sprite(Create_Ked_Sprite( this.Get_View, evt.Get_Id,
                                              evt.Get_Class,
                                              evt.Get_X, evt.Get_Y,
                                              evt.Get_Width, evt.Get_Height,
                                              evt.Is_Metaphysical, evt.Is_Physical,
                                              evt.Get_Lib_Name, evt.Get_Frame ));
        A_Ked_Sprite(sprite).Set_Attributes( evt.Get_Attributes );
        sprite.Set_Visible( True );
        this.Add_Sprite( sprite ); --, consume => False );
--      this.Set_Selection( A_Widget(sprite) );
    end Handle_Entity_Created;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Tile_Changed( this : access Ked_Scene;
                                   evt  : not null A_Tile_Changed_Event ) is
        tile  : constant Integer := evt.Get_Tile_ID;
        x     : constant Integer := evt.Get_X;
        y     : constant Integer := evt.Get_Y;
        layer : constant Integer := evt.Get_Layer;
    begin
        Scene(this.all).Handle_Tile_Changed( evt );
        if this.hasData then
            this.clipMap.Set_Tile( tile, x, y, layer );
        end if;
    end Handle_Tile_Changed;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_World_Loaded( this : access Ked_Scene;
                                   evt  : not null A_World_Loaded_Event ) is
    begin
        Scene(this.all).Handle_World_Loaded( evt );

        this.activeLayer := this.mapLayers'First;
        for i in this.mapLayers'First..this.mapLayers'Last loop
            if this.activeLayer = i then
                this.lightLevels(i) := Unsigned_8(
                    Constrain( Get_Pref( "scene", "active_light" ), 0, 255 ));
            else
                this.lightLevels(i) := Unsigned_8(
                    Constrain( Get_Pref( "scene", "inactive_light" ), 0, 255 ));
            end if;
        end loop;

        Delete( this.clipMap );
        this.clipMap := Create_Clip_Map( this.mapWidth, this.mapHeight,
                                         this.tileWidth, this.mapLayers,
                                         this.tileLib.Get_Name );
    end Handle_World_Loaded;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Exit( this : access Ked_Scene ) is
    begin
        Scene(this.all).On_Exit;
        A_Label(this.Get_View.Get_Widget( "lblScenePos" )).Set_Text( "" );
    end On_Exit;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Move( this : access Ked_Scene;
                             evt  : not null A_Mouse_Event ) is
        wx, wy : Integer;   -- world coordinates in pixels
        onMap  : Boolean;
    begin
        if this.hasData then
            wx := evt.Get_X;
            wy := evt.Get_Y;
            onMap := wx >= 0 and then wy >= 0 and then
                     wx < this.mapWidth * this.tileWidth and then
                     wy < this.mapHeight * this.tileWidth;

            if onMap then
                -- update the mouse coordinates label
                A_Label(this.Get_View.Get_Widget( "lblScenePos" )).Set_Text(
                    Trim( Integer'Image( wx ), Left ) &
                    ", " & Trim( Integer'Image( wy ), Left ) );
            else
                -- clear the mouse coordinates label
                A_Label(this.Get_View.Get_Widget( "lblScenePos" )).Set_Text( "" );
            end if;

            -- special case: the pointer tool primary drags the map
            if this.leftMouse and then A_Ked_View(this.Get_View).Get_Tool = Pointer_Tool then
                declare
                    x1, y1,
                    x2, y2 : Float;
                    fx, fy : Float;
                begin
                    if wx /= this.dragStartX or else wy /= this.dragStartY then
                        fx := this.wFocusX + Float(this.dragStartX - wx);
                        fy := this.wFocusY + Float(this.dragStartY - wy);
                        this.Get_Target_Bounds( x1, y1, x2, y2 );
                        this.Focus_On( Constrain( fx, x1, x2 ),
                                       Constrain( fy, y1, y2 ),
                                       centered => True );
                    end if;
                end;
                return;
            end if;

            if onMap then
                if this.leftMouse then
                    Queue_Apply_Tool_Map( Primary, this.activeModifiers, False,
                                          wx, wy, this.activeLayer );
                elsif this.rightMouse then
                    Queue_Apply_Tool_Map( Secondary, this.activeModifiers, False,
                                          wx, wy, this.activeLayer );
                elsif this.middleMouse then
                    Queue_Apply_Tool_Map( Tertiary, this.activeModifiers, False,
                                          wx, wy, this.activeLayer );
                end if;
            end if;
        end if;
    end On_Mouse_Move;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Press( this : access Ked_Scene;
                              evt  : not null A_Mouse_Button_Event ) is
        wx, wy : Integer;
        onMap  : Boolean;
    begin
        if this.hasData then
            if evt.Get_Button = Mouse_Left then
                this.leftMouse := True;
            elsif evt.Get_Button = Mouse_Right then
                this.rightMouse := True;
            elsif evt.Get_Button = Mouse_Middle then
                this.middleMouse := True;
            end if;
            this.activeModifiers := evt.Get_Modifiers;
            wx := evt.Get_X;
            wy := evt.Get_Y;
            onMap := wx >= 0 and then wy >= 0 and then
                     wx < this.mapWidth * this.tileWidth and then
                     wy < this.mapHeight * this.tileWidth;
            if evt.Get_Button = Mouse_Left then
                this.Set_Selection( null );
                if A_Ked_View(this.Get_View).Get_Tool = Pointer_Tool then
                    this.dragStartX := wx;
                    this.dragStartY := wy;
                elsif onMap then
                    Queue_Apply_Tool_Map( Primary, this.activeModifiers, True,
                                          wx, wy, this.activeLayer );
                end if;
            elsif evt.Get_Button = Mouse_Right then
                this.Set_Selection( null );
                if onMap then
                    Queue_Apply_Tool_Map( Secondary, this.activeModifiers, True,
                                          wx, wy, this.activeLayer );
                end if;
            elsif evt.Get_Button = Mouse_Middle then
                this.Set_Selection( null );
                if onMap then
                    Queue_Apply_Tool_Map( Tertiary, this.activeModifiers, True,
                                          wx, wy, this.activeLayer );
                end if;
            end if;
        end if;
    end On_Mouse_Press;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Release( this : access Ked_Scene;
                                evt  : not null A_Mouse_Button_Event ) is
    begin
        if evt.Get_Button = Mouse_Left then
            this.leftMouse := False;
        elsif evt.Get_Button = Mouse_Right then
            this.rightMouse := False;
        elsif evt.Get_Button = Mouse_Middle then
            this.middleMouse := False;
        end if;
    end On_Mouse_Release;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Scroll( this    : access Ked_Scene;
                               x, y    : Integer;
                               amount  : Integer;
                               handled : in out Boolean ) is
        x1, y1,
        x2, y2   : Float;
        dfx, dfy : Float;
        dz       : Float;
    begin
         dfx := Float(x) - this.wFocusX;    -- mouse distance from focus
         dfy := Float(y) - this.wFocusY;    --
         dz := this.zoom;
         if amount > 0 then
            for i in 1..amount*2 loop
                this.Zoom_Out;
            end loop;
            dz := this.zoom / dz;                   -- change in zoom
            this.Get_Target_Bounds( x1, y1, x2, y2 );
            this.Focus_On( Constrain( this.wFocusX + (dfx - dfx / dz), x1, x2 ),
                           Constrain( this.wFocusY + (dfy - dfy / dz), y1, y2 ) );
        else
            for i in amount*2..-1 loop
                this.Zoom_In;
            end loop;
            dz := this.zoom / dz;                   -- change in zoom
            this.Get_Target_Bounds( x1, y1, x2, y2 );
            this.Focus_On( Constrain( this.wFocusX + (dfx - dfx / dz), x1, x2 ),
                           Constrain( this.wFocusY + (dfy - dfy / dz), y1, y2 ) );
        end if;
        handled := True;
    end On_Mouse_Scroll;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Resize( this : access Ked_Scene ) is
        x1, y1,
        x2, y2 : Float;
    begin
        Scene(this.all).On_Resize;

        -- refocus, to keep the focus in-bounds. this needs to happen because
        -- the outer right/bottom bounds are based on the widget size.
        this.Get_Target_Bounds( x1, y1, x2, y2 );
        this.Focus_On( Constrain( this.wFocusX, x1, x2 ),
                       Constrain( this.wFocusY, y1, y2 ) );
    end On_Resize;

    ----------------------------------------------------------------------------

    procedure Set_Active_Layer( this  : not null access Ked_Scene'Class;
                                layer : Positive ) is
    begin
        if this.hasData and then layer in this.mapLayers'Range then
            if layer /= this.activeLayer then
                this.lightLevels(this.activeLayer) := Unsigned_8(
                    Constrain( Get_Pref( "scene", "inactive_light" ), 0, 255 ));
                this.lightLevels(layer) := Unsigned_8(
                    Constrain( Get_Pref( "scene", "active_light" ), 0, 255 ));
                this.activeLayer := layer;
            end if;
        end if;
    end Set_Active_Layer;

    ----------------------------------------------------------------------------

    procedure Set_Layer_Visible( this    : not null access Ked_Scene'Class;
                                 layer   : Positive;
                                 visible : Boolean ) is
    begin
        if this.hasData and then layer in this.visibleLayers'Range then
            if this.visibleLayers(layer) /= visible then
                this.visibleLayers(layer) := visible;
            end if;
        end if;
    end Set_Layer_Visible;

    ----------------------------------------------------------------------------

    procedure Shift_Focus( this : not null access Ked_Scene'Class; x,y : Float ) is
        x1, y1, x2, y2 : Float;
    begin
        -- keep the focus in-bounds
        this.Get_Target_Bounds( x1, y1, x2, y2 );
        this.Focus_On( Constrain( this.wFocusX + x, x1, x2 ),
                       Constrain( this.wFocusY + y, y1, y2 ) );
    end Shift_Focus;

    ----------------------------------------------------------------------------

    procedure Zoom_In( this : not null access Ked_Scene'Class ) is
    begin
        this.Zoom_With_Focus( Constrain( this.zoom * (1.0 + ZOOM_FACTOR), MIN_ZOOM, MAX_ZOOM ) );
    end Zoom_In;

    ----------------------------------------------------------------------------

    procedure Zoom_Out( this : not null access Ked_Scene'Class ) is
    begin
        this.Zoom_With_Focus( Constrain( this.zoom * (1.0 - ZOOM_FACTOR), MIN_ZOOM, MAX_ZOOM ) );
    end Zoom_Out;

    ----------------------------------------------------------------------------

    procedure Zoom_With_Focus( this   : not null access Ked_Scene'Class;
                               factor : Float ) is
    begin
        this.Set_Zoom( factor );
        this.Scroll_To( Integer(this.wFocusX - this.Get_Width / 2 / this.zoom),
                        Integer(this.wFocusY - this.Get_Height / 2 / this.zoom) );
    end Zoom_With_Focus;

begin

    -- light levels for active and inactive layers (0-255)
    Preferences.Set_Default( "scene", "active_light", 255 );
    Preferences.Set_Default( "scene", "inactive_light", 160 );
    Preferences.Set_Default( "scene", "scroll_speed", 3.0 );

end Widgets.Containers.Scenes.Ked;
