--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

--with Debugging;                         use Debugging;
with Events.Ked;                        use Events.Ked;
with Events.World;                      use Events.World;
with Game_Views;                        use Game_Views;
with Physics;                           use Physics;
with Resources;                         use Resources;
with Support;                           use Support;
with Themes.Palette;                    use Themes.Palette;
with Tiles;                             use Tiles;
with Tiles.Libraries.Manager;           use Tiles.Libraries.Manager;
with Tools;                             use Tools;
with Tools.Matrixbrushes;               use Tools.Matrixbrushes;
with Values.Casting;                    use Values.Casting;

package body Widgets.Palettes.Matrices is

    overriding
    function "="( l, r : Matrix_Rec ) return Boolean is
    begin
        return l.matrix = r.matrix;
    end "=";

    --==========================================================================

    function Create_Matrix_Palette( view : not null access Game_Views.Game_View'Class;
                                    id   : String ) return A_Palette is
        this : A_Palette := new Matrix_Palette;
    begin
        this.Construct( view, id );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Matrix_Palette;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Matrix_Palette;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
    begin
        Palette(this.all).Construct( view, id );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Set_Tool" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "World_Loaded" );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Matrix_Palette ) is
    begin
        this.matrices.Clear;
        Unload_Library( this.lib );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "World_Loaded" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Set_Tool" );
        Palette(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    function Get_Item_Count( this : access Matrix_Palette ) return Natural is
    begin
        return this.itemCount;
    end Get_Item_Count;

    ----------------------------------------------------------------------------

    overriding
    procedure Get_Item_Details( this   : access Matrix_Palette;
                                row,
                                col    : Natural;
                                bmp    : out A_Allegro_Bitmap;
                                border : out Allegro_Color ) is

        function Cast_Clip_Type is new Cast_Enum( Clip_Type );

        mrec : Matrix_Rec;
        tile : A_Tile;
    begin
        bmp := null;
        border := Themes.Palette.base_colors(Transparent);
        for m in 1..this.matrices.Length loop
            mrec := this.matrices.Element( Integer(m) );
            if row >= mrec.row and then row <= mrec.row + mrec.height - 1 and then
               col >= mrec.col and then col <= mrec.col + mrec.width - 1
            then
                tile := this.lib.Get_Tile( mrec.matrix(row - mrec.row + 1, col - mrec.col + 1) );
                bmp := tile.Get_Bitmap;
                case Cast_Clip_Type( tile.Get_Attribute( "clip" ), Passive ) is
                    when Wall   => border := Themes.Palette.base_colors(White);
                    when others => border := Themes.Palette.base_colors(Transparent);
                end case;
            end if;
        end loop;
    end Get_Item_Details;

    ----------------------------------------------------------------------------

    overriding
    function Get_Scroll_Inc_Y( this : access Matrix_Palette ) return Integer is
    begin
        return this.iconHeight + 1;
    end Get_Scroll_Inc_Y;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Matrix_Palette;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        if evt.Get_Id = WORLD_LOADED_ID then
            this.Set_Library_Name( A_World_Loaded_Event(evt).Get_Library_Name );
        elsif evt.Get_Id = SET_TOOL_ID then
            if A_Set_Tool_Event(evt).Get_Tool.all not in Matrixbrush'Class then
                this.selectedItem := 0;
            end if;
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Press( this : access Matrix_Palette;
                              evt  : not null A_Mouse_Button_Event ) is
        row  : constant Integer := (evt.Get_Y / (this.iconHeight + 1)) + 1;
        col  : constant Integer := (evt.Get_X / (this.iconWidth + 1)) + 1;
        mrec : Matrix_Rec;
        tool : A_Tool;
    begin
        for m in 1..this.matrices.Length loop
            mrec := this.matrices.Element( Integer(m) );
            if row >= mrec.row and then row <= mrec.row + mrec.height - 1 and then
                col >= mrec.col and then col <= mrec.col + mrec.width - 1
            then
                tool := Create_Matrixbrush( mrec.matrix,
                                            col - mrec.col,
                                            row - mrec.row );
                Queue_Set_Tool( tool );
                Delete( tool );
                this.selectedItem := (row - 1) * this.columns + col;
                exit;
            end if;
        end loop;
    end On_Mouse_Press;

    ----------------------------------------------------------------------------

    procedure Set_Library_Name( this : not null access Matrix_Palette'Class;
                                name : String ) is

        ------------------------------------------------------------------------

        procedure Organize_Matrices is

            --------------------------------------------------------------------

            -- Returns true if the matrix intersects with any other matrix.
            function Intersecting_Any( matrix : Matrix_Rec ) return Boolean is

                ----------------------------------------------------------------

                -- Returns true if the two matrix recs intersect at all.
                function Intersect( a, b : Matrix_Rec ) return Boolean is
                begin
                    if a.col > b.col + b.width - 1 or else    -- B is left of A
                       a.col + a.width - 1 < b.col or else    -- B is right of A
                       a.row > b.row + b.height - 1 or else   -- B is above A
                       a.row + a.height - 1 < b.row           -- B is below A
                    then
                        return False;
                    end if;
                    return True;
                end Intersect;

                ----------------------------------------------------------------

                other : Matrix_Rec;
            begin
                for i in 1..this.matrices.Length loop
                    other := this.matrices.Element( Positive(i) );
                    if Intersect( matrix, other ) then
                        return True;
                    end if;
                end loop;
                return False;
            end Intersecting_Any;

            --------------------------------------------------------------------

            mrec : Matrix_Rec;
            row  : Integer;
            done : Boolean := False;
        begin
            for i in 1..this.lib.Get_Matrix_Count loop
                mrec.matrix := this.lib.Get_Matrix( i );
                mrec.height := mrec.matrix'Length(1);
                mrec.width := mrec.matrix'Length(2);
                done := False;
                row := 1;
                loop
                    mrec.row := row;
                    for col in 1..Max( 1, this.columns - mrec.width + 1 ) loop
                        mrec.col := col;
                        if not Intersecting_Any( mrec ) then
                            if ((mrec.row + mrec.height - 1) - 1) * this.columns +
                                (mrec.col + mrec.width - 1) > this.itemCount
                            then
                                this.itemCount := ((mrec.row + mrec.height - 1) - 1) * this.columns + (mrec.col + mrec.width - 1);
                            end if;
                            done := True;
                            exit;
                        end if;
                    end loop;
                    exit when done;
                    row := row + 1;
                end loop;
                this.matrices.Append( mrec );
            end loop;
        end Organize_Matrices;

        ------------------------------------------------------------------------

    begin
        if this.lib = null or else name /= this.lib.Get_Name then
            Unload_Library( this.lib );
            this.matrices.Clear;
            this.itemCount := 0;
            this.lib := Load_Library( name, bitmaps => True );
            if this.lib /= null then
                Organize_Matrices;
            else
                raise RESOURCE_ERROR with "Error loading tile library <" & name & ">";
            end if;
            this.Scroll_To( 0, 0 );
            A_Widget(this).Pack;
        end if;
    end Set_Library_Name;

end Widgets.Palettes.Matrices;
