--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

--with Debugging;                         use Debugging;
with Events.Ked;                        use Events.Ked;
with Events.World;                      use Events.World;
with Game_Views;                        use Game_Views;
with Physics;                           use Physics;
with Resources;                         use Resources;
with Themes.Palette;                    use Themes.Palette;
with Tiles;                             use Tiles;
with Tiles.Libraries.Manager;           use Tiles.Libraries.Manager;
with Tools;                             use Tools;
with Tools.Paintbrushes;                use Tools.Paintbrushes;
with Values.Casting;                    use Values.Casting;

package body Widgets.Palettes.Tiles is

    function Create_Tile_Palette( view : not null access Game_Views.Game_View'Class;
                                  id   : String ) return A_Palette is
        this : A_Palette := new Tile_Palette;
    begin
        this.Construct( view, id );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Tile_Palette;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Tile_Palette;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
    begin
        Palette(this.all).Construct( view, id );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Set_Tool" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "World_Loaded" );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Tile_Palette ) is
    begin
        Unload_Library( this.lib );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "World_Loaded" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Set_Tool" );
        Palette(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    function Get_Item_Count( this : access Tile_Palette ) return Natural is
    begin
        return Natural(this.slots.Length);
    end Get_Item_Count;

    ----------------------------------------------------------------------------

    overriding
    procedure Get_Item_Details( this   : access Tile_Palette;
                                row,
                                col    : Natural;
                                bmp    : out A_Allegro_Bitmap;
                                border : out Allegro_Color ) is

        function Cast_Clip_Type is new Cast_Enum( Clip_Type );

        tile : A_Tile;
    begin
        bmp := null;
        border := base_colors(Transparent);
        if this.lib /= null then
            tile := this.lib.Get_Tile( this.slots.Element( (row - 1) * this.columns + col ) );
            bmp := tile.Get_Bitmap;
            case Cast_Clip_Type( tile.Get_Attribute( "clip" ), Passive ) is
                when Wall => border := Themes.Palette.base_colors(White);
                when others => border := Themes.Palette.base_colors(Transparent);
            end case;
        end if;
    end Get_Item_Details;

    ----------------------------------------------------------------------------

    overriding
    function Get_Scroll_Inc_Y( this : access Tile_Palette ) return Integer is
    begin
        return this.iconHeight + 1;
    end Get_Scroll_Inc_Y;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Tile_Palette;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
        slot : Integer := 0;
        id   : Integer;
        tool : A_Tool;
    begin
        if evt.Get_Id = WORLD_LOADED_ID then
            this.Set_Library_Name( A_World_Loaded_Event(evt).Get_Library_Name );
        elsif evt.Get_Id = SET_TOOL_ID then
            tool := A_Set_Tool_Event(evt).Get_Tool;
            if tool.all in Paintbrush'Class then
                id := Paintbrush(tool.all).Get_Tile;
                if id > 0 then
                    for i in 1..this.slots.Length loop
                        if id = this.slots.Element( Positive(i) ) then
                            slot := Integer(i);
                            exit;
                        end if;
                    end loop;
                end if;
            end if;
            this.selectedItem := slot;
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Press( this : access Tile_Palette;
                              evt  : not null A_Mouse_Button_Event ) is
        row    : constant Integer := (evt.Get_Y / (this.iconHeight + 1)) + 1;
        col    : constant Integer := (evt.Get_X / (this.iconWidth + 1)) + 1;
        tileId : Integer;
        tool   : A_Tool;
    begin
        if this.lib /= null then
            tileId := this.slots.Element( (row - 1) * this.columns + col );
            if tileId > 0 then
                tool := Create_Paintbrush( tileId );
                Queue_Set_Tool( tool );
                Delete( tool );
            end if;
        end if;
     end On_Mouse_Press;

    ----------------------------------------------------------------------------

    procedure Set_Library_Name( this : not null access Tile_Palette'Class;
                                name : String ) is
    begin
        if this.lib = null or else name /= this.lib.Get_Name then
            Unload_Library( this.lib );
            this.lib := Load_Library( name, bitmaps => True );
            if this.lib = null then
                raise RESOURCE_ERROR with "Error loading tile library <" & name & ">";
            end if;

            -- load the tile ids for the tile slots
            declare
                procedure Examine( slot : Positive; tile : A_Tile ) is
                    pragma Unreferenced( slot );
                begin
                    if tile /= null then
                        this.slots.Append( tile.Get_Id );
                    else
                        this.slots.Append( 0 );
                    end if;
                end Examine;
            begin
                this.slots.Clear;
                this.lib.Iterate_By_Slot( Examine'Unrestricted_Access );
            end;

            this.Scroll_To( 0, 0 );
            A_Widget(this).Pack;
        end if;
    end Set_Library_Name;

end Widgets.Palettes.Tiles;
