--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

--with Debugging;                        use Debugging;
with Drawing;                           use Drawing;
with Interfaces;                        use Interfaces;
with Themes.Palette;                    use Themes.Palette;

package body Widgets.Palettes is

    overriding
    procedure Construct( this : access Palette;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
    begin
        Widget(this.all).Construct( view, id );
        this.acceptFocus := False;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Palette ) is
        firstRow    : constant Integer := (this.vy1 / (this.iconHeight + 1)) + 1;
        lastRow     : constant Integer := (this.vy2 / (this.iconHeight + 1)) + 1;
        itemCount   : constant Integer := A_Palette(this).Get_Item_Count;
        selectedRow,
        selectedCol : Integer;
        drawX       : Integer;
        drawY       : Integer := (firstRow - 1) * (this.iconHeight + 1);
        sprite      : A_Allegro_Bitmap;
        itemBorder  : Allegro_Color;
        r, g, b     : Unsigned_8;
    begin
        this.Get_Selected( selectedRow, selectedCol );

        -- background color
        Rectfill( this.vx1, this.vy1, this.vx2, this.vy2, this.colors(Foreground) );

        for row in firstRow..lastRow loop
            exit when (row - 1) * this.columns > itemCount;

            drawX := 0;
            for col in 1..this.columns loop
                exit when (row - 1) * this.columns + col > itemCount;

                A_Palette(this).Get_Item_Details( row, col, sprite, itemBorder );
                if sprite /= null then
                    Draw_Bitmap_Stretched( sprite,
                                           drawX, drawY,
                                           this.iconWidth, this.iconHeight,
                                           proportional => True );
                end if;
                if itemBorder /= Themes.Palette.base_colors(Transparent) then
                    Rect( drawX, drawY,
                          drawX + this.iconWidth - 1,
                          drawY + this.iconHeight - 1, itemBorder );
                end if;

                if row = this.highlightRow and then col = this.highlightCol then
                    Al_Unmap_RGB( this.colors(Selected), r, g, b );
                    Rectfill( drawX, drawY,
                              drawX + this.iconWidth - 1,
                              drawY + this.iconHeight - 1,
                              Al_Map_RGBA( r, g, b, 176 ) );
                    Rect( drawX, drawY,
                          drawX + this.iconWidth - 1,
                          drawY + this.iconHeight - 1, this.colors(Selected) );
                elsif row = selectedRow and then col = selectedCol then
                    Rectfill( drawX, drawY,
                              drawX + this.iconWidth - 1,
                              drawY + this.iconHeight - 1,
                              Al_Map_RGBA( 160, 160, 160, 160 ) );
                    Rect( drawX, drawY,
                          drawX + this.iconWidth - 1,
                          drawY + this.iconHeight - 1, this.colors(Selected_Text) );
                end if;

                drawX := drawX + this.iconWidth + 1;
            end loop;

            drawY := drawY + this.iconHeight + 1;
        end loop;
    end Draw_Content;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Height( this : access Palette ) return Natural is
    begin
        if this.minHeight > 0 then
            return this.minHeight;
        else
            return (A_Palette(this).Get_Item_Count / this.columns + 1) *
                   (this.iconHeight + 1) - 1;
        end if;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Width( this : access Palette ) return Natural is
    begin
        if this.minWidth > 0 then
            return this.minWidth;
        else
            return this.columns * (this.iconWidth + 1) - 1;
        end if;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    overriding
    function Get_Scroll_Inc_X( this : access Palette ) return Integer is
    begin
        return this.iconWidth / 2;
    end Get_Scroll_Inc_X;

    ----------------------------------------------------------------------------

    overriding
    function Get_Scroll_Inc_Y( this : access Palette ) return Integer is
    begin
        return this.iconHeight / 2;
    end Get_Scroll_Inc_Y;

    ----------------------------------------------------------------------------

    procedure Get_Selected( this : not null access Palette'Class;
                            row,
                            col  : out Natural ) is
    begin
        row := 0;
        col := 0;
        if this.selectedItem > 0 then
            row := ((this.selectedItem - 1) / this.columns) + 1;
            col := ((this.selectedItem - 1) mod this.columns) + 1;
        end if;
    end Get_Selected;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Exit( this : access Palette ) is
    begin
        this.highlightRow := 0;
        this.highlightCol := 0;
    end On_Exit;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Mouse_Move( this : access Palette;
                             evt  : not null A_Mouse_Event ) is
        ex : constant Integer := evt.Get_X;
        ey : constant Integer := evt.Get_Y;
    begin
        if ex >= this.vx1 and then ex <= this.vx2 and then
           ey >= this.vy1 and then ey <= this.vy2
        then
            this.highlightRow := (ey / (this.iconHeight + 1)) + 1;
            this.highlightCol := (ex / (this.iconWidth + 1)) + 1;
        else
            this.highlightRow := 0;
            this.highlightCol := 0;
        end if;
    end On_Mouse_Move;

    ----------------------------------------------------------------------------

    procedure Set_Columns( this    : not null access Palette'Class;
                           columns : Natural ) is
    begin
        this.columns := columns;
        this.Pack;
    end Set_Columns;

    ----------------------------------------------------------------------------

    procedure Set_Icon_Size( this   : not null access Palette'Class;
                             width,
                             height : Positive ) is
    begin
        this.iconWidth := width;
        this.iconHeight := height;
        this.Pack;
    end Set_Icon_Size;

end Widgets.Palettes;
