--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Containers.Vectors;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Real_Time;                     use Ada.Real_Time;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Allegro.Bitmaps.Drawing;           use Allegro.Bitmaps.Drawing;
with Allegro.Bitmaps.Image_IO;          use Allegro.Bitmaps.Image_IO;
with Allegro.Bitmaps.Locking;           use Allegro.Bitmaps.Locking;
with Allegro.Color;                     use Allegro.Color;
with Allegro.Displays;                  use Allegro.Displays;
with Allegro.Drawing;                   use Allegro.Drawing;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Interfaces;                        use Interfaces;
with Support;                           use Support;
with Themes.Palette;                    use Themes.Palette;
with Tiles.Libraries.Manager;           use Tiles.Libraries.Manager;
with Worlds;                            use Worlds;

package body Worlds.Importing is

    function Import_World( imagePath,
                           libName,
                           domain    : String;
                           tolerance : Natural ) return A_World is

        UPDATE_DELAY : constant Time_Span := Seconds( 3 );

        package Int_Vectors is new Ada.Containers.Vectors(Positive, Natural, "=");

        ------------------------------------------------------------------------

        function Get_Tile_Width return Natural is
            map    : A_Map;
            result : Natural;
        begin
            map := Create_Map( 1, 1, (True, True, False) );
            result := map.Tile_Width;
            Delete( map );
            return result;
        exception
            when others =>
                Delete( map );
                raise;
        end Get_Tile_Width;

        ------------------------------------------------------------------------

        function Is_Transparent( bmp : A_Allegro_Bitmap ) return Boolean is
            r, g, b, a : Unsigned_8;
        begin
            if bmp = null then
                return False;
            end if;
            Al_Lock_Bitmap( bmp,
                            ALLEGRO_PIXEL_FORMAT_ANY_32_WITH_ALPHA,
                            ALLEGRO_LOCK_READONLY );
            for y in 0..(Al_Get_Bitmap_Height( bmp ) - 1) loop
                for x in 0..(Al_Get_Bitmap_Width( bmp ) - 1) loop
                    Al_Unmap_RGBA( Al_Get_Pixel( bmp, x, y ), r, g, b, a );
                    if a = 0 then
                        Al_Unlock_Bitmap( bmp );
                        return True;
                    end if;
                end loop;
            end loop;
            Al_Unlock_Bitmap( bmp );
            return False;
        end Is_Transparent;

        ------------------------------------------------------------------------

        function Match( a, b : A_Allegro_Bitmap ) return Boolean is
        begin
            if a = null and then b = null then
                return True;
            end if;

            if a = null or else b = null or else
               Al_Get_Bitmap_Width( a ) /= Al_Get_Bitmap_Width( b ) or else
               Al_Get_Bitmap_Height( a ) /= Al_Get_Bitmap_Height( b )
            then
                return False;
            end if;

            Al_Lock_Bitmap( a,
                            ALLEGRO_PIXEL_FORMAT_ANY_32_WITH_ALPHA,
                            ALLEGRO_LOCK_READONLY );
            Al_Lock_Bitmap( b,
                            ALLEGRO_PIXEL_FORMAT_ANY_32_WITH_ALPHA,
                            ALLEGRO_LOCK_READONLY );
            for y in 0..(Al_Get_Bitmap_Height( a ) - 1) loop
                for x in 0..(Al_Get_Bitmap_Width( a ) - 1) loop
                    if not Compare( Al_Get_Pixel( a, x, y ),
                                    Al_Get_Pixel( b, x, y ),
                                    tolerance )
                    then
                        Al_Unlock_Bitmap( b );
                        Al_Unlock_Bitmap( a );
                        return False;
                    end if;
                end loop;
            end loop;
            Al_Unlock_Bitmap( b );
            Al_Unlock_Bitmap( a );
            return True;
        end Match;
        ------------------------------------------------------------------------

        function Transparent_Match( bg, fg : A_Allegro_Bitmap ) return Boolean is
            fgp : Allegro_Color;
        begin
            if bg = null or else fg = null or else
               Al_Get_Bitmap_Width( bg ) /= Al_Get_Bitmap_Width( fg ) or else
               Al_Get_Bitmap_Height( bg ) /= Al_Get_Bitmap_Height( fg )
            then
                return False;
            end if;

            Al_Lock_Bitmap( bg,
                            ALLEGRO_PIXEL_FORMAT_ANY_32_WITH_ALPHA,
                            ALLEGRO_LOCK_READONLY );
            Al_Lock_Bitmap( fg,
                            ALLEGRO_PIXEL_FORMAT_ANY_32_WITH_ALPHA,
                            ALLEGRO_LOCK_READONLY );
            for y in 0..(Al_Get_Bitmap_Height( bg ) - 1) loop
                for x in 0..(Al_Get_Bitmap_Width( bg ) - 1) loop
                    fgp := Al_Get_Pixel( fg, x, y );
                    if fgp /= base_colors(Transparent) and then
                       not Compare( fgp, Al_Get_Pixel( bg, x, y ), tolerance )
                    then
                        Al_Unlock_Bitmap( fg );
                        Al_Unlock_Bitmap( bg );
                        return False;
                    end if;
                end loop;
            end loop;
            Al_Unlock_Bitmap( fg );
            Al_Unlock_Bitmap( bg );
            return True;
        end Transparent_Match;

        ------------------------------------------------------------------------

        procedure Update_Progress( percent, secsRemaining : Integer ) is
        begin
            null;
            pragma Debug( Dbg( "  " &
                               Image( percent ) & "% : " &
                               Image( secsRemaining ) & "s remaining",
                               D_GAME, INFO ) );
        end Update_Progress;

        ------------------------------------------------------------------------

        bmp          : A_Allegro_Bitmap := Al_Load_Bitmap( imagePath );
        lib          : A_Tile_Library := Load_Library( libName, bitmaps => True );
        world        : A_World := null;
        tw           : Natural;
        start        : constant Time := Clock;
        lastUpdate   : Time := start;      -- time of last progress update
        progress     : Natural;
        lastProgress : Natural := 0;       -- progress at last update
        matched      : Natural := 0;       -- tiles successfully matched
        tileIds      : Int_Vectors.Vector;
        transIds     : Int_Vectors.Vector;
        tmpBmp       : A_Allegro_Bitmap;
        clk          : Time;
        secsLeft     : Integer;
        flags        : Allegro_Bitmap_Flags;
    begin
        if bmp = null then
            raise Constraint_Error with "Unable to read file <" & imagePath & ">";
        elsif lib = null then
            raise Constraint_Error with "Unable to load library <" & libName & ">";
        elsif not Valid_Domain( domain ) then
            raise Constraint_Error with "Invalid domain <" & domain & ">";
        end if;

        tw := Get_Tile_Width;
        if Al_Get_Bitmap_Width( bmp ) mod tw /= 0 or else
           Al_Get_Bitmap_Height( bmp ) mod tw /= 0
        then
            raise Constraint_Error with "Bitmap dimensions must be a multiple of " &
                                        Image( tw );
        end if;

        declare
            procedure Examine( tile : not null A_Tile ) is
            begin
                tileIds.Append( tile.Get_Id );
            end Examine;
        begin
            lib.Iterate_By_Id( Examine'Unrestricted_Access );
        end;

        flags := Al_Get_New_Bitmap_Flags;
        Al_Set_New_Bitmap_Flags( ALLEGRO_MEMORY_BITMAP );
        tmpBmp := Al_Create_Bitmap( tw, tw );
        Al_Set_New_Bitmap_Flags( flags );
        for t in 1..tileIds.Length loop
            if Is_Transparent( lib.Get_Tile( tileIds.Element( Positive(t) ) ).Get_Bitmap ) then
                transIds.Append( tileIds.Element( Positive(t) ) );
            end if;
        end loop;

        world := Create_World( Al_Get_Bitmap_Width( bmp ) / tw,
                               Al_Get_Bitmap_Height( bmp ) / tw,
                               (True, True, False),
                               libName, domain );
        if world.Spawn_Entity( "player." & domain,
                               Float(Al_Get_Bitmap_Width( bmp )) / 2.0,
                               Float(Al_Get_Bitmap_Height( bmp )) / 2.0 ) = INVALID_ID
        then
            raise Constraint_Error with "Player allocator not registered for domain <" &
                                        domain & '>';
        end if;

        pragma Debug( Dbg( "Matching tiles...", D_GAME, INFO ) );
        for y in 0..(world.Get_Height_Tiles - 1) loop
            for x in 0..(world.Get_Width_Tiles - 1) loop

                declare
                    subBmp    : A_Allegro_Bitmap := Al_Create_Sub_Bitmap( bmp, x * tw, y * tw, tw, tw );
                    bgTile,
                    fgTile    : A_Tile;
                    matchedFg,
                    matchedBg : Boolean := False;
                    foundFg,
                    foundBg   : Boolean := False;
                begin
                    -- try to match just a background
                    for bg in 1..tileIds.Length loop
                        bgTile := lib.Get_Tile( tileIds.Element( Positive(bg) ) );
                        if bgTile.Get_Bitmap /= null then
                            if Match( subBmp, bgTile.Get_Bitmap ) then
                                world.Set_Tile( 1, x, y, bgTile.Get_Id, notify => False );
                                matchedBg := True;
                                foundBg := True;
                                exit;
                            end if;
                        end if;
                    end loop;

                    -- try to find a foreground tile that works
                    if not matchedBg then
                        for fg in 1..transIds.Length loop
                            fgTile := lib.Get_Tile( transIds.Element( Positive(fg) ) );
                            if Transparent_Match( subBmp, fgTile.Get_Bitmap ) then
                                -- try to combine a background with the foreground for a match
                                for bg in 1..tileIds.Length loop
                                    bgTile := lib.Get_Tile( tileIds.Element( Positive(bg) ) );
                                    if bgTile.Get_Bitmap /= null then
                                        Al_Set_Target_Bitmap( tmpBmp );
                                        Al_Clear_To_Color( Al_Map_RGB( 255, 0, 255 ) );
                                        Al_Draw_Bitmap( bgTile.Get_Bitmap, 0.0, 0.0, 0 );
                                        Al_Draw_Bitmap( fgTile.Get_Bitmap, 0.0, 0.0, 0 );
                                        Al_Set_Target_Bitmap( null );
                                        if Match( subBmp, tmpBmp ) then
                                            foundFg := not matchedFg;
                                            if not matchedFg then
                                                world.Set_Tile( 2, x, y, fgTile.Get_Id, notify => False );
                                            else
                                                -- found multiple foreground matches
                                                world.Set_Tile( 2, x, y, 0, notify => False );
                                            end if;
                                            foundBg := not matchedBg;
                                            if not matchedBg then
                                                world.Set_Tile( 1, x, y, bgTile.Get_Id, notify => False );
                                                matchedBg := True;
                                            else
                                                -- found multiple background matches
                                                world.Set_Tile( 1, x, y, 0, notify => False );
                                            end if;
                                        end if;
                                    end if;
                                end loop;
                                matchedFg := True;
                            end if;
                        end loop;
                    end if;

                    if foundFg or else foundBg then
                        matched := matched + 1;
                    end if;
                    Al_Destroy_Bitmap( subBmp );
                end;

                if x mod 20 = 0 then
                    clk := Clock;
                    if clk - lastUpdate >= UPDATE_DELAY then
                        progress := 100 * (y * world.Get_Width_Tiles + x + 1) /
                                          (world.Get_Width_Tiles * world.Get_Height_Tiles);
                        if progress /= lastProgress then
                            secsLeft := Integer(Long_Float(100.0 / Float(progress)) *
                                                Long_Float(To_Duration(clk - start)) -
                                                Long_Float(To_Duration(clk - start)));
                            Update_Progress( progress, secsLeft );
                            lastProgress := progress;
                            lastUpdate := clk;
                        end if;
                    end if;
                end if;

            end loop;
        end loop;
        Update_Progress( 100, 0 );
        pragma Debug( Dbg( "Matched " & Image( matched ) & " of " &
                           Image( world.Get_Width_Tiles * world.Get_Height_Tiles ) &
                           " tiles: " & Image( 100 * matched / (world.Get_Width_Tiles * world.Get_Height_Tiles) ) & '%',
                           D_GAME, INFO ) );

        pragma Debug( matched = 0, Dbg( "Failed to match any tiles.", D_GAME, INFO ) );

        Al_Destroy_Bitmap( tmpBmp );
        Al_Destroy_Bitmap( bmp );
        Unload_Library( lib );
        return world;
    exception
        when e : others =>
            Dbg( "Error importing world: " &
                 Exception_Name( e ) & " - " & Exception_Message( e ),
                 D_GAME, ERROR );
            Al_Destroy_Bitmap( tmpBmp );
            Al_Destroy_Bitmap( bmp );
            Unload_Library( lib );
            Delete( world );
            raise;
    end Import_World;

end Worlds.Importing;
