--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Text_IO;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Allegro.Color;                     use Allegro.Color;
with Applications;
--with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with GNAT.Directory_Operations;
with Engine_Debugging;                  use Engine_Debugging;
with Hashed_Strings;                    use Hashed_Strings;
with Preferences;                       use Preferences;
with Resources.Images;                  use Resources.Images;
with Support.Paths;                     use Support.Paths;
with Version;

pragma Elaborate_All( Applications );
pragma Elaborate_All( Preferences );
pragma Elaborate_All( Support.Paths );

package body Applications.Gui.Games.Keen is

    KEEN_DATA_DIR : constant String := App_Data_Directory &
                                       "Wellwood" & Slash &
                                       "Keen Galaxy" & Slash;

    KEEN_USER_DIR : constant String := Home_Directory &
                                       "Keen Galaxy" & Slash;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Application return A_Application is
        this : constant A_Game_Application := new Keen_Application;
    begin
        this.Construct( company         => "Kevin Wellwood",
                        name            => "keen",
                        userDir         => KEEN_USER_DIR,
                        defaultWidth    => 664,
                        defaultHeight   => 424,
                        defaultScale    => 1,
                        defaultWindowed => True );
        this.interactive := True;
        GNAT.Directory_Operations.Change_Dir( Execution_Directory );
        return A_Application(this);
    end Create_Application;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Loading_Screen( this : access Keen_Application ) is
        icon16,
        icon32,
        icon64  : A_Allegro_Bitmap;
        loading : A_Allegro_Bitmap;
    begin
        icon16 := Load_Image( "keen-16.png", "resources" );
        icon32 := Load_Image( "keen-32.png", "resources" );
        icon64 := Load_Image( "keen-64.png", "resources" );
        Al_Set_Display_Icons( this.display, icons => (icon16, icon32, icon64) );
        Al_Destroy_Bitmap( icon16 );
        Al_Destroy_Bitmap( icon32 );
        Al_Destroy_Bitmap( icon64 );

        loading := Load_Image( "loading.png", "resources" );
        if loading /= null then
            Clear_To_Color( Al_Map_RGB( 92, 92, 92 ) );
            Draw_Bitmap( loading,
                         (Get_Target_Width - Al_Get_Bitmap_Width( loading )) / 2,
                         (Get_Target_Height - Al_Get_Bitmap_Height( loading )) / 2 );
            Al_Destroy_Bitmap( loading );
        end if;
    end Draw_Loading_Screen;

    ----------------------------------------------------------------------------

    function Initialize( this : access Keen_Application ) return Boolean is

        procedure Display_Usage is
            use Ada.Text_IO;
        begin
            New_Line;
            Put_Line( this.Get_Name &
                      " v" & Version.Release &
                      " - " & this.Get_Company & " " &
                      "Copyright (C) " & Version.Copyright_Year );
            New_Line;
            Put_Line( "Usage:" );
            Put_Line( "  " & this.Get_Name & " [switches]" );
            New_Line;
            Put_Line( "Where optional [switches] are:" );
            Put_Line( "  -DV=<verbosity>          (Debugging Verbosity)" );
            Put_Line( "  -DV<system>=<verbosity>  (Debugging Verbosity per system)" );
            Put_Line( "    Where:" );
            Put_Line( "      <verbosity> is the verbosity level:" );
            Put_Line( "        e : Errors only (default)" );
            Put_Line( "        w : Warnings and errors only" );
            Put_Line( "        i : All info, warnings and errors" );
            Put_Line( "      <system> is the name of a specific debugging system:" );
            for i in Engine_Debugging.Systems'Range loop
                Put_Line( "        " & To_String( Engine_Debugging.Systems(i) ) );
            end loop;
            New_Line;
            Put_Line( "  -DD=<decoration>         (Debugging Decoration)" );
            Put_Line( "    Where:" );
            Put_Line( "      <decoration> is the level of decoration for debug output:" );
            Put_Line( "        0 : No decoration" );
            Put_Line( "        1 : Prefixed with current task" );
            Put_Line( "        2 : Prefixed with calling source unit and current task" );
            New_Line;
            Put_Line( "  -DF=<filename>           (Debugging File)" );
            Put_Line( "    Where:" );
            Put_Line( "      <filename> is the file name or path to write debugging information." );
            Put_Line( "        Debugging output will be appended to the file. If no path is given," );
            Put_Line( "        the file will be written relative to the application's configuration" );
            Put_Line( "        directory." );
            New_Line;
        end Display_Usage;

    begin
        -- Display usage and quit
        for i in 1..Argument_Count loop
            if Argument( i ) = "?" or else
               Argument( i ) = "-?" or else
               Argument( i ) = "-help"
            then
                Display_Usage;
                return False;
            end if;
        end loop;

        return Game_Application(this.all).Initialize;
    end Initialize;

    ----------------------------------------------------------------------------

    procedure Load_Resources( this : access Keen_Application ) is
        pragma Unreferenced( this );
    begin
        -- preload all sound effects
        Resources.Preload_Resource( "click.wav",        "audio" );
        Resources.Preload_Resource( "enter_level.wav",  "audio" );
        Resources.Preload_Resource( "exit_level.wav",   "audio" );
        Resources.Preload_Resource( "flag1.wav",        "audio" );
        Resources.Preload_Resource( "flag2.wav",        "audio" );
        Resources.Preload_Resource( "get_1up.wav",      "audio" );
        Resources.Preload_Resource( "get_drop.wav",     "audio" );
        Resources.Preload_Resource( "get_gem.wav",      "audio" );
        Resources.Preload_Resource( "get_gun.wav",      "audio" );
        Resources.Preload_Resource( "get_points.wav",   "audio" );
        Resources.Preload_Resource( "keen_dead.wav",    "audio" );
        Resources.Preload_Resource( "keen_die.wav",     "audio" );
        Resources.Preload_Resource( "keen_floor.wav",   "audio" );
        Resources.Preload_Resource( "keen_jump.wav",    "audio" );
        Resources.Preload_Resource( "keen_pogo.wav",    "audio" );
        Resources.Preload_Resource( "keen_shoot.wav",   "audio" );
        Resources.Preload_Resource( "menu_close.wav",   "audio" );
        Resources.Preload_Resource( "menu_open.wav",    "audio" );
        Resources.Preload_Resource( "paddle_hit1.wav",  "audio" );
        Resources.Preload_Resource( "paddle_hit2.wav",  "audio" );
        Resources.Preload_Resource( "paddle_miss.wav",  "audio" );
        Resources.Preload_Resource( "paddle_score.wav", "audio" );
        Resources.Preload_Resource( "paddle_wall.wav",  "audio" );
        Resources.Preload_Resource( "shot_hit.wav",     "audio" );
    end Load_Resources;

begin

    Register_Allocator( Create_Application'Access );

    Preferences.Set_Default( "application", "max_input_fps", 80 );
    Preferences.Set_Default( "application", "max_logic_fps", 120 );
    Preferences.Set_Default( "application", "max_render_fps", 60 );
    Preferences.Set_Default( "application", "editing_mode", False );
    Preferences.Set_Default( "application", "media", KEEN_DATA_DIR & "media" & Slash );

end Applications.Gui.Games.Keen;
