--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Actions;                           use Actions;
with Allegro.Color;                     use Allegro.Color;
with Allegro.Keyboard;                  use Allegro.Keyboard;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Audio;                      use Events.Audio;
with Events.Keen;                       use Events.Keen;
with Game_Views;                        use Game_Views;
with Keyboard;                          use Keyboard;
with Themes;                            use Themes;
with Values.Casting;                    use Values.Casting;
with Widgets.Buttons;                   use Widgets.Buttons;
with Widgets.Buttons.Simple_Listeners;
with Widgets.Containers.Game_Screens;   use Widgets.Containers.Game_Screens;
with Widgets.Containers.Game_Screens.Game_Play;
 use Widgets.Containers.Game_Screens.Game_Play;
with Widgets.Containers.Game_Screens.Menus.Main;
 use Widgets.Containers.Game_Screens.Menus.Main;
with Widgets.Containers.Game_Screens.Message;
 use Widgets.Containers.Game_Screens.Message;
with Widgets.Containers.Game_Screens.Retry;
 use Widgets.Containers.Game_Screens.Retry;
with Widgets.Containers.Game_Screens.Title;
 use Widgets.Containers.Game_Screens.Title;
with Widgets.Containers.Screen_Managers;
 use Widgets.Containers.Screen_Managers;
with Widgets.Simple_Key_Listeners;

pragma Elaborate_All( Game_Views );
pragma Elaborate_All( Widgets.Simple_Key_Listeners );
pragma Elaborate_All( Widgets.Buttons.Simple_Listeners );

package body Game_Views.Keen is

    WINDOW_TITLE : constant String := "Keen Galaxy";

    MIN_LOAD_DELAY : constant Time_Span := Seconds( 3 );

    package Button_Listeners is new Widgets.Buttons.Simple_Listeners(Keen_View);
    use Button_Listeners;

    package Key_Listeners is new Widgets.Simple_Key_Listeners(Keen_View);
    use Key_Listeners;

    ----------------------------------------------------------------------------

    -- Called when the user attempts to quit the game via a key shortcut. This
    -- procedure may do something else first, like ask if the game should be
    -- saved, etc. When the game should be quit, this procedure will queue an
    -- Exit_Application event.
    procedure Action_Quit( this   : not null access Keen_View'Class;
                           action : A_Key_Action );

    ----------------------------------------------------------------------------

    procedure Handle_Dialog_Event( this : not null access Keen_View'Class;
                                   evt  : not null A_Dialog_Event );

    procedure Handle_Message_Event( this : not null access Keen_View'Class;
                                    evt  : not null A_Message_Event );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Game_View( display : not null A_Allegro_Display ) return A_Game_View is
        this : A_Keen_View := new Keen_View;
    begin
        this.Construct( display, width => 320, height => 200 );
        return A_Game_View(this);
    exception
        when others =>
            Delete( A_Game_View(this) );
            raise;
    end Create_Game_View;

    ----------------------------------------------------------------------------

    procedure Action_Quit( this   : not null access Keen_View'Class;
                           action : A_Key_Action ) is
        pragma Unreferenced( action );
    begin
        this.Quit;
     end Action_Quit;

    ----------------------------------------------------------------------------

    procedure Handle_Dialog_Event( this : not null access Keen_View'Class;
                                   evt  : not null A_Dialog_Event ) is
        screen : A_Game_Screen;
    begin
        -- we're expecting a retry level dialog with two choices
        if evt.Get_Desc.Get_Name = "Retry" then
            screen := Create_Retry_Screen( this, "scrRetry",
                                           evt.Get_Dialog_Id,
                                           evt.Get_Desc.Get_Text,
                                           evt.Get_Desc.Get_Choice( 1 ),
                                           evt.Get_Desc.Get_Choice( 2 ) );
            A_Screen_Manager(this.Get_Widget( "screenManager" )).Add_Screen( screen );
        end if;
    end Handle_Dialog_Event;

    ----------------------------------------------------------------------------

    procedure Handle_Message_Event( this : not null access Keen_View'Class;
                                    evt  : not null A_Message_Event ) is
        msg : A_Game_Screen;
    begin
        msg := Create_Message_Screen( A_Game_View(this), "scrMessage", True );
        Message_Screen(msg.all).Set_Text_Align( evt.Get_Text_Align );
        Message_Screen(msg.all).Set_Text( evt.Get_Text );
        Message_Screen(msg.all).Set_Icon( evt.Get_Icon );
        Message_Screen(msg.all).Set_Closable( True );
        Message_Screen(msg.all).Set_Auto_Pause( True );
        Screen_Manager(this.Get_Widget( "screenManager" ).all).Add_Screen( msg );
    exception
        when others =>
            Dbg( "Error: A message is already on display", D_VIEW, ERROR );
    end Handle_Message_Event;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Close_Window( this    : access Keen_View;
                               allowed : in out Boolean ) is
        pragma Unreferenced( this );
    begin
        -- there's no "Save Game?" dialog; just quit
        allowed := True;
    end On_Close_Window;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Keen_View;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        if evt.Get_Id = MESSAGE_ID then
            this.Handle_Message_Event( A_Message_Event(evt) );

        elsif evt.Get_Id = DIALOG_ID then
            this.Handle_Dialog_Event( A_Dialog_Event(evt) );

        else
            Game_View(this.all).Handle_Event( evt, resp );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    overriding
    procedure Initialize_Widgets( this : access Keen_View;
                                  win  : not null A_Window ) is
        screenManager : A_Screen_Manager;
        screen        : A_Game_Screen;
    begin
        win.Set_Outer_Border( Al_Map_RGB( 0, 170, 170 ) );
        win.Add_Listener( Listener( ALLEGRO_KEY_F4, (ALT=>Yes,others=>No),
                                    Key_Press, this, Action_Quit'Access ) );
        win.Set_Title( WINDOW_TITLE );

        -- Screen Manager and initial game screens:
        -- Title (in front) -> Main Menu -> Gameplay
        screenManager := Create_Screen_Manager( this, "screenManager" );

            screen := Create_Game_Play_Screen( this, "scrGamePlay" );
            screenManager.Add_Screen( screen );

            screen := Create_Main_Menu_Screen( this, "scrMainMenu",
                                               gameInProgress => False );
            screenManager.Add_Screen( screen );

            screen := Create_Title_Screen( this, "scrTitle" );
            screenManager.Add_Screen( screen );

        win.Add_Widget( A_Widget(screenManager) );
    end Initialize_Widgets;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Initialize( this : access Keen_View ) is
    begin
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Dialog" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Message" );
    end On_Initialize;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Finalize( this : access Keen_View ) is
    begin
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Message" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Dialog" );
    end On_Finalize;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Game_Paused( this : access Keen_View; paused : Boolean ) is
    begin
        this.Get_Widget( "pnlPaused" ).Set_Visible( paused );
    end On_Game_Paused;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Game_State_Changed( this          : access Keen_View;
                                     isPlaying     : Boolean;
                                     isInterrupted : Boolean ) is
        screen : A_Game_Screen;
    begin
        if isPlaying then
            -- A game has begun - - - - - -

            -- clear paused state of a previous game session
            this.Get_Widget( "pnlPaused" ).Set_Visible( False );
            this.pausedByPlayer := False;
            this.pausedByMenu := 0;

            -- the menu screen was removed when the event to start the game
            -- was queued.
        else
            -- A game has ended - - - - - -
            Queue_Stop_Music;

            if not isInterrupted then
                -- the game ended naturally by Keen losing all his lives, so
                -- put the menu back up. otherwise, the game session was
                -- interrupted by the user to start a new game or load a game,
                -- so we don't want to put the menu back up.
                screen := Create_Main_Menu_Screen( A_Game_View(this), "scrMainMenu",
                                                   gameInProgress => False );
                A_Screen_Manager(this.Get_Widget( "screenManager" )).Add_Screen( screen );
            end if;
        end if;
    end On_Game_State_Changed;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Loading_Begin( this : access Keen_View ) is
        screen : A_Game_Screen;
    begin
        Queue_Stop_Music;

        screen := Create_Message_Screen( this, "scrLoading", transparent => False );
        Message_Screen(screen.all).Set_Text_Align( Align_Right );
        Message_Screen(screen.all).Set_Text( "One moment" );
        Message_Screen(screen.all).Set_Icon( "loading_keen1" );
        A_Screen_Manager(this.Get_Widget( "screenManager" )).Add_Screen( screen );

        -- the loading screen won't hide for at least MIN_LOAD_DELAY time
        this.loadingScreen := True;
        this.loadingScreenComplete := Clock + MIN_LOAD_DELAY;
    end On_Loading_Begin;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Loading_End( this    : access Keen_View;
                              success : Boolean;
                              message : String ) is
        pragma Unreferenced( message );
    begin
        -- don't hide the loading screen if it isn't shown
        if not this.loadingScreen then
            return;
        end if;

        -- hide the loading screen
        if Clock >= this.loadingScreenComplete then
            this.loadingScreen := False;
            A_Game_Screen(this.Get_Widget( "scrLoading" )).Exit_Screen;

            -- start playing music for the next world
            if Length( this.worldMusic ) > 0 then
                Queue_Play_Music( To_String( this.worldMusic ) );
            end if;

            -- the loading screen is done so let the game know the view is ready
            -- to play
            this.Ready_For_Play;

        -- the level failed to load
        elsif not success then
            Message_Screen(this.Get_Widget( "scrLoading" ).all).Set_Icon( "talking_keen" );
            Message_Screen(this.Get_Widget( "scrLoading" ).all).Set_Text( "Keen folds his arms and refuses to enter" );
            Queue_Play_Sound( "no" );
        end if;
    end On_Loading_End;

    ----------------------------------------------------------------------------

    overriding
    procedure On_World_Property_Changed( this  : access Keen_View;
                                         name  : String;
                                         value : Value_Ptr'Class ) is
    begin
        if name = "music" then
            this.worldMusic := To_Unbounded_String( Cast_String( value ) );
        elsif name = "introduction" then
            if this.loadingScreen then
                Message_Screen(this.Get_Widget( "scrLoading" ).all).Set_Text( Cast_String( value ) );
            end if;
        end if;
    end On_World_Property_Changed;

    ----------------------------------------------------------------------------

    procedure Pause_By_Menu( this    : not null access Keen_View'Class;
                             enabled : Boolean ) is
    begin
        if enabled then
            -- a menu paused game play
            this.pausedByMenu := this.pausedByMenu + 1;
            if this.pausedByMenu = 1 and then not this.pausedByPlayer then
                this.Pause_Game( pause => True );
            end if;
        else
            -- a menu resumed game play
            this.pausedByMenu := this.pausedByMenu - 1;
            if this.pausedByMenu < 0 then
                this.pausedByMenu := 0;
                Dbg( "Error: Resume without pause", D_VIEW, ERROR );
            else
                if this.pausedByMenu = 0 and then not this.pausedByPlayer then
                    this.Pause_Game( pause => False );
                end if;
            end if;
        end if;
    end Pause_By_Menu;

    ----------------------------------------------------------------------------

    procedure Pause_By_Player( this    : not null access Keen_View'Class;
                               enabled : Boolean ) is
    begin
        -- player can only pause/resume if the menu hasn't paused game play
        if this.pausedByMenu = 0 then
            if this.pausedByPlayer /= enabled then
                this.pausedByPlayer := enabled;
                this.Pause_Game( pause => this.pausedByPlayer );
            end if;
        end if;
    end Pause_By_Player;

    ----------------------------------------------------------------------------

    procedure Tick( this : access Keen_View; time : Tick_Time ) is
    begin
        Game_View(this.all).Tick( time );

        -- handle the loading screen
        if not this.loading then
            -- loading is finished; is it time to take down the loading screen?
            if this.loadingScreen and then Clock >= this.loadingScreenComplete then
                -- loading is complete and the minimum delay has elapsed;
                this.On_Loading_End( True, "" );
            end if;
        end if;
    end Tick;

begin

    Register_Allocator( Create_Game_View'Access );

end Game_Views.Keen;
