--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Numerics.Discrete_Random;
with Ada.Strings.Equal_Case_Insensitive;
--with Debugging;                         use Debugging;
with Entities;                          use Entities;
with Events.Audio;                      use Events.Audio;
with Events.Entities;                   use Events.Entities;
with Events.Game;                       use Events.Game;
with Events.Keen;                       use Events.Keen;
with Events.World;                      use Events.World;
with Games;                             use Games;
with Games.Sessions;                    use Games.Sessions;
with Support.Paths;                     use Support.Paths;
with Values.Booleans;                   use Values.Booleans;
with Values.Casting;                    use Values.Casting;
with Values.Numbers;                    use Values.Numbers;
with Values.Strings;                    use Values.Strings;
with Worlds;                            use Worlds;
with Worlds.Keen;                       use Worlds.Keen;

pragma Elaborate_All( Games.Sessions );

package body Games.Sessions.Keen is

    MAP_WORLD      : constant String := "shadowlands";
    MAP_WORLD_NAME : constant String := "Shadowlands";

    ----------------------------------------------------------------------------

    function Next_Dialog_Id return Integer;

    ----------------------------------------------------------------------------

    -- Enters a level, raising an exception on error loading world 'name'.
    procedure Enter_Level( this : not null access Keen_Session'Class;
                           name : String );

    -- Exits a level, returning to the overhead map. If 'success' is True, the
    -- current level will be marked as complete in the session vars. An
    -- exception will be raised on error loading the world.
    procedure Exit_Level( this    : not null access Keen_Session'Class;
                          success : Boolean );

    procedure Handle_Dialog_Response( this : not null access Keen_Session'Class;
                                      evt  : not null A_Dialog_Response_Event );

    procedure Handle_Entity_Entered_Tile( this : not null access Keen_Session'Class;
                                          evt  : not null A_Entity_Tile_Event );

    procedure Handle_Give_Ammo( this : not null access Keen_Session'Class;
                                evt  : not null A_Give_Ammo_Event );

    procedure Handle_Give_Drops( this : not null access Keen_Session'Class;
                                 evt  : not null A_Give_Drops_Event );

    procedure Handle_Give_Points( this : not null access Keen_Session'Class;
                                  evt  : not null A_Give_Points_Event );

    procedure Handle_Player_Died( this : not null access Keen_Session'Class;
                                  evt  : not null A_Event );

    procedure Handle_Set_Tile( this : not null access Keen_Session'Class;
                               evt  : not null A_Set_Tile_Event );

    procedure One_Up( this : not null access Keen_Session'Class );

    -- Reloads the current level. An exception will be raised on error loading
    -- the world.
    procedure Retry_Level( this : not null access Keen_Session'Class );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Next_Dialog_Id return Integer is
        package Random_32 is new Ada.Numerics.Discrete_Random( Integer );
        use Random_32;
        g : Generator;
    begin
        Reset( g );
        return Random( g );
    end Next_Dialog_Id;

    --==========================================================================

    function Create_Keen_Session( game : not null A_Game ) return A_Game_Session is
        this : constant A_Game_Session := new Keen_Session;
    begin
        this.Construct( game );
        return this;
    end Create_Keen_Session;

    ----------------------------------------------------------------------------

    -- package private
    procedure Enter_Level( this : not null access Keen_Session'Class;
                           name : String ) is
    begin
        this.Set_Var( "level.won", Create( "" ) );
        this.Load_World( name, fresh => True );
    end Enter_Level;

    ----------------------------------------------------------------------------

    -- package private
    procedure Exit_Level( this    : not null access Keen_Session'Class;
                          success : Boolean ) is
    begin
        -- mark the level as complete
        if success and then this.Get_World /= null then
            this.Set_Var( "level." & Remove_Extension( Cast_String( this.Get_World.Get_Property( "filename" ) ) ), Create( True ) );
            this.Set_Var( "level.won", Create( Remove_Extension( Cast_String( this.Get_World.Get_Property( "filename" ) ) ) ) );
        else
            -- the level was not completed
            this.Set_Var( "level.won", Create( "" ) );
        end if;

        -- activate the overhead map world
        this.Load_World( MAP_WORLD, fresh => False );
    end Exit_Level;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Keen_Session;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
    begin
        if evt.Get_Id = ENTITY_ENTERED_TILE_ID then
            this.Handle_Entity_Entered_Tile( A_Entity_Tile_Event(evt) );

        elsif evt.Get_Id = GIVE_POINTS_ID then
            this.Handle_Give_Points( A_Give_Points_Event(evt) );

        elsif evt.Get_Id = GIVE_DROPS_ID then
            this.Handle_Give_Drops( A_Give_Drops_Event(evt) );

        elsif evt.Get_Id = GIVE_AMMO_ID then
            this.Handle_Give_Ammo( A_Give_Ammo_Event(evt) );

        elsif evt.Get_Id = SET_TILE_ID then
            this.Handle_Set_Tile( A_Set_Tile_Event(evt) );

        elsif evt.Get_Id = PLAYER_DIED_ID then
            this.Handle_Player_Died( evt );

        elsif evt.Get_Id = DIALOG_RESPONSE_ID then
            this.Handle_Dialog_Response( A_Dialog_Response_Event(evt) );

        else
            Game_Session(this.all).Handle_Event( evt, resp );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    procedure Handle_Dialog_Response( this : not null access Keen_Session'Class;
                                      evt  : not null A_Dialog_Response_Event ) is
    begin
        if evt.Get_Dialog_Id = this.dialogId then
            this.Pause( False );
            this.dialogId := Integer'First;
            if evt.Get_Choice = 1 then
                -- reload the current world
                this.Retry_Level;
            else
                -- load the overhead map
                this.Exit_Level( success => False );
            end if;
        end if;
    end Handle_Dialog_Response;

    ----------------------------------------------------------------------------

    procedure Handle_Entity_Entered_Tile( this : not null access Keen_Session'Class;
                                          evt  : not null A_Entity_Tile_Event ) is
        world : constant A_World := this.Get_World;

        function Get_Action_At( x, y : Integer ) return Action_Type is
            function Cast_Action_Type is new Cast_Enum( Action_Type );

            id              : Natural;
            action          : Action_Type;
            effectiveAction : Action_Type := NONE;
        begin
            if x >= 0 and then y >= 0 and then
               x < world.Get_Width_Tiles and then y < world.Get_Height_Tiles
            then
                for layer in 1..2 loop   -- only the first two are physical
                    id := world.Get_Tile_Id( layer, x, y );
                    if id /= 0 then
                        action := Cast_Action_Type( world.Get_Library.Get_Tile( id ).Get_Attribute( "action" ), NONE );
                        if effectiveAction < action then
                            effectiveAction := action;
                        end if;
                    end if;
                end loop;
            end if;
            return effectiveAction;
        end Get_Action_At;

    begin
        if evt.Get_Id = world.Get_Player.Get_Id then
            --Dbg( "Player touched" & evt.Get_X'Img & "," & evt.Get_Y'Img );
            if Get_Action_At( evt.Get_X, evt.Get_Y ) = HAZZARD then
                -- todo: send a "Die" one-time directive instead
                world.Get_Player.Die;
            end if;
        end if;
    end Handle_Entity_Entered_Tile;

    ----------------------------------------------------------------------------

    procedure Handle_Give_Ammo( this : not null access Keen_Session'Class;
                                evt  : not null A_Give_Ammo_Event ) is
        ammo : Number_Ptr := As_Number( this.Get_Var( "ammo" ) );
    begin
        -- minimum of zero
        ammo := ammo + evt.Get_Amount;
        if ammo < 0 then
            ammo := Create( 0 );
        end if;
        this.Set_Var( "ammo", ammo );
    end Handle_Give_Ammo;

    ----------------------------------------------------------------------------

    procedure Handle_Give_Drops( this : not null access Keen_Session'Class;
                                 evt  : not null A_Give_Drops_Event ) is
        drops : Number_Ptr := As_Number( this.Get_Var( "drops" ) );
    begin
        drops := drops + evt.Get_Amount;
        while drops >= 100 loop
            drops := drops - 100;
            this.One_Up;
        end loop;
        this.Set_Var( "drops", drops );
    end Handle_Give_Drops;

    ----------------------------------------------------------------------------

    procedure Handle_Give_Points( this : not null access Keen_Session'Class;
                                  evt  : not null A_Give_Points_Event ) is
    begin
        this.Set_Var( "points", As_Number( this.Get_Var( "points" ) ) + evt.Get_Amount );
        if this.Get_Var( "points" ) >= this.Get_Var( "nextLife" ) then
            this.One_Up;

            -- The value of 'nextLife' doubles each time
            this.Set_Var( "nextLife", As_Number( this.Get_Var( "nextLife" ) ) * 2 );
        end if;
    end Handle_Give_Points;

    ----------------------------------------------------------------------------

    procedure Handle_Player_Died( this : not null access Keen_Session'Class;
                                  evt  : not null A_Event ) is
        pragma Unreferenced( evt );
        dialog : A_Dialog_Desc;
    begin
        if As_Number( this.Get_Var( "lives" ) ) > 0 then
            -- display the level retry dialog
            this.Set_Var( "lives", As_Number( this.Get_Var( "lives" ) ) - 1 );
            this.Pause( True );

            dialog := Create_Dialog_Desc( "Retry", Cast_String( this.Get_World.Get_Property( "title" ) ) );
            dialog.Add_Choice( "Try Again" );
            dialog.Add_Choice( "Exit to " & MAP_WORLD_NAME );
            this.dialogId := Next_Dialog_Id;
            Queue_Dialog( this.dialogId, dialog );
        else
            -- all out of lives, game over!
            Queue_End_Game( completed => True );
        end if;
    end Handle_Player_Died;

    ----------------------------------------------------------------------------

    procedure Handle_Set_Tile( this : not null access Keen_Session'Class;
                               evt  : not null A_Set_Tile_Event ) is
    begin
        if this.Get_World /= null then
            this.Get_World.Set_Tile( evt.Get_Layer,
                                     evt.Get_X, evt.Get_Y,
                                     evt.Get_Tile_ID,
                                     notify => True );
        end if;
    end Handle_Set_Tile;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Load_World( this : access Keen_Session; filename : String ) is
    begin
        if filename'Length = 0 or else Ada.Strings.Equal_Case_Insensitive( Remove_Extension( filename ), MAP_WORLD ) then
            -- loading the empty string or the overhead map is equivalent to
            -- exiting the level.
            this.Exit_Level( success => True );
        else
            this.Enter_Level( filename );
        end if;
    end On_Load_World;

    ----------------------------------------------------------------------------

    overriding
    procedure On_New_Game( this : access Keen_Session ) is
    begin
        -- register to listen for events
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Dialog_Response" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Entity_Entered_Tile" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Give_Ammo" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Give_Drops" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Give_Points" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Player_Died" );
        this.game.Get_Corral.Add_Listener( A_Event_Listener(this), "Set_Tile" );

        -- setup initial game state
        this.Set_Var( "ammo",     Create( 5 ) );
        this.Set_Var( "drops",    Create( 0 ) );
        this.Set_Var( "points",   Create( 0 ) );
        this.Set_Var( "nextLife", Create( 20_000 ) );
        this.Set_Var( "lives",    Create( 3 ) );
        this.Set_Var( "ancients", Create( 0 ) );
        this.Set_Var( "scuba",    Create( False ) );

        -- levels that have not yet been beaten
        this.Set_Var( "level.border",    Create( False ) );
        this.Set_Var( "level.bwb",       Create( False ) );
        this.Set_Var( "level.pit",       Create( False ) );
        this.Set_Var( "level.slugville", Create( False ) );
        this.Set_Var( "level.won",       Create( "" ) );

        -- load the overhead map world to begin
        this.Load_World( MAP_WORLD, fresh => True );
    end On_New_Game;

    ----------------------------------------------------------------------------

    overriding
    procedure On_End_Game( this      : access Keen_Session;
                           completed : Boolean ) is
        pragma Unreferenced( completed );
    begin
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Set_Tile" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Player_Died" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Give_Points" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Give_Drops" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Give_Ammo" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Entity_Entered_Tile" );
        this.game.Get_Corral.Remove_Listener( A_Event_Listener(this), "Dialog_Response" );
    end On_End_Game;

    ----------------------------------------------------------------------------

    procedure One_Up( this : not null access Keen_Session'Class ) is
    begin
        this.Set_Var( "lives", As_Number( this.Get_Var( "lives" ) ) + 1 );
        Queue_Play_Sound( "get_1up" );
    end One_Up;

    ----------------------------------------------------------------------------

    procedure Retry_Level( this : not null access Keen_Session'Class ) is
    begin
        -- load the current world again
        this.Set_Var( "level.won", Create( "" ) );
        this.Load_World( Cast_String( this.Get_World.Get_Property( "filename" ) ), fresh => True );
    end Retry_Level;

begin

    Register_Allocator( Create_Keen_Session'Access );

end Games.Sessions.Keen;
