--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Characters.Handling;           use Ada.Characters.Handling;
--with Debugging;                         use Debugging;
with Events.Entities;                   use Events.Entities;
with Events.Game;                       use Events.Game;
with Events.World;                      use Events.World;
with Game_Views;                        use Game_Views;
with Game_Views.Keen;                   use Game_Views.Keen;
with Keen_Directives;
with Preferences;                       use Preferences;
with Support;                           use Support;
with Values.Casting;                    use Values.Casting;
with Values.Numbers;                    use Values.Numbers;
with Widgets.Containers;                use Widgets.Containers;
with Widgets.Containers.Game_Screens.Menus.Main;
 use Widgets.Containers.Game_Screens.Menus.Main;
with Widgets.Containers.Keen_Panels;    use Widgets.Containers.Keen_Panels;
with Widgets.Containers.Scenes;         use Widgets.Containers.Scenes;
with Widgets.Containers.Scenes.Keen;    use Widgets.Containers.Scenes.Keen;
with Widgets.Containers.Screen_Managers;
 use Widgets.Containers.Screen_Managers;
with Widgets.Labels;                    use Widgets.Labels;
with Widgets.Progress_Boards;           use Widgets.Progress_Boards;
with Widgets.Progress_Boards.Simple_Listeners;
with Widgets.Scoreboards;               use Widgets.Scoreboards;
with Widgets.Simple_Key_Listeners;

pragma Elaborate_All( Preferences );
pragma Elaborate_All( Widgets.Simple_Key_Listeners );
pragma Elaborate_All( Widgets.Progress_Boards.Simple_Listeners );

package body Widgets.Containers.Game_Screens.Game_Play is

    package Key_Listeners is new Widgets.Simple_Key_Listeners(Game_Play_Screen);
    use Key_Listeners;

    package PBoard_Listeners is new Widgets.Progress_Boards.Simple_Listeners(Game_Play_Screen);
    use PBoard_Listeners;

    -- Called when the Escape key is pressed to return to the main menu.
    procedure Action_Back_To_Menu( this   : not null access Game_Play_Screen'Class;
                                   action : A_Key_Action );

    -- This is called when the progress board closes.
    procedure Action_Progress_Board_Closed( this   : not null access Game_Play_Screen'Class;
                                            action : A_PBoard_Action  );

    -- This is called when the progress board opens.
    procedure Action_Progress_Board_Opened( this : not null access Game_Play_Screen'Class;
                                            action : A_PBoard_Action );

    -- Called when a registered scene key is pressed/released/held.
    procedure Action_Scene_Key( this   : not null access Game_Play_Screen'Class;
                                action : A_Key_Action );

    -- Called when the Follow_Entity event is received for the scene to track
    -- the player entity.
    procedure Handle_Follow_Entity( this : not null access Game_Play_Screen'Class;
                                    evt  : not null A_Entity_Event );

    -- Called when a game session variable changes. This updates the progress
    -- board widget.
    procedure Handle_Game_Var_Changed( this : not null access Game_Play_Screen'Class;
                                       evt  : not null A_Game_Var_Changed_Event );

    -- Called when the viewport on the scene needs to scroll separately from
    -- movement by the target entity (e.g. looking up or down.)
    procedure Handle_Scroll_View( this : not null access Game_Play_Screen'Class;
                                  evt  : not null A_Scroll_View_Event );

    -- Handles World_Loaded events to begin ignoring any keys being held down.
    -- This prevents in advertent actions on play resuming when a player is
    -- holds down a key during the load screen.
    procedure Handle_World_Loaded( this : not null access Game_Play_Screen'Class;
                                   evt  : not null A_World_Loaded_Event );

    -- Called when a property of the world changes.
    procedure Handle_World_Property_Changed( this : not null access Game_Play_Screen'Class;
                                             evt  : not null A_World_Property_Event );

    -- Reads the key bindings from preferences.
    procedure Load_Bindings( this : not null access Game_Play_Screen'Class );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Game_Play_Screen( view : not null access Game_Views.Game_View'Class;
                                      id   : String ) return A_Game_Screen is
        this : A_Game_Screen := new Game_Play_Screen;
    begin
        this.Construct( view, id );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Game_Play_Screen;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Game_Play_Screen;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
        scene         : A_Scene;
        scoreboard    : A_Scoreboard;
        progressBoard : A_Progress_Board;
        kpanel        : A_Keen_Panel;
        label         : A_Label;
    begin
        Game_Screen(this.all).Construct( view, id, popup => False );
        this.Load_Bindings;
        this.Add_Listener( Listener( this.key(MENU), Key_Press, this, Action_Back_To_Menu'Access ) );

        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Follow_Entity" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Game_Var_Changed" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Scroll_View" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "World_Loaded" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "World_Property_Changed" );

        -- Create child widgets --

        -- game scene
        scene := Create_Scene( view, this.id & ":scene" );
        scene.Set_Layout_LTRB( 0, 0, 0, 0 );
        scene.Add_Listener( Listener( this.key(MOVE_LEFT),  Key_Press,   this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(MOVE_LEFT),  Key_Release, this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(MOVE_RIGHT), Key_Press,   this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(MOVE_RIGHT), Key_Release, this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(MOVE_UP),    Key_Press,   this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(MOVE_UP),    Key_Release, this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(MOVE_DOWN),  Key_Press,   this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(MOVE_DOWN),  Key_Release, this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(JUMP),       Key_Press,   this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(JUMP),       Key_Release, this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(SHOOT),      Key_Press,   this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(POGO),       Key_Press,   this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(PROGRESS),   Key_Press,   this, Action_Scene_Key'Access ) );
        scene.Add_Listener( Listener( this.key(PAUSE),      Key_Press,   this, Action_Scene_Key'Access ) );
        this.Add_Child( A_Widget(scene) );

        -- scoreboard
        scoreboard := Create_Scoreboard( view, this.id & ":scoreboard", "interface" );
        scoreboard.Set_Layout_LTWH( 7, 4, 0, 0 );
        scoreboard.Set_Next( this.id & ":scene" );
        this.Add_Child( A_Widget(scoreboard) );

        -- paused dialog
        kpanel := Create_Panel( view, "pnlPaused", "interface" );
        kpanel.Set_Layout_Center( 64, 40 );
        kpanel.Set_Next( this.id & ":scene" );
        kpanel.Set_Visible( False );

            -- paused text label
            label := Create_Label( view, this.id & ":lblPaused", "Paused" );
            label.Set_Layout_Center( 0, 0 );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            label.Set_Font( "dos", 10 );
            kpanel.Add_Widget( A_Widget(label) );

        this.Add_Child( A_Widget(kpanel) );

        -- drop down progress board
        -- progress board (popup) (with transition)
        progressBoard := Create_Progress_Board( view, this.id & ":progressBoard", "interface" );
        progressBoard.Add_Listener( Listener( Opened, this, Action_Progress_Board_Opened'Access ) );
        progressBoard.Add_Listener( Listener( Closed, this, Action_Progress_Board_Closed'Access ) );
        progressBoard.Set_Next( this.id & ":scene" );
        this.Add_Child( A_Widget(progressBoard) );
    exception
        when others =>
            Delete( A_Widget(kpanel) );
            Delete( A_Widget(label) );
            Delete( A_Widget(progressBoard) );
            Delete( A_Widget(scoreboard) );
            Delete( A_Widget(scene) );
            raise;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Game_Play_Screen ) is
    begin
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "World_Property_Changed" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "World_Loaded" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Scroll_View" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Game_Var_Changed" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Follow_Entity" );
        Game_Screen(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Action_Back_To_Menu( this   : not null access Game_Play_Screen'Class;
                             action : A_Key_Action ) is
        pragma Unreferenced( action );
        screen : A_Game_Screen;
    begin
        A_Keen_View(this.Get_View).Pause_By_Menu( enabled => True );
        screen := Create_Main_Menu_Screen( this.Get_View, "scrMainMenu",
                                           gameInProgress => True );
        A_Screen_Manager(this.parent).Add_Screen( screen );
    end Action_Back_To_Menu;

    ----------------------------------------------------------------------------

    procedure Action_Progress_Board_Closed( this   : not null access Game_Play_Screen'Class;
                                            action : A_PBoard_Action ) is
        pragma Unreferenced( action );
    begin
        A_Keen_View(this.Get_View).Pause_By_Menu( enabled => False );
    end Action_Progress_Board_Closed;

    ----------------------------------------------------------------------------

    procedure Action_Progress_Board_Opened( this   : not null access Game_Play_Screen'Class;
                                            action : A_PBoard_Action ) is
        pragma Unreferenced( action );
    begin
        A_Keen_View(this.Get_View).Pause_By_Menu( enabled => True );
    end Action_Progress_Board_Opened;

    ----------------------------------------------------------------------------

    procedure Action_Scene_Key( this   : not null access Game_Play_Screen'Class;
                                action : A_Key_Action ) is
        aid : constant Action_Id := action.Get_Id;
        key : constant Integer := action.Get_Key;
    begin
        if aid = Key_Press then -- - - - - - - - - - - - - - - - - - - - - - - -
            -- events from a key are only ignored until it's pressed again
            this.ignoreKey(key) := False;

            if key = this.key(PROGRESS) then
                A_Progress_Board(this.Get_View.Get_Widget( this.id & ":progressBoard" )).Toggle;
            elsif key = this.key(PAUSE) then
                A_Keen_View(this.Get_View).Pause_By_Player( not this.Get_View.Is_Paused );
            elsif key = this.key(MOVE_LEFT) then
                Queue_Entity_Directive( this.playerId, Keen_Directives.LEFT, Ongoing );
            elsif key = this.key(MOVE_RIGHT) then
                Queue_Entity_Directive( this.playerId, Keen_Directives.RIGHT, Ongoing );
            elsif key = this.key(MOVE_UP) then
                Queue_Entity_Directive( this.playerId, Keen_Directives.UP, Ongoing );
            elsif key = this.key(MOVE_DOWN) then
                Queue_Entity_Directive( this.playerId, Keen_Directives.DOWN, Ongoing );
            elsif key = this.key(JUMP) then
                Queue_Entity_Directive( this.playerId, Keen_Directives.JUMP, Ongoing );
            elsif key = this.key(POGO) then
                Queue_Entity_Directive( this.playerId, Keen_Directives.POGO, Once );
            elsif key = this.key(SHOOT) then
                Queue_Entity_Directive( this.playerId, Keen_Directives.SHOOT, Once );
            end if;

        elsif aid = Key_Release then  -- - - - - - - - - - - - - - - - - - - - -
            if this.ignoreKey(key) then
                return;
            end if;

            -- if the key activates an on-going directive, cancel the directive
            -- on key release. if the game is paused then these directives will
            -- be ignored.
            if key = this.key(MOVE_LEFT) then
                Queue_Entity_Directive( this.playerId, Keen_Directives.LEFT, Inactive );
            elsif key = this.key(MOVE_RIGHT) then
                Queue_Entity_Directive( this.playerId, Keen_Directives.RIGHT, Inactive );
            elsif key = this.key(MOVE_UP) then
                Queue_Entity_Directive( this.playerId, Keen_Directives.UP, Inactive );
            elsif key = this.key(MOVE_DOWN) then
                Queue_Entity_Directive( this.playerId, Keen_Directives.DOWN, Inactive );
            elsif key = this.key(JUMP) then
                Queue_Entity_Directive( this.playerId, Keen_Directives.JUMP, Inactive );
            end if;

        end if;
    end Action_Scene_Key;

    ----------------------------------------------------------------------------

    overriding
    procedure Activate( this : access Game_Play_Screen ) is
    begin
        Game_Screen(this.all).Activate;
        this.Get_View.Get_Widget( this.id & ":scene" ).Request_Focus;
    end Activate;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Game_Play_Screen;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        if evt /= null then
            if evt.Get_Id = SCROLL_VIEW_ID then
                this.Handle_Scroll_View( A_Scroll_View_Event(evt) );

            elsif evt.Get_Id = GAME_VAR_CHANGED_ID then
                this.Handle_Game_Var_Changed( A_Game_Var_Changed_Event(evt) );

            elsif evt.Get_Id = FOLLOW_ENTITY_ID then
                this.Handle_Follow_Entity( A_Entity_Event(evt) );

            elsif evt.Get_Id = WORLD_LOADED_ID then
                this.Handle_World_Loaded( A_World_Loaded_Event(evt) );

            elsif evt.Get_Id = WORLD_PROPERTY_CHANGED_ID then
                this.Handle_World_Property_Changed( A_World_Property_Event(evt) );

            end if;
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    procedure Handle_Follow_Entity( this : not null access Game_Play_Screen'Class;
                                    evt  : not null A_Entity_Event ) is
    begin
        this.playerId := evt.Get_Id;
    end Handle_Follow_Entity;

    ----------------------------------------------------------------------------

    procedure Handle_Game_Var_Changed( this : not null access Game_Play_Screen'Class;
                                       evt  : not null A_Game_Var_Changed_Event ) is
        var : constant String := To_Lower( evt.Get_Var );
        val : constant Number_Ptr := As_Number( evt.Get_Value );
    begin
        -- update the GUI (ex: scoreboard or progress board) with the new game
        -- variable value. Should those widgets themselves listen for these
        -- events and handle them without the view intervening?
        if var = "points" then
            A_Scoreboard(this.Get_View.Get_Widget( this.id & ":scoreboard" )).Set_Points( val.Get.To_Int );
            A_Progress_Board(this.Get_View.Get_Widget( this.id & ":progressBoard" )).Set_Points( val.Get.To_Int );
        elsif var = "drops" then
            A_Progress_Board(this.Get_View.Get_Widget( this.id & ":progressBoard" )).Set_Drops( val.Get.To_Int );
        elsif var = "ammo" then
            A_Scoreboard(this.Get_View.Get_Widget( this.id & ":scoreboard" )).Set_Ammo( val.Get.To_Int );
            A_Progress_Board(this.Get_View.Get_Widget( this.id & ":progressBoard" )).Set_Ammo( val.Get.To_Int );
        elsif var = "lives" then
            A_Scoreboard(this.Get_View.Get_Widget( this.id & ":scoreboard" )).Set_Lives( val.Get.To_Int );
            A_Progress_Board(this.Get_View.Get_Widget( this.id & ":progressBoard" )).Set_Lives( val.Get.To_Int );
        elsif var = "nextlife" then
            A_Progress_Board(this.Get_View.Get_Widget( this.id & ":progressBoard" )).Set_Next_Life( val.Get.To_Int );
        end if;
    end Handle_Game_Var_Changed;

    ----------------------------------------------------------------------------

    procedure Handle_Scroll_View( this : not null access Game_Play_Screen'Class;
                                  evt  : not null A_Scroll_View_Event ) is
    begin
        A_Scene(this.Get_View.Get_Widget( this.id & ":scene" )).Scroll( evt.Get_X, evt.Get_Y );
    end Handle_Scroll_View;

    ----------------------------------------------------------------------------

    procedure Handle_World_Loaded( this : not null access Game_Play_Screen'Class;
                                   evt  : not null A_World_Loaded_Event ) is
        pragma Unreferenced( evt );
    begin
        -- begin ignoring all action keys until the next press event
        this.ignoreKey(this.key(MOVE_UP))    := True;
        this.ignoreKey(this.key(MOVE_DOWN))  := True;
        this.ignoreKey(this.key(MOVE_LEFT))  := True;
        this.ignoreKey(this.key(MOVE_RIGHT)) := True;
        this.ignoreKey(this.key(JUMP))       := True;
        this.ignoreKey(this.key(POGO))       := True;
        this.ignoreKey(this.key(SHOOT))      := True;
        this.ignoreKey(this.key(PROGRESS))   := True;
    end Handle_World_Loaded;

    ----------------------------------------------------------------------------

    procedure Handle_World_Property_Changed( this : not null access Game_Play_Screen'Class;
                                             evt  : not null A_World_Property_Event ) is
    begin
        if evt.Get_Property_Name = "title" then
            A_Progress_Board(this.Get_View.Get_Widget( this.id & ":progressBoard" )).Set_Location( Cast_String( evt.Get_Value ) );
        end if;
    end Handle_World_Property_Changed;

    ----------------------------------------------------------------------------

    procedure Load_Bindings( this : not null access Game_Play_Screen'Class ) is
    begin
        for i in Key_Purpose'Range loop
            this.key(i) := Constrain(
                Integer'( Preferences.Get_Pref( "keyboard",
                                                "binding." & Key_Purpose'Image( i ) ) ),
                1, ALLEGRO_KEY_MAX - 1
            );
        end loop;
    end Load_Bindings;

begin

    Preferences.Set_Default( "scaling_filter", "Auto" );

    Preferences.Set_Default( "keyboard", "binding.move_up",     ALLEGRO_KEY_UP );
    Preferences.Set_Default( "keyboard", "binding.move_down",   ALLEGRO_KEY_DOWN );
    Preferences.Set_Default( "keyboard", "binding.move_left",   ALLEGRO_KEY_LEFT );
    Preferences.Set_Default( "keyboard", "binding.move_right",  ALLEGRO_KEY_RIGHT );

    -- The Windows controls have been temporarily disabled because Allegro 5
    -- does not send repeat events for Ctrl or Alt keys, like normal keys. Since
    -- jumping relies on multiple Jump_Start messages being sent while the key
    -- is held down, it can't work with just the press and release messages.
    --
    -- The plan is to change the way input is handled by the game so that only
    -- Key_Press and Key_Release events are sigificant. They will register an
    -- ongoing impulse with the Game object and it will poll the state of the
    -- impulse as necessary, instead of continually a start and then a stop.
#if WINDOWS then
    -- Standard controls for jump and pogo
    Preferences.Set_Default( "keyboard", "binding.jump",        ALLEGRO_KEY_LCTRL );
    Preferences.Set_Default( "keyboard", "binding.pogo",        ALLEGRO_KEY_ALT );
#elsif OSX then
    -- Control and Alt can't be used on OS X because they're linked to changing
    -- the active desktop when combined with the arrow keys, by default.
    Preferences.Set_Default( "keyboard", "binding.jump",        ALLEGRO_KEY_Z );
    Preferences.Set_Default( "keyboard", "binding.pogo",        ALLEGRO_KEY_X );
#end if;
    Preferences.Set_Default( "keyboard", "binding.shoot",       ALLEGRO_KEY_SPACE );
    Preferences.Set_Default( "keyboard", "binding.progress",    ALLEGRO_KEY_ENTER );
    Preferences.Set_Default( "keyboard", "binding.pause",       ALLEGRO_KEY_P );
    Preferences.Set_Default( "keyboard", "binding.menu",        ALLEGRO_KEY_ESCAPE );
    Preferences.Set_Default( "keyboard", "binding.menu_next",   ALLEGRO_KEY_DOWN );
    Preferences.Set_Default( "keyboard", "binding.menu_prev",   ALLEGRO_KEY_UP );
    Preferences.Set_Default( "keyboard", "binding.menu_select", ALLEGRO_KEY_ENTER );

end Widgets.Containers.Game_Screens.Game_Play;
