--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Allegro.Keyboard;                  use Allegro.Keyboard;
--with Debugging;                         use Debugging;
with Game_Views;                        use Game_Views;
with Game_Views.Keen;                   use Game_Views.Keen;
with Palette_16;                        use Palette_16;
with Support;                           use Support;
with Widgets.Labels;                    use Widgets.Labels;
with Widgets.Containers.Keen_Panels;    use Widgets.Containers.Keen_Panels;

package body Widgets.Containers.Game_Screens.Message is

    MAX_LINES : constant := 3;

    -- Recalculates the lines of text, putting as many words as will fit onto
    -- each line.
    procedure Layout_Text( this : not null access Message_Screen'Class );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Message_Screen( view        : not null access Game_Views.Game_View'Class;
                                    id          : String;
                                    transparent : Boolean := True ) return A_Game_Screen is
        this : A_Game_Screen := new Message_Screen;
    begin
        this.Construct( view, id, transparent );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Message_Screen;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this        : access Message_Screen;
                         view        : not null access Game_Views.Game_View'Class;
                         id          : String;
                         transparent : Boolean ) is
        kpanel : A_Keen_Panel;
        label  : A_Label;
    begin
        Game_Screen(this.all).Construct( view, id, popup => transparent );
        this.Add_Listener( A_Key_Listener(this) );
        this.Set_Color( Background, Black );

        kpanel := Create_Panel( view, this.id & ":pnlMain", "interface" );
        kpanel.Set_Layout_Center( 222, 78 );

            -- icon
            label := Create_Label( view, this.id & ":lblIcon", icon=> "" );
            label.Set_Layout_LTWH( 1, 11, 0, 0 );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            kpanel.Add_Widget( A_Widget(label) );

            -- text line 1
            label := Create_Label( view, this.id & ":lblLine1" );
            label.Set_Layout_LTRB( 57, 13, 4, -22 );
            label.Set_Align( Align_Center );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            label.Set_Font( "dos", 10 );
            kpanel.Add_Widget( A_Widget(label) );

            -- text line 2
            label := Create_Label( view, this.id & ":lblLine2" );
            label.Set_Layout_LTRB( 57, 23, 4, -32 );
            label.Set_Align( Align_Center );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            label.Set_Font( "dos", 10 );
            kpanel.Add_Widget( A_Widget(label) );

            -- text line 3
            label := Create_Label( view, this.id & ":lblLine3" );
            label.Set_Layout_LTRB( 57, 33, 4, -42 );
            label.Set_Align( Align_Center );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            label.Set_Font( "dos", 10 );
            kpanel.Add_Widget( A_Widget(label) );

        this.Add_Child( A_Widget(kpanel) );
    exception
        when others =>
            Delete( A_Widget(label) );
            Delete( A_Widget(kpanel) );
            raise;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Activate( this : access Message_Screen ) is
    begin
        Game_Screen(this.all).Activate;
        if this.autoPause then
            A_Keen_View(this.Get_View).Pause_By_Menu( enabled => True );
        end if;
    end Activate;

    ----------------------------------------------------------------------------

    overriding
    procedure Deactivate( this : access Message_Screen ) is
    begin
        Game_Screen(this.all).Deactivate;
        if this.autoPause then
            A_Keen_View(this.Get_View).Pause_By_Menu( enabled => False );
        end if;
    end Deactivate;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Action( this    : access Message_Screen;
                             action  : A_Key_Action;
                             handled : out Boolean ) is
    begin
        if this.closable then
            if action.Get_Key = ALLEGRO_KEY_ENTER or else
               action.Get_Key = ALLEGRO_KEY_SPACE or else
               action.Get_Key = ALLEGRO_KEY_ESCAPE
            then
                this.Exit_Screen;
            end if;
        end if;
        handled := True;
    end Handle_Action;

    ----------------------------------------------------------------------------

    procedure Layout_Text( this : not null access Message_Screen'Class ) is
        theme     : constant A_Theme := Themes.Get;
        line      : Positive := 1;
        label     : A_Label := A_Label(this.Get_View.Get_Widget( this.Get_Id & ":lblLine" & Image( line ) ));
        fontName  : constant String := label.Get_Font_Name;
        fontSize  : constant Positive := label.Get_Font_Size;
        lineWidth : constant Natural := label.Get_Content_Width;
        lineText  : Unbounded_String;

        ------------------------------------------------------------------------

        procedure Add_Word( word : String ) is
        begin
            -- add the word to the current line if we're on the last line or
            -- nothing is on this line yet or the text will fit on this line.
            if line = MAX_LINES or else
               Length( lineText ) = 0 or else
               theme.Get_Text_Width( To_String( lineText ) & ' ' & word, fontName, fontSize ) <= lineWidth
            then
                -- add word to the current line
                if Length( lineText ) = 0 then
                    Append( lineText, word );
                else
                    Append( lineText, ' ' & word );
                end if;
            else
                -- complete current line and add word to the next line
                label.Set_Text( To_String( lineText ) );
                lineText := To_Unbounded_String( word );
                line := line + 1;
                label := A_Label(this.Get_View.Get_Widget( this.Get_Id & ":lblLine" & Image( line ) ));
            end if;
        end Add_Word;

        ------------------------------------------------------------------------

    begin
        Iterate_Words( To_String( this.text ), Add_Word'Access );

        if Length( lineText ) > 0 then
            -- complete the phrase to the current line
            label.Set_Text( To_String( lineText ) );
            line := line + 1;
        end if;

        -- clear any remaining lines
        for i in line..MAX_LINES loop
            A_Label(this.Get_View.Get_Widget( this.Get_Id & ":lblLine" & Image( i ) )).Set_Text( "" );
        end loop;
    end Layout_Text;

    ----------------------------------------------------------------------------

    overriding
    procedure Pack( this : access Message_Screen ) is
    begin
        Game_Screen(this.all).Pack;
        this.Layout_Text;
    end Pack;

    ----------------------------------------------------------------------------

    procedure Set_Auto_Pause( this    : not null access Message_Screen'Class;
                              enabled : Boolean ) is
    begin
        this.autoPause := enabled;
    end Set_Auto_Pause;

    ----------------------------------------------------------------------------

    procedure Set_Closable( this     : not null access Message_Screen'Class;
                            closable : Boolean ) is
    begin
        this.closable := closable;
    end Set_Closable;

    ----------------------------------------------------------------------------

    procedure Set_Icon( this : not null access Message_Screen'Class; filename : String ) is
    begin
        A_Label(this.Get_View.Get_Widget( this.id & ":lblIcon" )).Set_Icon( filename );
    end Set_Icon;

    ----------------------------------------------------------------------------

    procedure Set_Text( this : not null access Message_Screen'Class; text : String ) is
    begin
        this.text := To_Unbounded_String( text );
        this.Layout_Text;
    end Set_Text;

    ----------------------------------------------------------------------------

    procedure Set_Text_Align( this  : not null access Message_Screen'Class;
                              align : Align_Type ) is
    begin
        if align /= this.textAlign then
            this.textAlign := align;

            case this.textAlign is
                when Align_Left =>
                    for i in 1..MAX_LINES loop
                        this.Get_View.Get_Widget( this.Get_Id & ":lblLine1" ).Set_Layout_LTRB( 4, 13, 57, -22 );
                        this.Get_View.Get_Widget( this.Get_Id & ":lblLine2" ).Set_Layout_LTRB( 4, 23, 57, -32 );
                        this.Get_View.Get_Widget( this.Get_Id & ":lblLine3" ).Set_Layout_LTRB( 4, 33, 57, -42 );
                        this.Get_View.Get_Widget( this.Get_Id & ":lblIcon" ).Set_Layout_LTWH( -1, 11, 0, 0 );
                    end loop;
                when Align_Center =>
                    for i in 1..MAX_LINES loop
                        this.Get_View.Get_Widget( this.Get_Id & ":lblLine1" ).Set_Layout_LTRB( 4, 13, 4, -22 );
                        this.Get_View.Get_Widget( this.Get_Id & ":lblLine2" ).Set_Layout_LTRB( 4, 23, 4, -32 );
                        this.Get_View.Get_Widget( this.Get_Id & ":lblLine3" ).Set_Layout_LTRB( 4, 33, 4, -42 );
                        -- don't move the icon; it should be off anyway. no room.
                    end loop;
                when Align_Right =>
                    this.Get_View.Get_Widget( this.Get_Id & ":lblLine1" ).Set_Layout_LTRB( 57, 13, 4, -22 );
                    this.Get_View.Get_Widget( this.Get_Id & ":lblLine2" ).Set_Layout_LTRB( 57, 23, 4, -32 );
                    this.Get_View.Get_Widget( this.Get_Id & ":lblLine3" ).Set_Layout_LTRB( 57, 33, 4, -42 );
                    this.Get_View.Get_Widget( this.Get_Id & ":lblIcon" ).Set_Layout_LTWH( 1, 11, 0, 0 );
            end case;

            this.Layout_Text;      -- re-adjust text in lines
        end if;
    end Set_Text_Align;

end Widgets.Containers.Game_Screens.Message;
