--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;

package body Widgets.Containers.Screen_Managers is

    function Create_Screen_Manager( view : not null access Game_Views.Game_View'Class;
                                    id   : String ) return A_Screen_Manager is
        this : A_Screen_Manager := new Screen_Manager;
    begin
        this.Construct( view, id );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Screen_Manager;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Screen_Manager;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
    begin
        Container(this.all).Construct( view, id );
    end Construct;

    ----------------------------------------------------------------------------

    procedure Add_Screen( this   : not null access Screen_Manager'Class;
                          screen : in out A_Game_Screen ) is
        use Widget_Lists;
        crsr : Cursor;
    begin
        pragma Debug( Dbg( "Added screen " & screen.To_String, D_GUI, INFO ) );
        screen.Set_Layout_LTRB( 0, 0, 0, 0 );
        screen.Set_Visible( False );
        if not screen.Is_Popup then
            -- if this is an opaque screen being added, then hide all the
            -- currently active screens behind it. there may be more than one if
            -- the front screen is a popup.
            crsr := this.children.Last;
            while Has_Element( crsr ) loop
                if A_Game_Screen(Element( crsr )).Is_Active then
                    A_Game_Screen(Element( crsr ) ).Deactivate;
                else
                    -- there are no more active screens
                    exit;
                end if;
                Previous( crsr );
            end loop;
        end if;
        this.Add_Child( A_Widget(screen), consume => False );
        if this.Is_Rooted then
            screen.Activate;
        end if;
        screen := null;
    exception
        when others =>
            screen := null;
            raise;
    end Add_Screen;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Shown( this : access Screen_Manager; shown : Boolean ) is
        use Widget_Lists;
        crsr : Widget_Lists.Cursor;
    begin
        if shown then
            crsr := this.children.Last;
            while Has_Element( crsr ) loop
                if A_Game_Screen(Element( crsr )).Is_Popup then
                    -- activate all the popups in front
                    A_Game_Screen(Element( crsr )).Activate;
                else
                    -- activate the first non-popup and leave
                    A_Game_Screen(Element( crsr )).Activate;
                    exit;
                end if;
                Previous( crsr );
            end loop;
        end if;
    end On_Shown;

    ----------------------------------------------------------------------------

    procedure Remove_Screen( this   : not null access Screen_Manager'Class;
                             screen : not null A_Game_Screen ) is
        use Widget_Lists;

        crsr    : Widget_Lists.Cursor;
        screen2 : A_Game_Screen;
    begin
        screen.Deactivate;
        this.Remove_Child( A_Widget(screen) );
        pragma Debug( Dbg( "Removed screen " & screen.To_String, D_GUI, INFO ) );

        -- activate the screens on the front of the screen manager, starting
        -- with the first opaque (non-popup) screen, activating forward to the
        -- top. this requires traversing the screens from the front toward the
        -- back to the first opaque screen, then traversing back toward the
        -- front, activating each screen along the way.
        crsr := this.children.Last;
        while Has_Element( crsr ) loop
            screen2 := A_Game_Screen(Element( crsr ));

            if not screen2.Is_Popup then
                -- traverse from here back to the front, activating screens
                while Has_Element( crsr ) loop
                    screen2 := A_Game_Screen(Element( crsr ));
                    screen2.Activate;
                    Next( crsr );
                end loop;
                exit;
            end if;

            -- screen is a popup, keep going backward
            Previous( crsr );
        end loop;
    end Remove_Screen;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Screen_Manager; time : Tick_Time ) is
        use Widget_Lists;
        pragma Unreferenced( time );

        updated : Boolean := True;
        crsr    : Widget_Lists.Cursor;
        screen  : A_Game_Screen;
    begin
        while updated loop
            updated := False;
            crsr := this.children.First;
            while Has_Element( crsr ) loop
                screen := A_Game_Screen(Element( crsr ));

                -- if the screen is exiting, remove it and re-iterate through
                -- the now-changed children list to look for more changes.
                if screen.Is_Exiting then
                    this.Remove_Screen( screen );
                    Delete( A_Widget(screen) );
                    updated := True;
                    exit;              -- loop back through .children
                end if;

                Next( crsr );
            end loop;
        end loop;
    end Tick;

end Widgets.Containers.Screen_Managers;
