--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Allegro.Keyboard;                  use Allegro.Keyboard;
with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Engine_Debugging;                  use Engine_Debugging;
with Palette_16;

package body Widgets.Menu_Enumerations is

    SPACING : constant Integer := 2;

    BLINK_DELAY : constant Time_Span := Milliseconds( 500 );

    ON_ICON       : constant String := "menu_dot_on";
    OFF_ICON      : constant String := "menu_dot_off";
    DISABLED_ICON : constant String := "menu_dot_disabled";

    procedure Dispatch_Enum_Action( this  : not null access Menu_Enumeration'Class;
                                    id    : Action_Id;
                                    index : Natural;
                                    text  : String );

    procedure Set_Icon( this : not null access Menu_Enumeration'Class; icon : String );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Menu_Enum( view : not null access Game_Views.Game_View'Class;
                               id   : String ) return A_Menu_Enumeration is
        this : A_Menu_Enumeration := new Menu_Enumeration;
    begin
        this.Construct( view, id );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Menu_Enum;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Menu_Enumeration;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
    begin
        Widget(this.all).Construct( view, id );
        this.border := Border_None;
        this.focusTime := Time_Span_Zero;
        this.Set_Color( Themes.Text, Palette_16.Green );
        this.Set_Color( Themes.Selected, Palette_16.Dark_Grey );
        this.Set_Color( Themes.Foreground, Palette_16.Dark_Grey );
        this.Set_Font( "menu", 10 );
        this.Set_Icon( OFF_ICON );
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add_Listener( this     : access Menu_Enumeration;
                            listener : not null A_Enum_Listener ) is
    begin
        this.Add_Listener( "Enum_Listener", A_Action_Listener(listener) );
    end Add_Listener;

    ----------------------------------------------------------------------------

    procedure Add_Option( this : not null access Menu_Enumeration'Class;
                          text : String ) is
    begin
        this.options.Append( text );
        if this.index = 0 then
            this.Set_Index( 1, fireAction => False );
        end if;
    end Add_Option;

    ----------------------------------------------------------------------------

    procedure Dispatch_Enum_Action( this  : not null access Menu_Enumeration'Class;
                                    id    : Action_Id;
                                    index : Natural;
                                    text  : String ) is
        action : A_Enum_Action := new Enum_Action;

        procedure Dispatch( listener : A_Action_Listener ) is
        begin
            A_Enum_Listener(listener).Handle_Action( action );
        end Dispatch;

    begin
        action.Construct( id, A_Widget(this), index, text );
        this.Iterate_Listeners( "Enum_Listener", Dispatch'Access );
        Delete( A_Object(action) );
    end Dispatch_Enum_Action;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Menu_Enumeration ) is
        theme     : constant A_Theme := Themes.Get;
        icon      : constant A_Allegro_Bitmap := theme.Get_Bitmap( this.icon );
        border    : constant Border_Type := this.border;
        bgColor   : constant Allegro_Color := this.colors(Foreground);
        textColor : constant Allegro_Color := this.colors(Text);
        bw        : constant Natural := theme.Border_Width;
        pad       : Natural := 0;
    begin
        if not this.transparent then
            theme.Draw_Box( 0, 0, this.cwidth - 1, this.cheight - 1, bgColor, border );
        end if;

        if this.focused and then this.border /= Border_None then
            Rect( bw, bw, this.cwidth - 1 - bw, this.cheight - 1 - bw,
                  this.colors(Selected) );
        end if;
        if this.border /= Border_None then
            pad := bw + SPACING;
        end if;
        theme.Draw_Label( pad, pad,
                          this.cwidth - 1 - pad,
                          this.cheight - 1 - pad,
                          icon, this.Get_Text,
                          To_String( this.fontName ), this.fontSize,
                          textColor, Align_Left );
    end Draw_Content;

    ----------------------------------------------------------------------------

    function Get_Index( this : not null access Menu_Enumeration'Class ) return Natural is
    begin
        return this.index;
    end Get_Index;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Height( this : access Menu_Enumeration ) return Natural is
        theme : A_Theme;
        icon  : A_Allegro_Bitmap;
        pad   : Natural := 0;
    begin
        if this.minHeight > 0 then
            return this.minHeight;
        else
            theme := Themes.Get;
            icon := theme.Get_Bitmap( this.icon );
            if this.border /= Border_None then
                pad := theme.Border_Width * 2;
            end if;
            return pad + theme.Label_Height( icon, this.Get_Text,
                                             To_String( this.fontName ), this.fontSize );
        end if;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Width( this : access Menu_Enumeration ) return Natural is
        theme : A_Theme;
        icon  : A_Allegro_Bitmap;
        pad   : Natural := 0;
    begin
        if this.minWidth > 0 then
            return this.minWidth;
        else
            theme := Themes.Get;
            icon := theme.Get_Bitmap( this.icon );
            if this.border /= Border_None then
                pad := theme.Border_Width * 2;
            end if;
            return pad + theme.Label_Width( icon, this.Get_Text,
                                            To_String( this.fontName ),
                                            this.fontSize );
        end if;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    function Get_Text( this : not null access Menu_Enumeration'Class ) return String is
    begin
        if this.index > 0 and then this.index <= Natural(this.options.Length) then
            return this.options.Element( this.index );
        end if;
        return "";
    end Get_Text;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Blur( this : access Menu_Enumeration ) is
    begin
        Widget(this.all).On_Blur;
        this.Set_Color( Themes.Text, Palette_16.Green );
        this.Set_Icon( OFF_ICON );
    end On_Blur;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Disabled( this : access Menu_Enumeration ) is
    begin
        this.Set_Icon( DISABLED_ICON );
    end On_Disabled;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Enabled( this : access Menu_Enumeration ) is
    begin
        this.Set_Icon( OFF_ICON );
    end On_Enabled;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Focus( this : access Menu_Enumeration ) is
    begin
        this.Set_Color( Themes.Text, Palette_16.Light_Green );
    end On_Focus;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Key_Press( this    : access Menu_Enumeration;
                            evt     : not null A_Key_Event;
                            handled : in out Boolean ) is
    begin
        if evt.Get_Key = ALLEGRO_KEY_DOWN then
            if this.Focus_Next then
                handled := True;
            end if;
        elsif evt.Get_Key = ALLEGRO_KEY_UP then
            if this.Focus_Prev then
                handled := True;
            end if;
        elsif evt.Get_Key = ALLEGRO_KEY_RIGHT then
            if this.index < Natural(this.options.Length) then
                this.Set_Index( this.index + 1 );
            else
                this.Set_Index( 1 );
            end if;
            handled := True;
        elsif evt.Get_Key = ALLEGRO_KEY_LEFT then
            if this.index > 1 then
                this.Set_Index( this.index - 1 );
            else
                this.Set_Index( Natural(this.options.Length) );
            end if;
            handled := True;
        elsif evt.Get_Key = ALLEGRO_KEY_TAB then
            -- prevent the default tab-focus-switching behavior
            handled := True;
        end if;
    end On_Key_Press;

    ----------------------------------------------------------------------------

    procedure Set_Icon( this : not null access Menu_Enumeration'Class; icon : String ) is
    begin
        if icon'Length > 0 then
            this.icon := Themes.Get.Get_ID( icon );
            pragma Debug( this.icon = 0,
                          Dbg( "Warning: Icon <" & icon & "> not found for widget " &
                               To_String( this ),
                               D_GUI, WARNING ) );
        else
            this.icon := 0;
        end if;
    end Set_Icon;

    ----------------------------------------------------------------------------

    procedure Set_Index( this       : not null access Menu_Enumeration'Class;
                         index      : Positive;
                         fireAction : Boolean := True ) is
    begin
        if index /= this.index and then index <= Natural(this.options.Length) then
            this.index := index;
            if fireAction then
                this.Dispatch_Enum_Action( Changed, this.index, this.Get_Text );
            end if;
        end if;
    end Set_Index;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Menu_Enumeration; time : Tick_Time ) is
    begin
        if this.focused then
            if ((time.total - this.focusTime) / BLINK_DELAY) mod 2 = 0 then
                this.Set_Icon( ON_ICON );
            else
                this.Set_Icon( OFF_ICON );
            end if;
        else
            this.focusTime := time.total;
        end if;
    end Tick;

    --==========================================================================

    not overriding
    procedure Construct( this   : access Enum_Action;
                         id     : Action_Id;
                         source : not null A_Widget;
                         index  : Natural;
                         text   : String ) is
    begin
        Action(this.all).Construct( id, source );
        this.index := index;
        this.text := To_Unbounded_String( text );
    end Construct;

    ----------------------------------------------------------------------------

    function Get_Index( this : not null access Enum_Action'Class ) return Natural is
    begin
        return this.index;
    end Get_Index;

    ----------------------------------------------------------------------------

    function Get_Text( this : not null access Enum_Action'Class ) return String is
    begin
        return To_String( this.text );
    end Get_Text;

end Widgets.Menu_Enumerations;
