--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Allegro.Keyboard;                  use Allegro.Keyboard;
--with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Events.Audio;                      use Events.Audio;
with Palette_16;                        use Palette_16;
with Resources;                         use Resources;
with Tiles.Libraries.Manager;           use Tiles.Libraries.Manager;

package body Widgets.Paddle_Wars is

    BALL_SPEED        : constant := 65.0;
    BALL_RESET_DELAY  : constant Time_Span := Seconds( 1 );

    KEEN_PADDLE_SPEED : constant := 80.0;
    COMP_PADDLE_SPEED : constant := 40.0;

    KEEN_SCORE_X : constant := 31;
    KEEN_SCORE_Y : constant := 2;

    COMP_SCORE_X : constant := 137;
    COMP_SCORE_Y : constant := 2;

    ----------------------------------------------------------------------------

    -- Resets the ball to the middle of the field of play and zeros its
    -- horizontal motion, but maintains the direction of its vertical motion.
    -- This occurs each time a player scores.
    procedure Reset_Ball( this : not null access Paddle_War'Class );

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Paddle_War( view    : not null access Game_Views.Game_View'Class;
                                id      : String;
                                libName : String ) return A_Paddle_War is
        this : A_Paddle_War := new Paddle_War;
    begin
        this.Construct( view, id, libName );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Paddle_War;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this    : access Paddle_War;
                         view    : not null access Game_Views.Game_View'Class;
                         id      : String;
                         libName : String ) is

        ------------------------------------------------------------------------

        -- Returns the tile ID for a tile name matching 'filename' in 'lib'. If
        -- no tile with a matching name can be found, a RESOURCE_ERROR will be
        -- raised.
        function Get_Tile_Id( lib      : A_Tile_Library;
                              filename : String ) return Natural is
            id : constant Natural := lib.Get_Id( filename );
        begin
            if id = 0 then
                raise RESOURCE_ERROR with "Error loading tile <" & filename &
                                          "> from library <" & lib.Get_Name & ">";
            end if;
            return id;
        end Get_Tile_Id;

        ------------------------------------------------------------------------

    begin
        Widget(this.all).Construct( view, id );
        this.Set_Color( Themes.Text, Palette_16.Light_Green );
        this.Set_Font( "menu", 10 );

        this.lib := Load_Library( libName, bitmaps => True );
        if this.lib /= null then
            this.backgroundId := Get_Tile_Id( this.lib, "paddlewar.png" );
            this.paddleId := Get_Tile_Id( this.lib, "paddle.png" );
            this.ballId := Get_Tile_Id( this.lib, "ball.png" );
        else
            raise RESOURCE_ERROR with "Error loading tile library <" & libName & ">";
        end if;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Paddle_War ) is
    begin
        Unload_Library( this.lib );
        Widget(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Paddle_War ) is
        theme : constant A_Theme := Themes.Get;
        bmp : A_Allegro_Bitmap;
    begin
        if this.lib = null then
            return;
        end if;

        -- draw the background
        Draw_Bitmap( this.lib.Get_Tile( this.backgroundId ).Get_Bitmap, 0, 0 );

        -- draw the scores
        Draw_String( Trim( Integer'Image( this.keenScore ), Left ),
                     KEEN_SCORE_X, KEEN_SCORE_Y,
                     theme.Get_Font( To_String( this.fontName ), this.fontSize ),
                     this.colors(Text) );
        Draw_String( Trim( Integer'Image( this.compScore ), Left ),
                     COMP_SCORE_X, COMP_SCORE_Y,
                     theme.Get_Font( To_String( this.fontName ), this.fontSize ),
                     this.colors(Text) );

        -- draw the paddles
        bmp := this.lib.Get_Tile( this.paddleId ).Get_Bitmap;
        Draw_Bitmap( bmp,
                     Integer(this.compPlayer.x - this.compPlayer.w2),
                     Integer(this.compPlayer.y - this.compPlayer.h2) );
        Draw_Bitmap( bmp,
                     Integer(this.keenPlayer.x - this.keenPlayer.w2),
                     Integer(this.keenPlayer.y - this.keenPlayer.h2) );

        -- draw the ball (if it's in or nearly inbounds)
        if this.ball.y + this.ball.h2 >= this.boundsY1 and then
           this.ball.y - this.ball.h2 <= this.boundsY2
        then
            Draw_Bitmap( this.lib.Get_Tile( this.ballId ).Get_Bitmap,
                         Integer(this.ball.x - this.ball.w2),
                         Integer(this.ball.y - this.ball.h2) );
        end if;
    end Draw_Content;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Height( this : access Paddle_War ) return Natural is
    begin
        if this.minHeight > 0 then
            return this.minHeight;
        elsif this.lib /= null and then this.backgroundId /= 0 then
            return Al_Get_Bitmap_Height( this.lib.Get_Tile( this.backgroundId ).Get_Bitmap );
        else
            return 0;
        end if;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Width( this : access Paddle_War ) return Natural is
    begin
        if this.minWidth > 0 then
            return this.minWidth;
        elsif this.lib /= null and then this.backgroundId /= 0 then
            return Al_Get_Bitmap_Width( this.lib.Get_Tile( this.backgroundId ).Get_Bitmap );
        else
            return 0;
        end if;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Key_Press( this    : access Paddle_War;
                            evt     : not null A_Key_Event;
                            handled : in out Boolean ) is
    begin
        if evt.Get_Key = ALLEGRO_KEY_LEFT then
            this.keenPlayer.xv := -KEEN_PADDLE_SPEED;
            handled := True;
        elsif evt.Get_Key = ALLEGRO_KEY_RIGHT then
            this.keenPlayer.xv := KEEN_PADDLE_SPEED;
            handled := True;
        end if;
    end On_Key_Press;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Key_Release( this    : access Paddle_War;
                              evt     : not null A_Key_Event;
                              handled : in out Boolean ) is
    begin
        if evt.Get_Key = ALLEGRO_KEY_LEFT then
            this.keenPlayer.xv := 0.0;
            handled := True;
        elsif evt.Get_Key = ALLEGRO_KEY_RIGHT then
            this.keenPlayer.xv := 0.0;
            handled := True;
        end if;
    end On_Key_Release;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Shown( this : access Paddle_War; shown : Boolean ) is
        pragma Unreferenced( shown );
    begin
        -- Automatically reset the game every time it is displayed after being
        -- hidden.
        this.Reset;
    end On_Shown;

    ----------------------------------------------------------------------------

    procedure Reset( this : not null access Paddle_War'Class ) is
        bmp : A_Allegro_Bitmap;
    begin
        -- get the size of the paddles
        bmp := this.lib.Get_Tile( this.paddleId ).Get_Bitmap;
        this.compPlayer.w := Float(Al_Get_Bitmap_Width( bmp ));
        this.compPlayer.h := Float(Al_Get_Bitmap_Height( bmp ));
        this.compPlayer.w2 := this.compPlayer.w / 2.0;
        this.compPlayer.h2 := this.compPlayer.h / 2.0;

        this.keenPlayer.w := Float(Al_Get_Bitmap_Width( bmp ));
        this.keenPlayer.h := Float(Al_Get_Bitmap_Height( bmp ));
        this.keenPlayer.w2 := this.keenPlayer.w / 2.0;
        this.keenPlayer.h2 := this.keenPlayer.h / 2.0;

        -- get the size of the ball
        bmp := this.lib.Get_Tile( this.ballId ).Get_Bitmap;
        this.ball.w := Float(Al_Get_Bitmap_Width( bmp ));
        this.ball.h := Float(Al_Get_Bitmap_Height( bmp ));
        this.ball.w2 := this.ball.w / 2.0;
        this.ball.h2 := this.ball.h / 2.0;

        -- determine the boundaries for the game
        this.boundsX1 := 3.0;
        this.boundsY1 := 16.0;
        this.boundsX2 := Float(this.Get_Content_Width - 3);
        this.boundsY2 := Float(this.Get_Content_Height - 13);

        -- reset the players
        this.compPlayer.x := this.boundsX1 + (this.boundsX2 - this.boundsX1) / 2.0;
        this.compPlayer.y := this.boundsY1;
        this.compPlayer.xv := 0.0;

        this.keenPlayer.x := this.compPlayer.x;
        this.keenPlayer.y := this.boundsY2;
        this.keenPlayer.xv := 0.0;

        -- reset the ball
        this.Reset_Ball;
        this.ball.yv := BALL_SPEED;   -- set the initial ball speed
    end Reset;

    ----------------------------------------------------------------------------

    procedure Reset_Ball( this : not null access Paddle_War'Class ) is
    begin
        this.ball.x := this.boundsX1 + (this.boundsX2 - this.boundsX1) / 2.0;
        this.ball.y := this.boundsY1 + (this.boundsY2 - this.boundsY1) / 2.0;
        this.ball.xv := 0.0;
        this.outOfBounds := False;
        this.nextBall := Time_Span_Last;
    end Reset_Ball;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Paddle_War; time : Tick_Time ) is
    begin
        -- move the human player's paddle
        this.keenPlayer.x := this.keenPlayer.x + this.keenPlayer.xv * Float(To_Duration( time.elapsed ));
        if this.keenPlayer.x < this.boundsX1 + this.keenPlayer.w2 then
            this.keenPlayer.x := this.boundsX1 + this.keenPlayer.w2;
        elsif this.keenPlayer.x > this.boundsX2 - this.keenPlayer.w2 then
            this.keenPlayer.x := this.boundsX2 - this.keenPlayer.w2;
        end if;

        -- move the computer player's paddle
        this.compPlayer.x := this.compPlayer.x + this.compPlayer.xv * Float(To_Duration( time.elapsed ));
        if this.outOfBounds then
            -- the ball is out of bounds, don't chase it
            this.compPlayer.xv := 0.0;
        else
            if this.ball.x >= this.compPlayer.x then
                -- chase to the right
                if this.ball.x - this.ball.w > this.compPlayer.x then
                    -- chase at top speed
                    this.compPlayer.xv := COMP_PADDLE_SPEED;
                else
                    -- try to keep up with the ball
                    this.compPlayer.xv := Float'Min( this.ball.xv, COMP_PADDLE_SPEED );
               end if;
            else
                -- chase to the left
                if this.ball.x + this.ball.w < this.compPlayer.x then
                    -- chase at top speed
                   this.compPlayer.xv := -COMP_PADDLE_SPEED;
                else
                    -- try to keep up with the ball
                    this.compPlayer.xv := Float'Max( this.ball.xv, -COMP_PADDLE_SPEED );
                end if;
            end if;
        end if;

        -- move the ball
        this.ball.y := this.ball.y + this.ball.yv * Float(To_Duration( time.elapsed ));
        this.ball.x := this.ball.x + this.ball.xv * Float(To_Duration( time.elapsed ));

        -- check if the ball bounces off the sides
        if this.ball.x < 0.0 then
            if this.ball.x - this.ball.w2 < this.boundsX1 then
                Queue_Play_Sound( "paddle_wall" );
                this.ball.xv := -this.ball.xv;
                this.ball.x := this.boundsX1 + this.ball.w2;
            end if;
        else
            if this.ball.x + this.ball.w2 > this.boundsX2 then
                Queue_Play_Sound( "paddle_wall" );
                this.ball.xv := -this.ball.xv;
                this.ball.x := this.boundsX2 - this.ball.w2;
            end if;
        end if;

        -- check of the ball goes out of bounds or hits a paddle
        if not this.outOfBounds then
            if this.ball.yv < 0.0 then
                -- check if the computer player missed it
                if this.ball.y < this.compPlayer.y then
                    Queue_Play_Sound( "paddle_score" );
                    this.keenScore := this.keenScore + 1;
                    this.outOfBounds := True;
                    this.nextBall := time.total + BALL_RESET_DELAY;
                -- check for collision with the computer player
                elsif this.ball.y - this.ball.h2 < this.compPlayer.y + this.compPlayer.h2 then        -- vertical
                    if this.ball.x + this.ball.w2 >= this.compPlayer.x - this.compPlayer.w2 and then  -- horizontal
                       this.ball.x - this.ball.w2 <= this.compPlayer.x + this.compPlayer.w2
                    then
                        -- change direction of the ball
                        Queue_Play_Sound( "paddle_hit2" );
                        this.ball.yv := -this.ball.yv;
                        this.ball.xv := BALL_SPEED * (this.ball.x - this.compPlayer.x) / (this.ball.w2 + this.compPlayer.w2);
                        this.ball.y := this.compPlayer.y + this.compPlayer.h2 + this.ball.h2;
                    end if;
                end if;
            else
                -- check if the human player missed it
                if this.ball.y > this.keenPlayer.y then
                    Queue_Play_Sound( "paddle_miss" );
                    this.compScore := this.compScore + 1;
                    this.outOfBounds := True;
                    this.nextBall := time.total + BALL_RESET_DELAY;
                -- check for collision with the human player
                elsif this.ball.y + this.ball.h2 > this.keenPlayer.y - this.keenPlayer.h2 then        -- vertical
                    if this.ball.x + this.ball.w2 >= this.keenPlayer.x - this.keenPlayer.w2 and then  -- horizontal
                       this.ball.x - this.ball.w2 <= this.keenPlayer.x + this.keenPlayer.w2
                    then
                        -- change direction of the ball
                        Queue_Play_Sound( "paddle_hit1" );
                        this.ball.yv := -this.ball.yv;
                        this.ball.xv := BALL_SPEED * (this.ball.x - this.keenPlayer.x) / (this.ball.w2 + this.keenPlayer.w2);
                        this.ball.y := this.keenPlayer.y - this.keenPlayer.h2 - this.ball.h2;
                    end if;
                end if;
            end if;
        end if;

        -- reset the ball if it's time
        if this.nextBall <= time.total then
            this.Reset_Ball;
        end if;
    end Tick;

end Widgets.Paddle_Wars;
