--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Actions;

private with Tiles;
private with Tiles.Libraries;

pragma Warnings( Off, Actions );
pragma Elaborate_All( Actions );

package Widgets.Progress_Boards is

    type PBoard_Action is new Action with private;
    type A_PBoard_Action is access all PBoard_Action'Class;

    Opened : constant Action_Id;
    Closed : constant Action_Id;

    ----------------------------------------------------------------------------

    -- This is the interface to implement in order to listen to progress board
    -- actions. When an action occurs, the listener will be notified of the type
    -- of action performed and the progress board that performed the action.
    type PBoard_Listener is limited interface and Action_Listener;
    type A_PBoard_Listener is access all PBoard_Listener'Class;

    procedure Handle_Action( this   : access PBoard_Listener;
                             action : A_PBoard_Action ) is abstract;

    ----------------------------------------------------------------------------

    type Progress_Board is new Widget and Animated with private;
    type A_Progress_Board is access all Progress_Board'Class;

    function Create_Progress_Board( view    : not null access Game_Views.Game_View'Class;
                                    id      : String;
                                    libName : String ) return A_Progress_Board;
    pragma Precondition( id'Length > 0 );
    pragma Precondition( libName'Length > 0 );
    pragma Postcondition( Create_Progress_Board'Result /= null );

    procedure Add_Listener( this     : access Progress_Board;
                            listener : not null A_PBoard_Listener );

    -- Returns true if the board is not currently closed. It may be fully open
    -- or sliding in either direction.
    function Is_Open( this : access Progress_Board ) return Boolean;

    procedure Remove_Listener( this     : access Progress_Board;
                               listener : not null A_PBoard_Listener );

    procedure Set_Ammo( this : not null access Progress_Board'Class; ammo : Natural );

    procedure Set_Ancients( this : not null access Progress_Board'Class; ancients : Natural );

    procedure Set_Blue_Key( this : not null access Progress_Board'Class; blueKey : Boolean );

    procedure Set_Drops( this : not null access Progress_Board'Class; drops : Natural );

    procedure Set_Green_Key( this : not null access Progress_Board'Class; greenKey : Boolean );

    procedure Set_Lives( this : not null access Progress_Board'Class; lives : Natural );

    procedure Set_Location( this : not null access Progress_Board'Class; location : String );

    procedure Set_Next_Life( this : not null access Progress_Board'Class; nextLife : Natural );

    procedure Set_Points( this : not null access Progress_Board'Class; points : Natural );

    procedure Set_Red_Key( this : not null access Progress_Board'Class; redKey : Boolean );

    procedure Set_Scuba( this : not null access Progress_Board'Class; scuba : Boolean );

    procedure Set_Yellow_Key( this : not null access Progress_Board'Class; yellowKey : Boolean );

    procedure Show( this : access Progress_Board; enabled : Boolean );

    -- Toggles the open/closed state of the progress board. Calling this is
    -- equivalent to this.Show( not this.Is_Open ).
    procedure Toggle( this : access Progress_Board );

private

    use Tiles;
    use Tiles.Libraries;

    ----------------------------------------------------------------------------

    type PBoard_Action is new Action with null record;

    Opened : constant Action_Id := To_Action_Id( "progress_board.opened" );
    Closed : constant Action_Id := To_Action_Id( "progress_board.closed" );

    ----------------------------------------------------------------------------

    type Progress_Board is new Widget and Animated with
        record
            lib          : A_Tile_Library := null;    -- may be null if library cant load
            backgroundId : Natural := 0;              -- id of background tile
            digitId      : Tile_Id_Array(0..9) := (others => 0);  -- tile ids for digit icons
            ammo         : Natural := 0;
            ancients     : Natural := 0;
            blueKey      : Boolean := False;
            drops        : Natural := 0;
            greenKey     : Boolean := False;
            lives        : Natural := 0;
            location     : Unbounded_String;
            points       : Natural := 0;
            nextLife     : Natural := 0;
            redKey       : Boolean := False;
            scuba        : Boolean := False;
            yellowKey    : Boolean := False;
            ancientId,
            redKeyId,
            blueKeyId,
            yellowKeyId,
            greenKeyId   : Natural := 0;   -- tile ids for key icons
            slideDelay   : Float := 0.3;   -- seconds to fully slide up/down
            slideDir     : Integer := 0;   -- <0 = up, >0 = down
            slidePos     : Float := -1.0;  -- bottom of the board within its parent
        end record;

    procedure Construct( this    : access Progress_Board;
                         view    : not null access Game_Views.Game_View'Class;
                         id      : String;
                         libName : String );
    pragma Precondition( id'Length > 0 );
    pragma Precondition( libName'Length > 0 );

    procedure Delete( this : in out Progress_Board );

    procedure Dispatch_Action( this : access Progress_Board; id : Action_Id );

    procedure Draw_Content( this : access Progress_Board );

    procedure Tick( this : access Progress_Board; time : Tick_Time );

end Widgets.Progress_Boards;
