--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Allegro.Displays;                  use Allegro.Displays;
--with Debugging;                         use Debugging;
with Drawing;                           use Drawing;
with Interfaces;                        use Interfaces;
with Resources;                         use Resources;
with Support;                           use Support;
with Tiles.Libraries.Manager;           use Tiles.Libraries.Manager;

package body Widgets.Scoreboards is

    DIGIT_WIDTH : constant := 8;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Scoreboard( view    : not null access Game_Views.Game_View'Class;
                                id      : String;
                                libName : String ) return A_Scoreboard is
        this : A_Scoreboard := new Scoreboard;
    begin
        this.Construct( view, id, libName );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Scoreboard;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this    : access Scoreboard;
                         view    : not null access Game_Views.Game_View'Class;
                         id      : String;
                         libName : String ) is
    begin
        Widget(this.all).Construct( view, id );
        this.acceptFocus := False;
        this.lib := Load_Library( libName, bitmaps => True );
        if this.lib /= null then
            for i in 0..9 loop
                this.digitId(i) := this.lib.Get_Id( "digit" & Image( i ) & ".png" );
            end loop;
            this.backgroundId := this.lib.Get_Id( "scoreboard.png" );
            if this.backgroundId > 0 then
                this.Set_Min_Width( Al_Get_Bitmap_Width( this.lib.Get_Tile( this.backgroundId ).Get_Bitmap ) );
                this.Set_Min_Height( Al_Get_Bitmap_Height( this.lib.Get_Tile( this.backgroundId ).Get_Bitmap ) );
            end if;
        else
            raise RESOURCE_ERROR with "Error loading tile library <" & libName & ">";
        end if;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Scoreboard ) is
    begin
        Unload_Library( this.lib );
        Widget(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Scoreboard ) is

        ------------------------------------------------------------------------

        -- Draws a number on the scoreboard, given the location of the upper
        -- left corner of the most significant digit and the maximum number of
        -- digits to draw. If the number needs more than the maximum number of
        -- digits, the largest number possible will be drawn using the maximum
        -- number of digits. (ex: digits = 2, n = 100, draws "99")
        procedure Draw_Number( number : Natural;
                               x, y   : Integer;
                               dgts   : Positive ) is
            number2 : Unsigned_64 := Unsigned_64(number);
            n       : Unsigned_64;
            exp     : Unsigned_64;
            digit   : Natural;
        begin
            if number2 >= 10 ** dgts then
                number2 := 10 ** dgts - 1;
            end if;
            n := number2;

            for i in reverse 0..dgts-1 loop
                exp := 10 ** i;
                if exp <= number2 or else i = 0 then
                    digit := Natural((n - (n mod exp)) / exp);
                    Draw_Bitmap( this.lib.Get_Tile( this.digitId(digit) ).Get_Bitmap,
                                 x + ((dgts - 1) - i) * DIGIT_WIDTH, y );
                    n := n - Unsigned_64(digit) * exp;
                end if;
            end loop;
        end Draw_Number;

        ------------------------------------------------------------------------

        bmp : A_Allegro_Bitmap;
    begin
        if this.lib = null then
            return;
        end if;

        Al_Hold_Bitmap_Drawing( True );

        -- draw the background
        bmp := this.lib.Get_Tile( this.backgroundId ).Get_Bitmap;
        if bmp /= null then
            Draw_Bitmap( bmp, 0, 0 );
        end if;

        Draw_Number( this.points, 6, 4, dgts => 9 );
        Draw_Number( this.lives, 22, 20, dgts => 2 );
        Draw_Number( this.ammo, 62, 20, dgts => 2 );

        Al_Hold_Bitmap_Drawing( False );
    end Draw_Content;

    ----------------------------------------------------------------------------

    procedure Set_Ammo( this : not null access Scoreboard'Class; ammo : Natural ) is
    begin
        if this.ammo /= ammo then
            this.ammo := ammo;
        end if;
    end Set_Ammo;

    ----------------------------------------------------------------------------

    procedure Set_Lives( this : not null access Scoreboard'Class; lives : Natural ) is
    begin
        if this.lives /= lives then
            this.lives := lives;
        end if;
    end Set_Lives;

    ----------------------------------------------------------------------------

    procedure Set_Points( this : not null access Scoreboard'Class; points : Natural ) is
    begin
        if this.points /= points then
            this.points := points;
        end if;
    end Set_Points;

end Widgets.Scoreboards;
