--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
--with Debugging;                         use Debugging;
with Entities.Poles;                    use Entities.Poles;
with Entities.Sprites.Keen4;            use Entities.Sprites.Keen4;
with Events.Audio;                      use Events.Audio;
with Events.Entities;                   use Events.Entities;
with Events.Game;                       use Events.Game;
with Keen_Directives;
with Physics;
with Values.Booleans;                   use Values.Booleans;
with Values.Numbers;                    use Values.Numbers;
with Worlds;                            use Worlds;

package body Entities.Players.Keen.States is

    -- Distance that Keen should move (in either axis) while dying, before death
    DEATH_DISTANCE : constant Float := 200.0;

    -- The distance from the bottom of a pole sprite where Keen is considered to
    -- be at the bottom. (see the .poleBottom field) Useful for determining when
    -- to stop Keen's movement to prevent him from falling off the bottom.
    POLE_BOTTOM_BUFFER : constant := 2.0;

    -- Speed of blaster shots in world units per second
    SHOT_SPEED : constant Float := 160.0;

    ----------------------------------------------------------------------------

    type Frame_Type is
        record
            filename : String_Access := null;
            id       : Integer := -1;
        end record;

    type Frame_Array is array (Integer range <>) of Frame_Type;
    type A_Frame_Array is access all Frame_Array;

    type Direction_Array is array (Direction_8 range <>) of A_Frame_Array;
    type A_Direction_Array is access all Direction_Array;

    type Animation( directional : Boolean ) is
        record
            looping    : Boolean;
            frameDelay : Time_Span;
            case directional is
                when True =>
                    dir : A_Direction_Array := null;
                when False =>
                    frames : A_Frame_Array := null;
            end case;
        end record;
    type A_Animation is access all Animation;

    ----------------------------------------------------------------------------

    type Visible_Action is
    (
        Standing,
        Standing_and_Shooting,
        Looking_Frames,
        Walking,
        Jumping,                -- jumping upward
        Jumping_Midair,         -- hanging in midair during a jump
        Falling,                -- falling downward
        Jumping_and_Shooting,   -- shooting in the air (jumping or falling)
        Pogoing_Midair,
        Pogoing_On_Ground,
        Hanging_From_Pole,
        Hanging_From_Pole_And_Shooting,
        Climbing_Pole,
        Hanging_From_Ledge,
        Climbing_Ledge,
        Looking_Bored,
        Checking_Watch_Bored,
        Shrugging_Bored,
        Sitting_Bored,
        Reading_Bored,
        Rising_Bored,
        Entering_Doorway,
        Flipping_Switch,
        Dying_Frames
    );

    ----------------------------------------------------------------------------

    type Idle_Animation is
        record
            visibleAction : Visible_Action;
            idleDelay     : Time_Span;
            anmDuration   : Time_Span;
        end record;

    type Idle_Animation_Array is array (Positive range <>) of Idle_Animation;

    INDEX_LOOKING_BORED   : constant := 1;      -- blinking at the camera
    INDEX_CHECKING_WATCH  : constant := 2;      -- checking his watch
    INDEX_SHRUGGING_BORED : constant := 3;      -- shrugging his shoulders
    INDEX_SITTING_BORED   : constant := 4;      -- sitting down to read
    INDEX_READING_BORED   : constant := 5;      -- reading a book while sitting

    idle_animations : constant Idle_Animation_Array := Idle_Animation_Array'(
        INDEX_LOOKING_BORED   => (Looking_Bored,        idleDelay => Milliseconds( 4000 ), anmDuration => Milliseconds( 700 )),
        INDEX_CHECKING_WATCH  => (Checking_Watch_Bored, idleDelay => Milliseconds( 7700 ), anmDuration => Milliseconds( 1000 )),
        INDEX_SHRUGGING_BORED => (Shrugging_Bored,      idleDelay => Milliseconds( 11750 ), anmDuration => Milliseconds( 125 ) * 35),
        INDEX_SITTING_BORED   => (Sitting_Bored,        idleDelay => Milliseconds( 19000 ), anmDuration => Milliseconds( 115 ) * 4),
        INDEX_READING_BORED   => (Reading_Bored,        idleDelay => Milliseconds( 19000 ), anmDuration => Seconds( 84600 ))
    );

    ----------------------------------------------------------------------------

    type Action_Frames is array (Visible_Action) of A_Animation;

    frames : constant Action_Frames := Action_Frames'(
        Standing => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("stand_left"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("stand_right"), others => <>)))),
        Standing_and_Shooting => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("shoot_left"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("shoot_right"), others => <>)),
                D8_Up    => new Frame_Array'(
                    1 => (filename => new String'("shoot_up"), others => <>)))),
        Looking_Frames => new Animation'(
            directional => True,
            looping     => False,
            frameDelay  => Milliseconds( 180 ),
            dir         => new Direction_Array'(
                D8_Up   => new Frame_Array'(
                    1 => (filename => new String'("look_up"), others => <>)),
                D8_Down => new Frame_Array'(
                    1 => (filename => new String'("look_down1"), others => <>),
                    2 => (filename => new String'("look_down2"), others => <>)))),
        Walking => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    (filename => new String'("walk_left1"), others => <>),
                    (filename => new String'("walk_left2"), others => <>),
                    (filename => new String'("walk_left3"), others => <>),
                    (filename => new String'("walk_left4"), others => <>)),
                D8_Right => new Frame_Array'(
                    (filename => new String'("walk_right1"), others => <>),
                    (filename => new String'("walk_right2"), others => <>),
                    (filename => new String'("walk_right3"), others => <>),
                    (filename => new String'("walk_right4"), others => <>)))),
        Jumping => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("jump_left1"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("jump_right1"), others => <>)))),
        Jumping_Midair => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("jump_left2"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("jump_right2"), others => <>)))),
        Falling => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("jump_left3"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("jump_right3"), others => <>)))),
        Jumping_and_Shooting => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left       => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_left"), others => <>)),
                D8_Right      => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_right"), others => <>)),
                D8_Up         => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_up"), others => <>)),
                D8_Up_Left    => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_up"), others => <>)),
                D8_Up_Right   => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_up"), others => <>)),
                D8_Down       => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_down"), others => <>)),
                D8_Down_Left  => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_down"), others => <>)),
                D8_Down_Right => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_down"), others => <>)))),
        Pogoing_Midair => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("pogo_left_air"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("pogo_right_air"), others => <>)))),
        Pogoing_On_Ground => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("pogo_left_ground"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("pogo_right_ground"), others => <>)))),
        Hanging_From_Pole => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("pole_left1"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("pole_right1"), others => <>)))),
        Hanging_From_Pole_And_Shooting => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left       => new Frame_Array'(
                    1 => (filename => new String'("shoot_pole_left"), others => <>)),
                D8_Right      => new Frame_Array'(
                    1 => (filename => new String'("shoot_pole_right"), others => <>)),
                D8_Up         => null,
                D8_Down       => null,
                D8_Down_Left  => new Frame_Array'(
                    1 => (filename => new String'("shoot_pole_down_left"), others => <>)),
                D8_Down_Right => new Frame_Array'(
                    1 => (filename => new String'("shoot_pole_down_right"), others => <>)),
                D8_Up_Left    => new Frame_Array'(
                    1 => (filename => new String'("shoot_pole_up_left"), others => <>)),
                D8_Up_Right   => new Frame_Array'(
                    1 => (filename => new String'("shoot_pole_up_right"), others => <>)))),
        Climbing_Pole => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    (filename => new String'("pole_left1"), others => <>),
                    (filename => new String'("pole_left2"), others => <>),
                    (filename => new String'("pole_left3"), others => <>)),
                D8_Right => new Frame_Array'(
                    (filename => new String'("pole_right1"), others => <>),
                    (filename => new String'("pole_right2"), others => <>),
                    (filename => new String'("pole_right3"), others => <>)),
                D8_Up    => null,
                D8_Down  => new Frame_Array'(
                    (filename => new String'("pole_down1"), others => <>),
                    (filename => new String'("pole_down2"), others => <>),
                    (filename => new String'("pole_down3"), others => <>),
                    (filename => new String'("pole_down4"), others => <>)))),
        Hanging_From_Ledge => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("ledge_left1"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("ledge_right1"), others => <>)))),
        Climbing_Ledge => new Animation'(
            directional => True,
            looping     => False,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    (filename => new String'("ledge_left2"), others => <>),
                    (filename => new String'("ledge_left3"), others => <>),
                    (filename => new String'("ledge_left4"), others => <>),
                    (filename => new String'("ledge_left5"), others => <>)),
                D8_Right => new Frame_Array'(
                    (filename => new String'("ledge_right2"), others => <>),
                    (filename => new String'("ledge_right3"), others => <>),
                    (filename => new String'("ledge_right4"), others => <>),
                    (filename => new String'("ledge_right5"), others => <>)))),
        Looking_Bored => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => idle_animations(INDEX_LOOKING_BORED).anmDuration,
            frames      => new Frame_Array'(
                1 => (filename => new String'("look_up"), others => <>))),
        Shrugging_Bored => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => idle_animations(INDEX_SHRUGGING_BORED).anmDuration / 35,
            frames      => new Frame_Array'(
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug2"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug2"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug1"), others => <>),
                (filename => new String'("bored_shrug3"), others => <>),
                (filename => new String'("bored_shrug3"), others => <>),
                (filename => new String'("bored_shrug3"), others => <>),
                (filename => new String'("bored_shrug3"), others => <>),
                (filename => new String'("bored_shrug3"), others => <>),
                (filename => new String'("bored_shrug3"), others => <>),
                (filename => new String'("bored_shrug3"), others => <>),
                (filename => new String'("bored_shrug3"), others => <>),
                (filename => new String'("bored_shrug3"), others => <>))),
        Checking_Watch_Bored => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => idle_animations(INDEX_CHECKING_WATCH).anmDuration,
            frames      => new Frame_Array'(
                1 => (filename => new String'("bored_watch"), others => <>))),
        Sitting_Bored => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => idle_animations(INDEX_SITTING_BORED).anmDuration / 4,
            frames      => new Frame_Array'(
                (filename => new String'("bored_sit1"), others => <>),
                (filename => new String'("bored_sit2"), others => <>),
                (filename => new String'("bored_sit3"), others => <>),
                (filename => new String'("bored_sit4"), others => <>))),
        Reading_Bored => new Animation'(
            directional => False,
            looping     => True,
            frameDelay  => Milliseconds( 250 ),
            frames      => new Frame_Array'(
                (filename => new String'("bored_read1"), others => <>),
                (filename => new String'("bored_read1"), others => <>),
                (filename => new String'("bored_read1"), others => <>),
                (filename => new String'("bored_read1"), others => <>),
                (filename => new String'("bored_read1"), others => <>),
                (filename => new String'("bored_read1"), others => <>),
                (filename => new String'("bored_read1"), others => <>),
                (filename => new String'("bored_read1"), others => <>),
                (filename => new String'("bored_read1"), others => <>),
                (filename => new String'("bored_read1"), others => <>),
                (filename => new String'("bored_read1"), others => <>),
                (filename => new String'("bored_read1"), others => <>),
                (filename => new String'("bored_read2"), others => <>),
                (filename => new String'("bored_read3"), others => <>))),
        Rising_Bored => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => Milliseconds( 125 ),
            frames      => new Frame_Array'(
                1 => (filename => new String'("bored_rise1"), others => <>),
                2 => (filename => new String'("bored_rise2"), others => <>))),
        Entering_Doorway => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => TELEPORT_DURATION / 5,
            frames      => new Frame_Array'(
                (filename => new String'("doorway1"), others => <>),
                (filename => new String'("doorway2"), others => <>),
                (filename => new String'("doorway3"), others => <>),
                (filename => new String'("doorway4"), others => <>),
                (filename => new String'("doorway5"), others => <>))),
        Flipping_Switch => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => Milliseconds( 115 ),
            frames      => new Frame_Array'(
                1 => (filename => new String'("switch"), others => <>))),
        Dying_Frames => new Animation'(
            directional => False,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            frames      => new Frame_Array'(
                1 => (filename => new String'("die1"), others => <>)))
    );

    ----------------------------------------------------------------------------

    function To_D8( xv : Float; dir : Direction_Type ) return Direction_8 is
    begin
        if xv < 0.0 then
            return To_D8( dir + Left );
        elsif xv > 0.0 then
            return To_D8( dir + Right );
        end if;
        return To_D8( dir );
    end To_D8;
    pragma Inline_Always( To_D8 );

    ----------------------------------------------------------------------------

    function To_X( xv : Float; dir : Direction_Type ) return Direction_8 is
    begin
        if xv < 0.0 then
            return D8_Left;
        elsif xv > 0.0 then
            return D8_Right;
        end if;
        return To_X( dir );
    end To_X;
    pragma Inline_Always( To_X );

    ----------------------------------------------------------------------------

    function Get_Frame_Id( lib       : not null A_Tile_Library;
                           action    : Visible_Action;
                           actionAge : Time_Span;
                           dir       : Direction_8 := D8_Up;
                           reversed  : Boolean := False ) return Natural is

        --------------------------------------------------------------------

        procedure Resolve_Id( id : out Natural; frame : in out Frame_Type ) is
        begin
            if frame.id < 0 then
                frame.id := lib.Get_Id( frame.filename.all );
            end if;
            id := frame.id;
        end Resolve_Id;

        --------------------------------------------------------------------

        frameArray : access Frame_Array;
        frameNum   : Natural := 0;
        id         : Natural := 0;
    begin
        if frames(action).directional then
            frameArray := frames(action).dir(dir);
        else
            frameArray := frames(action).frames;
        end if;

        if frameArray'Length = 1 then
            frameNum := 0;
        else
            if frames(action).looping then
                frameNum := (actionAge / frames(action).frameDelay) mod frameArray'Length;
            else
                frameNum := Natural'Min( (actionAge / frames(action).frameDelay), frameArray'Length - 1 );
            end if;
        end if;

        if reversed then
            Resolve_Id( id, frameArray(frameArray'Last - frameNum) );
        else
            Resolve_Id( id, frameArray(frameArray'First + frameNum) );
        end if;

        return id;
    end Get_Frame_Id;

    --==========================================================================

    -- singleton state instances
    MOVING       : constant A_State := new Moving_State'(action => Act_Moving);
    LOOKING      : constant A_State := new Looking_State'(action => Act_Looking);
    LOOKING_BACK : constant A_State := new Looking_Back_State'(action => Act_Looking_Back);
    TELEPORTING  : constant A_State := new Teleporting_State'(action => Act_Teleporting);
    IDLE         : constant A_State := new Idle_State'(action => Act_Idle);
    RISING       : constant A_State := new Rising_State'(action => Act_Rising);
    DYING        : constant A_State := new Dying_State'(action => Act_Dying);
    CLIMBING     : constant A_State := new Climbing_State'(action => Act_Climbing);

    --==========================================================================

    function Get_State( action : Action_Type ) return A_State is
    begin
        case action is
            when Act_Moving       => return MOVING;
            when Act_Looking      => return LOOKING;
            when Act_Looking_Back => return LOOKING_BACK;
            when Act_Teleporting  => return TELEPORTING;
            when Act_Idle         => return IDLE;
            when Act_Rising       => return RISING;
            when Act_Dying        => return DYING;
            when Act_Climbing     => return CLIMBING;
        end case;
    end Get_State;

    --==========================================================================

    function Get_Frame( this : Moving_State; k : Keen'Class ) return Natural is
        pragma Unreferenced( this );
    begin
        if k.onPogo then
            -- on the pogostick, in the air or on the ground
            if k.grounded then
                return Get_Frame_Id( k.lib, Pogoing_On_Ground, k.age - k.stateStart, To_X( k.xv, k.dir ) );
            else
                return Get_Frame_Id( k.lib, Pogoing_Midair, k.age - k.stateStart, To_X( k.xv, k.dir ) );
            end if;
        elsif k.grounded then
            -- standing or walking on the ground
            if k.shooting then
                if k.dir and Up then
                    -- shooting upward
                    return Get_Frame_Id( k.lib, Standing_and_Shooting, k.age - k.stateStart, To_Y( k.dir ) );
                else
                    -- shooting left or right
                    return Get_Frame_Id( k.lib, Standing_and_Shooting, k.age - k.stateStart, To_X( k.xv, k.dir ) );
                end if;
            elsif k.xv = 0.0 then
                return Get_Frame_Id( k.lib, Standing, k.age - k.stateStart, To_X( k.xv, k.dir ) );
            else
                return Get_Frame_Id( k.lib, Walking, k.age - k.stateStart, To_X( k.xv, k.dir ) );
            end if;
        else
            -- in the air
            if k.shooting then
                return Get_Frame_Id( k.lib, Jumping_and_Shooting, k.age - k.stateStart, To_D8( k.xv, k.dir ) );
            elsif k.yv < 0.0 then
                return Get_Frame_Id( k.lib, Jumping, k.age - k.stateStart, To_X( k.xv, k.dir ) );
            elsif k.yv < Get_Keen_Pref( "falling_speed" ) then    -- todo: cache k value
                return Get_Frame_Id( k.lib, Jumping_Midair, k.age - k.stateStart, To_X( k.xv, k.dir ) );
            else
                return Get_Frame_Id( k.lib, Falling, k.age - k.stateStart, To_X( k.xv, k.dir ) );
            end if;
        end if;
    end Get_Frame;

    ----------------------------------------------------------------------------

    function Get_Frame( this : Looking_State; k : Keen'Class ) return Natural is
        pragma Unreferenced( this );
    begin
        -- looking up or down
        if k.shooting then
            return Get_Frame_Id( k.lib, Standing_and_Shooting, k.age - k.stateStart, To_Y( k.dir ) );
        else
            return Get_Frame_Id( k.lib, Looking_Frames, k.age - k.stateStart, To_Y( k.dir ), reversed => False );
        end if;
    end Get_Frame;

    ----------------------------------------------------------------------------

    function Get_Frame( this : Looking_Back_State; k : Keen'Class ) return Natural is
        pragma Unreferenced( this );
    begin
        -- looking up or down
        if k.shooting then
            return Get_Frame_Id( k.lib, Standing_and_Shooting, k.age - k.stateStart, To_Y( k.dir ) );
        else
            return Get_Frame_Id( k.lib, Looking_Frames, k.age - k.stateStart, To_Y( k.dir ), reversed => True );
        end if;
    end Get_Frame;

    ----------------------------------------------------------------------------

    function Get_Frame( this : Teleporting_State; k : Keen'Class ) return Natural is
        pragma Unreferenced( this );
    begin
        -- walking into a teleporter
        return Get_Frame_Id( k.lib, Entering_Doorway, k.age - k.stateStart );
    end Get_Frame;

    ----------------------------------------------------------------------------

    function Get_Frame( this : Idle_State; k : Keen'Class ) return Natural is
        pragma Unreferenced( this );
    begin
        return Get_Frame_Id( k.lib, idle_animations(k.idleNumber).visibleAction, k.age - k.stateStart );
    end Get_Frame;

    ----------------------------------------------------------------------------

    function Get_Frame( this : Rising_State; k : Keen'Class ) return Natural is
        pragma Unreferenced( this );
    begin
        return Get_Frame_Id( k.lib, Rising_Bored, k.age - k.stateStart );
    end Get_Frame;

    ----------------------------------------------------------------------------

    function Get_Frame( this : Dying_State; k : Keen'Class ) return Natural is
        pragma Unreferenced( this );
    begin
        return Get_Frame_Id( k.lib, Dying_Frames, k.age - k.stateStart );
    end Get_Frame;

    ----------------------------------------------------------------------------

    function Get_Frame( this : Climbing_State; k : Keen'Class ) return Natural is
        pragma Unreferenced( this );
    begin
        if k.shooting then
            -- shooting in any of four directions
            return Get_Frame_Id( k.lib, Hanging_From_Pole_And_Shooting, k.age - k.stateStart, To_D8( k.dir ) );
        elsif k.yv < 0.0 then
            -- climbing up
            return Get_Frame_Id( k.lib, Climbing_Pole, k.age - k.stateStart, To_X( k.dir ) );
        elsif k.yv > 0.0 then
            -- sliding down
            return Get_Frame_Id( k.lib, Climbing_Pole, k.age - k.stateStart, D8_Down );
        else
            -- hanging still
            return Get_Frame_Id( k.lib, Hanging_From_Pole, k.age - k.stateStart, To_X( k.dir ) );
        end if;
    end Get_Frame;

    --==========================================================================

    procedure Update( this : Moving_State; k : A_Keen; time : Tick_Time ) is
        pragma Unreferenced( time );

        ------------------------------------------------------------------------

        -- starts Keen thrusting left or right if he isn't already thrusting in
        -- a direction, or busy shooting (walking, pogoing, or in-air)
        procedure Start_Thrusting( hDir : Cardinal_Direction ) is
            acc : Float;
        begin
            -- keen can begin thrusting if he isn't already thrusting in a
            -- direction and he isn't busy shooting.
            if (not k.moving(Left) and then not k.moving(Right)) and then
               not k.shooting
            then
                k.moving(hDir) := True;
                k.dir := k.dir + hDir;

                if k.onPogo then
                    -- keen is moving in the air on his pogo stick
                    acc := Get_Keen_Pref( "midair_pogo_acceleration" );
                elsif not k.grounded then
                    -- keen is moving in the air without his pogo stick
                    acc := Get_Keen_Pref( "midair_acceleration" );
                else
                    -- keen is walking along the ground
                    acc := Get_Keen_Pref( "walk_acceleration" );
                end if;

                Queue_Accelerate( k.id, hDir, Get_Keen_Pref( "walk_speed" ), acc );
            end if;
        end Start_Thrusting;

        ------------------------------------------------------------------------

        -- stops thrusting Keen left or right in the moving state, if he was
        -- thrusting in 'hDir' direction
        procedure Stop_Thrusting( hDir : Cardinal_Direction ) is
        begin
            -- stop moving if 'dir' is the direction keen was moving
            if k.moving(hDir) then
                k.moving(hDir) := False;
                Queue_Accelerate( k.id, hDir, 0.0, 0.0 );
            end if;
        end Stop_Thrusting;

        ------------------------------------------------------------------------

        -- starts Keen aiming up or down (on ground or in-air). Returns True if
        -- Keen's state changes to aiming.
        function Start_Aiming( vDir : Cardinal_Direction ) return Boolean is
        begin
            -- can't change aim while shooting
            if k.shooting then
                return False;
            end if;

            -- aims in the given direction
            k.dir := k.dir + vDir;

            -- change to the Looking state if he's standing on the ground
            if not k.jumping and then
               k.grounded and then
               not k.onPogo
            then
                -- if keen is standing still
                if not k.moving then
                    -- change to the looking up/down state
                    k.Change_State( LOOKING,
                                    frames(Looking_Frames).frameDelay *
                                        frames(Looking_Frames).dir(To_D8( vDir ))'Length );
                    return True;
                end if;
            end if;
            return False;
        end Start_Aiming;

        ------------------------------------------------------------------------

        -- stops Keen aiming up or down, if he was aiming (on ground or in-air)
        procedure Stop_Aiming( vDir : Cardinal_Direction ) is
        begin
            -- can't change aim while shooting
            if k.shooting then
                return;
            end if;

            if k.dir and vDir then
                k.dir := k.dir - vDir;
            end if;
        end Stop_Aiming;

        ------------------------------------------------------------------------

        -- Starts jump (thrusting upward), if possible. Upward acceleration
        -- begins at the start of the jump and is maintained until at least the
        -- minimum jump height is reached, or the maximum jump height is
        -- reached. When the jump begins, .jumping is set True for as long as
        -- the jump directive is given and .jumpApexMax >= 0 until the apex of
        -- the jump is reached. After the apex, Keen will be off the ground,
        -- which prevents Star_Jumping() from starting another jump in mid-air.
        -- Once he lands on the ground again, .jumping is still True if the
        -- directive is still being held, which prevents a double-jump unless
        -- he's on his pogo.
        procedure Start_Jumping is
        begin
            -- keen can only start jumping if he's standing on the ground and
            -- hasn't started a jump before now. The .jumpApexMax field is
            -- checked in addition to .pogo because this should only happen once
            -- per jump.
            if (not k.jumping or else (k.onPogo and k.jumpApexMax < 0.0)) and then
              k.grounded and then
               not k.shooting
            then
                k.jumping := True;
                if k.onPogo then
                    k.jumpApexMin := Float'Max( k.y - k.minPogoHeight, 0.0 );
                    k.jumpApexMax := Float'Max( k.y - k.maxPogoHeight, 0.0 );
                else
                    k.jumpApexMin := Float'Max( k.y - k.minJumpHeight, 0.0 );
                    k.jumpApexMax := Float'Max( k.y - k.maxJumpHeight, 0.0 );
                end if;
                Queue_Accelerate( k.id, Up,
                                  Get_Keen_Pref( "jump_speed" ),
                                  Get_Keen_Pref( "jump_acceleration" ) );
                if k.onPogo then
                    Queue_Play_Sound( "keen_pogo.wav" );
                else
                    Queue_Play_Sound( "keen_jump.wav" );
                end if;
            end if;
        end Start_Jumping;

        ------------------------------------------------------------------------

        -- Stops the upward thrust of a jump if keen is jumping and has
        -- surpassed the minimum jump height. this does not allow keen to stop
        -- a jump early and only do a little hop.
        procedure Stop_Jumping is
        begin
            -- check if he's allowed to stop the jump yet (he must jump higher
            -- than the minimum jump height unless he hit his head).
            if k.y <= k.jumpApexMin then
                k.jumpApexMin := -1.0;
                k.jumpApexMax := -1.0;
                Queue_Accelerate( k.id, Up, 0.0, 0.0 );
            end if;
        end Stop_Jumping;

        ------------------------------------------------------------------------

        function Try_To_Activate_Touching_Entities return Boolean is

            procedure Activate( e : A_Entity ) is
            begin
                e.Activate( activator => A_Entity(k) );
            end Activate;

        begin
            -- only activate entities if Keen is standing on the ground and not
            -- using his pogo or gun.
            if k.grounded and then
               not k.jumping and then
               not k.onPogo and then
               not k.shooting
            then
                k.Iterate_Touching( Activate'Access );
            end if;

            -- if the activation is successful, the entity being activated
            -- should call Teleport() etc. on Keen, changing his state.
            if k.state.action /= this.action then
                k.Set_Directive( Keen_Directives.UP, Inactive );
                return True;
            end if;
            return False;
        end Try_To_Activate_Touching_Entities;

        ------------------------------------------------------------------------

        function Try_To_Grab_Pole( vDir : Cardinal_Direction ) return Boolean is

            --------------------------------------------------------------------

            procedure Try_To_Grab( e : A_Entity ) is
            begin
                if not (e.all in Pole'Class) then
                    -- not a pole
                    return;
                end if;

                -- if keen is standing on the ground and trying to climb down but
                -- the pole doesn't extend into the ground, don't start climbing
                if not k.grounded or else
                   vDir = Up or else
                   k.y + Float(k.height + k.world.Tile_Width) / 2.0 <= e.y + Float(e.height) / 2.0
                then
                    -- keen must be no higher than the top of the pole and no
                    -- lower than the middle of his body at the bottom of the
                    -- pole.
                    --
                    -- we check one pixel higher than the pole top because keen
                    -- standing on top of the ground and is exactly 2 tiles
                    -- (32 px) tall. this causes his head to be one pixel higher
                    -- than the tile boundary where the pole ends when it sticks
                    -- 2 tiles up from the ground.
                    if k.y - Float(k.height) / 2.0 >= e.y - Float(e.height) / 2.0 - 1.0 and then
                       k.y < e.y + Float(e.height) / 2.0 - POLE_BOTTOM_BUFFER
                    then
                        -- the above check ensures Keen is not at the top of the
                        -- pole. we'll clear .poleTop if he starts climing.

                        -- keen must have the center of his body on the pole
                        if k.x >= e.x - Float(e.width) / 2.0 and then
                           k.x <= e.x + Float(e.width) / 2.0
                        then
                            -- stop moving and ignore gravity
                            k.moving(Up) := False;
                            k.moving(Down) := False;
                            k.moving(Left) := False;
                            k.moving(Right) := False;
                            Queue_Accelerate( k.id, Left, 0.0, 0.0 );
                            Queue_Move_Entity( k.id, e.Get_X - 1.0, k.Get_Y );
                            Queue_Set_Entity_Attribute( k.id, "physical", Create( False ) );

                            -- stop jumping if keen jumped onto the pole
                            if k.jumpApexMax >= 0.0 then
                                k.jumpApexMin := -1.0;
                                k.jumpApexMax := -1.0;
                                Queue_Accelerate( k.id, Up, 0.0, 0.0 );
                            end if;

                            k.poleId := e.Get_Id;
                            k.poleTop := False;    -- see comment above\
                            -- the .poleBottom flag starts a short buffer space
                            -- above the bottom of the pole because if its too
                            -- small, the player will move right past it if he's
                            -- climbing fast enough.
                            k.poleBottom := k.y >=
                                            e.y + (Float(e.height) / 2.0) -
                                            POLE_BOTTOM_BUFFER;

                            -- begin climbing action (Move_Start will begin moving)
                            k.Change_State( CLIMBING );
                        end if;
                    end if;
                end if;
            end Try_To_Grab;

            --------------------------------------------------------------------

        begin
            -- Keen can grab poles if:
            -- and he's not on his pogo stick or shooting his gun
            -- and he's either standing on the ground or falling downward
            if not k.onPogo and then
               not k.shooting and then
               (k.grounded or else k.yv > 0.0)
            then
                k.Iterate_Touching( Try_To_Grab'Access );
            end if;
            return (k.state.action = Act_Climbing);  -- grabbed a pole?
        end Try_To_Grab_Pole;

        ------------------------------------------------------------------------

        procedure Shoot_Gun is
        begin
            if k.shooting then
                return;
            end if;

            k.shooting := True;
            k.shootStart := k.age;
            k.Reset_Idle;

            k.onPogo := False;

            if k.grounded then
                Stop_Thrusting( Left );
                Stop_Thrusting( Right );
            end if;

            if As_Number( k.world.Get_Game_State.Get_Var( "ammo" ) ) > 0 then
                k.world.Get_Game_State.Set_Var( "ammo", As_Number( k.world.Get_Game_State.Get_Var( "ammo" ) ) - 1 );
                Queue_Play_Sound( "keen_shoot.wav" );
                if k.dir and Up then
                    Queue_Spawn_Entity( Stunner_Shot_Tag,
                                        k.x, k.y - Float(k.height / 2),
                                        yv => -SHOT_SPEED );
                elsif k.dir and Down then
                    Queue_Spawn_Entity( Stunner_Shot_Tag,
                                        k.x, k.y + Float(k.height / 2),
                                        yv => SHOT_SPEED );
                elsif k.dir and Left then
                    Queue_Spawn_Entity( Stunner_Shot_Tag,
                                        k.x - Float(k.width / 2), k.y,
                                        xv => -SHOT_SPEED );
                elsif k.dir and Right then
                    Queue_Spawn_Entity( Stunner_Shot_Tag,
                                        k.x + Float(k.width / 2), k.y,
                                        xv => SHOT_SPEED );
                end if;
            else
                Queue_Play_Sound( "click.wav" );
            end if;
        end Shoot_Gun;

        ------------------------------------------------------------------------

    begin
        if k.Directive( Keen_Directives.LEFT ) then
            Start_Thrusting( Left );
        else
            Stop_Thrusting( Left );
        end if;
        if k.Directive( Keen_Directives.RIGHT ) then
            Start_Thrusting( Right );
        else
            Stop_Thrusting( Right );
        end if;

        if k.Directive( Keen_Directives.UP ) then
            -- the UP directive activates entities
            -- or begins climbing a pole
            -- or aims upward
            if not Try_To_Activate_Touching_Entities then
                 if Try_To_Grab_Pole( Up ) then
                     return;
                 elsif Start_Aiming( Up ) then
                     return;
                 end if;
            end if;
        else
            Stop_Aiming( Up );
        end if;
        if k.Directive( Keen_Directives.DOWN ) then
            if Try_To_Grab_Pole( Up ) then
                return;
            elsif Start_Aiming( Down ) then
                return;
            end if;
        else
            Stop_Aiming( Down );
        end if;

        if k.Directive( Keen_Directives.POGO ) then
            -- toggle the pogo (this directive is not ongoing)
            k.onPogo := not k.onPogo;
        end if;

        if k.Directive( Keen_Directives.JUMP ) then
            Start_Jumping;
        else
            -- if the directive is not active anymore, reset the .jumping flag
            -- when keen is safely back on the ground. don't allow it to be
            -- reset in mid-air or he'll be able to do a double jump as soon as
            -- he touches the ground again. the check for .jumpApexMax < 0
            -- handles the very brief case where keen is accelerating upward but
            -- has not left the ground yet (just .grounded may be insufficient).
            if k.grounded and then k.jumpApexMax < 0.0 then
                k.jumping := False;
            end if;
            -- try to stop the jump; this will succeed if he's jumping but only
            -- after he reaches the minimum apex height.
            Stop_Jumping;
        end if;

        if k.Directive( Keen_Directives.SHOOT ) then
            Shoot_Gun;
        end if;

        -- these are given by Ked to change Keen's orientation
        if k.Directive( Keen_Directives.FACE_LEFT ) then
            k.dir := Dir_Left;
        elsif k.Directive( Keen_Directives.FACE_RIGHT ) then
            k.dir := Dir_Right;
        end if;

        -- - - - - - -

        if k.onPogo and then k.grounded then
            -- keen is on his pogo stick; he needs to auto-bounce
            Start_Jumping;
        end if;

        -- if keen is jumping up, stop the when when he reaches maximum height
        if k.y <= k.jumpApexMax then
            Stop_Jumping;
        end if;

        -- todo: could this be simplified to? Reset_Idle() should make it unnecessary
        if k.grounded and then
            not k.jumping and then
            not k.onPogo and then
            not k.moving and then
            not k.editingMode
        then
            -- keen is standing still on the ground
            -- check if its time to start an idle animation
            if k.age >= k.idleStart + idle_animations(k.idleNumber).idleDelay then
                k.Change_State( IDLE, idle_animations(k.idleNumber).anmDuration );
                return;
            end if;
        else
            k.Reset_Idle;
        end if;
    end Update; -- Moving_State

    ----------------------------------------------------------------------------

    procedure Update( this : Looking_State; k : A_Keen; time : Tick_Time ) is
        pragma Unreferenced( this );

        ------------------------------------------------------------------------

        -- 'vDir' is the vertical direction in which to stop looking
        procedure Stop_Looking( vDir : Cardinal_Direction ) is
        begin
            -- can't change aim while shooting
            if k.shooting then
                return;
            end if;

            -- keen is no longer looking in the direction that he was before
            -- begin the action of looking back to center. use one frame fewer
            -- in the animation because the last frame of the looking animation
            -- is the steady looking frame.
            k.Change_State( LOOKING_BACK,
                            frames(Looking_Frames).frameDelay *
                            (frames(Looking_Frames).dir(To_D8( vDir ))'Length - 1) );
        end Stop_Looking;

        ------------------------------------------------------------------------

        -- shoots while standing on the ground, looking up
        procedure Shoot_Up is
        begin
            if not k.shooting then
                if k.dir and Up then
                    k.shooting := True;
                    k.shootStart := k.age;
                    k.Reset_Idle;

                    if As_Number( k.world.Get_Game_State.Get_Var( "ammo" ) ) > 0 then
                        k.world.Get_Game_State.Set_Var( "ammo", As_Number( k.world.Get_Game_State.Get_Var( "ammo" ) ) - 1 );
                        Queue_Play_Sound( "keen_shoot.wav" );
                        Queue_Spawn_Entity( Stunner_Shot_Tag,
                                            k.x, k.y - Float(k.height / 2),
                                            yv => -SHOT_SPEED );
                    else
                        Queue_Play_Sound( "click.wav" );
                    end if;
                end if;
            end if;
        end Shoot_Up;

        ------------------------------------------------------------------------

    begin
        -- check if it's time to stop looking and go back to the moving state:
        -- 1) keen wants to start walking left or right OR
        -- 2) the looking directive is no longer active
        if k.Directive( Keen_Directives.LEFT ) or else
           k.Directive( Keen_Directives.RIGHT ) or else
           (not k.Directive( Keen_Directives.UP ) and then (k.dir and Up)) or else
           (not k.Directive( Keen_Directives.DOWN ) and then (k.dir and Down))
        then
            -- time to stop looking in the direction he was looking
            if k.dir and Up then
                Stop_Looking( Up );
            else
                Stop_Looking( Down );
            end if;
            return;
        end if;

        -- Keen can shoot up while he is standing looking up
        if k.Directive( Keen_Directives.SHOOT ) then
            Shoot_Up;
        end if;

        if k.Directive( Keen_Directives.JUMP ) then
            -- check if keen is allowed to start a jump
             if k.dir and Up then
                 -- if keen is looking up then switch to the moving action so he
                 -- can jump on the next Update. he can't jump if he is looking
                 -- down.
                 k.Change_State( MOVING );
                 return;
             end if;
        end if;

        -- - - - - - -

        -- if keen is steadily looking up or down, scroll the view. the
        -- is_busy() check makes sure the looking up/down animation finishes
        -- before the view starts to scroll.
        if not k.shooting and then not k.Is_Busy then
            Queue_Scroll_View( 0.0, Float(To_Duration( time.elapsed )) * (Float(k.dir.y) * k.lookSpeed) );
        end if;
        k.Reset_Idle;
    end Update; -- Looking_State

    ----------------------------------------------------------------------------

    procedure Update( this : Looking_Back_State; k : A_Keen; time : Tick_Time ) is
        pragma Unreferenced( this, time );
    begin
        -- if keen is done looking back then go back to moving
        if not k.Is_Busy then
            k.Change_State( MOVING );
            k.dir.y := 0;      -- clear the vertical direction component
        end if;
        k.Reset_Idle;
    end Update; -- Looking_Back_State

    ----------------------------------------------------------------------------

    procedure Update( this : Teleporting_State; k : A_Keen; time : Tick_Time ) is
        pragma Unreferenced( this, time );
    begin
        if k.Directive( Keen_Directives.FINISH_TELEPORT ) then
            -- recenters the camera on keen. we do this here because the event
            -- that was dispatched before this was a Move_Entity event that
            -- queued a Entity_Moved event. We want to follow that Entity_Moved
            -- event with a Follow_Event to refocus keen so he remains in the
            -- center of the camera, not near the edge.
            Queue_Follow_Entity( k.id );
            k.Change_State( MOVING );
            return;
        end if;

        -- check if he's done the teleporting animation yet
        if not k.Is_Busy then
            -- stop moving left/right toward the doorway
            -- jump to the destination
            --   this performs clipping and may queue a grounded event
            -- end the teleporting action after the jump has been made
            Queue_Accelerate( k.id, Left, 0.0, 0.0 );
            Queue_Move_Entity( k.id, k.teleportX, k.teleportY );
            Queue_Entity_Directive( k.id, Keen_Directives.FINISH_TELEPORT, Once );
        end if;
        k.Reset_Idle;
    end Update; -- Teleporting_State

    ----------------------------------------------------------------------------

    procedure Update( this : Idle_State; k : A_Keen; time : Tick_Time ) is
        pragma Unreferenced( this, time );
    begin
        -- basically any input will interrupt an idle animation
        if k.Directive( Keen_Directives.LEFT  ) or else
           k.Directive( Keen_Directives.RIGHT ) or else
           k.Directive( Keen_Directives.UP    ) or else
           k.Directive( Keen_Directives.DOWN  ) or else
           k.Directive( Keen_Directives.JUMP  ) or else
           k.Directive( Keen_Directives.POGO  )
        then
            -- if keen is not in the process of sitting down, he can be interrupted.
            -- if he's sitting and reading, he has to stand back up again first.
            -- otherwise, he can abort the idle animation immediately.
            if k.idleNumber = INDEX_READING_BORED then
                k.Change_State( RISING,
                               frames(Rising_Bored).frameDelay *
                               frames(Rising_Bored).frames'Length );
                k.Reset_Idle;
                return;
            elsif k.idleNumber /= INDEX_SITTING_BORED then
                k.Change_State( MOVING );
                k.Reset_Idle;
                return;
            end if;
        end if;

        -- - - - - - -

        -- stop being idle immediately if the floor disappears beneath keen
        if not k.grounded then
            k.Change_State( MOVING );
            k.Reset_Idle;
            return;

        -- check if the idle animation is done
        elsif not k.Is_Busy then
            -- set the number for the next idle animation
            if k.idleNumber = idle_animations'Last then
                k.Reset_Idle;
            else
                k.idleNumber := idle_animations'First + (k.idleNumber mod idle_animations'Last);
            end if;

            -- check if another idle animation should begin immediately
            if k.age >= k.idleStart + idle_animations(k.idleNumber).idleDelay then
                k.Change_State( IDLE, idle_animations(k.idleNumber).anmDuration );
            else
                k.Change_State( MOVING );
            end if;
            return;
        end if;
    end Update; -- Idle_State

    ----------------------------------------------------------------------------

    procedure Update( this : Rising_State; k : A_Keen; time : Tick_Time ) is
        pragma Unreferenced( this, time );
    begin
        -- if keen is done standing up then go back to moving
        if not k.Is_Busy then
            k.Change_State( MOVING );
        end if;
        k.Reset_Idle;
    end Update; -- Rising_State

    ----------------------------------------------------------------------------

    procedure Update( this : Dying_State; k : A_Keen; time : Tick_Time ) is
        pragma Unreferenced( this, time );
    begin
        -- cap the height of the death bounce
        if k.y <= k.jumpApexMin then
            k.jumping := False;
            k.jumpApexMin := -1.0;
            k.jumpApexMax := -1.0;
            Queue_Accelerate( k.id, Up, 0.0, 0.0 );
        end if;

        -- check for death complete
        if not k.dead and then
          (k.x > k.deathX + DEATH_DISTANCE or else
           k.y > k.deathY + DEATH_DISTANCE or else
           k.x >= Float(k.world.Get_Width - 1) or else
           k.y >= Float(k.world.Get_Height - 1))
        then
            Queue_Play_Sound( "keen_dead.wav" );

            k.deathX := k.x;
            k.deathY := k.y;
            k.dead := True;   -- temporary to keep sound from repeating
            Queue_Player_Died;
        end if;
        k.Reset_Idle;
    end Update; -- Dying_State

    ----------------------------------------------------------------------------

    procedure Update( this : Climbing_State; k : A_Keen; time : Tick_Time ) is
        pragma Unreferenced( this, time );

        ------------------------------------------------------------------------

        -- starts shimmying up or down the pole, if possible.
        procedure Start_Shimmying( vDir : Cardinal_Direction ) is
        begin
            -- can't change aim while shooting
            if k.shooting then
                return;
            end if;

            -- aim up or down
            k.dir := k.dir + vDir;

            -- begin shimmying up or down. he can begin moving if he isn't
            -- already moving in a direction and he's not climbing past the end
            -- of the pole.
             if (not k.moving(Up) and then not k.moving(Down)) and then
                (vDir = Down or else not k.poleTop)   -- can't climb off top
             then
                 -- start moving up or down
                 k.moving(vDir) := True;

                 if vDir = Up then
                     Queue_Accelerate( k.id, vDir,
                                       Get_Keen_Pref( "climb_speed" ),
                                       Get_Keen_Pref( "climb_acceleration" ) );
                 else
                     Queue_Accelerate( k.id, vDir,
                                       Get_Keen_Pref( "slide_speed" ),
                                       Get_Keen_Pref( "slide_acceleration" ) );
                 end if;
             end if;
        end Start_Shimmying;

        ------------------------------------------------------------------------

        -- stops shimmying up or down the pole. or, if shimmying progress was
        -- already impeded (reached the top), this just stops aiming up or down.
        procedure Stop_Shimmying( vDir : Cardinal_Direction ) is
        begin
            -- can't change aim while shooting
            if k.shooting then
                return;
            end if;

            if k.dir and vDir then
                -- stop looking up or down
                k.dir := k.dir - vDir;
                if k.moving(vDir) then
                    -- stop moving up or down
                    k.moving(vDir) := False;
                    Queue_Accelerate( k.id, vDir, 0.0, Physics.INSTANT_ACCELERATION );
                end if;
            end if;
        end Stop_Shimmying;

        ------------------------------------------------------------------------

        -- aims left or right while hanging onto the pole
        procedure Aim( hDir : Cardinal_Direction ) is
        begin
            -- keen can only look left/right on a pole
            if not k.shooting then
                k.dir := k.dir + hDir;
            end if;
        end Aim;

        ------------------------------------------------------------------------

        -- starts a jump from the pole
        function Start_Jumping return Boolean is
        begin
            -- no jumping while shooting or climbing down
            if not k.jumping and then
               not k.shooting and then
               not (k.dir and Down)
            then
                k.poleId := INVALID_ID;
                k.poleTop := False;
                k.poleBottom := False;

                Queue_Set_Entity_Attribute( k.id, "physical", Create( True ) );
                k.jumping := True;
                k.jumpApexMin := Float'Max( k.y - k.minJumpHeight, 0.0 );
                k.jumpApexMax := Float'Max( k.y - k.maxJumpHeight, 0.0 );
                Queue_Accelerate( k.id, Up,
                                  Get_Keen_Pref( "jump_speed" ),
                                  Get_Keen_Pref( "jump_acceleration" ) );
                Queue_Play_Sound( "keen_jump.wav" );
                k.Change_State( MOVING );
                return True;
            end if;
            return False;
        end Start_Jumping;

        ------------------------------------------------------------------------

        procedure Shoot_Gun is

            procedure Stop_Climbing( vDir : Cardinal_Direction ) is
            begin
                -- this is the same code as Stop_Shimmying() with the exception
                -- that the direction is not cleared. we want to stop the
                -- climbing movement but not clear the direction so that the
                -- correct frame can be drawn.
                if k.moving(vDir) then
                    k.moving(vDir) := False;
                    Queue_Accelerate( k.id, vDir, 0.0, Physics.INSTANT_ACCELERATION );
                end if;
            end Stop_Climbing;

        begin
            if k.shooting then
                return;
            end if;

            k.shooting := True;
            k.shootStart := k.age;
            k.Reset_Idle;

            Stop_Climbing( Up );
            Stop_Climbing( Down );

            if As_Number( k.world.Get_Game_State.Get_Var( "ammo" ) ) > 0 then
                k.world.Get_Game_State.Set_Var( "ammo", As_Number( k.world.Get_Game_State.Get_Var( "ammo" ) ) - 1 );
                Queue_Play_Sound( "keen_shoot.wav" );
                if k.dir and Up then
                    Queue_Spawn_Entity( Stunner_Shot_Tag,
                                        k.x, k.y - Float(k.height / 2),
                                        yv => -SHOT_SPEED );
                elsif k.dir and Down then
                    Queue_Spawn_Entity( Stunner_Shot_Tag,
                                        k.x, k.y + Float(k.height / 2),
                                        yv => SHOT_SPEED );
                elsif k.dir and Left then
                    Queue_Spawn_Entity( Stunner_Shot_Tag,
                                        k.x - Float(k.width / 2), k.y,
                                        xv => -SHOT_SPEED );
                elsif k.dir and Right then
                    Queue_Spawn_Entity( Stunner_Shot_Tag,
                                        k.x + Float(k.width / 2), k.y,
                                        xv => SHOT_SPEED );
                end if;
            else
                Queue_Play_Sound( "click.wav" );
            end if;
        end Shoot_Gun;

        ------------------------------------------------------------------------

        pole : A_Entity;
    begin
        if k.Directive( Keen_Directives.UP ) then
            Start_Shimmying( Up );
        else
            Stop_Shimmying( Up );
        end if;
        if k.Directive( Keen_Directives.DOWN ) then
            Start_Shimmying( Down );
        else
            Stop_Shimmying( Down );
        end if;

        if k.Directive( Keen_Directives.LEFT ) then
            Aim( Left );
        elsif k.Directive( Keen_Directives.RIGHT ) then
            Aim( Right );
        end if;

        if k.Directive( Keen_Directives.JUMP ) then
            if Start_Jumping then
                -- changed out of this state
                return;
            end if;
        else
            k.jumping := False;
        end if;

        if k.Directive( Keen_Directives.SHOOT ) then
            Shoot_Gun;
        end if;

        -- - - - - - -

        if k.moving(Up) then
            -- stop moving up if Keen moves past the end of the pole
            k.poleBottom := False;
            pole := k.world.Get_Entity( k.poleId );
            -- see the comment in Try_To_Climb for information about
            -- the area of contact that is checked here.
            if k.y - Float(k.height) / 2.0 < pole.y - Float(pole.height) / 2.0 - 1.0 then
                k.poleTop := True;
                Stop_Shimmying( Up );
            end if;

        elsif k.moving(Down) then
            -- detect if Keen reaches the bottom end of the pole
            k.poleTop := False;
            pole := k.world.Get_Entity( k.poleId );

            -- this test must match the test in the Try_To_Grab_Pole() procedure
            -- in the MOVING state, when climbing starts.
            -- Note: setting poleBottom doesn't really matter unless we want to
            -- prevent the player from climbing off the bottom of the pole. if
            -- that is the case, call Stop_Shimmying() when the player hits the
            -- bottom of the pole and comment out the falling off the pole logic
            -- that follows.
            k.poleBottom:= k.y >=
                           pole.y + Float(pole.height) / 2.0 - POLE_BOTTOM_BUFFER;

            if k.y > pole.y + Float(pole.height) / 2.0 - POLE_BOTTOM_BUFFER then
                -- fall off the bottom of the pole
                Stop_Shimmying( Down );
                k.moving(Down) := False;
                Queue_Set_Entity_Attribute( k.id, "physical", Create( True ) );
                k.Change_State( MOVING );
                return;
            end if;
        end if;
        k.Reset_Idle;
    end Update; -- Climbing_State

end Entities.Players.Keen.States;
