--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

--with Debugging;                         use Debugging;
with Entities.Players;
with Entities.Players.Keen.States;      use Entities.Players.Keen.States;
with Events.Audio;                      use Events.Audio;
with Events.Entities;                   use Events.Entities;
with Physics;
with Preferences;
with Values.Booleans;                   use Values.Booleans;

pragma Elaborate_All( Entities.Players );
pragma Elaborate_All( Preferences );

package body Entities.Players.Keen is

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Get_Keen_Pref( shortName : String ) return Float is
    begin
        return Preferences.Get_Pref( "development", "entity.player.keen." & shortName );
    end Get_Keen_Pref;
    pragma Inline_Always( Get_Keen_Pref );

    --==========================================================================

    function Create_Player return A_Entity is
        this : constant A_Entity := new Keen;
    begin
        this.Construct;
        return this;
    end Create_Player;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Keen ) is
    begin
        Player(this.all).Construct( 8, 32, Preferences.Get_Pref( "development", "entity.player.keen.library" ) );
        this.state := Get_State( Act_Moving );
        this.minJumpHeight := Get_Keen_Pref( "min_jump_height" );
        this.maxJumpHeight := Get_Keen_Pref( "max_jump_height" );
        this.minPogoHeight := Get_Keen_Pref( "min_pogo_height" );
        this.maxPogoHeight := Get_Keen_Pref( "max_pogo_height" );
        this.maxBounceHeight := Get_Keen_Pref( "max_bounce_height" );
        this.lookSpeed := Get_Keen_Pref( "look_speed" );
        this.Update_Frame( notify => False );
        this.Reset_Idle;
    end Construct;

    ----------------------------------------------------------------------------

    procedure Change_State( this   : not null access Keen'Class;
                            state  : not null A_State;
                            length : Time_Span := Time_Span_Zero ) is
    begin
        -- change keen's state, if 'state' is different. begin tracking how long
        -- he's been in this state (via .stateStart) and how long he must be
        -- in it (as .stateDuration).
        if this.state /= state then
            this.state := state;
            this.stateStart := this.age;
            this.stateDuration := length;
        end if;
    end Change_State;

    ----------------------------------------------------------------------------

    -- todo: "die" should be a directive sent to the entity that it can act on
    overriding
    procedure Die( this : access Keen ) is
    begin
        -- keen is temporarily invincible while teleporting
        if this.state.action = Act_Teleporting then
            return;
        end if;

        this.Change_State( Get_State( Act_Dying ) );
        this.deathX := this.x;
        this.deathY := this.y;
        this.jumping := True;
        this.jumpApexMin := Float'Max( this.y - this.maxBounceHeight, 0.0 );
        this.jumpApexMax := this.jumpApexMin;

        Queue_Follow_Entity( 0 );
        Queue_Set_Entity_Attribute( this.id, "clipped", Create( False ) );
        Queue_Accelerate( this.id, Up,
                          Get_Keen_Pref( "jump_speed" ),
                          Physics.INSTANT_ACCELERATION );
        Queue_Accelerate( this.id, Right,
                          Get_Keen_Pref( "walk_speed" ) / 2.0,
                          Physics.INSTANT_ACCELERATION );
        Queue_Play_Sound( "keen_die.wav" );
    end Die;

    ----------------------------------------------------------------------------

    function Is_Busy( this : not null access Keen'Class ) return Boolean is
    begin
        if this.stateDuration > Time_Span_Zero then
            return this.stateStart + this.stateDuration > this.age;
        end if;
        return False;
    end Is_Busy;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Keen is
        this : aliased Keen;
    begin
        this.Construct;
        Keen'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Keen ) is
    begin
        Player'Read( stream, Player(obj) );
        obj.jumping := Boolean'Input( stream );
        obj.jumpApexMin := Float'Input( stream );
        obj.jumpApexMax := Float'Input( stream );
        obj.onPogo := Boolean'Input( stream );
        obj.shooting := Boolean'Input( stream );
        obj.shootStart := Time_Span'Input( stream );
        obj.state := Get_State( Action_Type'Input( stream ) );
        obj.stateStart := Time_Span'Input( stream );
        obj.stateDuration := Time_Span'Input( stream );
        obj.idleStart := Time_Span'Input( stream );
        obj.idleNumber := Positive'Input( stream );
        obj.teleportX := Float'Input( stream );
        obj.teleportY := Float'Input( stream );
        obj.poleId := Entity_Id'Input( stream );
        obj.poleTop := Boolean'Input( stream );
        obj.poleBottom := Boolean'Input( stream );
        obj.deathX := Float'Input( stream );
        obj.deathY := Float'Input( stream );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Keen ) is
    begin
        Player'Write( stream, Player(obj) );
        Boolean'Output( stream, obj.jumping );
        Float'Output( stream, obj.jumpApexMin );
        Float'Output( stream, obj.jumpApexMax );
        Boolean'Output( stream, obj.onPogo );
        Boolean'Output( stream, obj.shooting );
        Time_Span'Output( stream, obj.shootStart );
        Action_Type'Output( stream, obj.state.action );
        Time_Span'Output( stream, obj.stateStart );
        Time_Span'Output( stream, obj.stateDuration );
        Time_Span'Output( stream, obj.idleStart );
        Positive'Output( stream, obj.idleNumber );
        Float'Output( stream, obj.teleportX );
        Float'Output( stream, obj.teleportY );
        Entity_Id'Output( stream, obj.poleId );
        Boolean'Output( stream, obj.poleTop );
        Boolean'Output( stream, obj.poleBottom );
        Float'Output( stream, obj.deathX );
        Float'Output( stream, obj.deathY );
    end Object_Write;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Hit_Wall( this         : access Keen;
                           dir          : Cardinal_Direction;
                           firstContact : Boolean ) is
        pragma Unreferenced( firstContact );
    begin
        if this.state.action = Act_Moving then
            if this.jumpApexMax >= 0.0 and then dir = Up then
                -- begin contents of Moving.Stop_Jumping
                this.jumpApexMin := -1.0;
                this.jumpApexMax := -1.0;
                Queue_Accelerate( this.id, Up, 0.0, 0.0 );
                -- end contents of Moving.Stop_Jumping
            end if;
        end if;
    end On_Hit_Wall;

    ----------------------------------------------------------------------------

    procedure Reset_Idle( this : not null access Keen'Class ) is
    begin
        this.idleStart := this.age;
        this.idleNumber := 1;
    end Reset_Idle;

    ----------------------------------------------------------------------------

    not overriding
    procedure Teleport( this         : access Keen;
                        fromX, fromY : Float;
                        toX, toY     : Float ) is
        pragma Unreferenced( fromY );
    begin
        this.dir := this.dir - Up;  -- have to look up to teleport; stop looking

        -- stop moving left or right and move toward the teleporter
        if this.moving(Left) then
            this.moving(Left) := False;
        elsif this.moving(Right) then
            this.moving(Right) := False;
        end if;
        if fromX < this.x then
            -- Keen moves 67% of the way to the center of the door
            Queue_Accelerate( this.id, Left,
                              ((this.x - fromX) * 0.67) /
                                  Float(To_Duration( TELEPORT_DURATION )),
                              Physics.INSTANT_ACCELERATION );
        elsif fromX > this.x then
            Queue_Accelerate( this.id, Right,
                              ((fromX - this.x) * 0.67) /
                                  Float(To_Duration( TELEPORT_DURATION )),
                              Physics.INSTANT_ACCELERATION );
        else
            Queue_Accelerate( this.id, Left, 0.0, Physics.INSTANT_ACCELERATION );
        end if;

        this.teleportX := toX;
        this.teleportY := toY;
        this.Change_State( Get_State( Act_Teleporting ), TELEPORT_DURATION );
    end Teleport;

    ----------------------------------------------------------------------------

    overriding
    procedure Update( this : access Keen; time : Tick_Time ) is
    begin
        -- update based on current state
        this.state.Update( A_Keen(this), time );

        if this.shooting and then this.age > this.shootStart + SHOOT_DURATION then
            this.shooting := False;
        end if;
    end Update;

    ----------------------------------------------------------------------------

    overriding
    procedure Update_Frame( this : access Keen; notify : Boolean := True ) is
    begin
        this.Set_Frame( this.state.Get_Frame( this.all ), notify );
    end Update_Frame;

begin

    Entities.Factory.Register_Class( "player.Platform", Create_Player'Access );

    Preferences.Set_Default( "development", "entity.player.keen.library", "keen" );

    Preferences.Set_Default( "development", "entity.player.keen.walk_speed",               92.0  );
    Preferences.Set_Default( "development", "entity.player.keen.walk_acceleration",        700.0 );
    Preferences.Set_Default( "development", "entity.player.keen.midair_acceleration",      500.0 );
    Preferences.Set_Default( "development", "entity.player.keen.midair_pogo_acceleration", 350.0 );

    Preferences.Set_Default( "development", "entity.player.keen.jump_speed",        192.0  );
    Preferences.Set_Default( "development", "entity.player.keen.jump_acceleration", 2400.0 );
    Preferences.Set_Default( "development", "entity.player.keen.falling_speed",     54.0   );

    Preferences.Set_Default( "development", "entity.player.keen.climb_speed",        48.0  );
    Preferences.Set_Default( "development", "entity.player.keen.climb_acceleration", 2400.0 );
    Preferences.Set_Default( "development", "entity.player.keen.slide_speed",        64.0 );
    Preferences.Set_Default( "development", "entity.player.keen.slide_acceleration", 2400.0 );

    Preferences.Set_Default( "development", "entity.player.keen.min_jump_height",   10.0 );
    Preferences.Set_Default( "development", "entity.player.keen.max_jump_height",   40.0 );
    Preferences.Set_Default( "development", "entity.player.keen.min_pogo_height",   16.0 );
    Preferences.Set_Default( "development", "entity.player.keen.max_pogo_height",   80.0 );
    Preferences.Set_Default( "development", "entity.player.keen.max_bounce_height", 26.0 );

    -- The scrolling speed in world coordinates per second when Keen looks up or down.
    Preferences.Set_Default( "development", "entity.player.keen.look_speed", 80.0 );

end Entities.Players.Keen;
