--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Tags;                          use Ada.Tags;
--with Debugging;                         use Debugging;
with Entities;
with Entities.Enemies;                  use Entities.Enemies;
with Events.Audio;                      use Events.Audio;
with Events.Entities;                   use Events.Entities;
with Physics;                           use Physics;
with Resources;                         use Resources;
with Tiles.Libraries.Manager;           use Tiles.Libraries.Manager;

pragma Elaborate_All( Entities );

package body Entities.Sprites.Keen4 is

    POINTS_LIFETIME : constant Time_Span := Milliseconds( 600 );

    blue_1up_frame        : Natural := 0;
    collect_gem_frame     : Natural := 0;
    collect_stunner_frame : Natural := 0;
    points_100_frame      : Natural := 0;
    points_200_frame      : Natural := 0;
    points_500_frame      : Natural := 0;
    points_1000_frame     : Natural := 0;
    points_2000_frame     : Natural := 0;
    points_5000_frame     : Natural := 0;
    yellow_1up_frame      : Natural := 0;

    -- Finds the sprite named 'filename' in the "sprites" tile library and
    -- stores its tile id in 'id'. If the id passed in is non-zero (because the
    -- id was previously found) then no work is done. Note that calling this is
    -- very expensive if the "sprites" library isn't already cached because the
    -- library will be loaded and unloaded each time.
    procedure Find_Sprite_Tile_Id( filename : String; id : in out Natural );

    function Create_Blue_1UP return A_Entity;
    function Create_Collect_Gem return A_Entity;
    function Create_Collect_Stunner return A_Entity;
    function Create_Points_100 return A_Entity;
    function Create_Points_200 return A_Entity;
    function Create_Points_500 return A_Entity;
    function Create_Points_1000 return A_Entity;
    function Create_Points_2000 return A_Entity;
    function Create_Points_5000 return A_Entity;
    function Create_Yellow_1UP return A_Entity;

    function Create_Collect_Item( tileId : Natural ) return A_Entity;

    ----------------------------------------------------------------------------

    collect_drop_frames : A_Tile_Id_Array := null;

    -- The access returned by this procedure should not be modified, it is for
    -- reference only! This function uses an unprotected library-level variable
    -- for caching the frames for Collect_Drop.
    function Frame_Array( this : not null access Collect_Drop'Class ) return A_Tile_Id_Array;

    ----------------------------------------------------------------------------

    flag_frames : A_Tile_Id_Array := null;

    -- The access returned by this procedure should not be modified, it is for
    -- reference only! This function uses an unprotected library-level variable
    -- for caching the frames for Flag.
    function Frame_Array( this : not null access Flag'Class ) return A_Tile_Id_Array;

    ----------------------------------------------------------------------------

    flag_up_frames : A_Tile_Id_Array := null;

    -- the difference between the life time and the fly time is the period of
    -- time the flag stops and lowers itself into the pot before disappearing,
    -- to be replaced with the stationary Flag sprite.
    FLAG_LIFE_TIME : constant := 1.05;     -- seconds
    FLAG_FLY_TIME  : constant := 0.75;     -- seconds
    FLAG_DISTANCE  : constant := 30.0;     -- pixels (from down and left)

    -- The access returned by this procedure should not be modified, it is for
    -- reference only! This function uses an unprotected library-level variable
    -- for caching the frames for Flag_Up.
    function Frame_Array( this : not null access Flag_Up'Class ) return A_Tile_Id_Array;

    ----------------------------------------------------------------------------

    stunner_blast_frames : A_Tile_Id_Array := null;

    -- The access returned by this procedure should not be modified, it is for
    -- reference only! This function uses an unprotected library-level variable
    -- for caching the frames for Stunner_Blast.
    function Frame_Array( this : not null access Stunner_Blast'Class ) return A_Tile_Id_Array;

    ----------------------------------------------------------------------------

    stunner_shot_frames : A_Tile_Id_Array := null;

    procedure Explode( this : not null access Stunner_Shot'Class );

    -- The access returned by this procedure should not be modified, it is for
    -- reference only! This function uses an unprotected library-level variable
    -- for caching the frames for Stunner_Shot.
    function Frame_Array( this : not null access Stunner_Shot'Class ) return A_Tile_Id_Array;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    procedure Find_Sprite_Tile_Id( filename : String; id : in out Natural ) is
        lib : A_Tile_Library;
    begin
        if id /= 0 then
            return;
        end if;

        lib := Load_Library( "sprites", bitmaps => False );
        if lib /= null then
            id := lib.Get_Id( filename );
            Unload_Library( lib );
        end if;
    end Find_Sprite_Tile_Id;

    ----------------------------------------------------------------------------

    function Create_Blue_1UP return A_Entity is
    begin
        Find_Sprite_Tile_Id( "1up-blue", blue_1up_frame );
        return Create_Collect_Item( blue_1up_frame );
    end Create_Blue_1UP;

    function Create_Collect_Gem return A_Entity is
    begin
        Find_Sprite_Tile_Id( "gemgrab", collect_gem_frame );
        return Create_Collect_Item( collect_gem_frame );
    end Create_Collect_Gem;

    function Create_Collect_Stunner return A_Entity is
    begin
        Find_Sprite_Tile_Id( "gungrab", collect_stunner_frame );
        return Create_Collect_Item( collect_stunner_frame );
    end Create_Collect_Stunner;

    function Create_Points_100 return A_Entity is
    begin
        Find_Sprite_Tile_Id( "100points", points_100_frame );
        return Create_Collect_Item( points_100_frame );
    end Create_Points_100;

    function Create_Points_200 return A_Entity is
    begin
        Find_Sprite_Tile_Id( "200points", points_200_frame );
        return Create_Collect_Item( points_200_frame );
    end Create_Points_200;

    function Create_Points_500 return A_Entity is
    begin
        Find_Sprite_Tile_Id( "500points", points_500_frame );
        return Create_Collect_Item( points_500_frame );
    end Create_Points_500;

    function Create_Points_1000 return A_Entity is
    begin
        Find_Sprite_Tile_Id( "1000points", points_1000_frame );
        return Create_Collect_Item( points_1000_frame );
    end Create_Points_1000;

    function Create_Points_2000 return A_Entity is
    begin
        Find_Sprite_Tile_Id( "2000points", points_2000_frame );
        return Create_Collect_Item( points_2000_frame );
    end Create_Points_2000;

    function Create_Points_5000 return A_Entity is
    begin
        Find_Sprite_Tile_Id( "5000points", points_5000_frame );
        return Create_Collect_Item( points_5000_frame );
    end Create_Points_5000;

    function Create_Yellow_1UP return A_Entity is
    begin
        Find_Sprite_Tile_Id( "1up-yellow", yellow_1up_frame );
        return Create_Collect_Item( yellow_1up_frame );
    end Create_Yellow_1UP;

    ----------------------------------------------------------------------------

    function Create_Collect_Item( tileId : Natural ) return A_Entity is
        this : A_Entity := new Collect_Item;
    begin
        Collect_Item(this.all).Construct( tileId );   -- may raise exception
        return this;
    exception
        when RESOURCE_ERROR =>
            Delete( this );
            raise;
    end Create_Collect_Item;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Collect_Item; tileId : Natural ) is
    begin
        Sprite(this.all).Construct( 0, 0, "sprites", (1 => tileId),
                                    lifeSpan => POINTS_LIFETIME );  -- may raise exception
        this.physical := False;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Collect_Item is
        this : aliased Collect_Item;
    begin
        this.Construct( 0 );   -- the frame array will be read, overriding 0
        Collect_Item'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Collect_Item ) is
    begin
        Sprite'Read( stream, Sprite(obj) );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Collect_Item ) is
    begin
        Sprite'Write( stream, Sprite(obj) );
    end Object_Write;

    --==========================================================================

    function Create_Collect_Drop return A_Entity is
        this : A_Entity := new Collect_Drop;
    begin
        this.Construct;
        return this;
    exception
        when RESOURCE_ERROR =>
            Delete( this );
            raise;
    end Create_Collect_Drop;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Collect_Drop ) is
    begin
        Sprite(this.all).Construct( 16, 16, "sprites", this.Frame_Array.all,
                                    lifeSpan => Milliseconds( 315 ) );
        this.physical := False;
    end Construct;

    ----------------------------------------------------------------------------

    function Frame_Array( this : not null access Collect_Drop'Class ) return A_Tile_Id_Array is
        pragma Unreferenced( this );
        lib : A_Tile_Library;
    begin
        if collect_drop_frames = null then
            lib := Load_Library( "sprites", bitmaps => False );
            if lib /= null then
                collect_drop_frames := new Tile_Id_Array'(lib.Get_Id( "drop5" ),
                                                          lib.Get_Id( "drop6" ),
                                                          lib.Get_Id( "drop7" ));
                Unload_Library( lib );
            else
                collect_drop_frames := new Tile_Id_Array'(0, 0, 0);
            end if;
        end if;
        return collect_drop_frames;
    end Frame_Array;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Collect_Drop is
        this : aliased Collect_Drop;
    begin
        this.Construct;
        Collect_Drop'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Collect_Drop ) is
    begin
        Sprite'Read( stream, Sprite(obj) );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Collect_Drop ) is
    begin
        Sprite'Write( stream, Sprite(obj) );
    end Object_Write;

    --==========================================================================

    function Create_Flag return A_Entity is
        this : A_Entity := new Flag;
    begin
        this.Construct;
        return this;
    exception
        when RESOURCE_ERROR =>
            Delete( this );
            raise;
    end Create_Flag;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Flag ) is
    begin
        Sprite(this.all).Construct( 16, 16, "sprites", this.Frame_Array.all,
                                    frameDelay => Milliseconds( 150 ) );
        this.physical := False;
        this.clipped := False;
    end Construct;

    ----------------------------------------------------------------------------

    function Frame_Array( this : not null access Flag'Class ) return A_Tile_Id_Array is
        pragma Unreferenced( this );
        lib : A_Tile_Library;
    begin
        if flag_frames = null then
            lib := Load_Library( "sprites", bitmaps => False );
            if lib /= null then
                flag_frames := new Tile_Id_Array'(lib.Get_Id( "flag1" ),
                                                  lib.Get_Id( "flag2" ),
                                                  lib.Get_Id( "flag3" ),
                                                  lib.Get_Id( "flag4" ));
                Unload_Library( lib );
            else
                flag_frames := new Tile_Id_Array'(1 => 0);
            end if;
        end if;
        return flag_frames;
    end Frame_Array;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Flag is
        this : aliased Flag;
    begin
        this.Construct;
        Flag'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Flag ) is
    begin
        Sprite'Read( stream, Sprite(obj) );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Flag ) is
    begin
        Sprite'Write( stream, Sprite(obj) );
    end Object_Write;

    --==========================================================================

    procedure Spawn_Raising_Flag( x, y : Float ) is
    begin
        Queue_Spawn_Entity( Flag_Up_Tag,
                            x - FLAG_DISTANCE, y + FLAG_DISTANCE,
                            xv =>  FLAG_DISTANCE / FLAG_FLY_TIME,
                            yv => -FLAG_DISTANCE / FLAG_FLY_TIME );
    end Spawn_Raising_Flag;

    ----------------------------------------------------------------------------

    function Create_Flag_Up return A_Entity is
        this : A_Entity := new Flag_Up;
    begin
        this.Construct;
        return this;
    exception
        when RESOURCE_ERROR =>
            Delete( this );
            raise;
    end Create_Flag_Up;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Flag_Up ) is
        lifetime : constant Time_Span := To_Time_Span( FLAG_LIFE_TIME );
    begin
        Sprite(this.all).Construct( 16, 16, "sprites", this.Frame_Array.all,
                                    lifeSpan => lifetime );
        this.physical := False;
        this.clipped := False;
    end Construct;

    ----------------------------------------------------------------------------

    function Frame_Array( this : not null access Flag_Up'Class ) return A_Tile_Id_Array is
        pragma Unreferenced( this );
        lib : A_Tile_Library;
    begin
        if flag_up_frames = null then
            lib := Load_Library( "sprites", bitmaps => False );
            if lib /= null then
                flag_up_frames := new Tile_Id_Array'(lib.Get_Id( "flag_raise1" ),
                                                     lib.Get_Id( "flag_raise2" ),
                                                     lib.Get_Id( "flag_raise3" ),
                                                     lib.Get_Id( "flag_raise4" ),
                                                     lib.Get_Id( "flag_raise5" ),
                                                     lib.Get_Id( "flag_raise6" ),
                                                     lib.Get_Id( "flag_raise7" ));
            else
                flag_up_frames := new Tile_Id_Array'(1..1 => 0);
            end if;
        end if;
        return flag_up_frames;
    end Frame_Array;

    ----------------------------------------------------------------------------

    overriding
    procedure Update( this : access Flag_Up; time : Tick_Time ) is
    begin
        Sprite(this.all).Update( time );

        if not this.playedSound then
            Queue_Play_Sound( "flag1.wav" );
            this.playedSound := True;
        end if;

        if this.age >= To_Time_Span( FLAG_FLY_TIME ) then
            Queue_Accelerate( this.id, Down, 0.0, Physics.INSTANT_ACCELERATION );
            Queue_Accelerate( this.id, Left, 0.0, Physics.INSTANT_ACCELERATION );
        end if;

        if this.age >= this.lifeSpan then
            Queue_Spawn_Entity( Flag_Tag, this.x, this.y );
            Queue_Play_Sound( "flag2.wav" );
        end if;
    end Update;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Flag_Up is
        this : aliased Flag_Up;
    begin
        this.Construct;
        Flag_Up'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Flag_Up ) is
    begin
        Sprite'Read( stream, Sprite(obj) );
        obj.playedSound := Boolean'Input( stream );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Flag_Up ) is
    begin
        Sprite'Write( stream, Sprite(obj) );
        Boolean'Output( stream, obj.playedSound );
    end Object_Write;

    --==========================================================================

    function Create_Stunner_Blast return A_Entity is
        this : A_Entity := new Stunner_Blast;
    begin
        this.Construct;
        return this;
    exception
        when RESOURCE_ERROR =>
            Delete( this );
            raise;
    end Create_Stunner_Blast;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Stunner_Blast ) is
    begin
        Sprite(this.all).Construct( 16, 16, "sprites", this.Frame_Array.all,
                                    lifeSpan => Milliseconds( 210 ) );
        this.physical := False;
    end Construct;

    ----------------------------------------------------------------------------

    function Frame_Array( this : not null access Stunner_Blast'Class ) return A_Tile_Id_Array is
        pragma Unreferenced( this );
        lib : A_Tile_Library;
    begin
        if stunner_blast_frames = null then
            lib := Load_Library( "sprites", bitmaps => False );
            stunner_blast_frames := new Tile_Id_Array'(1..2 => 0);
            if lib /= null then
                stunner_blast_frames(1) := lib.Get_Id( "phaze-blast1" );
                stunner_blast_frames(2) := lib.Get_Id( "phaze-blast2" );
                Unload_Library( lib );
            end if;
        end if;
        return stunner_blast_frames;
    end Frame_Array;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Stunner_Blast is
        this : aliased Stunner_Blast;
    begin
        this.Construct;
        Stunner_Blast'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Stunner_Blast ) is
    begin
        Sprite'Read( stream, Sprite(obj) );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Stunner_Blast ) is
    begin
        Sprite'Write( stream, Sprite(obj) );
    end Object_Write;

    --==========================================================================

    function Create_Stunner_Shot return A_Entity is
        this : A_Entity := new Stunner_Shot;
    begin
        this.Construct;
        return this;
    exception
        when RESOURCE_ERROR =>
            Delete( this );
            raise;
    end Create_Stunner_Shot;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Stunner_Shot ) is
    begin
        Sprite(this.all).Construct( 12, 12, "sprites", this.Frame_Array.all,
                                    frameDelay => Milliseconds( 115 ) );
        this.physical := False;
        this.clipped := True;
    end Construct;

    ----------------------------------------------------------------------------

    procedure Explode( this : not null access Stunner_Shot'Class ) is
    begin
        Queue_Delete_Entity( this.id );
        Queue_Spawn_Entity( External_Tag( Stunner_Blast'Tag ), this.x, this.y );
        Queue_Play_Sound( "shot_hit.wav" );
    end Explode;

    ----------------------------------------------------------------------------

    function Frame_Array( this : not null access Stunner_Shot'Class ) return A_Tile_Id_Array is
        pragma Unreferenced( this );
        lib : A_Tile_Library;
    begin
        if stunner_shot_frames = null then
            lib := Load_Library( "sprites", bitmaps => False );
            stunner_shot_frames := new Tile_Id_Array'(1..4 => 0);
            if lib /= null then
                stunner_shot_frames(1) := lib.Get_Id( "phaze1" );
                stunner_shot_frames(2) := lib.Get_Id( "phaze2" );
                stunner_shot_frames(3) := lib.Get_Id( "phaze3" );
                stunner_shot_frames(4) := lib.Get_Id( "phaze4" );
                Unload_Library( lib );
            end if;
        end if;
        return stunner_shot_frames;
    end Frame_Array;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Stunner_Shot is
        this : aliased Stunner_Shot;
    begin
        this.Construct;
        Stunner_Shot'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Stunner_Shot ) is
    begin
        Sprite'Read( stream, Sprite(obj) );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Stunner_Shot ) is
    begin
        Sprite'Write( stream, Sprite(obj) );
    end Object_Write;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Collide( this : access Stunner_Shot; e : not null A_Entity ) is
    begin
        if e.all in Enemy'Class then
            if A_Enemy(e).Is_Shootable then
                A_Enemy(e).Die;
                this.Explode;
            end if;
        end if;
    end On_Collide;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Hit_Wall( this         : access Stunner_Shot;
                           dir          : Cardinal_Direction;
                           firstContact : Boolean ) is
        pragma Unreferenced( dir );
    begin
        if firstContact then
            this.Explode;
        end if;
    end On_Hit_Wall;

begin

    Entities.Factory.Register_Class( Blue_1UP_Tag,        Create_Blue_1UP'Access );
    Entities.Factory.Register_Class( Collect_Drop_Tag,    Create_Collect_Drop'Access );
    Entities.Factory.Register_Class( Collect_Gem_Tag,     Create_Collect_Gem'Access );
    Entities.Factory.Register_Class( Collect_Stunner_Tag, Create_Collect_Stunner'Access );
    Entities.Factory.Register_Class( Flag_Tag,            Create_Flag'Access );
    Entities.Factory.Register_Class( Flag_Up_Tag,         Create_Flag_Up'Access );
    Entities.Factory.Register_Class( Points_100_Tag,      Create_Points_100'Access );
    Entities.Factory.Register_Class( Points_200_Tag,      Create_Points_200'Access );
    Entities.Factory.Register_Class( Points_500_Tag,      Create_Points_500'Access );
    Entities.Factory.Register_Class( Points_1000_Tag,     Create_Points_1000'Access );
    Entities.Factory.Register_Class( Points_2000_Tag,     Create_Points_2000'Access );
    Entities.Factory.Register_Class( Points_5000_Tag,     Create_Points_5000'Access );
    Entities.Factory.Register_Class( Stunner_Blast_Tag,   Create_Stunner_Blast'Access );
    Entities.Factory.Register_Class( Stunner_Shot_Tag,    Create_Stunner_Shot'Access );
    Entities.Factory.Register_Class( Yellow_1UP_Tag,      Create_Yellow_1UP'Access );

end Entities.Sprites.Keen4;
