--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Ada.Tags;                          use Ada.Tags;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Entities.Players;                  use Entities.Players;
with Entities.Players.Keen;             use Entities.Players.Keen;
with Entities.Sprites.Keen4;            use Entities.Sprites.Keen4;
with Entities.Triggers;
with Events.Audio;                      use Events.Audio;
with Events.Keen;                       use Events.Keen;
with Events.World;                      use Events.World;
with Resources;                         use Resources;
with Themes;                            use Themes;
with Values.Booleans;                   use Values.Booleans;
with Values.Nulls;                      use Values.Nulls;
with Values.Numbers;                    use Values.Numbers;
with Values.Strings;                    use Values.Strings;

pragma Elaborate_All( Entities.Triggers );

package body Entities.Triggers.Keen is

    ICON_FILENAME : constant String := "trigger_entity";

    ----------------------------------------------------------------------------

    -- Executes the 'EnterLevel' trigger function. Returns True if 'args' were
    -- acceptable.
    function Func_EnterLevel( this : not null access Keen_Trigger'Class;
                              args : Value_Array ) return Boolean;

    -- Executes the 'ExitLevel' trigger function. Returns True if 'args' were
    -- acceptable.
    function Func_ExitLevel( this : not null access Keen_Trigger'Class;
                             args : Value_Array ) return Boolean;

    -- Executes the 'Flag' trigger function. Returns True if 'args' were
    -- acceptable.
    function Func_Flag( this : not null access Keen_Trigger'Class;
                        args : Value_Array ) return Boolean;

    -- Executes the 'Message' trigger function. Returns True if 'args' were
    -- acceptable.
    function Func_Message( this : not null access Keen_Trigger'Class;
                           args : Value_Array ) return Boolean;

    -- Executes the 'SetTile' trigger function. Returns True if 'args' were
    -- acceptable.
    function Func_SetTile( this : not null access Keen_Trigger'Class;
                           args : Value_Array ) return Boolean;

    -- Executes the 'Sound' trigger function. Returns True if 'args' were
    -- acceptable.
    function Func_Sound( this : not null access Keen_Trigger'Class;
                         args : Value_Array ) return Boolean;

    -- Executes the 'Teleport' trigger function. Returns True if 'args' were
    -- acceptable.
    function Func_Teleport( this : not null access Keen_Trigger'Class;
                            args : Value_Array ) return Boolean;

    -- END PRIVATE DECLARATIONS --
    --==========================================================================

    function Create_Keen_Trigger return A_Entity is
        this : A_Entity := new Keen_Trigger;
    begin
        this.Construct;
        return this;
    exception
        when RESOURCE_ERROR =>
            Delete( this );
            raise;
    end Create_Keen_Trigger;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Keen_Trigger ) is
    begin
        Trigger(this.all).Construct( "icons", ICON_FILENAME );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    function Evaluate_Function( this      : access Keen_Trigger;
                                name      : String;
                                arguments : Value_Array ) return Value_Ptr is

        ------------------------------------------------------------------------

        -- Recursively return a string of the values in a Value_Array
        function To_String( values : Value_Array ) return String is
            str : Unbounded_String;
        begin
            for i in values'Range loop
                Append( str, Image( values(i) ) );
                if i < values'Last then
                    Append( str, ", " );
                end if;
            end loop;
            return To_String( str );
        end To_String;
        pragma Warnings( Off, To_String );

        ------------------------------------------------------------------------

        argsOk : Boolean;
    begin
        pragma Debug( Dbg( this.To_String & ".Evaluate_Function: " & name &
                           "( " & To_String( arguments ) & " )",
                           D_EVAL, INFO ) );

        if name = "EnterLevel" then
            argsOk := this.Func_EnterLevel( arguments );

        elsif name = "ExitLevel" then
            argsOk := this.Func_ExitLevel( arguments );

        elsif name = "Flag" then
            argsOk := this.Func_Flag( arguments );

        elsif name = "Message" then
            argsOk := this.Func_Message( arguments );

        elsif name = "SetTile" then
            argsOk := this.Func_SetTile( arguments );

        elsif name = "Sound" then
            argsOk := this.Func_Sound( arguments );

        elsif name = "Teleport" then
            argsOk := this.Func_Teleport( arguments );

        else
            return
                val : Value_Ptr := Trigger(this.all).Evaluate_Function( name, arguments )
            do
                if val = Values.Nul then
                    val := Create_Null;
                end if;
                if val = Create_Null then
                    null;
                    pragma Debug( Dbg( this.To_String &
                                  ": Unrecognized function " & name & "()",
                                  D_EVAL, WARNING ) );
                end if;
            end return;
        end if;

        if not argsOk then
            pragma Debug( Dbg( this.To_String &
                               ": Invalid arguments for " & name & "()",
                               D_EVAL, WARNING ) );
            return Create( False ).As_Value;
        else
            return Create( True ).As_Value;
        end if;
    end Evaluate_Function;

    ----------------------------------------------------------------------------

    -- EnterLevel(levelName : String)
    function Func_EnterLevel( this : not null access Keen_Trigger'Class;
                              args : Value_Array ) return Boolean is
    begin
        if args'Length /= 1 or else args(args'First).Get.Get_Type /= V_STRING then
            return False;
        end if;

        declare
            levelName : constant String := As_String( args(args'First) ).Get.To_String;
        begin
            if levelName'Length > 0 then
                Queue_Stop_Music;
                Queue_Play_Sound( "enter_level" );
                Queue_Load_World( levelName );
            end if;
            this.activatable := False;
        end;
        return True;
    end Func_EnterLevel;

    ----------------------------------------------------------------------------

    -- ExitLevel()
    function Func_ExitLevel( this : not null access Keen_Trigger'Class;
                             args : Value_Array ) return Boolean is
    begin
        if args'Length /= 0 then
            return False;
        end if;

        Queue_Stop_Music;
        Queue_Play_Sound( "exit_level" );
        Queue_Load_World( "" );
        this.activatable := False;
        return True;
    end Func_ExitLevel;

    ----------------------------------------------------------------------------

    -- Flag()
    function Func_Flag( this : not null access Keen_Trigger'Class;
                        args : Value_Array ) return Boolean is
    begin
        if args'Length /= 0 then
            return False;
        end if;

        Entities.Sprites.Keen4.Spawn_Raising_Flag( this.x, this.y );
        this.activatable := False;
        return True;
    end Func_Flag;

    ----------------------------------------------------------------------------

    -- Message(message : String)
    function Func_Message( this : not null access Keen_Trigger'Class;
                           args : Value_Array ) return Boolean is
        pragma Unreferenced( this );
        align : Align_Type := Align_Right;
    begin
        if args'Length < 1 or else args'Length > 3 then
            return False;
        end if;
        for i in args'Range loop
            if args(i).Get.Get_Type /= V_STRING then
                return False;
            end if;
        end loop;

        if args'Length = 3 then
            if As_String( args(args'First + 2) ) = "left" then
                align := Align_Left;
            elsif As_String( args(args'First + 2) ) = "center" then
                align := Align_Center;
            end if;
            Queue_Message( text => As_String( args(args'First) ).Get.To_String,
                           icon => As_String( args(args'First + 1) ).Get.To_String,
                           textAlign => align );
        elsif args'Length = 2 then
            Queue_Message( text => As_String( args(args'First) ).Get.To_String,
                           icon => As_String( args(args'First + 1) ).Get.To_String );
        else
            Queue_Message( text => As_String( args(args'First) ).Get.To_String );
        end if;
        return True;
    end Func_Message;

    ----------------------------------------------------------------------------

    -- SetTile(layer : Integer; id : Natural)
    function Func_SetTile( this : not null access Keen_Trigger'Class;
                           args : Value_Array ) return Boolean is
    begin
        if args'Length /= 2 or else
           args(args'First).Get.Get_Type /= V_NUMBER or else
           args(args'First+1).Get.Get_Type /= V_NUMBER
        then
            return False;
        end if;
        Queue_Set_Tile( As_Number( args(args'First) ).Get.To_Int, this.x, this.y,
                        As_Number( args(args'First+1) ).Get.To_Int );
        return True;
    end Func_SetTile;

    ----------------------------------------------------------------------------

    -- Sound(name : String)
    function Func_Sound( this : not null access Keen_Trigger'Class;
                         args : Value_Array ) return Boolean is
        pragma Unreferenced( this );
    begin
        if args'Length /= 1 then
            return False;
        end if;
        if args(args'First).Get.Get_Type /= V_STRING then
            return False;
        end if;

        Queue_Play_Sound( As_String( args(args'First) ).Get.To_String );
        return True;
    end Func_Sound;

    ----------------------------------------------------------------------------

    -- Teleport(destX : Integer; destY : Integer)
    function Func_Teleport( this : not null access Keen_Trigger'Class;
                            args : Value_Array ) return Boolean is
    begin
        if args'Length /= 2 or else
           args(args'First).Get.Get_Type /= V_NUMBER or else
           args(args'First+1).Get.Get_Type /= V_NUMBER
        then
            return False;
        end if;
        if this.activator.all in Players.Keen.Keen'Class then
                Players.Keen.Keen(this.activator.all).Teleport( this.x, this.y,
                                                                As_Number( args(args'First) ).Get.To_Float,
                                                                As_Number( args(args'First+1) ).Get.To_Float );
        end if;
        return True;
    end Func_Teleport;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Keen_Trigger is
        this : aliased Keen_Trigger;
    begin
        this.Construct;
        Keen_Trigger'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Keen_Trigger ) is
    begin
        Trigger'Read( stream, Trigger(obj) );
        obj.frame := obj.lib.Get_Id( ICON_FILENAME );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Keen_Trigger ) is
    begin
        Trigger'Write( stream, Trigger(obj) );
    end Object_Write;

begin

    Entities.Factory.Register_Class( External_Tag( Keen_Trigger'Tag ), Create_Keen_Trigger'Access );

end Entities.Triggers.Keen;
