--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Numerics.Elementary_Functions; use Ada.Numerics.Elementary_Functions;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Ada.Text_IO;                       use Ada.Text_IO;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Allegro.Bitmaps.Drawing;           use Allegro.Bitmaps.Drawing;
with Allegro.Bitmaps.Image_IO;          use Allegro.Bitmaps.Image_IO;
with Allegro.Bitmaps.Locking;           use Allegro.Bitmaps.Locking;
with Allegro.Color;                     use Allegro.Color;
with Allegro.Displays;                  use Allegro.Displays;
with Allegro.Drawing;                   use Allegro.Drawing;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;
with Interfaces;                        use Interfaces;
with Resources;                         use Resources;
with Resources.Images;                  use Resources.Images;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;

package body Tile_Operations is

    function Convert_To_BMP_24( filename : String ) return Boolean is
        ok     : Boolean := False;
        input  : A_Allegro_Bitmap;
        output : A_Allegro_Bitmap;
    begin
        Al_Set_New_Bitmap_Format( ALLEGRO_PIXEL_FORMAT_ANY_32_WITH_ALPHA );
        input := Al_Load_Bitmap( Normalize_Pathname( filename ) );
        if input /= null then
            Al_Convert_Mask_To_Alpha( input, Al_Map_RGB( 255, 0, 255 ) );

            -- since we're saving to a BMP, overlay the image on a magic pink
            -- background to preserve transparency in the 24 bit .BMP format.
            output := Al_Create_Bitmap( Al_Get_Bitmap_Width( input ), Al_Get_Bitmap_Height( input ) );
            if output = null then
                -- something went very wrong
                Al_Destroy_Bitmap( input );
                return False;
            end if;
            Al_Set_Target_Bitmap( output );
            Al_Clear_To_Color( Al_Map_RGBA( 255, 0, 255, 255 ) );
            Al_Draw_Bitmap( input, 0.0, 0.0, 0 );
            Al_Set_Target_Bitmap( null );

#if WINDOWS then
            -- Save the file with a lower-case filename. This is acceptable on a
            -- non case sensitive file system like Windows. This is done because
            -- convert is primarily useful for converting old 8-bit resources
            -- that were created in the DOS days and have upper-case filenames.
            if Al_Save_Bitmap( Remove_Extension( To_Lower( Normalize_Pathname( filename ) ) ) & ".bmp", output ) then
#else
            if Al_Save_Bitmap( Remove_Extension( Normalize_Pathname( filename ) ) & ".bmp", output ) then
#end if;
                ok := True;
            end if;

            Al_Destroy_Bitmap( output );
            Al_Destroy_Bitmap( input );
        end if;
        return ok;
    end Convert_To_BMP_24;

    ----------------------------------------------------------------------------

    function Convert_To_PNG( filename : String ) return Boolean is
        ok     : Boolean := False;
        input  : A_Allegro_Bitmap;
    begin
        Al_Set_New_Bitmap_Format( ALLEGRO_PIXEL_FORMAT_ANY_32_WITH_ALPHA );
        input := Al_Load_Bitmap( Normalize_Pathname( filename ) );
        if input /= null then
            Al_Convert_Mask_To_Alpha( input, Al_Map_RGB( 255, 0, 255 ) );

#if WINDOWS then
            -- Save the file with a lower-case filename. This is acceptable on a
            -- non case sensitive file system like Windows. This is done because
            -- convert is primarily useful for converting old 8-bit resources
            -- that were created in the DOS days and have upper-case filenames.
            if Al_Save_Bitmap( Remove_Extension( To_Lower( Normalize_Pathname( filename ) ) ) & ".png", input ) then
#else
            if Al_Save_Bitmap( Remove_Extension( Normalize_Pathname( filename ) ) & ".png", input ) then
#end if;
                ok := True;
            end if;

            Al_Destroy_Bitmap( input );
        end if;
        return ok;
    end Convert_To_PNG;

    ----------------------------------------------------------------------------

    procedure Cut( filename  : String;
                   width     : Positive;
                   spacing   : Natural := 0;
                   skipEmpty : Boolean := False;
                   prefix    : String := "";
                   format    : String := "";
                   indexPath : String := "";
                   startId   : Positive := 1;
                   minCols   : Positive := 1;
                   textOut   : access procedure( str : String ) ) is
        tileId  : Positive := startId;
        idxFile : File_Type;

        -- number of columns to write to the tile index
        columns  : Integer;

        -- number of digits in the largest tile id to write in the tile index;
        -- used for pretty file formatting
        idDigits : Positive;

        saveBMP : constant Boolean := format = "bmp" or else
                                      (format'Length = 0 and then
                                       To_Lower( Get_Extension( filename ) ) = "bmp");
        savePNG : constant Boolean := format = "png" or else
                                      (format'Length = 0 and then
                                       To_Lower( Get_Extension( filename ) ) = "png");

        ------------------------------------------------------------------------

        procedure Put_Line( str : String ) is
        begin
            if textOut /= null then
                textOut( "cut: " & str );
            end if;
        end Put_Line;

        ------------------------------------------------------------------------

        function Count_Digits( tileCount : Integer ) return Positive is
        begin
            if tileCount > 0 then
                return Integer(Float'Floor( Log( Float(tileCount), 10.0 ) )) + 1;
            end if;
            return 1;
        end Count_Digits;

        ------------------------------------------------------------------------

        function Is_Empty_Tile( bmp : not null A_Allegro_Bitmap ) return Boolean is
            r, g, b, a : Unsigned_8;
        begin
            Al_Lock_Bitmap( bmp, ALLEGRO_PIXEL_FORMAT_ANY_32_WITH_ALPHA, ALLEGRO_LOCK_READONLY );
            for y in 0..Al_Get_Bitmap_Height( bmp ) - 1 loop
                for x in 0..Al_Get_Bitmap_Width( bmp ) - 1 loop
                    Al_Unmap_RGBA( Al_Get_Pixel( bmp, x, y ), r, g, b, a );
                    if a /= 0 then
                        Al_Unlock_Bitmap( bmp );
                        return False;
                    end if;
                end loop;
            end loop;
            Al_Unlock_Bitmap( bmp );
            return True;
        end Is_Empty_Tile;

        ------------------------------------------------------------------------

        -- write 'bmp' as an image file. 'outName' may be a relative path but it
        -- must not include a file extension. the extension is automatically
        -- determined by 'format' or the extension of 'filename'.
        procedure Save_Tile( outName : String; bmp : A_Allegro_Bitmap ) is
            ok   : Boolean;
            pragma Warnings( Off, ok );
            ext  : Unbounded_String;
            temp : A_Allegro_Bitmap;
        begin
            if saveBMP then
                -- the format doesn't support an alpha channel so blit our
                -- alpha-channel image onto a magenta background so the color
                -- can be reinterpreted later as transparent.
                temp := Al_Create_Bitmap( Al_Get_Bitmap_Width( bmp ), Al_Get_Bitmap_Height( bmp ) );
                Al_Set_Target_Bitmap( temp );
                Al_Clear_To_Color( Al_Map_RGBA( 255, 0, 255, 255 ) );
                Al_Draw_Bitmap( bmp, 0.0, 0.0, 0 );
                Al_Set_Target_Bitmap( null );

                ext := To_Unbounded_String( "bmp" );
                ok := Al_Save_Bitmap( outName & ".bmp", temp );
                Al_Destroy_Bitmap( temp );
            elsif savePNG then
                ext := To_Unbounded_String( "png" );
                ok := Al_Save_Bitmap( outName & ".png", bmp );
            end if;
            if Is_Open( idxFile ) then
                Put_Line( idxFile, String'((idDigits - Count_Digits( tileId )) * ' ') &
                                   Image( tileId ) & ' ' &
                                   Get_Filename( outName & '.' & ext ) );
            end if;
        end Save_Tile;

        ------------------------------------------------------------------------

        function Tile_Number( row, col : Integer ) return String is
            result : Unbounded_String;
        begin
            result := To_Unbounded_String( Image( row ) );
            if row < 10 then
                result := "0" & result;
            end if;
            if col < 10 then
                result := result & "0";
            end if;
            result := result & Image( col );
            return To_String( result );
        end Tile_Number;

        ------------------------------------------------------------------------

        res    : A_Resource_File;
        input  : A_Allegro_Bitmap;
        output : A_Allegro_Bitmap;
    begin
        if not Is_Format_Supported( Get_Extension( filename ) ) then
            Put_Line( "Unsupported image format: " & Get_Extension( filename ) );
            return;
        end if;

        res := Load_Resource( Normalize_Pathname( filename ), "" );
        if res = null then
            Put_Line( "File not found: " & filename );
            return;
        end if;

        -- Load the raw file into memory
        input := Load_Image( res );
        Release( res );
        if input = null then
            Put_Line( "Error loading image from " & filename );
            return;
        end if;

        begin
            -- Create the tile index file for writing. This will overwrite an
            -- existing file.
            if indexPath'Length > 0 then
                Create( idxFile, Out_File, Normalize_Pathname( indexPath ) );
            end if;
        exception
            when others =>
                Put_Line( "Error creating " & Normalize_Pathname( indexPath ) );
        end;

        output := Al_Create_Bitmap( width, width );
        columns := Integer'Max( Div_Ceil( Al_Get_Bitmap_Width( input ), width + spacing ), minCols );
        idDigits := Count_Digits( (startId - 1) +
                                  (columns *  Div_Ceil( Al_Get_Bitmap_Height( input ),
                                                        width + spacing )) );

        declare
            row  : Integer := 1;
            col  : Integer := 1;
            x, y : Integer;
        begin
            y := 0;
            while y < Al_Get_Bitmap_Height( input ) loop
                x := 0;

                -- Note: while we loop on the height of the input image, the
                -- minCols argument allows the user to create a greater number
                -- of columns in the tile index file than there are columns in
                -- the input image (which would result in several empty columns
                -- on the right in the tile index).
                while col <= columns loop

                    -- Blit the cutout to the single tile bitmap
                    Al_Set_Target_Bitmap( output );
                    Al_Draw_Bitmap_Region( input,
                                           Float(x), Float(y),
                                           Float(width), Float(width),
                                           0.0, 0.0, 0 );

                    -- Save the tile to disk
                    --
                    -- if the current column is still within the input image and
                    -- the tile is not completely transparent, or we're not
                    -- skipping transparent, save the tile as a file
                    if x < Al_Get_Bitmap_Width( input ) and then
                       (not skipEmpty or else not Is_Empty_Tile( output ))
                    then
                        Save_Tile( Normalize_Pathname( prefix & Tile_Number( row, col ) ), output );
                    elsif Is_Open( idxFile ) then
                        -- otherwise just save a spot in the tile index
                        Put_Line( idxFile, String'((idDigits - 1) * ' ') & "0" );
                    end if;

                    -- Draw a row separator comment line
                    if ((tileId - startId) mod columns) = (columns - 1) then
                        Put_Line( idxFile, (80 * '-') );
                    end if;

                    tileId := tileId + 1;
                    col := col + 1;
                    x := x + width + spacing;
                end loop;
                col := 1;
                row := row + 1;
                y := y + width + spacing;
            end loop;
        end;

        Al_Set_Target_Bitmap( null );

        Al_Destroy_Bitmap( input );
        Al_Destroy_Bitmap( output );
        if Is_Open( idxFile ) then
            Close( idxFile );
        end if;
    end Cut;

end Tile_Operations;
