--
-- Copyright (c) 2012 Kevin Wellwood
-- All rights reserved.
--
-- This source code is distributed under the Modified BSD License. For terms and
-- conditions, see license.txt.
--

with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Text_IO;
with Applications;
with Debugging;                         use Debugging;
with GNAT.Directory_Operations.Iteration;
 use GNAT.Directory_Operations.Iteration;
with Preferences;                       use Preferences;
with Support.Paths;                     use Support.Paths;
with Worlds;                            use Worlds;

pragma Elaborate_All( Applications );
pragma Elaborate_All( Preferences );
pragma Elaborate_All( Support.Paths );

package body Applications.Shell.UpdateMap is

    function Create_Application return A_Application is
        this : constant A_Shell_Application := new UpdateMap_Application;
    begin
        this.Construct( company       => "Kevin Wellwood",
                        name          => "updatemap",
                        configDir     => "",
                        enableAllegro => False );
        return A_Application(this);
    end Create_Application;

    ----------------------------------------------------------------------------

    overriding
    function Run( this : access UpdateMap_Application ) return Integer is

        returnCode : Integer := 0;
        found      : Natural := 0;
        updated    : Natural := 0;

        ------------------------------------------------------------------------

        procedure Display_Title is
        begin
            Ada.Text_IO.New_Line;
            Ada.Text_IO.Put_Line( this.Get_Announce );
            Ada.Text_IO.New_Line;
        end Display_Title;

        ------------------------------------------------------------------------

        procedure Display_Usage is
        begin
            Ada.Text_IO.Put_Line( "Usage:" );
            Ada.Text_IO.Put_Line( "  " & this.Get_Name & " <map_path> | all" );
            Ada.Text_IO.New_Line;
            returnCode := 1;
        end Display_Usage;

        ------------------------------------------------------------------------

        procedure Update_Map( path : String ) is
            world : A_World := null;
        begin
            found := found + 1;
            world := Load_World( path );
            world.Save( path );
            Ada.Text_IO.Put_Line( "Updated " & path );
            updated := updated + 1;
        exception
            when FILE_NOT_FOUND =>
                Delete( world );
                Ada.Text_IO.Put_Line( "Skipping " & path & " - File not found" );
            when READ_EXCEPTION =>
                -- this exception occurs if the world identifier is not found,
                -- the file version is not readable, or the format is otherwise
                -- damaged.
                Delete( world );
            when e : others =>
                Delete( world );
                Ada.Text_IO.Put_Line( "Skipping " & path & " - Error loading world" );
                pragma Debug( Dbg( e, level => ERROR ) );
        end Update_Map;

        ------------------------------------------------------------------------

        procedure Examine( path  : String;
                           index : Positive;
                           quit  : in out Boolean ) is
            pragma Unreferenced( index, quit );
        begin
            Update_Map( path );
        end Examine;

        ------------------------------------------------------------------------

        procedure Iterate is new Wildcard_Iterator( Examine );
    begin
        Display_Title;

        if Argument_Count < 1 then
            Display_Usage;
            return returnCode;
        end if;

        if To_Lower( Argument( 1 ) ) = "all" then
            Ada.Text_IO.Put_Line( "Searching ." & Slash & " ..." );
            Iterate( "*." & World_Extension );

            Ada.Text_IO.Put_Line( "Searching " & Get_Pref( "application", "media" ) & " ..." );
            Iterate( Get_Pref( "application", "media" ) & "*." & World_Extension );

            Ada.Text_IO.Put_Line( "Searching " & Get_Pref( "application", "media" ) & "worlds" & Slash & " ..." );
            Iterate( Get_Pref( "application", "media" ) & "worlds" & Slash & "*." & World_Extension );

            if found = 0 then
                Ada.Text_IO. New_Line;
                Ada.Text_IO.Put_Line( "No files found." );
            elsif updated = 0 then
                Ada.Text_IO. New_Line;
                Ada.Text_IO.Put_Line( "No files need updating." );
            end if;
        else
            Update_Map( Add_Extension( Argument( 1 ), World_Extension ) );
        end if;

        return returnCode;
    end Run;

    ----------------------------------------------------------------------------

begin

    Register_Allocator( Create_Application'Access );

    Preferences.Set_Default( "application", "media", "media" & Slash );
    Preferences.Set_Default( "application", "editing_mode", True );

end Applications.Shell.UpdateMap;
